/**
 * Title:        MLReplace
 * Description:  A multi-line replace task for ant
 * Copyright:    Copyright (c) 2000 Andreas Heilwagen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * @author Andreas Heilwagen
 *
 * $Id: MLReplace.java,v 1.1 2001/10/28 12:01:47 aheilwag Exp $
 * 
 * $Log: MLReplace.java,v $
 * Revision 1.1  2001/10/28 12:01:47  aheilwag
 * moved build files from /tools/ant/etc to /build
 *
 * Revision 1.2  2000/11/10 20:04:26  aheilwag
 * added cvs tags
 *
 */


import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import java.io.*;
import java.util.*;


/**
 * Replaces all the occurrences of the given string token with the given
 * string value of the indicated files. The substring @newline@ is beeing
 * replaced by a newline.
 */

public class MLReplace extends MatchingTask {
    
    private File src = null;
    private String token = null;
    private String value = "";

    private File dir = null;
    
    /**
     * Do the execution.
     */
    public void execute() throws BuildException {
        
        if (token == null) {
            throw new BuildException("replace token must not be null");
        }

        if (src == null && dir == null) {
            throw new BuildException("Either the file or the dir attribute must be specified");
        }
        
        log("Replacing " + token + " --> " + value);

        if (src != null) {
            processFile(src);
        }
        
        if (dir != null) {
	    DirectoryScanner ds = super.getDirectoryScanner(dir);
            String[] srcs = ds.getIncludedFiles();

            for(int i=0; i<srcs.length; i++) {
                File file = new File(dir,srcs[i]); 
                processFile(file);
            }
        }
    }

    /**
     * Perform the replacement on the given file.
     *
     * The replacement is performed on a temporary file which then replaces the original file.
     *
     * @param src the source file
     */
    private void processFile(File src) throws BuildException {
        File temp = new File(src.getPath() + ".temp");

        if (temp.exists()) {
            throw new BuildException("MLReplace: temporary file " + temp.getPath() + " already exists");
        }

        try {
            BufferedReader br = new BufferedReader(new FileReader(src));
            BufferedWriter bw = new BufferedWriter(new FileWriter(temp));

            String line;
            
            while (true) {
                line = br.readLine();
                if (line == null) break;
                if (line.length() != 0) bw.write(replace(line));
                bw.newLine();
            }
             
            bw.flush();
            bw.close();
            br.close();
            
            src.delete();
            temp.renameTo(src);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new BuildException(ioe);
        }       
    }


    /**
     * Set the source file.
     */
    public void setFile(String file) {
        this.src = project.resolveFile(file);
    }

    /**
     * Set the source files path when using matching tasks.
     */
    public void setDir(String dirName) {
        dir = project.resolveFile(dirName);
    }

    /**
     * Set the string token to replace.
     */
    public void setToken(String token) {
        this.token = token;
    }

    /**
     * Set the string value to use as token replacement.
     */
    public void setValue(String value)
    {
        int          pos;
        StringBuffer buf = new StringBuffer (value);

        while ((pos = buf.toString ().indexOf ("@newline@ ")) >= 0)
          buf.replace (pos, pos + 10, "\n");
        this.value = buf.toString ();
    }

    /**
     * Perform the token substitution.
     */    
    private String replace (String orig) {
        StringBuffer buffer = new StringBuffer();
        int start = 0, end = 0;
        
        while ((end = orig.indexOf(token, start)) > -1) {
            buffer.append(orig.substring(start, end));
            buffer.append(value);
            start = end + token.length();
        }
        
        buffer.append(orig.substring(start));
        
        return buffer.toString();            
    }
}