

package gov.va.med.hds.cd.ccow.internal;

import java.awt.Component;
import java.util.*;

import com.sentillion.sdkweb.webcontextor.*;

import gov.va.med.hds.cd.ccow.*;
import gov.va.med.hds.cd.ccow.ui.*;

public class CCOWBroker implements IClinicalContextBroker {

    private static final Map EMPTY_MAP =
        Collections.unmodifiableMap(new HashMap());

    private String appName;
    private String appPasscode;
    private IContextMapper mapper;
    private CCOWBrokerParticipant participant;
    private InternalCCOWContext currentContext;

    private Set observers = Collections.synchronizedSet(new HashSet());


    public CCOWBroker(String appName, String appPasscode, IContextMapper mapper) {
        this.appName = appName;
        this.appPasscode = appPasscode;
        this.mapper = mapper;
        try {
            participant = new CCOWBrokerParticipant();            
        } catch (WebContextorException wce) {
            participant = null;
        }
        recomputeContext(null);
    }

    void dispose() {
        observers.clear();
        breakExternalLink(false);
    }

    public synchronized boolean isConnected() {
        return (participant != null);
    }

    public synchronized boolean joinExternalContext(boolean useOurContext) {
        return joinExternalContext(useOurContext, null);
    }
    
    public synchronized boolean joinExternalContext(boolean useOurContext,
                                                    Component parent) {
        if (participant != null) return true;
        Map ourContext = null;
        if (currentContext != null)
            ourContext = currentContext.getContext();

        try {
            participant = new CCOWBrokerParticipant();
        } catch (WebContextorException wce) {
            participant = null;
            return false;
        }

        if (useOurContext) {
            if (changeContextImpl(ourContext, parent) == false)
                breakExternalLink(false);
        }

        recomputeContext(ourContext);

        if (participant != null)
            contextChangeNotify();

        return (participant != null);
    }

    public void breakExternalLink() {
        breakExternalLink(true);
    }
    protected void breakExternalLink(boolean notify) {
        if (participant != null)
            participant.dispose();
        participant = null;
        if (notify)
            contextChangeNotify();
    }

    public void addContextObserver(IContextObserver observer) {
        observers.add(observer);
    }

    /** Get the contents of the clinical context.
     *
     * @return an immutable Map containing all the items in the
     *     clinical context.  All of the keys in the map will be of
     *     type {@link ContextItemName}; all of the values in the map
     *     will be of type String.  */
    public Map getContextItems() {
        if (currentContext == null)
            recomputeContext(null);

        return currentContext.getContext();
    }

    private void recomputeContext(Map defaultContext) {
        Map context = defaultContext;

        if (participant != null) try {
            context = participant.getCurrentContext();
        } catch (WebContextorException wce) {
            participant = null;
        }

        if (context != null)
            currentContext = new InternalCCOWContext(context, mapper);
        else
            currentContext = new InternalCCOWContext(mapper);
    }

    /** Request permission to change the clinical context.
     *
     * Desktop components should call this method as soon as the user
     * indicates their desire to initiate a context change. For
     * example, when the user requests to see a patient selection
     * dialog, this method should be called <b>before</b> displaying
     * the dialog.
     *
     * @param contextItemNames A list of context items that the caller
     *       might possibly change if permission were granted
     * @return an opaque token indicating permission to change the
     *       context.  If the user has decided to cancel the change,
     *       returns null.
     */
    public Object requestChange(Set contextItemNames) {
        Set itemNames =
            InternalCCOWContext.canonicalizeContextItemNames(contextItemNames);
        itemNames = Collections.unmodifiableSet(itemNames);
        contextChangeRequestToken = null;

        // look through our list of context observers and ask each
        // ContextParticipant to prepare for the pending change.  If
        // any participant returns false, cancel the request.
        Iterator i = observers.iterator();
        while (i.hasNext()) {
            Object observer = i.next();
            if (observer instanceof IContextParticipant) {
                IContextParticipant contextParticipant =
                    (IContextParticipant) observer;
                boolean shouldContinue =
                    contextParticipant.prepareForContextChange(itemNames);
                if (!shouldContinue) return null;
            }
        }
        contextChangeRequestToken = new Date();
        return contextChangeRequestToken;
    }
    private Object contextChangeRequestToken = null;



    /** Allows a context participant to change the clinical context.
     *
     * @param requestToken the permission object returned earlier by
     *       the {@link #requestChange(Set) requestChange} operation
     * @param the new items to store in the context. If keys in this
     *       Map are not {@link ContextItemName} objects, their toString
     *       method will be called to obtain the full name of the
     *       context item.  If values in the Map are not String
     *       objects, their toString method will be called to obtain
     *       the value of the context item.
     * @return true if the context change was successful.
     */
    public boolean changeContext(Object requestToken, Map newContext) {
        return changeContext(requestToken, newContext, null);   
    }
    public boolean changeContext(Object requestToken, Map newContext, 
                                 Component parent) 
    {
        if (contextChangeRequestToken == null ||
            contextChangeRequestToken != requestToken) return false;

        return changeContextImpl(newContext, parent);
    }
    private boolean changeContextImpl(Map newContext, Component parent) {
        try {
            Map proposedContext =
                currentContext.proposeContextChange(newContext);
            String ourResponse = surveyParticipants(proposedContext, -1);

            ContextChangeProposalResponse response = null;

            if (participant != null)
                response = participant.proposeContextChange(newContext);
            else
                response = new ContextChangeProposalResponse(null, false);

            if (ourResponse != null && ourResponse.length() != 0) {
                if (participant != null)
                    ourResponse = participant.getAppName()+": " + ourResponse;
                response.addResponse(ourResponse);
            }

            int nextStep = CCOWContinuationDialog.show(parent,
                                                       response.surveyResponses,
                                                       response.noContinue,
                                                       false);

            if (nextStep == CCOWContinuationDialog.CANCEL) {
                if (participant != null)
                    participant.cancelContextChange();
                return false;
            }

            if (nextStep == CCOWContinuationDialog.CONTINUE) {
                if (participant != null)
                    participant.commitContextChange();
                else {
                    currentContext.replaceContext(proposedContext);
                    contextChangeNotify();
                }
                return true;
            }

            if (nextStep == CCOWContinuationDialog.BREAK_LINK) {
                if (participant != null)
                    participant.cancelContextChange();
                breakExternalLink(false);
                currentContext.replaceContext(proposedContext);
                contextChangeNotify();
                return true;
            }
        } catch (Exception e) { }

        return false;
    }



    /** Survey all internal {@link ContextParticipant}s to see if they
     * would have a problem with a context change.
     *
     * @param pendingContext (optional parameter) a handle to the
     *     pending context; you may leave this null if you supply the
     *     <code>pendingContextCoupon</code>, and the pending context
     *     will be looked up for you.
     * @param pendingContextCoupon (optional parameter) the coupon for
     *     the pending context; this will be ignored if a non null
     *     value was passed in for the <code>pendingContext</code>
     * @return <code>null</code> if no components have a problem;
     *     otherwise returns a string describing the reprecussions of
     *     a context change
     */
    protected String surveyParticipants(Map pendingContext,
                                        long pendingContextCoupon)
        throws WebContextorException
    {

        // look through our list of context observers and ask each
        // ContextParticipant about the pending change.  Collect
        // these responses into a Set (eliminating duplicates)
        Set responses = new HashSet();
        Iterator i = observers.iterator();
        while (i.hasNext()) {
            Object observer = i.next();
            if (observer instanceof IContextParticipant) {
                if (pendingContext == null)
                    pendingContext =
                        participant.getPendingContext(pendingContextCoupon);

                IContextParticipant p = (IContextParticipant) observer;
                String response = p.reviewContextChange(pendingContext);
                if (response != null)
                    responses.add(response);
            }
        }

        // If no ContextParticipants have any problems with
        // the context change, reply back with null to
        // indicate no problems with the context change.
        if (responses.isEmpty()) return null;

        // Append all the responses we found into a single
        // string describing the problems we would have with
        // the context change.
        StringBuffer buf = new StringBuffer();
        i = responses.iterator();
        while (i.hasNext())
            buf.append(i.next()).append("; ");
        buf.setLength(buf.length() - 2);   // strip final "; "

        // Send back our reply
        return buf.toString();
    }



    /** Notify all our {@link ContextObserver}s that the context has
     * changed.
     */
    protected void contextChangeNotify() {

        // look through our list of context observers and notify
        // each of the context change.
        Iterator i = observers.iterator();
        while (i.hasNext()) {
            IContextObserver observer = (IContextObserver) i.next();
            observer.contextChanged();
        }

        // TODO: loop through all of the observers and ensure that
        // they are displaying the new context.  Disable
        // components which misbehave. (This will require hooks
        // into the desktop to disable misbehaving components.)
    }


    private class CCOWBrokerParticipant extends CCOWParticipant {

        public CCOWBrokerParticipant() throws WebContextorException {
            super(appName, appPasscode);
        }

        /** Respond to an externally initiated context change request.
         */
        public String contextChangesPending(long pendingContextCoupon) {
            try {
                // call our superclass's contextChangesPending method.
                super.contextChangesPending(pendingContextCoupon);
                return surveyParticipants(null, pendingContextCoupon);
            } catch (Exception e) {
                return null;
            }
        }

        /** Respond to an externally committed context change.
         */
        public void contextChangesAccepted(long newContextCoupon) {
            // call our superclass's contextChangesAccepted method.
            super.contextChangesAccepted(newContextCoupon);

            // refresh our copy of the current context
            recomputeContext(currentContext.getContext());

            contextChangeNotify();
        }
    }

}
