

package gov.va.med.hds.cd.ccow.internal;

import java.util.*;

import gov.va.med.hds.cd.ccow.*;

import com.sentillion.sdkweb.webcontextor.*;
import com.sentillion.sdkweb.webdesktop.*;


/** This class initiates and maintains a connection to CCOW, and
 * manages the grunt-work of interacting with the Sentillion
 * WebContextor object.
 */
public class CCOWParticipant implements SimpleCCOWParticipant {
    
    private static final Set PARTICIPANT_INSTANCES = new HashSet();

    private static WebLocator webLocator = new WebLocator();

    private WebContextor webContextor;
    private String[] blob;
    private String appName;
    private String appPasscode;
    private String participantUrl;
    private String contextManagerUrl;

    private Map currentContext = null;
    private Map pendingContext = null;


    public CCOWParticipant(String appName, String appPasscode)
        throws WebContextorException
    {
        init(appName, appPasscode, 10);
    }

    public CCOWParticipant(String appName, String appPasscode, int timeout)
        throws WebContextorException
    {
        init(appName, appPasscode, timeout);
    }

    private void init(String appName, String appPasscode, int timeout)
        throws WebContextorException
    {
        // Create an empty contextor blob
        blob = sa();

        // using currentTimeMillis is icky - is there a better way
        // to do this?
        this.appName = appName + "#" + System.currentTimeMillis();
        
        this.appPasscode = appPasscode;

        // register with the participant server and get our URL
        participantUrl =
            CCOWParticipantServer.getInstance().registerParticipant(this);

        try {
            // locate the URL of the context manager
            contextManagerUrl = webLocator.locateContextManager
                (null, participantUrl, timeout, sa(), sa(), sa());
        } catch (CMRegistryException cmre) {
            throw new GeneralFailureException(cmre.getMessage());
        }

        // create a WebContextor object and start it.
        webContextor = new WebContextor();
        webContextor.run(blob,              // the "contextor blob" object
                         contextManagerUrl, // the context manager's url
                         participantUrl,    // the participant's url
                         this.appName,      // the name of the application
                         this.appPasscode,  // the passcode of the application
                         true);             // survey me
        
        // add this participant to the global list of instances.
        PARTICIPANT_INSTANCES.add(this);
    }

    public void dispose() {
        CCOWParticipantServer.getInstance().unregisterParticipant(this);
        try {
            webContextor.stop(blob);
            PARTICIPANT_INSTANCES.remove(this);
        } catch (Exception e) {}
    }
    
    public static void disposeAll() {
        Iterator i = PARTICIPANT_INSTANCES.iterator();
        while (i.hasNext()) try {
            CCOWParticipant p = (CCOWParticipant) i.next();
            p.dispose();
        } catch (Exception e) {}
    }

    public String getAppName() { return appName; }


    private String[] sa() { return new String[1]; }


    public synchronized String contextChangesPending
        (long pendingContextCoupon)
    {
        pendingContext = null;
        return null;
    }

    public synchronized void contextChangesAccepted(long newContextCoupon) {
        currentContext = pendingContext = null;
    }

    public synchronized void contextChangesCanceled(long canceledContextCoupon)
    {
        pendingContext = null;
    }

    public synchronized void commonContextTerminated() {
        pendingContext = null;
    }

    public void ping() {}


    /** translates a Sentillion ContextItemCollection object into a
     * context Map that our users can use
     */
    public static Map translateContext(ContextItemCollection ctxt) {
        Map result = new HashMap();
        if (ctxt == null) return result;

        int numItems = ctxt.count();
        for (int i = 1;   i <= numItems;   i++) try {
            ContextItem item = ctxt.item(i);
            IContextItemName name =
                ContextItemNameFactory.getName(item.getName());
            String value = item.getValue();
            result.put(name, value);
        } catch (UnknownItemException uie) {}

        return Collections.unmodifiableMap(result);
    }

    /** translates a context Map into a ContextItemCollection object
     * that the Sentillion classes can use
     */
    public static ContextItemCollection translateContext(Map ctxt) {
        if (ctxt == null) return null;

        ContextItemCollection result = new ContextItemCollection();

        Iterator i = ctxt.entrySet().iterator();
        Map.Entry e;
        while (i.hasNext()) try {
            e = (Map.Entry) i.next();
            Object key = e.getKey();
            String name = null;
            if (key instanceof IContextItemName)
                name = ((IContextItemName)key).getName();
            else if (key == null)
                continue;
            else
                name = key.toString();

            Object valueObj = e.getValue();
            String value = null;
            if (valueObj != null)
                value = valueObj.toString();

            ContextItem item = new ContextItem(name, value);
            result.add(item);
        } catch (Exception ex) {}

        return result;
    }

    /** Get the contents of the current context.
     */
    public synchronized Map getCurrentContext() throws WebContextorException {
        if (currentContext == null)
            currentContext = translateContext
                (webContextor.getCurrentContext(blob));

        return currentContext;
    }

    /** Get the contents of the pending context.
     *
     * This method should only be called during the processing of the
     * contextChangesPending method. At any other time, it will most
     * likely return null.
     */
    public synchronized Map getPendingContext(long coupon)
        throws WebContextorException
    {
        if (pendingContext == null)
            pendingContext = translateContext
                (webContextor.getPendingContext(blob, Long.toString(coupon)));

        return pendingContext;
    }

    public void stop() throws WebContextorException {
        webContextor.stop(blob);
        pendingContext = null;
    }

    public ContextChangeProposalResponse proposeContextChange(Map newContext)
        throws WebContextorException
    {
        boolean [] noContinue = new boolean[1];
        String [] surveyResponses;
        ContextItemCollection newContextItemCollection =
            translateContext(newContext);

        webContextor.startContextChange(blob);
        pendingContext = null;
        surveyResponses = webContextor.endContextChange
            (blob, true, newContextItemCollection, noContinue);

        ContextChangeProposalResponse result =
            new ContextChangeProposalResponse(surveyResponses, noContinue[0]);
        return result;
    }

    public void commitContextChange() throws WebContextorException {
        pendingContext = currentContext = null;
        String[] newCoupon = sa();
        String listenerURLs =
            webContextor.commitContextChange(blob, newCoupon);
        if (listenerURLs != null && listenerURLs.length() > 0)
            (new WebNotifier()).notifyListeners
                (listenerURLs, newCoupon[0], 2);
        contextChangesAccepted(Long.parseLong(newCoupon[0]));
    }

    public void cancelContextChange() throws WebContextorException {
        webContextor.cancelContextChange(blob);
        pendingContext = null;
    }

    public void suspend() throws WebContextorException {
        webContextor.suspend(blob);
        pendingContext = null;
    }

    public void resume() throws WebContextorException {
        webContextor.resume(blob);
        pendingContext = currentContext = null;
    }

    public int getState() throws WebContextorException {
        return webContextor.getState(blob);
    }

    public String getContextCoupon() throws WebContextorException {
        return webContextor.getContextCoupon(blob);
    }

    public String getParticipantCoupon() throws WebContextorException {
        return webContextor.getParticipantCoupon(blob);
    }

    public String getContextManagerUrl() {
        return contextManagerUrl;
    }

    public static final int CsNotRunning = WebContextor.CsNotRunning;
    public static final int CsParticipating = WebContextor.CsParticipating;
    public static final int CsSuspended = WebContextor.CsSuspended;
}
