
package gov.va.med.hds.cd.ccow.internal;

import java.util.Hashtable;
import java.net.Socket;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.net.URLDecoder;
import java.net.InetAddress;
import java.util.Iterator;
import java.net.URLEncoder;


/** This class provides a basic implementation of the "Context
 * Participant Interface", as described in the <i>Web/HTTP Component
 * Technology Mapping</i> document of the <i>HL7 Context Management
 * Standard</i>.
 *
 * In particular, this class listens on a port for incoming HTTP
 * requests.  These requests are interpreted as one of the following
 * messages:
 *
 * <ul>
 *    <li>ContextChangesPending
 *    <li>ContextChangesAccepted
 *    <li>ContextChangesCanceled
 *    <li>CommonContextTerminated
 *    <li>Ping
 * </ul>
 *
 * This class forwards the messages it receives to an object of type
 * {@link SimpleCCOWParticipant}, which is responsible for handling
 * the message.
 *
 * This class is also capable of listening on a single port and
 * receiving messages on behalf of any number of distinct
 * {@link SimpleCCOWParticipant}s;  incoming messages can be correctly
 * routed to the appropriate object.
 */
class CCOWParticipantServer extends TinyWebServer {


    /** A number used to generate unique IDs for each
     *  {@link SimpleCCOWParticipant} */
    private long uniqueID = System.currentTimeMillis();

    /** A collection of the {@link SimpleCCOWParticipant}s we are
     * representing. Values in the map are objects of type
     * {@link SimpleCCOWParticipant}, while keys are Strings
     * containing the unique ID of the {@link SimpleCCOWParticipant}. */
    protected Hashtable participants = new Hashtable();

    /** The base URL of this web server, used to construct URLs for
     * each participant. */
    private String baseURL;




    protected class CCOWParticipantServerThread extends TinyWebThread {

        public CCOWParticipantServerThread(Socket clientSocket) {
            super(clientSocket);
            setDaemon(true);
        }

        /** Accept an incoming message, dispatch it to the appropriate
         * {@link SimpleCCOWParticipant}, and send back an
         * appropriately formatted reply.
         *
         * These messages will be GET requests with a path of the form
         *
         * <pre>/<i>ID</i>?interface=ContextParticipant&method=<i>name</i>&<i>args...</i></pre>
         *
         * where:<ul>
         *
         * <li><code><i>ID</i></code> is the unique ID of a registered
         *     {@link SimpleCCOWParticipant}
         *
         * <li><code><i>name</i></code> is one of the following
         *     messages:<ul>
         *     <li><code>ContextChangesPending</code>
         *     <li><code>ContextChangesAccepted</code>
         *     <li><code>ContextChangesCanceled</code>
         *     <li><code>CommonContextTerminated</code>
         *     <li><code>Ping</code></ul>
         *
         * <li><code><i>args...</i></code> are additional URL-encoded
         *     arguments to the message.
         *
         * </ul>
         *
         * Of course, the URL-encoded arguments can appear in any order
         * and will be interpreted correctly.
         */
        public void serviceRequest()
            throws TinyWebThreadException, IOException
        {
            // ensure that the path begins with a '/' character.
            if (path == null || path.length() == 0 || path.charAt(0) != '/')
                badRequest();

            // find the beginning of the query arguments
            int queryPos = path.indexOf('?');
            if (queryPos < 2) badRequest();

            // extract the participant id from the path, and find the
            // appropriate participant.
            String id = path.substring(1, queryPos);
            SimpleCCOWParticipant p = getParticipant(id);
            if (p == null)
                sendError(404, "Not Found", "No such participant.");

            // Extract and parse the query parameters
            String query = path.substring(queryPos + 1);
            if (query.length() == 0) badRequest();
            Map params = parseQuery(query);
            if (params == null) badRequest();

            // Ensure that the request indicated the correct value of
            // the "interface" parameter
            String interfaceName = (String) params.get("interface");
            if (!"ContextParticipant".equalsIgnoreCase(interfaceName))
                badRequest();

            // Get the value of the "method" parameter and take the
            // action specified.
            String method = (String) params.get("method");
            if (method == null) badRequest();

            if (method.equalsIgnoreCase("ContextChangesPending"))
                doContextChangesPending(p, params);

            else if (method.equalsIgnoreCase("ContextChangesAccepted"))
                doContextChangesAccepted(p, params);

            else if (method.equalsIgnoreCase("ContextChangesCanceled"))
                doContextChangesCanceled(p, params);

            else if (method.equalsIgnoreCase("CommonContextTerminated"))
                doCommonContextTerminated(p, params);

            else if (method.equalsIgnoreCase("Ping"))
                doPing(p, params);

            out.flush();
        }

        /** Send a reply to the HTTP client indicating that the HTTP
         * request was invalid.
         */
        protected void badRequest()
            throws TinyWebThreadException, IOException
        {
            sendError( 400, "Bad Request", "No request found." );
        }

        /** Parse a string of URL-encoded query parameters.
         *
         * @param query the query string of a URL.  This is a string
         *   of name=value pairs, separated by the '&amp;' character.
         *   Each name and value must be URL-encoded. <b>Note:</b>
         *   this implementation does not support repeated parameters
         *   with the same name.
         *
         * @return a collection mapping names to values. Since the
         *   CCOW spec explicitly indicates that parameter names are
         *   case insensitive, the names will automatically be
         *   translated to lowercase for easier lookup.
         */
        protected Map parseQuery(String query) {
            // error checking
            if (query == null || query.length() == 0) return null;

            // strip initial '?' if it is present
            if (query.charAt(0) == '?')
                query = query.substring(1);

            Map parameters = new HashMap();
            String param, name, val;
            int equalsPos;
            // break apart the string on "&" characters.
            StringTokenizer params = new StringTokenizer(query, "&");
            while (params.hasMoreTokens()) {
                param = params.nextToken();
                equalsPos = param.indexOf('=');
                try {
                    if (equalsPos == 0 || param.length() == 0)
                        continue;
                    else if (equalsPos == -1)
                        // if no value is present, use the value "true"
                        parameters.put
                            (urlDecode(param).toLowerCase(), "true");
                    else {
                        // extract the name and value, and save them
                        // in the results map.
                        name = urlDecode(param.substring(0,equalsPos));
                        val = urlDecode(param.substring(equalsPos+1));
                        parameters.put(name.toLowerCase(), val);
                    }
                } catch (Exception e) {
                    System.err.println("Malformed query parameter: " + param);
                }
            }
            return parameters;
        }

        /** Send a reply appropriate to the CCOW spec */
        protected void sendResponse(String body) throws IOException {
            sendHeaders(200, "OK", DEFAULT_TEXT_MIME_TYPE, body.length(), -1);
            out.write(body);
            out.flush();
        }

        /** Convenience routine to retrieve the context coupon from the
         * parameter map */
        protected long getCoupon(Map params)
            throws TinyWebThreadException, IOException
        {
            try {
                String couponStr = (String) params.get("contextcoupon");
                return Long.parseLong(couponStr);
            } catch (Exception ioe) {
                badRequest();
            }
            // this line is never reached - it just keeps the compiler happy.
            return -1;
        }

        /** Handle a ContextChangesPending message */
        protected void doContextChangesPending(SimpleCCOWParticipant p,
                                               Map params)
            throws TinyWebThreadException, IOException
        {
            String response = p.contextChangesPending(getCoupon(params));
            if (response == null || response.length() == 0)
                sendResponse("decision=accept"+
                             "&reason=");
            else
                sendResponse("decision=conditional_accept"+
                             "&reason="+urlEncode(response));
        }

        /** Handle a ContextChangesAccepted message */
        protected void doContextChangesAccepted(SimpleCCOWParticipant p,
                                                Map params)
            throws TinyWebThreadException, IOException
        {
            p.contextChangesAccepted(getCoupon(params));
            sendResponse("");
        }

        /** Handle a ContextChangesCanceled message */
        protected void doContextChangesCanceled(SimpleCCOWParticipant p,
                                                Map params)
            throws TinyWebThreadException, IOException
        {
            p.contextChangesCanceled(getCoupon(params));
            sendResponse("");
        }

        /** Handle a CommonContextTerminated message */
        protected void doCommonContextTerminated(SimpleCCOWParticipant p,
                                                 Map params)
            throws TinyWebThreadException, IOException
        {
            p.commonContextTerminated();
            sendResponse("");
        }

        /** Handle a Ping message */
        protected void doPing(SimpleCCOWParticipant p, Map params)
            throws TinyWebThreadException, IOException
        {
            p.ping();
            sendResponse("");
        }
    }
    
    protected static String urlEncode(String s) { 
        try {
            return URLEncoder.encode(s, "UTF-8");
        } catch (Exception e) {
            return null; 
        }
    }
    protected static String urlDecode(String s) { 
        try {
            return URLDecoder.decode(s, "UTF-8");
        } catch (Exception e) {
            return null;
        }
    }

    protected TinyWebThread createThread(Socket clientSocket) {
        return new CCOWParticipantServerThread(clientSocket);
    }

    /** Create a CCOWParticipantServer listening on any available port.
     *
     * This method is private in accordance with the singleton pattern.
     */
    private CCOWParticipantServer() throws IOException {
        super(0, true);
        calcBaseURL();
        setDaemon(true);
        start();
    }

    private static CCOWParticipantServer INSTANCE = null;
    static {
        try {
            INSTANCE = new CCOWParticipantServer();
        } catch (IOException ioe) {
            ioe.printStackTrace();
            INSTANCE = null;
        }
    }


    /** Return the (singleton) instance of the CCOWParticipantServer.  */
    public static CCOWParticipantServer getInstance() {
        return INSTANCE;
    }


    /** Calculate the base URL of this web server, used to construct
     * URLs for each participant. */
    private void calcBaseURL() throws IOException {
        StringBuffer result = new StringBuffer();
        result.append("http://");
        result.append(InetAddress.getLocalHost().getHostAddress());
        result.append(":");
        result.append(serverSocket.getLocalPort());
        result.append("/");
        baseURL = result.toString();
    }


    /** Register a {@link SimpleCCOWParticipant} with this server.
     *
     * Once registered, this server will accept CCOW messages on
     * behalf of the {@link SimpleCCOWParticipant} and forward them to
     * that object.  Calling this method multiple times for the same
     * {@link SimpleCCOWParticipant} has no effect.
     *
     * @param p the participant to register
     * @return the "participant URL" assigned to this participant.
     */
    public String registerParticipant(SimpleCCOWParticipant p) {
        synchronized (participants) {
            String result = getParticipantURL(p);
            if (result != null)
                return result;

            String id = Long.toString(uniqueID++);
            participants.put(id, p);
            return buildParticipantURL(id);
        }
    }


    /** Unregister a {@link SimpleCCOWParticipant} with this server.
     *
     * The {@link SimpleCCOWParticipant} will no longer receive
     * CCOW messages.
     *
     * @param p the participant to unregister
     */
    public void unregisterParticipant(SimpleCCOWParticipant p) {
        synchronized (participants) {
            String participantID = getParticipantID(p);
            if (participantID != null)
                participants.remove(participantID);
        }
    }

    /** Retrieve the "participant URL" of a previously registered
     * {@link SimpleCCOWParticipant}. (This URL is required in various
     * places by CCOW.)
     *
     * @param p a participant
     * @return the URL assigned to the participant, or null if the
     * participant has never been registered.
     */
    public String getParticipantURL(SimpleCCOWParticipant p) {
        String participantID = getParticipantID(p);
        if (participantID == null)
            return null;
        else
            return buildParticipantURL(participantID);
    }


    /** Construct the "participant URL" for a participant with the
     * given id. */
    protected String buildParticipantURL(String id) {
        return baseURL + id;
    }


    /** Lookup the unique ID for a given participant
     */
    protected String getParticipantID(SimpleCCOWParticipant p) {
        synchronized (participants) {
            Iterator i = participants.entrySet().iterator();
            while (i.hasNext()) {
                Map.Entry e = (Map.Entry) i.next();
                if (e.getValue() == p)
                    return (String) e.getKey();
            }
            return null;
        }
    }


    /** Get the participant with the given ID.  */
    protected SimpleCCOWParticipant getParticipant(String id) {
        synchronized (participants) {
            return (SimpleCCOWParticipant) participants.get(id);
        }
    }

    /** @link dependency */
    /*#SimpleCCOWParticipant lnkSimpleCCOWParticipant;*/
}
