/*
 * Created on Apr 14, 2003
 *
 */
package gov.va.med.hds.cd.ccow.desktop;

import java.io.File;
import java.util.LinkedList;

import gov.va.med.hds.cd.ccow.IChartLauncher;
import gov.va.med.hds.cd.m.IMumpsFactory;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExecutableExtension;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;

/** A default/dumb implementation of a chart launcher.
 * 
 * This launcher looks in the settings file for an element of the form:
 * <xmp>
 *    <chartApplication>
 *      <port>6000</port> 
 *      <exePath>c:\program files\vista\cprs\cprschart.exe</exePath>
 *      <exePath>c:\path\my_cprs.exe</exePath>
 *    </chartApplication>
 * </xmp>
 * 
 * The &lt;port&gt; element is optional; if it is missing, the chart will be
 * launched with the same port as the M connection.  (If the M connection is 
 * using VistALink, and your chart application is CPRS, this would almost
 * certainly be the wrong port.) 
 * 
 * Any number of &lt;exePath&gt; elements can be listed.  If they name a
 * nonexistent file, they will be ignored.  Any files that are found will
 * be tried in the order they are listed.
 * 
 * @author VHAISLTUMAD
 */
public class DefaultChartLauncher
	implements IChartLauncher, IExecutableExtension {

	public static final String SETTINGS_ROOT_ELEMENT_NAME = "chartApplication";
	public static final String SETTINGS_PORT_ELEMENT_NAME = "port";
	public static final String SETTINGS_EXE_PATH_ELEMENT_NAME = "exePath";

	private IRuntimeContext runtime;
	private File[] executables;
	private String serverName = null;
	private int port = -1;

	/** Setup and initialize an instance of this object.
	 */
	public void setInitializationData(
		IRuntimeContext runtime,
		IConfigurationElement config,
		String propertyName,
		Object data)
		throws Exception {

		this.runtime = runtime;

		// Lookup the server name and port of the M server we're connected to,
		// since we will need this information to launch the chart later. 
		try {
			IMumpsFactory mPlugin =
				(IMumpsFactory) runtime.getPluginInstance(
					IMumpsFactory.PLUGIN_ID);
			this.serverName = mPlugin.getServerName();
			this.port = mPlugin.getServerPort();
		} catch (PluginStartupException e) {
			this.serverName = null;
		} catch (PluginNotFoundException e) {
			this.serverName = null;
		}
		if (this.serverName == null)
			return;

		// retrieve the settings from the runtime. check for an alternative
		// port and look for declared executable paths.
		LinkedList executablePaths = new LinkedList();
		IConfigurationElement[] settings = runtime.getSettings();
		for (int i = 0; i < settings.length; i++) {
			if (settings[i].getName().equals(SETTINGS_ROOT_ELEMENT_NAME))
				parseSetting(settings[i], executablePaths);
		}

		this.executables = new File[executablePaths.size()];
		this.executables = (File[]) executablePaths.toArray(this.executables);
	}

	private void parseSetting(
		IConfigurationElement rootElem,
		LinkedList executablePaths) {
		parsePortSetting(rootElem);
		parsePathSettings(rootElem, executablePaths);
	}

	private void parsePathSettings(
		IConfigurationElement rootElem,
		LinkedList executablePaths) {
		// find all the children of our configuration element that are named
		// "executable".  Retrieve their "path" attribute and see if it 
		// points to a valid, existing file.
		IConfigurationElement[] exes =
			rootElem.getChildren(SETTINGS_EXE_PATH_ELEMENT_NAME);
		for (int i = 0; i < exes.length; i++) {
			File f = new File(exes[i].getValue());
			if (f.exists() && f.canRead())
				executablePaths.add(f);
		}
	}

	private void parsePortSetting(IConfigurationElement rootElem) {
		IConfigurationElement[] ports =
			rootElem.getChildren(SETTINGS_PORT_ELEMENT_NAME);
		if (ports.length == 0)
			return;

		try {
			this.port = Integer.parseInt(ports[0].getValue());
		} catch (NumberFormatException nfe) {
		}
	}

	/** Launch an instance of the CPRS chart.
	 * 
	 * @return true if the chart was launched successfully. 
	 */
	public boolean launchChart() {
		if (serverName == null || port == -1 || executables == null)
			return false;

		Runtime r = Runtime.getRuntime();

		for (int i = 0; i < executables.length; i++)
			try {
				File exe = executables[i];
				File dir = exe.getParentFile();
				String[] cmdLine =
					new String[] {
						exe.getPath(),
						"s=" + serverName,
						"p=" + port,
						"ccow=force" };
				Process p = r.exec(cmdLine, null, dir);
				return true;
			} catch (Exception e) {
				// do nothing - just try the next item in the list.
			}

		// we were unable to launch the chart.
		return false;
	}

}
