package gov.va.med.hds.cd.ccow.internal;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;

import gov.va.med.hds.cd.ccow.internal.CCOWParticipantServer;
import gov.va.med.hds.cd.ccow.internal.SimpleCCOWParticipant;
import gov.va.med.hds.cd.ccow.internal.TinyWebServer;

import junit.framework.TestCase;

public class CCOWParticipantServerTest extends TestCase {

    public CCOWParticipantServerTest(String name) {
        super(name);
    }

    protected CCOWParticipantServer server =
        CCOWParticipantServer.getInstance();

    protected MockParticipant participant;
    protected String participantURL;
    protected long coupon;
    protected String currentComplaint;
    protected boolean shouldThrowException;

    protected void setUp() {
        participant = new MockParticipant();
        participantURL = server.registerParticipant(participant);
        coupon = System.currentTimeMillis();
        shouldThrowException = false;
    }
    protected void tearDown() {
        server.unregisterParticipant(participant);
        participant = null;
    }

    private static byte[] slurpContents(InputStream in, boolean close)
        throws IOException
    {
        byte [] result = null;
        ByteArrayOutputStream slurpBuffer = new ByteArrayOutputStream();
        byte[] buffer = new byte[1024];
        int bytesRead;
        while ((bytesRead = in.read(buffer)) != -1)
            slurpBuffer.write(buffer, 0, bytesRead);
        result = slurpBuffer.toByteArray();
        if (close) try { in.close(); } catch (IOException ioe) {}
        return result;
    }

    private Map performConnection(String args) throws IOException {
        URL u = new URL(participantURL +
                        "?interface=ContextParticipant" + args);
        URLConnection conn = u.openConnection();
        conn.connect();
        int status = ((HttpURLConnection) conn).getResponseCode();

        Map result = null;

        if (status != 200) {
            result = new HashMap();
            result.put("STATUS", Integer.toString(status));

        } else {
            byte [] contents = slurpContents(conn.getInputStream(), true);
            String reply = new String(contents);
            result = TinyWebServer.parseParams(reply);
        }
        return result;
    }


    public void testContextChangesPendingAccept() throws IOException {
        currentComplaint = null;
        Map results =  performConnection
            ("&method=ContextChangesPending" +
             "&contextcoupon=" + coupon);
        assertEquals(2, results.size());
        assertEquals("accept", results.get("decision"));
        assertEquals("", results.get("reason"));
        assertEquals(participant.lastMethodCalled, "contextChangesPending");
        assertEquals(participant.lastMethodArgument, coupon);
    }


    public void testContextChangesPendingConditionalAccept()
        throws IOException
    {
        currentComplaint = MOCK_COMPLAINT;
        Map results =  performConnection
            ("&method=ContextChangesPending" +
             "&contextcoupon=" + coupon);
        assertEquals(2, results.size());
        assertEquals("conditional_accept", results.get("decision"));
        assertEquals(MOCK_COMPLAINT, results.get("reason"));
        assertEquals(participant.lastMethodCalled, "contextChangesPending");
        assertEquals(participant.lastMethodArgument, coupon);
    }


    public void testContextChangesAccepted() throws IOException {
        Map results =  performConnection
            ("&method=ContextChangesAccepted" +
             "&contextcoupon=" + coupon);
        assertEquals(0, results.size());
        assertEquals(participant.lastMethodCalled, "contextChangesAccepted");
        assertEquals(participant.lastMethodArgument, coupon);
    }


    public void testContextChangesCanceled() throws IOException {
        Map results =  performConnection
            ("&method=ContextChangesCanceled" +
             "&contextcoupon=" + coupon);
        assertEquals(0, results.size());
        assertEquals(participant.lastMethodCalled, "contextChangesCanceled");
        assertEquals(participant.lastMethodArgument, coupon);
    }


    public void testCommonContextTerminated() throws IOException {
        Map results =  performConnection("&method=CommonContextTerminated");
        assertEquals(0, results.size());
        assertEquals(participant.lastMethodCalled, "commonContextTerminated");
        assertEquals(participant.lastMethodArgument, -1);
    }


    public void testPing() throws IOException {
        Map results =  performConnection("&method=ping");
        assertEquals(0, results.size());
        assertEquals(participant.lastMethodCalled, "ping");
        assertEquals(participant.lastMethodArgument, -1);
    }


    public void testPingException() throws IOException {
        shouldThrowException = true;
        Map results =  performConnection("&method=ping");
        assertEquals(1, results.size());
        assertNotNull(results.get("STATUS"));
        assertEquals("500", results.get("STATUS"));
    }



    private class MockParticipant implements SimpleCCOWParticipant {

        public String lastMethodCalled = null;
        public long lastMethodArgument = -1;

        public String contextChangesPending(long pendingContextCoupon) {
            lastMethodCalled = "contextChangesPending";
            lastMethodArgument = pendingContextCoupon;
            return currentComplaint;
        }
        public void contextChangesAccepted(long newContextCoupon) {
            lastMethodCalled = "contextChangesAccepted";
            lastMethodArgument = newContextCoupon;
        }
        public void contextChangesCanceled(long canceledContextCoupon) {
            lastMethodCalled = "contextChangesCanceled";
            lastMethodArgument = canceledContextCoupon;
        }
        public void commonContextTerminated() {
            lastMethodCalled = "commonContextTerminated";
            lastMethodArgument = -1;
        }
        public void ping() {
            if (shouldThrowException)
                throw new RuntimeException();
            lastMethodCalled = "ping";
            lastMethodArgument = -1;
        }
    }
    private static final String MOCK_COMPLAINT = "This is a mock complaint."+
        " It contains special characters: (?&#/=<>)";
}
