

package gov.va.med.hds.cd.ccow;

import java.awt.Component;
import java.util.Map;
import java.util.Set;

/** Manages an instance of the clinical context, and its connection to CCOW.
 *
 * The clinician desktop framework will maintain a unique instance of
 * this class for each clinician desktop window.
 */
public interface IClinicalContextBroker {

    /** Get the contents of the clinical context.
     *
     * @returns an immutable Map containing all the items in the
     *     clinical context.  All of the keys in the map will be of
     *     type {@link ContextItemName}; all of the values in the map
     *     will be of type String.
     */
    public Map getContextItems();

    /** Request permission to change the clinical context.
     *
     * Desktop components should call this method as soon as the user
     * indicates their desire to initiate a context change. For
     * example, when the user requests to see a patient selection
     * dialog, this method should be called <b>before</b> displaying
     * the dialog.
     *
     * @param contextItemNames A list of context items that the caller
     *       might possibly change if permission were granted
     * @return an opaque token indicating permission to change the
     *       context.  If the user has decided to cancel the change,
     *       returns null.
     */
    public Object requestChange(Set contextItemNames);

    /** Allows a context participant to change the clinical context.
     *
     * @param requestToken the permission object returned earlier by
     *       the {@link #requestChange(Set)} operation
     * @param the new items to store in the context. If keys in this
     *       Map are not {@link ContextItemName} objects, their
     *       toString method will be called to obtain the full name of
     *       the context item.  If values in the Map are not String
     *       objects, their toString method will be called to obtain
     *       the value of the context item.
     * @return true if the context change was successful.
     */
    public boolean changeContext(Object requestToken, Map newContext);

    /** Allows a context participant to change the clinical context.
     *
     * @param requestToken the permission object returned earlier by
     *       the {@link #requestChange(Set)} operation
     * @param the new items to store in the context. If keys in this
     *       Map are not {@link ContextItemName} objects, their
     *       toString method will be called to obtain the full name of
     *       the context item.  If values in the Map are not String
     *       objects, their toString method will be called to obtain
     *       the value of the context item.
     * @param parent the user interface component or window which is
     *       initiating the context change.  (If the context change
     *       operation has a need to interact with the user, it will
     *       display modal dialogs over this component.)
     * @return true if the context change was successful.
     */
    public boolean changeContext(Object requestToken, Map newContext, 
                                 Component parent);


    /** Return true if the broker is currently connected to an external
     * clinical context
     */
    public boolean isConnected();

    /** Break the link to the external clinical context.
     */
    public void breakExternalLink();

    /** Rejoin the external clinical context.
     *
     * @param useOurContext true if we should synchronize the external
     *       context to our internal context; false if we should synchronize
     *       our internal context to the external context.
     * @return <code>true</code> if the broker is connected to the
     * clinical context on return.
     */
    public boolean joinExternalContext(boolean useOurContext);
    

    /** Rejoin the external clinical context.
     *
     * @param useOurContext true if we should synchronize the external
     *       context to our internal context; false if we should synchronize
     *       our internal context to the external context.
     * @param parent the user interface component or window which is
     *       initiating the context change.  (If the context change
     *       operation has a need to interact with the user, it will
     *       display modal dialogs over this component.)
     * @return <code>true</code> if the broker is connected to the
     * clinical context on return.
     */
    public boolean joinExternalContext(boolean useOurContext, Component parent);

}
