package gov.va.med.hds.cd.config.configDAO;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.ICallResultHandler;
import gov.va.med.hds.cd.config.IConfigCoreConstants;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.cd.config.ParameterCategoryException;
import gov.va.med.hds.cd.config.ResultHandlerException;
import gov.va.med.hds.cd.daoregistry.IDataAccessObject;
import gov.va.med.hds.cd.m.ICallRequest;
import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.m.IMumpsFactory;
import gov.va.med.hds.cd.m.IMumpsListener;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.util.mumps.StringUtil;
import gov.va.med.hds.util.swingworker.SwingWorker;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class MConfigDAO implements IConfigDAO, IDataAccessObject {
	private static final String PUT_PARAM_RPC_KEY = "PUT PARAM";
	private static final String REMOVE_PARAM_CAT_RPC_KEY = "REMOVE PARAM CAT";
	private static final String DELETE_PARAM_CAT_RPC_KEY = "DELETE PARAM CAT";
	private static final String REMOVE_PARAM_FROM_CAT_RPC_KEY = "REMOVE PARAM FROM CAT";
	private static final String ADD_PARAM_TO_CAT_RPC_KEY = "ADD PARAM CAT";
	private static final String INSERT_PARAM_CAT_RPC_KEY = "INSERT PARAM CAT";
	private static final String UPDATE_PARAM_CAT_RPC_KEY = "UPDATE PARAM CAT";
	private IMumpsFactory mConn = null;
	private IConfiguration configuration;
	private SwingWorker worker;
	private IMumps mumps = null;
	private IRuntimeContext runtime;
	private IConfigurationElement config;
	public MConfigDAO() {
	}

	/**
	 * @see gov.va.med.hds.cd.daoregistry.IDataAccessObject#init(gov.va.med.hds.cd.runtime.IRuntimeContext)
	 */
	public void setInitializationData(IRuntimeContext runtime, IConfigurationElement config, String propertyName, Object data) throws Exception {
		this.runtime = runtime;
		this.config = config;
	}

	private void assureMumpsIsCreated() throws ConfigDAOException {
		if (mumps != null)
			return;
		try {
			IPluginDescriptor desc = runtime.getPluginRegistry().getPluginDescriptor(IConfigCoreConstants.PLUGIN_ID);
			mumps = ((IMumpsFactory) runtime.getPluginInstance(IMumpsFactory.PLUGIN_ID)).create(desc);
		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#putParameterTree(gov.va.med.hds.cd.config.IParameterTree)
	 */
	public void putParameterTree(IParameterTree parameterTree) throws ConfigDAOException {
		IParameterCategory parameterRoot = (IParameterCategory) parameterTree.getRoot();
		updateParameterCategory(parameterRoot);
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#insertParameterCategory(gov.va.med.hds.cd.config.IParameterCategory, gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void insertParameterCategory(IParameterCategory pc, IParameterCategory parent) throws ConfigDAOException {
		assureMumpsIsCreated();

		//do depth-first traversal
		List children = pc.children();
		Iterator scIt = children.iterator();
		HashMap pcMap = new HashMap();

		int scSeq = 0;

		while (scIt.hasNext()) {
			IParameterCategory subCat = (IParameterCategory) scIt.next();
			if (!subCat.isLeaf()) {
				insertParameterCategory(subCat, pc);
				scSeq++;

				pcMap.put("3^" + scSeq + "^.01", Integer.toString(scSeq));
				pcMap.put("3^" + scSeq + "^.02", (subCat.getId() > 0 ? "`" + Long.toString(subCat.getId()) : ""));
			}
		}

		ArrayList callParams = new ArrayList();

		pcMap.put(".01", pc.getFullName());
		pcMap.put(".02", pc.getPluginId());
		pcMap.put(".03", pc.getName());
		pcMap.put(".04", (parent.getId() > 0 ? "`" + Long.toString(parent.getId()) : ""));
		pcMap.put(".05", Boolean.toString(pc.isPackageRoot()));
		pcMap.put("1", pc.getPreferencePage());

		//add parameters to list
		Map pars = (TreeMap) pc.parameters();
		int pCnt = 0;
		Iterator keysIt = (pars.keySet()).iterator();
		while (keysIt.hasNext()) {
			IParameter parameter = (IParameter) pars.get(keysIt.next());
			pCnt++;
			pcMap.put("2^" + pCnt + "^.01", parameter.getName());
			pcMap.put("2^" + pCnt + "^.02", (parameter.getId() > 0 ? "`" + Long.toString(parameter.getId()) : ""));
		}

		callParams.add(pcMap);

		String result = null;
		try {
			result = mumps.callRPC(INSERT_PARAM_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD INSERT PARAMETER CATEGORY failed with message: " + StringUtil.piece(result, '^', 2));
			} else {
				pc.setId(Long.parseLong(StringUtil.piece(result, '^', 2)));
				pc.setPersistent(true);
			}

		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	public void addParameterToCategory(IParameter parameter, IParameterCategory pc) throws ConfigDAOException {
		assureMumpsIsCreated();
		HashMap pcMap = new HashMap();
		ArrayList callParams = new ArrayList();

		if (pc.getParameter(parameter.getKey()) != null) {
			throw new ConfigDAOException("Parameter " + parameter + " is already part of " + pc);
		}

		int pCnt = pc.parameters().size() + 1;

		pcMap.put("IEN", Long.toString(pc.getId()));

		pcMap.put("2^" + pCnt + "^.01", parameter.getName());
		pcMap.put("2^" + pCnt + "^.02", (parameter.getId() > 0 ? "`" + Long.toString(parameter.getId()) : ""));

		callParams.add(pcMap);

		String result = null;
		try {
			result = mumps.callRPC(ADD_PARAM_TO_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD ADD PARAMETER TO CATEGORY failed with message: " + StringUtil.piece(result, '^', 2));
			}

		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#removeParameterFromCategory(gov.va.med.hds.cd.config.IParameter, gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void removeParameterFromCategory(IParameter parameter, IParameterCategory pc) throws ConfigDAOException {
		assureMumpsIsCreated();
		ArrayList callParams = new ArrayList();

		callParams.add(Long.toString(parameter.getId()));
		callParams.add(Long.toString(pc.getId()));

		String result = null;

		try {
			result = mumps.callRPC(REMOVE_PARAM_FROM_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD REMOVE PARAM FROM CAT failed with message: " + StringUtil.piece(result, '^', 2));
			}
		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#updateParameterCategory(gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void updateParameterCategory(IParameterCategory pc) throws ConfigDAOException {
		assureMumpsIsCreated();
		//do depth-first traversal
		List children = pc.children();
		Iterator scIt = children.iterator();
		HashMap pcMap = new HashMap();

		int scSeq = 0;

		while (scIt.hasNext()) {

			IParameterCategory subCat = (IParameterCategory) scIt.next();
			
			if (subCat.isMarkedForRemoval()) {
				break;
			}

			if (!subCat.isPersistent() || !(subCat.getId() > 0)) {
				insertParameterCategory(subCat, pc);
				break;
			}

			scSeq++;

			pcMap.put("3^" + scSeq + "^.01", Integer.toString(scSeq));
			pcMap.put("3^" + scSeq + "^.02", (subCat.getId() > 0 ? "`" + Long.toString(subCat.getId()) : ""));

		}

		if (pc.isMarkedForRemoval()) {
			deleteParameterCategory(pc, true);
			return;
		}
		
		if (!pc.isPersistent() || !(pc.getId() > 0)) {
			insertParameterCategory(pc, pc.getParent());
			return;
		}

		ArrayList callParams = new ArrayList();

		pcMap.put("IEN", Long.toString(pc.getId()));
		pcMap.put(".01", pc.getFullName());
		pcMap.put(".02", pc.getPluginId());
		pcMap.put(".03", pc.getName());
		pcMap.put(".04", (pc.getParentId() > 0 ? "`" + Long.toString(pc.getParentId()) : ""));
		pcMap.put(".05", Boolean.toString(pc.isPackageRoot()));
		pcMap.put("1", pc.getPreferencePage());

		//add parameters to list
		Map pars = (TreeMap) pc.parameters();
		int pCnt = 0;
		Iterator keysIt = (pars.keySet()).iterator();
		while (keysIt.hasNext()) {
			IParameter parameter = (IParameter) pars.get(keysIt.next());
			pCnt++;
			pcMap.put("2^" + pCnt + "^.01", parameter.getName());
			pcMap.put("2^" + pCnt + "^.02", (parameter.getId() > 0 ? "`" + Long.toString(parameter.getId()) : ""));
		}

		callParams.add(pcMap);

		String result = null;
		try {
			result = mumps.callRPC(UPDATE_PARAM_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD UPDATE PARAMETER CATEGORY failed with message: " + StringUtil.piece(result, '^', 2));
			}
		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#deleteParameterCategory(gov.va.med.hds.cd.config.IParameterCategory, boolean)
	 */
	public void deleteParameterCategory(IParameterCategory pc, boolean deleteSubCategories) throws ConfigDAOException {
		assureMumpsIsCreated();
		ArrayList callParams = new ArrayList();

		callParams.add(Long.toString(pc.getId()));
		callParams.add(deleteSubCategories == true ? "1" : "0");

		String result = null;

		try {
			result = mumps.callRPC(DELETE_PARAM_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD DELETE PARAMETER CATEGORY failed with message: " + StringUtil.piece(result, '^', 2));
			} else {
				pc.setId(Long.parseLong(StringUtil.piece(result, '^', 2)));
			}

		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}

	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#removeParameterCategory(gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void removeParameterCategory(IParameterCategory pc) throws ParameterCategoryException, ConfigDAOException {

		IParameterCategory parent = pc.getParent();
		removeParameterCategory(pc, parent);
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#removeParameterCategory(gov.va.med.hds.cd.config.IParameterCategory, gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void removeParameterCategory(IParameterCategory pc, IParameterCategory parent) throws ParameterCategoryException, NumberFormatException, ConfigDAOException {
		assureMumpsIsCreated();
		ArrayList callParams = new ArrayList();

		if (parent == null) {
			throw new ParameterCategoryException("Parameter Category " + pc.getName() + " has no parent, and so cannot be removed.");
		}
		callParams.add(Long.toString(pc.getId()));
		callParams.add(Long.toString(parent.getId()));

		String result = null;

		try {
			result = mumps.callRPC(REMOVE_PARAM_CAT_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(result, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD REMOVE PARAMETER CATEGORY failed with message: " + StringUtil.piece(result, '^', 2));
			} else {
				pc.setId(Long.parseLong(StringUtil.piece(result, '^', 2)));
			}

		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#putParameter(java.lang.String, java.lang.String)
	 */
	public void putParameter(String key, String value) throws ConfigDAOException {
		assureMumpsIsCreated();
		ArrayList callParams = new ArrayList();
		ArrayList val = new ArrayList();
		val.addAll(Arrays.asList(value.split("\n")));

		callParams.add(key);
		callParams.add(val);

		String error = null;
		try {
			error = mumps.callRPC(PUT_PARAM_RPC_KEY, callParams);
			if (Integer.parseInt(StringUtil.piece(error, '^', 1)) > 0) {
				throw new ConfigDAOException("Remote Procedure XHD PUT PARAMETER failed with message: " + StringUtil.piece(error, '^', 2));
			}
		} catch (Exception e) {
			throw new ConfigDAOException(e);
		}
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#setConfiguration(gov.va.med.hds.cd.config.IConfiguration)
	 */
	public void setConfiguration(IConfiguration configuration) {
		this.configuration = configuration;
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfigDAO#query(gov.va.med.hds.cd.m.ICallRequest, gov.va.med.hds.cd.config.ICallResultHandler, gov.va.med.hds.cd.config.IConfigDAOListener)
	 */
	public void query(final ICallRequest request, final ICallResultHandler handler, final IConfigDAOListener listener) {

		try {
			assureMumpsIsCreated();
			mumps.callRPC(request.getCallName(), request.getParams(), new IMumpsListener() {
				ConfigDAOEvent configDAOEvent = new ConfigDAOEvent(this, request);
				public void CallFailed(ICallRequest request, MumpsException e) {
					configDAOEvent.setException(new ConfigDAOException(e));
					listener.callFinished(configDAOEvent);
				}
				public void CallSucceeded(ICallRequest request, String result) {
					try {
						configDAOEvent.setResultObject(handler.processResult(result));
					} catch (ResultHandlerException e) {
						configDAOEvent.setException(e);
					}
					listener.callFinished(configDAOEvent);
				}
			});
		} catch (Exception e) {
			ConfigDAOEvent configDAOEvent = new ConfigDAOEvent(this, request);
			configDAOEvent.setException(e);
			listener.callFinished(configDAOEvent);
		}
	}

	/**
	 * Method queryWorker uses SwingWorker to issue call result on Swing
	 * Dispatch Thread.
	 * 
	 * @param request
	 * @param handler
	 * @param listener
	 */
	public void queryWorker(final ICallRequest request, final ICallResultHandler handler, final IConfigDAOListener listener) {

		worker = new SwingWorker() {
			ConfigDAOException queryException = null;
			String callResult = null;
			public synchronized Object construct() {
				try {
					assureMumpsIsCreated();
					mumps.callRPC(request.getCallName(), request.getParams(), new IMumpsListener() {
						public void CallFailed(ICallRequest request, MumpsException e) {
							queryException = new ConfigDAOException(e);
							synchronized (worker) {
								worker.notify();
							}
						}
						public void CallSucceeded(ICallRequest request, String result) {
							callResult = result;
							synchronized (worker) {
								worker.notify();
							}
						}
					});
					wait();
					// timeout
					if ((callResult == null) && (queryException == null)) {
						queryException = new ConfigDAOException("Call " + request.getCallName() + " timed out...");
					}
				} catch (Exception e) {
					queryException = new ConfigDAOException(e);
				}
				return callResult;
			}
			public void finished() {
				ConfigDAOEvent configDAOEvent = new ConfigDAOEvent(this, request);
				if (queryException != null) {
					configDAOEvent.setException(queryException);
					listener.callFinished(configDAOEvent);
					return;
				}
				try {
					configDAOEvent.setResultObject(handler.processResult(get().toString()));
				} catch (ResultHandlerException e) {
					queryException = new ConfigDAOException(e);
				}
				listener.callFinished(configDAOEvent);
			}
		};
		worker.start();
	}

}
