package gov.va.med.hds.cd.config.internal;

import java.util.ArrayList;
import java.util.EventListener;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.event.EventListenerList;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.ConfigurationChangeEvent;
import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.ConfigurationReadyEvent;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurationChangeListener;
import gov.va.med.hds.cd.config.IConfigurationReadyListener;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterChangeListener;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.cd.config.NonExistentParameterException;
import gov.va.med.hds.cd.config.ParameterCategoryException;
import gov.va.med.hds.cd.config.ParameterChangeEvent;
import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.security.CouldNotCreateUserInfoException;
import gov.va.med.hds.cd.security.ISecurity;
import gov.va.med.hds.cd.security.IUserInfo;
import gov.va.med.hds.util.mumps.StringUtil;
/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class Configuration
	implements IConfiguration, IConfigurationChangeListener {

	private static final String GET_CONFIGURATION_RPC_KEY = "GET CONFIGURATION";
	private EventListenerList configurationListeners = new EventListenerList();
	private IConfigDAO configDAO;
	private Map projectionMap = new HashMap();
	private Map indexMap = new HashMap();
	private Map catMap = new HashMap();
	private IRuntimeContext runtime;
	private IDaoRegistry daoRegistry;
	private ISecurity security;
	private IUserInfo userInfo;
	private String pluginID = "";
	private ConfigurationCallHandler configurationCallHandler;
	private ConfigurationCallHandler mutableConfigurationCallHandler;
	protected IParameterTree parameterTree;
	protected IParameterCategory parameterRoot;
	protected IParameterTree mutableParameterTree = null;
	protected IParameterCategory mutableParameterRoot = null;

	Configuration(
		String pluginID,
		IRuntimeContext runtime,
		IConfigurationReadyListener crl)
		throws ConfigurationException {

		this.runtime = runtime;
		this.pluginID = pluginID;

		try {
			daoRegistry =
				(IDaoRegistry) runtime.getPluginInstance(
					IDaoRegistry.PLUGIN_ID);
		} catch (PluginStartupException e) {
			throw new ConfigurationException(e);
		} catch (PluginNotFoundException e) {
			throw new ConfigurationException(e);
		}
		try {
			security =
				(ISecurity) runtime.getPluginInstance(ISecurity.PLUGIN_ID);
			try {
				userInfo = security.getUserInfo();
			} catch (CouldNotCreateUserInfoException e) {
				throw new ConfigurationException(e);
			}
		} catch (PluginStartupException e) {
			throw new ConfigurationException(e);
		} catch (PluginNotFoundException e) {
			throw new ConfigurationException(e);
		}

		try {
			configDAO =
				(IConfigDAO) daoRegistry.getDAO(IConfigDAO.INTERFACE_NAME);
			configDAO.setConfiguration((IConfiguration) this);
		} catch (Exception e) {
			runtime.getLogger(getClass()).error("Unable to get reference to instance of " + IConfigDAO.class.getName(), e);
		}
		addConfigurationReadyListener(crl);
		List params = new ArrayList();
		params.add(pluginID);
		configurationCallHandler = new ConfigurationCallHandler(this);
		configDAO.query(
			new CallRequest(GET_CONFIGURATION_RPC_KEY, params),
			configurationCallHandler,
			new ConfigurationCallListener());

	}

	/**
	 * Method buildProjection.
	 * @param parameterRoot
	 */
	void buildProjection(IParameterCategory pCat) {

		if (pCat == null)
			return;

		//copy pCat parameters Map to projectionMap
		Iterator keyIt = pCat.parameters().keySet().iterator();
		while (keyIt.hasNext()) {
			String fullKey = (String) keyIt.next();
			String ezKey = StringUtil.piece(fullKey, '^', 1);
			String value = pCat.getParameterValue(fullKey);
			projectionMap.put(ezKey, value);
			indexMap.put(ezKey, fullKey);
			catMap.put(fullKey, pCat);
		}

		/* iterate through pCat children & add their
		 * parameters to projectionMap */
		Iterator scIt = pCat.children().iterator();
		while (scIt.hasNext()) {
			IParameterCategory subCat = (ParameterCategory) scIt.next();
			if (subCat != null) {
				buildProjection(subCat);
			}
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addConfigurationChangeListener(IConfigurationChangeListener)
	 */
	public void addConfigurationChangeListener(IConfigurationChangeListener iConfigurationListener) {
		configurationListeners.add(
			IConfigurationChangeListener.class,
			iConfigurationListener);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeConfigurationChangeListener(IConfigurationChangeListener)
	 */
	public void removeConfigurationChangeListener(IConfigurationChangeListener iConfigurationListener) {
		configurationListeners.remove(
			IConfigurationChangeListener.class,
			iConfigurationListener);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addParameterChangeListener(gov.va.med.hds.cd.config.IParameterChangeListener)
	 */
	public void addParameterChangeListener(IParameterChangeListener pcl) {
		configurationListeners.add(IParameterChangeListener.class, pcl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeParameterChangeListener(gov.va.med.hds.cd.config.IParameterChangeListener)
	 */
	public void removeParameterChangeListener(IParameterChangeListener pcl) {
		configurationListeners.remove(IParameterChangeListener.class, pcl);
	}

	/**
	 * Method fireConfigurationChanged.
	 * @param modParam
	 * @param modParamCat
	 */
	protected void fireConfigurationChanged(
		IParameterCategory subject,
		IParameterCategory object,
		int verb) {

		EventListener[] ccl;

		ccl =
			configurationListeners.getListeners(
				IConfigurationChangeListener.class);

		if (ccl.length == 0) {
			return;
		}

		ConfigurationChangeEvent ccevent =
			new ConfigurationChangeEvent(this, subject, object, verb);

		for (int i = 0; i < ccl.length; ++i) {

			IConfigurationChangeListener listener =
				(IConfigurationChangeListener) ccl[i];
			listener.configurationChanged(ccevent);
		}
	}

	/**
	 * Method fireParameterChanged.
	 * @param parameter
	 */
	protected void fireParameterChanged(IParameter parameter) {

		EventListener[] pcl;

		pcl =
			configurationListeners.getListeners(IParameterChangeListener.class);

		if (pcl.length == 0) {
			return;
		}

		ParameterChangeEvent pcevent =
			new ParameterChangeEvent(this, parameter);

		for (int i = 0; i < pcl.length; ++i) {

			IParameterChangeListener listener =
				(IParameterChangeListener) pcl[i];
			listener.parameterChanged(pcevent);
		}

	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationChangeListener)
	 */
	public void addConfigurationReadyListener(IConfigurationReadyListener crl) {
		configurationListeners.add(IConfigurationReadyListener.class, crl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationChangeListener)
	 */
	public void removeConfigurationReadyListener(IConfigurationReadyListener crl) {
		configurationListeners.remove(IConfigurationReadyListener.class, crl);
	}

	/**
	 * Method fireConfigurationReady. Informs all registered
	 * ConfigurationReadyListners when the current instance has finished its
	 * query to the data layer.
	 */
	protected void fireConfigurationReady() {
		EventListener[] crl;

		crl =
			configurationListeners.getListeners(
				IConfigurationReadyListener.class);

		if (crl.length == 0) {
			return;
		}

		ConfigurationReadyEvent crevent = new ConfigurationReadyEvent(this);

		for (int i = 0; i < crl.length; ++i) {

			IConfigurationReadyListener listener =
				(IConfigurationReadyListener) crl[i];
			listener.configurationReady(crevent);
		}
	}

	/**
	 * Method fireConfigurationCallFailed.
	 * @param configurationException
	 */
	protected void fireConfigurationCallFailed(ConfigurationException configurationException) {
		EventListener[] crl;

		crl =
			configurationListeners.getListeners(
				IConfigurationReadyListener.class);

		if (crl.length == 0) {
			return;
		}

		for (int i = 0; i < crl.length; ++i) {

			IConfigurationReadyListener listener =
				(IConfigurationReadyListener) crl[i];
			listener.configurationCallFailed(configurationException);
		}
	}

	/**
	 * Returns the parameterTree.
	 * @return IParameterTree
	 */
	public IParameterTree getParameterTree() {
		if (parameterTree == null) {
			parameterRoot = getParameterRoot();
			parameterTree = new ParameterTree(mutableParameterRoot);
		}
		return parameterTree;
	}

	/**
	 * Returns the parameterRoot.
	 * @return IParameterCategory
	 */
	public IParameterCategory getParameterRoot() {
		if (parameterRoot == null) {
			parameterRoot = new ParameterCategory("Retrieving data...", this);
			parameterTree = new ParameterTree(mutableParameterRoot);
		}
		return parameterRoot;
	}

	/**
	 * Method commitChanges.
	 * @throws ConfigDAOException
	 */
	public void commitChanges() throws ConfigDAOException {
		configDAO.putParameterTree(parameterTree);
	}

	/**
	 * Method removeParameterCategory.
	 * @param pc
	 * @throws ConfigurationException
	 */
	public void removeParameterCategory(IParameterCategory pc)
		throws ConfigurationException {
		try {
			configDAO.removeParameterCategory(pc);
			fireConfigurationChanged(
				null,
				pc,
				ConfigurationChangeEvent.CATEGORY_REMOVED);
		} catch (ParameterCategoryException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getParameterValue(String)
	 */
	public String getParameterValue(String ezKey) {
		return (String) projectionMap.get(ezKey);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#putParameter(String, String)
	 */
	public void putParameter(String ezKey, String value)
		throws NonExistentParameterException, ConfigurationException {
		projectionMap.put(ezKey, value);
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException(
				"Parameter Category "
					+ pCat.getName()
					+ " not defined on Server");
		}
		pCat.putParameter(fullKey, value);
		try {
			configDAO.putParameter(fullKey, value);
			IParameter modParam = getParameter(ezKey);
			fireParameterChanged(modParam);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		} catch (NonExistentParameterException nPe) {
			throw new ConfigurationException(
				"Configuration Exception: " + nPe.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#putParameter(IParameter)
	 */
	public void putParameter(IParameter parameter)
		throws ConfigurationException {
		IParameterCategory pCat =
			getCategoryForParameter(
				StringUtil.piece(parameter.getKey(), '^', 1));
		if (!(pCat instanceof IParameterCategory)) {
			throw new ConfigurationException(
				"Category not found for Parameter: " + parameter.getName());
		}

		pCat.putParameter(parameter);

		try {
			configDAO.putParameter(parameter.getKey(), parameter.getValue());
			fireParameterChanged(parameter);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getCategoryForParameter(String)
	 */
	public IParameterCategory getCategoryForParameter(String ezKey) {
		String fullKey = (String) indexMap.get(ezKey);
		return (ParameterCategory) catMap.get(fullKey);
	}

	/**
	 * Method getListenerCount.
	 */
	public int getListenerCount(Class type) {
		return configurationListeners.getListenerCount(type);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getParameter(String)
	 */
	public IParameter getParameter(String ezKey)
		throws NonExistentParameterException {
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException("Parameter not defined on Server");
		}
		return pCat.getParameter(fullKey);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#updateParameterCategory(IParameterCategory)
	 */
	public void updateParameterCategory(IParameterCategory parameterCategory)
		throws ConfigDAOException {
		configDAO.updateParameterCategory(parameterCategory);
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfiguration#insertParameterCategory(gov.va.med.hds.cd.config.IParameterCategory, gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void insertParameterCategory(
		IParameterCategory pc,
		IParameterCategory parent)
		throws ConfigDAOException {
		configDAO.insertParameterCategory(pc, parent);
	}

	/**
	 * Method getUserInfo. Used internally in Configurator to reference user
	 * demographic information.
	 * @return IUserInfo
	 */
	public IUserInfo getUserInfo() {
		return userInfo;
	}

	class ConfigurationCallListener implements IConfigDAOListener {

		/**
		 * @see gov.va.med.hds.cd.config.IConfigDAOListener#callFinished(gov.va.med.hds.cd.config.ConfigDAOEvent)
		 */
		public void callFinished(ConfigDAOEvent configDAOEvent) {
			if (configDAOEvent.getException() != null) {
				fireConfigurationCallFailed(
					new ConfigurationException(configDAOEvent.getException()));
			} else {
				parameterRoot =
					(IParameterCategory) configDAOEvent.getResultObject();
				parameterTree = new ParameterTree(parameterRoot);
				buildProjection(parameterRoot);
				fireConfigurationReady();
			}
		}

	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return pluginID;
	}
	/**
	 * @see gov.va.med.hds.cd.config.IConfigurationChangeListener#configurationChanged(gov.va.med.hds.cd.config.ConfigurationChangeEvent)
	 */
	public void configurationChanged(ConfigurationChangeEvent cce) {
	}

}
