package gov.va.med.hds.cd.config.internal;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.ConfigurationReadyEvent;
import gov.va.med.hds.cd.config.ICallResultHandler;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurationReadyListener;
import gov.va.med.hds.cd.config.IConfigurator;
import gov.va.med.hds.cd.config.IParameterDefinitionCallListener;
import gov.va.med.hds.cd.config.ParameterDefinitionException;
import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.runtime.IPlugin;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class Configurator
	implements IConfigurator, IPlugin, IConfigDAOListener {

	private static final String GET_PARAETER_DEF_LIST_RPC_KEY = "PARAM DEF LIST";
	private Map configMap = new HashMap();
	private Map mutableConfigMap = new HashMap();
	private Map preferencePageMap = new HashMap();
	private IPluginDescriptor plugin;
	private IRuntimeContext runtime;
	private IDaoRegistry daoRegistry;
	private IConfigDAO configDAO;
	private List parameterDefinitions = null;
	private IParameterDefinitionCallListener parameterDefinitionListener;

	public Configurator() {
	}

	public IConfiguration getConfiguration(String pluginId, int timeOut)
			throws ConfigurationException {
		SynchronousConfigurationReadyListener l = new SynchronousConfigurationReadyListener();
		return l.getConfiguration(pluginId, timeOut);
	}

	public IConfiguration getConfiguration(
		String pluginId,
		IConfigurationReadyListener crl)
		throws ConfigurationException {

		IConfiguration pluginConfig;
		IConfiguration pluginMutableConfig;

		if (configMap.containsKey(pluginId)) {
			pluginConfig = (Configuration) configMap.get(pluginId);
			crl.configurationReady(new ConfigurationReadyEvent(pluginConfig));
		} else {
			pluginConfig = new Configuration(pluginId, runtime, crl);
			configMap.put(pluginId, pluginConfig);
		}

		if (mutableConfigMap.containsKey(pluginId)) {
			pluginMutableConfig =
				(IConfiguration) mutableConfigMap.get(pluginId);
			pluginMutableConfig.addConfigurationChangeListener(
				(Configuration) pluginConfig);
		}

		return pluginConfig;

	}

	public IConfiguration getMutableConfiguration(
		String pluginId,
		IConfigurationReadyListener crl)
		throws ConfigurationException {

		IConfiguration pluginMutableConfig;
		IConfiguration pluginConfig;

		if (mutableConfigMap.containsKey(pluginId)) {
			pluginMutableConfig =
				(IConfiguration) mutableConfigMap.get(pluginId);
		} else {
			pluginMutableConfig =
				new MutableConfiguration(pluginId, runtime, crl);
			mutableConfigMap.put(pluginId, pluginMutableConfig);
		}

		if (configMap.containsKey(pluginId)) {
			pluginConfig = (Configuration) configMap.get(pluginId);
			pluginMutableConfig.addConfigurationChangeListener(
				(Configuration) pluginConfig);
		}

		return pluginMutableConfig;

	}

	public void parameterDefinitionQuery(IParameterDefinitionCallListener listener)
		throws ConfigurationException {
		this.parameterDefinitionListener = listener;
		CallRequest request = new CallRequest(GET_PARAETER_DEF_LIST_RPC_KEY);
		ICallResultHandler handler = new ParameterDefinitionListHandler();
		try {
			daoRegistry =
				(IDaoRegistry) runtime.getPluginInstance(
					IDaoRegistry.PLUGIN_ID);
		} catch (PluginStartupException e) {
			throw new ConfigurationException(e);
		} catch (PluginNotFoundException e) {
			throw new ConfigurationException(e);
		}
		try {
			configDAO =
				(IConfigDAO) daoRegistry.getDAO(IConfigDAO.INTERFACE_NAME);
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}

		configDAO.query(request, handler, this);

	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfigDAOListener#callFinished(gov.va.med.hds.cd.config.ConfigDAOEvent)
	 */
	public void callFinished(ConfigDAOEvent configDAOEvent) {
		if (configDAOEvent.getException() != null) {
			parameterDefinitionListener.callFailed(
				new ParameterDefinitionException(
					configDAOEvent.getException()));
		} else {
			parameterDefinitionListener.callSucceeded(
				(List) configDAOEvent.getResultObject());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#shutdown()
	 */
	public void shutdown() throws Exception {
		// Configurator shutting down...
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#startup(IRuntimeContext, IPluginDescriptor)
	 */
	public void startup(IRuntimeContext runtime, IPluginDescriptor plugin)
		throws Exception {

		//Configurator starting up...

		//validate plugin
		if (!plugin.getUniqueIdentifier().equals(IConfigurator.PLUGIN_ID)) {
			throw new InvalidPluginDescriptorException(
				"invalid plugin passed to Configurator: "
					+ plugin.getUniqueIdentifier());
		}
		this.runtime = runtime;
		this.plugin = plugin;
	}

	private class SynchronousConfigurationReadyListener implements IConfigurationReadyListener {

		private IConfiguration configuration;
		private boolean gotResult = false;

		public IConfiguration getConfiguration(String pluginId, int timeOut) throws ConfigurationException {
			Configurator.this.getConfiguration(pluginId, this);
			synchronized(this) {
				if (gotResult == false)
					try {
						wait(timeOut);
					} catch (InterruptedException e) {
						throw new ConfigurationException(e);
					}
			}
			if (configuration == null) throw new ConfigurationException("time out!!!");
			return configuration;
		}

		public void configurationReady(ConfigurationReadyEvent cre) {
			configuration = cre.getConfiguration();
			synchronized(this) {
				gotResult = true;
				notify();
			}
		}

		public void configurationCallFailed(ConfigurationException ce) {
			configuration = null;
			synchronized(this) {
				gotResult = true;
				notify();
			}
		}
	}
}
