package gov.va.med.hds.cd.config.internal;

import java.util.ArrayList;
import java.util.EventListener;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.event.EventListenerList;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.ConfigurationChangeEvent;
import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.ConfigurationReadyEvent;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurationChangeListener;
import gov.va.med.hds.cd.config.IConfigurationReadyListener;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterChangeListener;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.cd.config.NonExistentParameterException;
import gov.va.med.hds.cd.config.ParameterCategoryException;
import gov.va.med.hds.cd.config.ParameterChangeEvent;
import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.security.CouldNotCreateUserInfoException;
import gov.va.med.hds.cd.security.ISecurity;
import gov.va.med.hds.cd.security.IUserInfo;
import gov.va.med.hds.util.mumps.StringUtil;
/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class MutableConfiguration
	implements IConfiguration, TreeModelListener {

	private EventListenerList configurationListeners = new EventListenerList();
	private IConfigDAO configDAO;
	private Map projectionMap = new HashMap();
	private Map indexMap = new HashMap();
	private Map catMap = new HashMap();
	private IRuntimeContext runtime;
	private IDaoRegistry daoRegistry;
	private ISecurity security;
	private IUserInfo userInfo;
	private String pluginID = "";
	private ConfigurationCallHandler mutableConfigurationCallHandler;
	private boolean dirty = false;
	protected IParameterTree mutableParameterTree = null;
	protected IParameterCategory mutableParameterRoot = null;

	MutableConfiguration(
		String pluginID,
		IRuntimeContext runtime,
		IConfigurationReadyListener crl)
		throws ConfigurationException {

		this.runtime = runtime;
		this.pluginID = pluginID;

		try {
			daoRegistry =
				(IDaoRegistry) runtime.getPluginInstance(
					IDaoRegistry.PLUGIN_ID);
		} catch (PluginStartupException e) {
			throw new ConfigurationException(e);
		} catch (PluginNotFoundException e) {
			throw new ConfigurationException(e);
		}
		try {
			security =
				(ISecurity) runtime.getPluginInstance(ISecurity.PLUGIN_ID);
			try {
				userInfo = security.getUserInfo();
			} catch (CouldNotCreateUserInfoException e) {
				throw new ConfigurationException(e);
			}
		} catch (PluginStartupException e) {
			throw new ConfigurationException(e);
		} catch (PluginNotFoundException e) {
			throw new ConfigurationException(e);
		}

		try {
			configDAO =
				(IConfigDAO) daoRegistry.getDAO(IConfigDAO.INTERFACE_NAME);
			configDAO.setConfiguration((IConfiguration) this);
		} catch (Exception e) {
			runtime.getLogger(getClass()).error("Unable to get instance of " + IConfigDAO.class.getName(), e);
		}
		addConfigurationReadyListener(crl);
		List params = new ArrayList();
		params.add(pluginID);
		mutableConfigurationCallHandler = new ConfigurationCallHandler(this);
		configDAO.query(
			new CallRequest("GET MUTABLE CONFIGURATION", params),
			mutableConfigurationCallHandler,
			new MutableConfigurationCallListener());

	}

	/**
	 * Method buildProjection.
	 * @param pCat
	 */
	void buildProjection(IParameterCategory pCat) {

		if (pCat == null)
			return;

		//copy pCat parameters Map to projectionMap
		Iterator keyIt = pCat.parameters().keySet().iterator();
		while (keyIt.hasNext()) {
			String fullKey = (String) keyIt.next();
			String ezKey = StringUtil.piece(fullKey, '^', 1);
			String value = pCat.getParameterValue(fullKey);
			projectionMap.put(ezKey, value);
			indexMap.put(ezKey, fullKey);
			catMap.put(fullKey, pCat);
		}

		/* iterate through pCat children & add their
		 * parameters to projectionMap */
		Iterator scIt = pCat.children().iterator();
		while (scIt.hasNext()) {
			IParameterCategory subCat = (ParameterCategory) scIt.next();
			if (subCat != null) {
				buildProjection(subCat);
			}
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addConfigurationChangeListener(IConfigurationChangeListener)
	 */
	public void addConfigurationChangeListener(IConfigurationChangeListener ccl) {
		configurationListeners.add(IConfigurationChangeListener.class, ccl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeConfigurationChangeListener(IConfigurationChangeListener)
	 */
	public void removeConfigurationChangeListener(IConfigurationChangeListener ccl) {
		configurationListeners.remove(IConfigurationChangeListener.class, ccl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addParameterChangeListener(gov.va.med.hds.cd.config.IParameterChangeListener)
	 */
	public void addParameterChangeListener(IParameterChangeListener pcl) {
		configurationListeners.add(IParameterChangeListener.class, pcl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeParameterChangeListener(gov.va.med.hds.cd.config.IParameterChangeListener)
	 */
	public void removeParameterChangeListener(IParameterChangeListener ccl) {
		configurationListeners.remove(IParameterChangeListener.class, ccl);
	}

	/**
	 * Method fireConfigurationChanged.
	 * @param modParam
	 * @param modParamCat
	 */
	protected void fireConfigurationChanged(
		IParameterCategory subject,
		IParameterCategory object,
		int verb) {

		EventListener[] ccl;

		ccl =
			configurationListeners.getListeners(
				IConfigurationChangeListener.class);

		if (ccl.length == 0) {
			return;
		}

		ConfigurationChangeEvent ccevent =
			new ConfigurationChangeEvent(this, subject, object, verb);

		for (int i = 0; i < ccl.length; ++i) {

			IConfigurationChangeListener listener =
				(IConfigurationChangeListener) ccl[i];
			listener.configurationChanged(ccevent);
		}

	}

	protected void fireParameterChanged(IParameter modParam) {

		EventListener[] pcl;

		pcl =
			configurationListeners.getListeners(IParameterChangeListener.class);

		if (pcl.length == 0) {
			return;
		}

		ParameterChangeEvent pcevent = new ParameterChangeEvent(this, modParam);

		for (int i = 0; i < pcl.length; ++i) {

			IParameterChangeListener listener =
				(IParameterChangeListener) pcl[i];
			listener.parameterChanged(pcevent);
		}

	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationChangeListener)
	 */
	public void addConfigurationReadyListener(IConfigurationReadyListener crl) {
		configurationListeners.add(IConfigurationReadyListener.class, crl);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationChangeListener)
	 */
	public void removeConfigurationReadyListener(IConfigurationReadyListener crl) {
		configurationListeners.remove(IConfigurationReadyListener.class, crl);
	}

	/**
	 * Method fireConfigurationReady.
	 */
	protected void fireConfigurationReady() {
		EventListener[] crl;

		crl =
			configurationListeners.getListeners(
				IConfigurationReadyListener.class);

		if (crl.length == 0) {
			return;
		}

		ConfigurationReadyEvent crevent = new ConfigurationReadyEvent(this);

		for (int i = 0; i < crl.length; ++i) {

			IConfigurationReadyListener listener =
				(IConfigurationReadyListener) crl[i];
			listener.configurationReady(crevent);
		}
	}

	/**
	 * Returns the parameterTree.
	 * @return IParameterTree
	 */
	public IParameterTree getParameterTree() {
		if (mutableParameterTree == null) {
			mutableParameterRoot = getParameterRoot();
			mutableParameterTree = new ParameterTree(mutableParameterRoot);
			((ParameterTree) mutableParameterTree).addTreeModelListener(this);
		}
		return mutableParameterTree;
	}

	/**
	 * Returns the parameterRoot.
	 * @return IParameterCategory
	 */
	public IParameterCategory getParameterRoot() {
		if (mutableParameterRoot == null) {
			mutableParameterRoot =
				new ParameterCategory("Retrieving data...", this);
			mutableParameterTree = new ParameterTree(mutableParameterRoot);
		}
		return mutableParameterRoot;
	}

	/**
	 * Method commitChanges.
	 * @throws ConfigDAOException
	 */
	public void commitChanges() throws ConfigDAOException {
		configDAO.putParameterTree(mutableParameterTree);
	}

	/**
	 * @param pc
	 * @throws ConfigurationException
	 */
	public void removeParameterCategory(IParameterCategory pc)
		throws ConfigurationException {
		IParameterCategory parentCategory = pc.getParent();
		removeParameterCategory(pc, parentCategory);
	}

	public void removeParameterCategory(
		IParameterCategory pc,
		IParameterCategory parentCategory)
		throws ConfigurationException {
		if (pc.isPackageRoot()) {
			throw new ConfigurationException("You may not remove the root category...");
		}

		if (!pc.isLeaf()) {
			try {
				configDAO.removeParameterCategory(pc, parentCategory);
				fireConfigurationChanged(
					pc,
					parentCategory,
					ConfigurationChangeEvent.CATEGORY_REMOVED);
			} catch (ParameterCategoryException e) {
				throw new ConfigurationException(
					"Configuration Exception: " + e.getMessage());
			} catch (ConfigDAOException e) {
				throw new ConfigurationException(
					"Configuration Exception: " + e.getMessage());
			}
		} else {
			ArrayList parameterList = pc.parametersAsArrayList();
			Iterator plIt = parameterList.iterator();

			if (plIt.hasNext()) {
				IParameter par = (IParameter) plIt.next();
				try {
					configDAO.removeParameterFromCategory(par, parentCategory);
				} catch (ConfigDAOException e) {
					throw new ConfigurationException(
						"Configuration Exception: " + e.getMessage());
				}
			}
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getParameterValue(String)
	 */
	public String getParameterValue(String ezKey) {
		return (String) projectionMap.get(ezKey);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#putParameter(String, String)
	 */
	public void putParameter(String ezKey, String value)
		throws NonExistentParameterException, ConfigurationException {
		projectionMap.put(ezKey, value);
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException("Parameter not defined on Server");
		}
		pCat.putParameter(fullKey, value);
		try {
			configDAO.putParameter(fullKey, value);
			IParameter modParam = getParameter(ezKey);
			fireParameterChanged(modParam);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		} catch (NonExistentParameterException nPe) {
			throw new ConfigurationException(
				"Configuration Exception: " + nPe.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#putParameter(IParameter)
	 */
	public void putParameter(IParameter parameter)
		throws ConfigurationException {
		IParameterCategory pCat =
			getCategoryForParameter(
				StringUtil.piece(parameter.getKey(), '^', 1));
		if (!(pCat instanceof IParameterCategory)) {
			throw new ConfigurationException(
				"Category not found for Parameter: " + parameter.getName());
		}

		pCat.putParameter(parameter);

		try {
			configDAO.putParameter(parameter.getKey(), parameter.getValue());
			fireParameterChanged(parameter);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getCategoryForParameter(String)
	 */
	public IParameterCategory getCategoryForParameter(String ezKey) {
		String fullKey = (String) indexMap.get(ezKey);
		return (ParameterCategory) catMap.get(fullKey);
	}

	/**
	 * Method getListenerCount.
	 */
	public int getListenerCount(Class type) {
		return configurationListeners.getListenerCount(type);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getParameter(String)
	 */
	public IParameter getParameter(String ezKey)
		throws NonExistentParameterException {
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException("Parameter not defined on Server");
		}
		return pCat.getParameter(fullKey);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#updateParameterCategory(IParameterCategory)
	 */
	public void updateParameterCategory(IParameterCategory parameterCategory)
		throws ConfigDAOException {
		configDAO.updateParameterCategory(parameterCategory);
	}

	public void insertParameterCategory(
		IParameterCategory pc,
		IParameterCategory parent)
		throws ConfigDAOException {
		ArrayList parameterList;
		if (!pc.isLeaf()) {
			configDAO.insertParameterCategory(pc, parent);
		} else {
			parameterList = pc.parametersAsArrayList();
			Iterator plIt = parameterList.iterator();

			if (plIt.hasNext()) {
				IParameter parameter = (IParameter) plIt.next();
				System.out.println("Adding " + parameter + " to " + parent);
				configDAO.addParameterToCategory(parameter, parent);
			}
		}
	}

	/**
	 * Method getUserInfo. Used internally in Configurator to reference user
	 * demographic information.
	 * @return IUserInfo
	 */
	public IUserInfo getUserInfo() {
		return userInfo;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getMutableParameterRoot()
	 */
	public IParameterCategory getMutableParameterRoot()
		throws ConfigurationException {
		if (mutableParameterRoot == null) {
			mutableParameterRoot =
				new ParameterCategory("Retrieving data...", this);
			mutableParameterTree = new ParameterTree(mutableParameterRoot);
		}
		return mutableParameterRoot;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getMutableParameterTree()
	 */
	public IParameterTree getMutableParameterTree()
		throws ConfigurationException {
		if (mutableParameterTree == null) {
			mutableParameterRoot = getMutableParameterRoot();
			mutableParameterTree = new ParameterTree(mutableParameterRoot);
		}
		return mutableParameterTree;
	}

	class MutableConfigurationCallListener implements IConfigDAOListener {

		/**
		 * @see gov.va.med.hds.cd.config.IConfigDAOListener#callFinished(gov.va.med.hds.cd.config.ConfigDAOEvent)
		 */
		public void callFinished(ConfigDAOEvent configDAOEvent) {
			if (configDAOEvent.getException() != null) {
				runtime.getLogger(this.getClass()).equals(configDAOEvent.getException());
				//TODO: consider what to do next with exception
			} else if (configDAOEvent.getResultObject() != null) {
				mutableParameterRoot =
					(IParameterCategory) configDAOEvent.getResultObject();
				mutableParameterTree = new ParameterTree(mutableParameterRoot);
				buildProjection(mutableParameterRoot);
				fireConfigurationReady();
			}
		}

	}
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return pluginID;
	}

	/* (non-Javadoc)
	 * @see javax.swing.event.TreeModelListener#treeNodesChanged(javax.swing.event.TreeModelEvent)
	 */
	public void treeNodesChanged(TreeModelEvent e) {
		// TODO Auto-generated method stub
		System.out.println("In MutableConfiguration.treeNodesChanged()...");
		IParameterCategory category =
			(IParameterCategory) (e.getTreePath().getLastPathComponent());

		try {
			int index = e.getChildIndices()[0];
			category = (IParameterCategory) (category.getChildAt(index));
			if (category.getId() > 0){
				category.setConfiguration(this);
				category.persist();
			}
		} catch (NullPointerException nPe) {
			// TODO add functional catch block
			nPe.printStackTrace();
		} catch (ConfigDAOException cDe) {
			// TODO Auto-generated catch block
			cDe.printStackTrace();
		}
	}

	/* (non-Javadoc)
	 * @see javax.swing.event.TreeModelListener#treeNodesInserted(javax.swing.event.TreeModelEvent)
	 */
	public void treeNodesInserted(TreeModelEvent e) {
		System.out.println("In MutableConfiguration.treeNodesInserted()...");
		IParameterCategory parent =
			(IParameterCategory) (e.getTreePath().getLastPathComponent());
		IParameterCategory category = null;
		Object[] children = e.getChildren();

		for (int i = 0; i < children.length; i++) {
			System.out.println(
				"Inserting "
					+ ((IParameterCategory) children[i]).toString()
					+ " into "
					+ parent.toString());
			try {
				insertParameterCategory(
					((IParameterCategory) children[i]),
					parent);
			} catch (ConfigDAOException ce) {
				//TODO Add functional catch block
				ce.printStackTrace();
			}
		}
	}

	/* (non-Javadoc)
	 * @see javax.swing.event.TreeModelListener#treeNodesRemoved(javax.swing.event.TreeModelEvent)
	 */
	public void treeNodesRemoved(TreeModelEvent e) {
		System.out.println("In MutableConfiguration.treeNodesRemoved()...");
		System.err.println("TreeModelEvent e = " + e);
		IParameterCategory parent =
			(IParameterCategory) (e.getTreePath().getLastPathComponent());
		Object[] children = e.getChildren();

		for (int i = 0; i < children.length; i++) {
			System.out.println(
				"Removing "
					+ ((IParameterCategory) children[i]).toString()
					+ " from "
					+ parent.toString());
			try {
				removeParameterCategory(
					((IParameterCategory) children[i]),
					parent);
			} catch (ConfigurationException ce) {
				//TODO Add functional catch block
				ce.printStackTrace();
			}
		}
	}

	/* (non-Javadoc)
	 * @see javax.swing.event.TreeModelListener#treeStructureChanged(javax.swing.event.TreeModelEvent)
	 */
	public void treeStructureChanged(TreeModelEvent e) {
		System.out.println("In MutableConfiguration.treeStructureChanged()...");
		System.out.println("TreeModelEvent e = " + e);
		IParameterCategory category =
			(IParameterCategory) (e.getTreePath().getLastPathComponent());

		/*
		 * If the event lists children, then the changed
		 * node is the child of the node we've already
		 * gotten.  Otherwise, the changed node and the
		 * specified node are the same.
		 */
		try {
			int index = e.getChildIndices()[0];
			category = (IParameterCategory) (category.getChildAt(index));
			if (dirty) {
				category.setConfiguration(this);
				category.persist();
			}
		} catch (NullPointerException nPe) {
			// TODO implement functional catch clause
			//			nPe.printStackTrace(); 
		} catch (ConfigDAOException cDe) {
			// TODO implement functional catch clause
			cDe.printStackTrace();
		}

		System.out.println("Current node: " + category.toString());
	}

}
