package gov.va.med.hds.cd.config.internal;

import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterDefinition;
import gov.va.med.hds.util.mumps.StringUtil;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class Parameter implements IParameter {
	private IParameterCategory category = null;
	private IConfiguration configuration;
	private String defaultValue;
	private String entity;
	private String extEntity;
	private String fullName;

	private long id;
	private String initialValue;
	private String instance;
	private String key;
	private boolean multiValued;
	private String name;
	private String value;
	private boolean wordProcessing;

	public Parameter(IConfiguration configuration, IParameterCategory pCat) {
		super();
		this.category = pCat;
		this.configuration = configuration;
	}

	/**
	 * @param pd
	 */
	public Parameter(IParameterDefinition pd) {
		super();
		this.name = pd.getDisplayText();
		this.fullName = pd.getFullName();
		this.id = pd.getId();
		this.multiValued = pd.isMultiValued();
	}

	/**
	 * Constructor Parameter.
	 * @param string
	 */
	public Parameter(
		String newName,
		IConfiguration configuration,
		IParameterCategory pCat) {
		super();
		name = newName;
		this.category = pCat;
		this.configuration = configuration;
	}

	/**
	 * Returns the category.
	 * @return IParameterCategory
	 */
	public IParameterCategory getCategory() {
		return category;
	}

	/**
	 * Returns the configuration.
	 * @return IConfiguration
	 */
	public IConfiguration getConfiguration() {
		return configuration;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getDefaultValue()
	 */
	public String getDefaultValue() {
		return defaultValue;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getEntity()
	 */
	public String getEntity() {
		return entity;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getExtEntity()
	 */
	public String getExtEntity() {
		return extEntity;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getFullName()
	 */
	public String getFullName() {
		return fullName;
	}

	/**
	 * Returns the id.
	 * @return long
	 */
	public long getId() {
		return id;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getInitialValue()
	 */
	public String getInitialValue() {
		return initialValue;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#getInstance()
	 */
	public String getInstance() {
		return instance;
	}

	/**
	 * Returns the key.
	 * @return String
	 */
	public String getKey() {
		return key;
	}

	/**
	 * Returns the name.
	 * @return String
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the value.
	 * @return String
	 */
	public String getValue() {
		return value;
	}

	/**
	 * Returns the multiValued.
	 * @return boolean
	 */
	public boolean isMultiValued() {
		return multiValued;
	}

	/**
	 * Returns the wordProcessing.
	 * @return boolean
	 */
	public boolean isWordProcessing() {
		return wordProcessing;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#persist()
	 */
	public void persist() throws ConfigurationException {
		if (configuration != null) {
			configuration.putParameter((IParameter) this);
		}
	}

	/**
	 * Sets the category.
	 * @param category The category to set
	 */
	public void setCategory(IParameterCategory category) {
		this.category = category;
	}

	/**
	 * Sets the configuration.
	 * @param configuration The configuration to set
	 */
	public void setConfiguration(IConfiguration configuration) {
		this.configuration = configuration;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setDefaultValue(String)
	 */
	public void setDefaultValue(String defaultValue) {
		this.defaultValue = defaultValue;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setEntity(String)
	 */
	public void setEntity(String entity) {
		this.entity = entity;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setExtEntity(String)
	 */
	public void setExtEntity(String extEntity) {
		this.extEntity = extEntity;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setFullName(String)
	 */
	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	/**
	 * Sets the id.
	 * @param id The id to set
	 */
	public void setId(long id) {
		this.id = id;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setInitialValue(String)
	 */
	public void setInitialValue(String initialValue) {
		this.initialValue = initialValue;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameter#setInstance(String)
	 */
	public void setInstance(String instance) {
		this.instance = instance;
	}

	/**
	 * Sets the key.
	 * @param key The key to set
	 */
	public void setKey(String key) {
		this.key = key;
		this.entity = StringUtil.piece(key, '^', 2);
		this.instance = StringUtil.piece(key, '^', 4);
	}

	/**
	 * Sets the multiValued.
	 * @param multiValued The multiValued to set
	 */
	public void setMultiValued(boolean multiValued) {
		this.multiValued = multiValued;
	}

	/**
	 * Sets the name.
	 * @param name The name to set
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Sets the value.
	 * @param value The value to set
	 */
	public void setValue(String value) {
		this.value = value;
	}

	/**
	 * Sets the wordProcessing.
	 * @param wordProcessing The wordProcessing to set
	 */
	public void setWordProcessing(boolean wordProcessing) {
		this.wordProcessing = wordProcessing;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return name;
	}

}
