package gov.va.med.hds.cd.config.internal;

import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterDefinition;
import gov.va.med.hds.cd.config.ParameterCategoryException;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.dom4j.Node;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom4j.Dom4jXPath;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class ParameterCategory implements IParameterCategory, Transferable, Comparable {

	static DataFlavor flavors[] = { CATEGORY_FLAVOR };
	private List children = new ArrayList();
	private IConfiguration configuration = null;
	private String fullName;
	private long id;
	private boolean leaf = false;
	private boolean markedForRemoval = false;

	private String name;
	private boolean packageRoot;
	private Map parameters = new TreeMap();
	private IParameterCategory parent;
	private long parentId;
	private boolean persistent = false;
	private String pluginId;
	private String preferencePage;
	private Map subCatIndex = new HashMap();
	
	/**
	 * Constructor ParameterCategory creates a "deep copy" of the instance
	 * referenced by <code>original</code>
	 * @param original
	 */
	public ParameterCategory(IParameterCategory original) {
		name = "Copy of " + original.getName();
		fullName = original.getFullName() + " COPY";
		pluginId = original.getPluginId();
		id = 0;
		parent = original.getParent();
		parentId = original.getParentId();
		packageRoot = false;
		preferencePage = original.getPreferencePage();
		markedForRemoval = false;
		persistent = false;
		configuration = original.getConfiguration();
		leaf = original.isLeaf();
		
		Iterator childIter = original.children().iterator();
		while (childIter.hasNext()) {
			IParameterCategory originalChild = (IParameterCategory) childIter.next();
			IParameterCategory copyChild = new ParameterCategory(originalChild);
			insert(copyChild);
		}
		 
//		parameters = new TreeMap();
	}

	public ParameterCategory(IParameterDefinition pDef) {
		this.name = pDef.getDisplayText();
		this.fullName = pDef.getFullName();
		this.packageRoot = false;
		this.leaf = true;
		this.id = 0L;
	}

	/**
	 * Constructor for root ParameterCategory.
	 * @param node
	 */
	public ParameterCategory(Node node, IConfiguration config)
		throws ParameterCategoryException {

		this.parent = null;
		this.configuration = config;

		if (node == null)
			throw new ParameterCategoryException("xml node parameterCategory should not be null!");

		try {
			id =
				Long.parseLong(
					((Node) (new Dom4jXPath("id")).selectSingleNode(node))
						.getStringValue());
			fullName =
				((Node) (new Dom4jXPath("fullName")).selectSingleNode(node))
					.getStringValue();
			pluginId =
				((Node) (new Dom4jXPath("pluginId")).selectSingleNode(node))
					.getStringValue();
			name =
				((Node) (new Dom4jXPath("name")).selectSingleNode(node))
					.getStringValue();
			parentId =
				Long.parseLong(
					((Node) (new Dom4jXPath("parentId"))
						.selectSingleNode(node))
						.getStringValue());
			packageRoot =
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("packageRoot"))
						.selectSingleNode(node))
						.getStringValue());
			preferencePage =
				((Node) (new Dom4jXPath("preferencePageId"))
					.selectSingleNode(node))
					.getStringValue();
			Node leafNode = (Node) (new Dom4jXPath("leaf")).selectSingleNode(node);
			if (leafNode != null) {
				leaf = "true".equalsIgnoreCase(leafNode.getStringValue());
			}
			
			persistent = true;
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"couldn't find name element",
				e);
		}

		buildParameters(node);

		buildChildren(node);

	}

	/**
	 * Constructor ParameterCategory.
	 * @param node
	 */
	public ParameterCategory(
		Node node,
		IParameterCategory parent,
		IConfiguration config)
		throws ParameterCategoryException {

		this.parent = parent;
		this.configuration = config;

		if (node == null)
			throw new ParameterCategoryException("xml node parameterCategory should not be null!");

		try {
			id =
				Long.parseLong(
					((Node) (new Dom4jXPath("id")).selectSingleNode(node))
						.getStringValue());
			fullName =
				((Node) (new Dom4jXPath("fullName")).selectSingleNode(node))
					.getStringValue();
			pluginId =
				((Node) (new Dom4jXPath("pluginId")).selectSingleNode(node))
					.getStringValue();
			name =
				((Node) (new Dom4jXPath("name")).selectSingleNode(node))
					.getStringValue();
			parentId =
				Long.parseLong(
					((Node) (new Dom4jXPath("parentId"))
						.selectSingleNode(node))
						.getStringValue());
			packageRoot =
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("packageRoot"))
						.selectSingleNode(node))
						.getStringValue());
			preferencePage =
				((Node) (new Dom4jXPath("preferencePageId"))
					.selectSingleNode(node))
					.getStringValue();
			Node leafNode = (Node) (new Dom4jXPath("leaf")).selectSingleNode(node);
			if (leafNode != null) {
				leaf = "true".equalsIgnoreCase(leafNode.getStringValue());
			}

			persistent = true;
			if (parent != null) {
				((ParameterCategory) parent).subCatIndex.put(fullName, this);
			}
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"couldn't find name element",
				e);
		}

		buildParameters(node);

		buildChildren(node);

	}

	public ParameterCategory(String name, IConfiguration config) {
		this.configuration = config;
		this.name = name;
	}

	public void addParameter(IParameter parameter) {
		parameters.put(parameter.getKey(), parameter);
	}

	private void addParameter(Node node) throws ParameterCategoryException {

		IParameter parameter = new Parameter(configuration, this);

		try {
			parameter.setId(
				Long.parseLong(
					((Node) (new Dom4jXPath("id")).selectSingleNode(node))
						.getStringValue()));
			parameter.setFullName(
				((Node) (new Dom4jXPath("fullName")).selectSingleNode(node))
					.getStringValue());
			parameter.setName(
				((Node) (new Dom4jXPath("name")).selectSingleNode(node))
					.getStringValue());
			parameter.setMultiValued(
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("multiValued"))
						.selectSingleNode(node))
						.getStringValue()));
			parameter.setWordProcessing(
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("wordProcessing"))
						.selectSingleNode(node))
						.getStringValue()));
			parameter.setKey(
				((Node) (new Dom4jXPath("key")).selectSingleNode(node))
					.getStringValue());

			Node extEntityNode =
				(Node) (new Dom4jXPath("extEntity")).selectSingleNode(node);
			if (extEntityNode != null) {
				parameter.setExtEntity(extEntityNode.getStringValue());
			}

			parameter.setValue(
				((Node) (new Dom4jXPath("value")).selectSingleNode(node))
					.getStringValue());
			// set initialValue of parameter
			parameter.setInitialValue(parameter.getValue());
			parameter.setDefaultValue(
				((Node) (new Dom4jXPath("defaultValue"))
					.selectSingleNode(node))
					.getStringValue());
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"couldn't find name element",
				e);
		}
		parameters.put(parameter.getKey(), parameter);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#addParameter(key, value)
	 */
	public void addParameter(String key, IParameter parameter) {
		parameters.put(key, parameter);
	}

	private void buildChildren(Node node) throws ParameterCategoryException {
		XPath xpath;
		try {
			xpath = (XPath) new Dom4jXPath("children/parameterCategory");
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"JaxenException at parameters/parameter.",
				e);
		}
		List results;
		try {
			results = xpath.selectNodes(node);
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"JaxenException at selectNodes(node).",
				e);
		}
		Iterator resultIter = results.iterator();

		if (!resultIter.hasNext())
			return;

		while (resultIter.hasNext()) {
			Object object = resultIter.next();
			if (object instanceof Node) {
				insert(
					new ParameterCategory(
						(Node) object,
						this,
						this.configuration));
			}
		}
	}

	private void buildParameters(Node node) throws ParameterCategoryException {
		XPath xpath;
		try {
			xpath = (XPath) new Dom4jXPath("parameters/parameter");
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"JaxenException at parameters/parameter.",
				e);
		}
		List results;
		try {
			results = xpath.selectNodes(node);
		} catch (JaxenException e) {
			throw new ParameterCategoryException(
				"JaxenException at selectNodes(node).",
				e);
		}
		Iterator resultIter = results.iterator();

		if (!resultIter.hasNext())
			return;

		while (resultIter.hasNext()) {
			Object object = resultIter.next();
			if (object instanceof Node) {
				addParameter((Node) object);
			}
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#children()
	 */
	public List children() {
		return Collections.unmodifiableList(children);
	}
	/**
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o) {
		if (!(o instanceof ParameterCategory)) {
			return 0;
		}
		return name.compareTo(((ParameterCategory) o).toString());
	}

	public boolean getAllowsChildren() {
		return !leaf;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getChild(String)
	 */
	public IParameterCategory getChild(String fullName) {
		return (IParameterCategory) subCatIndex.get(fullName);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getChildAt(int)
	 */
	public IParameterCategory getChildAt(int childIndex) {
		return (IParameterCategory) children.get(childIndex);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getChildCount()
	 */
	public int getChildCount() {
		return children.size();
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getConfiguration()
	 */
	public IConfiguration getConfiguration() {
		return configuration;
	}

	/**
	 * Returns the fullName.
	 * @return String
	 */
	public String getFullName() {
		return fullName;
	}
	/**
	 * Returns the id.
	 * @return long
	 */
	public long getId() {
		return id;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getIndex(IParameterCategory)
	 */
	public int getIndex(IParameterCategory node) {
		return children.indexOf(node);
	}

	/**
	 * Returns the name.
	 * @return String
	 */
	public String getName() {
		return name;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getParameter(String)
	 */
	public IParameter getParameter(String key) {
		return (IParameter) parameters.get(key);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#addParameter(key, value)
	 */
	public String getParameterValue(String key) {
		return ((IParameter) (parameters.get(key))).getValue();
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#getParent()
	 */
	public IParameterCategory getParent() {
		return parent;
	}

	/**
	 * Returns the parentId.
	 * @return long
	 */
	public long getParentId() {
		return parentId;
	}
	/**
	  * Returns the path from the root, to get to this node.  The last
	  * element in the path is this node.
	  *
	  * @return an array of TreeNode objects giving the path, where the
	  *         first element in the path is the root and the last
	  *         element is this node.
	  */
	public IParameterCategory[] getPath() {
		return getPathToRoot(this, 0);
	}

	/**
	 * Builds the parents of node up to and including the root node,
	 * where the original node is the last element in the returned array.
	 * The length of the returned array gives the node's depth in the
	 * tree.
	 * 
	 * @param aNode  the TreeNode to get the path for
	 * @param depth  an int giving the number of steps already taken towards
	 *        the root (on recursive calls), used to size the returned array
	 * @return an array of TreeNodes giving the path from the root to the
	 *         specified node 
	 */
	protected IParameterCategory[] getPathToRoot(
		IParameterCategory aNode,
		int depth) {
		IParameterCategory[] retNodes;

		/* Check for null, in case someone passed in a null node, or
		   they passed in an element that isn't rooted at root. */
		if (aNode == null) {
			if (depth == 0)
				return null;
			else
				retNodes = new IParameterCategory[depth];
		} else {
			depth++;
			retNodes = getPathToRoot(aNode.getParent(), depth);
			retNodes[retNodes.length - depth] = aNode;
		}
		return retNodes;
	}

	/**
	 * Returns the pluginId.
	 * @return String
	 */
	public String getPluginId() {
		return pluginId;
	}

	/**
	 * Returns the preferencePage.
	 * @return PreferencePanel
	 */
	public String getPreferencePage() {
		return (preferencePage == null ? "" : preferencePage);
	}
	/**
	 * @see java.awt.datatransfer.Transferable#getTransferData(java.awt.datatransfer.DataFlavor)
	 */
	public Object getTransferData(DataFlavor flavor)
		throws UnsupportedFlavorException, IOException {
		if (flavor.equals(CATEGORY_FLAVOR)) {
			return this;
		}
		throw new UnsupportedFlavorException(flavor);
	}
	/**
	 * @see java.awt.datatransfer.Transferable#getTransferDataFlavors()
	 */
	public DataFlavor[] getTransferDataFlavors() {
		return flavors;
	}

	/**
	 * overloaded insert (for when you just can't know which child you're inserting!
	 **/
	public void insert(IParameterCategory child) {
		children.add(child);
		child.setParent(this);
		child.setParentId(this.getId());
		Collections.sort(children);
		if (child.getFullName() != null) {
			subCatIndex.put(child.getFullName(), child);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#insert(IParameterCategory, int)
	 */
	public void insert(IParameterCategory child, int index) {
		children.add(index, child);
		child.setParent(this);
		child.setParentId(this.getId());
		Collections.sort(children);
		if (child.getFullName() != null) {
			subCatIndex.put(child.getFullName(), child);
		} 
	}
	/**
	 * @see java.awt.datatransfer.Transferable#isDataFlavorSupported(java.awt.datatransfer.DataFlavor)
	 */
	public boolean isDataFlavorSupported(DataFlavor flavor) {
		return flavor.equals(CATEGORY_FLAVOR);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#isLeaf()
	 */
	public boolean isLeaf() {
		return leaf;
	}

	/**
	 * Returns the markedForRemoval.
	 * @return boolean
	 */
	public boolean isMarkedForRemoval() {
		return markedForRemoval;
	}

	/**
	 * Returns the packageRoot.
	 * @return boolean
	 */
	public boolean isPackageRoot() {
		return packageRoot;
	}

	/**
	 * Returns the persistent.
	 * @return boolean
	 */
	public boolean isPersistent() {
		return persistent;
	}

	private void markChildrenToRemove(IParameterCategory paramTreeNode) {

		Iterator sPit = ((List) paramTreeNode.children()).iterator();

		while (sPit.hasNext()) {
			paramTreeNode.markToRemove((IParameterCategory) sPit.next());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#markToRemove(int)
	 */
	public void markToRemove(int index) {
		IParameterCategory child = (IParameterCategory) children.get(index);
		child.setMarkedForRemoval(true);
		markChildrenToRemove(child);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#markToRemove(IParameterCategory)
	 */
	public void markToRemove(IParameterCategory node) {

		if (node == null) {
			throw new IllegalArgumentException("argument is null");
		}

		if (!children.contains(node)) {
			throw new IllegalArgumentException("argument is not a child");
		}

		markToRemove(getIndex(node)); // linear search

	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#parameters()
	 */
	public Map parameters() {
		return parameters;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#parameters()
	 */
	public ArrayList parametersAsArrayList() {

		ArrayList paramList = new ArrayList();

		Iterator pIt = parameters.keySet().iterator();

		while (pIt.hasNext()) {
			paramList.add(parameters.get(pIt.next()));
		}

		return paramList;

	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#persist()
	 */
	public void persist() throws ConfigDAOException {
		if (configuration != null) {
			configuration.updateParameterCategory((IParameterCategory) this);
			if (!this.isPersistent() && !this.isMarkedForRemoval()) {
				setPersistent(true);				
			}
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#putParameter(IParameter)
	 */
	public void putParameter(IParameter parameter) {
		parameters.put(parameter.getKey(), parameter);
	}

	public void putParameter(String key, String value) {
		IParameter parameter = (IParameter) parameters.get(key);
		parameter.setValue(value);
		parameters.put(key, parameter);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#remove(int)
	 */
	public void remove(int index) {
		IParameterCategory child = (IParameterCategory) children.get(index);
		children.remove(index);
		child.setParent(null);
		Collections.sort(children);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#remove(IParameterCategory)
	 */
	public void remove(IParameterCategory node) {

		if (node == null) {
			throw new IllegalArgumentException("argument is null");
		}

		if (!children.contains(node)) {
			throw new IllegalArgumentException("argument is not a child");
		}

		remove(getIndex(node)); // linear search

	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#removeFromParent()
	 */
	public void removeFromParent() {
		if (parent != null) {
			((ParameterCategory) parent).subCatIndex.remove(this.fullName);
			parent.remove(this);
			parent = null;
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#remove(int)
	 */
	public void removeParameter(String key) {
		if (parameters.containsKey(key)) {
			parameters.remove(key);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#setConfiguration(IConfiguration)
	 */
	public void setConfiguration(IConfiguration configuration) {
		this.configuration = configuration;
	}

	/**
	 * Sets the fullName.
	 * @param fullName The fullName to set
	 */
	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	/**
	 * Sets the id.
	 * @param id The id to set
	 */
	public void setId(long id) {
		this.id = id;
	}

	/**
	 * Sets the leaf.
	 * @param leaf The leaf to set
	 */
	public void setLeaf(boolean isLeaf) {
		this.leaf = isLeaf;
	}

	/**
	 * Sets the markedForRemoval.
	 * @param markedForRemoval The markedForRemoval to set
	 */
	public void setMarkedForRemoval(boolean markedForRemoval) {
		this.markedForRemoval = markedForRemoval;
	}

	/**
	 * Sets the name.
	 * @param name The name to set
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Sets the packageRoot.
	 * @param packageRoot The packageRoot to set
	 */
	public void setPackageRoot(boolean packageRoot) {
		this.packageRoot = packageRoot;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IParameterCategory#setParent(IParameterCategory)
	 */
	public void setParent(IParameterCategory newParent) {
		parent = newParent;
	}

	/**
	 * Sets the parentId.
	 * @param parentId The parentId to set
	 */
	public void setParentId(long parentId) {
		this.parentId = parentId;
	}

	/**
	 * Sets the persistent.
	 * @param persistent The persistent to set
	 */
	public void setPersistent(boolean persistent) {
		this.persistent = persistent;
	}

	/**
	 * Sets the pluginId.
	 * @param pluginId The pluginId to set
	 */
	public void setPluginId(String pluginId) {
		this.pluginId = pluginId;
	}

	/**
	 * Sets the preferencePage.
	 * @param preferencePage The preferencePage to set
	 */
	public void setPreferencePage(String preferencePage) {
		this.preferencePage = preferencePage;
	}

	public String toString() {
		return name;
	}

	public void updateParameter(
		String oldKey,
		String newKey,
		String newValue) {
		IParameter parameter = (IParameter) parameters.get(oldKey);
		parameter.setKey(newKey);
		parameter.setValue(newValue);

		parameters.remove(oldKey);
		parameters.put(newKey, parameter);
	}

}
