package gov.va.med.hds.cd.config.internal;

import gov.va.med.hds.cd.config.IEntity;
import gov.va.med.hds.cd.config.IParameterDefinition;
import gov.va.med.hds.cd.config.ParameterDefinitionException;
import gov.va.med.hds.util.mumps.StringUtil;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.dom4j.Node;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom4j.Dom4jXPath;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class ParameterDefinition implements IParameterDefinition {
	private long id;
	private String fullName;
	private String displayText;
	private boolean multiValued;
	private String instanceTerm;
	private String valueTerm;
	private boolean prohibitEditing;
	private String valueType;
	private String valueDomain;
	private String valueHelp;
	private String valueValidationCode;
	private String valueScreenCode;
	private String instanceType;
	private String instanceDomain;
	private String instanceHelp;
	private String instanceValidationCode;
	private String instanceScreenCode;
	private String description;
	private List entities = new ArrayList();
	private List keywords = new ArrayList();
	private Map entityClassMap = new HashMap();

	public ParameterDefinition() {
		initializeEntityMap();
	}

	public ParameterDefinition(long id, String fullName) {
		this.id = id;
		this.fullName = fullName;
		initializeEntityMap();
	}
	
	public ParameterDefinition(String listString) throws ParameterDefinitionException {
		initializeEntityMap();
		this.id = Long.parseLong(StringUtil.piece(listString,'^',1));
		this.fullName = StringUtil.piece(listString,'^',2);
		this.displayText = StringUtil.piece(listString,'^',3);
		buildEntities(StringUtil.piece(listString,'^',4));
	}

	public ParameterDefinition(Node node) throws ParameterDefinitionException {
		
		initializeEntityMap();
		
		if (node == null)
			throw new ParameterDefinitionException("xml node parameterDefinition should not be null!");

		try {
			id =
				Long.parseLong(
					((Node) (new Dom4jXPath("id")).selectSingleNode(node))
						.getStringValue());
			fullName =
				((Node) (new Dom4jXPath("name")).selectSingleNode(node))
					.getStringValue();
			displayText =
				((Node) (new Dom4jXPath("displayText")).selectSingleNode(node))
					.getStringValue();
			multiValued =
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("multipleValued"))
						.selectSingleNode(node))
						.getStringValue());
			instanceTerm =
				((Node) (new Dom4jXPath("instanceTerm"))
					.selectSingleNode(node))
					.getStringValue();
			valueTerm =
				((Node) (new Dom4jXPath("valueTerm")).selectSingleNode(node))
					.getStringValue();
			prohibitEditing =
				"true".equalsIgnoreCase(
					((Node) (new Dom4jXPath("prohibitEditing"))
						.selectSingleNode(node))
						.getStringValue());
			valueType =
				((Node) (new Dom4jXPath("valueDataType"))
					.selectSingleNode(node))
					.getStringValue();
			valueDomain =
				((Node) (new Dom4jXPath("valueDomain")).selectSingleNode(node))
					.getStringValue();
			valueHelp =
				((Node) (new Dom4jXPath("valueHelp")).selectSingleNode(node))
					.getStringValue();
			valueValidationCode =
				((Node) (new Dom4jXPath("valueValidationCode"))
					.selectSingleNode(node))
					.getStringValue();
			valueScreenCode =
				((Node) (new Dom4jXPath("valueScreenCode"))
					.selectSingleNode(node))
					.getStringValue();
			instanceType =
				((Node) (new Dom4jXPath("instanceDataType"))
					.selectSingleNode(node))
					.getStringValue();
			instanceDomain =
				((Node) (new Dom4jXPath("instanceDomain"))
					.selectSingleNode(node))
					.getStringValue();
			instanceHelp =
				((Node) (new Dom4jXPath("instanceHelp"))
					.selectSingleNode(node))
					.getStringValue();
			instanceValidationCode =
				((Node) (new Dom4jXPath("instanceValidationCode"))
					.selectSingleNode(node))
					.getStringValue();
			instanceScreenCode =
				((Node) (new Dom4jXPath("instanceScreenCode"))
					.selectSingleNode(node))
					.getStringValue();
			description =
				(((Node) (new Dom4jXPath("description")).selectSingleNode(node))
					.getStringValue()).trim();
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(
				"couldn't find name element",
				e);
		}

		buildEntities(node);
		buildKeywords(node);
	}

	private void initializeEntityMap() {
        entityClassMap.put("CLASS", ClassEntity.class);
        entityClassMap.put("DEVICE", DeviceEntity.class);
        entityClassMap.put("DIVISION", DivisionEntity.class);
        entityClassMap.put("LOCATION", LocationEntity.class);
        entityClassMap.put("PACKAGE", PackageEntity.class);
        entityClassMap.put("ROOM-BED", RoomBedEntity.class);
        entityClassMap.put("SERVICE", ServiceEntity.class);
        entityClassMap.put("SYSTEM", SystemEntity.class);
        entityClassMap.put("TEAM", TeamEntity.class);
        entityClassMap.put("TEAM (OE/RR)", OERRTeamEntity.class);
        entityClassMap.put("USER", UserEntity.class);
	}

	/**
	 * Method buildEntities.
	 * @param entList
	 * @throws ParameterDefinitionException
	 */
	private void buildEntities(String entList) throws ParameterDefinitionException {
		for (int i = 0; i < StringUtil.length(entList,';'); i++) {
			addEntity(StringUtil.piece(entList,';', i+1));
		}
	}

	private void addEntity(String entString) throws ParameterDefinitionException {
		try {
			String entityFile = StringUtil.piece(entString,'|',2);
			Class entityClass = (Class) entityClassMap.get(entityFile);
			IEntity entity = null;
			if (entityClass != null) {
				entity = (IEntity) entityClass.newInstance();
				entity.setPrecedence(Float.parseFloat(StringUtil.piece(entString,'|',1)));
				entities.add(entity);
			}
		} catch (InstantiationException e) {
			throw new ParameterDefinitionException(e);
		} catch (IllegalAccessException e) {
			throw new ParameterDefinitionException(e);
		}

	}
	
	/**
	 * Method buildEntities.
	 * @param node
	 */
	private void buildEntities(Node node) throws ParameterDefinitionException {
		XPath xpath;
		try {
			xpath = (XPath) new Dom4jXPath("entities/entity");
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(
				"JaxenException at entities/entity.",
				e);
		}
		List results = null;
		try {
			results = xpath.selectNodes(node);
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(
				"JaxenException at selectNodes(node).",
				e);
		}

		if (results != null) {
			Iterator resultIter = results.iterator();

			if (!resultIter.hasNext())
				return;

			while (resultIter.hasNext()) {
				Object object = resultIter.next();
				if (object instanceof Node) {
					addEntity((Node) object);
				}
			}
		}
	}

	/**
	 * Method addEntity.
	 * @param node
	 */
	private void addEntity(Node node) throws ParameterDefinitionException {
		try {
			String entityFile =
				((Node) (new Dom4jXPath("entityFile")).selectSingleNode(node))
					.getStringValue();
			Class entityClass = (Class) entityClassMap.get(entityFile);
			IEntity entity = null;
			if (entityClass != null) {
				entity = (IEntity) entityClass.newInstance();
				entity.setPrecedence(
					Float.parseFloat(
						((Node) (new Dom4jXPath("precedence"))
							.selectSingleNode(node))
							.getStringValue()));
				entities.add(entity);
			}
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(e);
		} catch (InstantiationException e) {
			throw new ParameterDefinitionException(e);
		} catch (IllegalAccessException e) {
			throw new ParameterDefinitionException(e);
		}

	}

	/**
	 * Method buildEntities.
	 * @param node
	 */
	private void buildKeywords(Node node) throws ParameterDefinitionException {
		XPath xpath;
		try {
			xpath = (XPath) new Dom4jXPath("keywords/keyword");
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(
				"JaxenException at keywords.",
				e);
		}
		List results = null;
		try {
			results = xpath.selectNodes(node);
		} catch (JaxenException e) {
			throw new ParameterDefinitionException(
				"JaxenException at selectNodes(node).",
				e);
		}

		if (results != null) {
			Iterator resultIter = results.iterator();

			while (resultIter.hasNext()) {
				Object object = resultIter.next();
				if (object instanceof Node) {
					keywords.add(((Node) object).getStringValue());
				}
			}
		}
	}

	/**
	 * Method addEntity.
	 * @param node
	 */
	private void addKeyword(Node node) throws ParameterDefinitionException {
		String keyword = node.getStringValue();
		if (keyword != null) {
			keywords.add(keyword);
		}

	}

	/**
	 * Returns the description.
	 * @return String
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * Returns the displayText.
	 * @return String
	 */
	public String getDisplayText() {
		return displayText;
	}

	/**
	 * Returns the entities.
	 * @return List
	 */
	public List getEntities() {
		return entities;
	}

	/**
	 * Returns the fullName.
	 * @return String
	 */
	public String getFullName() {
		return fullName;
	}

	/**
	 * Returns the instanceDomain.
	 * @return String
	 */
	public String getInstanceDomain() {
		return instanceDomain;
	}

	/**
	 * Returns the instanceHelp.
	 * @return String
	 */
	public String getInstanceHelp() {
		return instanceHelp;
	}

	/**
	 * Returns the instanceScreenCode.
	 * @return String
	 */
	public String getInstanceScreenCode() {
		return instanceScreenCode;
	}

	/**
	 * Returns the instanceTerm.
	 * @return String
	 */
	public String getInstanceTerm() {
		return instanceTerm;
	}

	/**
	 * Returns the instanceType.
	 * @return String
	 */
	public String getInstanceType() {
		return instanceType;
	}

	/**
	 * Returns the instanceValidationCode.
	 * @return String
	 */
	public String getInstanceValidationCode() {
		return instanceValidationCode;
	}

	/**
	 * Returns the multiValued.
	 * @return boolean
	 */
	public boolean isMultiValued() {
		return multiValued;
	}

	/**
	 * Returns the prohibitEditing.
	 * @return boolean
	 */
	public boolean isProhibitEditing() {
		return prohibitEditing;
	}

	/**
	 * Returns the valueDomain.
	 * @return String
	 */
	public String getValueDomain() {
		return valueDomain;
	}

	/**
	 * Returns the valueHelp.
	 * @return String
	 */
	public String getValueHelp() {
		return valueHelp;
	}

	/**
	 * Returns the valueScreenCode.
	 * @return String
	 */
	public String getValueScreenCode() {
		return valueScreenCode;
	}

	/**
	 * Returns the valueTerm.
	 * @return String
	 */
	public String getValueTerm() {
		return valueTerm;
	}

	/**
	 * Returns the valueType.
	 * @return String
	 */
	public String getValueType() {
		return valueType;
	}

	/**
	 * Returns the valueValidationCode.
	 * @return String
	 */
	public String getValueValidationCode() {
		return valueValidationCode;
	}

	/**
	 * Sets the description.
	 * @param description The description to set
	 */
	public void setDescription(String description) {
		this.description = description;
	}

	/**
	 * Sets the displayText.
	 * @param displayText The displayText to set
	 */
	public void setDisplayText(String displayText) {
		this.displayText = displayText;
	}

	/**
	 * Sets the entities.
	 * @param entities The entities to set
	 */
	public void setEntities(List entities) {
		this.entities = entities;
	}

	public void addEntity(IEntity entity) {
		this.entities.add(entity);
	}

	public void removeEntity(int index) {
		this.entities.remove(index);
	}

	public void removeEntity(IEntity entity) {
		this.entities.remove(entity);
	}

	/**
	 * Sets the fullName.
	 * @param fullName The fullName to set
	 */
	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	/**
	 * Sets the instanceDomain.
	 * @param instanceDomain The instanceDomain to set
	 */
	public void setInstanceDomain(String instanceDomain) {
		this.instanceDomain = instanceDomain;
	}

	/**
	 * Sets the instanceHelp.
	 * @param instanceHelp The instanceHelp to set
	 */
	public void setInstanceHelp(String instanceHelp) {
		this.instanceHelp = instanceHelp;
	}

	/**
	 * Sets the instanceScreenCode.
	 * @param instanceScreenCode The instanceScreenCode to set
	 */
	public void setInstanceScreenCode(String instanceScreenCode) {
		this.instanceScreenCode = instanceScreenCode;
	}

	/**
	 * Sets the instanceTerm.
	 * @param instanceTerm The instanceTerm to set
	 */
	public void setInstanceTerm(String instanceTerm) {
		this.instanceTerm = instanceTerm;
	}

	/**
	 * Sets the instanceType.
	 * @param instanceType The instanceType to set
	 */
	public void setInstanceType(String instanceType) {
		this.instanceType = instanceType;
	}

	/**
	 * Sets the instanceValidationCode.
	 * @param instanceValidationCode The instanceValidationCode to set
	 */
	public void setInstanceValidationCode(String instanceValidationCode) {
		this.instanceValidationCode = instanceValidationCode;
	}

	/**
	 * Sets the multiValued.
	 * @param multiValued The multiValued to set
	 */
	public void setMultiValued(boolean multiValued) {
		this.multiValued = multiValued;
	}

	/**
	 * Sets the prohibitEditing.
	 * @param prohibitEditing The prohibitEditing to set
	 */
	public void setProhibitEditing(boolean prohibitEditing) {
		this.prohibitEditing = prohibitEditing;
	}

	/**
	 * Sets the valueDomain.
	 * @param valueDomain The valueDomain to set
	 */
	public void setValueDomain(String valueDomain) {
		this.valueDomain = valueDomain;
	}

	/**
	 * Sets the valueHelp.
	 * @param valueHelp The valueHelp to set
	 */
	public void setValueHelp(String valueHelp) {
		this.valueHelp = valueHelp;
	}

	/**
	 * Sets the valueScreenCode.
	 * @param valueScreenCode The valueScreenCode to set
	 */
	public void setValueScreenCode(String valueScreenCode) {
		this.valueScreenCode = valueScreenCode;
	}

	/**
	 * Sets the valueTerm.
	 * @param valueTerm The valueTerm to set
	 */
	public void setValueTerm(String valueTerm) {
		this.valueTerm = valueTerm;
	}

	/**
	 * Sets the valueType.
	 * @param valueType The valueType to set
	 */
	public void setValueType(String valueType) {
		this.valueType = valueType;
	}

	/**
	 * Sets the valueValidationCode.
	 * @param valueValidationCode The valueValidationCode to set
	 */
	public void setValueValidationCode(String valueValidationCode) {
		this.valueValidationCode = valueValidationCode;
	}

	public String toString() {
		return fullName;
	}

	/**
	 * Returns the id.
	 * @return long
	 */
	public long getId() {
		return id;
	}

	/**
	 * Sets the id.
	 * @param id The id to set
	 */
	public void setId(long id) {
		this.id = id;
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IParameterDefinition#getKeyWords()
	 */
	public List getKeyWords() {
		return keywords;
	}

}
