package gov.va.med.hds.cd.config.internal;

import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.util.mumps.StringUtil;

import java.util.EventListener;

import javax.swing.event.EventListenerList;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class ParameterTree implements TreeModel, IParameterTree {

	/** Root of the tree. */
	protected IParameterCategory root;

	/** Listeners. */
	protected EventListenerList listenerList = new EventListenerList();

	/**
	  * Creates a tree in which any node can have children.
	  *
	  * @param root a ParameterCategory object that is the root of the tree
	  * 
	  */
	public ParameterTree(IParameterCategory root) {
		this.root = root;
	}

	/**
	 * Sets the root to <code>root</code>. A null <code>root</code> implies
	 * the tree is to display nothing, and is legal.
	 */
	public void setRoot(IParameterCategory root) {
		IParameterCategory oldRoot = this.root;
		this.root = root;
		if (root == null && oldRoot != null) {
			fireTreeStructureChanged(this, null);
		} else {
			nodeStructureChanged(root);
		}
	}

	/**
	 * Returns the root of the tree.  Returns null only if the tree has
	 * no nodes.
	 *
	 * @return  the root of the tree
	 */
	public Object getRoot() {
		return root;
	}

	/**
	 * Returns the index of child in parent.
	 * If either the parent or child is <code>null</code>, returns -1.
	 * @param parent a note in the tree, obtained from this data source
	 * @param child the node we are interested in
	 * @return the index of the child in the parent, or -1
	 *    if either the parent or the child is <code>null</code>
	 */
	public int getIndexOfChild(
		IParameterCategory parent,
		IParameterCategory child) {
		if (parent == null || child == null)
			return -1;
		return parent.getIndex(child);
	}

	/**
	 * Returns the child of <I>parent</I> at index <I>index</I> in the parent's
	 * child array.  <I>parent</I> must be a node previously obtained from
	 * this data source. This should not return null if <i>index</i>
	 * is a valid index for <i>parent</i> (that is <i>index</i> >= 0 &&
	 * <i>index</i> < getChildCount(<i>parent</i>)).
	 *
	 * @param   parent  a node in the tree, obtained from this data source
	 * @return  the child of <I>parent</I> at index <I>index</I>
	 */
	public IParameterCategory getChild(IParameterCategory parent, int index) {
		return parent.getChildAt(index);
	}

	/**
	 * Returns the number of children of <I>parent</I>.  Returns 0 if the node
	 * is a leaf or if it has no children.  <I>parent</I> must be a node
	 * previously obtained from this data source.
	 *
	 * @param   parent  a node in the tree, obtained from this data source
	 * @return  the number of children of the node <I>parent</I>
	 */
	public int getChildCount(IParameterCategory parent) {
		return ((IParameterCategory) parent).getChildCount();
	}

	/**
	 * Invoke this method if you've modified the TreeNodes upon which this
	 * model depends.  The model will notify all of its listeners that the
	 * model has changed.
	 */
	public void reload() {
		reload(root);
	}

	public void insertNodeInto(
		IParameterCategory newChild,
		IParameterCategory parent) {

		parent.insert(newChild);

		int index = parent.getIndex(newChild);

		int[] newIndexes = new int[1];

		newIndexes[0] = index;

		nodesWereInserted(parent, newIndexes);

	}

	/**
	 * Invoked this to insert newChild at location index in parents children.
	 * This will then message nodesWereInserted to create the appropriate
	 * event. This is the preferred way to add children as it will create
	 * the appropriate event.
	 */
	public void insertNodeInto(
		IParameterCategory newChild,
		IParameterCategory parent,
		int index) {
		parent.insert(newChild, index);

		int[] newIndexes = new int[1];

		newIndexes[0] = index;
		nodesWereInserted(parent, newIndexes);
	}

	/**
	 * Message this to remove node from its parent. This will message
	 * nodesWereRemoved to create the appropriate event. This is the
	 * preferred way to remove a node as it handles the event creation
	 * for you.
	 */
	public void removeNodeFromParent(IParameterCategory node) {
		IParameterCategory parent = node.getParent();

		if (parent == null)
			throw new IllegalArgumentException("node does not have a parent.");

		int[] childIndex = new int[1];
		IParameterCategory[] removedArray = new IParameterCategory[1];

		childIndex[0] = parent.getIndex(node);
		parent.remove(childIndex[0]);
		removedArray[0] = node;
		nodesWereRemoved(parent, childIndex, removedArray);
	}

	/**
	  * Invoke this method after you've changed how node is to be
	  * represented in the tree.
	  */
	public void nodeChanged(IParameterCategory node) {
		if (listenerList != null && node != null) {
			IParameterCategory parent = node.getParent();

			if (parent != null) {
				int anIndex = parent.getIndex(node);
				if (anIndex != -1) {
					int[] cIndexs = new int[1];

					cIndexs[0] = anIndex;
					nodesChanged(parent, cIndexs);
				}
			} else if (node == getRoot()) {
				nodesChanged(node, null);
			}
		}
	}

	/**
	 * Invoke this method if you've modified the TreeNodes upon which this
	 * model depends.  The model will notify all of its listeners that the
	 * model has changed below the node <code>node</code> (PENDING).
	 */
	public void reload(IParameterCategory node) {
		if (node != null) {
			fireTreeStructureChanged(this, getPathToRoot(node), null, null);
		}
	}

	/**
	  * Invoke this method after you've inserted some TreeNodes into
	  * node.  childIndices should be the index of the new elements and
	  * must be sorted in ascending order.
	  */
	public void nodesWereInserted(
		IParameterCategory node,
		int[] childIndices) {
		if (listenerList != null
			&& node != null
			&& childIndices != null
			&& childIndices.length > 0) {
			int cCount = childIndices.length;
			IParameterCategory[] newChildren = new IParameterCategory[cCount];

			for (int counter = 0; counter < cCount; counter++)
				newChildren[counter] = node.getChildAt(childIndices[counter]);
			fireTreeNodesInserted(
				this,
				getPathToRoot(node),
				childIndices,
				newChildren);
		}
	}

	/**
	  * Invoke this method after you've removed some TreeNodes from
	  * node.  childIndices should be the index of the removed elements and
	  * must be sorted in ascending order. And removedChildren should be
	  * the array of the children objects that were removed.
	  */
	public void nodesWereRemoved(
		IParameterCategory node,
		int[] childIndices,
		IParameterCategory[] removedChildren) {
		if (node != null && childIndices != null) {
			fireTreeNodesRemoved(
				this,
				getPathToRoot(node),
				childIndices,
				removedChildren);
		}
	}

	/**
	  * Invoke this method after you've changed how the children identified by
	  * childIndicies are to be represented in the tree.
	  */
	public void nodesChanged(IParameterCategory node, int[] childIndices) {
		if (node != null) {
			if (childIndices != null) {
				int cCount = childIndices.length;

				if (cCount > 0) {
					IParameterCategory[] cChildren =
						new IParameterCategory[cCount];

					for (int counter = 0; counter < cCount; counter++)
						cChildren[counter] =
							node.getChildAt(childIndices[counter]);
					fireTreeNodesChanged(
						this,
						getPathToRoot(node),
						childIndices,
						cChildren);
				}
			} else if (node == getRoot()) {
				fireTreeNodesChanged(this, getPathToRoot(node), null, null);
			}
		}
	}

	/**
	  * Invoke this method if you've totally changed the children of
	  * node and its childrens children...  This will post a
	  * treeStructureChanged event.
	  */
	public void nodeStructureChanged(IParameterCategory node) {
		if (node != null) {
			fireTreeStructureChanged(this, getPathToRoot(node), null, null);
		}
	}

	/**
	 * Builds the parents of node up to and including the root node,
	 * where the original node is the last element in the returned array.
	 * The length of the returned array gives the node's depth in the
	 * tree.
	 * 
	 * @param aNode the TreeNode to get the path for
	 * @param an array of TreeNodes giving the path from the root to the
	 *        specified node. 
	 */
	public IParameterCategory[] getPathToRoot(IParameterCategory aNode) {
		return getPathToRoot(aNode, 0);
	}

	/**
	 * Builds the parents of node up to and including the root node,
	 * where the original node is the last element in the returned array.
	 * The length of the returned array gives the node's depth in the
	 * tree.
	 * 
	 * @param aNode  the TreeNode to get the path for
	 * @param depth  an int giving the number of steps already taken towards
	 *        the root (on recursive calls), used to size the returned array
	 * @return an array of TreeNodes giving the path from the root to the
	 *         specified node 
	 */
	protected IParameterCategory[] getPathToRoot(
		IParameterCategory aNode,
		int depth) {
		IParameterCategory[] retNodes;
		// This method recurses, traversing towards the root in order
		// size the array. On the way back, it fills in the nodes,
		// starting from the root and working back to the original node.

		/* Check for null, in case someone passed in a null node, or
		   they passed in an element that isn't rooted at root. */
		if (aNode == null) {
			if (depth == 0)
				return null;
			else
				retNodes = new IParameterCategory[depth];
		} else {
			depth++;
			if (aNode == root)
				retNodes = new IParameterCategory[depth];
			else
				retNodes = getPathToRoot(aNode.getParent(), depth);
			retNodes[retNodes.length - depth] = aNode;
		}
		return retNodes;
	}

	//
	//  Events
	//

	/**
	 * Adds a listener for the TreeModelEvent posted after the tree changes.
	 *
	 * @see     #removeTreeModelListener
	 * @param   l       the listener to add
	 */

	public void addTreeModelListener(TreeModelListener l) {
		listenerList.add(TreeModelListener.class, l);
	}

	/**
	 * Removes a listener previously added with <B>addTreeModelListener()</B>.
	 *
	 * @see     #addTreeModelListener
	 * @param   l       the listener to remove
	 */
	public void removeTreeModelListener(TreeModelListener l) {
		listenerList.remove(TreeModelListener.class, l);
	}

	/**
	 * Returns an array of all the tree model listeners
	 * registered on this model.
	 *
	 * @return all of this model's <code>TreeModelListener</code>s
	 *         or an empty
	 *         array if no tree model listeners are currently registered
	 *
	 * @see #addTreeModelListener
	 * @see #removeTreeModelListener
	 *
	 * @since 1.4
	 */
	public TreeModelListener[] getTreeModelListeners() {
		return (TreeModelListener[]) listenerList.getListeners(
			TreeModelListener.class);
	}

	/**
	 * Notifies all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 *
	 * @param source the node being changed
	 * @param path the path to the root node
	 * @param childIndicies the indices of the changed elements
	 * @param children the changed elements
	 * @see EventListenerList
	 */
	protected void fireTreeNodesChanged(
		IParameterTree source,
		IParameterCategory[] path,
		int[] childIndices,
		IParameterCategory[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListeners(TreeModelListener.class);
		TreeModelEvent e = null;
		// Process the listeners
		for (int i = 0; i < listeners.length; i++) {
			// Lazily create the event:
			if (e == null) {
				e = new TreeModelEvent(source, path, childIndices, children);
			}
			((TreeModelListener) listeners[i]).treeNodesChanged(e);
		}
	}

	/**
	 * Notifies all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 *
	 * @param source the node where new elements are being inserted
	 * @param path the path to the root node
	 * @param childIndicies the indices of the new elements
	 * @param children the new elements
	 * @see EventListenerList
	 */
	protected void fireTreeNodesInserted(
		IParameterTree source,
		IParameterCategory[] path,
		int[] childIndices,
		IParameterCategory[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeNodesInserted(e);
			}
		}
	}

	/**
	 * Notifies all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 *
	 * @param source the node where elements are being removed
	 * @param path the path to the root node
	 * @param childIndicies the indices of the removed elements
	 * @param children the removed elements
	 * @see EventListenerList
	 */
	protected void fireTreeNodesRemoved(
		IParameterTree source,
		IParameterCategory[] path,
		int[] childIndices,
		IParameterCategory[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeNodesRemoved(e);
			}
		}
	}

	/**
	 * Notifies all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 *
	 * @param source the node where the tree model has changed
	 * @param path the path to the root node
	 * @param childIndicies the indices of the affected elements
	 * @param children the affected elements
	 * @see EventListenerList
	 */
	protected void fireTreeStructureChanged(
		IParameterTree source,
		IParameterCategory[] path,
		int[] childIndices,
		IParameterCategory[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeStructureChanged(e);
			}
		}
	}

	/*
	 * Notifies all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 *
	 * @param source the node where the tree model has changed
	 * @param path the path to the root node
	 * @see EventListenerList
	 */
	private void fireTreeStructureChanged(
		IParameterTree source,
		TreePath path) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e = new TreeModelEvent(source, path);
				((TreeModelListener) listeners[i + 1]).treeStructureChanged(e);
			}
		}
	}

	/**
	 * Returns an array of all the objects currently registered
	 * as <code><em>Foo</em>Listener</code>s
	 * upon this model.
	 * <code><em>Foo</em>Listener</code>s are registered using the
	 * <code>add<em>Foo</em>Listener</code> method.
	 *
	 * <p>
	 *
	 * You can specify the <code>listenerType</code> argument
	 * with a class literal,
	 * such as
	 * <code><em>Foo</em>Listener.class</code>.
	 * For example, you can query a
	 * <code>DefaultTreeModel</code> <code>m</code>
	 * for its tree model listeners with the following code:
	 *
	 * <pre>TreeModelListener[] tmls = (TreeModelListener[])(m.getListeners(TreeModelListener.class));</pre>
	 *
	 * If no such listeners exist, this method returns an empty array.
	 *
	 * @param listenerType the type of listeners requested; this parameter
	 *          should specify an interface that descends from
	 *          <code>java.util.EventListener</code>
	 * @return an array of all objects registered as
	 *          <code><em>Foo</em>Listener</code>s on this component,
	 *          or an empty array if no such
	 *          listeners have been added
	 * @exception ClassCastException if <code>listenerType</code>
	 *          doesn't specify a class or interface that implements
	 *          <code>java.util.EventListener</code>
	 *
	 * @see #getTreeModelListeners
	 * 
	 * @since 1.3
	 */
	public EventListener[] getListeners(Class listenerType) {
		return listenerList.getListeners(listenerType);
	}
	/**
	 * @see javax.swing.tree.TreeModel#getChild(java.lang.Object, int)
	 */
	public Object getChild(Object parent, int index) {
		return ((IParameterCategory) parent).getChildAt(index);
	}
	/**
	 * @see javax.swing.tree.TreeModel#getChildCount(java.lang.Object)
	 */
	public int getChildCount(Object parent) {
		return ((IParameterCategory) parent).getChildCount();
	}
	/**
	 * @see javax.swing.tree.TreeModel#isLeaf(java.lang.Object)
	 */
	public boolean isLeaf(Object node) {
		return ((IParameterCategory) node).isLeaf();
	}
	/**
	 * @see javax.swing.tree.TreeModel#valueForPathChanged(javax.swing.tree.TreePath, java.lang.Object)
	 */
	public void valueForPathChanged(TreePath path, Object newValue) {
		IParameterCategory oldValue =
			(IParameterCategory) path.getLastPathComponent();
		oldValue.setName(newValue.toString());
		String fullName =
			StringUtil.piece(oldValue.getFullName(), ' ', 1)
				+ " "
				+ newValue.toString().toUpperCase();
		oldValue.setFullName(fullName);
		nodeChanged(oldValue);
	}
	/**
	 * @see javax.swing.tree.TreeModel#getIndexOfChild(java.lang.Object, java.lang.Object)
	 */
	public int getIndexOfChild(Object parent, Object child) {
		return ((IParameterCategory) parent).getIndex(
			(IParameterCategory) child);
	}

}
