package gov.va.med.hds.cd.config.itest;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.ICallResultHandler;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurator;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterDefinition;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.cd.config.ParameterCategoryException;
import gov.va.med.hds.cd.config.internal.ConfigurationCallHandler;
import gov.va.med.hds.cd.config.internal.Parameter;
import gov.va.med.hds.cd.config.internal.ParameterCategory;
import gov.va.med.hds.cd.config.internal.ParameterDefinitionListHandler;
import gov.va.med.hds.cd.config.internal.ParameterTree;
import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.runtime.test.RuntimeTestCase;
import gov.va.med.hds.cd.security.ISecurity;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Handler;
import java.util.logging.LogRecord;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class MConfigDAOTest extends RuntimeTestCase implements IConfigDAOListener {

	private static final String GET_CONFIGURATION_RPC_KEY = "GET CONFIGURATION";
	private static final String PARAM_DEF_LIST_RPC_KEY = "PARAM DEF LIST";
	private IConfigDAO configDAO = null;
	private ISecurity securityModule = null;
	private IDaoRegistry daoRegistry = null;
	private IParameterTree pTree = null;
	private IParameterCategory pRoot = null;
	private static String prismPluginId = "gov.va.med.hds.cd.prism";
	private TestHandler testHandler = new TestHandler();
	private IConfigurator configurator = null;
	private IConfiguration configuration = null;
	private List pDefs = null;
	private ConfigDAOEvent configDAOEvent = null;
	private Map pDefMap = new HashMap();

	/**
	 * Constructor for MConfigDAOTest.
	 * @param arg0
	 */
	public MConfigDAOTest(String arg0) {
		super(arg0);
	}

	public static void main(String[] args) {
		junit.textui.TestRunner.run(MConfigDAOTest.class);
	}

	/**
	 * @see TestCase#setUp()
	 */
	protected synchronized void setUp() throws Exception {
		super.setUp();
		super.setUp();
		parseRegistryResource("registry.xml");
		getRuntime().readSettings(getClass().getResourceAsStream("VistalinkSettings.xml"));

		securityModule =
			(ISecurity) getRuntime().getPluginInstance(ISecurity.PLUGIN_ID);
		securityModule.login();
		configurator = (IConfigurator) getRuntime().getPluginInstance(IConfigurator.PLUGIN_ID);
		daoRegistry = (IDaoRegistry) getRuntime().getPluginInstance(IDaoRegistry.PLUGIN_ID);
		configDAO = (IConfigDAO) daoRegistry.getDAO(IConfigDAO.INTERFACE_NAME);
		List params = new ArrayList();
		params.add(prismPluginId);
		CallRequest request = new CallRequest(GET_CONFIGURATION_RPC_KEY, params);
		ICallResultHandler handler = new ConfigurationCallHandler(configuration);
		configDAO.query(request, handler, this);
		wait();
		pRoot = (IParameterCategory) configDAOEvent.getResultObject();
		pTree = new ParameterTree(pRoot);
	}

	public void PutParameterTree() {
	
		try {
			configDAO.putParameterTree(pTree);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}

	}
	
	private void PutParameter(IParameter par) {
		String key = par.getKey();
		String value = par.getValue();
		try {
			configDAO.putParameter(key, value);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
	}

	private void PutLargeParameter(IParameter par) {
		String key = par.getKey();
		String value = par.getValue();
		try {
			configDAO.putParameter(key, value);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
	}
	
	public synchronized void testUpdateParameterCategory() {
		List params = new ArrayList();
		params.add(prismPluginId);
		CallRequest request = new CallRequest(GET_CONFIGURATION_RPC_KEY, params);
		ICallResultHandler handler = new ConfigurationCallHandler(configuration);
		
		//change fullName
		String tempName = pRoot.getFullName();
		pRoot.setFullName("XHD TEST PRISM PARAMETERS");
		try {
			configDAO.updateParameterCategory(pRoot);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
		
		//retrieve changes, check validity	
		configDAO.query(request, handler, this);
		try {
			wait();
			if (configDAOEvent.getException() != null) {
				fail("ResultHandlerException caught: " + configDAOEvent.getException());
			}
			pRoot = (IParameterCategory) configDAOEvent.getResultObject();
			assertEquals("XHD TEST PRISM PARAMETERS", pRoot.getFullName());
		} catch (InterruptedException e) {
			fail("InterruptedException caught: " + e.getMessage());
		}

		//restore fullName
		pRoot.setFullName(tempName);
		try {
			configDAO.updateParameterCategory(pRoot);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
		
		//retrieve changes, check validity
		configDAO.query(request, handler, this);
		try {
			wait();
			if (configDAOEvent.getException() != null) {
				fail("ResultHandlerException caught: " + configDAOEvent.getException());
			}
			pRoot = (IParameterCategory) configDAOEvent.getResultObject();
			assertEquals(tempName, pRoot.getFullName());
		} catch (InterruptedException e) {
			fail("InterruptedException caught: " + e.getMessage());
		}
				
	}

	public void testInsertParameterCategory() {
		IParameterCategory 	pc = new ParameterCategory("Test Category", null),
							parent = null,
							subPc = new ParameterCategory("Test subCategory", null),
							subSubPc = new ParameterCategory("Test subSubCategory", null);

		parameterDefinitionQuery();
		
		parent = pRoot;

		pc.setFullName("XHD TEST CATEGORY");
		pc.setParent(parent);
		pc.setParentId(parent.getId());
		pc.setPluginId(prismPluginId);
		pc.setPreferencePage("gov.va.med.hds.config.dialog.TestPrefPage");

		IParameterDefinition wpPdef = (IParameterDefinition) pDefMap.get("XHD TEST WP PARAMETERS");
		if (wpPdef == null) {
			fail("Parameter Definition for XHD TEST WP PARAMETERS not found!");
		}
		
		IParameter wpPar = (IParameter) new Parameter(wpPdef.getDisplayText(), null, pc);

		wpPar.setKey(wpPdef.getDisplayText() + "^UserId^" + wpPdef.getId() + "^1");
		wpPar.setValue(getTree());
		wpPar.setId(wpPdef.getId());
		wpPar.setFullName(wpPdef.getFullName());
		wpPar.setMultiValued(wpPdef.isMultiValued());
		wpPar.setWordProcessing(true);
		pc.addParameter(wpPar);

		subPc.setFullName("XHD TEST SUBCATEGORY");
		subPc.setParent(pc);
		subPc.setParentId(pc.getId());
		subPc.setPluginId(prismPluginId);
		subPc.setPreferencePage("gov.va.med.hds.config.dialog.TestPrefPage");

		IParameterDefinition scalarPdef = (IParameterDefinition) pDefMap.get("XHD TEST SCALAR PARAMETER");
		if (scalarPdef == null) {
			fail("Parameter Definition for XHD TEST SCALAR PARAMETER not found!");
		}
		
		IParameter scalarPar = (IParameter) new Parameter(scalarPdef.getDisplayText(), null, pc);

		scalarPar.setKey(scalarPdef.getDisplayText() + "^UserId^" + scalarPdef.getId() + "^1");
		scalarPar.setValue("99");
		scalarPar.setId(scalarPdef.getId());
		scalarPar.setFullName(scalarPdef.getFullName());
		scalarPar.setMultiValued(scalarPdef.isMultiValued());
		scalarPar.setWordProcessing(false);

		subPc.addParameter(scalarPar);
		pc.insert(subPc);
		
		subSubPc.setFullName("XHD TEST SUBSUBCATEGORY");
		subSubPc.setParent(subPc);
		subSubPc.setParentId(subPc.getId());
		subSubPc.setPluginId(IConfigurator.PLUGIN_ID);
		subSubPc.setPreferencePage("gov.va.med.hds.config.dialog.TestPrefPage");
		subSubPc.addParameter(wpPar);
		subPc.insert(subSubPc);
		
		try {
			configDAO.insertParameterCategory(pc, parent);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
		
		// while test category exists, test PutParameter and PutLargeParameter
		PutParameter(scalarPar);
		
		PutLargeParameter(wpPar);
		
	}

	public void testRemoveParameterCatetory() {
		IParameterCategory 	pc = null;
							
		pc = pRoot.getChild("XHD TEST CATEGORY");
		
		if (pc == null) {
			fail("Couldn't get XHD TEST CATEGORY from Root Category");
		}
		
		try {
			configDAO.removeParameterCategory(pc);
		} catch (ParameterCategoryException e) {
			fail("ParameterCategoryException caught: " + e.getMessage());
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
		// call testInsertParameterCategory to re-establish link
		testInsertParameterCategory();
		
	}
	
	public void testDeleteParameterCategory() {
		IParameterCategory 	pc = null;
		
		pc = pRoot.getChild("XHD TEST CATEGORY");
		
		if (pc == null) {
			fail("Couldn't get XHD TEST CATEGORY from Root Category");
		}
		
		try {
			configDAO.deleteParameterCategory(pc, true);
		} catch (ConfigDAOException e) {
			fail("ConfigDAOException caught: " + e.getMessage());
		}
	}

	private synchronized void parameterDefinitionQuery() {
		CallRequest request = new CallRequest(PARAM_DEF_LIST_RPC_KEY);
		ICallResultHandler handler = new ParameterDefinitionListHandler();
		
		configDAO.query(request, handler, this);
		try {
			wait();
			if (configDAOEvent.getException() != null) {
				fail("ResultHandlerException caught: " + configDAOEvent.getException());
			}
			pDefs = (List) configDAOEvent.getResultObject();
			assertTrue(pDefs.size() > 0);
			Iterator pdIt = pDefs.iterator();
			
			while (pdIt.hasNext()) {
				IParameterDefinition pd = (IParameterDefinition) pdIt.next();
				pDefMap.put(pd.getFullName(), pd);
			}
			 
		} catch (InterruptedException e) {
			fail("InterruptedException caught: " + e.getMessage());
		}
		
	}

	public synchronized void callFinished(ConfigDAOEvent configDAOEvent) {
		this.configDAOEvent = configDAOEvent;

		notify();
	}
	
	class TestHandler extends Handler {
		public void close() throws SecurityException {
		}
		public void flush() {
		}
		public void publish(LogRecord record) {
			fail(record.getMessage());
		}
	}
	
	private String getTree() {
		return "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
				"<configTree\n" +
					"xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n" +
					"xsi:noNamespaceSchemaLocation=\"C:\\reeng\\main\\modules\\config\\src\\gov\\va\\med\\hds\\config\\xml\\configTree.xsd\">\n" +
					"<pluginParameterCategory>\n" +
						"<id>1</id>\n" +
						"<name>Dashboard</name>\n" +
						"<preferencePage>gov.va.med.hds.dash.DBPrefPage</preferencePage>\n" +
						"<parameters>\n" +
							"<parameter>\n" +
								"<name>Look and Feel</name>\n" +
								"<key>Look and Feel^5000;DIC(4,^201^1</key>\n" +
								"<value>Motif</value>\n" +
							"</parameter>\n" +
							"<parameter>\n" +
								"<name>Selection Mode</name>\n" +
								"<key>Selection Mode^0^4749^1</key>\n" +
								"<value>Table-Navigator-Viewer</value>\n" +
							"</parameter>\n" +
							"<parameter>\n" +
								"<name>Window Mode</name>\n" +
								"<key>Window Mode^0^4750^1</key>\n" +
								"<value>Single Document Interface</value>\n" +
							"</parameter>\n" +
						"</parameters>\n" +
						"<children>\n" +
							"<parameterCategory>\n" +
								"<id>2</id>\n" +
								"<name>Patient Selection</name>\n" +
								"<preferencePage>gov.va.med.hds.dash.PtSelPrefPage</preferencePage>\n" +
								"<parameters>\n" +
									"<parameter>\n" +
										"<name>Patient Selection Mode</name>\n" +
										"<key>Patient Selection Mode^0^4751^1</key>\n" +
										"<value>single click</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Default Patient List</name>\n" +
										"<key>Default Patient List^375;DIC(9.4,^4752^1</key>\n" +
										"<value>Patient List A</value>\n" +
									"</parameter>\n" +
								"</parameters>\n" +
								"<children>\n" +
								"</children>\n" +
							"</parameterCategory>\n" +
							"<parameterCategory>\n" +
								"<id>3</id>\n" +
								"<name>Lab Collection Times</name>\n" +
								"<preferencePage>gov.va.med.hds.dash.LabPrefPage</preferencePage>\n" +
								"<parameters>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^28800</key>\n" +
										"<value>0810</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^36000</key>\n" +
										"<value>1030</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^39600</key>\n" +
										"<value>1130</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^66600</key>\n" +
										"<value>1900</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^73800</key>\n" +
										"<value>2100</value>\n" +
									"</parameter>\n" +
									"<parameter>\n" +
										"<name>Lab Collection Times</name>\n" +
										"<key>Lab Collection Times^5000;DIC(4,^21^75000</key>\n" +
										"<value>2350</value>\n" +
									"</parameter>\n" +
								"</parameters>\n" +
								"<children>\n" +
								"</children>\n" +
							"</parameterCategory>\n" +
						"</children>\n" +
					"</pluginParameterCategory>\n" +
				"</configTree>\n";
	}
		
	/**
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {
		securityModule.logout();
		super.tearDown();
	}

}
