package gov.va.med.hds.cd.config.test;

import gov.va.med.hds.cd.config.ConfigDAOEvent;
import gov.va.med.hds.cd.config.ConfigDAOException;
import gov.va.med.hds.cd.config.ConfigurationChangeEvent;
import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.IConfigDAO;
import gov.va.med.hds.cd.config.IConfigDAOListener;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurationChangeListener;
import gov.va.med.hds.cd.config.IConfigurationReadyListener;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.IParameterCategory;
import gov.va.med.hds.cd.config.IParameterChangeListener;
import gov.va.med.hds.cd.config.IParameterTree;
import gov.va.med.hds.cd.config.NonExistentParameterException;
import gov.va.med.hds.cd.config.ParameterCategoryException;
import gov.va.med.hds.cd.config.ParameterChangeEvent;
import gov.va.med.hds.cd.config.internal.ParameterCategory;
import gov.va.med.hds.cd.config.internal.ParameterTree;
import gov.va.med.hds.util.mumps.StringUtil;

import java.util.EventListener;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.event.EventListenerList;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class MockConfiguration implements IConfiguration {

	private EventListenerList configurationListeners = new EventListenerList();
	private IConfigDAO configDAO;
	private IParameterTree parameterTree;
	private IParameterCategory parameterRoot;
	private Map projectionMap = new HashMap();
	private Map indexMap = new HashMap();
	private Map catMap = new HashMap();

	MockConfiguration(String pluginID) throws ConfigurationException {

		configDAO = new MockConfigDAO();

		parameterTree = ((MockConfigDAO) configDAO).getParameterTree(pluginID);
		parameterRoot = ((MockConfigDAO) configDAO).getParameterRoot(pluginID);
		buildProjection(parameterRoot);
	}

	/**
	 * Method buildProjection.
	 * @param parameterRoot
	 */
	private void buildProjection(IParameterCategory pCat) {

		if (pCat == null)
			return;

		//copy pCat parameters Map to projectionMap
		Iterator keyIt = pCat.parameters().keySet().iterator();
		while (keyIt.hasNext()) {
			String fullKey = (String) keyIt.next();
			String ezKey = StringUtil.piece(fullKey, '^', 1);
			String value = pCat.getParameterValue(fullKey);
			projectionMap.put(ezKey, value);
			indexMap.put(ezKey, fullKey);
			catMap.put(fullKey, pCat);
		}

		/* iterate through pCat children & add their
		 * parameters to projectionMap */
		Iterator scIt = pCat.children().iterator();
		while (scIt.hasNext()) {
			IParameterCategory subCat = (ParameterCategory) scIt.next();
			if (subCat != null) {
				buildProjection(subCat);
			}
		}
	}

	public void addConfigurationChangeListener(IConfigurationChangeListener ccl) {
		configurationListeners.add(
			IConfigurationChangeListener.class,
			ccl);
	}

	public void removeConfigurationChangeListener(IConfigurationChangeListener ccl) {
		configurationListeners.remove(
			IConfigurationChangeListener.class,
			ccl);
	}

	public void addParameterChangeListener(IParameterChangeListener pcl) {
		configurationListeners.add(
			IParameterChangeListener.class,
			pcl);
	}

	public void removeParameterChangeListener(IParameterChangeListener pcl) {
		configurationListeners.remove(
			IParameterChangeListener.class,
			pcl);
	}

	protected void fireConfigurationChanged(
		IParameterCategory subject,
		IParameterCategory object,
		int verb) {

		EventListener[] ccl;

		ccl =
			configurationListeners.getListeners(
				IConfigurationChangeListener.class);

		if (ccl.length == 0) {
			return;
		}

		ConfigurationChangeEvent ccevent =
			new ConfigurationChangeEvent(this, subject, object, verb);

		for (int i = 0; i < ccl.length; ++i) {

			IConfigurationChangeListener listener =
				(IConfigurationChangeListener) ccl[i];
			listener.configurationChanged(ccevent);
		}

	}
	
	protected void fireParameterChanged(IParameter parameter) {

		EventListener[] pcl;

		pcl =
			configurationListeners.getListeners(
				IParameterChangeListener.class);

		if (pcl.length == 0) {
			return;
		}

		ParameterChangeEvent pcevent =
			new ParameterChangeEvent(this, parameter);

		for (int i = 0; i < pcl.length; ++i) {

			IParameterChangeListener listener =
				(IParameterChangeListener) pcl[i];
			listener.parameterChanged(pcevent);
		}

	}

	/**
	 * Returns the parameterTree.
	 * @return IParameterTree
	 */
	public IParameterTree getParameterTree() {
		return parameterTree;
	}

	/**
	 * Returns the parameterRoot.
	 * @return IParameterCategory
	 */
	public IParameterCategory getParameterRoot() {
		return parameterRoot;
	}

	public void commitChanges() throws ConfigDAOException {
		configDAO.putParameterTree(parameterTree);
	}

	public void removeParameterCategory(IParameterCategory pc)
		throws ConfigurationException {
		try {
			configDAO.removeParameterCategory(pc);
		} catch (ParameterCategoryException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		}
	}

	public String getParameterValue(String ezKey) {
		return (String) projectionMap.get(ezKey);
	}

	public void putParameter(String ezKey, String value)
		throws NonExistentParameterException, ConfigurationException {
		projectionMap.put(ezKey, value);
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException("Parameter not defined on Server");
		}
		pCat.putParameter(fullKey, value);
		try {
			configDAO.putParameter(fullKey, value);
			IParameter modParam = getParameter(ezKey);
			fireParameterChanged(modParam);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		} catch (NonExistentParameterException nPe) {
			throw new ConfigurationException(
				"Configuration Exception: " + nPe.getMessage());
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#putParameter(IParameter)
	 */
	public void putParameter(IParameter parameter)
		throws ConfigurationException {
		IParameterCategory pCat =
			getCategoryForParameter(
				StringUtil.piece(parameter.getKey(), '^', 1));
		if (!(pCat instanceof IParameterCategory)) {
			throw new ConfigurationException(
				"Category not found for Parameter: " + parameter.getName());
		}

		pCat.putParameter(parameter);

		try {
			configDAO.putParameter(parameter.getKey(), parameter.getValue());
			fireParameterChanged(parameter);
		} catch (ConfigDAOException e) {
			throw new ConfigurationException(
				"Configuration Exception: " + e.getMessage());
		}

	}

	public IParameterCategory getCategoryForParameter(String ezKey) {
		String fullKey = (String) indexMap.get(ezKey);
		return (ParameterCategory) catMap.get(fullKey);
	}

	/**
	 * Method getListenerCount.
	 */
	public int getListenerCount(Class type) {
		return configurationListeners.getListenerCount(type);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getParameter(String)
	 */
	public IParameter getParameter(String ezKey)
		throws NonExistentParameterException {
		String fullKey = (String) indexMap.get(ezKey);
		IParameterCategory pCat = (ParameterCategory) catMap.get(fullKey);
		// if can't find appropriate category, throw NonExistentParameterException
		if (pCat == null) {
			throw new NonExistentParameterException("Parameter not defined on Server");
		}
		return pCat.getParameter(fullKey);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#updateParameterCategory(IParameterCategory)
	 */
	public void updateParameterCategory(IParameterCategory parameterCategory)
		throws ConfigDAOException {
		configDAO.updateParameterCategory(parameterCategory);
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getMutableParameterRoot()
	 */
	public IParameterCategory getMutableParameterRoot()
		throws ConfigurationException {
		return null;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#getMutableParameterTree()
	 */
	public IParameterTree getMutableParameterTree()
		throws ConfigurationException {
		return null;
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#addConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationReadyListener)
	 */
	public void addConfigurationReadyListener(IConfigurationReadyListener crl) {
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#fireConfigurationReady()
	 */
	public void fireConfigurationReady() {
	}

	/**
	 * @see gov.va.med.hds.cd.config.IConfiguration#removeConfigurationReadyListener(gov.va.med.hds.cd.config.IConfigurationReadyListener)
	 */
	public void removeConfigurationReadyListener(IConfigurationReadyListener crl) {
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.config.IConfiguration#insertParameterCategory(gov.va.med.hds.cd.config.IParameterCategory, gov.va.med.hds.cd.config.IParameterCategory)
	 */
	public void insertParameterCategory(
		IParameterCategory pc,
		IParameterCategory parent)
		throws ConfigDAOException {
	}

	protected void fireConfigurationCallFailed(ConfigurationException configurationException) {
		EventListener[] crl;

		crl =
			configurationListeners.getListeners(
				IConfigurationReadyListener.class);

		if (crl.length == 0) {
			return;
		}

		for (int i = 0; i < crl.length; ++i) {

			IConfigurationReadyListener listener =
				(IConfigurationReadyListener) crl[i];
			listener.configurationCallFailed(configurationException);
		}
	}

	class ConfigurationCallListener implements IConfigDAOListener {
		/**
		 * @see gov.va.med.hds.cd.config.IConfigDAOListener#callFinished(gov.va.med.hds.cd.config.ConfigDAOEvent)
		 */
		public void callFinished(ConfigDAOEvent configDAOEvent) {
			if (configDAOEvent.getException() != null) {
				fireConfigurationCallFailed(
					new ConfigurationException(configDAOEvent.getException()));
			} else {
				parameterRoot =
					(IParameterCategory) configDAOEvent.getResultObject();
				parameterTree = new ParameterTree(parameterRoot);
				buildProjection(parameterRoot);
				fireConfigurationReady();
			}
		}
	}

}
