package gov.va.med.hds.cd.config;


import java.awt.datatransfer.DataFlavor;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public interface IParameterCategory {
	
	static final DataFlavor CATEGORY_FLAVOR =
		new DataFlavor(IParameterCategory.class, "Parameter Category");
	/**
	 * Returns the child <code>IParameterCategory</code> at index
	 * <code>childIndex</code>.
	 */
	IParameterCategory getChildAt(int childIndex);

	/**
	 * Returns the number of children <code>TreeNode</code>s the receiver
	 * contains.
	 */
	int getChildCount();

	/**
	 * Returns the parent <code>TreeNode</code> of the receiver.
	 */
	IParameterCategory getParent();

	/**
	 * Returns the index of <code>node</code> in the receivers children.
	 * If the receiver does not contain <code>node</code>, -1 will be
	 * returned.
	 */
	int getIndex(IParameterCategory node);

	/**
	 * Returns true if the receiver is a leaf.
	 */
	boolean isLeaf();

	/**
	 * Returns the children of the receiver as an <code>Enumeration</code>.
	 */
	List children();

	/**
	 * Returns the parameters of the receiver as an <code>Enumeration</code>.
	 */
	Map parameters();

	/**
	 * Returns the parameters of the receiver as an <code>Enumeration</code>.
	 */
	ArrayList parametersAsArrayList();

	/**
	 * overloaded insert (for when you just can't know which child you're inserting!
	 **/
	void insert(IParameterCategory child);

	/**
	 * Adds <code>child</code> to the receiver at <code>index</code>.
	 * <code>child</code> will be messaged with <code>setParent</code>.
	 */
	void insert(IParameterCategory child, int index);

	/**
	 * Removes the child at <code>index</code> from the receiver.
	 */
	void remove(int index);

	/**
	 * Removes <code>node</code> from the receiver. <code>setParent</code>
	 * will be messaged on <code>node</code>.
	 */
	void remove(IParameterCategory node);

	/**
	 * Marks the child at <code>index</code> for removal from the receiver.
	 */
	public void markToRemove(int index);

	/**
	 * Marks <code>node</code> for removal from the receiver. <code>setParent</code>
	 * will be messaged on <code>node</code>.
	 */
	public void markToRemove(IParameterCategory node);

	/**
	 * Adds <code>parameter</code> to the receiver at <code>key</code>.
	 * 
	 */
	void addParameter(String key, IParameter parameter);

	/**
	 * Overload of addParameter
	 **/
	void addParameter(IParameter parameter);

	/**
	 * Adds <code>parameter</code> to the receiver at <code>key</code>.
	 * 
	 */
	String getParameterValue(String key);

	/**
	 * Removes <code>parameter</code> from the receiver at <code>key</code>.
	 * 
	 */
	void removeParameter(String key);

	/**
	 * Removes the receiver from its parent.
	 */
	void removeFromParent();

	/**
	 * Sets the parent of the receiver to <code>newParent</code>.
	 */
	void setParent(IParameterCategory newParent);

	/**
	 * Sets the name of the receiver to <code>name</code>.
	 */
	void setName(String name);

	/**
	 * Returns the markedForRemoval.
	 * @return boolean
	 */
	boolean isMarkedForRemoval();

	/**
	 * Sets the markedForRemoval.
	 * @param markedForRemoval The markedForRemoval to set
	 */
	void setMarkedForRemoval(boolean markedForRemoval);

	/**
	 * Returns the name of the PreferencePage for this category
	 **/
	String getPreferencePage();

	void setPreferencePage(String preferencePage);

	/**
	 * evaluate whether a parameter category has be persisted in the
	 * backing store.
	 **/
	boolean isPersistent();

	/**
	 * Method setPersistent.
	 * @param b
	 */
	void setPersistent(boolean b);

	/**
	 * Sets the value of the parameter identified by key
	 * @param String key the unique string identifier of the parameter
	 * @param String value the value to filed for the parameter 
	 **/
	void putParameter(String key, String value);

	/**
	 * Sets the value of the parameter identified by key
	 * @param String key the unique string identifier of the parameter
	 * @param String value the value to filed for the parameter 
	 **/
	void putParameter(IParameter parameter);

	/**
	 * Updates Parameter values (and keys)
	 * @param oldKey the key that you're re-defining
	 * @param newKey the new key
	 * @param the value to be filed for the parameter
	 **/
	void updateParameter(String oldKey, String newKey, String newValue);

	/**
	 * Returns the id.
	 * @return long
	 */
	long getId();

	/**
	 * Returns the parentId.
	 * @return long
	 */
	long getParentId();

	/**
	 * Sets the id.
	 * @param id The id to set
	 */
	void setId(long id);

	/**
	 * Sets the parentId.
	 * @param parentId The parentId to set
	 */
	void setParentId(long parentId);

	/**
	 * Returns the fullName.
	 * @return String
	 */
	String getFullName();

	/**
	 * Returns the pluginId.
	 * @return String
	 */
	String getPluginId();

	/**
	 * Returns the name.
	 * @return String
	 */
	String getName();

	/**
	 * Sets the fullName.
	 * @param fullName The fullName to set
	 */
	void setFullName(String fullName);

	/**
	 * Sets the pluginId.
	 * @param pluginId The pluginId to set
	 */
	void setPluginId(String pluginId);

	/**
	 * Returns the packageRoot.
	 * @return boolean
	 */
	boolean isPackageRoot();

	/**
	 * Sets the packageRoot.
	 * @param packageRoot The packageRoot to set
	 */
	void setPackageRoot(boolean packageRoot);

	/**
	 * Method getChild.
	 * @param string
	 * @return IParameterCategory
	 */
	IParameterCategory getChild(String string);

	/**
	 * Sets the configuration.
	 * @param configuration The configuration to set
	 */
	void setConfiguration(IConfiguration configuration);

	/**
	 * Returns the configuration interface <code>IConfiguration</code> of
	 * the receiver.
	 */
	IConfiguration getConfiguration();
	
	/**
	 * Returns an <code>IParameter</code> reference, given its key
	 **/
	IParameter getParameter(String key);
	
	/**
	 * Calls configuration.updateParameterCategory((IParameterCategory) this)
	 * to save ParameterCategory to backing store
	 **/
	void persist() throws ConfigDAOException;

	/**
	 * Method getPath.
	 * @return IParameterCategory[]
	 */
	IParameterCategory[] getPath();
	
	/**
	 * Method setLeaf.
	 * @param isLeaf
	 */
	void setLeaf(boolean isLeaf);
	
}
