package gov.va.med.hds.cd.config;

import gov.va.med.hds.cd.spectrum.text.WholeNumberField;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.InputVerifier;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EtchedBorder;

/**
 *
 * @author Joel Russell
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 **/
public class IntegerFieldEditor
	extends FieldEditor
	implements IFieldEditor, ActionListener {

	private String labelText = "Specify value (0-100): ";
	private JPanel labelPanel = new JPanel(new GridLayout(0, 1));
	private JPanel fieldPanel = new JPanel(new GridLayout(0, 1));
	private int minValue = 0;
	private int maxValue = Integer.MAX_VALUE;

	protected IntegerFieldEditor() {
		super(new WholeNumberField(0, 7));
	}

	/** Creates new form ConfigPrefFieldEditor */
	public IntegerFieldEditor(String labelText, int min, int max) {
		super(new WholeNumberField(min, 7));
		this.label.setText(labelText);
		this.minValue = min;
		this.maxValue = max;
		initGUI();
	}

	private void initGUI() {
		errorText = "Enter a number between " + minValue + " and " + maxValue;
		field.setInputVerifier(new IntVerifier());
		((JTextField) field).addActionListener(this);
		field.setBorder(new EtchedBorder());
		
		labelPanel.add(label);

		fieldPanel.add(field);

		setLayout(new GridLayout(0, 2));
		add(labelPanel);
		add(fieldPanel);
	}

	/**
	 * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		JTextField tf = (JTextField) e.getSource();
		boolean isValid;
		isValid = isNumberValid(e.getActionCommand());

		if (!isValid) {
			tf.setBorder(new EtchedBorder(Color.RED, Color.RED));
			IntegerFieldEditor.this.showErrorMessage();
		} else {
			tf.setBorder(new EtchedBorder());
			IntegerFieldEditor.this.hideErrorMessage();

			if (parameter != null) {
				parameter.setValue(
					(isValid ? e.getActionCommand() : parameter.getValue()));
				setDirty(true);
			}
		}
	}

	private void commit() {
		
	}
	
	class IntVerifier extends InputVerifier {

		public boolean verify(JComponent input) {
			WholeNumberField tf = (WholeNumberField) input;
			return isNumberValid(tf.getText());
		}

		public boolean shouldYieldFocus(JComponent input) {
			WholeNumberField tf = (WholeNumberField) input;
			boolean result;
			result = super.shouldYieldFocus(input);

			if (!result) {
				tf.setBorder(new EtchedBorder(Color.RED, Color.RED));
				IntegerFieldEditor.this.showErrorMessage();
			} else {
				tf.setBorder(new EtchedBorder());
				IntegerFieldEditor.this.hideErrorMessage();
				if (parameter != null) {
					parameter.setValue(tf.getText());
					setDirty(true);
				}
			}
			return result;
		}

	}

	/**
	 * validates input.
	 **/
	protected boolean isNumberValid(String numberString) {
		try {
			int number = Integer.valueOf(numberString).intValue();
			if (number >= minValue && number <= maxValue) {
				return true;
			} else {
				return false;
			}
		} catch (NumberFormatException nfe) {
			return false;
		}
	}

	public void setValidRange(int minVal, int maxVal) {
		minValue = minVal;
		maxValue = maxVal;
	}

	/**
	 * @see gov.va.med.hds.cd.config.dialog.IFieldEditor#doApply()
	 */
	public void doApply() throws ConfigurationException {
		if (parameter != null) {
			parameter.persist();
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.dialog.IFieldEditor#loadDefault()
	 */
	public void loadDefault() {
		if (parameter != null) {
			String defaultParameterValue = parameter.getDefaultValue();
			try {
				int intValue = Integer.parseInt(defaultParameterValue);
				((WholeNumberField) field).setValue(intValue);
			} catch (NumberFormatException e) {
				((WholeNumberField) field).setText("");
			}
			parameter.setValue(defaultParameterValue);
			setDirty(true);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.config.dialog.IFieldEditor#loadInitialValue()
	 */
	public void loadInitialValue() {
		if (parameter != null) {
			((WholeNumberField) field).setValue(
				Integer.parseInt(parameter.getInitialValue()));
		}
	}

	/**
	 * Method setValue.
	 * @param value
	 */
	public void setValue(String value) {
		if (isNumberValid(value)) {
			((WholeNumberField) field).setValue(Integer.parseInt(value));
		}
	}
	
	/**
	 * Method getValue.
	 * @return String
	 */
	public String getValue() {
		String result = Integer.toString(((WholeNumberField) field).getValue());
		if (isNumberValid(result)) {
			return result;
		} else {
			return "";
		}
	}

}
