/*
 * Created on May 19, 2003
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 */
package gov.va.med.hds.dashboard.dao.m;

import gov.va.med.hds.dashboard.dao.IDashboardSignableDAO;
import gov.va.med.hds.dashboard.items.IDashboardItem;
import gov.va.med.hds.dashboard.itemtype.DashboardSignable;
import gov.va.med.hds.signature.ISignableItem;
import gov.va.med.hds.signature.dao.ISignatureDAO;
import gov.va.med.hds.util.mumps.StringUtil;

import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

/**
 *
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 * 
 * @author vhaislschras
 */
public class DashboardSignableDAO extends AbstractDashboardDAO
  implements IDashboardSignableDAO, ISignatureDAO
{
  protected static final String UNSIGNED_BY_ORDERID = "UNSIGNED BY ID";

  public DashboardSignableDAO()
  { 
    super();
    dbItemsByID = UNSIGNED_BY_ORDERID;
    dbItemsByPatient = null;
  }

  /* (non-Javadoc)
   * @see gov.va.med.hds.dashboard.dao.IDashboardItemDAO#parseItems(java.lang.Object)
   */
  public List parseItems(Object data)
  {
    List signables = new ArrayList();
    DashboardSignable txOrder = null;
    String[] resultData = ((String) data).split("\n");
    for (int l = 0; l < resultData.length; l++)
    {
      String[] line = resultData[l].split("=", 2);
      if ("Item".equalsIgnoreCase(line[0]))
        signables.add((txOrder = (DashboardSignable)parseItem(line[1])));
      else if ("Order".equalsIgnoreCase(line[0]))
        parseOrder(txOrder, line[1]);
      else if ("Text".equalsIgnoreCase(line[0]))
        parseText(txOrder, line[1]);
      else if ("Cmnt".equalsIgnoreCase(line[0]))
        txOrder.addComment(line[1]);
    }
    return signables;
  }

  private IDashboardItem parseItem(String data)
  {
    // The ordered item looks like:
    // [DOC|ORD]:ID^TITLE^YYYYMMDDHHmmss-TZ^[true|false]
    // ORD:ID^^YYYYMMDDHHmmss
    String[] signable = StringUtil.split(data, '^');
    IDashboardItem item = null;
    try
    {
      item = (IDashboardItem)dbItems.get(signable[0]);
      if (null == item)
      {
        item = (IDashboardItem)itemTypes.getItemType(ITEM_TYPE_ID).getInstance();
        item.setID(signable[0]);
      }
      item.setDate((-1 != signable[2].indexOf('.') ? (Object)StringUtil.fmDateToDate(signable[2]) : signable[2]));
      item.setShortTitle((0 != signable[1].length() ? signable[1] : 
        item.getFormattedDate()));
      if (3 < signable.length)
        ((DashboardSignable)item).setRequiresEncounter(Boolean.valueOf(signable[3]).booleanValue());
      item.setDirty(false);
      dbItems.put(item.getID(), item);
      return item;
    }
    catch (Exception e)
    {
      logger.error(e.getMessage(), e);
    }
    return item;
  }

  private void parseText(IDashboardItem item, String data)
  {
    // Text has a single line of text each row;
    StringBuffer text = (StringBuffer) item.getData();
    if (null == text)
      text = (StringBuffer)((DashboardSignable)item)
        .createDataType();
    else
      text.append("\n");
    text.append(data);
    item.setData(text);
  }

  /* (non-Javadoc)
   * @see gov.va.med.hds.signature.dao.ISignatureDAO#sign(Object)
   */
  public void sign(Object signature) throws SignatureException
  {
    List items = getItems(true);
    if (0 < items.size())
    {
      boolean success = true;
      StringBuffer unsuccessful = new StringBuffer();
      List signableItems = new ArrayList();
      for (Iterator iter = items.iterator(); iter.hasNext();)
      {
        ISignableItem item = (ISignableItem)iter.next();
        if (item.isReadyToSign())
          signableItems.add(item);
      }
      try
      {
        String sigCode = mumps.encrypt(signature.toString());
        List params = new ArrayList(4);
        List ids = getIdsFromItems(signableItems);
        params.add(0, provider);
        params.add(1, "");
        params.add(2, sigCode);
        params.add(3, ids);
        String signed = mumps.callRPC(IDashboardSignableDAO.SIGN_RPC, params);
        Iterator signeds = Arrays.asList(signed.trim().split("\n")).iterator();
        ids.clear();
        while (signeds.hasNext())
        {
          String[] item = ((String)signeds.next()).trim().split("\\^");
          if ("1".equals(item[1]))
          {
            DashboardSignable signedItem = (DashboardSignable)getItem(item[0]);
            signedItem.setSigned(true);
            signedItem.setDirty(false);
          }
          else
          {
            success = false;
            unsuccessful.append("\nOrder#" + getOrderNumber(item[0]));
            if (1 < item[1].length())
              unsuccessful.append(" - " + item[1].substring(1));
          }
        }
      }
      catch (Exception e)
      {
        logger.error(e.getLocalizedMessage(), e);
        unsuccessful.append("\n" + e.getMessage());
        success = false;
      }
      if (!success)
        throw new SignatureException(unsuccessful.toString());
    }
  }

  /* (non-Javadoc)
   * @see gov.va.med.hds.signature.dao.ISignatureDAO#getSignatureCandidates()
   */
  public synchronized List getSignatureCandidates()
  {
    List items = getItems(true);
    List candidates = new ArrayList();
    for (Iterator iter = items.iterator(); iter.hasNext();)
    {
      ISignableItem item = (ISignableItem)iter.next();
      if (item.isReadyToSign() && !item.isSigned())
        candidates.add(item);
    }
    return candidates;
  }

  private String getOrderNumber(String id)
  {
    return id.split(":")[1].split(";")[0];
  }

  /* (non-Javadoc)
   * @see gov.va.med.hds.dashboard.dao.IDashboardItemDAO#updateItems(boolean)
   */
  public void updateItems(boolean flush)
  {
    // If flush is true we want all items.
    // If flush is false we want only dirty items.
    List items = getItems(!flush);
    if (0 < items.size())
    {
      List param = new ArrayList();
      for (Iterator iter = items.iterator(); iter.hasNext();)
      {
        IDashboardItem item = (IDashboardItem)iter.next();
        param.add(item.getID());
      }
      updateItems(param, flush);
    }
  }

	/* 
   * Signable items are updated by the calling the sign method from the Signature module.
   * When updateItems is call, we need to remove all items that have been signed or 
   * are not selected to be signed.  Items that are selected to be signed need to stay around
   * because they are known to the Signature module and we don't want them to disappear.
   * (non-Javadoc)
	 * @see gov.va.med.hds.dashboard.dao.IDashboardItemDAO#updateItems(java.util.List, boolean)
	 */
	public void updateItems(List ids, boolean flush)
	{
    for (Iterator iter = ids.iterator(); iter.hasNext();)
    {
      DashboardSignable item = (DashboardSignable)getItem(iter.next());
      if (null == item) continue;

      // If flush is false, then just update the cached items that are signed
      // and dirty to not dirty.  There are probably no such items
      // because signing the items should set dirty to false.
      if (!flush)
      {
				if (item.isDirty() && item.isSigned())
          item.setDirty(false);
      }
      // If flush id true, then remove signed items and items that are
      // not signed and not selected for signing from the cache.
      else
      {
        if (!item.isReadyToSign() || item.isSigned())
          removeFromCache(item.getID());
      }
    }
	}

}
