unit uLauncher;

interface

uses
  SysUtils, Windows, Classes, registry, uUrlReader, uLogger;

type
  TLauncher = class
  private
    FUrl:        string;
    FAllParams:  string;
    FExeParams:  string;
    FJreParams:  string;
    FRunParams:  string;
    FJreMinimum: string;
    FLocalDir:   string;
    FSourceDir:  string;
    FRunJar:     string;
    FFiles:      TStringList;
    FUrlChoices: TStringList;
    FLogger: TLogger;
    function GetCurrentJre: string;
    function GetHighestJre: string;
    procedure LoadJreVersions(AList: TStrings);
    function GetJrePath: string;
    procedure BuildParamList(Dest: TStrings; const CmdString: string);
    procedure ParseParams(ParamList: TStringList);
    procedure SetAllParams(const Value: string);
    procedure SetUrlToRelativeFile(AUrlReader: TUrlReader; RelativeFile: string);
    function GetUpdateCmd: string;
    function GetUrlChoices: TStringList;
  protected
    property LocalDir: string read FLocalDir;
    property SourceDir: string read FSourceDir;
    property RunJar: string read FRunJar;
    property Files: TStringList read FFiles;
    procedure ParseConfig(AConfig: TStringList);
  public
    constructor Create;
    destructor Destroy; override;
    function FilesAreUpdated: Boolean;
    function RunJava: TProcessInformation;
    function UpdateSelf: Boolean;
    property AllParams:  string write SetAllParams;
    property ExeParams:  string read  FExeParams;
    property JreParams:  string read  FJreParams;
    property RunParams:  string read  FRunParams;
    property JrePath:    string read  GetJrePath;
    property JreMinimum: string read  FJreMinimum;
    property Url:        string read  FUrl write FUrl;
    property UpdateCmd:  string read  GetUpdateCmd;
    property UrlChoices: TStringList read GetUrlChoices;
    property Logger: TLogger write FLogger;
  end;

implementation

uses
  uFnString, uFnSystem, FileCtrl, Forms;

const
  JRE_MAIN_KEY        = 'SOFTWARE\JavaSoft\Java Runtime Environment';
  JRE_MINIMUM         = '1.4.1';
  JRE_CURRENT_VERSION = 'CurrentVersion';
  JRE_MICRO_VERSION   = 'MicroVersion';
  JRE_HOME            = 'JavaHome';
  URL_LINKS           = 'configuration\links.settings';

{ TLauncher }

constructor TLauncher.Create;
{ initialize based on the command line parameters that were passed in }
var
  i: Integer;
  ParamList: TStringList;
begin
  FJreMinimum := JRE_MINIMUM;
  ParamList := TStringList.Create;
  for i := 1 to ParamCount do ParamList.Add(ParamStr(i));
  ParseParams(ParamList);
  ParamList.Free;
  FFiles      := TStringList.Create;
  FUrlChoices := TStringList.Create;
end;

destructor TLauncher.Destroy;
begin
  FFiles.Free;
  FUrlChoices.Free;
  inherited;
end;

procedure TLauncher.BuildParamList(Dest: TStrings; const CmdString: string);
var
  i: Integer;
  x: string;
begin
  FLogger.Debug(self, 'BuildParamList start');
  i := 1;
  x := '';
  Dest.Clear;
  while i <= Length(CmdString) do
  begin
    if CmdString[i] = '"' then
    begin
      Inc(i);
      while (i <= Length(CmdString)) and (CmdString[i] <> '"') do
      begin
        x := x + CmdString[i];
        Inc(i);
      end;
      Inc(i);
    end
    else if CmdString[i] = ' ' then
    begin
      Dest.Add(x);
      x := '';
      Inc(i);
    end else
    begin
      x := x + CmdString[i];
      Inc(i);
    end;
  end;
  if Length(x) > 0 then Dest.Add(x);
  FLogger.Debug(self, 'BuildParamList end');
end;

procedure TLauncher.ParseParams(ParamList: TStringList);
var
  i: Integer;
begin
  FUrl       := '';
  FExeParams := '';
  FJreParams := '';
  FRunParams := '';
  for i := 0 to Pred(ParamList.Count) do
  begin
    if CharAt(ParamList[i], 1) <> '-'
      then FUrl := ParamList[i]
    else if Pos('-exe=', LowerCase(ParamList[i])) = 1
      then FExeParams := Copy(ParamList[i], 6, Length(ParamList[i]))
    else if Pos('-jre=', LowerCase(ParamList[i])) = 1
      then FJreParams := Copy(ParamList[i], 6, Length(ParamList[i]))
    else
      FRunParams := FRunParams + '"' + ParamList[i] + '" ';
  end;
  FRunParams := Trim(FRunParams);
end;

procedure TLauncher.SetAllParams(const Value: string);
var
  ParamList: TStringList;
begin
  FAllParams := Value;
  ParamList := TStringList.Create;
  BuildParamList(ParamList, Value);
  ParseParams(ParamList);
  ParamList.Free;
end;

procedure TLauncher.LoadJreVersions(AList: TStrings);
var
  Registry: TRegistry;
begin
  FLogger.Debug(self, 'LoadJreVersions start');
  Registry := TRegistry.Create;
  try
    Registry.RootKey := HKEY_LOCAL_MACHINE;
    if Registry.OpenKeyReadOnly(JRE_MAIN_KEY) then Registry.GetKeyNames(AList);
    Registry.CloseKey;
  finally
    Registry.Free;
  end;
  FLogger.Debug(self, 'LoadJreVersions end');
end;

function TLauncher.GetCurrentJre: string;
begin
  FLogger.Debug(self, 'GetCurrentJre start');
  Result := RegReadStr(JRE_MAIN_KEY, JRE_CURRENT_VERSION);
  FLogger.Debug(self, 'GetCurrentJre end');
end;

function TLauncher.GetHighestJre: string;
var
  i: Integer;
  VersionList: TStringList;
begin
  FLogger.Debug(self, 'GetHighestJre start');
  Result := '';
  VersionList := TStringList.Create;
  try
    LoadJreVersions(VersionList);
    for i := 0 to Pred(VersionList.Count) do
      if CompareVersionJava(VersionList[i], Result) > 0 then Result := VersionList[i];
  finally
    VersionList.Free;
  end;
  FLogger.Debug(self, 'GetHighestJre end');
end;

function TLauncher.GetJrePath: string;
var
  Version: string;
begin
  FLogger.Debug(self, 'GetJrePath start');
  Result := '';
  Version := GetCurrentJre;
  if CompareVersionJava(Version, JreMinimum) < 0 then
  begin
    Version := GetHighestJre;
    if CompareVersionJava(Version, JreMinimum) < 0 then Version := '';
  end;
  if Length(Version) > 0
    then Result := RegReadStr(JRE_MAIN_KEY + '\' + Version, JRE_HOME);
  FLogger.Debug(self, 'GetJrePath end');
end;

procedure TLauncher.ParseConfig(AConfig: TStringList);
var
  i: Integer;
begin
  FLogger.Debug(self, 'ParseConfig start');
  i := 0;
  while (i < AConfig.Count) and (Pos('<startup ', AConfig[i]) = 0) do Inc(i);
  if i >= AConfig.Count then Exit;
  FLocalDir := StringBetween(AConfig[i], 'plugin="', '"');
  FSourceDir := StringBetween(AConfig[i], 'source="', '"');
  if(FSourceDir = '') then
    FSourceDir := FLocalDir;
  FRunJar   := StringBetween(AConfig[i], 'run="',    '"');
  while (i < AConfig.Count) and (Pos('</files>', AConfig[i]) = 0) do
  begin
    if Pos('<file>', AConfig[i]) > 0
      then FFiles.Add(StringBetween(AConfig[i], '<file>', '</file>'));
    Inc(i);
  end;
  FLogger.Debug(self, 'ParseConfig end');
end;

function TLauncher.FilesAreUpdated: Boolean;
var
  i: Integer;
  aPath, aFileName, forceFileName: string;
  config: TStringList;
  urlReader: TUrlReader;
  forceUpdate: boolean;

begin
  FLogger.Debug(self, 'FilesAreUpdated start');
  Result := False;
  config    := TStringList.Create;
  urlReader := TUrlReader.Create;
  urlReader.Logger := FLogger;
  urlReader.Url := FUrl;
  urlReader.LoadToStrings(config);
  ParseConfig(config);
  if (Length(FRunJar) = 0) or (Length(FLocalDir) = 0) then Exit;
  if not DirectoryExists('plugins')              then CreateDir('plugins');
  if not DirectoryExists('plugins\' + FLocalDir) then CreateDir('plugins\' + FLocalDir);
  aPath := urlReader.Path;
  forceFileName := ExpandFileName('plugins\' + FLocalDir + '\force.update');
  forceUpdate := FileExists(forceFileName);
  for i := 0 to Pred(FFiles.Count) do
  begin
    aFileName := ExpandFileName('plugins\' + FLocalDir + '\' + FFiles[i]);
    if (forceUpdate or (not FileExists(aFileName))) then
    begin
      if (Copy(aPath, 1, 4) = 'http') or (Copy(aPath, 1, 3) = 'ftp')
        then urlReader.Url := aPath + 'plugins/' + FSourceDir + '/' + FFiles[i]
        else urlReader.Url := aPath + 'plugins\' + FSourceDir + '\' + FFiles[i];
      urlReader.CopyToFile(aFileName);
    end;
  end;
  if(forceUpdate) then
    SysUtils.DeleteFile(forceFileName);
  Result := True;
  FLogger.Debug(self, 'FilesAreUpdated end');
end;

const
  TITLE: PChar = PChar('HealtheVet Desktop');

function TLauncher.RunJava: TProcessInformation;
var
  StartInfo: TStartupInfo;
//  ProcInfo: TProcessInformation;
  jreExe, jreCmd, cmd: string;
begin
  FLogger.Debug(self, 'RunJava start');
  if FRunJar = '' then Exit;
  if FLocalDir = '' then Exit;
  jreExe := GetJrePath + '\bin\javaw.exe';
  jreCmd := '-jar "plugins\' + FLocalDir + '\' + FRunJar + '" "' + FUrl + '" ' + FRunParams;
  FillChar(StartInfo, SizeOf(StartInfo), 0);
  StartInfo.CB := SizeOf(StartInfo);
  cmd := jreExe + ' ' + jreCmd;
  StartInfo.lpTitle := @TITLE;
  CreateProcess(nil, PChar(cmd), nil, nil, False,
    DETACHED_PROCESS or NORMAL_PRIORITY_CLASS, nil, nil, StartInfo, Result);
  FLogger.Debug(self, 'JRE Parms:'+cmd);
// *** The above form was used to fix an unknown error that caused the -jar to not be passed
// *** to the JVM.  Not sure why this happened, verious forms of arguments were tired, none
// *** of them worked - it seems to have something to do with the CreateProcess API stripping
// *** off parameters starting with a - at the first of the command.
//  CreateProcess(PChar(jreExe), PChar(jreCmd), nil, nil, False,
//    DETACHED_PROCESS or NORMAL_PRIORITY_CLASS, nil, nil, StartInfo, ProcInfo);
  FLogger.Debug(self, 'RunJava end');
end;

procedure TLauncher.SetUrlToRelativeFile(AUrlReader: TUrlReader; RelativeFile: string);
var
  serverPath: string;
begin
  FLogger.Debug(self, 'SetUrlToRelativeFile start');
  AUrlReader.Url := FUrl;               // reset to original path
  serverPath := AUrlReader.Path;
  if (Copy(serverPath, 1, 4) = 'http') or (Copy(serverPath, 1, 3) = 'ftp')
    then StringReplace(RelativeFile, '\', '/', [rfReplaceAll])
    else StringReplace(RelativeFile, '/', '\', [rfReplaceAll]);
  AUrlReader.Url := serverPath + RelativeFile;
  FLogger.Debug(self, 'SetUrlToRelativeFile end');
end;

function TLauncher.UpdateSelf: Boolean;
var
  exeSettings: TStringList;
  urlReader: TUrlReader;
  serverVersion, localVersion, localExe: string;
begin
  FLogger.Debug(self, 'UpdateSelf start');
  Result := FALSE;
  if Pos('noupdate', FExeParams) > 0 then Exit;
  exeSettings := TStringList.Create;
  urlReader := TUrlReader.Create;
  urlReader.Logger := FLogger;
  localExe := FullToFilePart(Application.ExeName);
  SetUrlToRelativeFile(urlReader, 'client/' + Piece(localExe, '.', 1) + '.settings');
  urlReader.LoadToStrings(exeSettings);
  serverVersion := exeSettings.Values['version'];
  localVersion := ClientVersion(Application.ExeName);
  if CompareVersionJava(serverVersion, localVersion) > 0 then
  begin
    SetUrlToRelativeFile(urlReader, 'client/' + localExe);
    urlReader.CopyToFile(Piece(localExe, '.', 1) + '.tmp');
    SetUrlToRelativeFile(urlReader, 'client/RunUpdated.exe');
    urlReader.CopyToFile('RunUpdated.exe');
    Result := TRUE;
  end;
  FLogger.Debug(self, 'UpdateSelf end');
end;

function TLauncher.GetUpdateCmd: string;
begin
  FLogger.Debug(self, 'GetUpdateCmd start');
  Result := '';
  if FileExists('RunUpdated.exe') then
  begin
    Result := 'RunUpdated.exe ';
    Result := Result + Piece(FullToFilePart(Application.ExeName), '.', 1);
    Result := Result + ' "' + FUrl + '"';
    if Length(FExeParams) > 0
      then Result := Result + ' -exe="' + FExeParams + ' noupdate"'
      else Result := Result + ' -exe="noupdate"';
    if Length(FJreParams) > 0
      then Result := Result + ' -jre="' + FJreParams + '"';
    if Length(FRunParams) > 0
      then Result := Result + ' ' + FRunParams;
  end;
  FLogger.Debug(self, 'GetUpdateCmd end');
end;

function TLauncher.GetUrlChoices: TStringList;
begin
  FLogger.Debug(self, 'GetUrlChoices start');
  if FileExists('link.settings') then FUrlChoices.LoadFromFile('link.settings');
  Result := FUrlChoices;
  FLogger.Debug(self, 'GetUrlChoices end');
end;

end.
