unit uUrlReader;

interface

uses
  SysUtils, Classes, Windows, forms, WinInet, uLogger;

type
  TUrlReader = class
  private
    FErrorMsg: string;
    FFileName: string;
    FUrl: string;
    FLogger: TLogger;
    procedure InetCopyToFile(const TheFile: string);
    procedure FileCopyToFile(const TheFile: string);
    procedure InetLoadToStrings(Dest: TStrings);
    procedure FileLoadToStrings(Dest: TStrings);
    procedure SetUrl(const Value: string);
    function GetPath: string;
  public
    constructor Create;
    destructor Destroy; override;
    procedure CopyToFile(const TheFile: string);
    procedure LoadToStrings(Dest: TStrings);
    property ErrorMsg: string read FErrorMsg;
    property Path: string read GetPath;
    property Url: string read FUrl write SetUrl;
    property Logger: TLogger write FLogger;
  end;

implementation

uses uFnString, FileCtrl;

const
  AGENT_NAME    = 'HealtheVet Desktop';
  TX_ERR_NOFILE = 'File not found: ';
  TX_ERR_COPY   = 'File copy failed';
  TX_ERR_HTTP   = 'HTTP file copy failed';

{ TUrlReader -- private methods }

procedure TUrlReader.FileCopyToFile(const TheFile: string);
{ copies from a OS file source to a local file }
var
  fullFileName: string;
begin
  FLogger.Debug(self, 'FileCopyToFile start; file='+TheFile);
  fullFileName := ExpandFileName(FFileName);
  if not CopyFile(PChar(fullFileName), PChar(TheFile), FALSE) then FErrorMsg := TX_ERR_COPY;
  FLogger.Debug(self, 'FileCopyToFile end; file='+TheFile);
end;

procedure TUrlReader.FileLoadToStrings(Dest: TStrings);
{ loads an OS file source into TStrings }
begin
  FLogger.Debug(self, 'FileLoadToStrings start');
  Dest.LoadFromFile(FFileName);
  FLogger.Debug(self, 'FileLoadToStrings end');
end;

procedure TUrlReader.InetCopyToFile(const TheFile: string);
{ copies a file via HTTP or FTP to a local file }
const
  BUFFER_SIZE = 1024;
var
  hSession, hURL: HInternet;
  Buffer: array[1..BUFFER_SIZE] of Byte;
  BufferLen: DWORD;
  f: File;
begin
  FLogger.Debug(self, 'InetCopyToFile start; file='+TheFile);
  //hSession := InternetOpen(PChar(sAppName), INTERNET_OPEN_TYPE_PRECONFIG, nil, nil, INTERNET_FLAG_OFFLINE);
  hSession := InternetOpen(AGENT_NAME, INTERNET_OPEN_TYPE_PRECONFIG, nil, nil, 0);
  try
    FErrorMsg := TX_ERR_HTTP;
    hURL := InternetOpenURL(hSession, PChar(FUrl), nil, 0, 0, 0);
    if Assigned(hURL) then
    begin
      try
        AssignFile(f, TheFile);
        Rewrite(f,1);
        repeat
          InternetReadFile(hURL, @Buffer, SizeOf(Buffer), BufferLen);
          BlockWrite(f, Buffer, BufferLen);
        until BufferLen = 0;
        CloseFile(f);
        FErrorMsg := '';
      finally
        InternetCloseHandle(hURL);
      end; {try}
    end; {if Assigned}
  finally
    InternetCloseHandle(hSession);
  end;
  FLogger.Debug(self, 'InetCopyToFile end; file='+TheFile);
end;

procedure TUrlReader.InetLoadToStrings(Dest: TStrings);
{ loads a file via HTTP or FTP into TStrings }
const
  BUFFER_SIZE = 1024;
var
  i: Integer;
  x: string;
  hSession, hURL: HInternet;
  Buffer: array[1..BUFFER_SIZE] of Byte;
  BufferLen: DWORD;
begin
  FLogger.Debug(self, 'InetLoadToStrings start');
  hSession := InternetOpen(AGENT_NAME, INTERNET_OPEN_TYPE_PRECONFIG, nil, nil, 0);
  try
    FErrorMsg := TX_ERR_HTTP;
    hURL := InternetOpenURL(hSession, PChar(FUrl), nil, 0, 0, 0);
    if Assigned(hURL) then
    begin
      try
        x := '';
        repeat
          InternetReadFile(hURL, @Buffer, SizeOf(Buffer), BufferLen);
          for i := 1 to BufferLen do x := x + Chr(Buffer[i]);
        until BufferLen = 0;
        Dest.Text := x;
        FErrorMsg := '';
      finally
        InternetCloseHandle(hURL);
      end; {try}
    end; {if Assigned}
  finally
    InternetCloseHandle(hSession);
  end;
  FLogger.Debug(self, 'InetLoadToStrings end');
end;

procedure TUrlReader.SetUrl(const Value: string);
{ save URL, convert to file name if this is a file: URL }
var
  protocol: string;
begin
  FLogger.Debug(self, 'SetUrl start; value='+Value);
  FUrl := value;
  FFileName := '';
  FErrorMsg := '';
  protocol := '';
  if Pos(':/', Value) > 0 then protocol := Copy(Value, 1, Pos(':/', Value)+1);
  if (protocol = 'ftp:/') or (protocol = 'http:/') then Exit;
  // no further parsing necessary for FTP or HTTP protocols
  FFileName := Value;
  if protocol = 'file:/' then
  begin
    FFileName := Copy(FFileName, 9, Length(FFileName));
    FFileName := StringReplace(FFileName, '/', '\', [rfReplaceAll]);
  end;
  if not FileExists(FFileName) then FErrorMsg := TX_ERR_NOFILE + FUrl;
  FLogger.Debug(self, 'SetUrl end; value='+Value);
end;

function TUrlReader.GetPath: string;
var
  i: Integer;
begin
  FLogger.Debug(self, 'GetPath start');
  if Length(FFileName) > 0 then
  begin
    for i := Length(FFileName) downto 1 do if FFileName[i] = '\' then break;
    Result := Copy(FFileName, 1, i);
  end else
  begin
    for i := Length(FUrl)      downto 1 do if FUrl[i]      = '/' then break;
    Result := Copy(FUrl, 1, i);
  end;
  FLogger.Debug(self, 'GetPath end');
end;

{ TUrlReader -- public methods }

constructor TUrlReader.Create;
begin

end;

destructor TUrlReader.Destroy;
begin
  inherited;

end;

procedure TUrlReader.CopyToFile(const TheFile: string);
begin
  if Length(FErrorMsg) > 0 then Exit;
  if Length(FFileName) > 0
    then FileCopyToFile(TheFile)
    else InetCopyToFile(TheFile);
end;

procedure TUrlReader.LoadToStrings(Dest: TStrings);
begin
  if Length(FErrorMsg) > 0 then Exit;
  if Length(FFileName) > 0
    then FileLoadToStrings(Dest)
    else InetLoadToStrings(Dest);
end;

end.
