unit uLauncherTest;

interface

uses
  SysUtils, Classes, TestFrameWork, uLauncher;

type
  // We use a Test Launcher to gain access to the protected classes inside TLauncher
  TTestLauncher = class (TLauncher);

  TLauncherTest = class(TTestCase)
  private
    FLauncher: TTestLauncher;
  protected
    procedure Setup; override;
    procedure TearDown; override;
  published
    procedure TestCreate;
    procedure TestParse;
    procedure TestJREPath;
    procedure TestURLChoices;
    procedure TestSetAllParams;
    procedure TestUpdateFiles;
    procedure TestRunJava;
    procedure TestUpdatingSelf;
  end;

implementation

{ TLauncherTest }

procedure TLauncherTest.Setup;
begin
  inherited;
  FLauncher := TTestLauncher.Create;
end;

procedure TLauncherTest.TearDown;
begin
  inherited;
  FLauncher.Free;
end;

procedure TLauncherTest.TestCreate;
begin
  Check(FLauncher <> nil);
  Check(FLauncher.JREMinimum = '1.4.1');
  // the following assumes this was started with the following command line parameters:
  // test\files\testconfig.xml -exe="pe1 pe2" -jre="pj1 pj2" -pb1 -pb2
  Check(FLauncher.Url = 'test\files\testconfig.xml');
  Check(FLauncher.ExeParams = 'pe1 pe2', 'exe params');
  Check(FLauncher.JreParams = 'pj1 pj2', 'jre params');
  Check(FLauncher.RunParams = '"-pb1" "-pb2"', 'run params:' + FLauncher.RunParams);
end;

procedure TLauncherTest.TestJREPath;
begin
  // this assumes the path to the JRE begins with C:\Program Files\Java\j2re1.4.1_02
  Check(Pos('j2re1.4.1', FLauncher.JrePath) > 0, FLauncher.JrePath);
end;

procedure TLauncherTest.TestParse;
var
  list: TStringList;
  sIndex: integer;
  fIndex: integer;

begin
  list := TStringList.Create;
  try
    list.Add('<?xml version="1.0" encoding="UTF-8"?>');
    list.Add('<configuration crc="12345">');
    sIndex := list.Add('  <startup plugin="a.b.c.d_1.2.3" run="testing.jar" source="testSource">');
    list.Add('    <files>');
    list.Add('      <file>testing.one</file>');
    list.Add('      <file>two.three</file>');
    fIndex := list.Add('');
    list.Add('    </files>');
    list.Add('  </startup>');
    list.Add('</configuration>');
    FLauncher.ParseConfig(list);
    CheckEquals('a.b.c.d_1.2.3', FLauncher.LocalDir, 'Failed to parse Local Directory');
    CheckEquals('testSource', FLauncher.SourceDir, 'Failed to parse Source Directory');
    CheckEquals('testing.jar', FLauncher.RunJar, 'Failed to parse Run Jar');
    CheckEquals(2, FLauncher.Files.Count, 'Failed to parse all Files');
    Check(FLauncher.Files.IndexOf('testing.one') >= 0, 'Failed to find first file');
    Check(FLauncher.Files.IndexOf('two.three') >= 0, 'Failed to find second file');
    list[sIndex] := '<startup plugin="plugin.two" run="AnotherTest.jar">';
    list[fIndex] := '<file>some.unknown.file</file>';
    FLauncher.Files.Clear;
    FLauncher.ParseConfig(list);
    CheckEquals('plugin.two', FLauncher.LocalDir, 'Failed to parse Local Directory');
    CheckEquals('plugin.two', FLauncher.SourceDir, 'Failed to parse Source Directory');
    CheckEquals('AnotherTest.jar', FLauncher.RunJar, 'Failed to parse Run Jar');
    CheckEquals(3, FLauncher.Files.Count, 'Failed to parse all Files');
    Check(FLauncher.Files.IndexOf('testing.one') >= 0, 'Failed to find first file');
    Check(FLauncher.Files.IndexOf('two.three') >= 0, 'Failed to find second file');
    Check(FLauncher.Files.IndexOf('some.unknown.file') >= 0, 'Failed to find third file');
  finally
    list.free;
  end;
end;

procedure TLauncherTest.TestRunJava;
begin
//  Check(FLauncher.FilesAreUpdated);
//  FLauncher.RunJava;
end;

procedure TLauncherTest.TestSetAllParams;
begin

end;

procedure TLauncherTest.TestUpdateFiles;
begin
  Check(FLauncher.FilesAreUpdated);
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\boot.jar'), 'no boot.jar');
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\mum.jar'), 'no mum.jar');
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\manifest.xml'), 'no manifest.xml');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\boot.jar');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\mum.jar');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\manifest.xml');
  FLauncher.AllParams := 'http://localhost/testconfig.xml';
  Check(FLauncher.FilesAreUpdated);
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\boot.jar'), 'no boot.jar http');
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\mum.jar'), 'no mum.jar http');
  Check(FileExists('plugins\gov.va.med.hds.cd.mum_0.1.2\manifest.xml'), 'no manifest.xml http');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\boot.jar');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\mum.jar');
  DeleteFile('plugins\gov.va.med.hds.cd.mum_0.1.2\manifest.xml');
end;

procedure TLauncherTest.TestUpdatingSelf;
var
  aList: TStringList;
begin
  DeleteFile('aLaunchTest.tmp');
  DeleteFile('RunUpdated.exe');
  aList := TStringList.Create;
  aList.Add('exeversion 1.0.1');
  aList.SaveToFile('test\files\client\aLaunchTest.exe');
  aList.Clear;
  aList.Add('version=1.0.1');
  aList.SaveToFile('test\files\client\aLaunchTest.settings');
  aList.Clear;
  Check(FLauncher.UpdateSelf = FALSE, 'should not expect update');
  Check(FileExists('aLaunchTest.tmp') = FALSE, 'aLaunchTest should not be local');
  Check(FileExists('RunUpdated.exe')  = FALSE, 'RunUpdated should not be local');
  //
  aList := TStringList.Create;
  aList.Add('exeversion 1.0.2');
  aList.SaveToFile('test\files\client\aLaunchTest.exe');
  aList.Clear;
  aList.Add('version=1.0.2');
  aList.SaveToFile('test\files\client\aLaunchTest.settings');
  aList.Clear;
  Check(FLauncher.UpdateSelf = TRUE, 'update expected');
  Check(FileExists('aLaunchTest.tmp'), 'local aLaunchTest missing');
  Check(FileExists('RunUpdated.exe'), 'local RunUpdated missing');
  aList.LoadFromFile('aLaunchTest.tmp');
  Check(aList[0] = 'exeversion 1.0.2', aList[0]);
  Check(FLauncher.UpdateCmd =
    'RunUpdated.exe aLaunchTest "test\files\testconfig.xml" -exe="pe1 pe2 noupdate" -jre="pj1 pj2" "-pb1" "-pb2"',
     FLauncher.UpdateCmd);
  DeleteFile('aLaunchTest.tmp');
  DeleteFile('RunUpdated.exe');
  aList.Free;
end;

procedure TLauncherTest.TestURLChoices;
var
  aList: TStringList;
begin
  DeleteFile('link.settings');
  Check(FLauncher.UrlChoices.Count = 0, 'should not have url choices');
  aList := TStringList.Create;
  aList.Add('Production System = http://prodserver/config.sys');
  aList.Add('Test System = http://testserver/testconfig.sys');
  aList.SaveToFile('link.settings');
  Check(FLauncher.UrlChoices.Count = 2,
    'choices count = ' + IntToStr(FLauncher.UrlChoices.Count));
  Check(FLauncher.UrlChoices[1] = 'Test System = http://testserver/testconfig.sys',
    'wrong choice text');
  DeleteFile('link.settings');
  aList.Free;
end;

initialization
 TestFramework.RegisterTest(TLauncherTest.Suite);

end.
