unit uUrlReaderTest;

interface

uses
  SysUtils, Classes, TestFrameWork, uUrlReader;

type
  TUrlReaderTest = class(TTestCase)
  private
    FUrlReader: TUrlReader;
    procedure CreateTempFile(const AFileName: string);
  protected
    procedure Setup; override;
    procedure TearDown; override;
  published
    procedure TestCreate;
    procedure TestSetUrl;
    procedure TestLoadUrl;
    procedure TestCopyUrl;
    procedure TestReadPath;
  end;

implementation

const
  FILE_TEMP    = 'UrlReaderTestTemp.xml';
  FILE_SRC     = 'UrlReaderTestSrc.xml';
  FILE_SRC_URL = 'file:///UrlReaderTestSrc.xml';
  FILE_DEST    = 'UrlReaderTestDest.xml';
  HTTP_SRC_URL = 'http://localhost/UrlReaderTestSrc.txt';

{ TUrlReaderTest }

procedure TUrlReaderTest.Setup;
begin
  inherited;
  FUrlReader := TUrlReader.Create;
end;

procedure TUrlReaderTest.TearDown;
begin
  inherited;
  FUrlReader.Free;
end;

procedure TUrlReaderTest.CreateTempFile(const AFileName: string);
var
  temp: TStringList;
begin
  temp := TStringList.Create;
  temp.Add('line 1');
  temp.Add('line 2');
  temp.Add('line 3');
  temp.SaveToFile(AFileName);
  temp.Free;
end;

procedure TUrlReaderTest.TestCreate;
begin
  Check(FUrlReader <> nil);
  Check(FUrlReader.Url = '');
  Check(FUrlReader.ErrorMsg = '');
end;

procedure TUrlReaderTest.TestSetUrl;
begin
  CreateTempFile(FILE_TEMP);
  FUrlReader.Url := FILE_TEMP;
  Check(FUrlReader.Url = FILE_TEMP);
  Check(FUrlReader.ErrorMsg = '');
  Check(DeleteFile(FILE_TEMP), 'Unable to delete file');
  FUrlReader.Url := FILE_TEMP;
  Check(FUrlReader.Url = FILE_TEMP);
  Check(Length(FUrlReader.ErrorMsg) > 0, 'Should have "File not found" error');
end;

procedure TUrlReaderTest.TestCopyUrl;
var
  ListSrc, ListDest: TStringList;
begin
  // URL is local file name
  CreateTempFile(FILE_SRC);
  FUrlReader.Url := FILE_SRC;
  FUrlReader.CopyToFile(FILE_DEST);
  Check(FUrlReader.ErrorMsg = '', FUrlReader.ErrorMsg);
  ListSrc  := TStringList.Create;
  ListSrc.LoadFromFile(FILE_SRC);
  ListDest := TStringList.Create;
  ListDest.LoadFromFile(FILE_DEST);
  Check(ListSrc.Text = ListDest.Text, 'files differ');
  DeleteFile(FILE_SRC);
  DeleteFile(FILE_DEST);
  // URL is file:/ protocol
  CreateTempFile(FILE_SRC);
  FUrlReader.Url := FILE_SRC_URL;
  FUrlReader.CopyToFile(FILE_DEST);
  Check(FUrlReader.ErrorMsg = '', FUrlReader.ErrorMsg);
  ListSrc  := TStringList.Create;
  ListSrc.LoadFromFile(FILE_SRC);
  ListDest := TStringList.Create;
  ListDest.LoadFromFile(FILE_DEST);
  Check(ListSrc.Text = ListDest.Text, 'files differ');
  DeleteFile(FILE_SRC);
  DeleteFile(FILE_DEST);
  // URL is http:/ protocol
  FUrlReader.Url := HTTP_SRC_URL;
  FUrlReader.CopyToFile(FILE_DEST);
  Check(FUrlReader.ErrorMsg = '', FUrlReader.ErrorMsg);
  CreateTempFile(FILE_SRC);
  ListSrc  := TStringList.Create;
  ListSrc.LoadFromFile(FILE_SRC);
  ListDest := TStringList.Create;
  ListDest.LoadFromFile(FILE_DEST);
  Check(ListSrc.Text = ListDest.Text, 'files differ');
  DeleteFile(FILE_SRC);
  DeleteFile(FILE_DEST);
end;

procedure TUrlReaderTest.TestLoadUrl;
var
  ListSrc, ListDest: TStringList;
begin
  // URL is local file name
  CreateTempFile(FILE_SRC);
  ListDest := TStringList.Create;
  FUrlReader.Url := FILE_SRC;
  FUrlReader.LoadToStrings(ListDest);
  Check(FUrlReader.ErrorMsg = '', FUrlReader.ErrorMsg);
  ListSrc  := TStringList.Create;
  ListSrc.LoadFromFile(FILE_SRC);
  Check(ListSrc.Text = ListDest.Text, 'files differ');
  DeleteFile(FILE_SRC);
  // URL is local file name
  ListDest := TStringList.Create;
  FUrlReader.Url := HTTP_SRC_URL;
  FUrlReader.LoadToStrings(ListDest);
  Check(FUrlReader.ErrorMsg = '', FUrlReader.ErrorMsg);
  CreateTempFile(FILE_SRC);
  ListSrc  := TStringList.Create;
  ListSrc.LoadFromFile(FILE_SRC);
  Check(ListSrc.Text = ListDest.Text, 'files differ');
  DeleteFile(FILE_SRC);
end;

procedure TUrlReaderTest.TestReadPath;
begin
  FUrlReader.Url := FILE_SRC;
  Check(FUrlReader.Path = '', FUrlReader.Path);
  FUrlReader.Url := FILE_SRC_URL;
  Check(FUrlReader.Path = '', FUrlReader.Path);
  FUrlReader.Url := HTTP_SRC_URL;
  Check(FUrlReader.Path = 'http://localhost/', FUrlReader.Path);
end;

initialization
 TestFramework.RegisterTest(TUrlReaderTest.Suite);

end.
