package gov.va.med.hds.cd.m.internal;
import gov.va.med.hds.cd.m.CallException;
import gov.va.med.hds.cd.m.ConnectException;
import gov.va.med.hds.cd.m.CreateException;
import gov.va.med.hds.cd.m.IBroker;
import gov.va.med.hds.cd.m.SetContextException;
import gov.va.med.hds.cd.m.internal.broker.jni.COMException;
import gov.va.med.hds.cd.m.internal.broker.jni.Connector;
import gov.va.med.hds.cd.m.internal.common.Delimiters;
import gov.va.med.hds.cd.security.UserCancelledAuthenticationException;
import gov.va.med.hds.util.DisposeException;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

/**
 * Provides an RPC Shared Broker based implementation of the IBroker interface.
 * 
 * Broker is a singleton.
 */
class Broker implements IBroker {
	private static final int PROGRAMMER_DEBUG_CONNECT_TIME_LIMIT_PADDING = 120;
	/**
	 * Returns a unique instance of the Broker.
	 * 
	 * @return IBroker
	 * @throws CreateException
	 */
	public static IBroker getInstance() throws CreateException {
		if (instance == null) {
			instance = new Broker();
		}
		instance.initialize();
		return instance;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setLogin(String)
	 */
	public void setLogin(String login) {
		if (login != null && login.length() > 0) {
			setSilentLogin(true);
		} else {
			setSilentLogin(false);
		}
		this.login = login;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setPassword(String)
	 */
	public void setPassword(String password) {
		this.password = password;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setDivision(String)
	 */
	public void setDivision(String division) {
		this.division = division;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setSilentLogin(boolean)
	 */
	public void setSilentLogin(boolean silent) {
		this.silentLogin = silent;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setConnectTimeLimit(int)
	 */
	public void setConnectTimeLimit(int connectTimeLimitInSeconds) {
		this.connectTimeLimit = connectTimeLimitInSeconds;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setServerName(String)
	 */
	public void setServerName(String serverName) {
		this.serverName = serverName;
	}

	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setServerPort(int)
	 */
	public void setServerPort(int serverPort) {
		this.serverPort = serverPort;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IConnection#connect()
	 */
	public void connect() throws ConnectException, UserCancelledAuthenticationException {
		String servPort = serverName + ":" + String.valueOf(serverPort);
		String loginString = null;
		if (silentLogin)
			loginString = connector.createLoginStr(login, password, division);

		long timeLimit = connectTimeLimit;
		if (wantDebug)
			timeLimit += PROGRAMMER_DEBUG_CONNECT_TIME_LIMIT_PADDING;
		try {
			connector.connect(
				clientName,
				servPort,
				timeLimit,
				!silentLogin,
				loginString,
				wantDebug);
		} catch (gov.va.med.hds.cd.m.internal.broker.jni.UserCancelledAuthenticationException e) {
			throw new gov.va.med.hds.cd.security.UserCancelledAuthenticationException();
		} catch (Exception e) {
			throw new ConnectException(e);
		}
		connected = true;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setRPCTimeLimit(int)
	 */
	public void setRPCTimeLimit(int rpcTimeLimitInMillis) {
		this.rpcTimeLimit = rpcTimeLimitInMillis;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setContext(String)
	 */
	public void setContext(String context) throws SetContextException {
		try {
			connector.setContext(context);
		} catch (Exception e) {
			throw new SetContextException(e.getMessage(), e);
		}
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setClientName(String)
	 */
	public void setClientName(String clientName) {
		this.clientName = clientName;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#getWantDebug()
	 */
	public boolean getWantDebug() {
		return wantDebug;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#setWantDebug(boolean)
	 */
	public void setWantDebug(boolean wantDebug) {
		this.wantDebug = wantDebug;
	}
	/**
	 * @see gov.va.med.hds.cd.m.broker.IBroker#call(String, Object)
	 */
	public String call(String rpcName, Collection params) throws CallException {
		if (!connected)
			throw new CallException("Broker is not connected. Please connect before calling.");
		String result = null;
		try {
			result = connector.call(rpcName, createParamString(params), rpcTimeLimit);
		} catch (Exception e) {
			throw new CallException(e);
		}
		if (result != null) {
			if (result.endsWith(TRIM_END)) {
				result = result.substring(0, result.length() - 2);
			}
		}

		return result;
	}
	/**
	 * @see gov.va.med.hds.common.IDisposable#dispose()
	 */
	public void dispose() throws DisposeException {
		if (initialized == true) {
			connector.dispose();
			connector = null;
			connected = false;
			initialized = false;
		}
	}
	protected void finalize() throws DisposeException {
		dispose();
	}
	/**
	 * Constructor for Broker.
	 */
	protected Broker() throws CreateException {
		initialize();
	}
	private void initialize() throws CreateException {
		if (connector == null) {
			try {
				connector = new Connector();
				connected = false;
			} catch (COMException e) {
				throw new CreateException(e);
			}
			initialized = true;
		}
	}
	private String createParamString(Collection params) {
		if (params == null)
			return "";
		StringBuffer str = new StringBuffer();
		Iterator it = params.iterator();
		Object o;
		while (it.hasNext()) {
			o = it.next();
			if (o instanceof Map) {
				writeMapToBuffer(str, (Map) o);
			} else if (o instanceof Collection) {
				writeCollectionToBuffer(str, (Collection) o);
			} else {
				writeValueToBuffer(str, o.toString());
			}
			str.append(Delimiters.SEP_GS);
		}
		return str.toString();
	}
	/**
	 * Method writeValueToBuffer.
	 * @param str
	 * @param string
	 */
	private void writeValueToBuffer(StringBuffer str, String value) {
		str.append(Delimiters.LITERAL_PARAM_MARKER);
		str.append(Delimiters.SEP_FS);
		str.append(value);
	}
	/**
	 * Method writeCollectionToBuffer.
	 * @param str
	 * @param collection
	 */
	private void writeCollectionToBuffer(StringBuffer str, Collection collection) {
		str.append(Delimiters.LIST_PARAM_MARKER);
		str.append(Delimiters.SEP_FS);
		Iterator it = collection.iterator();
		Object o;
		int subscript = 1;
		while (it.hasNext()) {
			o = it.next();
			str.append(subscript);
			str.append(Delimiters.SEP_US);
			str.append(o.toString());
			str.append(Delimiters.SEP_RS);
			subscript++;
		}
	}
	/**
	 * Method writeMapToBuffer.
	 * @param str
	 * @param map
	 */
	private void writeMapToBuffer(StringBuffer str, Map map) {
		str.append(Delimiters.LIST_PARAM_MARKER);
		str.append(Delimiters.SEP_FS);
		Iterator it = map.keySet().iterator();
		Object o;
		while (it.hasNext()) {
			o = it.next();
			str.append("\"" + o.toString() + "\"");
			str.append(Delimiters.SEP_US);
			str.append(map.get(o).toString());
			str.append(Delimiters.SEP_RS);
		}
	}
	private static final String TRIM_END = "\r\n";
	private static final String DEFAULT_CLIENT_NAME = "Broker";
	private Connector connector = null;
	private boolean initialized = false;
	private String login = null;
	private String password = null;
	private String division = null;
	private String clientName = DEFAULT_CLIENT_NAME;
	private String serverName = null;
	private int serverPort = 0;
	private boolean silentLogin = false;
	private int connectTimeLimit = 0;
	private int rpcTimeLimit = 0;
	private static Broker instance = null;
	private boolean connected = false;
	private boolean wantDebug = false;
}
