package gov.va.med.hds.cd.m.internal;

import gov.va.med.hds.cd.m.IConnectRequest;

/**
 * Implementation of a connect request that connects to a server with possible login and password
 */
class BrokerConnectRequest implements IConnectRequest {
	private static final int DEFAULT_TIME_LIMIT = 120;
	/**
	 * Connect request given server name and server port. 
	 * This request implies that the user will be prompted interactively
	 * for any identity credentials.
	 * 
	 * @param serverName server name
	 * @param serverPort port on the server which to connect
	 */
	public BrokerConnectRequest(String serverName, int serverPort) {
		setServerName(serverName);
		setServerPort(serverPort);
	}
	/**
	 * Connect request given server name, server port, and connect time limit.
	 * This request implies that the user will be prompted interactively for any
	 * identity credentials.
	 * 
	 * @param serverName
	 * @param serverPort
	 * @param timeLimit in seconds
	 */
	public BrokerConnectRequest(String serverName, int serverPort, int timeLimit) {
		setServerName(serverName);
		setServerPort(serverPort);
		this.timeLimit = timeLimit;
	}
	/**
	 * This connect request implies that the user will be promted interactively
	 * for any identity credentials. It also allows for programmer level debugging.
	 * 
	 * @param serverName server name
	 * @param serverPort port on the server which to connect
	 * @param wantProgrammerDebug set to true if the programmer needs to debug across
	 * the connection. Set to false when programmer level debugging is not neccesary.
	 */
	public BrokerConnectRequest(String serverName, int serverPort, boolean wantProgrammerDebug) {
		setServerName(serverName);
		setServerPort(serverPort);
		setWantDebug(wantProgrammerDebug);
	}
	/**
	 * This connect request implies that the user will be promted interactively
	 * for any identity credentials. It also allows for programmer level debugging.
	 * Allow for connect time limit as a parameter.
	 * 
	 * @param serverName
	 * @param serverPort
	 * @param wantProgrammerDebug
	 * @param timeLimit in seconds
	 */
	public BrokerConnectRequest(
		String serverName,
		int serverPort,
		boolean wantProgrammerDebug,
		int timeLimit) {
		setServerName(serverName);
		setServerPort(serverPort);
		setWantDebug(wantProgrammerDebug);
		this.timeLimit = timeLimit;
	}
	/**
	 * This connect request provides information for a silent login. Since the login and password
	 * are provided there will be no user interactive prompting.
	 * 
	 * @param serverName server name
	 * @param serverPort port on the server which to connect
	 * @param login login string
	 * @param password password string
	 */
	public BrokerConnectRequest(String serverName, int serverPort, String login, String password) {
		setServerName(serverName);
		setServerPort(serverPort);
		setLogin(login);
		setPassword(password);
	}
	/**
	 * This connect request provides information for a silent login. Since the login and password
	 * are provided there will be no user interactive prompting.
	 * Allows for connect time limit as a parameter.
	 * 
	 * @param serverName
	 * @param serverPort
	 * @param login
	 * @param password
	 * @param timeLimit in seconds
	 */
	public BrokerConnectRequest(
		String serverName,
		int serverPort,
		String login,
		String password,
		int timeLimit) {
		setServerName(serverName);
		setServerPort(serverPort);
		setLogin(login);
		setPassword(password);
		this.timeLimit = timeLimit;
	}
	/**
	 * This connect request provides information for a silent login. Since the login and password
	 * are provided there will be no user interactive prompting. It also allows for programmer level debugging.
	 * 
	 * @param serverName server name
	 * @param serverPort port on the server which to connect
	 * @param login login string
	 * @param wantProgrammerDebug set to true if the programmer needs to debug across
	 * the connection. Set to false when programmer level debugging is not neccesary.
	 */
	public BrokerConnectRequest(
		String serverName,
		int serverPort,
		String login,
		String password,
		boolean wantProgrammerDebug) {
		setServerName(serverName);
		setServerPort(serverPort);
		setLogin(login);
		setPassword(password);
		setWantDebug(wantProgrammerDebug);
	}
	/**
	 * This connect request provides information for a silent login. Since the login and password
	 * are provided there will be no user interactive prompting. It also allows for programmer level debugging.
	 * Allows for connect time limit as a parameter.
	 * 
	 * @param serverName
	 * @param serverPort
	 * @param login
	 * @param password
	 * @param wantProgrammerDebug
	 * @param timeLimit in seconds
	 */
	public BrokerConnectRequest(
		String serverName,
		int serverPort,
		String login,
		String password,
		boolean wantProgrammerDebug,
		int timeLimit) {
		setServerName(serverName);
		setServerPort(serverPort);
		setLogin(login);
		setPassword(password);
		setWantDebug(wantProgrammerDebug);
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#setLogin(String)
	 */
	public void setLogin(String login) {
		this.login = login;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#setPassword(String)
	 */
	public void setPassword(String password) {
		this.password = password;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#setServerName(String)
	 */
	public void setServerName(String serverName) {
		this.serverName = serverName;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getServerName()
	 */
	public String getServerName() {
		return serverName;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#setServerPort(int)
	 */
	public void setServerPort(int serverPort) {
		this.serverPort = serverPort;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getServerPort()
	 */
	public int getServerPort() {
		return serverPort;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getLogin()
	 */
	public String getLogin() {
		return login;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getPassword()
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getWantDebug()
	 */
	public boolean getWantDebug() {
		return wantDebug;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#setWantDebug(boolean)
	 */
	public void setWantDebug(boolean wantDebug) {
		this.wantDebug = wantDebug;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IConnectRequest#getTimeLimit()
	 */
	public int getTimeLimit() {
		return timeLimit;
	}

	private String login = null;
	private String password = null;
	private String serverName = null;
	private int serverPort = 0;
	private boolean wantDebug = false;
	private int timeLimit = DEFAULT_TIME_LIMIT;
}
