package gov.va.med.hds.cd.m.internal;
import gov.va.med.hds.cd.m.CallQueueException;
import gov.va.med.hds.cd.m.CallRequestCacheException;
import gov.va.med.hds.cd.m.CallRequestException;
import gov.va.med.hds.cd.m.ConnectException;
import gov.va.med.hds.cd.m.DisconnectException;
import gov.va.med.hds.cd.m.ICallQueue;
import gov.va.med.hds.cd.m.ICallRequest;
import gov.va.med.hds.cd.m.IConnectListener;
import gov.va.med.hds.cd.m.IConnectRequest;
import gov.va.med.hds.cd.m.IDisconnectListener;
import gov.va.med.hds.cd.m.IMumpsListener;
import gov.va.med.hds.cd.m.IRPCLogListener;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.m.RPCLogEventObject;
import gov.va.med.hds.cd.m.TimedResult;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.security.UserCancelledAuthenticationException;
import gov.va.med.hds.util.ExceptionStackTrace;
import gov.va.med.hds.util.collections.Fifo;
import gov.va.med.hds.util.collections.FifoException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
/**
 * Implements an ICallQueue with the VistALink Mumps services
 */
public abstract class CallQueueBase implements ICallQueue {
	static final String DEFAULT_RPC_CONTEXT = "XHDXC DESKTOP";
	private static final String CALLQUEUE_PROCESSOR_THREAD_NAME = "CallQueue Processor";
	private static final String CALLQUEUE_DISPATCH_THREAD_NAME = "CallQueue Dispatch";
	private static final int DISCONNECT_TIME_LIMIT = 15000;

	private MumpsConnectionSettings settings;
	private CallRequestCache callCache = null;

	private boolean connected = false;
	private IConnectRequest connectPending = null;
	private IConnectListener connectPendingListener = null;
	private Exception connectDoneException = null;
	boolean connectTimedOut;

	private static final int MAX_FIFO_ENTRIES = 300;
	private Fifo callFifo = null;
	private Fifo resultFifo = null;

	private boolean destroyPending = false;
	private IDisconnectListener destroyPendingListener = null;
	private boolean disconnectDone = false;
	private DisconnectException disconnectDoneException = null;
	boolean disconnectTimedOut;

	private MumpsException callDoneException = null;
	private String callDoneResult = null;
	boolean callTimedOut;

	private Thread processorThread;
	private Thread dispatchThread;

	private ArrayList logListeners = new ArrayList();

	private IRuntimeContext runtime;

	private int rpcTimeLimitOverride = 0;
	private boolean useRpcTimeLimitOverride = false;

	/**
	 * Constructor
	 * @param settings
	 * @throws CallQueueException
	 */
	public CallQueueBase(IRuntimeContext runtime, MumpsConnectionSettings settings)
		throws CallQueueException {
		this.runtime = runtime;

		/**
		 * Call Processor Thread
		 * This thread processes RPC calls placed on the call queue.
		 * It is responsible for making the calls.
		 */
		class CallProcessor extends Thread {
			public void run() {
				try {
					processConnectInit();
				} catch (Exception e) {
					System.out.println(ExceptionStackTrace.getStackTrace(e));
				}
				callFifo = new Fifo(MAX_FIFO_ENTRIES);
				while (true) {
					try {
						serviceCallFifo();
					} catch (Exception e) {
						CallQueueBase.this.runtime.getLogger(this.getClass().getName()).error(
							"Servicing call FIFO failed.",
							e);
					}
					try {
						synchronized (this) {
							this.wait();
						}
					} catch (InterruptedException e) {
						CallQueueBase.this.runtime.getLogger(this.getClass().getName()).error(
							CALLQUEUE_PROCESSOR_THREAD_NAME + " thread was interrupted.",
							e);
					}
				}
			}
		}

		/**
		 * Dispatch Processor Thread
		 * This thread dispatches the results from processing
		 * RPC calls.
		 */
		class DispatchProcessor extends Thread {
			public void run() {
				resultFifo = new Fifo(MAX_FIFO_ENTRIES);
				while (true) {
					try {
						serviceResultFifo();
					} catch (Exception e) {
						CallQueueBase.this.runtime.getLogger(this.getClass().getName()).error(
							"Failed processing dispatch FIFO.",
							e);
					}
					try {
						synchronized (this) {
							this.wait();
						}
					} catch (InterruptedException e) {
						CallQueueBase.this.runtime.getLogger(this.getClass().getName()).error(
							CALLQUEUE_DISPATCH_THREAD_NAME + " thread was interrupted.",
							e);
					}
				}
			}
		}
		this.settings = settings;

		try {
			createCache();
		} catch (CallRequestCacheException e) {
			throw new CallQueueException(e);
		}

		processorThread = new CallProcessor();
		processorThread.setName(CALLQUEUE_PROCESSOR_THREAD_NAME);
		processorThread.start();

		dispatchThread = new DispatchProcessor();
		dispatchThread.setName(CALLQUEUE_DISPATCH_THREAD_NAME);
		dispatchThread.start();
	}

	/**
	 * @see gov.va.med.hds.cd.m.ICallQueue#enableTimeLimitOverride(int)
	 */
	public void enableTimeLimitOverride(int value) {
		useRpcTimeLimitOverride = true;
		rpcTimeLimitOverride = value;
	}

	/**
	 * @see gov.va.med.hds.cd.m.ICallQueue#disableTimeLimitOverride()
	 */
	public void disableTimeLimitOverride() {
		useRpcTimeLimitOverride = false;
	}

	protected abstract void processConnectInit() throws CallQueueProcessConnectInitException;
	protected abstract String processRequest(ICallRequest request) throws Exception;
	protected abstract void processConnect(IConnectRequest connectRequest)
		throws ConnectException, UserCancelledAuthenticationException;
	protected abstract void processDisconnect() throws Exception;
	/**
	 * @see gov.va.med.hds.cd.m.ICallQueue#add(ICallRequest, ICallQueueListener)
	 */
	synchronized public void add(ICallRequest request, IMumpsListener listener)
		throws CallQueueException {
		if (!connected)
			throw new CallQueueException("CallQueue must be connected by security before adds can be accomplished.");
		try {
			callFifo.add(new FifoRequestEntry(request, listener));
			synchronized (processorThread) {
				processorThread.notify();
			}
		} catch (FifoException e) {
			throw new CallQueueException(e);
		}
	}
	/**
	 * @see gov.va.med.hds.cd.m.ICallQueue#add(ICallRequest)
	 */
	synchronized public String add(ICallRequest request) throws MumpsException {
		callDoneException = null;
		callDoneResult = null;
		callTimedOut = true;
		add(request, new IMumpsListener() {
			synchronized public void CallFailed(ICallRequest request, MumpsException e) {
				callDoneException = e;
				callTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
			synchronized public void CallSucceeded(ICallRequest request, String result) {
				callDoneResult = result;
				callTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
		});
		try {
			if(useRpcTimeLimitOverride)
				wait(rpcTimeLimitOverride);
			else
				wait(request.getCallTimeLimit());
		} catch (InterruptedException e) {
			throw new CallQueueException(e);
		}
		if (callTimedOut) {
			// Figure out a way to start up a new connection
			// needs to be abstracted to Vistalink can do this different than Shared Broker
			throw new CallQueueException(
				"Call "
					+ request.getCallName()
					+ " timed out after "
					+ request.getCallTimeLimit()
					+ " millis.");
		}

		if (callDoneException != null)
			throw callDoneException;
		return callDoneResult;
	}

	/**
	 * 
	 * @see gov.va.med.hds.cd.m.ICallQueue#addLogListener(gov.va.med.hds.cd.m.IRPCLogListener)
	 */
	public void addLogListener(IRPCLogListener listener) throws CallQueueException {
		logListeners.add(listener);
	}
	/**
	 * 
	 * @see gov.va.med.hds.cd.m.ICallQueue#removeLogListener(gov.va.med.hds.cd.m.IRPCLogListener)
	 */
	public void removeLogListener(IRPCLogListener listener) throws CallQueueException {
		logListeners.remove(listener);
	}
	private void serviceResultFifo() throws FifoException {
		while (resultFifo.getCount() > 0) {
			FifoResultEntry result = (FifoResultEntry) resultFifo.pop();
			notifyListeners(result);
			result.dispatch();
		}
	}
	private void notifyListeners(FifoResultEntry result) throws FifoException {
		if (logListeners.size() > 0) {
			RPCLogEventObject data;
			try {
				data =
					new RPCLogEventObject(
						this,
						result.request.copy(),
						result.result,
						result.exception);
			} catch (CallRequestException e) {
				throw new FifoException(e);
			}
			for (int i = 0; i < logListeners.size(); i++)
				 ((IRPCLogListener) logListeners.get(i)).callExecuted(data);
		}
	}
	private void serviceCallFifo() throws Exception {
		while (connectPending != null || destroyPending || callFifo.getCount() > 0) {
			if (destroyPending) {
				processDestroyPending();
			} else if (connectPending != null) {
				processConnectPending();
			} else {
				processFifo();
			}
		}
	}
	private void processFifo() throws Exception {
		String result = null;
		FifoRequestEntry request = null;
		try {
			request = (FifoRequestEntry) callFifo.pop();
			long startTime = System.currentTimeMillis();
			result = processRequest(request.request);
			long elapsedTime = System.currentTimeMillis() - startTime;
			TimedResult timedResult = new TimedResult(result, elapsedTime);
			addToCache(request.request, timedResult);
			resultFifo.add(
				new FifoResultEntry(request.request, request.listener, timedResult, null));
		} catch (Exception e) {
			resultFifo.add(
				new FifoResultEntry(
					request.request,
					request.listener,
					null,
					new CallQueueException(
						"Call Failed \"" + request.request.getCallName() + "\"",
						e)));
		}

		synchronized (dispatchThread) {
			dispatchThread.notify();
		}
	}
	private void processConnectPending() throws FifoException {
		try {
			processConnect(connectPending);
		} catch (ConnectException e) {
			connectPendingListener.ConnectFailed(connectPending, e);
		} catch (UserCancelledAuthenticationException e) {
			connectPendingListener.ConnectFailed(connectPending, e);
		}
		for (int i = 0; i < callFifo.getCount(); i++) {
			FifoRequestEntry request = null;
			request = (FifoRequestEntry) callFifo.pop();
			if (request.listener != null)
				request.listener.CallFailed(
					request.request,
					new CallQueueException("New Connection is being established. Call request rejected."));
		}

		if (connectPendingListener != null) {
			connected = true;
			connectPendingListener.ConnectSucceeded(connectPending);
		}
		connectPending = null;
		connectPendingListener = null;
	}
	private void processDestroyPending() throws FifoException, DisconnectException {
		for (int i = 0; i < callFifo.getCount(); i++) {
			FifoRequestEntry request = (FifoRequestEntry) callFifo.pop();
			if (request.listener != null)
				request.listener.CallFailed(
					request.request,
					new CallQueueException("Connection is being destroyed. Call request rejected."));
		}
		if (connectPending != null) {
			if (connectPendingListener != null)
				connectPendingListener.ConnectFailed(
					connectPending,
					new ConnectException("Connection is being destroyed. This connect request has been cancelled."));
			connectPending = null;
			connectPendingListener = null;
		}
		try {
			processDisconnect();
		} catch (Exception e) {
			destroyPendingListener.DisconnectFailed(new DisconnectException(e));
		}
		if (destroyPendingListener != null) {
			connected = false;
			destroyPendingListener.DisconnectSucceeded();
		}
		destroyPendingListener = null;
		destroyPending = false;
	}
	synchronized void disconnect(IDisconnectListener listener) throws DisconnectException {
		destroyPending = true;
		destroyPendingListener = listener;
		try {
			destroyCache();
		} catch (CallRequestCacheException e) {
			throw new DisconnectException(e);
		}
		synchronized (processorThread) {
			processorThread.notify();
		}
	}
	synchronized void disconnect() throws DisconnectException {
		disconnectDone = false;
		disconnectDoneException = null;
		disconnectTimedOut = true;
		disconnect(new IDisconnectListener() {
			synchronized public void DisconnectFailed(DisconnectException e) {
				disconnectDone = true;
				disconnectDoneException = e;
				disconnectTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
			synchronized public void DisconnectSucceeded() {
				disconnectDone = true;
				disconnectTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
		});
		try {
			wait(DISCONNECT_TIME_LIMIT);
		} catch (InterruptedException e) {
			throw new DisconnectException(e);
		}
		if (disconnectTimedOut)
			throw new DisconnectException(
				"Disconnect timed out after " + DISCONNECT_TIME_LIMIT + " milliseconds.");

		if (disconnectDoneException != null)
			throw disconnectDoneException;
	}
	protected synchronized void connect(IConnectRequest request, IConnectListener listener) {
		connectPending = request;
		connectPendingListener = listener;
		synchronized (processorThread) {
			processorThread.notify();
		}
	}
	protected synchronized void connect(IConnectRequest request) throws Exception {
		connectTimedOut = true;
		connectDoneException = null;
		connect(request, new IConnectListener() {
			synchronized public void ConnectFailed(IConnectRequest request, Exception e) {
				connectDoneException = e;
				connectTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
			synchronized public void ConnectSucceeded(IConnectRequest request) {
				connectTimedOut = false;
				synchronized (CallQueueBase.this) {
					CallQueueBase.this.notify();
				}
			}
		});

		try {
			wait(request.getTimeLimit() * 1000);
		} catch (InterruptedException e) {
			throw new ConnectException(e);
		}
		if (connectTimedOut)
			throw new ConnectException(
				"Connection timed out after " + request.getTimeLimit() + " milliseconds.");

		if (connectDoneException != null)
			throw connectDoneException;
	}
	private void destroyCache() throws CallRequestCacheException {
		if (callCache != null) {
			File file = new File(settings.getRecordFileName());
			FileOutputStream out = null;
			try {
				out = new FileOutputStream(file);
			} catch (Exception e) {
				throw new CallRequestCacheException(e);
			}
			callCache.write(out);
			try {
				out.flush();
				out.close();
			} catch (IOException e) {
				throw new CallRequestCacheException(e);
			}
			callCache = null;
		}
	}
	private void createCache() throws CallRequestCacheException {
		if (settings.isRecord()) {
			callCache = new CallRequestCache(runtime);
			if (settings.isRecordAppend()) {
				File file = new File(settings.getRecordFileName());
				FileInputStream input;
				try {
					input = new FileInputStream(file);
				} catch (FileNotFoundException e) {
					throw new CallRequestCacheException(e);
				}
				callCache.load(input);
			}
		}
	}
	private void addToCache(ICallRequest request, TimedResult result)
		throws CallRequestCacheException {
		if (settings.isRecord()) {
			if (!callCache.contains(request))
				callCache.add(request, result);
		}
	}

}
class FifoRequestEntry {
	public FifoRequestEntry(ICallRequest request, IMumpsListener listener) {
		this.request = request;
		this.listener = listener;
	}
	public ICallRequest request;
	public IMumpsListener listener;
}
class FifoResultEntry {
	public FifoResultEntry(
		ICallRequest request,
		IMumpsListener listener,
		TimedResult result,
		CallQueueException e) {
		this.request = request;
		this.listener = listener;
		this.result = result;
		this.exception = e;
	}
	public void dispatch() {
		if (exception != null)
			listener.CallFailed(request, exception);
		else
			listener.CallSucceeded(request, result.result);
	}
	public ICallRequest request = null;
	public IMumpsListener listener = null;
	public TimedResult result = null;
	public CallQueueException exception = null;
}
