package gov.va.med.hds.cd.m.internal;

import gov.va.med.hds.cd.m.*;
import gov.va.med.hds.cd.m.CallRequestCacheEntry;
import gov.va.med.hds.cd.m.CallRequestCacheException;
import gov.va.med.hds.cd.m.CallRequestException;
import gov.va.med.hds.cd.m.ICallRequest;
import gov.va.med.hds.cd.m.TimedResult;
import gov.va.med.hds.cd.m.ValidationFailedException;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.security.ISecurity;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.security.auth.Subject;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;
import org.iso_relax.verifier.Schema;
import org.iso_relax.verifier.Verifier;
import org.iso_relax.verifier.VerifierConfigurationException;
import org.iso_relax.verifier.VerifierFactory;
import org.jaxen.JaxenException;
import org.jaxen.dom4j.Dom4jXPath;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.sun.msv.verifier.jarv.TheFactoryImpl;

/**
 * Provides a call request cache based on xml input and output
 * */
public class CallRequestCache implements ICallRequestCache {
	private static final boolean PRINT_DEBUG_INFO = false;
	public static final String VALIDATING_SCHEMA_PATH =
		"/gov/va/med/hds/cd/m/internal/xml/CallRequestCacheSchema.xsd";

	public CallRequestCache(IRuntimeContext runtime) throws CallRequestCacheException {
		this.runtime = runtime;
		try {
			compileSchema();
		} catch (ValidationFailedException e) {
			throw new CallRequestCacheException(e);
		}
	}
	/**
	 * @see gov.va.med.hds.cd.m.ICallRequestCache#getResult(gov.va.med.hds.cd.m.ICallRequest)
	 */
	public TimedResult getResult(ICallRequest callRequest) {
		if (PRINT_DEBUG_INFO) {
			printHashCodes(callRequest);
		}
		return ((TimedResult) cache.get(callRequest));
	}
	/**
	 * @see gov.va.med.hds.cd.m.ICallRequestCache#add(gov.va.med.hds.cd.m.ICallRequest, gov.va.med.hds.cd.m.TimedResult)
	 */
	public void add(ICallRequest request, TimedResult result) throws CallRequestCacheException {
		if (!cache.containsKey(request))
			try {
				cache.put(request.copy(), result);
			} catch (CallRequestException e) {
				throw new CallRequestCacheException(e);
			}
	}
	/**
	 * @see gov.va.med.hds.cd.m.ICallRequestCache#contains(gov.va.med.hds.cd.m.ICallRequest)
	 */
	public boolean contains(ICallRequest request) {
		return cache.containsKey(request);
	}
	/**
	 * Writes the contents of the cache to the stream
	 * 
	 * @param stream
	 * @throws CallRequestCacheException
	 */
	public void write(OutputStream stream) throws CallRequestCacheException {
		addUserInfoToCache();
		writeMapElements(stream);
	}
	/**
	 * Loads the xml contents of the stream into the cache
	 * 
	 * @param stream
	 * @throws CallRequestCacheException
	 */
	public void load(InputStream stream) throws CallRequestCacheException {
		SAXReader reader = new SAXReader();
		Document document;
		try {
			document = reader.read(stream);
		} catch (Exception e) {
			throw new CallRequestCacheException(e);
		}
		try {
			String userInfoSelectStr =
				"/" + CALL_CACHE_ROOT_ELEMENT_NAME + "/" + USERINFO_ELEMENT_NAME;
			Node userInfoNode =
				(Node) (new Dom4jXPath(userInfoSelectStr).selectSingleNode(document));
			userInfo = new MumpsUserInfo(userInfoNode);
		} catch (JaxenException e) {
			throw new CallRequestCacheException(e);
		}
		List list;
		try {
			String callEntrySelectStr =
				"/"
					+ CALL_CACHE_ROOT_ELEMENT_NAME
					+ "/"
					+ CALLS_ELEMENT_NAME
					+ "/"
					+ CallRequestCacheEntry.CALL_ENTRY_ROOT_ELEMENT_NAME;
			list = (new Dom4jXPath(callEntrySelectStr)).selectNodes(document);
		} catch (JaxenException e) {
			throw new CallRequestCacheException(e);
		}

		Iterator it = list.iterator();
		Node entry;
		while (it.hasNext()) {
			entry = (Node) it.next();
			addToCache(entry);
		}
	}
	/**
	 * Validates the input stream against the schema stream
	 * 
	 * @param stream
	 * @throws ValidationFailedException
	 */
	public void validate(InputStream stream) throws ValidationFailedException {
		Verifier verifier;
		try {
			verifier = (Verifier) schema.newVerifier();
		} catch (VerifierConfigurationException e) {
			throw new ValidationFailedException(e);
		}
		verifier.setErrorHandler(new ErrorHandler() {
			public void error(SAXParseException arg0) throws SAXException {
				throw new SAXException(arg0);
			}
			public void fatalError(SAXParseException arg0) throws SAXException {
				throw new SAXException(arg0);
			}
			public void warning(SAXParseException arg0) throws SAXException {
			}
		});

		boolean verified;
		try {
			verified = verifier.verify(new InputSource(stream));
		} catch (Exception e) {
			throw new ValidationFailedException(
				"Failed validation against schema file " + VALIDATING_SCHEMA_PATH + " ",
				e);
		}
	}
	private void compileSchema() throws ValidationFailedException {
		InputStream schemaStream = getClass().getResourceAsStream(VALIDATING_SCHEMA_PATH);
		VerifierFactory factory = new TheFactoryImpl();
		try {
			schema = factory.compileSchema(schemaStream);
		} catch (Exception e) {
			throw new ValidationFailedException(e);
		}
	}
	public IMumpsUserInfo retrieveUserInfo() {
		return userInfo;
	}
	private void printHashCodes(ICallRequest callRequest) {
		Iterator it = cache.keySet().iterator();
		System.out.println(
			"Request hashCode(" + callRequest.getCallName() + ") " + callRequest.hashCode());
		while (it.hasNext()) {
			ICallRequest key = (ICallRequest) it.next();
			if (callRequest.equals(key))
				System.out.println(
					"("
						+ key.getCallName()
						+ ")"
						+ key.hashCode()
						+ " == request("
						+ callRequest.getCallName()
						+ ")"
						+ callRequest.hashCode());
			else
				System.out.println(
					"("
						+ key.getCallName()
						+ ")"
						+ key.hashCode()
						+ " != request("
						+ callRequest.getCallName()
						+ ")"
						+ callRequest.hashCode());
		}
	}

	private void addToCache(Node entry) throws CallRequestCacheException {
		CallRequestCacheEntry item = new CallRequestCacheEntry(entry);
		add(item.getRequest(), item.getResult());
	}
	private void addUserInfoToCache() throws CallRequestCacheException {
		if (runtime != null) {
			try {
				ISecurity security = (ISecurity) runtime.getPluginInstance(ISecurity.PLUGIN_ID);
				Subject subject = security.getSubject();
				Set principals = subject.getPrincipals(IMumpsUserInfo.class);
				if (principals.size() != 1)
					throw new CallRequestCacheException(
						"There must be 1 and only 1 principal in the subjects that are a subclass of "
							+ IMumpsUserInfo.class.getName()
							+ " There were "
							+ principals.size());
				userInfo = new MumpsUserInfo((IMumpsUserInfo) principals.iterator().next());
			} catch (Exception e) {
				throw new CallRequestCacheException(e);
			}
		}
	}
	private void writeMapElements(OutputStream stream) throws CallRequestCacheException {
		Document document = DocumentHelper.createDocument();
		Element root = document.addElement(CALL_CACHE_ROOT_ELEMENT_NAME);
		if (userInfo == null)
			userInfo = new MumpsUserInfo();

		userInfo.writeXml(root);
		Element calls = root.addElement(CALLS_ELEMENT_NAME);
		Set keys = cache.keySet();
		Iterator it = keys.iterator();
		ICallRequest request;
		TimedResult result;
		CallRequestCacheEntry entry;

		while (it.hasNext()) {
			request = (ICallRequest) it.next();
			result = (TimedResult) cache.get(request);
			entry = new CallRequestCacheEntry(request, result);
			entry.construct(calls);
		}
		XMLWriter writer;
		try {
			writer = new XMLWriter(stream, OutputFormat.createPrettyPrint());
		} catch (UnsupportedEncodingException e) {
			throw new CallRequestCacheException("Couldn't create an XMLWriter", e);
		}
		try {
			writer.write(document);
		} catch (IOException e) {
			throw new CallRequestCacheException("Write the map elements", e);
		}
	}
	private HashMap cache = new HashMap();
	private MumpsUserInfo userInfo = null;
	private IRuntimeContext runtime = null;
	private Schema schema = null;
}

class MumpsUserInfo implements IMumpsUserInfo {
	public MumpsUserInfo() {
		divisionIEN = "UNDEFINED";
		divisionStationName = "UNDEFINED";
		divisionStationNumber = "UNDEFINED";
		dTime = "UNDEFINED";
		duz = "UNDEFINED";
		newPerson01 = "UNDEFINED";
		firstName = "UNDEFINED";
		lastName = "UNDEFINED";
		middleName = "UNDEFINED";
		prefix = "UNDEFINED";
		suffix = "UNDEFINED";
		title = "UNDEFINED";
		language = "UNDEFINED";
		serviceSection = "UNDEFINED";
	}
	public MumpsUserInfo(Node userInfoNode) throws JaxenException {
		divisionIEN =
			((Node) (new Dom4jXPath(ICallRequestCache.DIVISIONIEN_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		divisionStationName =
			((Node) (new Dom4jXPath(ICallRequestCache.DIVISIONSTATIONNAME_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		divisionStationNumber =
			((Node) (new Dom4jXPath(ICallRequestCache.DIVISIONSTATIONNUMBER_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		dTime =
			((Node) (new Dom4jXPath(ICallRequestCache.DTIME_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		duz =
			((Node) (new Dom4jXPath(ICallRequestCache.DUZ_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		newPerson01 =
			((Node) (new Dom4jXPath(ICallRequestCache.NEWPERSON01_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		firstName =
			((Node) (new Dom4jXPath(ICallRequestCache.FIRSTNAME_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		lastName =
			((Node) (new Dom4jXPath(ICallRequestCache.LASTNAME_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		middleName =
			((Node) (new Dom4jXPath(ICallRequestCache.MIDDLENAME_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		prefix =
			((Node) (new Dom4jXPath(ICallRequestCache.PREFIX_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		suffix =
			((Node) (new Dom4jXPath(ICallRequestCache.SUFFIX_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		title =
			((Node) (new Dom4jXPath(ICallRequestCache.TITLE_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		language =
			((Node) (new Dom4jXPath(ICallRequestCache.LANGUAGE_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
		serviceSection =
			((Node) (new Dom4jXPath(ICallRequestCache.SERVICESECTION_ELEMENT_NAME))
				.selectSingleNode(userInfoNode))
				.getText();
	}
	public MumpsUserInfo(IMumpsUserInfo info) {
		divisionIEN = info.getDivisionIEN();
		divisionStationName = info.getDivisionStationName();
		divisionStationNumber = info.getDivisionStationNumber();
		dTime = info.getDTime();
		duz = info.getDuz();
		newPerson01 = info.getNewPerson01();
		firstName = info.getFirstName();
		lastName = info.getLastName();
		middleName = info.getMiddleName();
		prefix = info.getPrefix();
		suffix = info.getSuffix();
		title = info.getTitle();
		language = info.getLanguage();
		serviceSection = info.getServiceSection();
	}
	public void writeXml(Element parent) {
		Element root = parent.addElement(ICallRequestCache.USERINFO_ELEMENT_NAME);
		Element elem = root.addElement(ICallRequestCache.DIVISIONIEN_ELEMENT_NAME);
		elem.addText(divisionIEN);
		elem = root.addElement(ICallRequestCache.DIVISIONSTATIONNAME_ELEMENT_NAME);
		elem.addText(divisionStationName);
		elem = root.addElement(ICallRequestCache.DIVISIONSTATIONNUMBER_ELEMENT_NAME);
		elem.addText(divisionStationNumber);
		elem = root.addElement(ICallRequestCache.DTIME_ELEMENT_NAME);
		elem.addText(dTime);
		elem = root.addElement(ICallRequestCache.DUZ_ELEMENT_NAME);
		elem.addText(duz);
		elem = root.addElement(ICallRequestCache.NEWPERSON01_ELEMENT_NAME);
		elem.addText(newPerson01);
		elem = root.addElement(ICallRequestCache.FIRSTNAME_ELEMENT_NAME);
		elem.addText(firstName);
		elem = root.addElement(ICallRequestCache.LASTNAME_ELEMENT_NAME);
		elem.addText(lastName);
		elem = root.addElement(ICallRequestCache.MIDDLENAME_ELEMENT_NAME);
		elem.addText(middleName);
		elem = root.addElement(ICallRequestCache.PREFIX_ELEMENT_NAME);
		elem.addText(prefix);
		elem = root.addElement(ICallRequestCache.SUFFIX_ELEMENT_NAME);
		elem.addText(suffix);
		elem = root.addElement(ICallRequestCache.TITLE_ELEMENT_NAME);
		elem.addText(title);
		elem = root.addElement(ICallRequestCache.LANGUAGE_ELEMENT_NAME);
		elem.addText(language);
		elem = root.addElement(ICallRequestCache.SERVICESECTION_ELEMENT_NAME);
		elem.addText(serviceSection);
	}
	public String getDivisionIEN() {
		return divisionIEN;
	}

	public String getDivisionStationName() {
		return divisionStationName;
	}

	public String getDivisionStationNumber() {
		return divisionStationNumber;
	}

	public String getDTime() {
		return dTime;
	}

	public String getDuz() {
		return duz;
	}

	public String getNewPerson01() {
		return newPerson01;
	}

	public String getFirstName() {
		return firstName;
	}

	public String getLastName() {
		return lastName;
	}

	public String getMiddleName() {
		return middleName;
	}

	public String getPrefix() {
		return prefix;
	}

	public String getSuffix() {
		return suffix;
	}

	public String getTitle() {
		return title;
	}

	public String getLanguage() {
		return language;
	}

	public String getServiceSection() {
		return serviceSection;
	}

	private String divisionIEN;
	private String divisionStationName;
	private String divisionStationNumber;
	private String dTime;
	private String duz;
	private String newPerson01;
	private String firstName;
	private String lastName;
	private String middleName;
	private String prefix;
	private String suffix;
	private String title;
	private String language;
	private String serviceSection;
	private IRuntimeContext runtime;
}
