package gov.va.med.hds.cd.m.internal;
import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.security.AuthenticationFailedException;
import gov.va.med.hds.cd.security.ConnectionFailedException;
import gov.va.med.hds.cd.security.DisconnectFailedException;
import gov.va.med.hds.cd.security.UserCancelledAuthenticationException;
import gov.va.med.hds.util.ExceptionStackTrace;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;
/**
 * Provides basic login module functionality for security based on MUMPS.
 */
public abstract class LoginModuleBase implements LoginModule {
	public static final String SERVER_NAME_KEY = "server";
	public static final String SERVER_PORT_KEY = "port";
	static final String RUNTIME_CONTEXT_KEY = "runtime";
	static final String MAIN_FRAME_KEY = "frame";
	/**
	 * @see javax.security.auth.spi.LoginModule#initialize(Subject, CallbackHandler, Map, Map)
	 */
	public void initialize(
		Subject subject,
		CallbackHandler callbackHandler,
		Map sharedState,
		Map options) {
		serverName = (String) options.get(SERVER_NAME_KEY);
		String port = (String) options.get(SERVER_PORT_KEY);
		if (port != null)
			serverPort = Integer.parseInt(port);
		runtime = (IRuntimeContext) options.get(RUNTIME_CONTEXT_KEY);
		this.subject = subject;
	}
	/**
	 * @see javax.security.auth.spi.LoginModule#login()
	 */
	public boolean login() throws LoginException {
		boolean result = false;
		createConnections();
		result = authenticate();
		return result;
	}
	/**
	 * @see javax.security.auth.spi.LoginModule#commit()
	 */
	public boolean commit() throws LoginException {
		return true;
	}
	/**
	 * @see javax.security.auth.spi.LoginModule#abort()
	 */
	public boolean abort() throws LoginException {
		try {
			shutdownConnections();
		} catch (Exception e) {
			throw new LoginException(ExceptionStackTrace.getStackTrace(e));
		}
		return true;
	}

	/**
	 * @see javax.security.auth.spi.LoginModule#logout()
	 */
	public boolean logout() throws LoginException {
		try {
			shutdownConnections();
			Set allPrincipals = subject.getPrincipals();
			Set principalsIAdded = subject.getPrincipals(getPrincipalClass());
			Iterator iterator = principalsIAdded.iterator();
			while (iterator.hasNext())
				allPrincipals.remove(iterator.next());
		} catch (Exception e) {
			throw new LoginException(ExceptionStackTrace.getStackTrace(e));
		}
		return true;
	}
	/**
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize() throws Exception {
		shutdownConnections();
	}
	/**
	 * Authenticates user credentials
	 * 
	 * @return boolean returns true if successful, false otherwise
	 * @throws AuthenticationFailedException
	 */
	protected abstract boolean authenticate() throws AuthenticationFailedException;
	/**
	 * Shuts down any underlying security connections
	 * 
	 * @throws DisconnectFailedException
	 */
	protected abstract void shutdownConnections() throws DisconnectFailedException;
	/**
	 * Creates any underlying security connections
	 * 
	 * @throws ConnectionFailedException
	 */
	protected abstract void createConnections()
		throws ConnectionFailedException, UserCancelledAuthenticationException;
	/**
	 * Gets the class type for the principal that will be added to the Subject
	 * on successful login
	 * 
	 * @return Class
	 */
	protected abstract Class getPrincipalClass();

	protected String serverName = "";
	protected int serverPort = 0;
	protected IRuntimeContext runtime = null;
	protected Subject subject = null;
	protected IMumps mumps = null;
}
