package gov.va.med.hds.cd.m.internal;

import gov.va.med.hds.cd.m.CallQueueException;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.m.EncryptException;
import gov.va.med.hds.cd.m.ICallQueue;
import gov.va.med.hds.cd.m.IMumpsListener;
import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.m.IOptionContext;
import gov.va.med.hds.cd.m.IRPCLogListener;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.m.RPCDoesNotExistInPluginDescriptorsException;
import gov.va.med.hds.util.ExceptionStackTrace;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Provides a base implemetation for rpc calls
 */
public class Mumps implements IMumps {
	private ICallQueue queue;
	private IOptionContext context;
	private Map rpcKeyValues;
	/**
	 * @param queue
	 * @param context
	 */
	public Mumps(ICallQueue queue, IOptionContext context, Map rpcKeyValues) {
		this.queue = queue;
		this.context = context;
		this.rpcKeyValues = rpcKeyValues;
	}
	private void handleError(String rpcKey, Exception e) throws MumpsException {
		throw new MumpsException(
			"RPC " + lookupRPCName(rpcKey) + " failed.\n" + ExceptionStackTrace.getStackTrace(e));
	}
	private String lookupRPCName(String key) throws RPCDoesNotExistInPluginDescriptorsException {
		String result = (String) rpcKeyValues.get(key);
		if (result == null)
			throw new RPCDoesNotExistInPluginDescriptorsException(
				"No RPC for key " + key + " exists");
		return result;
	}
	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.lang.Object[])
	 */
	public String callRPC(String rpcKey, Object[] args) throws MumpsException {
		String result = null;
		ArrayList params = null;
		if (args != null) {
			params = new ArrayList();
			for (int i = 0; i < args.length; i++)
				params.add(args[i]);
		}
		try {
			result = queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey), params));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.util.List)
	 */
	public String callRPC(String rpcKey, List params) throws MumpsException {
		String result = null;
		try {
			return queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey), params));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.lang.Object[], gov.va.med.hds.cd.m.ICallQueueListener)
	 */
	public void callRPC(String rpcKey, Object[] args, IMumpsListener listener)
		throws MumpsException {
		String result = null;
		ArrayList params = null;
		if (args != null) {
			params = new ArrayList();
			for (int i = 0; i < args.length; i++)
				params.add(args[i]);
		}
		try {
			queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey), params), listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.util.List, gov.va.med.hds.cd.m.ICallQueueListener)
	 */
	public void callRPC(String rpcKey, List params, IMumpsListener listener)
		throws MumpsException {
		String result = null;
		try {
			queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey), params), listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String)
	 */
	public String callRPC(String rpcKey) throws MumpsException {
		String result = null;
		try {
			return queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey)));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, gov.va.med.hds.cd.m.ICallQueueListener)
	 */
	public void callRPC(String rpcKey, IMumpsListener listener) throws MumpsException {
		String result = null;
		try {
			queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey)), listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#addLogListener(gov.va.med.hds.cd.m.IRPCLogListener)
	 */
	public void addLogListener(IRPCLogListener listener) throws MumpsException {
		try {
			queue.addLogListener(listener);
		} catch (CallQueueException e) {
			throw new MumpsException();
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#removeLogListener(gov.va.med.hds.cd.m.IRPCLogListener)
	 */
	public void removeLogListener(IRPCLogListener listener) throws MumpsException {
		try {
			queue.removeLogListener(listener);
		} catch (CallQueueException e) {
			throw new MumpsException();
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#encrypt(java.lang.String)
	 */
	public String encrypt(String data) throws EncryptException {
		return queue.encrypt(data);
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#decrypt(java.lang.String)
	 */
	public String decrypt(String data) throws EncryptException {
		return queue.decrypt(data);
	}

	/**
	 * Gets the internal call queue
	 * @return
	 */
	ICallQueue getCallQueue() {
		return queue;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, int)
	 */
	public String callRPC(String rpcKey, int timeLimit) throws MumpsException {
		String result = null;
		try {
			return queue.add(new CallRequest(context.getName(), lookupRPCName(rpcKey), timeLimit));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.lang.Object[], int)
	 */
	public String callRPC(String rpcKey, Object[] args, int timeLimit) throws MumpsException {
		String result = null;
		ArrayList params = null;
		if (args != null) {
			params = new ArrayList();
			for (int i = 0; i < args.length; i++)
				params.add(args[i]);
		}
		try {
			return queue.add(
				new CallRequest(context.getName(), lookupRPCName(rpcKey), params, timeLimit));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.util.List, int)
	 */
	public String callRPC(String rpcKey, List params, int timeLimit) throws MumpsException {
		String result = null;
		try {
			return queue.add(
				new CallRequest(context.getName(), lookupRPCName(rpcKey), params, timeLimit));
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
		return result;
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, gov.va.med.hds.cd.m.ICallQueueListener, int)
	 */
	public void callRPC(String rpcKey, IMumpsListener listener, int timeLimit)
		throws MumpsException {
		String result = null;
		try {
			queue.add(
				new CallRequest(context.getName(), lookupRPCName(rpcKey), timeLimit),
				listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.lang.Object[], gov.va.med.hds.cd.m.ICallQueueListener, int)
	 */
	public void callRPC(String rpcKey, Object[] args, IMumpsListener listener, int timeLimit)
		throws MumpsException {
		String result = null;
		ArrayList params = null;
		if (args != null) {
			params = new ArrayList();
			for (int i = 0; i < args.length; i++)
				params.add(args[i]);
		}
		try {
			queue.add(
				new CallRequest(context.getName(), lookupRPCName(rpcKey), params, timeLimit),
				listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.m.IMumps#callRPC(java.lang.String, java.util.List, gov.va.med.hds.cd.m.ICallQueueListener, int)
	 */
	public void callRPC(String rpcKey, List params, IMumpsListener listener, int timeLimit)
		throws MumpsException {
		String result = null;
		try {
			queue.add(
				new CallRequest(context.getName(), lookupRPCName(rpcKey), params, timeLimit),
				listener);
		} catch (CallQueueException e) {
			handleError(rpcKey, e);
		}
	}
}
