package gov.va.med.hds.cd.m.internal;

import gov.va.med.foundations.rpc.RpcNotInContextFaultException;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.m.ICallQueue;
import gov.va.med.hds.cd.m.IOptionContext;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.security.InvalidOptionException;
import gov.va.med.hds.util.ExceptionStackTrace;
import gov.va.med.hds.util.mumps.StringUtil;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * This class validates the context versions against the contexts stored on the server.
 * It incompatible versions are found, an InvalidContextException is thrown.
 */
public class OptionContextValidator {

	private static final String OPTION_VERSION_VALIDATING_RPC = "XHDX VERSRV";
	/**
	 * Constructor for OptionContextValidator 
	 */
	public OptionContextValidator() {
		super();
	}

	/**
	 * Validates the context versions against the contexts stored on the server.
		 * It incompatible versions are found, an InvalidContextException is thrown.
	 * @param contexts
	 * @param queue
	 */
	public void validate(IOptionContext[] contexts, ICallQueue queue, IOptionContext pluginContext)
		throws InvalidContextException, InvalidOptionException {
		SimpleList contextList = new SimpleList();
		ListMap map = new ListMap();
		HashMap missing = new HashMap();
		targetContext = pluginContext;
		initLists(contexts, contextList, map, missing);
		String[] data = getServerVersions(queue, contextList);
		validate(map, missing, data);
	}

	private void validate(ListMap map, HashMap missing, String[] data)
		throws InvalidContextException {
		for (int i = 0; i < data.length; i++) {
			String name = StringUtil.piece(data[i], '^', 1);
			Version version = getVersion(StringUtil.piece(data[i], '^', 2));
			List list = map.get(name);
			if (list.size() > 0)
				missing.remove(name);
			for (int j = 0; j < list.size(); j++) {
				IOptionContext context = (IOptionContext) list.get(j);
				if (!context.getVersionMatch().matches(version, context.getVersion()))
					throw new InvalidContextException(getErrorMessage(context, name, version));
			}
		}
		if (missing.size() > 0) {
			String name = (String) missing.keySet().iterator().next();
			IOptionContext context = (IOptionContext) missing.get(name);
			throw new InvalidContextException(getErrorMessage(context, null, null));
		}
	}

	private String[] getServerVersions(ICallQueue queue, SimpleList contextList)
		throws InvalidContextException, InvalidOptionException {
		String result;
		List params = new ArrayList();
		params.add(contextList.getList());
		try {
			result = queue.add(new CallRequest(targetContext.getName(), OPTION_VERSION_VALIDATING_RPC, params));
		} catch (MumpsException e) {
			if (e.getCause() instanceof RpcNotInContextFaultException)
				throw new InvalidOptionException(e.getMessage());
			else
				throw new InvalidContextException(ExceptionStackTrace.getStackTrace(e));
		}
		String[] data = result.split("\n");
		return data;
	}

	private void initLists(
		IOptionContext[] contexts,
		SimpleList contextList,
		ListMap map,
		HashMap missing) {
		for (int i = 0; i < contexts.length; i++) {
			map.add(contexts[i].getName(), contexts[i]);
			contextList.add(contexts[i].getName());
			missing.put(contexts[i].getName(), contexts[i]);
		}
	}

	private String getErrorMessage(IOptionContext context, String name, Version version) {
		StringBuffer buf = new StringBuffer();
		buf.append("The ");
		buf.append(context.getPluginDescriptor().getUniqueIdentifier());
		buf.append(" v");
		buf.append(context.getPluginDescriptor().getVersion().toString());
		buf.append(" plugin specifies an option context requirement on ");
		buf.append(context.getName());
		buf.append(" v");
		buf.append(context.getVersion());
		buf.append(".  ");
		if (name == null) {
			buf.append("No such context exists on the server.");
		} else {
			buf.append("The server version of ");
			buf.append(name);
			buf.append(" ");
			buf.append(version.toString());
			buf.append(".");
		}
		return buf.toString();
	}

	private Version getVersion(String version) {
		Version result;
		try {
			result = new Version(version);
		} catch (VersionFormatException e) {
			if (noVersion == null)
				noVersion = new Version(EMPTY_VERSION);
			result = noVersion;
		}
		return result;
	}

	private static final String EMPTY_VERSION = "0.0";
	private Version noVersion;
	private IOptionContext targetContext;
}