package gov.va.med.hds.cd.m.internal;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;

import gov.va.med.foundations.security.vistalink.VistaKernelPrincipal;
import gov.va.med.foundations.security.vistalink.VistaKernelPrincipalImpl;
import gov.va.med.foundations.security.vistalink.VistaLoginModule;
import gov.va.med.foundations.security.vistalink.VistaLoginModuleNoPathToListenerException;
import gov.va.med.foundations.security.vistalink.VistaLoginModuleTooManyInvalidAttemptsException;
import gov.va.med.foundations.security.vistalink.VistaLoginModuleUserCancelledException;
import gov.va.med.foundations.security.vistalink.VistaLoginModuleUserTimedOutException;
import gov.va.med.hds.cd.m.DisconnectException;
import gov.va.med.hds.cd.m.IMumpsFactory;
import gov.va.med.hds.cd.m.IOptionContext;
import gov.va.med.hds.cd.m.NoOptionContextInPluginDescriptorException;
import gov.va.med.hds.cd.m.OptionContext;
import gov.va.med.hds.cd.m.OptionContextExtensionConfigurationException;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.security.AuthenticationFailedException;
import gov.va.med.hds.cd.security.ConnectionFailedException;
import gov.va.med.hds.cd.security.InvalidOptionException;
import gov.va.med.hds.cd.security.LogoutException;
import gov.va.med.hds.cd.security.OptionValidationException;
import gov.va.med.hds.cd.security.TooManyInvalidLoginAttemptsException;
import gov.va.med.hds.cd.security.UserCancelledAuthenticationException;

/**
 * Abstract implementation of Vistalink login module
 */
public abstract class VistalinkLoginModuleBase extends LoginModuleBase {
	protected static final String VLJ_SERVER_ADDRESS_KEY = VistaLoginModule.SERVER_ADDRESS_KEY;
	protected static final String VLJ_SERVER_PORT_KEY = VistaLoginModule.SERVER_PORT_KEY;
	protected static final String VLJ_ACCESS_CODE_KEY = "ACCESS_CODE_KEY";
	protected static final String VLJ_VERIFY_CODE_KEY = "VERIFY_CODE_KEY";
	protected static final String VLJ_DIVISION_KEY = "DIVISION_KEY";
	private static final String OPTION_TEST_RPC_NAME = "USER DEMOS";

	private int CONNECT_TIMEOUT_IN_SECONDS = 120;
	/**
	 * base constructor. Decorates a VistaAVLoginModule to wrap the Vistalinks default login module.
	 */
	protected VistalinkLoginModuleBase() {
		loginModule = new VistaLoginModule();
	}
	/**
	 * @see javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject, javax.security.auth.callback.CallbackHandler, java.util.Map, java.util.Map)
	 */
	public void initialize(
		Subject subject,
		CallbackHandler callbackHandler,
		Map sharedState,
		Map options) {
		super.initialize(subject, null, sharedState, options);
		serverName = options.get(VLJ_SERVER_ADDRESS_KEY).toString();
		String port = (String) options.get(VLJ_SERVER_PORT_KEY);
		if (port != null)
			serverPort = Integer.parseInt(port);

		loginModule.initialize(subject, callbackHandler, sharedState, options);
	}
	/* (non-Javadoc)
	 * @see javax.security.auth.spi.LoginModule#login()
	 */
	public boolean login() throws LoginException {
		boolean result = false;
		try {
			loginModule.login();
			result = true;
		} catch (VistaLoginModuleUserCancelledException e) {
			LoginException exp = new UserCancelledAuthenticationException("User cancelled authentication");
			exp.initCause(e);
			throw exp;
		} catch (VistaLoginModuleTooManyInvalidAttemptsException e) {
			LoginException exp = new TooManyInvalidLoginAttemptsException("Too many invalid login attempts");
			exp.initCause(e);
			throw exp;
		} catch (VistaLoginModuleUserTimedOutException e) {
			LoginException exp = new AuthenticationFailedException("User login timed out");
			exp.initCause(e);
			throw exp;
		} catch (VistaLoginModuleNoPathToListenerException e){ 
			LoginException exp = new AuthenticationFailedException("Could not find Mumps server: "+serverName+"  port: "+serverPort);
			exp.initCause(e);
			throw exp;
		} catch (LoginException e) {
			LoginException exp = new LoginException();
			exp.initCause(e);
			throw exp;
		}
		return result;
	}
	/* (non-Javadoc)
	 * @see javax.security.auth.spi.LoginModule#commit()
	 * 
	 * Strips out the Vistalink supplied principal and replaces it with a principal supplied by the framework
	 */
	public boolean commit() throws LoginException {
		IPluginDescriptor desc = null;
		if (loginModule.commit()) {
			VistaKernelPrincipalImpl vistaPrincipal;
			IMumpsFactory conn = null;
			try {
				Set principals = subject.getPrincipals(VistaKernelPrincipal.class);
				if (principals.size() != 1)
					throw new LoginException(
						"There were "
							+ principals.size()
							+ " principals of type "
							+ VistaKernelPrincipal.class
							+ ". There must be one and only one.");
				vistaPrincipal = (VistaKernelPrincipalImpl) principals.iterator().next();
				subject.getPrincipals().remove(vistaPrincipal);
				subject.getPrincipals().add(new VistalinkMumpsPrincipal(vistaPrincipal));

				desc = runtime.getPluginRegistry().getPluginDescriptor(IMumpsFactory.PLUGIN_ID);
				IMumpsFactory factory =
					(IMumpsFactory) runtime.getPluginInstance(IMumpsFactory.PLUGIN_ID);
				mumps = factory.create(desc);
			} catch (Exception e) {
				LoginException exp = new LoginException("Couldn't setup principals");
				exp.initCause(e);
				throw exp;
			}
			try {
				((VistalinkCallQueue) ((Mumps) mumps).getCallQueue()).connect(
					new VistalinkConnectRequest(vistaPrincipal, CONNECT_TIMEOUT_IN_SECONDS));
			} catch (Exception e) {
				LoginException exp = new ConnectionFailedException("Could not connect");
				exp.initCause(e);
				throw exp;
			}
			try {
				validateOptionContexts(desc);
			} catch (InvalidOptionException e) {
				String message =
					"User "
						+ vistaPrincipal.getUserDemographicValue(
							VistaKernelPrincipal.KEY_NAME_FAMILYLAST)
						+ ","
						+ vistaPrincipal.getUserDemographicValue(
							VistaKernelPrincipal.KEY_NAME_GIVENFIRST)
						+ " has not been allocated the "
						+ CallQueueBase.DEFAULT_RPC_CONTEXT
						+ " option.\n"
						+ "Please consult a system administrator for access to this option.";
				LoginException exp =new OptionValidationException(message);
				exp.initCause(e); 
				throw exp;
			}
			return true;
		}
		return false;
	}
	private void validateOptionContexts(IPluginDescriptor plugin)
		throws InvalidContextException, InvalidOptionException {
		IOptionContext contexts[] = null;
		IOptionContext oc = null;
		LinkedList ctxts = new LinkedList();
		IPluginDescriptor[] descs = runtime.getPluginRegistry().getPluginDescriptors();
		for (int i = 0; i < descs.length; i++) {
			try {
				oc = new OptionContext(descs[i]);
			} catch (OptionContextExtensionConfigurationException e) {
				InvalidContextException exp = new InvalidContextException(
					"Plugin "
						+ descs[i].getUniqueIdentifier()
						+ " has an invalid optionContext configuration.\n");
				exp.initCause(e);
				throw exp;
			} catch (NoOptionContextInPluginDescriptorException e) {
				continue;
			}
			ctxts.add(oc);
		}
		if (ctxts.size() > 0) {
			contexts = new IOptionContext[ctxts.size()];
			Iterator it = ctxts.iterator();
			int count = 0;
			while (it.hasNext())
				contexts[count++] = (IOptionContext) it.next();
		}
		IOptionContext pluginContext;
		try {
			pluginContext = new OptionContext(plugin);
		} catch (OptionContextExtensionConfigurationException e) {
			InvalidContextException exp = new InvalidContextException(
				"Plugin "
					+ plugin.getUniqueIdentifier()
					+ " needs to have a valid optionContext\n");
			exp.initCause(e);
			throw exp;
		} catch (NoOptionContextInPluginDescriptorException e) {
			InvalidContextException exp = new InvalidContextException(
				"Plugin "
					+ plugin.getUniqueIdentifier()
					+ " needs to have a valid optionContext\n");
			exp.initCause(e);
			throw exp;
		}

		(new OptionContextValidator()).validate(
			contexts,
			((Mumps) mumps).getCallQueue(),
			pluginContext);
	}
	/* (non-Javadoc)
	 * @see javax.security.auth.spi.LoginModule#abort()
	 */
	public boolean abort() throws LoginException {
		return loginModule.abort();
	}
	/* (non-Javadoc)
	 * @see javax.security.auth.spi.LoginModule#logout()
	 */
	public boolean logout() throws LoginException {
		try {
			((VistalinkCallQueue) (((Mumps) mumps).getCallQueue())).disconnect();
		} catch (DisconnectException e) {
			LoginException exp = new LoginException(
				"Couldn't disconnect a VistalinkCallQueue\n");
			exp.initCause(e);
			throw exp;
		}
		boolean result = true;

		try {
			result = super.logout() && loginModule.logout();
		} catch (Exception e) {
			LogoutException exp = new LogoutException(
				"Vistalink Logout failed\n");
			exp.initCause(e);
			throw exp;
		}
		return result;
	}
	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.m.internal.LoginModuleBase#authenticate()
	 */
	protected boolean authenticate() {
		return false;
	}
	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.m.internal.LoginModuleBase#shutdownConnections()
	 */
	protected void shutdownConnections() {
	}

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.m.internal.LoginModuleBase#createConnections()
	 */
	protected void createConnections() {
	}
	/**
	 * @param login
	 * @param password
	 * @return boolean
	 * @throws ConnectionFailedException
	 */
	protected boolean callMumpsAuthenticate(String login, String password)
		throws ConnectionFailedException {
		return false;
	}
	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.m.internal.LoginModuleBase#getPrincipalClass()
	 */
	protected Class getPrincipalClass() {
		return VistalinkMumpsPrincipal.class;
	}
	protected VistaLoginModule loginModule;
}
