package gov.va.med.hds.cd.m.internal.broker.jni;
import gov.va.med.hds.util.DisposeException;
import gov.va.med.hds.util.IDisposable;
import gov.va.med.hds.cd.m.internal.common.Delimiters;
/**
 * Connector is a thin wrapper class around the JNI layer for 
 * the RPCSharedBroker. It provides connectivity to a MUMPS database
 */
public class Connector implements IDisposable {
	/**
	 * Constructor for Connector.
	 */
	public Connector() throws COMException {
		initialize();
		initialized = true;
	}
	/**
	 * Initializes the Connector. Must be called before any other calls. 
	 * COM and other dependencies are initialized here.
	 * @throws COMException
	 */
	protected native void initialize() throws COMException;
	/**
	 * Unitializes the Connector. Must be called after all calls to close up 
	 * any COM and other dependencies
	 * @throws ConnectorNotInitializedException
	 */
	protected native void unInitialize() throws ConnectorNotInitializedException;
	/**
	 * Method connect.
	 * @param clientName
	 * @param serverPort
	 * @param connectTimeLimit
	 * @param connectorCollectsCredentials
	 * @param loginStr
	 * @param wantDebug
	 * @throws ConnectorException
	 * @throws COMException
	 * @throws GeneralFailureException
	 * @throws CouldNotConnectException
	 * @throws ConnectorNotInitializedException
	 * @throws ConnectorNotConnectedException
	 * @throws UserCancelledAuthenticationException
	 */
	public native void connect(
		String clientName,
		String serverPort,
		long connectTimeLimit,
		boolean connectorCollectsCredentials,
		String loginStr,
		boolean wantDebug)
		throws
			ConnectorException,
			COMException,
			GeneralFailureException,
			CouldNotConnectException,
			ConnectorNotInitializedException,
			ConnectorNotConnectedException,
			UserCancelledAuthenticationException;
	/**
	 * Disconnects from the current connection.
	 * 
	 * @throws COMException
	 * @throws ConnectorException
	 * @throws ConnectorNotInitializedException
	 */
	protected native void disconnect()
		throws COMException, ConnectorException, ConnectorNotInitializedException;
	/**
	 * Sets the current rpc call context
	 * 
	 * @param optionName
	 * @throws COMException
	 * @throws ConnectorException
	 * @throws GeneralFailureException
	 * @throws CouldNotSetContextException
	 * @throws ConnectorNotInitializedException
	 */
	public native void setContext(String optionName)
		throws
			COMException,
			ConnectorException,
			GeneralFailureException,
			CouldNotSetContextException,
			ConnectorNotInitializedException;
	/**
	 * Makes the actual rpc call. The results are returned in a string. The results
	 * should be parsed by the caller.
	 * @param rpcName
	 * @param rpcParams
	 * @param rpcTimeLimit
	 * @return String
	 * @throws COMException
	 * @throws ConnectorException
	 * @throws GeneralFailureException
	 * @throws ConnectorNotInitializedException
	 */
	public native String call(String rpcName, String rpcParams, long rpcTimeLimit)
		throws
			COMException, ConnectorException, GeneralFailureException, ConnectorNotInitializedException;
	/**
	 * @see gov.va.med.hds.common.IDisposable#dispose()
	 */
	public void dispose() throws DisposeException {
		if (initialized == true) {
			try {
				unInitialize();
			} catch (ConnectorNotInitializedException e) {
				throw new DisposeException("unInitialize failed", e);
			}
			initialized = false;
		}
	}
	/**
	 * Gets the connection id for the currently connected client
	 * 
	 * @return int
	 */
	public int getConnectionClientId() {
		return connectionClientId;
	}
	/**
	 * Gets the last id for the last RPC called.
	 * 
	 * @return int
	 */
	public int getLastRPCCallId() {
		return lastRPCCallId;
	}
	/**
	 * Creates a login string from the given parameters. This string
	 * should be passed into Connect.connect in the loginStr parameter.
	 * 
	 * @param accessCode
	 * @param verifyCode
	 * @param division
	 * @return String
	 */
	public String createLoginStr(String accessCode, String verifyCode, String division) {
		StringBuffer buf = new StringBuffer();
		// (1) LoginHandle goes here
		buf.append(Delimiters.SEP_FS);
		// (2) NTToken goes here
		buf.append(Delimiters.SEP_FS);
		// (3) AccessCode goes here
		buf.append(accessCode);
		buf.append(Delimiters.SEP_FS);
		// (4) VerifyCode goes here
		buf.append(verifyCode);
		buf.append(Delimiters.SEP_FS);
		// (5) Division goes here
		buf.append(division);
		buf.append(Delimiters.SEP_FS);
		// (6) ModeVal goes here
		//     1 - lmAVCodes
		//     2 - lmAppHandle
		//     3 - lmNTToken
		buf.append(1);
		buf.append(Delimiters.SEP_FS);
		// (7) DivList goes here
		buf.append(Delimiters.SEP_FS);
		// (8) MultiDivision goes here
		buf.append(Delimiters.SEP_FS);
		// (9) DUZ goes here
		buf.append(Delimiters.SEP_FS);
		// (10) PrompDivision goes here
		buf.append(Delimiters.SEP_FS);
		// (11) ErrorText goes here
		buf.append(Delimiters.SEP_FS);
		return buf.toString();
	}
	static {
		System.loadLibrary("ConnectorC++");
	}

	protected void finalize() throws DisposeException {
		dispose();
	}
	private int connectionClientId = 0;
	private int lastRPCCallId = 0;
	public static void main(String[] args) {
		Connector conn = null;
		try {
			conn = new Connector();
		} catch (Exception e) {
			System.out.println(e.getMessage());
			try {
				conn.dispose();
			} catch (DisposeException exp) {
				System.out.println(exp.getMessage());
			}
		}
		try {
			testConnect(conn);
			testSetContext(conn);
			testCall(conn);
			conn.disconnect();
			System.out.println("ran successfully");
		} catch (Exception e) {
			System.out.println(e.getMessage());
		} finally {
			try {
				conn.unInitialize();
			} catch (ConnectorNotInitializedException e) {
				System.out.println(e.getMessage());
			}
		}
	}
	private void ReportAdditionalConnectResults(int uniqueClientId) {
		connectionClientId = uniqueClientId;
	}
	private void ReportAdditionalCallResults(int uniqueRPCId) {
		lastRPCCallId = uniqueRPCId;
	}
	private static void testSetContext(Connector conn)
		throws
			COMException,
			ConnectorException,
			CouldNotSetContextException,
			GeneralFailureException,
			ConnectorNotInitializedException {
		conn.setContext("ORRCM CLIENT");
	}
	private static void testConnect(Connector conn) throws Exception {
		String clientName = "Connector Test";
		String serverPort = "isl-reengkm:9000";
		int connectTimeLimit = 0;
		boolean connectorCollectsCredentials = true;
		boolean wantDebug = false;
		String loginStr = "";
		conn.connect(
			clientName,
			serverPort,
			connectTimeLimit,
			connectorCollectsCredentials,
			loginStr,
			wantDebug);
	}
	private static void testCall(Connector conn)
		throws
			COMException,
			ConnectorException,
			GeneralFailureException,
			ConnectorNotInitializedException {
		String rpcName = "ORRC PATIENT LIST ALL";
		String params = "";
		int rpcTimeLimit = 300;
		String results = conn.call(rpcName, params, rpcTimeLimit);
		System.out.println("sb.call results = " + results);
	}
	boolean initialized = false;
}
