package gov.va.med.hds.cd.m.internal.itest;
import gov.va.med.hds.cd.m.ICallRequest;
import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.m.IMumpsFactory;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.m.test.CallFactory;
import gov.va.med.hds.cd.m.test.TestConstants;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.test.RuntimeTestCase;
import gov.va.med.hds.cd.runtime.test.TestRuntime;
import gov.va.med.hds.cd.security.ISecurity;

import java.io.IOException;
import java.util.logging.Handler;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import junit.framework.TestCase;
/**
 * This test excersizes both SharedBroker and VistALink Connections 
 * and prints results to standard out. 
 * This test can be used to collect performance comparison data
 * between the SharedBroker and VistALink
 */
public class BrokerVsVistalinkPerformanceTest extends TestCase {
	public BrokerVsVistalinkPerformanceTest(String arg0) throws Exception {
		super(arg0);
		noParamSimpleRequest =
			CallFactory.getInstance().create(CallFactory.NO_PARAM_SIMPLE).request;
		singleParamSimpleRequest =
			CallFactory.getInstance().create(CallFactory.SINGLE_PARAM_SIMPLE).request;
		largeResultsRequest =
			CallFactory.getInstance().create(CallFactory.LARGE_PARAM_LARGE_RESULT).request;
		xmlDataInOutRequest =
			CallFactory.getInstance().create(CallFactory.LARGE_XML_PARAM_LARGE_XML_RESULT).request;
	}
	private static int CONNECT_DISCONNECT_ITERATIONS = 1;
	private static int XML_DATA_INOUT_ITERATIONS = 1;
	private static int NO_PARAM_SIMPLE_ITERATIONS = 1;
	private static int SINGLE_PARAM_SIMPLE_ITERATIONS = 1;
	private static int LARGE_RESULTS_ITERATIONS = 1;
	private static int FOUR_CALL_SEQUENCE_ITERATIONS = 1;

	private static boolean USE_FULL_ITERATIONS = false;

	static {
		if (USE_FULL_ITERATIONS) {
			CONNECT_DISCONNECT_ITERATIONS = 10;
			XML_DATA_INOUT_ITERATIONS = 100;
			NO_PARAM_SIMPLE_ITERATIONS = 100;
			SINGLE_PARAM_SIMPLE_ITERATIONS = 100;
			LARGE_RESULTS_ITERATIONS = 100;
			FOUR_CALL_SEQUENCE_ITERATIONS = 100;
		}
	}

	private final static float MS_PER_SECOND = 1000.0f;
	protected void setUp() throws Exception {
		super.setUp();
		System.out.println(
			"\tConnect (ave "
				+ CONNECT_DISCONNECT_ITERATIONS
				+ " times)"
				+ "\tNoParamSimple (ave "
				+ NO_PARAM_SIMPLE_ITERATIONS
				+ " times)"
				+ "\tSingleParamSimple (ave "
				+ SINGLE_PARAM_SIMPLE_ITERATIONS
				+ " times)"
				+ "\tLargeResults (ave "
				+ LARGE_RESULTS_ITERATIONS
				+ " times)"
				+ "\tXmlDataInOut (ave "
				+ XML_DATA_INOUT_ITERATIONS
				+ " times)"
				+ "\tFourCallSequence (ave "
				+ FOUR_CALL_SEQUENCE_ITERATIONS
				+ " times)"
				+ "\tDisconnect (ave "
				+ CONNECT_DISCONNECT_ITERATIONS
				+ " times)");
	}
	private final static String BROKER_TEST_NAME = "Shared Broker";
	private final static String VISTALINK_TEST_NAME = "VistALink";
	public void testBrokerPerformance() throws Exception {
		Handler handler;
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).setUseParentHandlers(false);
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).addHandler(handler = new Handler() {
			public void publish(LogRecord record) {
				fail(record.toString());
			}
			public void flush() {
			}
			public void close() {
			}
		});
		TestRuntime runtime = getRuntime(TestConstants.DEFAULT_REGISTRY_PATH);
		runtime.readSettings(getClass().getResourceAsStream("BrokerSettings.xml"));
		runTests(BROKER_TEST_NAME, runtime);
		runtime.shutdown();
		runtime = null;
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).removeHandler(handler);
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).setUseParentHandlers(true);
	}
	private TestRuntime getRuntime(String registryFile) {
		TestRuntime runtime = new TestRuntime();
		runtime.readPluginDescriptors(getClass().getResourceAsStream(registryFile));
		return runtime;
	}
	public void testVistalinkPerformance() throws Exception {
		Handler handler;
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).setUseParentHandlers(false);
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).addHandler(handler = new Handler() {
			public void publish(LogRecord record) {
				fail(record.toString());
			}
			public void flush() {
			}
			public void close() {
			}
		});
		TestRuntime runtime = getRuntime(TestConstants.DEFAULT_REGISTRY_PATH);
		runtime.readSettings(getClass().getResourceAsStream("VistalinkSettings.xml"));
		runTests(VISTALINK_TEST_NAME, runtime);
		runtime.shutdown();
		runtime = null;
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).removeHandler(handler);
		Logger.getLogger(RuntimeTestCase.RUNTIME_PACKAGE).setUseParentHandlers(true);
	}

	private long averageConnectMS = 0;
	private long averageDisconnectMS = 0;
	private long averageXmlDataInOutMS = 0;
	private long averageNoParamSimpleMS = 0;
	private long averageSingleParamSimpleMS = 0;
	private long averageLargeResultsMS = 0;
	private long averageFourCallSequenceMS = 0;

	private void runTests(String testName, IRuntimeContext runtime) throws Exception {
		ISecurity securityModule = (ISecurity) runtime.getPluginInstance(ISecurity.PLUGIN_ID);
		//		calculateAverageConnectDisconnectTimesMS(securityModule, CONNECT_DISCONNECT_ITERATIONS);
		securityModule.login();
		IPluginDescriptor desc =
			runtime.getPluginRegistry().getPluginDescriptor(IMumpsFactory.PLUGIN_ID);
		IMumps mumps =
			((IMumpsFactory) runtime.getPluginInstance(IMumpsFactory.PLUGIN_ID)).create(desc);

		calculateAverageXmlDataInOutTimeMS(mumps, XML_DATA_INOUT_ITERATIONS);

		calculateNoParamSimpleTimeMS(mumps, NO_PARAM_SIMPLE_ITERATIONS);

		calculateSingleParamSimpleTimeMS(mumps, SINGLE_PARAM_SIMPLE_ITERATIONS);

		calculateLargeResultsTimeMS(mumps, LARGE_RESULTS_ITERATIONS);

		calculateFourCallSequence(mumps, FOUR_CALL_SEQUENCE_ITERATIONS);

		securityModule.logout();
		System.out.println(
			testName
				+ "\t"
				+ ((float) averageConnectMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageNoParamSimpleMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageSingleParamSimpleMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageLargeResultsMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageXmlDataInOutMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageFourCallSequenceMS) / MS_PER_SECOND
				+ "\t"
				+ ((float) averageDisconnectMS) / MS_PER_SECOND);
	}

	private void calculateFourCallSequence(IMumps mumps, int iterations) throws Exception {
		long timeStamp;
		String result;
		averageFourCallSequenceMS = 0;
		for (long i = 0; i < iterations; i++) {
			timeStamp = System.currentTimeMillis();
			result =
				mumps.callRPC(noParamSimpleRequest.getCallName(), noParamSimpleRequest.getParams());
			assertTrue(result.length() > 0);
			result =
				mumps.callRPC(
					singleParamSimpleRequest.getCallName(),
					singleParamSimpleRequest.getParams());
			assertTrue(result.length() > 0);
			result =
				mumps.callRPC(largeResultsRequest.getCallName(), largeResultsRequest.getParams());
			assertTrue(result.length() > 0);
			result =
				mumps.callRPC(xmlDataInOutRequest.getCallName(), xmlDataInOutRequest.getParams());
			assertTrue(result.length() > 0);
			averageFourCallSequenceMS += System.currentTimeMillis() - timeStamp;
		}
		averageFourCallSequenceMS /= iterations;
	}
	private void calculateLargeResultsTimeMS(IMumps mumps, int iterations) throws MumpsException {
		String result;
		long timeStamp;
		averageLargeResultsMS = 0;
		for (long i = 0; i < iterations; i++) {
			timeStamp = System.currentTimeMillis();
			result =
				mumps.callRPC(largeResultsRequest.getCallName(), largeResultsRequest.getParams());
			printInfo(largeResultsRequest, result, "LargeResults");
			assertTrue(result.length() > 0);
			averageLargeResultsMS += System.currentTimeMillis() - timeStamp;
		}
		averageLargeResultsMS /= iterations;
	}
	private void calculateSingleParamSimpleTimeMS(IMumps mumps, int iterations)
		throws MumpsException {
		String result;
		long timeStamp;
		averageSingleParamSimpleMS = 0;
		for (long i = 0; i < iterations; i++) {
			timeStamp = System.currentTimeMillis();
			result =
				mumps.callRPC(
					singleParamSimpleRequest.getCallName(),
					singleParamSimpleRequest.getParams());
			printInfo(singleParamSimpleRequest, result, "SingleParamSimple");
			assertTrue(result.length() > 0);
			averageSingleParamSimpleMS += System.currentTimeMillis() - timeStamp;
		}
		averageSingleParamSimpleMS /= iterations;
	}

	private void printInfo(ICallRequest request, String result, String label) {
		//		System.out.println("\nTEST NAME: "+label);
		//		System.out.println("CALL NAME: "+request.getCallName());
		//		System.out.print("PARAMS: ");
		//		if(request.getParams() != null){
		//			System.out.print("\n");
		//			for(int i=0;i<request.getParams().size();i++){
		//				System.out.println("PARAM("+i+"): "+request.getParams().get(i).toString());
		//			}
		//		}else{
		//			System.out.println("NONE");
		//		}
		//		System.out.println("RESULT:\n"+result+"\n");
	}
	private void calculateNoParamSimpleTimeMS(IMumps mumps, int iterations) throws MumpsException {
		String result;
		long timeStamp;
		averageNoParamSimpleMS = 0;
		for (long i = 0; i < iterations; i++) {
			timeStamp = System.currentTimeMillis();
			result =
				mumps.callRPC(noParamSimpleRequest.getCallName(), noParamSimpleRequest.getParams());
			printInfo(noParamSimpleRequest, result, "NoParamSimple");
			assertTrue(result.length() > 0);
			averageNoParamSimpleMS += System.currentTimeMillis() - timeStamp;
		}
		averageNoParamSimpleMS /= iterations;
	}

	private void calculateAverageXmlDataInOutTimeMS(IMumps mumps, int iterations)
		throws IOException, MumpsException {
		long timeStamp;
		averageXmlDataInOutMS = 0;
		String result;
		for (long i = 0; i < iterations; i++) {
			timeStamp = System.currentTimeMillis();
			result =
				mumps.callRPC(xmlDataInOutRequest.getCallName(), xmlDataInOutRequest.getParams());
			printInfo(xmlDataInOutRequest, result, "XmlDataInOut");
			assertTrue(result.length() > 0);
			averageXmlDataInOutMS += System.currentTimeMillis() - timeStamp;
		}
		averageXmlDataInOutMS /= iterations;
	}

	private void calculateAverageConnectDisconnectTimesMS(
		ISecurity securityModule,
		long iterations)
		throws Exception {
		long timeStamp1, timeStamp2;
		averageConnectMS = 0;
		averageDisconnectMS = 0;
		for (long i = 0; i < iterations; i++) {
			timeStamp1 = System.currentTimeMillis();
			securityModule.login();
			timeStamp2 = System.currentTimeMillis();
			averageConnectMS += timeStamp2 - timeStamp1;
			securityModule.logout();
			averageDisconnectMS += System.currentTimeMillis() - timeStamp2;
		}
		averageConnectMS /= iterations;
		averageDisconnectMS /= iterations;
	}
	protected void tearDown() throws Exception {
		super.tearDown();
	}
	private ICallRequest noParamSimpleRequest;
	private ICallRequest singleParamSimpleRequest;
	private ICallRequest largeResultsRequest;
	private ICallRequest xmlDataInOutRequest;
}
