package gov.va.med.hds.cd.m.test;

import gov.va.med.hds.cd.m.CallQueueException;
import gov.va.med.hds.cd.m.CallRequest;
import gov.va.med.hds.cd.m.CallRequestExpectedResultPair;
import gov.va.med.hds.cd.m.ICallRequest;
import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.m.IMumpsListener;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.util.io.StreamUtilities;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

/**
 * A factory that generates RPC calls for testing purposes
 */
public class CallFactory {
	private static final String ECHO_RPC_KEY = "ECHO";
	private static final String EXAMPLE_TEXT_RPC_KEY = "TEXT";
	private static final String TEST_INPUT_LIST_RPC_KEY = "LIST";
	public static final int NO_PARAM_SIMPLE = 0;
	public static final int SINGLE_PARAM_SIMPLE = 1;
	public static final int LARGE_PARAM_LARGE_RESULT = 2;
	public static final int XML_PARAM_XML_RESULT = 3;
	public static final int PARAM_COLLECTION = 4;
	public static final int PARAM_MAP = 5;
	public static final int PARAM_SET = 6;
	public static final int LARGE_XML_PARAM_LARGE_XML_RESULT = 7;

	/**
	 * Gets an instance of the call factory
	 * @return
	 * @throws Exception
	 */
	public static CallFactory getInstance() throws Exception {
		if (instance == null)
			instance = new CallFactory();
		return instance;
	}
	/**
	 * Creates a call request pair for the specified RPC call factory index
	 * @param index
	 * @return
	 */
	public CallRequestExpectedResultPair create(int index) {
		return (CallRequestExpectedResultPair) calls.get(index);
	}
	/**
	 * Gets the number of unique RPCs in the Call factory
	 * @return
	 */
	public int getCount() {
		return calls.size();
	}
	/**
	 * Executes the call suite on the given IMumps
	 * @param mumps
	 * @throws Exception
	 */
	public void executeCallSuite(IMumps mumps) throws Exception {
		testSimpleCallAsync(mumps);
		testSimpleCallSync(mumps);
		testAFewSingleParamCallAsync(mumps);
		testAFewSingleParamCallSync(mumps);
		testAFewHeavyCallsSync(mumps);
		testParameterCollectionCall(mumps);
		testParameterMapCall(mumps);
		testParameterSetCall(mumps);
	}
	protected CallFactory() throws Exception {
		generateCalls();
	}
	private void generateCalls() throws Exception {
		calls.add(NO_PARAM_SIMPLE, createNoParamSimple());
		calls.add(SINGLE_PARAM_SIMPLE, createSingleParamSimple());
		calls.add(LARGE_PARAM_LARGE_RESULT, createLargeParamLargeResult());
		calls.add(XML_PARAM_XML_RESULT, createXmlParamXmlResult());
		calls.add(PARAM_COLLECTION, createParamCollection());
		calls.add(PARAM_MAP, createParamMap());
		calls.add(PARAM_SET, createParamSet());
		calls.add(LARGE_XML_PARAM_LARGE_XML_RESULT, createLargeXmlParamLargeXmlResult());
	}
	private void testSimpleCallAsync(IMumps mumps) throws Exception {
		callDone = false;
		callDoneException = null;
		callDoneResult = null;
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.NO_PARAM_SIMPLE);
		mumps.callRPC(pair.request.getCallName(), pair.request.getParams(), new IMumpsListener() {
			public void CallFailed(ICallRequest request, MumpsException e) {
				callDone = true;
				callDoneException = e;
				synchronized (CallFactory.this) {
					CallFactory.this.notify();
				}
			}

			public void CallSucceeded(ICallRequest request, String result) {
				callDone = true;
				callDoneResult = result;
				synchronized (CallFactory.this) {
					CallFactory.this.notify();
				}
			}
		});
		synchronized (this) {
			wait(WAIT_TIME_IN_MS);
		}
		if (!callDone) {
			throw new Exception("Call did not complete in time");
		};
		if (callDoneException != null) {
			throw callDoneException;
		}
		if (!pair.matchesExpectedResult(callDoneResult)) {
			throw new Exception(
				"Result failed expected\nResult:\n"
					+ callDoneResult
					+ "\nDoes not match expected result:\n"
					+ pair.expectedResult);
		}
	}
	private void testSimpleCallSync(IMumps mumps) throws Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.NO_PARAM_SIMPLE);
		String result = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
		if (!pair.matchesExpectedResult(result)) {
			throw new Exception(
				"Result failed expected\nResult:\n"
					+ result
					+ "\nDoes not match expected result:\n"
					+ pair.expectedResult);
		}
	}
	private void testAFewSingleParamCallAsync(IMumps mumps) throws Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.SINGLE_PARAM_SIMPLE);
		for (int i = 0; i < 3; i++) {
			callDone = false;
			callDoneException = null;
			callDoneResult = null;
			mumps
				.callRPC(
					pair.request.getCallName(),
					pair.request.getParams(),
					new IMumpsListener() {
				public void CallFailed(ICallRequest request, MumpsException e) {
					callDone = true;
					callDoneException = e;
					synchronized (CallFactory.this) {
						CallFactory.this.notify();
					}
				}

				public void CallSucceeded(ICallRequest request, String result) {
					callDone = true;
					callDoneResult = result;
					synchronized (CallFactory.this) {
						CallFactory.this.notify();
					}
				}
			});
			synchronized (this) {
				wait(WAIT_TIME_IN_MS);
			}
			if (!callDone) {
				throw new Exception("Call did not complete in time");
			};
			if (callDoneException != null) {
				throw callDoneException;
			}

			if (!pair.matchesExpectedResult(callDoneResult)) {
				throw new Exception(
					"Result failed expected\nResult:\n"
						+ callDoneResult
						+ "\nDoes not match expected result:\n"
						+ pair.expectedResult);
			}
		}
	}
	private void testAFewSingleParamCallSync(IMumps mumps) throws Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.SINGLE_PARAM_SIMPLE);
		for (int i = 0; i < 3; i++) {
			callDoneResult = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
			if (!pair.matchesExpectedResult(callDoneResult)) {
				throw new Exception(
					"Result failed expected\nResult:\n"
						+ callDoneResult
						+ "\nDoes not match expected result:\n"
						+ pair.expectedResult);
			}
		}
	}
	private void testAFewHeavyCallsSync(IMumps mumps) throws Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.LARGE_PARAM_LARGE_RESULT);
		for (int i = 0; i < 3; i++) {
			callDoneResult = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
			if (!pair.matchesExpectedResult(callDoneResult)) {
				throw new Exception(
					"Result failed expected\nResult:\n"
						+ callDoneResult
						+ "\nDoes not match expected result:\n"
						+ pair.expectedResult);
			}
		}
	}
	private void testParameterCollectionCall(IMumps mumps) throws Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.PARAM_COLLECTION);
		callDoneResult = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
		if (!pair.matchesExpectedResult(callDoneResult)) {
			throw new Exception(
				"Result failed expected\nResult:\n"
					+ callDoneResult
					+ "\nDoes not match expected result:\n"
					+ pair.expectedResult);
		}
	}

	private void testParameterMapCall(IMumps mumps) throws CallQueueException, Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.PARAM_MAP);
		callDoneResult = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
		if (!pair.matchesExpectedResult(callDoneResult)) {
			throw new Exception(
				"Result failed expected\nResult:\n"
					+ callDoneResult
					+ "\nDoes not match expected result:\n"
					+ pair.expectedResult);
		}
	}
	private void testParameterSetCall(IMumps mumps) throws CallQueueException, Exception {
		CallRequestExpectedResultPair pair =
			CallFactory.getInstance().create(CallFactory.PARAM_SET);
		callDoneResult = mumps.callRPC(pair.request.getCallName(), pair.request.getParams());
		if (!pair.matchesExpectedResult(callDoneResult)) {
			throw new Exception(
				"Result failed expected\nResult:\n"
					+ callDoneResult
					+ "\nDoes not match expected result:\n"
					+ pair.expectedResult);
		}
	}
	private static final String XML_PARAM_DATA_FILENAME = "LargeXMLData.xml";
	private CallRequestExpectedResultPair createLargeXmlParamLargeXmlResult() throws Exception {
		InputStreamReader reader =
			new InputStreamReader(getClass().getResourceAsStream(XML_PARAM_DATA_FILENAME));
		final int DATA_BUFF_LEN = 100;
		char[] data = new char[DATA_BUFF_LEN];
		boolean done = true;
		int charsRead;
		StringBuffer dataBuff = new StringBuffer();
		while (true) {
			charsRead = reader.read(data);
			if (charsRead == -1)
				break;
			dataBuff.append(data, 0, charsRead);
		}
		String xmlData = dataBuff.toString();
		String[] xmlLines = xmlData.split("\n");
		ArrayList params = new ArrayList();
		ArrayList list = new ArrayList();
		for (int i = 0; i < xmlLines.length; i++)
			list.add(xmlLines[i]);
		params.add(list);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) {
				return result.equals(this.expectedResult);
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), xmlData);
	}

	private CallRequestExpectedResultPair createXmlParamXmlResult() throws Exception {
		InputStream stream =
			getClass().getResourceAsStream(
				"/gov/va/med/hds/cd/m/internal/test/CallFactoryXmlData.xml");
		String xmlData = new String(StreamUtilities.convertStreamToByteArray(stream));
		String[] xmlLines = xmlData.split("\n");
		ArrayList params = new ArrayList();
		ArrayList list = new ArrayList();
		for (int i = 0; i < xmlLines.length; i++)
			list.add(xmlLines[i].replace('\r', '\n'));
		params.add(list);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) {
				return result.length() > 0;
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), xmlData);
	}
	private CallRequestExpectedResultPair createNoParamSimple() {
		final String EXPECTED_RESULT =
			"This file is used as a repository of server-based procedures in the contextof the Client/Server architecture.  By using the Remote Procedure Call (RPC)Broker, applications running on client workstations can invoke (call) theprocedures in this file to be executed by the server and the results will bereturned to the client application.Each remote procedure entry is associated with an entry point (ROUTINE withoptional TAG).  Calls to these procedures can include parameters of differentvalue types.  The resulting value of the call can be either a string, a listof strings or a word processing string as indicated by the RETURN VALUE TYPEfield (.04).The remote procedure may be available for use by anyone or its use may berestricted to one or more application.  The range of availability is indicatedby the AVAILABILITY field.  IF THERE IS NO ENTRY IN THE AVAILABILITY FIELD,then the procedure is assumed to be PUBLIC.A remote procedure may be removed from service for a period of time by settingthe INACTIVE field.  A request for use of a procedure which is marked inactivewill result in an error being returned to the originating application.";
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) throws Exception {
				// We get slightly different results from the Shared Broker verses Vistalink
				// The Shared Broker puts "\r\n" at the end of each item in a list
				// where as Vistalink does not
				// Also there appears to be an extra \n on the Vistalink call
				// So for the tests sake lets strip off the '\r' and compare
				// everything but the last character.
				// Supposably Vistalink should replace the Shared Broker so data results
				// should be consistent for client code based on Vistalink.
				result = result.replaceAll("\r\n", "\n");
				result = result.substring(0, expectedResult.length());
				return result.equals(expectedResult);
			}
		};
		return new Pair(new CallRequest(EXAMPLE_TEXT_RPC_KEY), EXPECTED_RESULT);
	}
	private CallRequestExpectedResultPair createSingleParamSimple() {
		final String EXPECTED_RESULT =
			"This is the result for a simple single param call. It is also the only param.";
		ArrayList params = new ArrayList();
		params.add(EXPECTED_RESULT);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) {
				return result.equals(expectedResult);
			}
		};
		return new Pair(new CallRequest(ECHO_RPC_KEY, params), EXPECTED_RESULT);
	}
	private CallRequestExpectedResultPair createLargeParamLargeResult() {
		final String EXPECTED_RESULT = "";
		final String BASE_STRING = "Here is some test text for line: ";
		final int STRING_COUNT = 300;
		ArrayList params = new ArrayList();
		ArrayList param1 = new ArrayList();
		for (int i = 0; i < STRING_COUNT; i++) {
			param1.add(BASE_STRING + i);
		}
		StringBuffer expectedResult = new StringBuffer();
		for (int i = 0; i < STRING_COUNT; i++) {
			expectedResult.append(BASE_STRING + i + "\n");
		}
		params.add(param1);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) throws Exception {
				return result.length() > BASE_STRING.length() * 300;
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), EXPECTED_RESULT);
	}
	private CallRequestExpectedResultPair createParamCollection() {
		final String EXPECTED_RESULT =
			"<?xml version=\"1.0\" encoding=\"UTF-8\" ?><result><resultTable name=\"test_list\" rowCount=\"4\" columnCount=\"2\"><columns><c name=\"name\" type=\"string\"/><c name=\"value\" type=\"string\"/></columns><rows><r><c>1</c><c>Item 1</c></r><r><c>2</c><c>Item 2</c></r><r><c>3</c><c>Item 3</c></r><r><c>4</c><c>Item 4</c></r></rows></resultTable></result>";
		ArrayList params = new ArrayList();
		ArrayList list = new ArrayList();
		list.add("Item 1");
		list.add("Item 2");
		list.add("Item 3");
		list.add("Item 4");
		params.add(list);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) throws Exception {
				return result.equals(expectedResult);
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), EXPECTED_RESULT);
	}
	private CallRequestExpectedResultPair createParamMap() {
		final String EXPECTED_RESULT =
			"<?xml version=\"1.0\" encoding=\"UTF-8\" ?><result><resultTable name=\"test_list\" rowCount=\"4\" columnCount=\"2\"><columns><c name=\"name\" type=\"string\"/><c name=\"value\" type=\"string\"/></columns><rows><r><c>first</c><c>Item 1</c></r><r><c>fourth</c><c>Item 4</c></r><r><c>second</c><c>Item 2</c></r><r><c>third</c><c>Item 3</c></r></rows></resultTable></result>";
		ArrayList params = new ArrayList();
		HashMap map = new HashMap();
		map.put("first", "Item 1");
		map.put("second", "Item 2");
		map.put("third", "Item 3");
		map.put("fourth", "Item 4");
		params.add(map);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) throws Exception {
				return result.equals(expectedResult);
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), EXPECTED_RESULT);
	}
	private CallRequestExpectedResultPair createParamSet() {
		final String EXPECTED_RESULT =
			"<?xml version=\"1.0\" encoding=\"UTF-8\" ?><result><resultTable name=\"test_list\" rowCount=\"4\" columnCount=\"2\"><columns><c name=\"name\" type=\"string\"/><c name=\"value\" type=\"string\"/></columns><rows><r><c>1</c><c>Item 2</c></r><r><c>2</c><c>Item 4</c></r><r><c>3</c><c>Item 1</c></r><r><c>4</c><c>Item 3</c></r></rows></resultTable></result>";
		ArrayList params = new ArrayList();
		HashSet set = new HashSet();
		set.add("Item 2");
		set.add("Item 1");
		set.add("Item 4");
		set.add("Item 3");
		params.add(set);
		class Pair extends CallRequestExpectedResultPair {
			public Pair(ICallRequest request, String expectedResult) {
				super(request, expectedResult);
			}
			public boolean matchesExpectedResult(String result) throws Exception {
				return result.equals(expectedResult);
			}
		};
		return new Pair(new CallRequest(TEST_INPUT_LIST_RPC_KEY, params), EXPECTED_RESULT);
	}

	private static CallFactory instance = null;
	private ArrayList calls = new ArrayList();
	private boolean callDone = false;
	private MumpsException callDoneException = null;
	private String callDoneResult = null;
	private static final int WAIT_TIME_IN_MS = 5000;
}
