package gov.va.med.hds.cd.m;
import gov.va.med.hds.util.mumps.StringUtil;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.dom4j.Element;
import org.dom4j.Node;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom4j.Dom4jXPath;
/**
 * Call request that reflects a MUMPS RPC call signature (rpc name and parameters).
 */
public class CallRequest implements ICallRequest {
	private static final String PARAM_ELEMENT_NAME = "param";
	private static final String NAME_ELEMENT_NAME = "name";
	private static final String CONTEXT_ELEMENT_NAME = "context";
	private static final String PARAMS_ELEMENT_NAME = "params";
	private static final String SIMPLE_ELEMENT_NAME = "simple";
	private static final String LIST_ELEMENT_NAME = "list";
	private static final String MAP_ELEMENT_NAME = "map";
	private static final String SET_ELEMENT_NAME = "set";
	private static final String LIST_ITEM_ELEMENT_NAME = "item";
	private static final String MAP_ITEM_ELEMENT_NAME = "item";
	private static final String MAP_ITEM_KEY_ELEMENT_NAME = "key";
	private static final String MAP_ITEM_VALUE_ELEMENT_NAME = "value";

	private int timeLimit = TIME_LIMIT_DEFAULT_IN_MILLIS;

	private String name = "";
	private List params = null;
	private String context = "";
	private int hashCodeValue = 0;

	/**
	 * Default timeout for a given call request
	 */
	public static final int TIME_LIMIT_DEFAULT_IN_MILLIS = 30000;

	/**
	 * Constructs a call request with a name and no parameters.
	 * @param callName
	 */
	public CallRequest(String callName) {
		setCallName(callName);
		computeHashCode();
	}
	/**
	 * Constructs a call request given a context and call nae
	 * @param context
	 * @param callName
	 */
	public CallRequest(String context, String callName) {
		setContext(context);
		setCallName(callName);
		computeHashCode();
	}
	/**
	 * Constructs a call request given a context and call nae
	 * @param context
	 * @param callName
	 * @param timeLimit
	 */
	public CallRequest(String context, String callName, int timeLimit) {
		setContext(context);
		setCallName(callName);
		setCallTimeLimit(timeLimit);
		computeHashCode();
	}
	/**
	 * Constructs a call request with a name, no parameters, and time limit.
	 * @param callName
	 * @param timeLimit in millis
	 */
	public CallRequest(String callName, int timeLimit) {
		setCallName(callName);
		setCallTimeLimit(timeLimit);
		computeHashCode();
	}
	/**
	 * Constructs a call request with a name and parameters. Default 
	 * properties are filled in.
	 * 
	 * @param callName name of the RPC
	 * @param params list of parameters
	 */
	public CallRequest(String callName, List params) {
		setCallName(callName);
		setParams(params);
		computeHashCode();
	}
	/**
	 * Constructs a call request with a name, parameters, and time limit.
	 * Default properties are filled in.
	 * @param callName
	 * @param params
	 * @param timeLimit in millis
	 */
	public CallRequest(String callName, List params, int timeLimit) {
		setCallName(callName);
		setParams(params);
		setCallTimeLimit(timeLimit);
		computeHashCode();
	}
	/**
	 * Constructs a call request with a name, parameters and an RPC context
	 * 
	 * @param context the RPC context
	 * @param callName the RPC call name
	 * @param params the RPC parameters
	 */
	public CallRequest(String context, String callName, List params) {
		setContext(context);
		setCallName(callName);
		setParams(params);

		computeHashCode();
	}
	/**
	 * Constructs a call request with a name, parameters, an RPC context, and
	 * time limit
	 * @param context
	 * @param callName
	 * @param params
	 * @param timeLimit in millis
	 */
	public CallRequest(String context, String callName, List params, int timeLimit) {
		setContext(context);
		setCallName(callName);
		setParams(params);
		setCallTimeLimit(timeLimit);
		computeHashCode();
	}

	/**
	 * Constructs a call request from an XML node
	 * @param context
	 * @param name
	 * @param paramsNode
	 * @throws CallRequestException
	 */
	public CallRequest(String context, String name, Node paramsNode) throws CallRequestException {
		setContext(context);
		setCallName(name);
		setupParams(paramsNode);
		computeHashCode();
	}
	/**
	 * Constructs a call request from an XML node
	 * @param paramsNode node containing the params XML
	 */
	public CallRequest(Node node) throws CallRequestException {
		try {
			setCallName(
				((Node) (new Dom4jXPath(NAME_ELEMENT_NAME)).selectSingleNode(node)).getText());

			Node contextNode = (Node) (new Dom4jXPath(CONTEXT_ELEMENT_NAME)).selectSingleNode(node);
			if (contextNode != null)
				setContext(contextNode.getText());

			setupParams((Node) (new Dom4jXPath(PARAMS_ELEMENT_NAME)).selectSingleNode(node));
			computeHashCode();
		} catch (Exception e) {
			throw new CallRequestException(e);
		}
	}

	/**
	 * Constructor CallRequest.
	 */
	protected CallRequest() {
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return hashCodeValue;
	}
	/**
	 * @see java.lang.Object#equals(Object)
	 */
	public boolean equals(Object obj) {
		CallRequest that = (CallRequest) obj;
		return (name.equals(that.name) && context.equals(that.context) && paramsEqual(that));
	}

	/**
	 * Writes the contents to an XML element
	 * @param parent
	 */
	public void write(Element parent) {
		constructName(parent);
		constructContext(parent);
		constructParams(parent);
	}
	private void constructContext(Element parent) {
		if (!context.equals("")) {
			Element contextElement = parent.addElement(CONTEXT_ELEMENT_NAME);
			contextElement.addText(getContext());
		}
	}
	private void constructName(Element parent) {
		Element name = parent.addElement(NAME_ELEMENT_NAME);
		name.addText(getCallName());
	}
	private void constructParams(Element parent) {
		Element params = parent.addElement(PARAMS_ELEMENT_NAME);
		if (getParams() != null) {
			Iterator it = getParams().iterator();
			while (it.hasNext())
				constructParam(it.next(), params);
		}
	}
	private void constructParam(Object o, Element parent) {
		Element param = parent.addElement(CallRequest.PARAM_ELEMENT_NAME);
		if (o instanceof List) {
			Element list = param.addElement(CallRequest.LIST_ELEMENT_NAME);
			Iterator it = ((List) o).iterator();
			while (it.hasNext()) {
				Element item = list.addElement(CallRequest.LIST_ITEM_ELEMENT_NAME);
				item.addCDATA(StringUtil.replaceCDATATags(it.next().toString()));
			}
		} else if (o instanceof Set) {
			Element set = param.addElement(CallRequest.SET_ELEMENT_NAME);
			Iterator it = ((Set) o).iterator();
			while (it.hasNext()) {
				Element item = set.addElement(CallRequest.LIST_ITEM_ELEMENT_NAME);
				item.addCDATA(StringUtil.replaceCDATATags(it.next().toString()));
			}
		} else if (o instanceof Map) {
			Element mapElem = param.addElement(CallRequest.MAP_ELEMENT_NAME);
			Map map = (Map) o;
			Set set = map.keySet();
			Iterator it = set.iterator();
			while (it.hasNext()) {
				Object key = it.next();
				Object value = map.get(key);
				Element item = mapElem.addElement(CallRequest.MAP_ITEM_ELEMENT_NAME);
				Element keyElem = item.addElement(CallRequest.MAP_ITEM_KEY_ELEMENT_NAME);
				keyElem.addText(key.toString());
				Element valueElem = item.addElement(CallRequest.MAP_ITEM_VALUE_ELEMENT_NAME);
				valueElem.addCDATA(StringUtil.replaceCDATATags(value.toString()));
			}
		} else {
			Element simple = param.addElement(CallRequest.SIMPLE_ELEMENT_NAME);
			simple.addCDATA(StringUtil.replaceCDATATags(o.toString()));
		}
	}

	private boolean paramsEqual(CallRequest that) {
		if (params == null && that.params == null)
			return true;
		if (params != null && that.params == null)
			return false;
		if (params == null && that.params != null)
			return false;
		if (params.size() != that.params.size())
			return false;
		Iterator it = params.iterator();
		Iterator thatIt = that.params.iterator();
		while (it.hasNext()) {
			if (!paramsAreEqual(it.next(), thatIt.next()))
				return false;
		}

		return true;
	}

	private boolean paramsAreEqual(Object o1, Object o2) {
		if (o1 instanceof List && o2 instanceof List)
			return testListEquals((List) o1, (List) o2);
		else if (o1 instanceof Map && o2 instanceof Map)
			return testMapEquals((Map) o1, (Map) o2);
		else if (o1 instanceof Set && o2 instanceof Set)
			return testSetEquals((Set) o1, (Set) o2);
		else
			return o1.toString().equals(o2.toString());
	}
	private boolean testSetEquals(Set s1, Set s2) {
		if (s1.size() != s2.size())
			return false;
		Iterator it = s1.iterator();
		while (it.hasNext()) {
			if (!isInSet(it.next().toString(), s2))
				return false;
		}
		return true;
	}
	private boolean isInSet(String s, Set set) {
		Iterator it = set.iterator();
		while (it.hasNext()) {
			if (s.equals(it.next().toString()))
				return true;
		}
		return false;
	}

	private boolean testMapEquals(Map m1, Map m2) {
		if (m1.size() != m2.size())
			return false;
		Iterator it1 = m1.keySet().iterator();
		Iterator it2 = m2.keySet().iterator();
		Object it1Key;
		while (it1.hasNext()) {
			it1Key = it1.next();
			if (!keyMatchesKeyAndValue(it1Key.toString(), m1.get(it1Key).toString(), m2))
				return false;
		}
		return true;
	}

	private boolean keyMatchesKeyAndValue(String key, String value, Map map) {
		Iterator it = map.keySet().iterator();
		Object matchKey;
		Object keyValue;
		while (it.hasNext()) {
			matchKey = it.next();
			if (key.equals(matchKey.toString())) {
				keyValue = map.get(matchKey);
				if (keyValue != null && value.equals(keyValue.toString()))
					return true;
			}
		}
		return false;
	}

	private boolean testListEquals(List c1, List c2) {
		if (c1.size() != c2.size())
			return false;
		Iterator c1It = c1.iterator();
		Iterator c2It = c2.iterator();
		while (c1It.hasNext()) {
			if (!c1It.next().toString().equals(c2It.next().toString()))
				return false;
		}
		return true;
	}

	private void computeHashCode() {
		hashCodeValue = name.hashCode() + context.hashCode() + computeParamsHashCode();
	}

	private int computeParamsHashCode() {
		int result = 0;
		if (params != null && params.size() > 0) {
			Iterator it = params.iterator();
			int val;
			while (it.hasNext()) {
				Object o = it.next();
				if (o instanceof Collection) {
					val = computeCollectionHashCode((Collection) o);
					result += val;
				} else if (o instanceof Map) {
					val = computeMapHashCode((Map) o);
					result += val;
				} else {
					val = o.toString().hashCode();
					result += val;
				}
			}
		}
		return result;
	}

	private int computeMapHashCode(Map map) {
		int result = 0;
		Iterator it = map.keySet().iterator();
		while (it.hasNext()) {
			Object key = it.next();
			result += key.toString().hashCode();
			result += map.get(key).toString().hashCode();
		}
		return result;
	}

	private int computeCollectionHashCode(Collection list) {
		int result = 0;
		Iterator it = list.iterator();
		Object next;
		int val;
		while (it.hasNext()) {
			next = it.next();
			val = next.toString().hashCode();
			result += val;
		}
		return result;
	}

	private void setupParams(Node paramsNode) throws CallRequestException {
		try {
			List p = (new Dom4jXPath(PARAM_ELEMENT_NAME)).selectNodes(paramsNode);
			if (p.size() > 0) {
				params = new ArrayList(p.size());
				Iterator iterator = p.iterator();
				while (iterator.hasNext()) {
					addParam((Node) iterator.next());
				}
			}
		} catch (Exception e) {
			throw new CallRequestException(e);
		}
	}

	private void addParam(Node node) throws Exception {
		Node simple = (Node) (new Dom4jXPath(SIMPLE_ELEMENT_NAME)).selectSingleNode(node);
		if (simple != null) {
			params.add(simple.getText());
			return;
		}
		Node list = (Node) (new Dom4jXPath(LIST_ELEMENT_NAME)).selectSingleNode(node);
		if (list != null) {
			addListParam(list);
			return;
		}
		Node map = (Node) (new Dom4jXPath(MAP_ELEMENT_NAME)).selectSingleNode(node);
		if (map != null) {
			addMapParam(map);
			return;
		}
		Node set = (Node) (new Dom4jXPath(SET_ELEMENT_NAME)).selectSingleNode(node);
		if (set != null) {
			addSetParam(set);
			return;
		}
	}

	private void addMapParam(Node node) throws JaxenException {
		List mapNodes = (new Dom4jXPath(MAP_ITEM_ELEMENT_NAME)).selectNodes(node);
		Map map = new HashMap();
		Iterator it = mapNodes.iterator();
		XPath keyPath = (XPath) new Dom4jXPath(MAP_ITEM_KEY_ELEMENT_NAME);
		XPath valuePath = (XPath) new Dom4jXPath(MAP_ITEM_VALUE_ELEMENT_NAME);
		while (it.hasNext()) {
			Node item = ((Node) it.next());
			String key = ((Node) keyPath.selectSingleNode(item)).getText();
			String value =
				StringUtil.restoreCDATATags(((Node) valuePath.selectSingleNode(item)).getText());
			map.put(key, value);
		}
		params.add(map);
	}

	private void addListParam(Node node) throws JaxenException {
		List listNodes = (new Dom4jXPath(LIST_ITEM_ELEMENT_NAME)).selectNodes(node);
		ArrayList list = new ArrayList(listNodes.size());
		Iterator it = listNodes.iterator();
		while (it.hasNext()) {
			String text = StringUtil.restoreCDATATags(((Node) it.next()).getText());
			list.add(text);
		}
		params.add(list);
	}

	private void addSetParam(Node node) throws JaxenException {
		List listNodes = (new Dom4jXPath(LIST_ITEM_ELEMENT_NAME)).selectNodes(node);
		Set set = new HashSet(listNodes.size());
		Iterator it = listNodes.iterator();
		while (it.hasNext()) {
			String text = StringUtil.restoreCDATATags(((Node) it.next()).getText());
			set.add(text);
		}
		params.add(set);
	}

	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#setCallTimeLimit(int)
	 */
	public void setCallTimeLimit(int timeLimitInMillis) {
		this.timeLimit = timeLimitInMillis;
		computeHashCode();
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#getCallTimeLimit()
	 */
	public int getCallTimeLimit() {
		return timeLimit;
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#setCallName(String)
	 */
	public void setCallName(String name) {
		this.name = name;
		computeHashCode();
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#getCallName()
	 */
	public String getCallName() {
		return name;
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#setParams(Collection)
	 */
	public void setParams(List params) {
		this.params = params;
		computeHashCode();
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#getParams()
	 */
	public List getParams() {
		return params;
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#setContext(String)
	 */
	public void setContext(String context) {
		this.context = context;
		computeHashCode();
	}
	/**
	 * @see gov.va.med.hds.cd.m.callqueue.ICallRequest#getContext()
	 */
	public String getContext() {
		return context;
	}
	/**
	 * @see gov.va.med.hds.cd.m.ICallRequest#copy()
	 */
	public ICallRequest copy() throws CallRequestException {
		CallRequest that = new CallRequest();
		that.name = this.name;
		that.context = this.context;
		try {
			that.setParams(createParamsDeepCopy());
		} catch (Exception e) {
			throw new CallRequestException("Could not perform params deep copy", e);
		}
		return that;
	}

	/**
	 * Creates a deep copy of params.
	 * @return List
	 */
	private List createParamsDeepCopy() throws IllegalAccessException, InstantiationException {
		if (params == null || params.size() == 0)
			return null;
		ArrayList list = new ArrayList(params.size());
		for (int i = 0; i < params.size(); i++) {
			list.add(createParamDeepCopy(params.get(i)));
		}
		return list;
	}

	/**
	 * creates a deep copy of single param
	 * @param object
	 * @return Object
	 */
	private Object createParamDeepCopy(Object object)
		throws IllegalAccessException, InstantiationException {
		if (object == null)
			return null;
		if (object instanceof List) {
			List list = (List) object;
			if (list.size() == 0)
				return null;
			Object[] objects = list.toArray();
			Object[] newObjects = new Object[objects.length];
			for (int i = 0; i < objects.length; i++) {
				newObjects[i] = objects[i].toString();
			}
			return Arrays.asList(newObjects);
		} else if (object instanceof Set) {
			Set set = (Set) object;
			if (set.size() == 0)
				return null;
			Set newSet = new HashSet(set.size());
			newSet.addAll(set);
			return newSet;
		} else if (object instanceof Map) {
			Map newMap = (Map) object.getClass().newInstance();
			newMap.putAll((Map) object);
			return newMap;
		} else {
			return object.toString();
		}
	}

}
