package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.IPluginExceptionManager;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

import java.io.IOException;
import java.io.InputStream;

/**
 * Root class for analyzing collections of class files
 */
public abstract class ClassCollectionAnalyzer implements IConfigConstants {

    /**
     * Constructor for ClassCollectionAnalyzer.
     */
    public ClassCollectionAnalyzer(ClassDataCache cache, IPluginExceptionManager errorManager) {
        super();
        this.errorManager = errorManager;
        this.cache = cache;
        cfa = new ClassFileAnalyzer();
        data = new ClassData();
    }

    /**
     * Returns a list of Strings indicating the classes in the last analyzed
     * jar file.
     */
    public SimpleList getClasses() {
        return data.getClasses();
    }

    /**
     * Returns a list of Strings indicating the classes the classes in the last
     * analyzed jar file are dependant upon.
     * @return Collection
     */
    public SimpleList getDependencies() {
        return data.getDependencies();
    }

    /**
     * Assigns an object to be notified when each class in the jar file is
     * encountered.
     * @param listener
     */
    public void assignClassListener(IClassListener listener) {
        classListener = listener;
    }

    /**
     * Returns a list of Strings indicating the packages in the last analyzed jar file.
     */
    public SimpleList getPackages() {
        return data.getPackages();
    }

    private void checkClasses(ClassFileAnalyzer cfa) {
        if (cfa.isStartup()) {
            if (data.getStartupClass() != null)
                data.setMultipleStartupClasses();
            else
                data.setStartupClass(cfa.getClassName());
        }
        if (cfa.isApplication()) {
            if (data.getApplicationClass() != null)
                data.setMultipleApplicationClasses();
            else
                data.setApplicationClass(cfa.getClassName());
        }
        if (classListener != null) {
            classListener.classFound(cfa.getClassName());
        }
    }

    /**
     * Resets the analyzer for the next set of files.
     * @param checkClasses
     */
    protected void reset() {
        data.clear();
    }

    private void addDependencies(SimpleList list) {
        data.getDependencies().addAll(list);
    }

    private void addClass(String cls) {
        if (!"".equals(cls)) {
            data.getClasses().add(cls);
            int idx = cls.lastIndexOf('.');
            String pkg;
            if (idx > 0)
                pkg = cls.substring(0, idx);
            else
                pkg = IBootConstants.DEFAULT_PACKAGE;
            addPackage(pkg);
        }
    }

    /**
     * Adds a package to the package list
     * @param pkg
     */
    public void addPackage(String pkg) {
        data.getPackages().add(pkg);
    }

    /**
     * If a class in the last analyzed jar file implements the IApplication
     * interface, this method returns the name of that class.  If no class
     * implements the IApplication interface, this method returns null. 
     * @return String
     */
    public String getApplicationClass() {
        return data.getApplicationClass();
    }

    /**
     * If a class in the last analyzed jar file implements the IStartup
     * interface, this method returns the name of that class. If no class
     * implements the IStartup interface, this method returns null.
     * @return String
     */
    public String getStartupClass() {
        return data.getStartupClass();
    }

    /**
     * Analyzes the class contained in the input stream.
     * @param inputStream - the stream containing the class data to be analyzed
     * @param closeStream - if true, will close the input stream on completion
     * @throws InvalidFileStreamException
     */
    protected void analyzeClass(InputStream inputStream, boolean closeStream)
        throws InvalidFileStreamException {
        cfa.analyzeClass(inputStream, closeStream);
        addClass(cfa.getClassName());
        addDependencies(cfa.getDependencies());
        checkClasses(cfa);
    }

    /**
     * Throws an error with the correctly formatted error message
     * @param e
     * @throws InvalidPluginException
     */
    protected void error(Exception e) throws InvalidPluginException {
        if (errorManager == null)
            throw new InvalidPluginException(e);
        else
            errorManager.badPlugin(e);
    }

    /**
     * Validates the collection of classes against invalid conditions
     * @throws InvalidPluginException
     */
    protected void validateClasses() throws InvalidPluginException {
        if (data.hasMultipleStartupClasses())
            if (errorManager == null)
                throw new InvalidPluginException(MULTIPLE_STARTUP_CLASSES);
            else
                errorManager.badPlugin(MULTIPLE_STARTUP_CLASSES);
        if (data.hasMultipleApplicationClasses())
            if (errorManager == null)
                throw new InvalidPluginException(MULTIPLE_APPLICATION_CLASSES);
            else
                errorManager.badPlugin(MULTIPLE_APPLICATION_CLASSES);
    }

    /**
     * Returns true if the correct class information is not contained in the cache
     * @param id
     * @param size
     * @param date
     */
    protected boolean needsClassData(String id, long size, long date) {
    	if(cache == null)
    		return true;
        try {
            return cache.needsClassData(id, size, date, data);
        } catch (IOException e) {
        }
        return true;
    }

    /**
     * Saves class data in the cache
     */
    protected void saveClassData() {
    	if(cache != null)
        	cache.cacheClassData(data);
    }

    private ClassData data;
    private ClassDataCache cache;
    private IPluginExceptionManager errorManager;
    private IClassListener classListener;
    private ClassFileAnalyzer cfa;
}
