package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

/**
 * This data class holds information gathered by a ClassCollectionAnalyzer about a collection of
 * class information  
 */
public class ClassData {

    private SimpleList classes;
    private SimpleList dependencies;
    private SimpleList packages;
    private String applicationClass;
    private String startupClass;
    private boolean multipleApplicationClasses;
    private boolean multipleStartupClasses;
    private String id;
    private long size;
    private long modified;
    private StringBuffer buf;

    /**
     * Constructor for ClassData 
     */
    public ClassData() {
        super();
        buf = new StringBuffer();
        classes = new SimpleList();
        packages = new SimpleList();
        dependencies = new SimpleList();
        clear();
    }

    /**
     * Clears the data contents
     */
    public void clear() {
        classes.clear();
        dependencies.clear();
        packages.clear();
        applicationClass = null;
        startupClass = null;
        multipleApplicationClasses = false;
        multipleStartupClasses = false;
        id = null;
        size = 0;
        modified = 0;
    }

    /**
     * Copies data from one ClassData to another
     * @param data
     */
    public void copy(ClassData data) {
        clear();
        applicationClass = data.applicationClass;
        startupClass = data.startupClass;
        multipleApplicationClasses = data.multipleApplicationClasses;
        multipleStartupClasses = data.multipleStartupClasses;
        classes.addAll(data.classes);
        packages.addAll(data.packages);
        dependencies.addAll(data.dependencies);
        id = data.id;
        size = data.size;
        modified = data.modified;
    }

    /**
     * Returns a new instance of this class with the same data as the cloned object
     */
    public ClassData createClone() {
        ClassData result = new ClassData();
        result.copy(this);
        return result;
    }

    protected static final String CLASS_DATA_ELEMENT = "data";
    private static final String ID_ATTRIBUTE = "id";
    private static final String SIZE_ATTRIBUTE = "size";
    private static final String DATE_ATTRIBUTE = "date";

    private static final String APP_CLASS_ATTRIBUTE = "appClass";
    private static final String STARTUP_CLASS_ATTRIBUTE = "startupClass";
    private static final String MULTI_APP_ATTRIBUTE = "multipleApps";
    private static final String MULTI_STARTUP_ATTRIBUTE = "multipleStartups";
    private static final String FLAG_VALUE = "true";

    private static final String CLASSES_ELEMENT = "classes";
    private static final String PACKAGES_ELEMENT = "packages";
    private static final String REQUIRES_ELEMENT = "requires";
    private static final String DELIM = "*";

    /**
     * Saves the class data to the XML cache.  Returns the index of the saved node.
     * @param xml
     * @param parent
     */
    public int save(ITreeData xml, int parent) {
        if (id != null) {
            int node = xml.createChild(parent, CLASS_DATA_ELEMENT);
            xml.setAttribute(node, ID_ATTRIBUTE, id);
            if (size > 0)
                xml.setAttribute(node, SIZE_ATTRIBUTE, Long.toString(size));
            if (modified > 0)
                xml.setAttribute(node, DATE_ATTRIBUTE, Long.toString(modified));
            if (startupClass != null)
                xml.setAttribute(node, STARTUP_CLASS_ATTRIBUTE, startupClass);
            if (applicationClass != null)
                xml.setAttribute(node, APP_CLASS_ATTRIBUTE, applicationClass);
            if (multipleStartupClasses)
                xml.setAttribute(node, MULTI_STARTUP_ATTRIBUTE, FLAG_VALUE);
            if (multipleApplicationClasses)
                xml.setAttribute(node, MULTI_APP_ATTRIBUTE, FLAG_VALUE);
            addList(classes, xml, node, CLASSES_ELEMENT);
            addList(packages, xml, node, PACKAGES_ELEMENT);
            addList(dependencies, xml, node, REQUIRES_ELEMENT);
            return node;
        }
        return -1;
    }

    private void addList(SimpleList list, ITreeData xml, int node, String tag) {
        if (list.size() > 0) {
            buf.setLength(0);
            for (int i = 0; i < list.size(); i++) {
                buf.append(DELIM);
                buf.append((String) list.get(i));
                if (i == (list.size() - 1))
                    buf.append(DELIM);
            }
            xml.setValue(xml.createChild(node, tag), buf.substring(0, buf.length()));
            buf.setLength(0);
        }
    }

    private void loadList(SimpleList list, ITreeData xml, int node, String tag) {
        int child = xml.findChildID(node, tag);
        if (child >= 0) {
            String value = xml.getValue(child);
            if (value != null) {
                int i = value.indexOf(DELIM);
                if (i >= 0) {
                    boolean working;
                    do {
                        working = false;
                        int j = value.indexOf(DELIM, i + 1);
                        if (j > i) {
                            list.add(value.substring(i + 1, j));
                            i = j;
                            working = true;
                        }
                    } while (working);
                }
            }
        }
    }

    /**
     * Loads class data from the XML cache 
     * @param xml
     * @param node
     */
    public void load(ITreeData xml, int node) {
        clear();
        setID(xml.getAttribute(node, ID_ATTRIBUTE));
        setSize(getLong(xml, node, SIZE_ATTRIBUTE));
        setLastModified(getLong(xml, node, DATE_ATTRIBUTE));
        setStartupClass(xml.getAttribute(node, STARTUP_CLASS_ATTRIBUTE));
        setApplicationClass(xml.getAttribute(node, APP_CLASS_ATTRIBUTE));
        multipleStartupClasses =
            (FLAG_VALUE.equals(xml.getAttribute(node, MULTI_STARTUP_ATTRIBUTE)));
        multipleApplicationClasses =
            (FLAG_VALUE.equals(xml.getAttribute(node, MULTI_APP_ATTRIBUTE)));
        loadList(classes, xml, node, CLASSES_ELEMENT);
        loadList(packages, xml, node, PACKAGES_ELEMENT);
        loadList(dependencies, xml, node, REQUIRES_ELEMENT);
    }

    private long getLong(ITreeData xml, int node, String attribute) {
        String value = xml.getAttribute(node, attribute);
        if (value == null)
            return 0;
        try {
            return Long.parseLong(value);
        } catch (NumberFormatException e) {
            return 0;
        }
    }

    /**
     * Returns the application class
     */
    public String getApplicationClass() {
        return applicationClass;
    }

    /**
     * Returns the list of classes contained in the class collection
     */
    public SimpleList getClasses() {
        return classes;
    }

    /**
     * Returns the list of classes that classes in this collection reference and are dependent on
     */
    public SimpleList getDependencies() {
        return dependencies;
    }

    /**
     * Returns the unique collection id, usually the relative jar file name
     */
    public String getID() {
        return id;
    }

    /**
     * Returns true if the error condition exists where more than one class in the collection 
     * implements the IApplication interface
     */
    public boolean hasMultipleApplicationClasses() {
        return multipleApplicationClasses;
    }

    /**
     * Returns true if the error condition exists where more than one class in the collection 
     * implements the IStartup interface
     */
    public boolean hasMultipleStartupClasses() {
        return multipleStartupClasses;
    }

    /**
     * Returns the list of packages for classes in the collection 
     */
    public SimpleList getPackages() {
        return packages;
    }

    /**
     * Returns the size of the jar file
     */
    public long getSize() {
        return size;
    }

    /**
     * Returns the class implementing the IStartup interface
     */
    public String getStartupClass() {
        return startupClass;
    }

    /**
     * Sets the value of the class implementing the IApplication interface
     * @param string
     */
    public void setApplicationClass(String string) {
        applicationClass = string;
    }

    /**
     * Sets the collection id
     * @param string
     */
    public void setID(String string) {
        id = string;
    }

    /**
     * Sets the error condition where multiple classes implement the IApplication interface
     */
    public void setMultipleApplicationClasses() {
        multipleApplicationClasses = true;
    }

    /**
     * Sets the error condition where multiple classes implement the IStartup interface 
     */
    public void setMultipleStartupClasses() {
        multipleStartupClasses = true;
    }

    /**
     * Sets the size of the collection 
     */
    public void setSize(long size) {
        this.size = size;
    }

    /**
     * Sets the class implementing the IStartup interface 
     * @param string
     */
    public void setStartupClass(String string) {
        startupClass = string;
    }

    /**
     * Returns the date/time stamp of the last modified date on the jar file
     */
    public long getLastModified() {
        return modified;
    }

    /**
     * Sets the date/time stamp of the last modified date on the jar file 
     */
    public void setLastModified(long l) {
        modified = l;
    }
}
