package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.ISavableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.KeyedList;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.io.Writer;
import java.net.URL;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This class manages a cache of class information contained in class collections (usually jars). 
 */
public class ClassDataCache {

    private KeyedList map;
    private File cacheFile;
    private boolean modified;
    private ISavableTreeData xml;
    private dataHandler handler;

    /**
     * Constructor for ClassDataCache 
     */
    public ClassDataCache(File rootDir) {
        super();
        map = new KeyedList();
        if (rootDir != null) {
            if (rootDir.getName().equals(IBootConstants.PLUGINS_DIRECTORY))
                rootDir = rootDir.getParentFile();
            cacheFile = new File(rootDir, IMUMConstants.PLUGIN_CLASS_DATA_CACHE_FILENAME);
            if (cacheFile.exists()) {
                try {
                    map.clear();
                    if (handler == null)
                        handler = new dataHandler();
                    SAXParserFactory.newInstance().newSAXParser().parse(cacheFile, handler);
                } catch (FileNotFoundException e) {
                } catch (SAXException e) {
                } catch (IOException e) {
                } catch (ParserConfigurationException e) {
                } catch (FactoryConfigurationError e) {
                }
            }
        }
        modified = false;
    }

    /**
     * Clears the cache 
     */
    public void clear() {
        map.clear();
    }

    /**
     * Loads the cache from a string
     * @param data
     */
    public void loadCache(String data) {
        try {
            map.clear();
            if (handler == null)
                handler = new dataHandler();
            SAXParserFactory.newInstance().newSAXParser().parse(
                new InputSource(new StringReader(data)),
                handler);
            modified = false;
        } catch (SAXException e) {
        } catch (IOException e) {
        } catch (ParserConfigurationException e) {
        } catch (FactoryConfigurationError e) {
        }
    }

    /**
     * Saves the cache to an external file
     * @throws IOException
     */
    public void save() throws IOException {
        if (modified && (cacheFile != null)) {
            try {
                generateXML();
                xml.save(cacheFile);
                xml.clear();
                modified = false;
            } catch (TreeDataException e) {
            }
        }
    }

    /**
     * Saves the cache to the specified writer
     * @param writer
     * @throws IOException
     */
    public void saveCache(Writer writer) throws IOException {
        if (modified) {
            try {
                generateXML();
                xml.save(writer);
                xml.clear();
                modified = false;
            } catch (TreeDataException e) {
            }
        }
    }

    private void generateXML() throws TreeDataException {
        if (xml == null)
            xml = TreeDataFactory.newSimpleXMLTreeData();
        else
            xml.clear();
        updateXML(xml);
    }

    private static final String CACHE_ELEMENT = "cache";

    private void updateXML(ISavableTreeData xml) {
        xml.clear();
        int root = xml.createRoot(CACHE_ELEMENT);
        for (int i = 0; i < map.size(); i++) {
            ClassData data = (ClassData) map.get(i);
            data.save(xml, root);
        }
    }

//    private void parseXML(ITreeData xml) {
//        int root = xml.getRootID();
//        if (CACHE_ELEMENT.equals(xml.getName(root))) {
//            int node = -1;
//            do {
//                node = xml.findChildID(root, ClassData.CLASS_DATA_ELEMENT, node);
//                if (node >= 0) {
//                    ClassData data = new ClassData();
//                    data.load(xml, node);
//                    cacheClassData(data);
//                }
//            } while (node >= 0);
//        }
//    }

	/**
	 * If the specified id, size and modified date are in the cache, the data object is populated
	 * with information from the cache, and this method returns false.  If it is not in the cache,
	 * this method returns true.
	 * @param id
	 * @param size
	 * @param lastModified
	 * @param data
	 * @return boolean
	 * @throws IOException
	 */
    public boolean needsClassData(String id, long size, long lastModified, ClassData data)
        throws IOException {
        initData(id, size, lastModified, data);
        if (id == null)
            return true;
        if ((size <= 0) || (lastModified <= 0))
            return true;
        ClassData cache = (ClassData) map.get(id);
        if ((cache == null)
            || (cache.getSize() != data.getSize())
            || (cache.getLastModified() != data.getLastModified()))
            return true;
        data.copy(cache);
        return false;
    }

    private void initData(String id, long size, long lastModified, ClassData data)
        throws IOException {
        data.clear();
        data.setID(id);
        data.setSize(size);
        data.setLastModified(lastModified);
    }

    /**
     * Stores class data in the cache.
     * @param classData
     */
    public void cacheClassData(ClassData classData) {
        if (classData.getID() != null) {
            ClassData current = (ClassData) map.get(classData.getID());
            if (current == null)
                map.add(classData.getID(), classData.createClone());
            else
                current.copy(classData);
            modified = true;
        }
    }

    protected static class Utils {

        private static final String pluginDir = "/" + IBootConstants.PLUGINS_DIRECTORY + "/";
        public static String getID(URL url) {
            String path = url.getPath();
            int idx = path.toLowerCase().lastIndexOf(pluginDir);
            if (idx >= 0)
                return path.substring(idx + pluginDir.length());
            return path;
        }

        private static final String mainDir = File.separator + "main" + File.separator;
        public static String getID(File file) {
            String path = file.getAbsolutePath();
            int idx = path.toLowerCase().lastIndexOf(mainDir);
            if (idx >= 0)
                return path.substring(idx + mainDir.length());
            return path;
        }
    }

    private class dataHandler extends DefaultHandler {
        private int root;
        private int node;
        private StringBuffer buffer;
        private ISavableTreeData xml;
        private ClassData data;

        public dataHandler() {
            super();
            buffer = new StringBuffer();
            data = new ClassData();
            try {
                xml = TreeDataFactory.newSimpleXMLTreeData();
            } catch (TreeDataException e) {
            }
        }

        public void startElement(String uri, String localName, String qName, Attributes attributes)
            throws SAXException {
            if (ClassData.CLASS_DATA_ELEMENT.equals(qName)) {
                xml.clear();
                int root = xml.createRoot(qName);
                node = root;
            } else {
                node = xml.createChild(root, qName);
            }
            if (attributes != null) {
                for (int i = 0; i < attributes.getLength(); i++) {
                    xml.setAttribute(node, attributes.getQName(i), attributes.getValue(i));
                }
            }
            buffer.setLength(0);
        }

        public void endElement(String uri, String localName, String qName) throws SAXException {
            if (xml.hasRoot() && (xml.getName(node).equals(qName))) {
                if (buffer.length() > 0) {
                    xml.setValue(node, buffer.substring(0, buffer.length()));
                }
            }
            if (ClassData.CLASS_DATA_ELEMENT.equals(qName) && (xml.hasRoot())) {
                data.load(xml, root);
                cacheClassData(data);
                xml.clear();
            }
            buffer.setLength(0);
        }

        public void characters(char ch[], int start, int length) throws SAXException {
            buffer.append(ch, start, length);
        }
    }
}
