package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.IApplication;
import gov.va.med.hds.cd.mum.boot.IStartup;
import gov.va.med.hds.cd.mum.boot.utils.IIgnoreStartup;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.mum.utils.IIgnoreApplication;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

/**
 * This class is used to analyze .class files and determine various
 * characteristics of the class.
 */
public class ClassFileAnalyzer implements IConfigConstants {

    private static boolean useIgnoreInterfaces = true;

    /**
     * Used to temporarily disable the checking for IIgnoreStartup and
     * IIgnoreApplication interfaces.  This is used only for unit testing
     * purposes.
     * @param value
     */
    public static void SetUseIgnoreInterfaces(boolean value) {
        useIgnoreInterfaces = value;
    }

    private static final int CONSTANT_Class = 7;
    private static final int CONSTANT_Fieldref = 9;
    private static final int CONSTANT_Methodref = 10;
    private static final int CONSTANT_InterfaceMethodref = 11;
    private static final int CONSTANT_String = 8;
    private static final int CONSTANT_Integer = 3;
    private static final int CONSTANT_Float = 4;
    private static final int CONSTANT_Long = 5;
    private static final int CONSTANT_Double = 6;
    private static final int CONSTANT_NameAndType = 12;
    private static final int CONSTANT_Utf8 = 1;
    private static final int CONSTANT_TYPE_COUNT = 14;
    private static final int CONSTANT_INVALID = 13;
    private static final int ACC_INTERFACE = 0x0200;
    private static final int ACC_ABSTRACT = 0x0400;

    private static final int[] dependantTypes = { CONSTANT_Class, CONSTANT_NameAndType };

    private static final String startupClass = IStartup.class.getName().replace('.', '/');
    private static final String ignoreStartupClass =
        IIgnoreStartup.class.getName().replace('.', '/');
    private static final String applicationClass = IApplication.class.getName().replace('.', '/');
    private static final String ignoreApplicationClass =
        IIgnoreApplication.class.getName().replace('.', '/');

    private boolean hasIgnoreApplication;
    private boolean hasIgnoreStartup;
    private boolean isApplication;
    private boolean isStartup;
    private String className;
    private SimpleList dependencies;

    /**
     * Constructor for ClassFileAnalyzer.
     */
    public ClassFileAnalyzer() {
        super();
        dependencies = new SimpleList();
		reset();
    }

    /**
     * Analyzes the class stream and determines class characteristics.  Each
     * call to this method resets the class with new values appropriate to the
     * analyzed class.  This method automatically closes the input stream when
     * finished.
     * @param stream - the class data to analyze (presumably from a .class file)
     * @throws InvalidFileStreamException
     */
    public void analyzeClass(InputStream stream) throws InvalidFileStreamException {
        analyzeClass(stream, true);
    }

    /**
     * Analyzes the class stream and determines class characteristics.  Each
     * call to this method resets the class with new values appropriate to the
     * analyzed class.
     * @param stream - the class data to analyze (presumable from a .class file)
     * @throws InvalidFileStreamException
     * @param stream
     * @param close - when true, the input stream is automatically closed when
     * finished
     * @throws InvalidFileStreamException
     */
    public void analyzeClass(InputStream dataStream, boolean close)
        throws InvalidFileStreamException {
        reset();

        try {
            InputStream stream = dataStream;
            try {
                if (!(stream instanceof BufferedInputStream))
                    stream = new BufferedInputStream(stream);
                if (!(stream instanceof DataInputStream))
                    stream = new DataInputStream(stream);
                decodeData((DataInputStream) stream);
            } finally {
                if (close)
                    stream.close();
            }
        } catch (IOException e) {
            badFile();
        }
    }

    private void reset() {
        hasIgnoreApplication = false;
        hasIgnoreStartup = false;
        isApplication = false;
        isStartup = false;
        className = "";
        dependencies.clear();
    }

    private void decodeData(DataInputStream data) throws InvalidFileStreamException {
        try {
            verifyClassSignature(data);
            readVersion(data);
            int count = getConstantPoolSize(data);
            processConstantPool(data, count);
            processClassData(data);
        } catch (IOException e) {
            throw new InvalidFileStreamException(FILE_IO_ERROR, e);
        }
    }

    private void readVersion(DataInputStream data) throws IOException {
        data.read(buffer, 0, 4);
    }

    private void processClassData(DataInputStream data)
        throws IOException, InvalidFileStreamException {
        int accessFlags = data.readShort();
        boolean isImplementableClass =
            (((accessFlags & ACC_INTERFACE) == 0) && ((accessFlags & ACC_ABSTRACT) == 0));
        int classIndex = data.readUnsignedShort() - 1;
        if ((classIndex < 0) || (classIndex >= Constant.Constants().size()))
            badFile();
        Constant con = (Constant) Constant.Constants().get(classIndex);
        if (con != null) {
            String temp = con.getString(0);
            className = temp.replace('/', '.');
        } else
            badFile();
        if (isImplementableClass) {
            data.read(buffer, 0, 2); // read superclass
            int interfaceCount = data.readUnsignedShort();
            if (interfaceCount > 0) {
                for (int i = 0; i < interfaceCount; i++) {
                    int index = data.readUnsignedShort() - 1;
                    con = (Constant) Constant.Constants().get(index);
                    String temp = con.getString(0);
                    if (startupClass.equals(temp))
                        isStartup = true;
                    if (applicationClass.equals(temp))
                        isApplication = true;
                    if (ignoreStartupClass.equals(temp))
                        hasIgnoreStartup = true;
                    if (ignoreApplicationClass.equals(temp))
                        hasIgnoreApplication = true;
                }
            }
        }
    }

    private static void badFile() throws InvalidFileStreamException {
        badFile("", null);
    }

    private static void badFile(String msg, Exception e) throws InvalidFileStreamException {
        if ("".equals(msg))
            msg = BAD_FILE;
        if (e == null)
            throw new InvalidFileStreamException(msg);
        else
            throw new InvalidFileStreamException(msg, e);
    }

    private void processConstantPool(DataInputStream data, int count)
        throws IOException, InvalidFileStreamException {
        Constant.Reset(count);
        int i = 0;
        while (i < count) {
            Constant.loadConstant(data);
            if (Constant.twoEntryConstant())
                i++;
            i++;
        }
        addDependencies();
    }

    private int getConstantPoolSize(DataInputStream data) throws IOException {
        return data.readUnsignedShort() - 1;
    }

    private void verifyClassSignature(DataInputStream data)
        throws IOException, InvalidFileStreamException {
        data.read(buffer, 0, 4); // read magic
        if ((buffer[0] != (byte) 0xCA)
            || (buffer[1] != (byte) 0xFE)
            || (buffer[2] != (byte) 0xBA)
            || (buffer[3] != (byte) 0xBE))
            badFile();
    }

    private void addDependencies() {
        List list = Constant.Constants();
        for (int i = 0; i < dependantTypes.length; i++) {
            for (int j = 0; j < list.size(); j++) {
                Constant con = (Constant) list.get(j);
                if ((con != null) && (con.code == dependantTypes[i]))
                    addDependency(con);
            }
        }
    }

    private void addDependency(Constant con) {
        addDependantString(con.getString(0));
        addTypeDependency(con.getString(1));
    }

    private void addDependantString(String str) {
        if ("".equals(str))
            return;
        boolean add = true;
        for (int i = 0; i < encodedSpec.length; i++) {
            if (str.startsWith(encodedSpec[i])) {
                addTypeDependency(str);
                add = false;
                break;
            }
        }
        if (add)
            addDependency(str);
    }

    private void addDependency(String s) {
        if ((!"".equals(s)) && (s.indexOf('/') > 0)) {
            String str = s.replace('/', '.');
            if ((!"".equals(str)) && (!str.startsWith(BOOT_PACKAGE))) {
                if (!JVMAnalyzer.getClasses().contains(str))
                    dependencies.add(str);
            }
        }
    }

    private final char refChar = 'L';
    private final char endrefChar = ';';
    private static final String[] encodedSpec = { "[", "(" };

    private void addTypeDependency(String s) {
        if ("".equals(s))
            return;
        int i = 0, len = s.length(), j;
        while (i < len) {
            if (s.charAt(i) == refChar) {
                j = s.indexOf(endrefChar, i);
                if (j > i) {
                    addDependency(s.substring(i + 1, j));
                    i = j;
                } else
                    i++;
            } else
                i++;
        }
    }

    /**
     * Returns the Collection of classes the last analyzed class is dependant upon.
     * @return Collection
     */
    public SimpleList getDependencies() {
        return dependencies;
    }

    /**
     * Returns the last analyzed class name.
     * @return String
     */
    public String getClassName() {
        return className;
    }

    /**
     * Returns true if the last analyzed class implements the IStartup interface
     * @see gov.va.med.hds.cd.mum.utils.IStartup
     * @return boolean
     */
    public boolean isStartup() {
        boolean result = isStartup;
        if (result && useIgnoreInterfaces && hasIgnoreStartup)
            result = false;
        return result;
    }

    /**
     * Returns true if the last analyzed class implements the IApplication
     * interface.
     * @see gov.va.med.hds.cd.mum.IApplication
     * @return boolean
     */
    public boolean isApplication() {
        boolean result = isApplication;
        if (result && useIgnoreInterfaces && hasIgnoreApplication)
            result = false;
        return result;
    }

    private byte[] buffer = new byte[4];

    private static class Constant {
        byte code;
        String value;
        boolean isUTF;
        int index1;
        int index2;

        static int maxCount;
        private static ArrayList pool = new ArrayList();
        private static ArrayList used = new ArrayList();

        private Constant() {
        }

        private static ConstantDataReader[] reader;

        static {
            StringConstantDataReader strReader = new StringConstantDataReader();
            DoubleStringConstantDataReader dblStrReader = new DoubleStringConstantDataReader();
            NullConstantDataReader nullReader = new NullConstantDataReader();
            reader = new ConstantDataReader[CONSTANT_TYPE_COUNT];
            for (int i = 0; i < reader.length; i++)
                reader[i] = nullReader;
            reader[CONSTANT_Class] = strReader;
            reader[CONSTANT_String] = strReader;
            reader[CONSTANT_Fieldref] = dblStrReader;
            reader[CONSTANT_Methodref] = dblStrReader;
            reader[CONSTANT_InterfaceMethodref] = dblStrReader;
            reader[CONSTANT_NameAndType] = dblStrReader;
            reader[CONSTANT_Integer] = new IntConstantDataReader();
            reader[CONSTANT_Float] = new FloatConstantDataReader();
            reader[CONSTANT_Long] = new LongConstantDataReader();
            reader[CONSTANT_Double] = new DoubleConstantDataReader();
            reader[CONSTANT_Utf8] = new UTFConstantDataReader();
        }

        public static void Reset(int count) {
            maxCount = count;
            for (int i = 0; i < used.size(); i++) {
                Object obj = used.get(i);
                if (obj != null)
                    pool.add(obj);
            }
            used.clear();
        }

        private static boolean lastConstantWasTwoEntryConstant;

        public static boolean twoEntryConstant() {
            return lastConstantWasTwoEntryConstant;
        }

        public static void loadConstant(DataInputStream data) throws InvalidFileStreamException {

            lastConstantWasTwoEntryConstant = false;
            Constant con;
            if (pool.size() > 0)
                con = getConstantFromPool(data);
            else
                con = getNewConstant(data);
            used.add(con);
        }

        private static Constant getNewConstant(DataInputStream data)
            throws InvalidFileStreamException {
            Constant con = new Constant();
            if (!con.canLoadConstant(data)) {
                pool.add(con);
                checkForTwoEntryConstant(con);
                con = null;
            }
            return con;
        }

        private static Constant getConstantFromPool(DataInputStream data)
            throws InvalidFileStreamException {
            Constant con = null;
            int cnt = pool.size();
            con = (Constant) pool.get(cnt - 1);
            if (con.canLoadConstant(data))
                pool.remove(cnt - 1);
            else {
                checkForTwoEntryConstant(con);
                con = null;
            }
            return con;
        }
        private static void checkForTwoEntryConstant(Constant con) {
            if (con.twoEntries) {
                used.add(null);
                lastConstantWasTwoEntryConstant = true;
            }
        }

        boolean twoEntries;

        public boolean canLoadConstant(DataInputStream data) throws InvalidFileStreamException {
            try {
                code = data.readByte();
                if ((code < 0) || (code >= CONSTANT_TYPE_COUNT))
                    code = CONSTANT_INVALID;
                return reader[code].canLoadConstant(this, data);
            } catch (IOException e) {
                throw new InvalidFileStreamException(e.getMessage(), e);
            }
        }

        public byte getCode() {
            return code;
        }

        public String getString(int index) {
            if (isUTF) {
                return value;
            }
            Constant con = null;
            if ((index == 0) && (index1 >= 0))
                con = (Constant) used.get(index1);
            else if ((index == 1) && (index2 >= 0))
                con = (Constant) used.get(index2);
            if (con == null) {
                return "";
            } else {
                return con.getString(index);
            }
        }

        public int count() {
            if (index2 >= 0)
                return 2;
            else
                return 1;
        }

        public static List Constants() {
            return used;
        }
    }

    private static abstract class ConstantDataReader {

        public boolean canLoadConstant(Constant con, DataInputStream data)
            throws InvalidFileStreamException {
            con.value = "";
            con.index1 = -1;
            con.index2 = -1;
            con.isUTF = false;
            con.twoEntries = false;
            boolean result = false;
            try {
                result = canLoad(con, data);
            } catch (IOException e) {
                badFile(e);
            }
            return result;
        }

        protected abstract boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException;

        protected void badFile() throws InvalidFileStreamException {
            throw new InvalidFileStreamException();
        }

        protected void badFile(Exception e) throws InvalidFileStreamException {
            throw new InvalidFileStreamException(e.getMessage(), e);
        }

    }

    private static class StringConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            con.index1 = data.readUnsignedShort() - 1;
            if ((con.index1 < 0) || (con.index1 >= Constant.maxCount))
                super.badFile();
            return true;
        }
    }

    private static class DoubleStringConstantDataReader extends StringConstantDataReader {
        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            super.canLoad(con, data);
            con.index2 = data.readUnsignedShort() - 1;
            if ((con.index2 < 0) || (con.index2 >= Constant.maxCount))
                super.badFile();
            return true;
        }
    }

    private static class IntConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            data.readInt();
            return false;
        }
    }

    private static class FloatConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            data.readFloat();
            return false;
        }
    }

    private static class LongConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            data.readLong();
            con.twoEntries = true;
            return false;
        }
    }

    private static class DoubleConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            data.readDouble();
            con.twoEntries = true;
            return false;
        }
    }

    private static class UTFConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            con.value = data.readUTF();
            con.isUTF = true;
            return true;
        }
    }

    private static class NullConstantDataReader extends ConstantDataReader {

        protected boolean canLoad(Constant con, DataInputStream data)
            throws InvalidFileStreamException, IOException {
            super.badFile();
            return false;
        }
    }
}