package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.Utils;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Provides a command line interface to the MUM.Config, allowing the building of
 * configuration files from the command line instead of from a user interface.
 */
public class CommandLineUtility implements IMUMConstants, IBootConstants {

    public static final String PATH_ARG = OPTION_PREFIX + "path" + ARG_ASSIGNMENT;
    public static final String MAPC_ARG = OPTION_PREFIX + "mapc";
    public static final String MAPP_ARG = OPTION_PREFIX + "mapp";
    public static final String SETTINGS_ARG = OPTION_PREFIX + "settings" + ARG_ASSIGNMENT;
    public static final String HELP_ARG = OPTION_PREFIX + "?";
    public static final String DEFAULT_CONFIG_FILE = "configuration.xml";
    public static final String HELP_TEXT =
        "Usage: mumconfig [options]"
            + EOL
            + "Options:"
            + EOL
            + "  "
            + PATH_ARG
            + "gold               An optional path for the gold directory where"
            + EOL
            + "                           the plugins subdirectory can be found.  If not"
            + EOL
            + "                           specified, the current directory is used."
            + EOL
            + "  "
            + SETTINGS_ARG
            + "filename.xml   An optional path and filename of an xml file"
            + EOL
            + "                           containing the configuration settings section."
            + EOL
            + "  configfilename.xml       The name of the configuration file to create."
            + EOL
            + "                           If no file name is specified, a default file"
            + EOL
            + "                           name of \"configuration.xml\" is used."
            + EOL
            + "  "
            + MAPC_ARG
            + "                    Creates a .mapc file of plugin/class accessability"
            + EOL
            + "  "
            + MAPP_ARG
            + "                    Creates a .mapp file of plugin/package accessability"
            + EOL
            + "  "
            + HELP_ARG
            + "                       Displays this help message."
            + EOL;

    /**
     * Executes the command line interface to the MUM.Config.
     * @param args
     */
    public static void main(String[] args) {
        if (Utils.hasArgument(args, HELP_ARG)) {
            System.out.print(HELP_TEXT);
            return;
        }
        String path = Utils.getArgument(args, PATH_ARG);
        String settings = Utils.getArgument(args, SETTINGS_ARG);
        String xml = Utils.getArgument(args, null);
        manager = PluginManagerFactory.newPluginManager();
        if (path == null)
            path = System.getProperty(USER_CURRENT_DIR_PROPERTY);
        manager.assignGoldDirectory(new File(path));
        if (xml == null)
            xml = DEFAULT_CONFIG_FILE;
        File configFile = new File(xml);
        URL settingsURL = null;
        if (settings != null) {
            try {
                settingsURL = MUMUtils.newURL(settings);
            } catch (MalformedURLException e) {
            }
        }
        Configuration config;
        try {
            config = new Configuration(manager);
            if (config.errors().hasErrors())
                config.errors().printErrors(System.out);
            else {
                boolean doMap;
                boolean mapClasses = Utils.hasArgument(args, MAPC_ARG);
                if (mapClasses)
                    doMap = true;
                else
                    doMap = Utils.hasArgument(args, MAPP_ARG);

                config.saveXML(configFile, settingsURL, doMap, mapClasses);
            }
        } catch (PluginLocatorException e) {
            error(e);
        } catch (InvalidConfigurationException e) {
            error(e);
        } catch (IOException e) {
            error(e);
        }
    }

    private static StringBuffer errorBuffer;

    private static void error(Throwable e) {
        if (errorBuffer == null)
            errorBuffer = new StringBuffer();
        String message;
        if (e.getMessage() == null)
            message = e.getClass().getName();
        else {
            message = e.getMessage();
        }
        if (errorBuffer.indexOf(message) < 0) {
            errorBuffer.append(message);
            errorBuffer.append(EOL);
        }
        if (e.getCause() != null)
            error(e.getCause());
        else {
            System.out.println(errorBuffer.substring(0, errorBuffer.length()));
            errorBuffer.setLength(0);
        }
    }

    private static IPluginManager manager;

}
