package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.DependentPlugin;
import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.PluginID;
import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Represents an existing plugin found in the gold directory.  Combines
 * information from the actual, installed plugin, and from the plugin's
 * manifest.xml file.
 */
public class ConfigPlugin extends DependentPlugin implements IConfigPlugin, IConfigConstants {

    /**
     * Creates a ConfigPlugin object based on the manifest and existing file
     * information.
     */
    public ConfigPlugin(ClassDataCache cache, URL manifestURL) throws InvalidPluginException {
        super();
        if (manifestURL == null)
            throw new NullPointerException();
        init(cache, manifestURL.getPath());
        parseManifestURL(manifestURL);
        parseJars();
        resolvePublications();
    }

    /**
     * Constructor for ConfigPlugin 
     */
    public ConfigPlugin(ClassDataCache cache, File manifestFile, boolean parseDirectory)
        throws InvalidPluginException {
        super();
        if (manifestFile == null)
            throw new NullPointerException();
        init(cache, manifestFile.getPath().replace(File.separatorChar, '/'));
        try {
            parseManifestURL(manifestFile.toURL());
        } catch (MalformedURLException e) {
            badPlugin(e);
        }
        if (parseDirectory)
            parseDir(manifestFile);
        resolvePublications();
    }

    private void resolvePublications() throws InvalidPluginException {
        publications.resolvePublications(classes, getPublished());
        Collections.sort(getPublished().getList());
    }

    private void init(ClassDataCache cache, String path) {
        parseManifestPath(path);
        files = new ArrayList();
        classes = new SimpleList();
        startupClass = null;
        publications = new PluginPublicationManager(this);
        classDependencies = new SimpleList();
        extensions = new SimpleList();
        extensionPoints = new SimpleList();
        declaredClassDependencies = new SimpleList();
        pluginDependencies = new ListMap();
        this.cache = cache;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#publishAllClasses()
     */
    public boolean publishAllClasses() {
        return publications.publishAllClasses();
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getClasses()
     */
    public SimpleList getClasses() {
        return classes;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getFile(int)
     */
    public ConfigFileEntry getFile(int index) {
        return (ConfigFileEntry) files.get(index);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IBasePlugin#getFileCount()
     */
    public int getFileCount() {
        return files.size();
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getIgnoreCode()
     */
    public int getIgnoreCode() {
        return ignoreCode;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getClassDependencies()
     */
    public SimpleList getClassDependencies() {
        return classDependencies;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getPluginDependencies()
     */
    public ListMap getPluginDependencies() {
        return pluginDependencies;
    }

    /**
     * Updates the dependency list with plugin prerequisites from the
     * manifest.
     * @param xml - The passed in manifest xml.
     * @throws InvalidPluginException
     */
    protected void getPrerequisites(ITreeData xml) throws InvalidPluginException {
        int root = xml.getRootID();
        int child = xml.findChildID(root, IPluginConstants.PREREQUISITES_ELEMENT);
        if (child >= 0) {
            if (xml.findChildID(root, IPluginConstants.PREREQUISITES_ELEMENT, child) >= 0)
                badPlugin(IMUMConstants.MULTIPLE_PREREQUISITES_SECTIONS);
            getPlugins(xml, child);
        }
    }

    private void getPlugins(ITreeData xml, int plugins) throws InvalidPluginException {
        for (int i = 0; i < xml.getChildCount(plugins); i++) {
            int child = xml.getChildID(plugins, i);
            if (IPluginConstants.PREREQUISITE_ELEMENT.equals(xml.getName(child))) {
                String plugin = xml.getAttribute(child, IPluginConstants.PLUGIN_ELEMENT);
                if ((plugin == null) || (plugin.length() == 0))
                    badPlugin(IMUMConstants.INVALID_PREREQUISITES_ENTRY);
                Version ver = null;
                try {
                    ver = new Version(xml.getAttribute(child, IPluginConstants.VERSION_ATTRIBUTE));
                } catch (VersionFormatException e) {
                    badPlugin(IMUMConstants.INVALID_PREREQUISITES_ENTRY, e);
                }
                if (pluginDependencies.get(plugin).size() > 0)
                    badPlugin(IMUMConstants.DUPLICATE_DEPENDENCY_TXT + plugin);
                if (plugin.equals(getID()))
                    badPlugin(IMUMConstants.SELF_DEPENDENCY_TXT);
                PluginDependency dependency =
                    PluginDependency.getDependency(
                        PluginID.getPluginID(plugin, ver.toString()),
                        xml.getAttribute(child, IPluginConstants.MATCH_ATTRIBUTE),
                        IPluginConstants.EXPORT_VALUE.equals(
                            xml.getAttribute(child, IPluginConstants.EXPORT_ATTRIBUTE)));
                pluginDependencies.add(dependency.getPluginID().getID(), dependency);
            }
        }
    }

    private void parseManifestPath(String path) {
        setPluginPath(path);
        StringBuffer buf = new StringBuffer(path);
        int idx = buf.lastIndexOf("/");
        if (idx >= 0)
            buf.setLength(idx);
        idx = buf.lastIndexOf("/");
        if (idx >= 0)
            setSourceDirectory(buf.substring(idx + 1));
    }

    private void parseManifestURL(URL url) throws InvalidPluginException {
        files.add(new ConfigFileEntry(url));
        ITreeData xml = getManifestXML(url);
        int root = xml.getRootID();
        getPluginInfo(xml, root);
        getDistributionInfo(xml, root, url);
        getPrerequisites(xml);
        getPublished(xml, root);
        setPluginPath(getDirectory());
        getExtensionInfo(xml, root);
    }

    private void getPluginInfo(ITreeData xml, int root) throws InvalidPluginException {
        String temp = xml.getAttribute(root, IPluginConstants.ID_ATTRIBUTE);
        if ((temp == null) || (temp.length() == 0))
            badPlugin(MISSING_PLUGIN_ID);
        setID(temp);
        try {
            setVersion(new Version(xml.getAttribute(root, IPluginConstants.VERSION_ATTRIBUTE)));
        } catch (VersionFormatException e) {
            badPlugin(e);
        }
        temp = xml.getAttribute(root, IPluginConstants.MODULE_ATTRIBUTE);
        if ((temp == null) || (temp.length() == 0))
            temp = getID();
        setModule(temp);
        temp = xml.getAttribute(root, IGNORE_ATTRIBUTE);
        ignoreCode = IGNORE_NONE;
        if (temp != null) {
            if (temp.equals(IGNORE_UNAVAILABLE_VALUE))
                ignoreCode = IGNORE_UNAVAILABLE;
        }
        declaredClassDependencies.add(xml.getAttribute(root, IPluginConstants.CLASS_ATTRIBUTE));
        setDirectory(getID() + VERSION_SEPARATOR + getVersion().toString());
        setPluginPath(getDirectory() + "/" + IBootConstants.MANIFEST_FILE);
        setName(xml.getAttribute(root, IPluginConstants.NAME_ATTRIBUTE));
        initPlugin();
    }

    private void getDistributionInfo(ITreeData xml, int root, URL url)
        throws InvalidPluginException {
        int dist = xml.findChildID(root, DISTRIBUTION_ELEMENT);
        StringBuffer buf = new StringBuffer(url.toString());
        int idx = buf.lastIndexOf("/");
        int len = 0;
        if (idx >= 0)
            len = idx + 1;
        if (dist >= 0) {
            if (xml.findChildID(root, DISTRIBUTION_ELEMENT, dist) >= 0)
                badPlugin(MULTIPLE_DISTRIBUTION_SECTIONS);
            getFiles(xml, dist, buf, len);
        }
    }

    private void getFiles(ITreeData xml, int dist, StringBuffer buf, int len)
        throws InvalidPluginException {
        for (int i = 0; i < xml.getChildCount(dist); i++) {
            getFile(buf, len, xml, xml.getChildID(dist, i));
        }
    }

    private void getFile(StringBuffer buf, int len, ITreeData xml, int file)
        throws InvalidPluginException {
        if (FILE_ELEMENT.equals(xml.getName(file))) {
            buf.setLength(len);
            String fname = xml.getAttribute(file, FILENAME_ATTRIBUTE);
            if ((fname == null) || (fname.length() == 0))
                badPlugin(INVALID_FILE_ENTRY);
            buf.append(fname);
            try {
                files.add(new ConfigFileEntry(MUMUtils.newURL(buf.substring(0, buf.length()))));
            } catch (MalformedURLException e) {
                badPlugin(INVALID_FILE_ENTRY, e);
            }
        }
    }

    private static final String BOOT_JAR_URL = "/" + BOOT_JAR;

    private void parseJars() throws InvalidPluginException {
        JarFileAnalyzer analyzer = null;
        for (int i = 0; i < files.size(); i++) {
            ConfigFileEntry entry = (ConfigFileEntry) files.get(i);
            if (entry.getURL().getPath().endsWith(IBootConstants.JAR_EXT)) {
                if (analyzer == null)
                    analyzer = new JarFileAnalyzer(cache, this);
                analyzer.analyzeJar(entry.getURL());
                setPackages(entry, analyzer);
                getDependencies(analyzer);
                checkClasses(analyzer);
                if (analyzer.getPackages().contains(BOOT_PACKAGE_NAME)) {
                    if (entry.getURL().getPath().endsWith(BOOT_JAR_URL)) {
                        for (int j = 0; j < analyzer.getDependencies().size(); j++) {
                            String cls = (String) analyzer.getDependencies().get(j);
                            if (!cls.startsWith(BOOT_PACKAGE))
                                throw new InvalidPluginException(
                                    ILLEGAL_BOOT_DEPENDENCY_START
                                        + cls
                                        + ILLEGAL_BOOT_DEPENDENCY_END);
                        }
                    }
                }
            }
        }
    }

    private void parseDir(File manifestFile) throws InvalidPluginException {
        File file = new File(manifestFile.getParentFile(), "bin");
        if (!file.exists())
            file = new File(manifestFile.getParentFile(), "classes");
        if (!file.exists()) {
            File[] files = manifestFile.getParentFile().listFiles();
            JarFileAnalyzer analyzer = null;
            for (int i = 0; i < files.length; i++) {
                if (files[i].getName().toLowerCase().endsWith(IBootConstants.JAR_EXT)) {
                    if (analyzer == null)
                        analyzer = new JarFileAnalyzer(cache, this);
                    try {
                        analyzer.analyzeJar(files[i].toURL());
                        getDependencies(analyzer);
                        checkClasses(analyzer);
                    } catch (MalformedURLException e) {
                        badPlugin(e);
                    }
                }
            }
        } else {
            DirectoryAnalyzer analyzer = new DirectoryAnalyzer(cache, this);
            analyzer.analyzeDirectory(file);
            getDependencies(analyzer);
            checkClasses(analyzer);
        }
    }

    private void setPackages(ConfigFileEntry entry, JarFileAnalyzer analyzer) {
        SimpleList pkgs = analyzer.getPackages();
        if (pkgs.size() > 0)
            entry.setPackages(new SimpleList(pkgs));
    }

    private void getDependencies(ClassCollectionAnalyzer analyzer) {
        classes.addAll(analyzer.getClasses());
        for (int i = 0; i < analyzer.getDependencies().size(); i++) {
            classDependencies.add((String) analyzer.getDependencies().get(i));
        }
    }

    private void checkClasses(ClassCollectionAnalyzer analyzer) throws InvalidPluginException {
        if (analyzer.getApplicationClass() != null) {
            if (getApplicationClassName() != null)
                badPlugin(MULTIPLE_APPLICATION_CLASSES);
            setApplicationClass(analyzer.getApplicationClass());
        }
        if (analyzer.getStartupClass() != null) {
            if (getStartupClassName() != null)
                badPlugin(MULTIPLE_STARTUP_CLASSES);
            setStartupClass(analyzer.getStartupClass());
        }
    }

    /**
     * Sets the startup class
     * @param startupClass
     */
    protected void setStartupClass(String startupClass) {
        this.startupClass = startupClass;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getStartupClassName()
     */
    public String getStartupClassName() {
        return startupClass;
    }

    private void getPublished(ITreeData xml, int root) throws InvalidPluginException {
        int pub = xml.findChildID(root, IPluginConstants.PUBLISHED_ELEMENT);
        if (pub >= 0) {
            if (xml.findChildID(root, IPluginConstants.PUBLISHED_ELEMENT, pub) >= 0)
                badPlugin(MULTIPLE_PUBLISHED_SECTIONS);
            getPublishedClasses(xml, pub);
        }
    }

    private void getPublishedClasses(ITreeData xml, int pub) throws InvalidPluginException {
        int publishCount = 0;
        for (int i = 0; i < xml.getChildCount(pub); i++) {
            getPublishedClass(xml, xml.getChildID(pub, i));
            publishCount++;
        }
        if (publications.publishAllClasses() && (publishCount != 1))
            badPlugin(TOO_MANY_PUBLISHED_ENTRIES);
    }

    private void getPublishedClass(ITreeData xml, int root) throws InvalidPluginException {
        if (IPluginConstants.PUBLISH_ELEMENT.equals(xml.getName(root))) {
            String cls = xml.getAttribute(root, IPluginConstants.CLASS_ATTRIBUTE);
            if ((cls == null) || (cls.length() == 0))
                badPlugin(INVALID_PUBLISHED_ENTRY);
            publications.addPublication(cls);
        }
    }

    private void getExtensionInfo(ITreeData xml, int root) {
        int node = xml.findChildID(root, IPluginConstants.EXTENSION_POINTS_ELEMENT);
        int idx;
        if (node >= 0) {
            idx = -1;
            do {
                idx = xml.findChildID(node, IPluginConstants.EXTENSION_POINT_ELEMENT, idx);
                if (idx >= 0)
                    extensionPoints.add(
                        getID() + '.' + xml.getAttribute(idx, IPluginConstants.ID_ATTRIBUTE));
            } while (idx >= 0);
        }
        node = xml.findChildID(root, IPluginConstants.EXTENSIONS_ELEMENT);
        if (node >= 0) {
            idx = -1;
            do {
                idx = xml.findChildID(node, IPluginConstants.EXTENSION_ELEMENT, idx);
                if (idx >= 0)
                    extensions.add(xml.getAttribute(idx, IPluginConstants.EXTENDS_ATTRIBUTE));
            } while (idx >= 0);
        }
    }

    /**
     * Returns the plugin id and version.
     */
    public String toString() {
        return getID() + " v" + getVersion().toString();
    }

    private List files;
    private PluginPublicationManager publications;
    private SimpleList classes;
    private int ignoreCode;
    private String startupClass;
    private SimpleList classDependencies;
    private SimpleList extensions;
    private SimpleList extensionPoints;
    private SimpleList declaredClassDependencies;
    private ListMap pluginDependencies;
    private ClassDataCache cache;

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getExtensionPoints()
     */
    public SimpleList getExtensionPoints() {
        return extensionPoints;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getExtensions()
     */
    public SimpleList getExtensions() {
        return extensions;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IConfigPlugin#getDeclaredClassDependencies()
     */
    public SimpleList getDeclaredClassDependencies() {
        return declaredClassDependencies;
    }
}
