package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.utils.ILoadableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.ISavableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.KeyedList;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;
import gov.va.med.hds.cd.mum.utils.ChecksumCalculator;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * A class containing a collection of IConfigPlugins that, together, represent a
 * plugin workstation configuration.
 */
public class Configuration implements IConfigConstants, Comparator {

    /**
     * Creates a configuration using IConfigPlugins stored in the
     * IPluginManager.
     * @param manager
     * @throws PluginLocatorException
     */
    public Configuration(IPluginManager manager) throws PluginLocatorException {
        super();
        plugins = new ListMap();
        errors = new ErrorManager();
        tempErrors = new ErrorManager();
        crc = new ChecksumCalculator();
        decisionList = new ArrayList();
        buffer = new StringBuffer();
        tempList = new SimpleList();
        parentList = new ListMap();
        tempMap = new ListMap();
        validatedChains = new SimpleList();
        this.manager = manager;
        manager.load();
        manager.loadPluginInformation();
        List keys = manager.getPlugins().getKeys();
        for (int i = 0; i < keys.size(); i++) {
            Object key = keys.get(i);
            List list = manager.getPlugins().get(key);
            for (int j = 0; j < list.size(); j++) {
                IConfigPlugin plugin = (IConfigPlugin) list.get(j);
                plugins.add(plugin.getID(), plugin);
            }
        }
        temp = new ArrayList(plugins.size());
        validate();
    }

    /**
     * Validates the configuration to determine if there are any configuration
     * errors.
     */
    public void validate() {
        errors.clear();
        validateStartup();
        validateApplication();
        updatePlugins(); // call before validate classes!
        validateClasses();
        validateExtensions();
        validateLibraries();
    }

    private void validateStartup() {
        IConfigPlugin startPlugin = null;
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            if (plugin.getStartupClassName() != null) {
                if (startPlugin == null)
                    startPlugin = plugin;
                else
                    errors.addError(Error.MULTIPLE_STARTUP_PLUGINS, startPlugin, plugin);
            }
        }
        if (startPlugin == null)
            errors.addError(Error.MISSING_STARTUP_PLUGIN);
        else {
            boolean noStartupJar = true;
            for (int i = 0; i < startPlugin.getFileCount(); i++) {
                if (getFileName(startPlugin.getFile(i)).equals(BOOT_JAR)) {
                    noStartupJar = false;
                    break;
                }
            }
            if (noStartupJar)
                errors.addError(Error.STARTUP_PLUGIN_MISSING_BOOT_JAR, startPlugin);
        }
    }

    private void validateApplication() {
        IConfigPlugin appPlugin = null;
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            if (plugin.getApplicationClassName() != null) {
                if (appPlugin == null)
                    appPlugin = plugin;
                else
                    errors.addError(Error.MULTIPLE_APPLICATION_PLUGINS, appPlugin, plugin);
            }
        }
        if (appPlugin == null)
            errors.addError(Error.MISSING_APPLICATION_PLUGIN);
    }

    private void updatePlugins() {
        updatePluginDependencies();
        makeDecisions();
        validateCyclicDependencies();
    }

    private void validateCyclicDependencies() {
        tempList.clear();
        validatedChains.clear();
        if (hasCycles()) {
            for (int i = 0; i < plugins.size(); i++) {
                IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
                if (!validatedChains.contains(plugin)) {
                    tempList.clear();
                    check4Cycle(plugin);
                }
            }
        }
    }

    private void check4Cycle(IConfigPlugin plugin) {
        validatedChains.add(plugin);
        if (tempList.contains(plugin)) {
            int last = tempList.getList().indexOf(plugin);
            Error error =
                errors.addError(Error.CYCLIC_DEPENDENCY, getSortedMessage(tempList, last));
            if (error.getPluginList().size() == 0) {
                error.addPlugin(plugin);
                for (int i = tempList.size() - 1; i >= last; i--)
                    error.addPlugin((IConfigPlugin) tempList.get(i));
            }
        } else {
            List parents = parentList.get(plugin);
            tempList.add(plugin);
            int length = tempList.size();
            for (int i = 0; i < parents.size(); i++) {
                if (i > 0) {
                    while (tempList.size() > length)
                        tempList.remove(tempList.size() - 1);
                }
                check4Cycle((IConfigPlugin) parents.get(i));
            }
        }
    }

    private boolean hasCycles() {
        tempMap.clear();
        parentList.clear();
        temp.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin parent = (IConfigPlugin) plugins.get(i);
            temp.add(parent);
            for (int j = 0; j < parent.requiredPlugins().size(); j++) {
                IConfigPlugin child = (IConfigPlugin) parent.requiredPlugins().get(j);
                tempMap.add(parent, child);
                parentList.add(child, parent);
            }
        }
        boolean changes;
        do {
            changes = false;
            int i = 0;
            while (i < temp.size()) {
                IConfigPlugin plugin = (IConfigPlugin) temp.get(i);
                boolean remove1 = removeEmptyLists(plugin, parentList, tempMap);
                boolean remove2 = removeEmptyLists(plugin, tempMap, parentList);
                if (remove1 || remove2) {
                    temp.remove(i);
                    changes = true;
                } else
                    i++;
            }
        }
        while (changes);
        return ((tempMap.size() > 0) || (parentList.size() > 0));
    }

    private boolean removeEmptyLists(IConfigPlugin plugin, ListMap list1, ListMap list2) {
        boolean remove = false;
        List list = list2.get(plugin);
        if (list.size() == 0) {
            remove = true;
            list = list1.get(plugin);
            for (int j = 0; j < list.size(); j++) {
                IConfigPlugin other = (IConfigPlugin) list.get(j);
                list2.remove(other, plugin);
            }
            list1.removeList(plugin);
        }
        return remove;
    }

    private String getSortedMessage(SimpleList list, int last) {
        if (sortList == null)
            sortList = new ArrayList();
        else
            sortList.clear();
        for (int i = list.size() - 1; i >= last; i--)
            sortList.add(((IConfigPlugin) list.get(i)).getDirectory());
        Collections.sort(sortList);
        buffer.setLength(0);
        for (int i = 0; i < sortList.size(); i++) {
            buffer.append((String) sortList.get(i));
            buffer.append(IBootConstants.EOL);
        }
        String sortedMessage = buffer.substring(0, buffer.length());
        return sortedMessage;
    }

    private void makeDecisions() {
        for (int i = 0; i < decisionList.size(); i++)
             ((VersionDecision) decisionList.get(i)).restart();
        int idx = 0;
        boolean working;
        do {
            tempErrors.clear();
            addDependentPlugins();
            addExportedPlugins();
            working = ((decisionList.size() > 0) && tempErrors.hasErrors());
            if (working) {
                do {
                    VersionDecision decision = (VersionDecision) decisionList.get(idx);
                    if (decision.endOfChoices()) {
                        decision.restart();
                        idx++;
                        working = (idx < decisionList.size());
                    } else {
                        decision.makeNextChoice();
                        idx = 0;
                        break;
                    }
                } while (working);
            }
        }
        while (working);
        errors.addErrors(tempErrors);
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            Collections.sort(plugin.requiredPlugins().getList(), this);
        }
    }

    public int compare(Object o1, Object o2) {
        return ((IConfigPlugin) o1).getDirectory().compareTo(((IConfigPlugin) o2).getDirectory());
    }

    private void addExportedPlugins() {
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            int idx = 0;
            while (idx < plugin.requiredPlugins().size()) {
                IConfigPlugin required = (IConfigPlugin) plugin.requiredPlugins().get(idx);
                for (int j = 0; j < required.getPluginDependencies().size(); j++) {
                    PluginDependency dependency =
                        (PluginDependency) required.getPluginDependencies().get(j);
                    if (dependency.export()) {
                        IConfigPlugin requirement = getPluginSatisfyingDependency(dependency);
                        IConfigPlugin current =
                            (IConfigPlugin) plugin.requiredPlugins().get(
                                dependency.getPluginID().getID());
                        if (current == null)
                            plugin.requiredPlugins().add(
                                dependency.getPluginID().getID(),
                                requirement);
                        else if (current != requirement) {
                            tempErrors.addError(
                                Error.MULTI_VERSION_DEPENDENCY,
                                null,
                                requirement,
                                current,
                                plugin);
                        }
                    }
                }
                idx++;
            }
        }
    }

    private void addDependentPlugins() {
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            plugin.requiredPlugins().clear();
            for (int j = 0; j < plugin.getPluginDependencies().size(); j++) {
                PluginDependency dependency =
                    (PluginDependency) plugin.getPluginDependencies().get(j);
                IConfigPlugin result = getPluginSatisfyingDependency(dependency);
                plugin.requiredPlugins().add(dependency.getPluginID().getID(), result);
            }
        }
    }

    private IConfigPlugin getPluginSatisfyingDependency(PluginDependency dependency) {
        if (dependency.getPlugins().size() == 0)
            return null;
        if (dependency.getPlugins().size() == 1)
            return (IConfigPlugin) dependency.getPlugins().get(0);
        VersionDecision pending = (VersionDecision) dependency.getData();
        if (pending != null)
            return pending.getChoice();
        return null;
    }

    private void updatePluginDependencies() {
        decisionList.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            for (int j = 0; j < plugin.getPluginDependencies().size(); j++) {
                PluginDependency dependency =
                    (PluginDependency) plugin.getPluginDependencies().get(j);
                dependency.getPlugins().clear();
                dependency.setData(null);
                findDependentPlugins(dependency);
                if (dependency.getPlugins().size() == 0) {
                    logDependencyError(plugin, dependency);
                }
            }
        }
    }

    private void findDependentPlugins(PluginDependency dependency) {
        List list = plugins.get(dependency.getPluginID().getID());
        for (int i1 = 0; i1 < list.size(); i1++) {
            IConfigPlugin plugin = (IConfigPlugin) list.get(i1);
            if (dependency.satisifiedBy(plugin.getPluginID())) {
                dependency.addPlugin(plugin);
                if (dependency.getPlugins().size() == 2) {
                    VersionDecision decision = new VersionDecision(plugin, dependency);
                    dependency.setData(decision);
                    decisionList.add(decision);
                }
            }
        }
    }

    private void logDependencyError(IConfigPlugin plugin, PluginDependency dependency) {
        String id = dependency.getPluginID().toString();
        if (manager.satisfiesDependency(dependency)) {
            Error error = errors.addError(Error.UNRESOLVED_DEPENDENT_PLUGIN, id, plugin);
            List lst = manager.getPlugins().get(dependency.getPluginID().getID());
            for (int k = 0; k < lst.size(); k++) {
                IConfigPlugin plgn = (IConfigPlugin) lst.get(k);
                if (dependency.satisifiedBy(plgn.getPluginID()))
                    error.addFound((IConfigPlugin) lst.get(k));
            }
        } else
            errors.addError(Error.UNAVAILABLE_DEPENDENT_PLUGIN, id, plugin);
    }

    private void validateClasses() {
        tempMap.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            for (int j = 0; j < plugin.getClasses().size(); j++) {
                tempMap.add(plugin.getClasses().get(j), plugin);
            }
        }
        List list = null;
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            KeyedList dependencies = plugin.requiredPlugins();
            validateClassList(plugin.getClassDependencies(), dependencies, plugin, list, false);
            validateClassList(
                plugin.getDeclaredClassDependencies(),
                dependencies,
                plugin,
                list,
                true);
        }
    }

    private void validateClassList(
        SimpleList classes,
        KeyedList dependencies,
        IConfigPlugin plugin,
        List list,
        boolean declaredClass) {
        for (int j = 0; j < classes.size(); j++) {
            String cls = (String) classes.get(j);
            boolean found = plugin.getClasses().contains(cls);
            if (!found)
                found = findClass(cls, plugin, dependencies, declaredClass);
            if (!found)
                classNotFoundError(list, plugin, cls, declaredClass);
        }
    }

    private void classNotFoundError(
        List list,
        IConfigPlugin plugin,
        String cls,
        boolean declaredClass) {
        List tempList = tempMap.get(cls);
        Error.ErrorType errorCode;
        if (tempList.size() > 0) {
            if (declaredClass)
                errorCode = Error.UNREACHABLE_DECLARED_CLASS;
            else
                errorCode = Error.UNREACHABLE_DEPENDENT_CLASS;
            Error error = errors.addError(errorCode, cls, plugin);
            for (int k = 0; k < tempList.size(); k++)
                error.addFound((IConfigPlugin) tempList.get(k));

        } else {
            if (manager.getClasses().contains(cls)) {
                if (declaredClass)
                    errorCode = Error.UNRESOLVED_DECLARED_CLASS;
                else
                    errorCode = Error.UNRESOLVED_DEPENDENT_CLASS;
                Error error = errors.addError(errorCode, cls, plugin);
                if (list == null)
                    list = new ArrayList();
                else
                    list.clear();
                manager.getClassPlugins(list, cls);
                for (int k = 0; k < list.size(); k++)
                    error.addFound((IConfigPlugin) list.get(k));
            } else {
                boolean doIt = true;
                if (declaredClass)
                    errorCode = Error.UNAVAILABLE_DECLARED_CLASS;
                else {
                    errorCode = Error.UNAVAILABLE_DEPENDENT_CLASS;
                    doIt = (plugin.getIgnoreCode() != IGNORE_UNAVAILABLE);
                }
                if (doIt)
                    errors.addError(errorCode, cls, plugin);
            }
        }
    }

    private boolean findClass(
        String cls,
        IConfigPlugin plugin,
        KeyedList dependencies,
        boolean declaredClass) {
        boolean found = false;
        for (int k = 0; k < dependencies.size(); k++) {
            IConfigPlugin plugin2 = (IConfigPlugin) dependencies.get(k);
            if (plugin2 != null) {
                found = plugin2.getClasses().contains(cls);
                if (found) {
                    validateClassPublication(cls, plugin, plugin2, declaredClass);
                    break;
                }
            }
        }
        return found;
    }

    private void validateClassPublication(
        String cls,
        IConfigPlugin plugin,
        IConfigPlugin plugin2,
        boolean declaredClass) {
        boolean published = plugin2.publishAllClasses();
        if (!published)
            published = plugin.getModule().equals(plugin2.getModule());
        if (!published)
            published = plugin2.getPublished().contains(cls);
        if (!published) {
            Error.ErrorType errorCode;
            if (declaredClass)
                errorCode = Error.UNPUBLISHED_DECLARED_CLASS;
            else
                errorCode = Error.UNPUBLISHED_DEPENDENT_CLASS;
            errors.addError(errorCode, cls, plugin, null, plugin2);
        }
    }

    private void validateExtensions() {
        tempMap.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            for (int j = 0; j < plugin.getExtensionPoints().size(); j++) {
                tempMap.add(plugin.getExtensionPoints().get(j), plugin);
            }
        }
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            for (int j = 0; j < plugin.getExtensions().size(); j++) {
                String key = (String) plugin.getExtensions().get(j);
                boolean missing = !plugin.getExtensionPoints().contains(key);
                if (missing) {
                    for (int k = 0; k < plugin.requiredPlugins().size(); k++) {
                        IConfigPlugin child = (IConfigPlugin) plugin.requiredPlugins().get(k);
                        if (child.getExtensionPoints().contains(key)) {
                            missing = false;
                            break;
                        }
                    }
                    if (missing) {
                        List list = tempMap.get(key);
                        if (list.size() > 0) {
                            Error error =
                                errors.addError(Error.UNREACHABLE_EXTENSION_POINT, key, plugin);
                            for (int k = 0; k < list.size(); k++)
                                error.addFound((IConfigPlugin) list.get(k));

                        } else {
                            list = manager.getExtensionPointPlugins(key);
                            if (list.size() == 0)
                                errors.addError(Error.UNAVAILABLE_EXTENSION_POINT, key, plugin);
                            else {
                                Error error =
                                    errors.addError(Error.UNRESOLVED_EXTENSION_POINT, key, plugin);
                                for (int k = 0; k < list.size(); k++)
                                    error.addFound((IConfigPlugin) list.get(k));
                            }
                        }
                    }
                }
            }
        }
    }

    private void validateLibraries() {
        List keys = plugins.getKeys();
        for (int i = 0; i < keys.size(); i++) {
            String key = (String) keys.get(i);
            List list = plugins.get(key);
            if (list.size() > 1) {
                boolean bad = false;
                for (int j = 0; j < list.size(); j++) {
                    IConfigPlugin plugin = (IConfigPlugin) list.get(j);
                    if ((plugin.getExtensionPoints().size() > 0)
                        || (plugin.getExtensions().size() > 0)) {
                        bad = true;
                        break;
                    }
                }
                if (bad) {
                    Error error = errors.addError(Error.MULTI_VERSION_HAS_EXTENSIONS, key);
                    for (int j = 0; j < list.size(); j++) {
                        error.addPlugin((IConfigPlugin) list.get(j));
                    }
                }
            }
        }
    }

    /**
     * The list of IConfigPlugins in the configuration.
     * @return List
     */
    public ListMap getPlugins() {
        return plugins;
    }

    /**
     * The error manager used to represent configuration errors
     * @return ErrorManager
     */
    public ErrorManager errors() {
        return errors;
    }

    /**
     * Clears the configuration of all plugins.
     */
    public void clear() {
        plugins.clear();
        errors.clear();
        validate();
    }

    /**
     * Saves a valid configuration to the specified file.
     * @param file - the location of the xml file to create.
     * @throws InvalidConfigurationException
     * @throws IOException
     */
    public void saveXML(File file) throws InvalidConfigurationException, IOException {
        saveXML(file, null, false, false);
    }

    /**
     * Saves a valid configuration to the specified file.  If the optional
     * settings URL points to an XML file with a settings root element, the
     * contents of this file will be copied into the first part of the
     * configuration file.
     * @param file - the location of the xml file to create.
     * @param settings - the URL of an xml file with a settings root element.
     * @param doMap - when true, a map file is created detailing dependencies 
     * @param mapClasses - when doMap is true, the map file created will show classes if
     * mapClasses is true, or packages if mapClasses is false.
     * @throws InvalidConfigurationException
     * @throws IOException
     */
    public void saveXML(File file, URL settings, boolean doMap, boolean mapClasses)
        throws InvalidConfigurationException, IOException {
        InputStream settingsData = null;
        if (settings != null)
            settingsData = settings.openConnection().getInputStream();
        OutputStream out = new BufferedOutputStream(new FileOutputStream(file));
        Writer map = null;
        if (doMap) {
            StringBuffer buf = new StringBuffer(file.getPath());
            int idx = buf.lastIndexOf(".");
            buf.setLength(idx + 1);
            buf.append("map");
            if (mapClasses)
                buf.append("c");
            else
                buf.append("p");
            File mapFile = new File(buf.toString());
            map = new BufferedWriter(new FileWriter(mapFile));
        }
        buildXML(out, settingsData, map, mapClasses);
        out.close();
    }

    /**
     * Saves a valid configuration to the specified file.  If the optional
     * settings input stream contains XML data with a settings root element, the
     * contents of this data will be copied into the first part of the
     * configuration file.
     * @param data - the output stream to contain the built xml.
     * @param settings - optional settings xml
     * @param mapData - the output stream to send the map information to - can be null
     * @param mapClasses - if mapData is not null, this determines if classes or packages are
     * mapped in the map file.
     * @throws InvalidConfigurationException
     */
    public void buildXML(
        OutputStream data,
        InputStream settings,
        Writer mapData,
        boolean mapClasses)
        throws InvalidConfigurationException {
        validate();
        try {
            if (errors.hasErrors())
                throw new InvalidConfigurationException();
            if (data == null)
                return;
            ITreeData settingsXML = getSettings(settings);
            ByteArrayOutputStream output = new ByteArrayOutputStream();
            buildXML(output, settingsXML, 0, null, false);
            ByteArrayInputStream input = new ByteArrayInputStream(output.toByteArray());
            crc.calcConfigFile(input);
            long crcValue = crc.getValue();

            buildXML(data, settingsXML, crcValue, mapData, mapClasses);
        } finally {
            if (data != null) {
                try {
                    data.close();
                } catch (IOException e) {
                }
            }
            if (settings != null) {
                try {
                    settings.close();
                } catch (IOException e) {
                }
            }
        }
    }

    private ITreeData getSettings(InputStream settings) {
        ILoadableTreeData settingsXML = null;
        if (settings != null) {
            try {
                settingsXML = TreeDataFactory.newSAXTreeData();
                settingsXML.ignoreWhitespace(true);
                settingsXML.load(settings);
            } catch (TreeDataException e) {
            }
        }
        return settingsXML;
    }

    private void buildXML(
        OutputStream data,
        ITreeData settings,
        long configCRC,
        Writer mapData,
        boolean mapClasses)
        throws InvalidConfigurationException {
        if (data != null) {
            crc.reset();
            ISavableTreeData doc = getNewDocument();
            generateXML(doc, settings, configCRC);
            outputDocument(data, doc, mapData, mapClasses);
        }
    }

    private void generateXML(ISavableTreeData doc, ITreeData settings, long configCRC) {
        int root = doc.createRoot(CONFIGURATION_ELEMENT);
        doc.setAttribute(root, CRC_ATTRIBUTE, Long.toString(configCRC));
        addStartupInformation(doc, root);
        if (settings != null)
            doc.createChildCopy(settings, settings.getRootID(), root);
        for (int i = 0; i < plugins.size(); i++)
            addPluginToXML((IConfigPlugin) plugins.get(i), doc, root);
    }

    private void addStartupInformation(ISavableTreeData doc, int root) {
        startupPlugin = null;
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            if (plugin.getStartupClassName() != null) {
                startupPlugin = plugin;
                int startup = doc.findChildID(root, STARTUP_ELEMENT);
                if (startup < 0)
                    startup = doc.createChild(root, STARTUP_ELEMENT);
                doc.setAttribute(startup, IPluginConstants.PLUGIN_ELEMENT, plugin.getDirectory());
                doc.setAttribute(startup, RUN_ATTRIBUTE, BOOT_JAR);
                if (!plugin.getDirectory().equals(plugin.getSourceDirectory()))
                    doc.setAttribute(startup, SOURCE_ATTRIBUTE, plugin.getSourceDirectory());
                int files = doc.createChild(startup, FILES_ELEMENT);
                for (int j = 0; j < plugin.getFileCount(); j++) {
                    int file = doc.createChild(files, FILE_ELEMENT);
                    doc.setValue(file, getFileName(plugin.getFile(j)));
                }
                break;
            }
        }
    }

    private void addPluginToXML(IConfigPlugin plugin, ITreeData doc, int root) {
        int pluginElement = doc.createChild(root, IPluginConstants.PLUGIN_ELEMENT);
        if (plugin.getName() != null)
            doc.setAttribute(pluginElement, IPluginConstants.NAME_ATTRIBUTE, plugin.getName());
        doc.setAttribute(pluginElement, DIRECTORY_ATTRIBUTE, plugin.getDirectory());
        if ((plugin.getModule().length() > 0) && (!plugin.getModule().equals(plugin.getID())))
            doc.setAttribute(pluginElement, IPluginConstants.MODULE_ATTRIBUTE, plugin.getModule());
        if (!plugin.getDirectory().equals(plugin.getSourceDirectory()))
            doc.setAttribute(pluginElement, SOURCE_ATTRIBUTE, plugin.getSourceDirectory());
        addApplicationNode(plugin, doc, pluginElement);
        boolean notStartup = (plugin != startupPlugin);
        for (int i = 0; i < plugin.getFileCount(); i++) {
            addFileNode(doc, pluginElement, plugin.getFile(i), notStartup);
        }
        for (int i = 0; i < plugin.requiredPlugins().size(); i++) {
            IConfigPlugin dependentPlugin = (IConfigPlugin) plugin.requiredPlugins().get(i);
            int dependency = doc.createChild(pluginElement, IPluginConstants.PREREQUISITE_ELEMENT);
            doc.setAttribute(
                dependency,
                IPluginConstants.PLUGIN_ELEMENT,
                dependentPlugin.getDirectory());
        }
        addPublishedNodes(plugin, doc, pluginElement);
    }

    private void addFileNode(
        ITreeData doc,
        int plugin,
        ConfigFileEntry entry,
        boolean notStartup) {
        String filename = getFileName(entry);
        int file = doc.createChild(plugin, FILE_ELEMENT);
        crc.calcCRC(entry.getURL());
        doc.setAttribute(file, FILENAME_ATTRIBUTE, filename);
        doc.setAttribute(file, CRC_ATTRIBUTE, Long.toString(crc.getValue()));
        doc.setAttribute(file, SIZE_ATTRIBUTE, getURLLength(entry.getURL()));
        if (entry.getPackages() != null) {
            for (int i = 0; i < entry.getPackages().size(); i++) {
                String pkgName = (String) entry.getPackages().get(i);
                if (notStartup
                    || ((pkgName.indexOf(BOOT_PACKAGE) < 0)
                        && (!pkgName.equals(BOOT_PACKAGE_NAME)))) {
                    int pkg = doc.createChild(file, PACKAGE_ELEMENT);
                    doc.setValue(pkg, pkgName);
                }
            }
        }
    }

    private String getFileName(ConfigFileEntry entry) {
        String filename = entry.getURL().getFile();
        int idx = filename.lastIndexOf("/");
        if (idx >= 0)
            filename = filename.substring(idx + 1);
        return filename;
    }

    private String getURLLength(URL url) {
        int len = 0;
        try {
            len = url.openConnection().getContentLength();
        } catch (IOException e) {
        }
        if (len <= 0) {
            try {
                InputStream in;
                in = url.openConnection().getInputStream();
                try {
                    len = in.available();
                } finally {
                    in.close();
                }
            } catch (IOException e) {
            }
        }
        if (len > 0)
            return Integer.toString(len);
        else
            return "0";
    }

    private void addPublishedNodes(IConfigPlugin plugin, ITreeData doc, int pluginElement) {
        if (plugin.publishAllClasses() || (plugin.getPublished().size() > 0)) {
            if (plugin.publishAllClasses()) {
                addPublication(doc, pluginElement, WILDCARD);
            } else {
                for (int i = 0; i < plugin.getPublished().size(); i++) {
                    addPublication(doc, pluginElement, (String) plugin.getPublished().get(i));
                }
            }
        }
    }

    private void addPublication(ITreeData doc, int plugin, String value) {
        int publish = doc.createChild(plugin, IPluginConstants.PUBLISH_ELEMENT);
        doc.setAttribute(publish, IPluginConstants.CLASS_ATTRIBUTE, value);
    }

    private void addApplicationNode(IConfigPlugin plugin, ITreeData doc, int pluginElement) {
        if (plugin.getApplicationClassName() != null) {
            int app = doc.createChild(pluginElement, APPLICATION_ELEMENT);
            doc.setAttribute(
                app,
                IPluginConstants.CLASS_ATTRIBUTE,
                plugin.getApplicationClassName());
        }
    }

    private ISavableTreeData getNewDocument() throws InvalidConfigurationException {
        try {
            return TreeDataFactory.newSimpleXMLTreeData();
        } catch (TreeDataException e) {
            throw new InvalidConfigurationException(e);
        }
    }

    private void outputDocument(
        OutputStream data,
        ISavableTreeData doc,
        Writer mapData,
        boolean mapClasses)
        throws InvalidConfigurationException {
        try {
            doc.save(new OutputStreamWriter(data));
        } catch (IOException e) {
            throw new InvalidConfigurationException(e);
        }
        if (mapData != null) {
            try {
                mapData.write("***** Plugin / ");
                if (mapClasses)
                    mapData.write("Class");
                else
                    mapData.write("Package");
                mapData.write(" Accessability Map  *****");
                mapData.write(IBootConstants.EOL);
                mapData.write(IBootConstants.EOL);
                SimpleList classes = new SimpleList();
                for (int i = 0; i < plugins.size(); i++) {
                    IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
                    mapData.write(IBootConstants.EOL);
                    mapData.write("Plugin: ");
                    mapData.write(plugin.toString());
                    mapData.write(IBootConstants.EOL);
                    classes.clear();
                    updateClassMap(classes, plugin, plugin.getModule(), mapClasses);
                    for (int j = 0; j < plugin.requiredPlugins().size(); j++) {
                        IConfigPlugin child = (IConfigPlugin) plugin.requiredPlugins().get(j);
                        updateClassMap(classes, child, plugin.getModule(), mapClasses);
                    }
                    Collections.sort(classes.getList());
                    for (int j = 0; j < classes.size(); j++) {
                        mapData.write("  ");
                        mapData.write((String) classes.get(j));
                        mapData.write(IBootConstants.EOL);
                    }
                }
            } catch (IOException e) {
            } finally {
                try {
                    mapData.close();
                } catch (IOException e) {
                }
            }
        }
    }

    private void updateClassMap(
        SimpleList classes,
        IConfigPlugin plugin,
        String module,
        boolean mapClasses) {
        boolean all = ((plugin.publishAllClasses()) || (plugin.getModule().equals(module)));
        for (int i = 0; i < plugin.getClasses().size(); i++) {
            String cls = (String) plugin.getClasses().get(i);
            if (all || (plugin.getPublished().contains(cls))) {
                if (mapClasses)
                    classes.add(cls);
                else {
                    int idx = cls.lastIndexOf('.');
                    if (idx >= 0)
                        classes.add(cls.substring(0, idx));
                    else
                        classes.add("<DEFAULT PACKAGE>");
                }
            }
        }
    }

    private ErrorManager errors;
    private ErrorManager tempErrors;
    private ListMap plugins;
    private IPluginManager manager;
    private ChecksumCalculator crc;
    private List decisionList;
    private List sortList;
    private StringBuffer buffer;
    private ListMap parentList;
    private ListMap tempMap;
    private List temp;
    private SimpleList validatedChains;
    private SimpleList tempList;
    private IConfigPlugin startupPlugin;

    private static class VersionDecision {
        IConfigPlugin plugin;
        PluginDependency dependency;
        IConfigPlugin choice;
        SimpleList remainingChoices;
        protected VersionDecision(IConfigPlugin plugin, PluginDependency dependency) {
            super();
            this.plugin = plugin;
            this.dependency = dependency;
            remainingChoices = new SimpleList();
        }

        protected IConfigPlugin getChoice() {
            return choice;
        }

        protected boolean endOfChoices() {
            return (remainingChoices.size() == 0);
        }

        protected void restart() {
            remainingChoices.clear();
            remainingChoices.addAll(dependency.getPlugins());
            makeNextChoice();
        }

        protected IConfigPlugin makeNextChoice() {
            choice = (IConfigPlugin) remainingChoices.get(0);
            for (int i = 1; i < remainingChoices.size(); i++) {
                IConfigPlugin possible = (IConfigPlugin) remainingChoices.get(i);
                if (possible.getVersion().isGreaterThan(choice.getVersion()))
                    choice = possible;
            }
            remainingChoices.remove(choice);
            return choice;
        }
    }
}