package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.util.Collections;
import java.util.Comparator;

/**
 * This is the base configuration error class
 */
public abstract class Error {

    public static final ErrorType MISSING_STARTUP_PLUGIN = new MissingStartupPlugin();
    public static final ErrorType MISSING_APPLICATION_PLUGIN = new MissingApplicationPlugin();
    public static final ErrorType MULTIPLE_STARTUP_PLUGINS = new MultipleStartupPlugins();
    public static final ErrorType MULTIPLE_APPLICATION_PLUGINS = new MultipleApplicationPlugins();
    public static final ErrorType STARTUP_PLUGIN_MISSING_BOOT_JAR =
        new StartupPluginMissingBootJar();

    public static final ErrorType UNAVAILABLE_DEPENDENT_PLUGIN = new UnavailableDependentPlugin();
    public static final ErrorType UNRESOLVED_DEPENDENT_PLUGIN = new UnresolvedDependentPlugin();

    public static final ErrorType UNAVAILABLE_EXTENSION_POINT = new UnavailableExtensionPoint();
    public static final ErrorType UNRESOLVED_EXTENSION_POINT = new UnresolvedExtensionPoint();
    public static final ErrorType UNREACHABLE_EXTENSION_POINT = new UnreachableExtensionPoint();

    public static final ErrorType UNAVAILABLE_DEPENDENT_CLASS = new UnavailableDependentClass();
    public static final ErrorType UNRESOLVED_DEPENDENT_CLASS = new UnresolvedDependentClass();
    public static final ErrorType UNREACHABLE_DEPENDENT_CLASS = new UnreachableDependentClass();
    public static final ErrorType UNPUBLISHED_DEPENDENT_CLASS = new UnpublishedDependentClass();

    public static final ErrorType UNAVAILABLE_DECLARED_CLASS = new UnavailableDeclaredClass();
    public static final ErrorType UNRESOLVED_DECLARED_CLASS = new UnresolvedDeclaredClass();
    public static final ErrorType UNREACHABLE_DECLARED_CLASS = new UnreachableDeclaredClass();
    public static final ErrorType UNPUBLISHED_DECLARED_CLASS = new UnpublishedDeclaredClass();

    public static final ErrorType CYCLIC_DEPENDENCY = new CyclicDependency();
    public static final ErrorType MULTI_VERSION_DEPENDENCY = new MultiVersionDependency();
    public static final ErrorType MULTI_VERSION_HAS_EXTENSIONS = new MultiVersionHasExtensions();

    /**
     * Constructor for Error 
     */
    protected Error(ErrorType errorType, boolean pluginList, boolean foundList) {
        super();
        this.errorType = errorType;
        if (pluginList)
            this.pluginList = new SimpleList();
        if (foundList)
            this.foundList = new SimpleList();
    }

    private SimpleList foundList;
    private SimpleList pluginList;
    private ErrorType errorType;
    private String text;

    /**
     * Resets the text varaible 
     */
    protected void textChanged() {
        text = null;
    }

    /**
     * Returns the error type 
     * @return ErrorType
     */
    public ErrorType getErrorType() {
        return errorType;
    }

    /**
     * Returns the bit mask error code for the error
     * @return long
     */
    protected long getErrorCode() {
        return errorType.getErrorCode();
    }

    /**
     * Returns the list of plugins in the Found list, if one exists  - this may be null for some
     * subclasses
     * @return SimpleList
     */
    public SimpleList getFoundList() {
        return foundList;
    }

    /**
     * Returns the list of plugins in the Plugin list, if one exists  - this may be null for some
     * subclasses
     * @return SimpleList
     */
    public SimpleList getPluginList() {
        return pluginList;
    }

    /**
     * Adds a plugin to the plugin list, if one exists
     * @param plugin
     */
    public void addPlugin(IConfigPlugin plugin) {
        if (plugin != null) {
            pluginList.add(plugin);
            textChanged();
        }
    }

    /**
     * Adds a plugin to the found list, if one exists
     * @param plugin
     */
    public void addFound(IConfigPlugin plugin) {
        if (plugin != null) {
            foundList.add(plugin);
            textChanged();
        }
    }

    /**
     * Gets the error message
     * @return String
     */
    public abstract String getMessage();
    
    /**
     * Builds the toString value
     */
    protected abstract String generateStringValue();

    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        if (text == null)
            text = generateStringValue();
        return text;
    }

    /**
     * This is the base error type class.  All descendent classes are singlgeton classes that act as
     * error class factories.
     */
    public static abstract class ErrorType {

        private static int errorCount = 0;

        /**
         * Constructor for ErrorType 
         */
        protected ErrorType() {
            super();
            errorNumber = errorCount;
            errorCount++;
            errorCode = 1 << errorNumber;
            if (errorCode < 0)
                throw new IllegalStateException("Too many error codes");
            errorClassName = MUMUtils.getBaseClassName(this.getClass().getName());
        }

        /**
         * Returns the unique error number for this error type 
         * @return int
         */
        protected int getErrorNumber() {
            return errorNumber;
        }

        /**
         * Creates an error class of this error type
         * @return Error
         */
        protected abstract Error createError();

        private long errorCode;
        private int errorNumber;

        /**
         * Returns the bit mask error code for this error type
         * @return long
         */
        protected long getErrorCode() {
            return errorCode;
        }

        /**
         * Indicates that errors of this error type with the same error message are combined, with 
         * the expectation that additional errors add plugins to the found or plugin lists.  When a
         * descendent class sets this to false, each error, even those with identicle error messages,
         * create new error classes.  
         * @return boolean
         */
        protected boolean shareMessage() {
            return true;
        }

        protected String errorClassName;
    
        /**
         * Return the base class name (without package) of the error class 
         */
        public String getErrorClassName() {
            return errorClassName;
        }

        /**
         * Use this only for Unit Testing
         */
        protected static void startErrorCountTest() {
            oldCount = errorCount;
            errorCount = 0;
        }

        /**
         * Use this only for Unit Testing
         */
        protected static void stopErrorCountTest() {
            errorCount = oldCount;
        }

        private static int oldCount;
    }

    /**
     * InternalError
     */
    private static abstract class InternalError
        extends Error
        implements IConfigConstants, Comparator {

        protected InternalError(ErrorType errorType, boolean pluginList, boolean foundList) {
            super(errorType, pluginList, foundList);
        }

        protected abstract String getErrorDescription();

        public String generateStringValue() {
            StringBuffer buffer = new StringBuffer();
            buffer.append("Error: ");
            buffer.append(getErrorType().getErrorClassName());
            buffer.append(IBootConstants.EOL);
            buffer.append("Cause: ");
            buffer.append(getErrorDescription());
            generateMessageText(buffer);
            generateOtherText(buffer);
            buffer.append(IBootConstants.EOL);
            return buffer.toString();
        }

        protected void generateMessageText(StringBuffer buffer) {
            String msg = getMessage();
            if (msg != null) {
                buffer.append(IBootConstants.EOL);
                buffer.append("  ");
                buffer.append(msg);
            }
        }

        protected void generateOtherText(StringBuffer buffer) {
            generatePluginText(buffer, "");
        }

        protected void generatePluginText(StringBuffer buffer, String text) {
            generateListText(buffer, super.getPluginList(), text, true);
        }

        protected void generateUnsortedPluginText(StringBuffer buffer, String text) {
            generateListText(buffer, super.getPluginList(), text, false);
        }

        protected void generateFoundText(StringBuffer buffer, String text) {
            generateListText(buffer, super.getFoundList(), text, true);
        }

        private void generateListText(
            StringBuffer buffer,
            SimpleList list,
            String text,
            boolean sort) {
            if ((list != null) && (list.size() > 0)) {
                if (sort && (list.size() > 1)) {
                    Collections.sort(list.getList(), this);
                }
                if (text != null) {
                    buffer.append(IBootConstants.EOL);
                    buffer.append("  Plugin");
                    if (list.size() > 1)
                        buffer.append("s");
                    if (text.length() > 0) {
                        buffer.append(" ");
                        buffer.append(text);
                    }
                    buffer.append(':');
                }
                for (int i = 0; i < list.size(); i++) {
                    buffer.append(IBootConstants.EOL);
                    buffer.append("    ");
                    buffer.append(((IConfigPlugin) list.get(i)).toString());
                }
            }
        }

        public int compare(Object o1, Object o2) {
            return o1.toString().compareTo(o2.toString());
        }
    }

    /**
     * MessageError
     */
	static abstract class MessageError extends InternalError {

        private String message;

        protected MessageError(ErrorType errorType, boolean pluginList, boolean foundList) {
            super(errorType, pluginList, foundList);
        }

        public String getMessage() {
            return message;
        }

        protected void setMessage(String string) {
            if (message != string) {
                message = string;
                textChanged();
            }
        }
    }

    /**
     * Multiple Interface Error
     */
    private abstract static class MultipleInterfaceError extends InternalError {

        protected MultipleInterfaceError(ErrorType errorType) {
            super(errorType, true, false);
        }

        protected void generateOtherText(StringBuffer buffer) {
            buffer.append(IBootConstants.EOL);
            buffer.append(PLUGIN_IMPLEMENTING_TXT);
            buffer.append(getInterfaceName());
            buffer.append(INTERFACE_TXT);
            buffer.append(":");
            generatePluginText(buffer, null);
        }

        protected abstract String getInterfaceName();
    }

    /**
     * Dependency Error
     */
    private abstract static class DependencyError extends MessageError {

        protected DependencyError(ErrorType errorType, boolean foundList) {
            super(errorType, true, foundList);
        }

        protected void generateOtherText(StringBuffer buffer) {
            generatePluginText(buffer, PLUGIN_REQUIRING_DEPENDENCY_TXT);
            generateFoundText(buffer, PLUGIN_SATISFYING_DEPENDENCY_TXT);
        }
    }

    /**
     * Plugin Dependency Error
     */
    private abstract static class PluginDependencyError extends DependencyError {

        protected PluginDependencyError(ErrorType errorType, boolean foundList) {
            super(errorType, foundList);
        }

        protected void setMessage(String string) {
            super.setMessage(PLUGIN_DEPENDENCY_TXT + string);
        }
    }

    /**
     * Class Dependency Error
     */
    private abstract static class ClassDependencyError extends DependencyError {

        protected ClassDependencyError(ErrorType errorType, boolean foundList) {
            super(errorType, foundList);
        }

        protected void setMessage(String string) {
            super.setMessage(CLASS_DEPENDENCY_TXT + string);
        }

        protected void rootSetMessage(String string) {
            super.setMessage(string);
        }
    }
    /**
     * Extension Point Error
     */
    private abstract static class ExtensionPointError extends MessageError {

        protected ExtensionPointError(ErrorType errorType, boolean foundList) {
            super(errorType, true, foundList);
        }

        protected void setMessage(String string) {
            super.setMessage(EXENSION_POINT_TXT + string);
        }

        protected void generateOtherText(StringBuffer buffer) {
            generatePluginText(buffer, PLUGIN_EXTENDING_EXTENSION_POINT_TXT);
            generateFoundText(buffer, PLUGIN_IMPLEMENTING_EXTENSION_POINT_TXT);
        }
    }

    /**
     * MissingStartupPlugin error classes
     */
	private static class MissingStartupPlugin extends ErrorType {

        protected Error createError() {
            return new MissingStartupPluginError(this);
        }
    }

    private static class MissingStartupPluginError extends InternalError {

        protected MissingStartupPluginError(ErrorType errorType) {
            super(errorType, false, false);
        }

        public String getMessage() {
            return MISSING_STARTUP_PLUGIN_TXT;
        }

        protected String getErrorDescription() {
            return MISSING_STARTUP_PLUGIN_ERROR_DESCRIPTION;
        }
    }

    /**
     * MissingApplicationPlugin error classes
     */
	private static class MissingApplicationPlugin extends ErrorType {

        protected Error createError() {
            return new MissingApplicationPluginError(this);
        }

    }

    private static class MissingApplicationPluginError extends InternalError {

        protected MissingApplicationPluginError(ErrorType errorType) {
            super(errorType, false, false);
        }

        public String getMessage() {
            return MISSING_APPLICATION_PLUGINS_TXT;
        }

        protected String getErrorDescription() {
            return MISSING_APPLICATION_PLUGIN_ERROR_DESCRIPTION;
        }
    }

    /**
     * MultipleStartupPlugins error classes
     */
	private static class MultipleStartupPlugins extends ErrorType {

        protected Error createError() {
            return new MultipleStartupPluginsError(this);
        }
    }

    private static class MultipleStartupPluginsError extends MultipleInterfaceError {

        protected MultipleStartupPluginsError(ErrorType errorType) {
            super(errorType);
        }

        public String getMessage() {
            return MULTIPLE_STARTUP_PLUGINS_TXT;
        }

        protected String getErrorDescription() {
            return MULTIPLE_STARTUP_PLUGINS_ERROR_DESCRIPTION;
        }

        protected String getInterfaceName() {
            return STARTUP_INTERFACE_BASE_NAME;
        }
    }

    /**
     * MultipleApplicationPlugins error classes
     */
	private static class MultipleApplicationPlugins extends ErrorType {

        protected Error createError() {
            return new MultipleApplicationPluginsError(this);
        }

    }

    private static class MultipleApplicationPluginsError extends MultipleInterfaceError {

        protected MultipleApplicationPluginsError(ErrorType errorType) {
            super(errorType);
        }

        public String getMessage() {
            return MULTIPLE_APPLICATION_PLUGINS_TXT;
        }

        protected String getErrorDescription() {
            return MULTIPLE_APPLICATION_PLUGINS_ERROR_DESCRIPTION;
        }

        protected String getInterfaceName() {
            return APPLICATION_INTERFACE_BASE_NAME;
        }
    }

    /**
     * UnavailableDependentPlugin error classes
     */
	private static class UnavailableDependentPlugin extends ErrorType {

        protected Error createError() {
            return new UnavailableDependentPluginError(this);
        }

    }

    private static class UnavailableDependentPluginError extends PluginDependencyError {

        protected UnavailableDependentPluginError(ErrorType errorType) {
            super(errorType, false);
        }

        protected String getErrorDescription() {
            return UNAVAILABLE_DEPENDENT_PLUGIN_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnavailableDependentClass error classes
     */
	private static class UnavailableDependentClass extends ErrorType {

        protected Error createError() {
            return new UnavailableDependentClassError(this);
        }

    }

    private static class UnavailableDependentClassError extends ClassDependencyError {

        protected UnavailableDependentClassError(ErrorType errorType) {
            super(errorType, false);
        }

        protected String getErrorDescription() {
            return UNAVAILABLE_DEPENDENT_CLASS_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnresolvedDependentPlugin error classes
     */
	private static class UnresolvedDependentPlugin extends ErrorType {

        protected Error createError() {
            return new UnresolvedDependentPluginError(this);
        }

    }

    private static class UnresolvedDependentPluginError extends PluginDependencyError {

        protected UnresolvedDependentPluginError(ErrorType errorType) {
            super(errorType, true);
        }

        protected String getErrorDescription() {
            return UNRESOLVED_DEPENDENT_PLUGIN_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnresolvedDependentClass error classes
     */
	private static class UnresolvedDependentClass extends ErrorType {

        protected Error createError() {
            return new UnresolvedDependentClassError(this);
        }

    }

    private static class UnresolvedDependentClassError extends ClassDependencyError {

        protected UnresolvedDependentClassError(ErrorType errorType) {
            super(errorType, true);
        }

        protected String getErrorDescription() {
            return UNRESOLVED_DEPENDENT_CLASS_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnreachableDependentClass error classes
     */
	private static class UnreachableDependentClass extends ErrorType {

        protected Error createError() {
            return new UnreachableDependentClassError(this);
        }

    }

    private static class UnreachableDependentClassError extends ClassDependencyError {

        protected UnreachableDependentClassError(ErrorType errorType) {
            super(errorType, true);
        }

        protected String getErrorDescription() {
            return UNREACHABLE_DEPENDENT_CLASS_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnpublishedDependentClass error classes
     */
	private static class UnpublishedDependentClass extends ErrorType {

        protected Error createError() {
            return new UnpublishedDependentClassError(this);
        }
    }

    private static class UnpublishedDependentClassError extends MessageError {

        protected UnpublishedDependentClassError(ErrorType errorType) {
            super(errorType, true, true);
        }

        protected String getErrorDescription() {
            return UNPUBLISHED_DEPENDENT_CLASS_ERROR_DESCRIPTION;
        }

        protected void setMessage(String string) {
            super.setMessage(CLASS_TXT + string);
        }

        protected void rootSetMessage(String string) {
            super.setMessage(string);
        }

        protected void generateOtherText(StringBuffer buffer) {
            generatePluginText(buffer, PLUGIN_REQUIRING_DEPENDENCY_TXT);
            generateFoundText(buffer, PLUGIN_SATISFYING_UNPUBLISHED_DEPENDENCY_TXT);
        }
    }

    /**
     * UnavailableDeclaredClass error classes
     */
	private static class UnavailableDeclaredClass extends ErrorType {

        protected Error createError() {
            return new UnavailableDeclaredClassError(this);
        }
    }

    private static class UnavailableDeclaredClassError extends UnavailableDependentClassError {

        protected UnavailableDeclaredClassError(ErrorType errorType) {
            super(errorType);
        }

        protected void setMessage(String string) {
            rootSetMessage(DECLARED_CLASS_DEPENDENCY_TXT + string);
        }
    }

    /**
     * UnresolvedDeclaredClass error classes
     */
	private static class UnresolvedDeclaredClass extends ErrorType {

        protected Error createError() {
            return new UnresolvedDeclaredClassError(this);
        }
    }

    private static class UnresolvedDeclaredClassError extends UnresolvedDependentClassError {

        protected UnresolvedDeclaredClassError(ErrorType errorType) {
            super(errorType);
        }

        protected void setMessage(String string) {
            rootSetMessage(DECLARED_CLASS_DEPENDENCY_TXT + string);
        }

    }

    /**
     * UnreachableDeclaredClass error classes
     */
	private static class UnreachableDeclaredClass extends ErrorType {

        protected Error createError() {
            return new UnreachableDeclaredClassError(this);
        }
    }

    private static class UnreachableDeclaredClassError extends UnreachableDependentClassError {

        protected UnreachableDeclaredClassError(ErrorType errorType) {
            super(errorType);
        }

        protected void setMessage(String string) {
            rootSetMessage(DECLARED_CLASS_DEPENDENCY_TXT + string);
        }

    }

    /**
     * UnpublishedDeclaredClass error classes
     */
	private static class UnpublishedDeclaredClass extends ErrorType {

        protected Error createError() {
            return new UnpublishedDeclaredClassError(this);
        }
    }

    private static class UnpublishedDeclaredClassError extends UnpublishedDependentClassError {

        protected UnpublishedDeclaredClassError(ErrorType errorType) {
            super(errorType);
        }

        protected void setMessage(String string) {
            rootSetMessage(DECLARED_CLASS_DEPENDENCY_TXT + string);
        }
    }

    /**
     * StartupPluginMissingBootJar error classes
     */
	private static class StartupPluginMissingBootJar extends ErrorType {

        protected Error createError() {
            return new StartupPluginMissingBootJarError(this);
        }

    }

    private static class StartupPluginMissingBootJarError extends InternalError {

        protected StartupPluginMissingBootJarError(ErrorType errorType) {
            super(errorType, true, false);
        }

        public String getMessage() {
            return STARTUP_PLUGIN_MISSING_BOOT_TXT;
        }

        protected String getErrorDescription() {
            return STARTUP_PLUGIN_MISSING_BOOT_JAR_ERROR_DESCRIPTION;
        }
    }

    /**
     * CyclicDependency error classes
     */
	private static class CyclicDependency extends ErrorType {

        protected Error createError() {
            return new CyclicDependencyError(this);
        }
    }

    private static class CyclicDependencyError extends MessageError {

        protected CyclicDependencyError(ErrorType errorType) {
            super(errorType, true, false);
        }

        public String getMessage() {
            return null;
        }

        protected String getErrorDescription() {
            return CYCLIC_DEPENDENCY_ERROR_DESCRIPTION;
        }

        protected void generateOtherText(StringBuffer buffer) {
            generateUnsortedPluginText(buffer, null);
            if (getPluginList().size() > 0) {
                buffer.append(IBootConstants.EOL);
                buffer.append("    ");
                buffer.append(((IConfigPlugin) getPluginList().get(0)).toString());
            }
        }
    }

    /**
     * MultiVersionDependency error classes
     */
	private static class MultiVersionDependency extends ErrorType {

        protected Error createError() {
            return new MultiVersionDependencyError(this);
        }

        protected boolean shareMessage() {
            return false;
        }
    }

    private static class MultiVersionDependencyError extends InternalError {

        protected MultiVersionDependencyError(ErrorType errorType) {
            super(errorType, true, true);
        }

        public String getMessage() {
            return MULTI_VERSION_DEPENDENCY_TXT;
        }

        protected String getErrorDescription() {
            return MULTI_VERSION_DEPENDENCY_ERROR_DESCRIPTION;
        }

        protected void generateOtherText(StringBuffer buffer) {
            generateFoundText(buffer, PLUGIN_REQUIRING_DEPENDENCY_TXT);
            generatePluginText(buffer, PLUGIN_SATISFYING_DEPENDENCY_TXT);
        }
    }

    /**
     * MultiVersionDependency error classes
     */
	private static class MultiVersionHasExtensions extends ErrorType {

        protected Error createError() {
            return new MultiVersionHasExtensionsError(this);
        }

        protected boolean shareMessage() {
            return false;
        }
    }

    private static class MultiVersionHasExtensionsError extends InternalError {

        protected MultiVersionHasExtensionsError(ErrorType errorType) {
            super(errorType, true, false);
        }

        public String getMessage() {
            return null;
        }

        protected String getErrorDescription() {
            return MULTI_VERSION_HAS_EXTENSIONS_DESCRIPTION;
        }
    }

    /**
     * UnavailableExtensionPoint error classes
     */
	private static class UnavailableExtensionPoint extends ErrorType {

        protected Error createError() {
            return new UnavailableExtensionPointError(this);
        }
    }

    private static class UnavailableExtensionPointError extends ExtensionPointError {

        protected UnavailableExtensionPointError(ErrorType errorType) {
            super(errorType, false);
        }

        protected String getErrorDescription() {
            return UNAVAILABLE_EXTENSION_POINT_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnresolvedExtensionPoint error classes
     */
	private static class UnresolvedExtensionPoint extends ErrorType {

        protected Error createError() {
            return new UnresolvedExtensionPointError(this);
        }

    }

    private static class UnresolvedExtensionPointError extends ExtensionPointError {

        protected UnresolvedExtensionPointError(ErrorType errorType) {
            super(errorType, true);
        }

        protected String getErrorDescription() {
            return UNRESOLVED_EXTENSION_POINT_ERROR_DESCRIPTION;
        }
    }

    /**
     * UnreachableExtensionPoint error classes
     */
	private static class UnreachableExtensionPoint extends ErrorType {

        protected Error createError() {
            return new UnreachableExtensionPointError(this);
        }

    }

    private static class UnreachableExtensionPointError extends ExtensionPointError {

        protected UnreachableExtensionPointError(ErrorType errorType) {
            super(errorType, true);
        }

        protected String getErrorDescription() {
            return UNREACHABLE_EXTENSION_POINT_ERROR_DESCRIPTION;
        }
    }
}
