package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.List;

/**
 * This class locates all available plugins found in the gold directory.  It is
 * subsequently used by Configurations to provide a list of available plugins.
 */
public class PluginManager implements IPluginManager, IConfigConstants {

    private ListMap plugins;
    private SimpleList classes;
    private ListMap extensionPoints;
    private File goldDir;
    private boolean loaded;
    private ClassDataCache cache;

    protected PluginManager() {
        super();
        plugins = new ListMap();
        classes = new SimpleList();
		extensionPoints = new ListMap();
        loaded = false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#assignGoldDirectory(java.io.File)
     */
    public void assignGoldDirectory(File goldDir) {
        this.goldDir = goldDir;
        cache = new ClassDataCache(goldDir);
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#load()
     */
    public void load() throws PluginLocatorException {
        if (!loaded) {
            reset();
            ConfigPluginLocatorFactory.newConfigPluginLocator(
                goldDir,
                cache).createDistributionPlugins(
                plugins);
            Collections.sort(plugins.getKeys());
            loaded = true;
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#reset()
     */
    public void reset() {
        plugins.clear();
        classes.clear();
		extensionPoints.clear();
        loaded = false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#getPlugins()
     */
    public ListMap getPlugins() {
        return plugins;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#isLoaded()
     */
    public boolean isLoaded() {
        return loaded;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#loadElements()
     */
    public void loadPluginInformation() {
        classes.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            classes.addAll(plugin.getClasses());
            for (int j = 0; j < plugin.getExtensionPoints().size(); j++) {
            	extensionPoints.add(plugin.getExtensionPoints().get(j), plugin);                
            }
        }
        try {
            if (cache != null)
                cache.save();
        } catch (IOException e) {
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#supportsPlugin(gov.va.med.hds.cd.mum.PluginElement)
     */
    public boolean satisfiesDependency(PluginDependency dependency) {
        List temp = plugins.get(dependency.getPluginID().getID());
        for (int i = 0; i < temp.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) temp.get(i);
            if (dependency.satisifiedBy(plugin.getPluginID()))
                return true;
        }
        return false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#getClassPlugin(gov.va.med.hds.cd.mum.RootElement)
     */
    public void getClassPlugins(List list, String cls) {
        list.clear();
        for (int i = 0; i < plugins.size(); i++) {
            IConfigPlugin plugin = (IConfigPlugin) plugins.get(i);
            if (plugin.getClasses().contains(cls))
                list.add(plugin);
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.config.IPluginManager#getClasses()
     */
    public SimpleList getClasses() {
        return classes;
    }

    public List getExtensionPointPlugins(String extensionPoint) {
        return extensionPoints.get(extensionPoint);
    }
}
