package gov.va.med.hds.cd.mum.config;

import gov.va.med.hds.cd.mum.IPluginExceptionManager;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

/**
 * This class is used to resolve published entries in a plugin's manifest.xml
 * file.  If converts published manifest entries, which may include wildcards,
 * into a specific list of classes to be published by the plugin.
 */
public class PluginPublicationManager implements IConfigConstants {

	/**
	 * Constructor for PluginPublicationManager.
	 * @param errorManager - the error manager to call when an error occurs.
	 */
	public PluginPublicationManager(IPluginExceptionManager errorManager) {
		super();
		this.errorManager = errorManager;
		publishedPatterns = new SimpleList();
		patternsFound = new SimpleList();
		publishedClasses = new SimpleList();
	}

	/**
	 * Clears out all publication data.
	 */
	public void reset() {
		clear();
		publishAllClasses = false;
		hasClassList = false;
	}

	private void clear() {
		publishedPatterns.clear();
		patternsFound.clear();
		publishedClasses.clear();
	}

	private SimpleList publishedPatterns;
	private SimpleList patternsFound;
	private SimpleList publishedClasses;
	private boolean publishAllClasses;
	private boolean hasClassList;
	private IPluginExceptionManager errorManager;

	/**
	 * Returns true if all classes in the plugin are to be published.
	 * @return boolean
	 */
	public boolean publishAllClasses() {
		return publishAllClasses;
	}

	/**
	 * Returns true if there are any published statements in the manifest, other
	 * than the publish statement that publishes all classes.
	 * @return boolean
	 */
	public boolean hasClassList() {
		return hasClassList;
	}

	/**
	 * Adds a publish entry from the manifest, and parses the entry for wildcard
	 * information.
	 * @param publication
	 * @throws InvalidPluginException
	 */
	public void addPublication(String publication)
		throws InvalidPluginException {
		if (publication.equals(WILDCARD)) {
			publishAllClasses = true;
		} else {
			if (publication.indexOf(WILDCARD) >= 0) {
				validatePattern(publication);
				publication =
					publication.substring(0, publication.length() - 1);
				publishedPatterns.add(publication);
				hasClassList = true;
			} else {
				publishedClasses.add(publication);
				hasClassList = true;
			}
		}
	}

	/**
	 * Compares publication information with a list of actual classes, and
	 * updates the list of classes to publish accordingly.  This method will
	 * also cause a exception to be raised if publications exist that are not
	 * satisifies by the class list.
	 * @param classList - the Collection of classes to validate against.
	 * @param published - the Collection to populate with published classes
	 * @throws InvalidPluginException
	 */
	public void resolvePublications(SimpleList classList, SimpleList published)
		throws InvalidPluginException {
		try {
			if (hasClassList) {
				for (int i = 0; i < classList.size(); i++)
					updatePublications((String) classList.get(i), published);
				resolveMissingClasses();
				resolveUnusedPatterns();
			}
		} finally {
			clear();
		}
	}

	private void validatePattern(String cls) throws InvalidPluginException {
		if (cls.indexOf(WILDCARD) != cls.length() - 1)
			errorManager.badPlugin(INVALID_WILDCARD);
	}

	private void resolveUnusedPatterns() throws InvalidPluginException {
		if ((publishedPatterns != null) && (publishedPatterns.size() > 0)) {
			if ((patternsFound == null) || (patternsFound.size() < 1))
				errorManager.badPlugin(
					NO_CLASSES_FOUND
						+ (String) publishedPatterns.get(0)
						+ WILDCARD);
			for (int i = 0; i < publishedPatterns.size(); i++) {
				if (!patternsFound.contains(publishedPatterns.get(i))) {
					errorManager.badPlugin(
						NO_CLASSES_FOUND
							+ (String) publishedPatterns.get(i)
							+ WILDCARD);
				}
			}
		}
	}

	private void resolveMissingClasses() throws InvalidPluginException {
		if ((publishedClasses != null) && (publishedClasses.size() > 0)) {
			errorManager.badPlugin(
				(String) publishedClasses.get(0) + CLASS_NOT_FOUND,
				new ClassNotFoundException());
		}
	}

	private void updatePublications(String className, SimpleList published) {
		if ((publishedClasses != null)
			&& (publishedClasses.contains(className))) {
			published.add(className);
			publishedClasses.remove(className);
		} else {
			if (publishedPatterns != null) {
				for (int i = 0; i < publishedPatterns.size(); i++) {
					String prefix = (String) publishedPatterns.get(i);
					if (className.startsWith(prefix)) {
						String suffix =
							className.substring(
								prefix.length(),
								className.length());
						if (suffix.indexOf(".") < 0) {
							published.add(className);
							if (patternsFound == null)
								patternsFound = new SimpleList();
							patternsFound.add(prefix);
						}
					}
				}
			}
		}
	}
}