package gov.va.med.hds.cd.mum.config.test;

import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.PluginID;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.config.ConfigPluginFactory;
import gov.va.med.hds.cd.mum.config.ConfigPluginLocatorFactory;
import gov.va.med.hds.cd.mum.config.Configuration;
import gov.va.med.hds.cd.mum.config.Error;
import gov.va.med.hds.cd.mum.config.ErrorManager;
import gov.va.med.hds.cd.mum.config.IConfigPlugin;
import gov.va.med.hds.cd.mum.config.IConfigPluginLocator;
import gov.va.med.hds.cd.mum.config.IPluginManager;
import gov.va.med.hds.cd.mum.config.PluginLocatorException;
import gov.va.med.hds.cd.mum.config.PluginManagerFactory;
import gov.va.med.hds.cd.mum.test.IMUMTestConstants;
import gov.va.med.hds.cd.mum.utils.ChecksumCalculator;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.util.ArrayList;

import junit.framework.TestCase;

public class ConfigurationTest
    extends TestCase
    implements IConfigPluginLocator, IMUMTestConstants {

    public void testMissing() throws Exception {
        init();
        config.clear();
        config.validate();
        assertError(Error.MISSING_STARTUP_PLUGIN);
        assertError(Error.MISSING_APPLICATION_PLUGIN);
        assertEquals(2, config.errors().getErrors().size());
        getError(Error.MISSING_STARTUP_PLUGIN);
        assertNull(error.getPluginList());
        assertNull(error.getFoundList());
        getError(Error.MISSING_APPLICATION_PLUGIN);
        assertNull(error.getPluginList());
        assertNull(error.getFoundList());
        addPlugin(mjEditor1);
        addPlugin(mjTemplateFields1);
        config.validate();
        assertNoError(Error.MISSING_STARTUP_PLUGIN);
        assertNoError(Error.MISSING_APPLICATION_PLUGIN);
    }

    public void testDuplicates() throws Exception {
        init();
        config.clear();
        assertNoError(Error.MULTIPLE_STARTUP_PLUGINS);
        assertNoError(Error.MULTIPLE_APPLICATION_PLUGINS);
        reconfig(STARTUP_PACKAGES);
        assertError(Error.MULTIPLE_STARTUP_PLUGINS);
        assertNoError(Error.MULTIPLE_APPLICATION_PLUGINS);
        getError(Error.MULTIPLE_STARTUP_PLUGINS);
        assertEquals(3, error.getPluginList().size());
        assertNull(error.getFoundList());
        assertNotNull(error);
        for (int i = 0; i < STARTUP_PACKAGES.length; i++)
            assertTrue(error.getPluginList().contains(getPlugin(STARTUP_PACKAGES[i])));

        reconfig(APPLICATION_PACKAGES);
        assertNoError(Error.MULTIPLE_STARTUP_PLUGINS);
        assertError(Error.MULTIPLE_APPLICATION_PLUGINS);
        getError(Error.MULTIPLE_APPLICATION_PLUGINS);
        assertEquals(3, error.getPluginList().size());
        assertNull(error.getFoundList());
        assertNotNull(error);
        for (int i = 0; i < APPLICATION_PACKAGES.length; i++)
            assertTrue(error.getPluginList().contains(getPlugin(APPLICATION_PACKAGES[i])));
    }

    public void testUnresolvedPlugins() throws Exception {
        init();
        config.clear();
        assertNoError(Error.UNRESOLVED_DEPENDENT_PLUGIN);
        IConfigPlugin plugin = addPlugin(mjEditor4);
        config.validate();
        assertError(Error.UNRESOLVED_DEPENDENT_PLUGIN);
        getError(Error.UNRESOLVED_DEPENDENT_PLUGIN);
        assertNotNull(error);
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getPluginList().contains(plugin));
        addPlugin(mjNoteCore2);
        config.validate();
        assertError(Error.UNRESOLVED_DEPENDENT_PLUGIN);
        addPlugin(mjNoteCore1);
        addPlugin(mjRuntime);
        addPlugin(mjLib2);
        config.validate();
        assertNoError(Error.UNRESOLVED_DEPENDENT_PLUGIN);
    }

    public void testUnavailablePlugins() throws Exception {
        init();
        config.clear();
        assertNoError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        addPlugin(mjEditor4);
        config.validate();
        assertNoError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        IConfigPlugin plugin = addPlugin(mjEditor1);
        config.validate();
        assertError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        getError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        assertNotNull(error);
        assertNull(error.getFoundList());
        assertTrue(error.getPluginList().contains(plugin));
        config.clear();
        assertNoError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        plugin = addPlugin(mjEditor2);
        config.validate();
        assertError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        getError(Error.UNAVAILABLE_DEPENDENT_PLUGIN);
        assertNotNull(error);
        assertNull(error.getFoundList());
        assertTrue(error.getPluginList().contains(plugin));
    }

    public void testUnresolvedClasses() throws Exception {
        init();
        config.clear();
        assertNoError(Error.UNRESOLVED_DEPENDENT_CLASS);
        IConfigPlugin plugin = addPlugin(mjEditor4);
        IConfigPlugin plugin1 = getPlugin(mjNoteCore1);
        IConfigPlugin plugin2 = getPlugin(mjLib2);
        config.validate();
        assertError(Error.UNRESOLVED_DEPENDENT_CLASS);
        int count = 0;
        boolean ok = false;
        for (int i = 0; i < config.errors().getErrors().size(); i++) {
            error = (Error) config.errors().getErrors().get(i);
            if (error.getErrorType() == Error.UNRESOLVED_DEPENDENT_CLASS) {
                count++;
                assertTrue(error.getPluginList().contains(plugin));
                if (error.getMessage().indexOf("tiu.notes.internal") >= 0) {
                    ok = true;
                    assertTrue(error.getFoundList().contains(plugin1));
                } else if (error.getMessage().indexOf("util.three") >= 0) {
                    ok = true;
                    assertTrue(error.getFoundList().contains(plugin2));
                }
            }
        }
        if (!ok) {
            outputErrors();
            fail("didn't find expected classes");
        }
        assertEquals(2, count);
        addPlugin(mjNoteCore1);
        addPlugin(mjLib2);
        config.validate();
        assertNoError(Error.UNRESOLVED_DEPENDENT_CLASS);
    }

    public void testUnreachableClasses() throws Exception {
        init();
        assertNoError(Error.UNREACHABLE_DEPENDENT_CLASS);
        IConfigPlugin plugin1 = getPlugin(mjNoteCore1);
        IConfigPlugin plugin2 = getPlugin(mjInMedOrders);
        IConfigPlugin plugin3 = getPlugin(mjLib2);
        String cls = "unreachable.test.Class";
        plugin1.getClasses().add(cls);
        plugin2.getClassDependencies().add(cls);
        plugin3.getClassDependencies().add(cls);
        config.validate();
        assertError(Error.UNREACHABLE_DEPENDENT_CLASS);
        getError(Error.UNREACHABLE_DEPENDENT_CLASS);
        assertNotNull(error);
        assertTrue(error.getMessage().indexOf(cls) >= 0);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(plugin2));
        assertTrue(error.getPluginList().contains(plugin3));
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getFoundList().contains(plugin1));
    }

    public void testUnavailableClasses() throws Exception {
        init();
        config.clear();
        assertNoError(Error.UNAVAILABLE_DEPENDENT_CLASS);
        IConfigPlugin plugin = addPlugin(mjEditor1);
        config.validate();
        assertError(Error.UNAVAILABLE_DEPENDENT_CLASS);
        getError(Error.UNAVAILABLE_DEPENDENT_CLASS);
        assertNotNull(error);
        assertNull(error.getFoundList());
        assertTrue(error.getPluginList().contains(plugin));
    }

    public void testUnpublishedClasses() throws Exception {
        init();
        config.clear();
        assertNoError(Error.UNPUBLISHED_DEPENDENT_CLASS);
        IConfigPlugin plugin = addPlugin(mjEditor1);
        IConfigPlugin plugin1 = addPlugin(mjNoteCore1);
        //		IConfigPlugin plugin2 = addPlugin(mjLib2);
        config.validate();
        assertError(Error.UNPUBLISHED_DEPENDENT_CLASS);
        getError(Error.UNPUBLISHED_DEPENDENT_CLASS);
        assertNotNull(error);
        assertTrue(error.getPluginList().contains(plugin));
        assertTrue(error.getFoundList().contains(plugin1));
    }

    public void testCompleteConfiguration() throws Exception {
        init();
        assertEquals(21, config.getPlugins().size());
        assertHasErrors();
        reconfig(GOOD_CONFIG);
        assertNoErrors();
        reconfig(EXISTING_CONFIG);
        assertNoErrors();
    }

    public void testXMLFailure() throws Exception {
        init();
        reconfig(GOOD_CONFIG);
        IConfigPlugin plugin = getPlugin(mjRemCore);
        config.getPlugins().remove(plugin.getID(), plugin);
        try {
            config.buildXML(null, null, null, false);
            fail("Should have thrown InvalidConfigurationException");
        } catch (InvalidConfigurationException e) {
        }
        addPlugin(mjRemCore);
        config.buildXML(null, null, null, false);
    }

    public void testXMLBuild() throws Exception {
        init();
        reconfig(GOOD_CONFIG);
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        config.buildXML(data, null, null, false);
        String buffer = data.toString();
        assertEquals(goodConfigXML, buffer);
        ChecksumCalculator crc = new ChecksumCalculator();
        ByteArrayInputStream stream = new ByteArrayInputStream(data.toByteArray());
        crc.calcConfigFile(stream);
        assertEquals(crcValue, crc.getValue());
    }

    public void testSettingsXMLBuild() throws Exception {
        init();
        reconfig(GOOD_CONFIG);
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        URL settings = this.getClass().getClassLoader().getResource(root + "settings.xml");
        assertNotNull(settings);
        config.buildXML(data, settings.openConnection().getInputStream(), null, false);
        String buffer = data.toString();
        assertEquals(goodConfigXMLWithSettings, buffer);
        ChecksumCalculator crc = new ChecksumCalculator();
        ByteArrayInputStream stream = new ByteArrayInputStream(data.toByteArray());
        crc.calcConfigFile(stream);
        assertEquals(settingsCrcValue, crc.getValue());
    }

    public void testMissingStartupJar() throws Exception {
        init();
        MockConfigPlugin.setNoStartupJar(true);
        try {
            mgr.reset();
            config = new Configuration(mgr);
            reconfig(GOOD_CONFIG);
        } finally {
            MockConfigPlugin.setNoStartupJar(false);
        }
        assertError(Error.STARTUP_PLUGIN_MISSING_BOOT_JAR);
    }

    public void testRequiredDependencyLists() throws Exception {
        init();
        reconfig(GOOD_CONFIG);
        IConfigPlugin pEditor4 = getPlugin(mjEditor4);
        IConfigPlugin pNoteCore1 = getPlugin(mjNoteCore1);
        IConfigPlugin pLib1 = getPlugin(mjLib1);
        IConfigPlugin pLib2 = getPlugin(mjLib2);
        IConfigPlugin pTemplates1 = getPlugin(mjTemplates1);
        IConfigPlugin pTemplatesCore = getPlugin(mjTemplatesCore);
        IConfigPlugin pTemplateFields2 = getPlugin(mjTemplateFields2);
        IConfigPlugin pEncounterForms = getPlugin(mjEncounterForms);
        IConfigPlugin pEncounters = getPlugin(mjEncounters);
        IConfigPlugin pRemCore = getPlugin(mjRemCore);
        IConfigPlugin pRemDialogs2 = getPlugin(mjRemDialogs2);
        IConfigPlugin pInMedOrders = getPlugin(mjInMedOrders);
        IConfigPlugin pOutMedOrders = getPlugin(mjOutMedOrders);
        IConfigPlugin pMedCore = getPlugin(mjMedCore);
        IConfigPlugin pRuntime = getPlugin(mjRuntime);

        assertEquals(2, pEditor4.requiredPlugins().size());
        assertTrue(pEditor4.requiredPlugins().containsValue(pNoteCore1));
        assertTrue(pEditor4.requiredPlugins().containsValue(pLib2));

        assertEquals(1, pNoteCore1.requiredPlugins().size());
        assertTrue(pNoteCore1.requiredPlugins().containsValue(pRuntime));

        assertEquals(0, pLib1.requiredPlugins().size());

        assertEquals(0, pLib2.requiredPlugins().size());

        assertEquals(1, pTemplates1.requiredPlugins().size());
        assertTrue(pTemplates1.requiredPlugins().containsValue(pTemplatesCore));

        assertEquals(0, pTemplatesCore.requiredPlugins().size());

        assertEquals(1, pTemplateFields2.requiredPlugins().size());
        assertTrue(pTemplateFields2.requiredPlugins().containsValue(pTemplatesCore));

        assertEquals(2, pEncounterForms.requiredPlugins().size());
        assertTrue(pEncounterForms.requiredPlugins().containsValue(pEncounters));
        assertTrue(pEncounterForms.requiredPlugins().containsValue(pLib2));

        assertEquals(1, pEncounters.requiredPlugins().size());
        assertTrue(pEncounters.requiredPlugins().containsValue(pLib2));

        assertEquals(1, pRemCore.requiredPlugins().size());
        assertTrue(pRemCore.requiredPlugins().containsValue(pLib2));

        assertEquals(5, pRemDialogs2.requiredPlugins().size());
        assertTrue(pRemDialogs2.requiredPlugins().containsValue(pRemCore));
        assertTrue(pRemDialogs2.requiredPlugins().containsValue(pOutMedOrders));
        assertTrue(pRemDialogs2.requiredPlugins().containsValue(pInMedOrders));
        assertTrue(pRemDialogs2.requiredPlugins().containsValue(pEncounters));
        assertTrue(pRemDialogs2.requiredPlugins().containsValue(pMedCore));

        assertEquals(1, pInMedOrders.requiredPlugins().size());
        assertTrue(pInMedOrders.requiredPlugins().containsValue(pMedCore));

        assertEquals(1, pOutMedOrders.requiredPlugins().size());
        assertTrue(pOutMedOrders.requiredPlugins().containsValue(pMedCore));

        assertEquals(0, pMedCore.requiredPlugins().size());

        assertEquals(0, pRuntime.requiredPlugins().size());
    }

    public void testCyclicDependendies() throws Exception {
        init(cycleManifestList);
        config.clear();
        assertNoError(Error.CYCLIC_DEPENDENCY);
        IConfigPlugin plugin1 = addPlugin(cycleManifest1);
        IConfigPlugin plugin2 = addPlugin(cycleManifest2);
        config.validate();
        assertError(Error.CYCLIC_DEPENDENCY);
        getError(Error.CYCLIC_DEPENDENCY);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertNull(error.getFoundList());
        assertTrue(error.getPluginList().get(0) == plugin1);
        assertTrue(error.getPluginList().get(1) == plugin2);

        config.clear();
        assertNoError(Error.CYCLIC_DEPENDENCY);
        plugin1 = addPlugin(cycleManifest3);
        plugin2 = addPlugin(cycleManifest4);
        IConfigPlugin plugin3 = addPlugin(cycleManifest5);
        IConfigPlugin plugin4 = addPlugin(cycleManifest6);
        config.validate();
        assertError(Error.CYCLIC_DEPENDENCY);
        getError(Error.CYCLIC_DEPENDENCY);
        assertNotNull(error);
        assertEquals(4, error.getPluginList().size());
        assertNull(error.getFoundList());
        assertTrue(error.getPluginList().get(0) == plugin1);
        assertTrue(error.getPluginList().get(1) == plugin2);
        assertTrue(error.getPluginList().get(2) == plugin4);
        assertTrue(error.getPluginList().get(3) == plugin3);

        config.clear();
        assertNoError(Error.CYCLIC_DEPENDENCY);
        addPlugin(cycleManifest11);
        addPlugin(cycleManifest10);
        IConfigPlugin pluginC = addPlugin(cycleManifest12);
        IConfigPlugin pluginD = addPlugin(cycleManifest13);
        addPlugin(cycleManifest9);
        IConfigPlugin pluginF = addPlugin(cycleManifest7);
        IConfigPlugin pluginG = addPlugin(cycleManifest8);
        config.validate();
        assertError(Error.CYCLIC_DEPENDENCY);
        getError(Error.CYCLIC_DEPENDENCY);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(pluginC));
        assertTrue(error.getPluginList().contains(pluginD));
        getNextError(Error.CYCLIC_DEPENDENCY);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(pluginF));
        assertTrue(error.getPluginList().contains(pluginG));
    }

    public void testExportedDependencies() throws Exception {
        init(exportManifestList);
        assertNoError(Error.MULTI_VERSION_DEPENDENCY);
        IConfigPlugin plugin1 = getPlugin(exportManifest1);
        IConfigPlugin plugin2 = getPlugin(exportManifest2);
        IConfigPlugin plugin3 = getPlugin(exportManifest3);
        IConfigPlugin plugin4 = getPlugin(exportManifest4);
        IConfigPlugin plugin5 = getPlugin(exportManifest5);

        assertEquals(0, plugin3.requiredPlugins().size());

        assertEquals(1, plugin2.requiredPlugins().size());
        assertTrue(plugin2.requiredPlugins().containsValue(plugin3));

        assertEquals(1, plugin1.getPluginDependencies().size());
        assertEquals(2, plugin1.requiredPlugins().size());
        assertTrue(plugin1.requiredPlugins().containsValue(plugin2));
        assertTrue(plugin1.requiredPlugins().containsValue(plugin3));

        assertEquals(3, plugin4.requiredPlugins().size());
        assertTrue(plugin4.requiredPlugins().containsValue(plugin1));
        assertTrue(plugin4.requiredPlugins().containsValue(plugin2));
        assertTrue(plugin4.requiredPlugins().containsValue(plugin3));

        assertEquals(4, plugin5.requiredPlugins().size());
        assertTrue(plugin5.requiredPlugins().containsValue(plugin1));
        assertTrue(plugin5.requiredPlugins().containsValue(plugin2));
        assertTrue(plugin5.requiredPlugins().containsValue(plugin3));
        assertTrue(plugin5.requiredPlugins().containsValue(plugin4));
    }

    public void testMultiVersionDependency() throws Exception {
        init(multiVersionDependency);
        IConfigPlugin a = getPlugin(multiA);
        IConfigPlugin c = getPlugin(multiC);
        IConfigPlugin b1 = getPlugin(multiB1);
        IConfigPlugin b2 = getPlugin(multiB2);
        IConfigPlugin b17 = getPlugin(multiB17);
        assertNoError(Error.MULTI_VERSION_DEPENDENCY);
        assertTrue(a.requiredPlugins().containsValue(b17));
        assertTrue(c.requiredPlugins().containsValue(b17));

        config.getPlugins().remove(b17.getID(), b17);
        config.validate();
        assertError(Error.MULTI_VERSION_DEPENDENCY);
        getError(Error.MULTI_VERSION_DEPENDENCY);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(b1));
        assertTrue(error.getPluginList().contains(b2));
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getFoundList().contains(a));
    }

    public void testExportMapping() throws Exception {
        init(abcList);
        IConfigPlugin a = getPlugin(pluginA);
        IConfigPlugin b = getPlugin(pluginB);
        IConfigPlugin c = getPlugin(pluginC);
        assertTrue(a.requiredPlugins().containsValue(b));
        assertTrue(a.requiredPlugins().containsValue(c));
        if (config.errors().getErrors().size() != 2)
            assertNoErrors();
        getError(Error.MISSING_APPLICATION_PLUGIN);
        assertNotNull(error);
        getError(Error.MISSING_STARTUP_PLUGIN);
        assertNotNull(error);
    }

    public void testExtensionPointErrors() throws Exception {
        init(epList);
        IConfigPlugin ep1 = getPlugin(plgn1);
        IConfigPlugin ep2 = getPlugin(plgn2);
        IConfigPlugin ep3 = getPlugin(plgn3);
        IConfigPlugin ep4 = getPlugin(plgn4);
        assertError(Error.UNAVAILABLE_EXTENSION_POINT);
        assertError(Error.UNREACHABLE_EXTENSION_POINT);
        assertNoError(Error.UNRESOLVED_EXTENSION_POINT);

        getError(Error.UNAVAILABLE_EXTENSION_POINT);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep2));
        assertTrue(error.getPluginList().contains(ep3));
        assertTrue(error.getMessage().indexOf("a.b.c.ep1.eee") >= 0);
        getNextError(Error.UNAVAILABLE_EXTENSION_POINT);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep3));
        assertTrue(error.getPluginList().contains(ep4));
        assertTrue(error.getMessage().indexOf("r.s.t.uvw.xyz") >= 0);
        getNextError(Error.UNAVAILABLE_EXTENSION_POINT);
        assertNull(error);

        getError(Error.UNREACHABLE_EXTENSION_POINT);
        assertNotNull(error);
        assertTrue(error.getMessage().indexOf("a.b.c.ep1.aaa") >= 0);
        assertEquals(1, error.getFoundList().size());
        assertEquals(ep1, error.getFoundList().get(0));
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep4));
        getNextError(Error.UNREACHABLE_EXTENSION_POINT);
        assertNotNull(error);
        assertTrue(error.getMessage().indexOf("d.e.f.ep2.ddd") >= 0);
        assertEquals(1, error.getFoundList().size());
        assertEquals(ep2, error.getFoundList().get(0));
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep4));
        getNextError(Error.UNREACHABLE_EXTENSION_POINT);
        assertNull(error);

        config.getPlugins().remove(ep1.getID(), ep1);
        config.validate();
        assertError(Error.UNRESOLVED_EXTENSION_POINT);
        getError(Error.UNRESOLVED_EXTENSION_POINT);
        assertNotNull(error);
        assertEquals(1, error.getFoundList().size());
        assertEquals(ep1, error.getFoundList().get(0));
        assertEquals(3, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep2));
        assertTrue(error.getPluginList().contains(ep3));
        assertTrue(error.getPluginList().contains(ep4));
    }

    public void testMultiVersionNotLibrary() throws Exception {
        init(mveList);
        assertError(Error.MULTI_VERSION_HAS_EXTENSIONS);
        IConfigPlugin ep1 = getPlugin(plgn1);
        IConfigPlugin ep2 = getPlugin(plgn2);
        IConfigPlugin ep3 = getPlugin(plgn3);
        IConfigPlugin ep4 = getPlugin(plgn4);
        getError(Error.MULTI_VERSION_HAS_EXTENSIONS);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep1));
        assertTrue(error.getPluginList().contains(ep2));
        getNextError(Error.MULTI_VERSION_HAS_EXTENSIONS);
        assertNotNull(error);
        assertEquals(2, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep3));
        assertTrue(error.getPluginList().contains(ep4));
        getNextError(Error.MULTI_VERSION_HAS_EXTENSIONS);
        assertNull(error);
    }

    public void test() throws Exception {
        init(deList);
        IConfigPlugin ep1 = getPlugin(plgn1);
        IConfigPlugin ep2 = getPlugin(plgn2);
        IConfigPlugin ep3 = getPlugin(plgn3);
        IConfigPlugin ep4 = getPlugin(plgn4);
        config.getPlugins().clear();
        config.getPlugins().add(ep1.getID(), ep1);
        config.getPlugins().add(ep2.getID(), ep2);
        config.validate();
        assertError(Error.UNAVAILABLE_DECLARED_CLASS);
        getError(Error.UNAVAILABLE_DECLARED_CLASS);
        assertNotNull(error);
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep1));

        assertNoError(Error.UNPUBLISHED_DECLARED_CLASS);
        assertNoError(Error.UNREACHABLE_DECLARED_CLASS);
        assertNoError(Error.UNRESOLVED_DECLARED_CLASS);
        ep4.getClasses().add("a.b.c.MyClass");
        mgr.loadPluginInformation();
        config.validate();
        assertNoError(Error.UNAVAILABLE_DECLARED_CLASS);
        assertError(Error.UNRESOLVED_DECLARED_CLASS);
        getError(Error.UNRESOLVED_DECLARED_CLASS);
        assertNotNull(error);
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep1));
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getFoundList().contains(ep4));

        config.getPlugins().add(ep4.getID(), ep4);
        config.validate();
        assertNoError(Error.UNAVAILABLE_DECLARED_CLASS);
        assertNoError(Error.UNRESOLVED_DECLARED_CLASS);
        assertError(Error.UNREACHABLE_DECLARED_CLASS);
        error = null;
        getError(Error.UNREACHABLE_DECLARED_CLASS);
        assertNotNull(error);
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep1));
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getFoundList().contains(ep4));

        config.getPlugins().remove(ep1.getID(), ep1);
        config.getPlugins().add(ep3.getID(), ep3);
        config.validate();
        assertNoError(Error.UNREACHABLE_DECLARED_CLASS);
        assertNoError(Error.UNRESOLVED_DECLARED_CLASS);
        assertNoError(Error.UNAVAILABLE_DECLARED_CLASS);
        assertError(Error.UNPUBLISHED_DECLARED_CLASS);
        error = null;
        getError(Error.UNPUBLISHED_DECLARED_CLASS);
        assertNotNull(error);
        assertEquals(1, error.getPluginList().size());
        assertTrue(error.getPluginList().contains(ep3));
        assertEquals(1, error.getFoundList().size());
        assertTrue(error.getFoundList().contains(ep4));

        ep4.getPublished().add("a.b.c.MyClass");
        config.validate();
        assertNoError(Error.UNREACHABLE_DECLARED_CLASS);
        assertNoError(Error.UNRESOLVED_DECLARED_CLASS);
        assertNoError(Error.UNAVAILABLE_DECLARED_CLASS);
        assertNoError(Error.UNPUBLISHED_DECLARED_CLASS);
    }

    //********************
    //*** end of tests ***
    //********************

    private static final int cycleManifest1 = 0;
    private static final int cycleManifest2 = 1;
    private static final int cycleManifest3 = 2;
    private static final int cycleManifest4 = 3;
    private static final int cycleManifest5 = 4;
    private static final int cycleManifest6 = 5;
    private static final int cycleManifest7 = 6;
    private static final int cycleManifest8 = 7;
    private static final int cycleManifest9 = 8;
    private static final int cycleManifest10 = 9;
    private static final int cycleManifest11 = 10;
    private static final int cycleManifest12 = 11;
    private static final int cycleManifest13 = 12;

    private static final String cmRoot = "manifests/cycleManifest";
    private static final String[] cycleManifestList =
        {
            cmRoot + "1.xml",
            cmRoot + "2.xml",
            cmRoot + "3.xml",
            cmRoot + "4.xml",
            cmRoot + "5.xml",
            cmRoot + "6.xml",
            cmRoot + "7.xml",
            cmRoot + "8.xml",
            cmRoot + "9.xml",
            cmRoot + "10.xml",
            cmRoot + "11.xml",
            cmRoot + "12.xml",
            cmRoot + "13.xml" };

    private static final int exportManifest1 = 0;
    private static final int exportManifest2 = 1;
    private static final int exportManifest3 = 2;
    private static final int exportManifest4 = 3;
    private static final int exportManifest5 = 4;

    private static final String emRoot = "manifests/exportManifest";
    private static final String[] exportManifestList =
        {
            emRoot + "1.xml",
            emRoot + "2.xml",
            emRoot + "3.xml",
            emRoot + "4.xml",
            emRoot + "5.xml" };

    private static final int multiA = 0;
    private static final int multiC = 1;
    private static final int multiB1 = 2;
    private static final int multiB2 = 3;
    private static final int multiB17 = 4;

    private static final String mvRoot = "manifests/multiVersionDependency";
    private static final String[] multiVersionDependency =
        {
            mvRoot + "1.xml",
            mvRoot + "2.xml",
            mvRoot + "3.xml",
            mvRoot + "4.xml",
            mvRoot + "5.xml" };

    private static final String root = "gov/va/med/hds/cd/mum/config/testdata/";

    private static final String[] abcList =
        {
            "plugins/a/" + MANIFEST_FILE,
            "plugins/b/" + MANIFEST_FILE,
            "plugins/c/" + MANIFEST_FILE };

    private static final int pluginA = 0;
    private static final int pluginB = 1;
    private static final int pluginC = 2;

    private static final String epRoot = "manifests/extensionPoint";
    private static final String[] epList =
        { epRoot + "1.xml", epRoot + "2.xml", epRoot + "3.xml", epRoot + "4.xml", };

    private static final String mveRoot = "manifests/multiVersionExtension";
    private static final String[] mveList =
        { mveRoot + "1.xml", mveRoot + "2.xml", mveRoot + "3.xml", mveRoot + "4.xml", };

    private static final String deRoot = "manifests/declaredError";
    private static final String[] deList =
        { deRoot + "1.xml", deRoot + "2.xml", deRoot + "3.xml", deRoot + "4.xml", };

    private static final int plgn1 = 0;
    private static final int plgn2 = 1;
    private static final int plgn3 = 2;
    private static final int plgn4 = 3;

    Configuration config;
    ArrayList list;
    IPluginManager mgr;

    public ConfigurationTest(String arg0) {
        super(arg0);
    }

    protected void setUp() throws Exception {
        super.setUp();
        PluginID.reset();
        list = new ArrayList();
    }

    protected void tearDown() throws Exception {
        super.tearDown();
        ConfigPluginLocatorFactory.assignConfigPluginLocator(null);
    }

    private String[] manifestList;

    protected void init() throws PluginLocatorException {
        init(null);
    }

    protected void init(String[] manifestList) throws PluginLocatorException {
        if (manifestList == null) {
            ConfigPluginLocatorFactory.assignConfigPluginLocator(new MockPluginLocator());
            manifestList = null;
        } else {
            this.manifestList = manifestList;
            ConfigPluginLocatorFactory.assignConfigPluginLocator(this);
        }
        mgr = PluginManagerFactory.newPluginManager();
        mgr.reset();
        config = new Configuration(mgr);
    }

    private void reconfig(int[] newConfig) throws Exception {
        config.getPlugins().clear();
        for (int i = 0; i < newConfig.length; i++)
            addPlugin(newConfig[i]);
        config.validate();
    }

    private IConfigPlugin addPlugin(int index) throws Exception {
        IConfigPlugin plugin = getPlugin(index);
        config.getPlugins().add(plugin.getID(), plugin);
        return plugin;
    }

    private IConfigPlugin getPlugin(int index) throws Exception {
        String name;
        if (manifestList == null) {
            name = (String) MOCKJAR_DATA[index][miDir];
            for (int i = 0; i < mgr.getPlugins().size(); i++) {
                IConfigPlugin plugin = (IConfigPlugin) mgr.getPlugins().get(i);
                if (plugin.getDirectory().equals(name)) {
                    return plugin;
                }
            }
        } else {
            name = manifestList[index];
            for (int i = 0; i < mgr.getPlugins().size(); i++) {
                IConfigPlugin plugin = (IConfigPlugin) mgr.getPlugins().get(i);
                if (plugin.getFile(0).getURL().getFile().endsWith(name))
                    return plugin;

            }
        }
        throw new Exception("plugin " + name + " not found");
    }

    private URL getURL(String manifestFileName) {
        return this.getClass().getClassLoader().getResource(root + manifestFileName);
    }

    private Error error;
    private int lastIndex = 0;

    private void getNextError(Error.ErrorType code) {
        getError(code, lastIndex);
    }
    private void getError(Error.ErrorType code) {
        getError(code, 0);
    }

    private void getError(Error.ErrorType code, int start) {
        for (int i = start; i < config.errors().getErrors().size(); i++) {
            Error temp = (Error) config.errors().getErrors().get(i);
            if (temp.getErrorType() == code) {
                error = temp;
                lastIndex = i + 1;
                return;
            }
        }
        error = null;
    }

    private void assertHasErrors() {
        assertHasErrors(true);
    }

    public static void assertHasErrors(ErrorManager mgr) {
        assertHasErrors(mgr, true);
    }

    private void assertNoErrors() {
        assertHasErrors(false);
    }

    public static void assertNoErrors(ErrorManager mgr) {
        assertHasErrors(mgr, false);
    }

    private void assertHasErrors(boolean hasErrors) {
        assertHasErrors(config.errors(), hasErrors);
    }

    public static void assertHasErrors(ErrorManager mgr, boolean hasErrors) {
        if (hasErrors != mgr.hasErrors())
            outputErrors(mgr);
        if ((!hasErrors) && (mgr.hasErrors()))
            mgr.printErrors(System.out);
        assertEquals(hasErrors, mgr.hasErrors());
    }

    private void assertNoError(Error.ErrorType code) {
        assertError(false, code);
    }

    private void assertError(Error.ErrorType code) {
        assertError(true, code);
    }

    private void assertError(boolean hasError, Error.ErrorType code) {
        boolean reallyHasError = config.errors().hasError(code);
        if (hasError != reallyHasError)
            outputErrors();
        assertEquals(hasError, reallyHasError);
    }

    private void outputErrors() {
        outputErrors(config.errors());
    }

    public static void outputErrors(ErrorManager mgr) {
        for (int i = 0; i < mgr.getErrors().size(); i++) {
            Error error = (Error) mgr.getErrors().get(i);
            System.out.println(error.toString());
        }
    }

    private long crcValue = 2673664136L;

    private long settingsCrcValue = 2054235276L;

    private String initialXML =
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + WIN_EOL + "<configuration crc=\"";

    private String startupXML =
        "\">"
            + WIN_EOL
            + "  <startup plugin=\"NoteCore_2.0.0\" run=\"boot.jar\" source=\"NoteCore1\">"
            + WIN_EOL
            + "    <files>"
            + WIN_EOL
            + "      <file>manifest.xml</file>"
            + WIN_EOL
            + "      <file>boot.jar</file>"
            + WIN_EOL
            + "      <file>mum.jar</file>"
            + WIN_EOL
            + "    </files>"
            + WIN_EOL
            + "  </startup>"
            + WIN_EOL;

    private String settingsXML =
        "  <settings>"
            + WIN_EOL
            + "    <testing/>"
            + WIN_EOL
            + "    <testing/>"
            + WIN_EOL
            + "    <one>"
            + WIN_EOL
            + "      <two>"
            + WIN_EOL
            + "        <three/>"
            + WIN_EOL
            + "      </two>"
            + WIN_EOL
            + "    </one>"
            + WIN_EOL
            + "  </settings>"
            + WIN_EOL;

    private String finalXML =
        "  <plugin name=\"Inpatient Med Orders\" directory=\"InpatientMedOrders_1.0.0\" module=\"Orders\" source=\"Inpatient1\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"inpatientmedorders.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>orders.inpatient</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"OrdersCore_2.0.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"My Lib\" directory=\"MyLib_1.0.0\" source=\"Lib1\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"mylib.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>util.one</package>"
            + WIN_EOL
            + "      <package>util.two</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <publish class=\"*\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"My Lib\" directory=\"MyLib_2.0.0\" source=\"Lib2\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"mylib.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>util.one</package>"
            + WIN_EOL
            + "      <package>util.two</package>"
            + WIN_EOL
            + "      <package>util.three</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <publish class=\"*\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Note Core\" directory=\"NoteCore_2.0.0\" module=\"Notes\" source=\"NoteCore1\">"
            + WIN_EOL
            + "    <application class=\"mumtest.Application\"/>"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"boot.jar\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"mum.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>tiu.notes</package>"
            + WIN_EOL
            + "      <package>tiu.notes.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"Runtime_1.0.0\"/>"
            + WIN_EOL
            + "    <publish class=\"mumtest.mum.TestStartup\"/>"
            + WIN_EOL
            + "    <publish class=\"mumtest.mum.TestApplication\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Note Editor\" directory=\"NoteEditor_3.0.0\" module=\"Notes\" source=\"NoteEditor3\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"noteeditor.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>tiu.notes.editor</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"MyLib_2.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"NoteCore_2.0.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Orders Core\" directory=\"OrdersCore_2.0.0\" module=\"Orders\" source=\"Orders\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"orderscore.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>orders</package>"
            + WIN_EOL
            + "      <package>orders.internal</package>"
            + WIN_EOL
            + "      <package>orders.meds</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <publish class=\"orders.DummyClass\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Outpatient Med Orders\" directory=\"OutpatientMedOrders_1.0.0\" module=\"Orders\" source=\"Outpatient1\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"outpatientmedorders.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>orders.outpatient</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"OrdersCore_2.0.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"PCE\" directory=\"PCE_1.0.0\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"pce.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>pce</package>"
            + WIN_EOL
            + "      <package>pce.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"MyLib_2.0.0\"/>"
            + WIN_EOL
            + "    <publish class=\"pce.DummyClass\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"PCE Forms\" directory=\"PCEForms_1.0.0\" module=\"PCE\" source=\"pceForms_1.0.0\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"pceforms.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>pce</package>"
            + WIN_EOL
            + "      <package>pce.forms</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"MyLib_2.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"PCE_1.0.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Rem Dialogs\" directory=\"RemDialogs_1.0.0\" module=\"Reminders\" source=\"RemDlgs2\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"remdialogs.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>reminders.dialogs</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"InpatientMedOrders_1.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"OrdersCore_2.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"OutpatientMedOrders_1.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"PCE_1.0.0\"/>"
            + WIN_EOL
            + "    <requires plugin=\"ReminderCore_1.0.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Reminder Core\" directory=\"ReminderCore_1.0.0\" module=\"Reminders\" source=\"RemCore\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"remindercore.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>reminders</package>"
            + WIN_EOL
            + "      <package>reminders.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"MyLib_2.0.0\"/>"
            + WIN_EOL
            + "    <publish class=\"*\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Runtime\" directory=\"Runtime_1.0.0\" source=\"Runtime1\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"runtime.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>gov.va.med.hds.cd.mum.update</package>"
            + WIN_EOL
            + "      <package>mum.common</package>"
            + WIN_EOL
            + "      <package>gov.va.med.hds.cd.mum</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <publish class=\"gov.va.med.hds.cd.mum.update.DummyClass\"/>"
            + WIN_EOL
            + "    <publish class=\"gov.va.med.hds.cd.mum.DummyClass\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Template Core\" directory=\"TemplateCore_1.2.0\" module=\"Templates\" source=\"TemplatesCore1\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"templatecore.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>tiu.templates</package>"
            + WIN_EOL
            + "      <package>tiu.templates.core</package>"
            + WIN_EOL
            + "      <package>tiu.templates.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Template Fields\" directory=\"TemplateFields_1.5.0\" module=\"Templates\" source=\"TemplateFields2\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"templatefields.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>tiu.templates</package>"
            + WIN_EOL
            + "      <package>tiu.templates.fields</package>"
            + WIN_EOL
            + "      <package>tiu.templates.fields.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"TemplateCore_1.2.0\"/>"
            + WIN_EOL
            + "    <publish class=\"tiu.templates.DummyClass\"/>"
            + WIN_EOL
            + "    <publish class=\"tiu.templates.fields.DummyClass\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "  <plugin name=\"Templates\" directory=\"Templates_2.0.0\" source=\"Templates2\">"
            + WIN_EOL
            + "    <file filename=\"manifest.xml\" crc=\"0\" size=\"0\"/>"
            + WIN_EOL
            + "    <file filename=\"templates.jar\" crc=\"0\" size=\"0\">"
            + WIN_EOL
            + "      <package>tiu.templates</package>"
            + WIN_EOL
            + "      <package>tiu.templates.internal</package>"
            + WIN_EOL
            + "    </file>"
            + WIN_EOL
            + "    <requires plugin=\"TemplateCore_1.2.0\"/>"
            + WIN_EOL
            + "  </plugin>"
            + WIN_EOL
            + "</configuration>";

    private String goodConfigXML = initialXML + Long.toString(crcValue) + startupXML + finalXML;

    private String goodConfigXMLWithSettings =
        initialXML + Long.toString(settingsCrcValue) + startupXML + settingsXML + finalXML;

    public void createDistributionPlugins(ListMap list) throws PluginLocatorException {
        for (int i = 0; i < manifestList.length; i++) {
            URL manifestURL = getURL(manifestList[i]);
            IConfigPlugin plugin;
            try {
                plugin = ConfigPluginFactory.newConfigPlugin(null, manifestURL);
            } catch (InvalidPluginException e) {
                throw new PluginLocatorException(e);
            }
            list.add(plugin.getID(), plugin);
        }
    }
}
