package gov.va.med.hds.cd.mum.test;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import junit.framework.TestCase;

public class VersionTest extends TestCase {

	public VersionTest(String name) {
		super(name);
	}

	public void testCreateFromInts() {
		Version v = new Version(2, 3, 1);
		assertEquals(2, v.getMajorComponent());
		assertEquals(3, v.getMinorComponent());
		assertEquals(1, v.getServiceComponent());
		assertEquals("", v.getQualifierComponent());
		assertTrue(v.equals(v));
	}
	
	public void testCreateFromString() {
		Version v = new Version("3");
		assertEquals(3, v.getMajorComponent());
		assertEquals(0, v.getMinorComponent());
		assertEquals(0, v.getServiceComponent());
		assertEquals("", v.getQualifierComponent());
		assertTrue(v.equals(v));
		
		v = new Version("1.9");
		assertEquals(1, v.getMajorComponent());
		assertEquals(9, v.getMinorComponent());
		assertEquals(0, v.getServiceComponent());
		assertEquals("", v.getQualifierComponent());
		assertTrue(v.equals(v));
		
		v = new Version("1.0.127564");
		assertEquals(1, v.getMajorComponent());
		assertEquals(0, v.getMinorComponent());
		assertEquals(127564, v.getServiceComponent());
		assertEquals("", v.getQualifierComponent());
		assertTrue(v.equals(v));
		
		v = new Version("3.7.2.build-127J");
		assertEquals(3, v.getMajorComponent());
		assertEquals(7, v.getMinorComponent());
		assertEquals(2, v.getServiceComponent());
		assertEquals("build-127J", v.getQualifierComponent());
		assertTrue(v.equals(v));
	}
	
	public void testCreateFromIntsNegativeComponentFailures() {
		try {
			Version v = new Version(-2, 3, 1);
			fail("Expected exception for negative major component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version(2, -3, 1);
			fail("Expected exception for negative minor component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version(2, 3, -1);
			fail("Expected exception for negative service component");
		} catch (VersionFormatException e) { }
	}
	
	public void testCreateFromStringEmptyFailures() {
		try {
			Version v = new Version(null);
			fail("Expected exception for empty version");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("");
			fail("Expected exception for empty version");
		} catch (VersionFormatException e) { }
	}
	
	public void testCreateFromStringDotFailures() {
		try {
			Version v = new Version(".34");
			fail("Expected exception for version starting with .");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("34.");
			fail("Expected exception for version ending with .");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("123.34..2354");
			fail("Expected exception for version containg a double .");
		} catch (VersionFormatException e) { }
	}
	
	public void testCreateFromStringNegativeComponentFailures() {
		try {
			Version v = new Version("-2.34.0.sds");
			fail("Expected exception for negative major component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("34.-56.234.gfsdf");
			fail("Expected exception for negative minor component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("543.65.-342.3fdv");
			fail("Expected exception for negative service component");
		} catch (VersionFormatException e) { }
	}
	
	public void testCreateFromStringNonNumericComponentFailures() {
		try {
			Version v = new Version("87&.234.35.23");
			fail("Expected exception for non-numeric major component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("34.sdf.234.gfsdf");
			fail("Expected exception for non-numeric minor component");
		} catch (VersionFormatException e) { }
		try {
			Version v = new Version("543.65.#235.3fdv");
			fail("Expected exception for non-numeric service component");
		} catch (VersionFormatException e) { }
	}
	
	public void testToString() {
		Version vi = new Version("3.7.2.build-127J");
		vi.equals(new Version(vi.toString()));
	}
	
	public void testCompatibilityComparison() {
		Version v1 = new Version("1.1.3.build-126J");
		
		Version v2 = new Version("2.0.1.build-17J");
		assertFalse(v2.isCompatibleWith(v1));
		
		v2 = new Version("1.2.1.build-198");
		assertTrue(v2.isCompatibleWith(v1));
		v2 = new Version("1.0.1.build-198");
		assertFalse(v2.isCompatibleWith(v1));
		
		v2 = new Version("1.1.7.build-198");
		assertTrue(v2.isCompatibleWith(v1));
		v2 = new Version("1.1.1.build-198");
		assertFalse(v2.isCompatibleWith(v1));
		
		v2 = new Version("1.1.3.build-198");
		assertTrue(v2.isCompatibleWith(v1));
		v2 = new Version("1.1.3.build-111J");
		assertFalse(v2.isCompatibleWith(v1));
	}
	
	public void testEquivalencyComparison() {
		Version v1 = new Version("1.1.3.build-126J");
		
		Version v2 = new Version("2.0.1.build-17J");
		assertFalse(v2.isEquivalentTo(v1));
		v2 = new Version("1.2.1.build-198");
		assertFalse(v2.isEquivalentTo(v1));
		v2 = new Version("1.0.1.build-198");
		assertFalse(v2.isEquivalentTo(v1));
		
		v2 = new Version("1.1.1.build-198");
		assertFalse(v2.isEquivalentTo(v1));
		v2 = new Version("1.1.5.build-198");
		assertTrue(v2.isEquivalentTo(v1));
		
		v2 = new Version("1.1.3.build-198");
		assertTrue(v2.isEquivalentTo(v1));
		v2 = new Version("1.1.3.build-12");
		assertFalse(v2.isEquivalentTo(v1));
	}
	
	public void testPerfectEqualityComparison() {
		Version v1 = new Version("3.7.2.build-127J");
		Version v2 = new Version("3.7.2.build-127J");
		assertTrue(v1.isPerfect(v2));
		v2 = new Version("2.9.1.build-198");
		assertFalse(v1.isPerfect(v2));
	}
	
	public void testGreaterThanComparison() {
		Version v1 = new Version("3.7.2.build-127J");
		Version v2 = new Version("3.7.2.build-127J");
		assertFalse(v2.isGreaterThan(v1));
		
		v2 = new Version("3.7.2.build-110J");
		assertFalse(v2.isGreaterThan(v1));
		v2 = new Version("3.7.2.build-206J");
		assertTrue(v2.isGreaterThan(v1));
		
		v2 = new Version("3.7.1");
		assertFalse(v2.isGreaterThan(v1));
		v2 = new Version("3.7.3");
		assertTrue(v2.isGreaterThan(v1));
		
		v2 = new Version("3.6");
		assertFalse(v2.isGreaterThan(v1));
		v2 = new Version("3.8");
		assertTrue(v2.isGreaterThan(v1));
		
		v2 = new Version("2");
		assertFalse(v2.isGreaterThan(v1));
		v2 = new Version("4");
		assertTrue(v2.isGreaterThan(v1));
	}
	
	public void testGreaterOrEqualToComparison() {
		Version v1 = new Version("3.7.2.build-127J");
		Version v2 = new Version("3.7.2.build-127J");
		assertTrue(v2.isGreaterOrEqualTo(v1));
		
		v2 = new Version("3.7.2.build-110J");
		assertFalse(v2.isGreaterOrEqualTo(v1));
		v2 = new Version("3.7.2.build-206J");
		assertTrue(v2.isGreaterOrEqualTo(v1));
		
		v2 = new Version("3.7.1");
		assertFalse(v2.isGreaterOrEqualTo(v1));
		v2 = new Version("3.7.3");
		assertTrue(v2.isGreaterOrEqualTo(v1));
		
		v2 = new Version("3.6");
		assertFalse(v2.isGreaterOrEqualTo(v1));
		v2 = new Version("3.8");
		assertTrue(v2.isGreaterOrEqualTo(v1));
		
		v2 = new Version("2");
		assertFalse(v2.isGreaterOrEqualTo(v1));
		v2 = new Version("4");
		assertTrue(v2.isGreaterOrEqualTo(v1));
	}
}
