package gov.va.med.hds.cd.mum.update.test;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.Settings;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.test.TestUtils;
import gov.va.med.hds.cd.mum.update.ConfigReader;
import gov.va.med.hds.cd.mum.update.ConfigReaderFactory;
import gov.va.med.hds.cd.mum.update.IConfigReader;
import gov.va.med.hds.cd.mum.update.IUpdateConstants;
import gov.va.med.hds.cd.mum.update.PluginSpec;

import java.net.MalformedURLException;
import java.net.URL;

import junit.framework.TestCase;

public class ConfigReaderTest
    extends TestCase
    implements IUpdateConstants, IMUMConstants, IBootConstants {

    public void testConfig1() throws InvalidConfigurationException {
        assertEquals(0, plugins.size());
        assertEquals(0, reader.getCRC());
        loadConfig(utils.CONFIG1);
        assertEquals(3, plugins.size());
        assertEquals(1001667165L, reader.getCRC());
        
        getPlugin("one.two.three.four.five_1.0.0");
        assertEquals("one.two.three.four.five", plugin.getID());
        assertEquals("1.0.0", plugin.getVersion().toString());
        assertEquals("one.two.three", plugin.getModule());
        assertEquals(plugin, reader.getStartupPlugin());
        assertNull(plugin.getApplicationClassName());
        assertFalse(plugin.publishAllClasses());
        assertEquals(2, plugin.getPublished().size());
        assertTrue(plugin.getPublished().contains("this.is.a.test.SomeClass"));
        assertTrue(plugin.getPublished().contains("this.is.a.test.SomeOtherClass"));
        assertEquals(2, plugin.getFileCount());
        assertFile(1232101, "manifest.xml");
        assertFile(
            1232102,
            "one.jar",
            new String[] {
                "pkg.one.two.three",
                "pkg.one.two.three.four",
                "pkg.one.two.four",
                "pkg.whatever" });

        getPlugin("one.two.three.four.five.six_0.0.0.whatever");
        assertEquals("one.two.three.four.five.six", plugin.getID());
        assertEquals("0.0.0.whatever", plugin.getVersion().toString());
        assertEquals("some.other.module", plugin.getModule());
        assertNull(plugin.getApplicationClassName());
        assertTrue(plugin.publishAllClasses());
        assertEquals(0, plugin.getPublished().size());
        assertEquals(4, plugin.getFileCount());
        assertFile(1232103, "manifest.xml");
        assertFile(1232104, "one.jar", new String[] { "pkg.one.two.three", "pkg.this.is.a.test" });
        assertFile(1232105, "two.jar");
        assertFile(
            1232106,
            "three.jar",
            new String[] { "testing.default.package", DEFAULT_PACKAGE });

        getPlugin("this.is.a.test_1.2.3.4");
        assertEquals("this.is.a.test", plugin.getID());
        assertEquals("1.2.3.4", plugin.getVersion().toString());
        assertEquals("this.is.a.test", plugin.getModule());
        assertEquals("some.pacakge.Application", plugin.getApplicationClassName());
        assertFalse(plugin.publishAllClasses());
        assertEquals(0, plugin.getPublished().size());
        assertEquals(1, plugin.getFileCount());
        assertFile(1232107, "manifest.xml");

		Settings settings = reader.getSettings();
		assertNotNull(settings);
		assertNotNull(settings.data());
		assertTrue(settings.data().hasRoot());
		int root = settings.data().getRootID();
		assertEquals("settings", settings.data().getName(root));
		assertEquals(0, settings.data().getChildCount(root));
		assertEquals(0, settings.data().getAttributeCount(root));        
    }

    public void testErrorConditions() throws MalformedURLException {
        try {
            reader.readConfiguration(null, null);
            fail("Should have thrown NullPointerException");
        } catch (InvalidConfigurationException e) {
            fail(e.getMessage());
        } catch (NullPointerException e) {
        }
        try {
            reader.readConfiguration(new URL("file:/fdsaf"), null);
            fail("Should have thrown NullPointerException");
        } catch (InvalidConfigurationException e) {
            fail(e.getMessage());
        } catch (NullPointerException e) {
        }
        try {
            reader.readConfiguration(null, plugins);
            fail("Should have thrown NullPointerException");
        } catch (InvalidConfigurationException e) {
            fail(e.getMessage());
        } catch (NullPointerException e) {
        }
        Class treeDataClass = TreeDataException.class;
        assertError(treeDataClass, utils.MISSING_ROOT);
        assertError(NO_CONFIG_ROOT, utils.INVALID_ROOT);
        assertError(MISSING_CONFIG_CRC, utils.MISSING_CRC);
        Class numClass = NumberFormatException.class;
        assertError(MISSING_CONFIG_CRC, numClass, utils.INVALID_CRC);
        assertError(MULTIPLE_SETTINGS, utils.MULTIPLE_SETTINGS);
        assertError(MISSING_STARTUP_PLUGIN_TXT, utils.MISSING_STARTUP);
        assertError(MISSING_APPLICATION_PLUGINS_TXT, utils.MISSING_APP);
        assertError(MISSING_MANIFEST_FILE, utils.MISSING_MANIFEST);
        assertError(MULTIPLE_STARTUP_CLASSES, utils.MULTIPLE_STARTUPS);
        assertError(MULTIPLE_APPLICATION_CLASSES, utils.MULTIPLE_APPS);
        if (ConfigReader.ENABLE_CRC_CHECKING)
            assertError(INCORRET_CONFIGURATION_CRC, utils.INCORRECT_CRC);
    }

    //********************
    //*** end of tests ***
    //********************

    public ConfigReaderTest(String arg0) {
        super(arg0);
        reader = ConfigReaderFactory.newConfigReader();
        plugins = new ListMap();
    }

    protected void setUp() throws Exception {
        super.setUp();
        utils = new TestUtils(this);
    }

    protected void tearDown() throws Exception {
        super.tearDown();
    }

    private IConfigReader reader;
    private ListMap plugins;
    private PluginSpec plugin;
    private TestUtils utils;

    private void loadConfig(String configFileName) throws InvalidConfigurationException {
        reader.readConfiguration(utils.getConfigURL(configFileName), plugins);
    }

    private void getPlugin(String dirName) {
        for (int i = 0; i < plugins.size(); i++) {
            PluginSpec spec = (PluginSpec) plugins.get(i);
            if (spec.getDirectory().equals(dirName)) {
                plugin = spec;
                return;
            }
        }
        fail("Should have had plugin " + dirName);
    }

    private void assertFile(long crc, String file) {
        PluginSpecTest.assertFile(plugin, crc, file, null);
    }

    private void assertFile(long crc, String file, String[] pkgs) {
        PluginSpecTest.assertFile(plugin, crc, file, pkgs);
    }

    private void assertError(Class exceptionClass, String filename) {
        assertError("", exceptionClass, filename);
    }
    private void assertError(String message, String filename) {
        assertError(message, null, filename);
    }

    private void assertError(String message, Class exceptionClass, String filename) {
        URL url = null;
        if (filename != null)
            url = utils.getConfigURL(filename);
        try {
            reader.readConfiguration(url, plugins);
            fail("Should have thrown InvalidConfigurationException");
        } catch (InvalidConfigurationException e) {
            if (message.length() > 0)
                assertEquals(message, e.getMessage());
            if (exceptionClass != null)
                assertTrue(e.getCause().getClass() == exceptionClass);
        }
    }

}
