package gov.va.med.hds.cd.mum;

import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.utils.ILoadableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.KeyedList;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;

/**
 * Root class implementing the core specification for plugin classes needing
 * dependency information.
 */
public abstract class DependentPlugin implements IDependentPlugin {

    /**
     * Constructor for DependentPlugin 
     */
    public DependentPlugin() {
        super();
        published = new SimpleList();
        requiredPlugins = new KeyedList();
    }

    /**
     * @see gov.va.med.hds.cd.mum.IBasePlugin#getDirectory()
     */
    public String getDirectory() {
        return directory;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IBasePlugin#getID()
     */
    public String getID() {
        return id;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getModule()
     */
    public String getModule() {
        return module;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IBasePlugin#getVersion()
     */
    public Version getVersion() {
        return version;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getApplicationClassName()
     */
    public String getApplicationClassName() {
        return applicationClass;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getPublished()
     */
    public SimpleList getPublished() {
        return published;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getPluginID()
     */
    public PluginID getPluginID() {
        return pluginID;
    }

    /**
     * Changes the plugin installation directory.
     * @param directory
     */
    protected void setDirectory(String directory) {
        this.directory = directory;
    }

    /**
     * Changes the plugin id.
     * @param id
     */
    protected void setID(String id) {
        this.id = id;
    }

    /**
     * Changes the module of which the plugin is a member.
     * @param module
     */
    protected void setModule(String module) {
        this.module = module;
    }

    /**
     * Changes the plugin version.
     * @param version
     */
    protected void setVersion(Version version) {
        this.version = version;
    }

    /**
     * Changes the name of the plugin class implementing the IApplication
     * interface.
     * @param applicationClass
     */
    protected void setApplicationClass(String applicationClass) {
        this.applicationClass = applicationClass;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#requiredPlugins()
     */
    public KeyedList requiredPlugins() {
        return requiredPlugins;
    }
    /**
     * Assignes a PluginID to the plugin class, based on the current id
     * and version.  The value generated is returned by the getPluginID method.
     */
    protected void initPlugin() {
        pluginID = PluginID.getPluginID(getID(), getVersion());
    }

    /**
     * Returns the xml of the manifest file specified in the url parameter.
     * @param url - url of the manifest file.
     * @return ITreeData
     * @throws InvalidPluginException
     */
    protected ITreeData getManifestXML(URL url) throws InvalidPluginException {
        InputStream in = getManifestInputStream(url);
        ITreeData xml = getManifestDocument(in);
        if (!IPluginConstants.PLUGIN_ELEMENT.equals(xml.getName(xml.getRootID())))
            badPlugin(IMUMConstants.INVALID_MANIFEST);
        return xml;
    }

    private ITreeData getManifestDocument(InputStream in) throws InvalidPluginException {
        ILoadableTreeData xml = null;
        try {
            try {
                xml = TreeDataFactory.newSAXTreeData();
                xml.load(in);
            } catch (TreeDataException e) {
                badPlugin(e);
            }
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                }
            }
        }
        return xml;
    }

    private InputStream getManifestInputStream(URL url) throws InvalidPluginException {
        URLConnection conn;
        InputStream in = null;
        try {
            conn = url.openConnection();
            conn.connect();
            in = new BufferedInputStream(conn.getInputStream());
        } catch (IOException e) {
            badPlugin(e);
        }
        return in;
    }

    /**
     * Specifies the absolute path of the plugin, used in error messages.
     * @param path
     */
    protected void setPluginPath(String path) {
        pluginPath = path;
    }

    private String getPluginPath() {
        return pluginPath;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IPluginExceptionManager#badPlugin(java.lang.String, java.lang.Throwable)
     */
    public void badPlugin(String message, Throwable cause) throws InvalidPluginException {
        StringBuffer buf = new StringBuffer(IMUMConstants.PLUGIN_ERROR);
        buf.append(getPluginPath());
        buf.append(IBootConstants.EOL);
        if (message != null)
            buf.append(message);
        throw new InvalidPluginException(buf.toString(), cause);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IPluginExceptionManager#badPlugin(java.lang.Throwable)
     */
    public void badPlugin(Throwable cause) throws InvalidPluginException {
        badPlugin(null, cause);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IPluginExceptionManager#badPlugin(java.lang.String)
     */
    public void badPlugin(String message) throws InvalidPluginException {
        badPlugin(message, null);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IPluginExceptionManager#badPlugin()
     */
    public void badPlugin() throws InvalidPluginException {
        badPlugin(null, null);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getSourceDirectory()
     */
    public String getSourceDirectory() {
        return sourceDirectory;
    }

    /**
     * Sets the source directory.
     * @param sourceDirectory - the source directory to set
     */
    protected void setSourceDirectory(String sourceDirectory) {
        this.sourceDirectory = sourceDirectory;
    }

    /**
     * @see gov.va.med.hds.cd.mum.IDependentPlugin#getName()
     */
    public String getName() {
        return name;
    }

    /**
     * Sets the human readable name of the plugin
     */
    protected void setName(String name) {
        this.name = name;
    }

    private String module;
    private String id;
    private Version version;
    private String directory;
    private String applicationClass;
    private PluginID pluginID;
    private SimpleList published;
    private String pluginPath;
    private String sourceDirectory;
    private KeyedList requiredPlugins;
    private String name;
}
