package gov.va.med.hds.cd.mum.boot;

import gov.va.med.hds.cd.mum.boot.loader.ClassLoaderManager;
import gov.va.med.hds.cd.mum.boot.loader.IDesktopClassLoader;
import gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin;
import gov.va.med.hds.cd.mum.boot.utils.Utils;

import java.io.File;
import java.util.Arrays;

/**
 * The first class called when launching the application
 */
public class Boot implements IBootConstants {

    private static IAbortManager manager;
    private static IStartupInfoFactory factory;
    private static String startupClass;
    private static boolean doSplashScreen = true;

    /**
     * Assigns an alternate abort manager to the boot - used mainly for testing.
     * @param abortManager
     */
    public static void assignAbortManager(IAbortManager abortManager) {
        manager = abortManager;
    }

    /**
     * Assigns an alternate startup info factory to the boot - used mainly for testing 
     * @param factory
     */
    public static void assignStartupInfoFactory(IStartupInfoFactory factory) {
        Boot.factory = factory;
    }

    /**
     * Assigns an alterante startup class to the boot - used mainly for testing 
     * @param startupClass
     */
    public static void assignStartupClass(String startupClass) {
        Boot.startupClass = startupClass;
    }

    /**
     * Launches the application.
     * @param args - command line arguments
     */
    public static void main(String[] args) {
    	final String[] argList = args;
    	DesktopThreadGroup tg = new DesktopThreadGroup();
    	Thread t = new Thread(tg, new Runnable() {
			public void run() {
				main(Boot.class.getClassLoader(), argList);
			}
    	});
    	DesktopThreadGroup.addUncaughtExceptionHandler(new DialogUncaughtExceptionHandler());
    	t.start();
    }

    /**
     * Alternate main used in unit testing that allows the correct class loader
     * to be used.  Called by the primary main method.
     * @param parentLoader
     * @param args
     */
    public static void main(ClassLoader parentLoader, String[] args) {
        init();
        IStartupInfo info = null;
        try {
            validateVersion();
//            ConsoleOutput.install();
//			LogOutput.ins
            info = factory.createStartupInfo(args);
            createSplashScreen(info);
            File startupJar = new File(info.getStartupDirectory(), STARTUP_JAR);
            if ((!startupJar.exists()) || (startupJar.isDirectory()))
                throw new BootFailureException(NO_STARTUP_JAR_FOUND);
            try {
                ClassLoaderManager.init(args);
                ClassLoader loader = (ClassLoader) ClassLoaderManager.getClassLoader(parentLoader);
                IDesktopPlugin plugin = new StartupDesktopPlugin(startupJar);
                ((IDesktopClassLoader) loader).assignPlugin(plugin);
                info.setLoader((ClassLoader) loader);
                Class startupClass = loader.loadClass(Boot.startupClass);
                IStartup startup = (IStartup) startupClass.newInstance();
                startup.run(info);
            } catch (Exception e) {
                if ((e instanceof SilentAbortException)
                    || ((e instanceof UpdateAuthorizationFailureException)))
                    throw e;
                throw new BootFailureException(UNEXPECTED_STARTUP_FAILURE, e);
            }
        } catch (Throwable e) {
            if ((info != null) && (info.getSplashScreen() != null))
                info.getSplashScreen().close();
            manager.abort(e, true);
        }
    }

    private static void createSplashScreen(IStartupInfo info) {
        if (doSplashScreen && (!Utils.hasArgument(info.getArgs(), SILENT_UPDATE_TAG))) {
            ISplashScreen splash = StatusNotifierFactory.newSplashScreen();
            info.setSplashScreen(splash);
        }
    }

    private static void init() {
        if (manager == null)
            manager = new AbortManager();
        if (factory == null)
            factory = new StartupInfoFactory();
        if (startupClass == null)
            startupClass = STARTUP_CLASS;
    }

    private static void validateVersion() {
        String ver = System.getProperty(JVM_VERSION_PROPERTY);
        int major = 0;
        int minor = 0;
        int build = 0;
        String[] version = versionSafeSplit(ver);
        try {
            if (version[0] != null)
                major = Integer.parseInt(version[0]);
            if (version[1] != null)
                minor = Integer.parseInt(version[1]);
            if (version[2] != null)
                build = Integer.parseInt(version[2]);
        } catch (NumberFormatException e) {
            badVersion();
        }
        if (valid(major, JVM_MINIMUM_VERSION_MAJOR))
            return;
        if (valid(minor, JVM_MINIMUM_VERSION_MINOR))
            return;
        valid(build, JVM_MINIMUM_VERSION_BUILD);
    }

    private static String[] versionSafeSplit(String ver) {
        int idx = ver.indexOf("_");
        if (idx >= 0)
            ver = ver.substring(0, idx);
        String[] version = new String[3];
        Arrays.fill(version, null);
        idx = ver.indexOf(".");
        if (idx > 0) {
            version[0] = ver.substring(0, idx);
            int idx1 = ver.indexOf(".", idx + 1);
            if (idx1 > 0) {
                version[1] = ver.substring(idx + 1, idx1);
                version[2] = ver.substring(idx1 + 1);
            }
        }
        return version;
    }

    private static boolean valid(int current, int needed) {
        if (current > needed)
            return true;
        if (current < needed)
            badVersion();
        return false;
    }

    private static void badVersion() {
        throw new IncompatibleVersionException(
            JVM_VERSION_ERROR_1 + System.getProperty(JVM_VERSION_PROPERTY) + JVM_VERSION_ERROR_2);
    }
    
    /**
     * Allows shutting off of the splash screen for testing purposes
     */
    public static void setDoSplashScreen(boolean doSplash) {
        doSplashScreen = doSplash;
    }
}
