package gov.va.med.hds.cd.mum.boot;

import gov.va.med.hds.cd.mum.boot.plaf.desktop.LookAndFeelLauncher;

import java.util.ArrayList;
import java.util.List;

/**
 * Used to control the splash screen, initially load the splash screen on a seperate thread, and
 * to send messages to the splash screen as needed. 
 */
public class Notifier implements IStatusManager {

	/**
     * Constructor for Notifier 
     */
    public Notifier() {
		running = true;
		Thread thread = new SplashThread(this);
		thread.setName("Splash Screen Thread");
		thread.start();
	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusManager#ensureWindowDisplayed()
     */
    public void ensureWindowDisplayed() {
		if (running && (splash == null)) {
			while (true) {
				synchronized (monitor) {
					try {
						monitor.wait();
						return;
					} catch (InterruptedException e) {
					}
				}
			}
		}
	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#showProgressBar(long, java.lang.String)
     */
    public synchronized void showProgressBar(long maxValue, String message) {
		if (splash == null)
			cacheCommand(SHOW_PROGRESS_BAR, maxValue, message);
		else
			splash.showProgressBar(maxValue, message);

	}
	
	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#updateProgressBar(long)
     */
    public synchronized void updateProgressBar(long value) {
		if (splash == null)
			cacheCommand(UPDATE_PROGRESS_BAR, value, null);
		else
			splash.updateProgressBar(value);

	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#updateProgressBarMax(long)
     */
    public synchronized void updateProgressBarMax(long value) {
		if (splash == null)
			cacheCommand(UPDATE_PROGRESS_BAR_MAX, value, null);
		else
			splash.updateProgressBarMax(value);
	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#updateProgressMessage(java.lang.String)
     */
    public synchronized void updateProgressMessage(String message) {
		if (splash == null)
			cacheCommand(UPDATE_PROGRESS_MESSAGE, 0, message);
		else
			splash.updateProgressMessage(message);
	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#hideProgressBar()
     */
    public synchronized void hideProgressBar() {
		if (splash == null)
			cacheCommand(HIDE_PROGRESS_BAR, 0, null);
		else
			splash.hideProgressBar();

	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#displayMessage(java.lang.String)
     */
    public synchronized void displayMessage(String message) {
		if (splash == null)
			cacheCommand(DISPLAY_MESSAGE, 0, message);
		else
			splash.displayMessage(message);

	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#setVersion(java.lang.String)
     */
    public synchronized void setVersion(String version) {
		if (splash == null)
			cacheCommand(SET_VERSION, 0, version);
		else
			splash.setVersion(version);

	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#showCancelButton(boolean)
     */
    public synchronized void showCancelButton(boolean value) {
		if (splash == null) {
			long show;
			if (value)
				show = 1;
			else
				show = 0;
			cacheCommand(SHOW_CANCEL_MESSAGE, show, null);
		} else
			splash.showCancelButton(value);

	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#isCanceled()
     */
    public synchronized boolean isCanceled() {
		if (splash == null)
			return false;
		else
			return splash.isCanceled();
	}

	/**
     * @see gov.va.med.hds.cd.mum.boot.ISplashScreen#close()
     */
    public synchronized void close() {
		if (splash != null) {
			ensureWindowDisplayed();
			splash.close();
			splash = null;
			if (commands != null) {
				commands.clear();
				commands = null;
			}
			closePending = false;
			running = false;
		} else
			closePending = true;
	}

	private static final int SHOW_PROGRESS_BAR = 1;
	private static final int UPDATE_PROGRESS_BAR = 2;
	private static final int UPDATE_PROGRESS_BAR_MAX = 3;
	private static final int UPDATE_PROGRESS_MESSAGE = 4;
	private static final int HIDE_PROGRESS_BAR = 5;
	private static final int SET_VERSION = 6;
	private static final int DISPLAY_MESSAGE = 7;
	private static final int SHOW_CANCEL_MESSAGE = 8;
	private static final int SET_CANCEL_MESSAGE = 9;

	private void cacheCommand(int command, long value, String text) {
		if (commands == null)
			commands = new ArrayList();
		commands.add(new Command(command, value, text));
	}

	synchronized void init() {
		try {
			if (splash != null)
				return;
			LookAndFeelLauncher.initDefaultLookAndFeel();
			splash = new Splash();
			if (closePending)
				close();
			else if (commands != null) {
				splash.beginDisplayUpdate();
				try {
					for (int i = 0; i < commands.size(); i++) {
						Command cmd = (Command) commands.get(i);
						switch (cmd.command) {
							case SHOW_PROGRESS_BAR :
								splash.showProgressBar(cmd.value, cmd.text);
								break;
							case UPDATE_PROGRESS_BAR :
								splash.updateProgressBar(cmd.value);
								break;
							case UPDATE_PROGRESS_BAR_MAX :
								splash.updateProgressBarMax(cmd.value);
								break;
							case UPDATE_PROGRESS_MESSAGE :
								splash.updateProgressMessage(cmd.text);
								break;
							case HIDE_PROGRESS_BAR :
								splash.hideProgressBar();
								break;
							case SET_VERSION :
								splash.setVersion(cmd.text);
								break;
							case DISPLAY_MESSAGE :
								splash.displayMessage(cmd.text);
								break;
							case SHOW_CANCEL_MESSAGE :
								splash.showCancelButton(cmd.value == 1);
								break;
							case SET_CANCEL_MESSAGE :
								splash.setCancelMessage(cmd.text);
								break;
						}
					}
				} finally {
					splash.endDisplayUpdate();
				}
				commands.clear();
				commands = null;
			}
		} finally {
			synchronized (monitor) {
				monitor.notifyAll();
			}
		}
	}

	private boolean closePending = false;
	private List commands;
	private Object monitor = new Object();
	private boolean running;
	private Splash splash;

	private class Command {
		int command;
		long value;
		String text;
		public Command(int command, long value, String text) {
			this.command = command;
			this.value = value;
			this.text = text;
		}
	}

	private class SplashThread extends Thread {

		public SplashThread(Notifier notifier) {
			this.notifier = notifier;
		}

		public void run() {
			try {
				notifier.init();
			} finally {
				notifier = null;
			}
		}

		private Notifier notifier;
	}

	/**
	 * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#beginDisplayUpdate()
	 */
	public void beginDisplayUpdate() {
		if(splash != null)
			splash.beginDisplayUpdate();		
	}

	/**
	 * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#endDisplayUpdate()
	 */
	public void endDisplayUpdate() {
		if(splash != null)
			splash.endDisplayUpdate();
	}

	/**
	 * @see gov.va.med.hds.cd.mum.boot.IStatusNotifier#setCancelMessage(java.lang.String)
	 */
	public void setCancelMessage(String message) {
		if (splash == null)
			cacheCommand(SET_CANCEL_MESSAGE, 0, message);
		else
			splash.setCancelMessage(message);
	}
}
