package gov.va.med.hds.cd.mum.boot.loader;

import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.Utils;

/**
 * This class provides a single source of class loader interaction that is independent of the
 * type of class loader being used.
 */
public class ClassLoaderManager {

    private static boolean forceSharedLoader = false;

    private static boolean useSharedLoader;
    private static SharedClassLoader sharedLoader;
    private static boolean initialize = true;

    /**
     * Forces the use of the shared class loader - used mainly in unit testing
     */
    public static void setForceSharedLoader(boolean force) {
        forceSharedLoader = force;
        initialize = true;
    }

    /**
     * Returns true if the shared class loader is being forced - used mainly in unit testing
     */
    public static boolean getForcedSharedLoader() {
        return forceSharedLoader;
    }

    /**
     * Resets the shared loader so that supsequent calls to getClassLoader will return a new 
     * instance of the shared loader, if the shared loader is being used - used mainly in unit 
     * testing 
     */
    public static void resetSharedLoader() {
        sharedLoader = null;
    }

    /**
     * This method must be called prior calls to getClassLoader, getDelegatingLoader and the
     * assignClassLoaders methods.  By looking at the command line arguments passed into the
     * application, this method determine if a single, shared class loader is to be used by all
     * plugins, or if each plugin is to receive it's own class loader. 
     */
    public static void init(String[] args) {
        if (initialize) {
            if (forceSharedLoader)
                useSharedLoader = true;
            else
                useSharedLoader = Utils.hasArgument(args, IBootConstants.LOAD_ALL_COMMAND_TAG);
            initialize = false;
            if (Utils.hasArgument(args, IBootConstants.LOAD_WEB_COMMAND_TAG))
                useSharedLoader = false;
            if (Utils.hasArgument(args, IBootConstants.TEST_TEXT_COMMAND_TAG)
                || Utils.hasArgument(args, IBootConstants.TEST_UI_COMMAND_TAG))
                useSharedLoader = true;
        }
    }

    private static void ensureInitialization() {
        if (initialize)
            throw new RuntimeException("Must call init before any other calls in ClassLoaderManager");
    }

    /**
     * Returns an instance of the classloader that is to be linked to each plugin.  This may be a
     * shared instance of a shared class loader, or a new instance of a plugin class loader.
     */
    public static IDesktopClassLoader getClassLoader(ClassLoader parent) {
        ensureInitialization();
        return getLoader(parent);
    }

    private static IDesktopClassLoader getLoader(ClassLoader parent) {
        if (useSharedLoader) {
            if (sharedLoader == null)
                sharedLoader = new SharedClassLoader(parent);
            return sharedLoader;
        } else {
            LinkedPluginLoader result = new LinkedPluginLoader(parent);
            if (IClassLoaderErrorListener.ENABLED)
                result.setErrorListener(SimpleClassLoaderErrorListener.getDefault());
            return result;
        }
    }

    /**
     * Once all the class loaders have been assigned to the plugins, this method must be called to
     * link class loaders together. 
     * @param startupLoader - the first class loader instantiated, used to load the mum.jar 
     * @param plugins - the list of plugins and thier class loaders
     * @throws ClassLoaderException
     */
    public static void assignClassLoaders(ClassLoader startupLoader, ListMap plugins)
        throws ClassLoaderException {
        ClassLoader parent = startupLoader.getParent();
        ensureInitialization();
        for (int i = 0; i < plugins.size(); i++) {
            IDesktopPlugin plugin = (IDesktopPlugin) plugins.get(i);
            if (plugin.getClassLoader() == null) {
                IDesktopClassLoader loader = getLoader(parent);
                loader.assignPlugin(plugin);
            }
        }
        if (!useSharedLoader) {
            for (int i = 0; i < plugins.size(); i++) {
                IDesktopPlugin plugin = (IDesktopPlugin) plugins.get(i);
                if (plugin.getClassLoader() instanceof ILinkedClassLoader)
                     ((ILinkedClassLoader) plugin.getClassLoader()).establishLink();
            }
        }
    }

	/**
     * Returns either a new instance of the delegating class loader, or the shared class loader.  
     * The delegating class loader can be used to provide a single class loader that can be passed
     * to third party libraries that delegates calls off to the various plugin class loaders as
     * needed. 
     */
    public static IDelegatingClassLoader getDelegatingLoader() {
		return getDelegatingLoader(ClassLoaderManager.class.getClassLoader());
	}

    /**
     * Returns an instance of the delegating class loader, with the specified parent loader.
     */
    public static IDelegatingClassLoader getDelegatingLoader(ClassLoader parent) {
        ensureInitialization();
        if (useSharedLoader) {
            if (sharedLoader == null)
                sharedLoader = new SharedClassLoader(parent);
            return sharedLoader;
        }
        return new DelegatingClassLoader(parent);
    }
}