package gov.va.med.hds.cd.mum.boot.loader;

import gov.va.med.hds.cd.mum.boot.IBootConstants;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

/**
 * Used internally by the LinkedPluginLoader
 */
public class FileLoader implements IFileLoader, IBootConstants {

    /**
     * Constructor for FileLoader.
     */
    protected FileLoader(File file) {
        super();
        this.file = file;
        String fName = file.getName().toLowerCase();
        isZip =
            (fName.endsWith(IBootConstants.JAR_EXT)
                || fName.endsWith(IBootConstants.ZIP_EXT));
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#getFile()
     */
    public File getFile() {
        return file;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#getData(java.lang.String)
     */
    public IDataBuffer getData(String name) throws IOException {
        ensureZipFile();
        if (zip != null) {
            ZipEntry entry = zip.getEntry(name);
            if (entry != null) {
                InputStream in = new BufferedInputStream(zip.getInputStream(entry));
                try {
                    out.reset();
                    int bytesRead;
                    while ((bytesRead = in.read(buffer)) != -1)
                        out.write(buffer, 0, bytesRead);
                } finally {
                    in.close();
                }
                return out;
            }
        }
        return null;
    }

    private void ensureZipFile() {
        if ((isZip) && (zip == null)) {
            try {
                zip = new ZipFile(file);
                if (buffer == null)
                    buffer = new byte[BUFFER_SIZE];
                if (out == null)
                    out = new DataBuffer(BUFFER_SIZE);
            } catch (ZipException e) {
                zip = null;
            } catch (IOException e) {
                zip = null;
            }
        }
        return;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#getZipFile()
     */
    public ZipFile getZipFile() {
        ensureZipFile();
        return zip;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#getAbsolutePath()
     */
    public String getAbsolutePath() {
        if (path == null)
            path = file.getAbsolutePath();
        return path;
    }

    private final File file;
    private URL url;
    private ZipFile zip;
    private final boolean isZip;
    private byte[] buffer;
    private DataBuffer out;
    private final int BUFFER_SIZE = 1024;
    private String path;

    private static class DataBuffer extends ByteArrayOutputStream implements IDataBuffer {

        /**
         * Constructor for DataBuffer.
         * @param size
         */
        public DataBuffer(int size) {
            super(size);
        }

        /**
         * Exposes the actual bytes (for efficiency purposes)
         * @return byte[]
         */
        public byte[] getBytes() {
            return buf;
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#isZipFile()
     */
    public boolean isZipFile() {
        return isZip;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#hasData(java.lang.String)
     */
    public boolean hasData(String name) {
        ensureZipFile();
        if (zip != null) {
            ZipEntry entry = zip.getEntry(name);
            return (entry != null);
        }
        return false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoader#getURL()
     */
    public URL getURL() {
        if (url == null) {
            try {
                url = file.toURL();
            } catch (MalformedURLException e) {
            }
        }
        return url;
    }
}
