package gov.va.med.hds.cd.mum.boot.loader;

import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Class Loader used for all plugins
 */
public class LinkedPluginLoader
    extends LinkedClassLoader
    implements IDesktopClassLoader, ILinkedClassLoader {

    private IDesktopPlugin plugin;
    private String sortKey;
    private String module;
    private SimpleList published;
    private SimpleList packageList;
    private boolean publishAllClasses;
    private IClassLoaderErrorListener errorListener;

    /**
     * Constructor for LinkedPluginLoader 
     */
    protected LinkedPluginLoader(ClassLoader parent) {
        super(parent);
        packageList = new SimpleList();
        errorListener = null;
    }

    /**
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    protected synchronized Class loadClass(String name, boolean resolve)
        throws ClassNotFoundException {
        try {
            return super.loadClass(name, resolve);
        } catch (ClassNotFoundException e) {
            if (IClassLoaderErrorListener.ENABLED && errorListener != null)
                errorListener.classNotFound(name, sortKey, module);
            throw e;
        }
    }

    private void addFile(File file, SimpleList packages) throws ClassLoaderException {
        if (file == null)
            return;
        if (file.getName().endsWith(IBootConstants.JAR_EXT)) {
            try {
                super.addURL(file.toURL());
            } catch (MalformedURLException e) {
                throw new ClassLoaderException(e);
            }
            packageList.addAll(packages);
            sortKey = file.getAbsolutePath();
            if (packages != null) {
                for (int i = 0; i < packages.size(); i++)
                    getLinks().add(packages.get(i), this);
            }
        } else
            assignOtherFile(file);
    }

    private void link(IDesktopPlugin plugin) {
        LinkedPluginLoader loader = (LinkedPluginLoader) plugin.getClassLoader();
        for (int i = 0; i < loader.packageList.size(); i++) {
            getLinks().add(loader.packageList.get(i), loader);
        }
        if (loader.getOtherFiles() != null) {
            for (int i = 0; i < loader.getOtherFiles().size(); i++)
                assignOtherFile((File) loader.getOtherFiles().get(i));
        }
    }

    /**
     * Adds a listener to detect class loading errors
     */
    protected void setErrorListener(IClassLoaderErrorListener errorListener) {
        this.errorListener = errorListener;
    }

	Class loadPluginClass(LinkedPluginLoader requestor, String className, boolean resolve)
		throws ClassNotFoundException {
		boolean loadClass = (requestor == this);
		if (!loadClass) {
			loadClass = publishAllClasses;
			if (!loadClass) {
				loadClass = this.module.equals(requestor.module);
				if (!loadClass)
					loadClass = published.contains(className);
			}
		}
		if (loadClass)
			return internalLoadClass(className, resolve);
		return null;
	}

    /**
     * @see java.lang.ClassLoader#findResource(java.lang.String)
     */
    public URL findResource(String name) {
        URL result = super.findResource(name);
        if ((result == null) && (IClassLoaderErrorListener.ENABLED) && (errorListener != null))
            errorListener.resourceNotFound(name, sortKey, module);
        return result;
    }

    URL findPluginResource(String name) {
        return internalFindResource(name);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopClassLoader#assignPlugin(gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin)
     */
    public void assignPlugin(IDesktopPlugin plugin) throws ClassLoaderException {
        this.plugin = plugin;
        this.module = plugin.getModule();
        this.published = plugin.getPublishedClasses();
        this.publishAllClasses = plugin.publishAllClasses();
        for (int i = 0; i < plugin.getFileCount(); i++)
            addFile(plugin.getPhysicalFile(i), plugin.getFilePackages(i));
        plugin.setClassLoader(this);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.ILinkedClassLoader#establishLink()
     */
    public void establishLink() {
        for (int i = 0; i < plugin.getLinkedPlugins().size(); i++) {
            link((IDesktopPlugin) plugin.getLinkedPlugins().get(i));
        }
        this.plugin = null;
        ensureLinkOrder();
    }

	LinkedPluginLoader getRequestor(Object obj) {
		return this;
	}

    LinkedPluginLoader getLoader(Object obj) {
        return (LinkedPluginLoader) obj;
    }

    public String getSortKey() {
        return sortKey;
    }

}
