package gov.va.med.hds.cd.mum.boot.loader;

import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.Utils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 * Shared class loader that loads all plugin classes  
 */
public class SharedClassLoader
    extends ClassLoader
    implements IDesktopClassLoader, IDelegatingClassLoader, IFileLoaderFactory, IBootConstants {

	/**
     * Constructor for SharedClassLoader 
     */
    protected SharedClassLoader(ClassLoader parent) {
        super(parent);
        factory = this;
    }

	/**
     * Constructor for SharedClassLoader 
     */
    protected SharedClassLoader(ClassLoader parent, IFileLoaderFactory factory) {
        super(parent);
        this.factory = factory;
    }

    /**
     * Adds a file to the list of files to load classes from
     * @param file
     * @throws FileNotFoundException
     */
    public void registerFile(File file) throws FileNotFoundException {
        validateFile(file);
        IFileLoader loader = factory.getNewLoader(file);
        if (loader.isZipFile()) {
            if (useMaps) {
                dumpSingleLoader();
                autoDetectPackages(loader);
            } else {
                singleLoader = loader;
                useMaps = true;
            }
        } else {
            if (otherFiles == null)
                otherFiles = new ArrayList();
            otherFiles.add(loader);
        }
    }

    private void validateFile(File file) throws FileNotFoundException {
        if (file == null)
            throw new NullPointerException();
        if ((!file.exists()) || (file.isDirectory()))
            throw new FileNotFoundException(file.getAbsolutePath());
    }

    /**
     * Adds a file to the list of files to load classes from
     * @param file
     * @param packages - list of packages to map the file to
     * @throws FileNotFoundException
     */
    public void registerFile(File file, List packages) throws FileNotFoundException {
        if (packages == null)
            registerFile(file);
        else {
            validateFile(file);
            dumpSingleLoader();
            if (packageMap == null)
                packageMap = new ListMap();
            IFileLoader loader = factory.getNewLoader(file);
            for (int i = 0; i < packages.size(); i++) {
                addPackage(((String) packages.get(i)).replace('.', '/'), loader);
            }
        }
    }

    private void addPackage(String key, IFileLoader loader) {
        if (key == null)
            return;
        List list = packageMap.add(key, loader);
        if (list.size() > 1) {
            if (comparer == null)
                comparer = new FileLoaderComparer();
            Collections.sort(list, comparer);
        }
    }

    private void dumpSingleLoader() {
        if (singleLoader != null) {
            if (packageMap == null)
                packageMap = new ListMap();
            autoDetectPackages(singleLoader);
            singleLoader = null;
        }
    }
    private void autoDetectPackages(IFileLoader loader) {
        ZipFile zip = loader.getZipFile();
        if (zip != null) {
            for (Enumeration enum = zip.entries(); enum.hasMoreElements();) {
                ZipEntry entry = (ZipEntry) enum.nextElement();
                String pkg = Utils.getPackageFromZipEntry(entry);
                if (pkg != null)
                    addPackage(pkg.replace('.', '/'), loader);
            }
        }
    }

    /**
     * @see java.lang.ClassLoader#findClass(java.lang.String)
     */
    protected Class findClass(final String name) throws ClassNotFoundException {
        String fileName = name.replace('.', '/').concat(IBootConstants.CLASS_EXT);
        List list = null;
        IFileLoader loader = null;
        Object obj = getLoaderOrList(fileName);
        if (obj instanceof IFileLoader)
            loader = (IFileLoader) obj;
        else
            list = (List) obj;
        if ((loader != null) || (list != null)) {
            int idx = 0;
            do {
                if (list != null) {
                    loader = (IFileLoader) list.get(idx);
                    idx++;
                }
                try {
                    IDataBuffer data = loader.getData(fileName);
                    if (data != null)
                        return defineClass(name, data.getBytes(), 0, data.size());
                } catch (IOException e) {
                }
            } while ((list != null) && (idx < list.size()));
        }
        throw new ClassNotFoundException(name);
    }

    private Object getLoaderOrList(final String name) {
        if (singleLoader == null) {
            IFileLoader loader = null;
            if (packageMap != null) {
                int idx = name.lastIndexOf('/');
                String pkg;
                if (idx > 0)
                    pkg = name.substring(0, idx);
                else
                    pkg = IBootConstants.DEFAULT_PACKAGE;
                List list = packageMap.get(pkg);
                if (list.size() == 1)
                    return list.get(0);
                else if (list.size() == 0)
                    return null;
                return list;
            }
            return loader;
        } else
            return singleLoader;
    }
    /**
     * @see java.lang.ClassLoader#findLibrary(java.lang.String)
     */
    protected String findLibrary(final String libname) {
        if (otherFiles != null) {
            String lib = System.mapLibraryName(libname);
            for (int i = 0; i < otherFiles.size(); i++) {
                File file = ((IFileLoader) otherFiles.get(i)).getFile();
                if (lib.equals(file.getName()))
                    return file.getAbsolutePath();
            }
        }
        return null;
    }

    /**
     * @see java.lang.ClassLoader#findResource(java.lang.String)
     */
    protected URL findResource(final String name) {
        List list = null;
        IFileLoader loader = null;
        Object obj = getLoaderOrList(name);
        if (obj instanceof IFileLoader)
            loader = (IFileLoader) obj;
        else
            list = (List) obj;
        if ((loader != null) || (list != null)) {
            int idx = 0;
            do {
                if (list != null) {
                    loader = (IFileLoader) list.get(idx);
                    idx++;
                }
                if (loader.hasData(name))
                    try {
                        return new URL(
                            IBootConstants.JAR_PROTOCOL
                                + loader.getFile().toURL().toString()
                                + "!/"
                                + name);
                    } catch (MalformedURLException e) {
                    }
            } while ((list != null) && (idx < list.size()));
        }
        if (otherFiles != null) {
            for (int i = 0; i < otherFiles.size(); i++) {
                loader = (IFileLoader) otherFiles.get(i);
                URL url = loader.getURL();
                if (name.equals(url.getPath()))
                    return url;
            }
        }
        return null;
    }

    /**
     * Closes all the jar file classes to close file locks
     */
    public void dispose() {
        close(singleLoader);
        if (packageMap != null) {
            for (int i = 0; i < packageMap.size(); i++) {
                close((IFileLoader) packageMap.get(i));
            }
        }
    }

    private void close(IFileLoader loader) {
        if ((loader != null) && (loader.isZipFile()) && (loader.getZipFile() != null))
            try {
                loader.getZipFile().close();
            } catch (IOException e) {
            }
    }

    private IFileLoader singleLoader;
    private IFileLoaderFactory factory;
    private ListMap packageMap;
    private List otherFiles;
    private boolean useMaps = false;
    private Comparator comparer;

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IFileLoaderFactory#getNewLoader(java.io.File)
     */
    public IFileLoader getNewLoader(File file) {
        return new FileLoader(file);
    }

    private static class FileLoaderComparer implements Comparator {

        public int compare(Object o1, Object o2) {
            return ((IFileLoader) o1).getAbsolutePath().compareTo(
                ((IFileLoader) o2).getAbsolutePath());
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopClassLoader#assignPlugin(gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin)
     */
    public void assignPlugin(IDesktopPlugin plugin) throws ClassLoaderException {
        if (plugin.getClassLoader() == null) {
            for (int i = 0; i < plugin.getFileCount(); i++)
                try {
                    File file = plugin.getPhysicalFile(i);
                    if (file != null) {
                        List packages;
                        if (plugin.getFilePackages(i) == null)
                            packages = null;
                        else
                            packages = plugin.getFilePackages(i).getList();
                        registerFile(file, packages);
                    }
                } catch (FileNotFoundException e) {
                    throw new ClassLoaderException(e);
                }
            plugin.setClassLoader(this);
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDelegatingClassLoader#addLoader(gov.va.med.hds.cd.mum.boot.loader.LinkedPluginLoader)
     */
    public void addLoader(LinkedPluginLoader loader) {
    }
}
