package gov.va.med.hds.cd.mum.boot.loader;

import java.io.File;

/**
 * Listner used to track down class loading issues 
 */
public class SimpleClassLoaderErrorListener
    implements IClassLoaderErrorListener {
        
    
    private static SimpleClassLoaderErrorListener DEFAULT =
        new SimpleClassLoaderErrorListener();
        
    public static IClassLoaderErrorListener getDefault() {
        return DEFAULT;
    }
    

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IClassLoaderErrorListener#classNotFound(java.lang.String, java.lang.String, java.lang.String)
     */
    public void classNotFound(
        String className,
        String requestorKey,
        String moduleName) {

        Throwable t = new Throwable();
        if (!shouldIgnore(IGNORED_CLASS_PATTERNS, className, t)) {
            // print a missing class message.
            System.err.println("\nClass Not Found: " + className);
            System.err.println("\tRequested By: " 
                + cleanupRequestorKey(requestorKey)
                + " (of module " + moduleName +")");
            printStackTrace(t);
        }
    }
    

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IClassLoaderErrorListener#resourceNotFound(java.lang.String, java.lang.String, java.lang.String)
     */
    public void resourceNotFound(
        String resourceName,
        String requestorKey,
        String moduleName) {
            
        Throwable t = new Throwable();
        if (!shouldIgnore(IGNORED_RESOURCE_PATTERNS, resourceName, t)) {
            // print a missing resource message.
            System.err.println("\nResource Not Found: " + resourceName);
            System.err.println("\tRequested By: " 
                + cleanupRequestorKey(requestorKey)
                + " (of module " + moduleName +")");
            printStackTrace(t);
        }
    }
    

    private String cleanupRequestorKey(String key) {
        if (key == null) return null;
        
        int end = key.lastIndexOf(File.separatorChar);
        if (end == -1) return key;
        
        int beg = key.lastIndexOf(File.separatorChar, end-1);
        if (beg == -1) return key;
        
        return key.substring(beg+1, end);
    }


    private static final String[][] IGNORED_CLASS_PATTERNS = {
        { null, "java.util.ResourceBundle", "getBundle" },
        { null, "gov.va.med.hds.cd.url.internal.ProviderExtension", "buildDynamicProvider" },
        { "BeanInfo", "java.beans.Introspector", "findExplicitBeanInfo" },
        { "Listener", "gov.va.med.hds.cd.workspace.internal.Multicaster", "getListenerTypes" },
        { "_en_US", null, null },
        { "a.b.c.NonExistantClass", null, null}
    };


    private static final String[][] IGNORED_RESOURCE_PATTERNS = {
        { "_en", "java.util.ResourceBundle", "getBundle" },
        { "_en", "javax.help.HelpUtilities", "getLocalizedResource" },
        { "log4j.xml", "org.apache.log4j.helpers.Loader", "getResource" },
        { "log4j.properties", "org.apache.log4j.helpers.Loader", "getResource" },
        { null, "gov.va.med.hds.cd.url.internal.ProviderExtension", "getClassPathProvider" }
    };
    
    
    private boolean shouldIgnore(String[][] patterns, String name, Throwable t) {
        for (int i = 0; i < patterns.length; i++) {
            if (patterns[i][0] != null && name.indexOf(patterns[i][0]) == -1)
                continue;
            if (stackContains(t, patterns[i][1], patterns[i][2]))
                return true;
        }
        return false;
    }
    
    
    private boolean stackContains(Throwable t, String className, String methodName) {
        StackTraceElement[] stack = t.getStackTrace();
        for (int i = 0; i < stack.length; i++) {
            StackTraceElement element = stack[i];
            if ((className == null || className.equals(element.getClassName()) &&
                (methodName == null || methodName.equals(element.getMethodName()))))
                return true;
        }
        return false;
    }
    
    
    private void printStackTrace(Throwable t) {
        StackTraceElement[] stack = t.getStackTrace();
        for (int i = 2; i < stack.length; i++) {            
            StackTraceElement element = stack[i];
            if (element.getClassName().startsWith("gov.va.med.hds.cd.mum.boot.loader"))
                continue;
            System.err.println("\t" + element);
        }
    }

}
