package gov.va.med.hds.cd.mum.boot.utils;

/**
 * A simple, fast xml parsing interface.  This interface relies on node id
 * values to locate data, eliminating the need for multiple objects represting
 * different nodes in the xml tree.
 */
public interface ITreeData {

	/**
	 * Returns an id value representing the root node
	 * @return int
	 */
	int getRootID();

	/**
	 * Returns the node's xml tag
	 * @param id
	 * @return String
	 */
	String getName(int id);

	/**
	 * Returns the number of attributes owned by the node
	 * @param id
	 * @return int
	 */
	int getAttributeCount(int id);

	/**
	 * Returns the name (xml tag) of the node's attribute
	 * @param id
	 * @param idx
	 * @return String
	 */
	String getAttributeName(int id, int idx);

	/**
	 * Returns the value of the node's attribute
	 * @param id
	 * @param idx
	 * @return String
	 */
	String getAttributeValue(int id, int idx);

	/**
	 * Returns the value of the node's attribute with the passed in attribute
	 * name
	 * @param id
	 * @param name
	 * @return String
	 */
	String getAttribute(int id, String name);

	/**
	 * Returns the value of the xml node.
	 * @param id
	 * @return String
	 */
	String getValue(int id);

	/**
	 * Returns the number of child nodes a given node contains
	 * @param id
	 * @return int
	 */
	int getChildCount(int id);

	/**
	 * Returns the id of a child node.  The id is needed to get the node's name,
	 * attributes and value.
	 * @param id
	 * @param idx
	 * @return int
	 */
	int getChildID(int id, int idx);

	/**
	 * Returns the id of the first child node of the indicated parent with the
	 * passed in name.  If no child is found, the returned id will be -1;
	 * @param id
	 * @param name
	 * @return int
	 */
	int findChildID(int id, String name);

	/**
	 * Returns the id of the child node of the indicated parent with the passed
	 * in name, starting after the passed in child.  If currentChild is -1,
	 * the first child with the passed in name is returned.  If no child is
	 * found, the returned id will be -1;
	 * @param id
	 * @param name
	 * @param currentChild
	 * @return int
	 */
	int findChildID(int id, String name, int currentChild);

	/**
	 * When set to true (the default), whitespace between nodes is ignored.
	 * @param value
	 */
	void ignoreWhitespace(boolean value);

	/**
	 * Resets the object as if no xml had been parsed.  load should call clear
	 * automatically.
	 */
	void clear();

	/**
	 * Adds a child node to the xml.  If parent is less than 0, the child will
	 * be the root child, unless one already exists, in which case an exception
	 * will be thrown.
	 * @param parent
	 * @return int
	 */
	int createChild(int parent, String name);

	/**
	 * Creates the root node.  If a root node already exists, an exception is
	 * thrown.
	 * @param name
	 * @return int
	 */
	int createRoot(String name);

	/**
	 * Deletes a child and removes if from the parent node
	 * @param id
	 */
	void removeChild(int parent, int child);

	/**
	 * Creates a child of the parent that is a duplicate of the from node.  All
	 * children nodes are copied as well.
	 * @param from
	 * @param to
	 * @return int
	 */
	int createChildCopy(int from, int parent);
	
	/**
	 * Creates a child of the parent that is a duplicate of the from node in the
	 * passed in ITreeData structure.  All children nodes are copied as well.
	 * @param data
	 * @param from
	 * @param parent
	 * @return int
	 */
	int createChildCopy(ITreeData data, int from, int parent);	

	/**
	 * Adds an attribute to the associated node
	 * @param id
	 * @param name
	 * @param value
	 */
	void setAttribute(int id, String name, String value);

	/**
	 * Sets the node's value
	 * @param id
	 * @param value
	 */
	void setValue(int id, String value);
	
	/**
	 * Returns true if the tree data has a root node.
	 */
	boolean hasRoot();
}