package gov.va.med.hds.cd.mum.boot.utils;

import gov.va.med.hds.cd.mum.boot.IBootConstants;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.HashMap;

/**
 * This class manages the settings values specified both in the config file and the optional
 * .settings file.
 */
public class Settings {

    ITreeData data;

    /**
     * Constructor for Settings 
     */
    public Settings(ITreeData xml) {
        super();
        try {
            data = TreeDataFactory.newSimpleXMLTreeData();
        } catch (TreeDataException e) {
        }
        int idx;
        if (xml == null)
            idx = -1;
        else
            idx = xml.findChildID(xml.getRootID(), IBootConstants.SETTINGS_ELEMENT);
        if (idx < 0)
            data.createRoot(IBootConstants.SETTINGS_ELEMENT);
        else
            data.createChildCopy(xml, idx, -1);
    }

    /**
     * Returns the ITreeData of the combined settings
     */
    public ITreeData data() {
        return data;
    }

    /**
     * Adds command line parameters to the settings if they have been mapped in the settings 
     * specifications.  Mapping occurs in the config file settings node, by adding a <tt>clparam</tt>
     * attribute to the XML that you want to allow overing by command line parameters.  This allows 
     * command line parameters with the same <tt>-param=</tt> param name to replace the node value.
     * To replace the value of an attribute, use the form <tt>attribute-clparam</tt> as the attribute name. 
     * For example, if the XML in the config file looked like this:
     * <p><tt><pre>
     *   &lt;settings&gt;
     *     &lt;server port="8000" clparam="server" port-clparam="port"&gt;localhost&lt;/server&gt;
     *   &lt;/settings&gt;</tt></pre>
     * <p>and the command line parameters <tt>-port=1234 -server=1.2.3.4</tt> were used, a call to this
     * method would result in the following XML: 
     * <p><tt><pre>
     *   &lt;settings&gt;
     *     &lt;server port="1234" clparam="server" port-clparam="port"&gt;1.2.3.4&lt;/server&gt;
     *   &lt;/settings&gt; </tt></pre>
     * @param args - the command line arguments to merge
     */
    public void mergeWithCommandLineParameters(String[] args) {
        if ((args != null) && (args.length > 0)) {
            int root = data.getRootID();
            applyCommandLineParamsToNode(root, args);
        }
    }

    /**
     * Reads name/value pair data from the reader (presumably from the .settings file) and merges 
     * values into the tree data if the names have been mapped in the settings specifications.
     * Mapping occurs in the config file settings node, by adding a <tt>setting</tt>
     * attribute to the XML that you want to allow overing by name/value pairs in the .settings file.
     * This allows settings with the same name to replace the node value. To replace the value of an 
     * attribute, use the form <tt>attribute-setting</tt> as the attribute name. 
     * For example, if the XML in the config file looked like this:
     * <p><tt><pre>
     *   &lt;settings&gt;
     *     &lt;server port="8000" setting="server" port-setting="port"&gt;localhost&lt;/server&gt;
     *   &lt;/settings&gt;</tt></pre>
     * <p>and the .settings file looked like this:
     * <p><tt><pre>
     *   Server = 1.2.3.4
     *   Port = 1234</tt></pre>
     * <p>a call to this method would result in the following XML: 
     * <p><tt><pre>
     *   &lt;settings&gt;
     *     &lt;server port="1234" setting="server" port-setting="port"&gt;1.2.3.4&lt;/server&gt;
     *   &lt;/settings&gt;</tt></pre>
     * @param input - the reader that contains the .settings name/value pair data
     */
    public void mergeWithSettings(Reader input) {
        if (input == null)
            return;
        HashMap map = new HashMap();
        BufferedReader reader = getBufferedReader(input);
        try {
            readNameValuePairs(map, reader);
        } finally {
            try {
                reader.close();
            } catch (IOException e) {
            }
        }
        applySettingsMap(data.getRootID(), map);
    }

    private void readNameValuePairs(HashMap map, BufferedReader reader) {
        String line;
        do {
            try {
                line = reader.readLine();
                if (line != null) {
                    int idx = line.indexOf(IBootConstants.ARG_ASSIGNMENT);
                    if (idx > 0) {
                        map.put(
                            line.substring(0, idx).trim().toLowerCase(),
                            line.substring(idx + 1).trim());
                    }
                }
            } catch (IOException e) {
                line = null;
            }
        } while (line != null);
    }

    private BufferedReader getBufferedReader(Reader input) {
        BufferedReader reader;
        if (input instanceof BufferedReader)
            reader = (BufferedReader) input;
        else
            reader = new BufferedReader(input);
        return reader;
    }

    private void applySettingsMap(int root, HashMap map) {
        int count = data.getAttributeCount(root);
        for (int i = 0; i < count; i++) {
            String name = data.getAttributeName(root, i);
            if (name.endsWith(IBootConstants.SETTING_ATTRIBUTE)) {
                String key = data.getAttributeValue(root, i);
                saveValue(
                    root,
                    name,
                    (String) map.get(key.trim().toLowerCase()),
                    IBootConstants.SETTING_ATTRIBUTE,
                    IBootConstants.SETTING_SUB_ATTRIBUTE);
            }
        }
        applySettingsToChildren(root, map);
    }

    private void applySettingsToChildren(int root, HashMap map) {
        int children = data.getChildCount(root);
        if (children > 0) {
            for (int i = 0; i < children; i++) {
                applySettingsMap(data.getChildID(root, i), map);
            }
        }
    }

    private void applyCommandLineParamsToNode(int root, String[] commands) {
        int count = data.getAttributeCount(root);
        for (int i = 0; i < count; i++) {
            String name = data.getAttributeName(root, i);
            if (name.endsWith(IBootConstants.COMMAND_LINE_PARAM_ATTRIBUTE)) {
                String key = data.getAttributeValue(root, i);
                String[] cmds = key.split(",");
                for (int j = 0; j < cmds.length; j++) {
                    String prefix =
                        IBootConstants.OPTION_PREFIX
                            + cmds[j]
                            + IBootConstants.ARG_ASSIGNMENT;
                    int arg = Utils.getArgumentIndex(commands, prefix);
                    if (arg >= 0) {
                        saveValue(
                            root,
                            name,
                            Utils.getArgumentValue(commands[arg], prefix),
                            IBootConstants.COMMAND_LINE_PARAM_ATTRIBUTE,
                            IBootConstants.COMMAND_LINE_PARAM_SUB_ATTRIBUTE);
                    }
                }
            }
        }
        applyCommandLineParamsToChildren(root, commands);
    }

    private void saveValue(
        int root,
        String name,
        String value,
        String attribTag,
        String attribSubTag) {
        if (value != null) {
            if (name.equals(attribTag)) {
                data.setValue(root, value);
            } else if (name.endsWith(attribSubTag)) {
                String attribName = name.substring(0, name.length() - attribSubTag.length());
                data.setAttribute(root, attribName, value);
            }
        }
    }

    private void applyCommandLineParamsToChildren(int root, String[] commands) {
        int children = data.getChildCount(root);
        if (children > 0) {
            for (int i = 0; i < children; i++) {
                applyCommandLineParamsToNode(data.getChildID(root, i), commands);
            }
        }
    }

}
