package gov.va.med.hds.cd.mum.boot.utils;

import gov.va.med.hds.cd.mum.boot.IBootConstants;

import java.util.ArrayList;
import java.util.List;

/**
 * A simple, fast xml parsing class.  This class relies on node id values to
 * locate data, eliminating the need for multiple objects represting different
 * nodes in the xml tree.
 */
public abstract class TreeData implements ITreeData, IBootConstants {

    protected final List data = new ArrayList();
    protected final IntListCache children = new IntListCache();
    protected final IntList elementIndex = new IntList();
    protected final IntList dataIndex = new IntList();
    protected final IntListCache attribs = new IntListCache();
    private boolean ignoreWhitespace = true;
	protected final BufList strbuf = new BufList();
    protected final IntList current = new IntList();

    public TreeData() {
        super();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#clear()
     */
    public void clear() {
        data.clear();
        children.clear();
        elementIndex.clear();
        dataIndex.clear();
        attribs.clear();
		current.clear();
		strbuf.clear();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getRootID()
     */
    public int getRootID() {
        return elementIndex.get(0);
    }
    
    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#hasRoot()
     */
    public boolean hasRoot() {
        return (elementIndex.size() > 0);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getName(int)
     */
    public String getName(int id) {
        return (String) data.get(elementIndex.get(id));
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getAttributeCount(int)
     */
    public int getAttributeCount(int id) {
        IntList attrList = attribs.getExisting(id); 
        return (attrList == null ? 0 : attrList.size());
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getAttributeName(int, int)
     */
    public String getAttributeName(int id, int idx) {
        IntList attrList = attribs.getExisting(id);
        if (attrList == null)
            throw new IndexOutOfBoundsException("Node has no attributes"); 
        return (String) data.get(attrList.get(idx));
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getAttributeValue(int, int)
     */
    public String getAttributeValue(int id, int idx) {
        IntList attrList = attribs.getExisting(id);
        if (attrList == null)
            throw new IndexOutOfBoundsException("Node has no attributes"); 
        return (String) data.get(attrList.get(idx) + 1);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getAttribute(int, java.lang.String)
     */
    public String getAttribute(int id, String name) {
        IntList attrList = attribs.getExisting(id);
        if (attrList != null)
    		for (int i = attrList.size();   i-- > 0;  ) {
                if (name.equals(data.get(attrList.get(i))))
                    return (String) data.get(attrList.get(i) + 1);
            }
        return null;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getValue(int)
     */
    public String getValue(int id) {
        if (dataIndex.get(id) < 0)
            return null;
        else
            return (String) data.get(dataIndex.get(id));
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getChildCount(int)
     */
    public int getChildCount(int id) {
        IntList kids = children.getExisting(id);
        if (kids == null)
            return 0;
        else
            return kids.size();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#getChildID(int, int)
     */
    public int getChildID(int id, int idx) {
        IntList kids = children.getExisting(id);
        if (kids == null)
            throw new IndexOutOfBoundsException("Node has no children"); 
        return kids.get(idx);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#findChildID(int, java.lang.String)
     */
    public int findChildID(int id, String name) {
        return findChildID(id, name, -1);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#findChildID(int, java.lang.String, int)
     */
    public int findChildID(int id, String name, int currentChild) {
        IntList kids = children.getExisting(id);
        if (kids != null) {
            boolean found = (currentChild < 0);
            for (int i = 0;   i < kids.size();   i++) {
                int child = kids.get(i);
                if (found) {
                    if (name.equals(getName(child)))
                        return child;
                } else {
                    found = (child == currentChild);
                }
            }
        }
        return -1;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#ignoreWhitespace(boolean)
     */
    public void ignoreWhitespace(boolean value) {
        ignoreWhitespace = value;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#createChild(int, java.lang.String)
     */
    public int createChild(int parent, String name) {
        validateTag(name);
        int idx = elementIndex.size();
        if (parent >= 0)
            children.get(parent).add(idx);
        else if (idx > 0)
            throw new RuntimeException(
                "XML Data already has a root element",
                new TreeDataException());
        attribs.clear(idx);
        children.clear(idx);
        dataIndex.set(idx, -1);
        elementIndex.add(data.size());
        data.add(name);
        return idx;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#createRoot(java.lang.String)
     */
    public int createRoot(String name) {
        return createChild(-1, name);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#createChildCopy(int, int)
     */
    public int createChildCopy(int from, int parent) {
        return createChildCopy(this, from, parent);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#createChildCopy(gov.va.med.hds.cd.mum.boot.utils.ITreeData, int, int)
     */
    public int createChildCopy(ITreeData data, int from, int parent) {
        int result = createChild(parent, data.getName(from));
        setValue(result, data.getValue(from));
        for (int i = 0; i < data.getAttributeCount(from); i++)
            setAttribute(result, data.getAttributeName(from, i), data.getAttributeValue(from, i));
        for (int i = 0; i < data.getChildCount(from); i++)
            createChildCopy(data, data.getChildID(from, i), result);
        return result;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#removeChild(int, int)
     */
    public void removeChild(int parent, int child) {
        if (parent < 0)
            clear();
        else {
            IntList list = children.getExisting(parent);
            if (list != null) {
                for (int i = 0; i < list.size(); i++) {
                    if (list.get(i) == child) {
                        list.remove(i);
                        return;
                    }
                }
            }
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#setAttribute(int, java.lang.String, java.lang.String)
     */
    public void setAttribute(int id, String name, String value) {
        validateTag(name);
        IntList attrList = attribs.get(id);
		for (int i = 0; i < attrList.size(); i++) {
            int idx = attrList.get(i);
            if (name.equals(data.get(idx))) {
                if (value == null) {
                    attrList.remove(i);
                    data.set(idx, null);
                    data.set(idx + 1, null);
                } else {
                    data.set(idx + 1, value);
                }
                return;
            }
        }
        if (value != null) {
            attrList.add(data.size());
            data.add(name);
            data.add(value);
        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.utils.ITreeData#setValue(int, java.lang.String)
     */
    public void setValue(int id, String value) {
        if (dataIndex.get(id) < 0) {
            if (value != null) {
                dataIndex.set(id, data.size());
                data.add(value);
            }
        } else {
            data.set(dataIndex.get(id), value);
        }
    }

    protected void error() throws TreeDataException {
        throw new TreeDataException("Invalid Tree Data Format");
    }

    protected boolean getIgnoreWhitespace() {
        return ignoreWhitespace;
    }

    private void validateTag(String name) {
        if (!Utils.validTag(name))
            badName();
    }

    private void badName() {
        throw new RuntimeException(INVALID_TAG_NAME_ERROR);
    }

    protected static class IntListCache {

        public IntListCache() {
            this(10);
        }

        public IntListCache(int initialCapacity) {
            super();
            if (initialCapacity < 0)
                throw new IllegalArgumentException("Illegal Capacity: " + initialCapacity);
            data = new IntList[initialCapacity];
        }

        public void clear() {
            for (int i = 0; i < size; i++) {
                if (data[i] != null)
                    data[i].clear();
            }
        }

        public final IntList getExisting(int index) {
            return (index < size ? data[index] : null);
        }

        public IntList get(int index) {
            ensureCapacity(index + 1);
            if (data[index] == null)
                data[index] = new IntList();
            if (size <= index)
                size = index + 1;
            return data[index];
        }

        private transient IntList data[];
        private int size;
        private void ensureCapacity(int minCapacity) {
            int oldCapacity = data.length;
            if (minCapacity > oldCapacity) {
                IntList oldData[] = data;
                int newCapacity = (oldCapacity * 3) / 2 + 1;
                if (newCapacity < minCapacity)
                    newCapacity = minCapacity;
                data = new IntList[newCapacity];
                System.arraycopy(oldData, 0, data, 0, size);
            }
        }

        public int size() {
            return size;
        }

        public void clear(int index) {
            ensureCapacity(index + 1);
            if (data[index] != null)
                data[index].clear();
        }

    }

    public static class BufList {

        public BufList() {
            this(4);
        }

        public BufList(int initialCapacity) {
            super();
            if (initialCapacity < 0)
                throw new IllegalArgumentException("Illegal Capacity: " + initialCapacity);
            data = new StringBuffer[initialCapacity];
        }

        public void clear() {
            for (int i = 0; i < size; i++)
                if (data[i] != null)
                    data[i].setLength(0);
        }

        public StringBuffer get(int index) {
            rangeCheck(index);
            return data[index];
        }

        public StringBuffer ensureBuffer(int index) {
            ensureCapacity(index + 1);
            if (data[index] == null)
                data[index] = new StringBuffer();
            if (size <= index)
                size = index + 1;
            return data[index];
        }

        private transient StringBuffer data[];
        private int size;
        private void ensureCapacity(int minCapacity) {
            int oldCapacity = data.length;
            if (minCapacity > oldCapacity) {
                StringBuffer oldData[] = data;
                int newCapacity = (oldCapacity * 3) / 2 + 1;
                if (newCapacity < minCapacity)
                    newCapacity = minCapacity;
                data = new StringBuffer[newCapacity];
                System.arraycopy(oldData, 0, data, 0, size);
            }
        }

        public int size() {
            return size;
        }

        public boolean hasData(int index) {
            ensureCapacity(index + 1);
            return ((data[index] != null) && (data[index].length() > 0));
        }

        private void rangeCheck(int index) {
            if ((index < 0) || (index >= size)) {
                throw new IndexOutOfBoundsException("Index: " + index + ", Size: " + size);
            }
        }
    }
}
