package gov.va.med.hds.cd.mum.boot.utils;

import gov.va.med.hds.cd.mum.boot.IBootConstants;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

/**
 * Contains general utility functions used in the boot packages
 */
public class Utils implements IBootConstants {

    private static final String META_DIR;
    private static final String[] encodedChars = { "&apos;", "&quot;", "&gt;", "&lt;", "&amp;" };
    private static final String[] decodedChars = { "'", "\"", ">", "<", "&" };
    private static final int[] encodedCharLength = new int[encodedChars.length];

    static {
        int idx = JarFile.MANIFEST_NAME.lastIndexOf('/');
        META_DIR = JarFile.MANIFEST_NAME.substring(0, idx);

        for (int i = 0; i < encodedCharLength.length; i++)
            encodedCharLength[i] = encodedChars[i].length();
    }

    /**
     * Returns the pacakge name from an entry in the zip file, based on it's directory
     */
    public static String getPackageFromZipEntry(ZipEntry entry) {
        String name = entry.getName();
        int idx = name.lastIndexOf('/');
        String pkg;
        if (idx > 0)
            pkg = name.substring(0, idx);
        else
            pkg = DEFAULT_PACKAGE;
        if (!META_DIR.equals(pkg))
            return pkg.replace('/', '.');
        return null;
    }

    /**
     * Populates a list with the packages contained in a zip file, based on the zip directories
     * @param zip
     * @param packages
     */
    public static void getPackages(ZipFile zip, SimpleList packages) {
        for (Enumeration enum = zip.entries(); enum.hasMoreElements();) {
            packages.add(getPackageFromZipEntry((ZipEntry) enum.nextElement()));
        }
    }

    private static final String numberChars = "0123456789";
    private static final String validExistingChars = numberChars + "_;&-";
    private static final String validAddingChars = numberChars + "_-";

    /**
     * Returns true if the passed in character can be used as an XML tag character
     * @param ch
     */
    public static boolean validTagChar(char ch) {
        return validTagChar(ch, validExistingChars);
    }

    /**
     * Returns true if the passed in character can be used as an XML tag character.
     * @param ch
     * @param validChars - optional list of characters to search other than alphabetical
     */
    public static boolean validTagChar(char ch, String validChars) {
        if ((ch >= 'a') && (ch <= 'z'))
            return true;
        if ((ch >= 'A') && (ch <= 'Z'))
            return true;
        return (validChars.indexOf(ch) >= 0);
    }

    /**
     * Returns true if the passed in string can be used as an XML tag
     * @param name
     */
    public static boolean validTag(String name) {
        if (name.length() > 0) {
            if (numberChars.indexOf(name.charAt(0)) >= 0)
                return false;
            for (int i = 0; i < name.length(); i++) {
                if (!validTagChar(name.charAt(i), validAddingChars))
                    return false;
            }
        } else
            return false;
        return true;
    }

    /**
     * Converts XML encoded strings as their actual characters
     * @param temp
     */
    public static void decodeData(StringBuffer temp) {
        int idx = 0;
        for (int i = 0; i < encodedChars.length; i++) {
            do {
                idx = temp.indexOf(encodedChars[i], idx);
                if (idx >= 0) {
                    temp.replace(idx, idx + encodedCharLength[i], decodedChars[i]);
                    idx -= encodedCharLength[i] - 1;
                }
            } while (idx >= 0);
        }
    }

    /**
     * Encodes prohibited XML characters into thier encoded forms
     * @param temp
     */
    public static void encodedData(StringBuffer temp) {
        int idx = 0;
        for (int i = decodedChars.length - 1; i >= 0; i--) {
            do {
                idx = temp.indexOf(decodedChars[i], idx);
                if (idx >= 0) {
                    temp.replace(idx, idx + 1, encodedChars[i]);
                    idx += encodedCharLength[i];
                }
            } while (idx >= 0);
        }
    }

    /**
     * Populates a list with the packages contained in a zip file, based on the zip directories
     * @param zipFile
     * @param packages
     */
    public static void getPackages(File zipFile, SimpleList packages) {
        try {
            ZipFile zip = new ZipFile(zipFile);
            try {
                getPackages(zip, packages);
            } finally {
                zip.close();
            }
        } catch (ZipException e) {
        } catch (IOException e) {
        }
    }

    /**
     * Returns true if the passed in argument list contains an argument with the appropriate tag.
     * @param args
     * @param tag
     */
    public static boolean hasArgument(String[] args, String tag) {
        return (getArgumentIndex(args, tag) >= 0);
    }

    /**
     * Returns the value of a command line argument
     * @param args
     * @param tag
     * @return String
     */
    public static String getArgument(String[] args, String tag) {
        int idx = getArgumentIndex(args, tag);
        if (idx < 0)
            return null;
        else
            return getArgumentValue(args[idx], tag);
    }

    /**
     * Returns the argument value
     * @param arg
     * @param tag
     * @return String
     */
    public static String getArgumentValue(String arg, String tag) {
        if (tag == null)
            return arg;
        else
            return arg.substring(tag.length());
    }
    /**
     * Returns the index of the passed in argument.
     * @param args - the list of arguments to search
     * @param tag - the command line option to search for.  If null, the first
     * option that does not start with a "-" will be returned
     * @return int - the index into the arguments that contains the desited
     * command line option, or -1 if the option was not found.
     */
    public static int getArgumentIndex(String[] args, String tag) {
        if ((args != null) && (args.length > 0)) {
            boolean isOption = (tag != null);
            boolean hasValue;
            String lcTag;
            if (isOption) {
                lcTag = tag.toLowerCase();
                hasValue = tag.endsWith(ARG_ASSIGNMENT);
            } else {
                lcTag = null;
                hasValue = false;
            }
            for (int i = 0; i < args.length; i++) {
                if (args[i] != null) {
                    if (isOption) {
                        String arg = args[i].toLowerCase();
                        if (hasValue) {
                            if (arg.startsWith(lcTag))
                                return i;
                        } else {
                            if (arg.equals(lcTag))
                                return i;
                        }
                    } else if (!args[i].startsWith(OPTION_PREFIX))
                        return i;
                }
            }
        }
        return -1;
    }

    /**
     * Returns a list of arguments without the argument with the passed in index
     * @param args
     * @param index
     * @return String[]
     */
    public static String[] extractArguemnt(String[] args, int index) {
        String[] result = new String[args.length - 1];
        int idx = 0;
        for (int i = 0; i < args.length; i++) {
            if (i != index) {
                result[idx] = args[i];
                idx++;
            }
        }
        return result;
    }

    /**
     * Returns a list of arguments with the new argument inserted into the
     * desired position
     * @param args
     * @param newArgument
     * @param index
     * @return String[]
     */
    public static String[] insertArgument(String[] args, String newArgument, int index) {
        int len;
        if (args == null)
            len = 1;
        else
            len = args.length + 1;
        String[] result = new String[len];
        int idx = 0;
        for (int i = 0; i < len; i++) {
            if (i == index) {
                result[i] = newArgument;
            } else {
                result[i] = args[idx];
                idx++;
            }
        }
        return result;
    }

    /**
     * Returns an array of arguments with the new argument inserted into the
     * desired position.
     * @param args
     * @param newArgument
     * @param index
     * @return String[]
     */
    public static String[] insertArgument(List args, String newArgument, int index) {
        int len;
        if (args == null)
            len = 1;
        else
            len = args.size() + 1;
        String[] result = new String[len];
        int idx = 0;
        for (int i = 0; i < len; i++) {
            if (i == index) {
                result[i] = newArgument;
            } else {
                result[i] = (String) args.get(idx);
                idx++;
            }
        }
        return result;
    }

    /**
     * Returns the physical directory where a class is loaded from.
     * @param resource
     */
    public static String getPhysicalDirectory(Class resource) {
        String name = resource.getName().replace('.', '/') + IBootConstants.CLASS_EXT;
        StringBuffer temp = null;
        URL url = resource.getClassLoader().getResource(name);
        URL url2;
        try {
            url2 = new URL(url.getPath()); // strip secondary protocol
        } catch (MalformedURLException e) {
            url2 = url;
        }
        try {
            temp = new StringBuffer(URLDecoder.decode(url2.getPath(), "UTF-8"));
        } catch (UnsupportedEncodingException e) {
        }
        int idx = temp.indexOf("!");
        if (idx >= 0)
            temp.setLength(idx);
        idx = temp.lastIndexOf("/");
        if (idx >= 0)
            temp.setLength(idx);
        return temp.toString();
    }

}
