package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IApplication;
import gov.va.med.hds.cd.mum.IPluginSource;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IStartupInfo;
import gov.va.med.hds.cd.mum.boot.loader.ClassLoaderManager;
import gov.va.med.hds.cd.mum.boot.loader.IDesktopClassLoader;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Class used to launch the application after all MUM activities have completed.
 */
public class ApplicationLauncher
    implements IApplicationLauncher, IUpdateConstants, IBootConstants {

    public ApplicationLauncher() {
        super();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IApplicationLauncher#launchApplication(java.net.URL, gov.va.med.hds.cd.mum.boot.utils.ListMap, gov.va.med.hds.cd.mum.boot.IStartupInfo, gov.va.med.hds.cd.mum.update.IPluginSpec)
     */
    public void launchApplication(
        URL configFile,
        ListMap specs,
        IStartupInfo info,
        IPluginSpec startupSpec)
        throws Throwable {
        assignPluginDependencies(specs, info);
        ClassLoaderManager.init(info.getArgs());
        if (info.getLoader() instanceof IDesktopClassLoader)
             ((IDesktopClassLoader) info.getLoader()).assignPlugin(startupSpec);
        ClassLoaderManager.assignClassLoaders(info.getLoader(), specs);
        if (runTests(info)) {
            releaseSingletons();
            executeTests(specs, info);
        } else {
            IApplication app = getAppObject(specs);
            Startup.notify(info, LAUNCHING_APP_MESSAGE);
            getSettingsXML(configFile, info);
            IPluginSource[] source = getPluginSource(specs);
            releaseSingletons();
            ITreeData settings = null;
            if (info.getSettings() != null)
                settings = info.getSettings().data();
            app.run(info.getArgs(), settings, source, info.getSplashScreen());
        }
    }

    private void releaseSingletons() {
        TreeDataFactory.releaseSingleton();
    }

    private boolean runTests(IStartupInfo info) {
        if (getTestClass(TEST_TEXT_COMMAND_TAG, info) != null)
            return true;
        else if (getTestClass(TEST_UI_COMMAND_TAG, info) != null)
            return true;
        else
            return false;
    }

    private String getTestClass(String tag, IStartupInfo info) {
        if (info.getArgs() != null) {
            for (int i = 0; i < info.getArgs().length; i++) {
                String arg = info.getArgs()[i];
                if ((arg != null) && (arg.startsWith(tag)))
                    return info.getArgs()[i].substring(tag.length());
            }
        }
        return null;
    }

    private void executeTests(ListMap specs, IStartupInfo info) throws TestFailureException {
        String junitClassName = JUNIT_TEXT_TEST_CLASS;
        String testClassName = getTestClass(TEST_TEXT_COMMAND_TAG, info);
        int cnt = 1;
        if (testClassName == null) {
            junitClassName = JUNIT_UI_TEST_CLASS;
            testClassName = getTestClass(TEST_UI_COMMAND_TAG, info);
            cnt++;
        }
        for (int i = 0; i < specs.size(); i++) {
            IPluginSpec spec = (IPluginSpec) specs.get(i);
            if (spec.getID().toLowerCase().equals(JUNIT)) {
                try {
                    Class junit = spec.getClassLoader().loadClass(junitClassName);
                    String[] args = new String[cnt];
                    args[0] = testClassName;
                    if (cnt > 1)
                        args[1] = "-noloading";
                    Method main = junit.getMethod(MAIN_METHOD, new Class[] { args.getClass()});
                    if (info.getSplashScreen() != null)
                        info.getSplashScreen().close();
                    ClassLoaderManager.resetSharedLoader();
                    main.invoke(null, new Object[] { args });
                } catch (ClassNotFoundException e) {
                    throw new TestFailureException(JUNIT_RUN_ERROR, e);
                } catch (NoSuchMethodException e) {
                    throw new TestFailureException(JUNIT_RUN_ERROR, e);
                } catch (IllegalAccessException e) {
                    throw new TestFailureException(JUNIT_RUN_ERROR, e);
                } catch (InvocationTargetException e) {
                    throw new TestFailureException(JUNIT_RUN_ERROR, e);
                }
                return;
            }
        }
        throw new IllegalArgumentException(JUNIT_PLUGIN_NOT_FOUND);
    }

    private void getSettingsXML(URL configFile, IStartupInfo info) throws MalformedURLException {
        if ((info.getSettings() != null) && (configFile != null)) {
            String path = configFile.toString();
            if (path.toLowerCase().endsWith(IBootConstants.XML_EXT)) {
                path =
                    path.substring(0, path.length() - IBootConstants.XML_EXT.length()).concat(
                        IBootConstants.SETTINGS_EXT);
                URL settingsFile = MUMUtils.newURL(path);
                Reader reader;
                try {
                    reader = new InputStreamReader(settingsFile.openConnection().getInputStream());
                } catch (IOException e) {
                    reader = null;
                }
                info.getSettings().mergeWithSettings(reader);
                info.getSettings().mergeWithCommandLineParameters(info.getArgs());
            }
        }
    }

    private IPluginSource[] getPluginSource(ListMap specs) {
        IPluginSource[] source = new IPluginSource[specs.size()];
        for (int i = 0; i < specs.size(); i++)
            source[i] = PluginClassFactory.newPluginSource((IPluginSpec) specs.get(i));
        return source;
    }

    /**
     * Assigns dependencies to the plugin specifications.  Used internally, made
     * public here for testing purposes only.
     */
    public void assignPluginDependencies(ListMap specs, IStartupInfo info)
        throws InvalidPluginException, MalformedURLException {
        for (int i = 0; i < specs.size(); i++) {
            IPluginSpec spec = (IPluginSpec) specs.get(i);
            spec.getFiles(info.getRootDirectory());
            spec.loadManifest(spec.getManifestURL());
        }
    }

    private IApplication getAppObject(ListMap specs)
        throws ClassNotFoundException, InstantiationException, IllegalAccessException {
        for (int i = 0; i < specs.size(); i++) {
            IPluginSpec spec = (IPluginSpec) specs.get(i);
            if ((spec.getApplicationClassName() != null)
                && (spec.getApplicationClassName().length() > 0)) {
                Class appClass =
                    ((ClassLoader) spec.getClassLoader()).loadClass(spec.getApplicationClassName());
                return (IApplication) appClass.newInstance();
            }
        }
        return null;
    }

}
