package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.Settings;
import gov.va.med.hds.cd.mum.utils.ChecksumCalculator;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

/**
 * Gets plugin information from the configuration file
 */
public class ConfigReader extends ConfigParser implements IConfigReader {

    public static final boolean ENABLE_CRC_CHECKING = true;

    /**
     * Constructor for ConfigReader 
     */
    protected ConfigReader() {
        super();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IConfigReader#getCRC()
     */
    public long getCRC() {
        return crc;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IConfigReader#readConfiguration(java.net.URL, gov.va.med.hds.cd.mum.boot.utils.ListMap)
     */
    public void readConfiguration(URL configFile, ListMap plugins)
        throws InvalidConfigurationException {
        if ((configFile == null) || (plugins == null))
            throw new NullPointerException();
        reset();
        ByteArrayInputStream input = getInputBuffer(configFile);
        ITreeData xml = getDocument(input);
        int root = parseRoot(xml);
        parseStartup(xml, root);
        parseChildren(plugins, xml, root);
        ChecksumCalculator calc = new ChecksumCalculator();
        input.reset();
        calc.calcConfigFile(input);
        if (ENABLE_CRC_CHECKING) {
            if (calc.getValue() != crc) {
//                				badConfig(INCORRET_CONFIGURATION_CRC+ " ("+ Long.toString(calc.getValue()));
                badConfig(INCORRET_CONFIGURATION_CRC);
            }
        }
        input.reset();
        settings = new Settings(xml);
    }

    private void reset() {
        startupPlugin = null;
        startupPluginName = null;
        crc = 0;
    }

    private void parseStartup(ITreeData xml, int root) throws InvalidConfigurationException {
        startupPluginName = null;
        int startup = xml.findChildID(root, IMUMConstants.STARTUP_ELEMENT);
        if (startup >= 0) {
            startupPluginName = xml.getAttribute(startup, IPluginConstants.PLUGIN_ELEMENT);
            if (xml.findChildID(root, IMUMConstants.STARTUP_ELEMENT, startup) >= 0)
                badConfig(IMUMConstants.MULTIPLE_STARTUP_CLASSES);
        }
    }

    private void parseChildren(ListMap plugins, ITreeData xml, int root)
        throws InvalidConfigurationException {
        verifyValidSettings(xml);
        int appCount = 0;
        for (int i = 0; i < xml.getChildCount(root); i++) {
            int child = xml.getChildID(root, i);
            if (IPluginConstants.PLUGIN_ELEMENT.equals(xml.getName(child))) {
                IPluginSpec plugin = PluginClassFactory.newPluginSpec(xml, child);
                if (plugin.getApplicationClassName() != null)
                    appCount++;
                plugins.add(plugin.getID(), plugin);
                if (plugin.getDirectory().equals(startupPluginName)) {
                    if (startupPlugin == null)
                        startupPlugin = plugin;
                    else
                        badConfig(IMUMConstants.MULTIPLE_STARTUP_CLASSES);
                }
            }
        }
        if (startupPlugin == null)
            badConfig(IMUMConstants.MISSING_STARTUP_PLUGIN_TXT);
        if (appCount == 0)
            badConfig(IMUMConstants.MISSING_APPLICATION_PLUGINS_TXT);
        if (appCount > 1)
            badConfig(IMUMConstants.MULTIPLE_APPLICATION_CLASSES);
    }

    private void verifyValidSettings(ITreeData xml)
        throws InvalidConfigurationException {
        validateSettings(xml);
    }

    private int parseRoot(ITreeData xml) throws InvalidConfigurationException {
        int root = xml.getRootID();
        if (root < 0)
            badConfig(NO_CONFIG_ROOT);
        if (!IMUMConstants.CONFIGURATION_ELEMENT.equals(xml.getName(root)))
            badConfig(NO_CONFIG_ROOT);
        String crcValue = xml.getAttribute(root, IMUMConstants.CRC_ATTRIBUTE);
        if ((crcValue == null) || (crcValue.length() == 0))
            badConfig(MISSING_CONFIG_CRC);
        try {
            crc = Long.parseLong(crcValue);
        } catch (NumberFormatException e) {
            badConfig(MISSING_CONFIG_CRC, e);
        }
        if (crc == 0)
            badConfig(MISSING_CONFIG_CRC);
        return root;
    }

    private ByteArrayInputStream getInputBuffer(URL configFile)
        throws InvalidConfigurationException {
        ByteArrayInputStream input = null;
        try {
            input =
                new ByteArrayInputStream(
                    convertStreamToByteArray(configFile.openConnection().getInputStream()));
        } catch (IOException e) {
            badConfig(e);
        }
        return input;
    }

    private long crc;

    private byte[] convertStreamToByteArray(InputStream in) throws IOException {
        byte[] result = null;
        try {
            int bufSize = 1024;
            ByteArrayOutputStream slurpBuffer = new ByteArrayOutputStream();
            byte[] buffer = new byte[bufSize];
            int bytesRead;
            while ((bytesRead = in.read(buffer)) != -1)
                slurpBuffer.write(buffer, 0, bytesRead);
            result = slurpBuffer.toByteArray();
        } finally {
            try {
                in.close();
            } catch (IOException ioe) {
            }
        }
        return result;
    }

    private IPluginSpec startupPlugin;
    private String startupPluginName;
    private Settings settings;

    /**
     * @see gov.va.med.hds.cd.mum.update.IConfigReader#getStartupPlugin()
     */
    public IPluginSpec getStartupPlugin() {
        return startupPlugin;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IConfigReader#getSettings()
     */
    public Settings getSettings() {
        return settings;
    }
}
