package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.utils.ChecksumCalculator;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class representing an plugin already installed on the workstation.
 */
public class ExistingPlugin implements IExistingPlugin, IUpdateConstants {

	/**
     * Constructor for ExistingPlugin 
	 * @param pluginDir - the directory where the plugin is installed.
	 * @throws FileNotFoundException
	 * @throws InvalidPluginDirectoryException
	 */
	protected ExistingPlugin(File pluginDir)
		throws FileNotFoundException, InvalidPluginDirectoryException {
		super();
		if (pluginDir == null)
			badDir(null);
		if (!pluginDir.exists())
			throw new FileNotFoundException();
		if (!pluginDir.isDirectory())
			badDir(FILE_NOT_DIRECTORY);
		parse(pluginDir);
		getFiles(pluginDir);
	}

	private void parse(File pluginDir) throws InvalidPluginDirectoryException {
		directory = pluginDir.getName();
		int idx = directory.indexOf(IMUMConstants.VERSION_SEPARATOR);
		if ((idx <= 0) || (idx == (directory.length() - 1)))
			badDir(INVALID_DIRECTORY_ENTRY);
		id = directory.substring(0, idx);
		try {
			version = new Version(directory.substring(idx + 1));
		} catch (VersionFormatException e) {
			badDir(INVALID_PLUGIN_VERSION, e);
		}
	}

	private void getFiles(File pluginDir)
		throws InvalidPluginDirectoryException {
		files = new ArrayList();
		ChecksumCalculator calc = new ChecksumCalculator();
		File[] list = pluginDir.listFiles();
		boolean noManifest = true;
		for (int i = 0; i < list.length; i++) {
			if (list[i].getName().equals(IBootConstants.MANIFEST_FILE))
				noManifest = false;
			long crc = calc.calcCachedChecksum(list[i]);
			files.add(new ExistingFileEntry(list[i], crc, list[i].length()));
		}
		if (noManifest)
			badDir(MISSING_MANIFEST_FILE);
	}

	private void badDir(String message)
		throws InvalidPluginDirectoryException {
		badDir(message, null);
	}

	private void badDir(String message, Throwable cause)
		throws InvalidPluginDirectoryException {
		if (message == null)
			throw new InvalidPluginDirectoryException(cause);
		else
			throw new InvalidPluginDirectoryException(message, cause);
	}

	/**
	 * @see gov.va.med.hds.cd.mum.update.IExistingPlugin#getFile(int)
	 */
	public ExistingFileEntry getFile(int index) {
		return (ExistingFileEntry) files.get(index);
	}

	/**
	 * @see gov.va.med.hds.cd.mum.IBasePlugin#getID()
	 */
	public String getID() {
		return id;
	}

	/**
	 * @see gov.va.med.hds.cd.mum.IBasePlugin#getVersion()
	 */
	public Version getVersion() {
		return version;
	}

	/**
	 * @see gov.va.med.hds.cd.mum.IBasePlugin#getDirectory()
	 */
	public String getDirectory() {
		return directory;
	}

	/**
	 * @see gov.va.med.hds.cd.mum.IBasePlugin#getFileCount()
	 */
	public int getFileCount() {
		return files.size();
	}

	/**
	 * @see gov.va.med.hds.cd.mum.update.IExistingPlugin#getFiles()
	 */
	public File[] getFiles() {
		File[] fileList = new File[getFileCount()];
		for (int i = 0; i < fileList.length; i++)
			fileList[i] = getFile(i).getFile();
		return fileList;
	}

	private String id;
	private String directory;
	private List files;
	private Version version;
}
