package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.DependentPlugin;
import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.InvalidPluginException;
import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.loader.IDesktopClassLoader;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.KeyedList;
import gov.va.med.hds.cd.mum.boot.utils.SimpleList;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

/**
 * Represents the plugin specification as detailed in the configuration file.
 */
public class PluginSpec extends DependentPlugin implements IPluginSpec, IUpdateConstants {

    /**
     * Constructor for PluginSpec 
     * @param xml - the config file xml
     * @param plugin - the node in the config file representing the plugin
     * @throws InvalidConfigurationException
     */
    protected PluginSpec(ITreeData xml, int plugin) throws InvalidConfigurationException {
        super();
        parse(xml, plugin);
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#publishAllClasses()
     */
    public boolean publishAllClasses() {
        return publishAllClasses;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#getFile(int)
     */
    public SpecFileEntry getFile(int index) {
        return (SpecFileEntry) fileEntries.get(index);
    }

    /**
     * @see gov.va.med.hds.cd.mum.IBasePlugin#getFileCount()
     */
    public int getFileCount() {
        return fileEntries.size();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#getClassLoader()
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#setClassLoader(java.lang.ClassLoader)
     */
    public void setClassLoader(ClassLoader classLoader) {
        if ((classLoader == null) || (classLoader instanceof IDesktopClassLoader))
            this.classLoader = classLoader;
        else
            throw new IllegalArgumentException(INVALID_CLASS_LOADER);
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#getFiles(java.io.File)
     */
    public File[] getFiles(File rootDirectory) {
        if ((files == null)
            || (fileRoot == null)
            || (!fileRoot.toString().equals(rootDirectory.toString()))) {
            files = new File[getFileCount()];
            fileRoot = rootDirectory;
            for (int j = 0; j < getFileCount(); j++) {
                files[j] =
                    new File(
                        rootDirectory.getPath()
                            + File.separator
                            + IBootConstants.PLUGINS_DIRECTORY
                            + File.separator
                            + getDirectory()
                            + File.separator
                            + getFile(j).getFileName());
            }
        }
        return files;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#getURLs(java.io.File)
     */
    public URL[] getURLs(File rootDirectory) {
        File[] files = getFiles(rootDirectory);
        URL[] results = new URL[files.length];
        for (int i = 0; i < files.length; i++) {
            try {
                results[i] = files[i].toURL();
            } catch (MalformedURLException e) {
            }
        }
        return results;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#getManifestURL()
     */
    public URL getManifestURL() {
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                if (files[i].getName().equals(IBootConstants.MANIFEST_FILE))
                    try {
                        return files[i].toURL();
                    } catch (MalformedURLException e) {
                        return null;
                    }

            }
        }
        return null;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#getManifestXML()
     */
    public ITreeData getManifestXML() {
        return manifestXML;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#loadManifest(java.net.URL)
     */
    public void loadManifest(URL manifest) throws InvalidPluginException {
        initPlugin();
        manifestXML = getManifestXML(manifest);
    }

    private void parse(ITreeData xml, int plugin) throws InvalidConfigurationException {
        if (plugin < 0)
            badConfig(PLUGIN_NODE_MISSING);
        if (xml == null)
            throw new InvalidConfigurationException();
        getPluginInformation(xml, plugin);
        getClassInformation(xml, plugin);
        getPublishedInformation(xml, plugin);
        getFileInformation(xml, plugin);
        getRequirements(xml, plugin);
    }

    private void getClassInformation(ITreeData xml, int plugin)
        throws InvalidConfigurationException {
        setApplicationClass(
            getClassEntry(
                xml,
                plugin,
                IMUMConstants.APPLICATION_ELEMENT,
                MULTIPLE_APPLICATION_ENTRIES,
                MISSING_APPLICATION_CLASS));
    }

    private void getFileInformation(ITreeData xml, int plugin)
        throws InvalidConfigurationException {
        int node = -1;
        boolean noManfiest = true;
        fileEntries = new ArrayList();
        do {
            node = xml.findChildID(plugin, IMUMConstants.FILE_ELEMENT, node);
            if (node >= 0) {
                SpecFileEntry entry =
                    new SpecFileEntry(
                        getFilename(xml, node),
                        getCRC(xml, node),
                        getSize(xml, node));
                addPackages(entry, xml, node);
                if (entry.getFileName().equals(IBootConstants.MANIFEST_FILE))
                    noManfiest = false;
                fileEntries.add(entry);
            }
        } while (node >= 0);
        if (noManfiest)
            badConfig(MISSING_MANIFEST_FILE);
    }

    private void getRequirements(ITreeData xml, int plugin) {
        int node = -1;
        do {
            node = xml.findChildID(plugin, IPluginConstants.PREREQUISITE_ELEMENT, node);
            if (node >= 0)
                requiredPlugins().add(xml.getAttribute(node, IPluginConstants.PLUGIN_ELEMENT), "");
        } while (node >= 0);
    }

    private void addPackages(SpecFileEntry entry, ITreeData xml, int node) {
        int pkg = -1;
        do {
            pkg = xml.findChildID(node, IMUMConstants.PACKAGE_ELEMENT, pkg);
            if (pkg >= 0) {
                String pkgName = xml.getValue(pkg);
                if (pkgName == null)
                    pkgName = IBootConstants.DEFAULT_PACKAGE;
                entry.addPackage(pkgName);
            }
        } while (pkg >= 0);
    }

    private String getFilename(ITreeData xml, int node) throws InvalidConfigurationException {
        String filename = xml.getAttribute(node, IMUMConstants.FILENAME_ATTRIBUTE);
        if ((filename == null) || (filename.length() == 0))
            badConfig(INVALID_FILE_ENTRY);
        return filename;
    }

    private long getCRC(ITreeData xml, int node) throws InvalidConfigurationException {
        String crcValue = xml.getAttribute(node, IMUMConstants.CRC_ATTRIBUTE);
        if ((crcValue == null) || (crcValue.length() == 0))
            badConfig(INVALID_FILE_ENTRY);
        long crc = 0;
        try {
            crc = Long.parseLong(crcValue);
        } catch (NumberFormatException e) {
            badConfig(INVALID_FILE_ENTRY, e);
        }
        return crc;
    }

    private long getSize(ITreeData xml, int node) throws InvalidConfigurationException {
        long size = 0;
        String sizeValue = xml.getAttribute(node, IMUMConstants.SIZE_ATTRIBUTE);
        if ((sizeValue != null) && (sizeValue.length() > 0))
            try {
                size = Long.parseLong(sizeValue);
            } catch (NumberFormatException e) {
                badConfig(INVALID_FILE_ENTRY, e);
            }
        return size;
    }

    private void getPublishedInformation(ITreeData xml, int plugin)
        throws InvalidConfigurationException {
        publishAllClasses = false;
        int count = 0;
        int node = -1;
        do {
            node = xml.findChildID(plugin, IPluginConstants.PUBLISH_ELEMENT, node);
            if (node >= 0) {
                count++;
                String cls = xml.getAttribute(node, IPluginConstants.CLASS_ATTRIBUTE);
                if ((cls == null) || (cls.length() == 0))
                    badConfig(MISSING_PUBLISHED_CLASS);
                if (cls.equals(IMUMConstants.WILDCARD))
                    publishAllClasses = true;
                else {
                    if (cls.indexOf(IMUMConstants.WILDCARD) >= 0)
                        badConfig(NO_WILDCARD_ENTRIES);
                    if (getPublished().contains(cls))
                        badConfig(DUPLICATE_CLASS_ENTRIES);
                    getPublished().add(cls);
                }
            }
        } while (node >= 0);
        if (publishAllClasses && (count != 1))
            badConfig(IMUMConstants.TOO_MANY_PUBLISHED_ENTRIES);
    }

    private String getClassEntry(
        ITreeData xml,
        int plugin,
        String tag,
        String dupError,
        String missingError)
        throws InvalidConfigurationException {
        String result = null;
        int node = xml.findChildID(plugin, tag);
        if (node >= 0) {
            result = xml.getAttribute(node, IPluginConstants.CLASS_ATTRIBUTE);
            if (xml.findChildID(plugin, tag, node) >= 0)
                badConfig(dupError);
            if ((result == null) || (result.length() == 0))
                badConfig(missingError);
        }
        return result;
    }

    private void getPluginInformation(ITreeData xml, int plugin)
        throws InvalidConfigurationException {
        setName(xml.getAttribute(plugin, IPluginConstants.NAME_ATTRIBUTE));
        String temp = xml.getAttribute(plugin, IMUMConstants.DIRECTORY_ATTRIBUTE);
        if (temp == null)
            badConfig(MISSING_DIRECTORY);
        int idx = temp.indexOf(IMUMConstants.VERSION_SEPARATOR);
        if ((idx <= 0) || (idx == (temp.length() - 1)))
            badConfig(INVALID_DIRECTORY_ENTRY);
        setDirectory(temp);
        setID(temp.substring(0, idx));
        try {
            setVersion(new Version(getDirectory().substring(idx + 1)));
        } catch (VersionFormatException e) {
            badConfig(INVALID_PLUGIN_VERSION, e);
        }
        temp = xml.getAttribute(plugin, IPluginConstants.MODULE_ATTRIBUTE);
        if (temp == null)
            temp = getID();
        else {
            if (temp.length() == 0)
                badConfig(INVALID_MODULE_ENTRY);
        }
        setModule(temp);
        temp = xml.getAttribute(plugin, IMUMConstants.SOURCE_ATTRIBUTE);
        if (temp == null)
            temp = getDirectory();
        setSourceDirectory(temp);
    }

    private void badConfig(String message) throws InvalidConfigurationException {
        throw new InvalidConfigurationException(message);
    }

    private void badConfig(String message, Exception cause) throws InvalidConfigurationException {
        throw new InvalidConfigurationException(message, cause);
    }

    private boolean publishAllClasses;
    private List fileEntries;
    private File[] files;
    private File fileRoot;
    private ClassLoader classLoader;
    private ITreeData manifestXML;
    private String stringValue;

    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        if (stringValue == null)
            stringValue = getID() + " v" + getVersion().toString();
        return stringValue;
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#getPhysicalFile(int)
     */
    public File getPhysicalFile(int index) {
        return files[index];
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#getFilePackages(int)
     */
    public SimpleList getFilePackages(int index) {
        return ((SpecFileEntry) fileEntries.get(index)).getPackages();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#getPublishedClasses()
     */
    public SimpleList getPublishedClasses() {
        return getPublished();
    }

    /**
     * @see gov.va.med.hds.cd.mum.boot.loader.IDesktopPlugin#getLinkedPlugins()
     */
    public KeyedList getLinkedPlugins() {
        return requiredPlugins();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#assignNeededUpdates(gov.va.med.hds.cd.mum.update.IExistingPlugin)
     */
    public void assignNeededUpdates(IExistingPlugin plugin) {
        for (int i = 0; i < fileEntries.size(); i++) {
            SpecFileEntry specEntry = getFile(i);
            specEntry.setUpdate(true);
            if (plugin != null) {
                for (int j = 0; j < plugin.getFileCount(); j++) {
                    ExistingFileEntry entry = plugin.getFile(j);
                    if (specEntry.getFileName().equals(entry.getFile().getName())) {
                        if ((entry.getCRC() == specEntry.getCRC())
                            && (entry.getSize() == specEntry.getSize()))
                            specEntry.setUpdate(false);
                        break;
                    }
                }
            }

        }
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#hasUpdates()
     */
    public boolean hasUpdates() {
        for (int i = 0; i < fileEntries.size(); i++) {
            SpecFileEntry entry = getFile(i);
            if (entry.getUpdate())
                return true;
        }
        return false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginSpec#samePlugin(gov.va.med.hds.cd.mum.update.IExistingPlugin)
     */
    public boolean samePlugin(IExistingPlugin plugin) {
        return plugin.getDirectory().equals(getDirectory())
            && plugin.getID().equals(getID())
            && plugin.getVersion().equals(getVersion());
    }
}
