package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.boot.IStatusNotifier;
import gov.va.med.hds.cd.mum.boot.SilentAbortException;
import gov.va.med.hds.cd.mum.boot.UpdateAuthorizationFailureException;
import gov.va.med.hds.cd.mum.utils.ChecksumCalculator;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Used to update plugins from the server to the workstation.
 */
public class PluginUpdater implements IPluginUpdater, IUpdateConstants {

    /**
     * Constructor for PluginUpdater 
     */
    protected PluginUpdater() {
        super();
        buf = new StringBuffer();
        calc = new ChecksumCalculator();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUpdater#update(java.net.URL, java.io.File, gov.va.med.hds.cd.mum.update.IPluginSpec)
     */
    public File[] update(URL fromPluginDir, File toPluginDir, IPluginSpec plugin)
        throws UpdateException, UpdateAuthorizationFailureException {
        if ((fromPluginDir == null) || (plugin == null) || (toPluginDir == null))
            throw new NullPointerException();
        if (!updating)
            throw new UpdateException(UPDATER_NOT_STARTED_ERROR);
        notify(plugin);
        File pluginDir = new File(updateDirectory, plugin.getDirectory());
        try {
            pluginDir.mkdirs();
        } catch (SecurityException e) {
            throw new UpdateException(IMUMConstants.WRITE_FAILURE_ERROR, e);
        }
        File[] results = getResults(plugin, toPluginDir);
        File[] updates = getUpdates(plugin, updateDirectory);
        copyPlugin(fromPluginDir, pluginDir, plugin, updates);
        cleanUp(plugin, pluginDir, new File(toPluginDir, plugin.getDirectory()), results, updates);
        updateDirectory.delete();
        return results;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUpdater#beginUpdate(java.io.File, java.lang.String, gov.va.med.hds.cd.mum.boot.IStatusNotifier, long)
     */
    public void beginUpdate(
        File rootDirectory,
        String message,
        IStatusNotifier notifier,
        long size)
        throws UpdateException {
        if (updating)
            throw new UpdateException(UPDATER_NOT_STOPPED_ERROR);
        validateUpdateAuthority(rootDirectory);
        this.notifier = notifier;
        if (notifier != null) {
            notifier.beginDisplayUpdate();
            try {
                notifier.showCancelButton(true);
                notifier.setCancelMessage("Canceling Plug-in Update...");
                notifier.showProgressBar(size, message);
            } finally {
                notifier.endDisplayUpdate();
            }
        }
        updating = true;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUpdater#endUpdate(gov.va.med.hds.cd.mum.boot.IStatusNotifier)
     */
    public void endUpdate(IStatusNotifier notifier) {
        if (updating) {
            MUMUtils.deleteTree(updateDirectory);
            if (notifier != null) {
                notifier.beginDisplayUpdate();
                try {
                    notifier.showCancelButton(false);
                    notifier.displayMessage(null);
                    notifier.hideProgressBar();
                } finally {
                    notifier.endDisplayUpdate();
                }
            }
            updating = false;
        }
    }

    private void validateUpdateAuthority(File root) throws UpdateException {
        updateDirectory = new File(root, IMUMConstants.UPDATE_DIRECTORY);
        File flagFile = new File(updateDirectory, UPDATE_PROGRESS_FILE);
        if (flagFile.exists()) {
            if (flagFile.lastModified() < (System.currentTimeMillis() - UPDATE_TIMEOUT))
                flagFile.delete();
            else
                throw new UnauthorizedUpdateException(UPDATE_IN_PROGRESS_ERROR);
        }
        updateDirectory.mkdirs();
        try {
            flagFile.createNewFile();
            flagFile.deleteOnExit();
        } catch (IOException e) {
        }
    }

    private void copyPlugin(URL fromPluginDir, File pluginDir, IPluginSpec plugin, File[] updates)
        throws UpdateException, UpdateAuthorizationFailureException {
        int count = 0;
        boolean copying = true;
        while (copying) {
            copyFiles(plugin, fromPluginDir, updates);
            if (listener != null)
                listener.copiedToUpdateFolder(plugin);
            copying = copiedFilesInvalid(updates, plugin);
            if (copying) {
                clearUpdates(updates);
                pluginDir.delete();
                count++;
                if (count >= MAX_COPY_TRIES)
                    throw new UpdateException(
                        COPY_FAILED_ERROR + plugin.toString() + COPY_FAILED_REASON);
            }
        }
    }

    private void clearUpdates(File[] updates) {
        for (int i = 0; i < updates.length; i++) {
            updates[i].delete();
        }
    }

    private boolean copiedFilesInvalid(File[] updates, IPluginSpec plugin) {
        boolean invalid = false;
        for (int i = 0; i < plugin.getFileCount(); i++) {
        	if(updates[i] == null) continue;
            SpecFileEntry entry = plugin.getFile(i);
            try {
                calc.calcCRC(updates[i].toURL());
            } catch (MalformedURLException e) {
                invalid = true;
                break;
            }
            if (entry.getCRC() != calc.getValue()) {
                invalid = true;
                break;
            }
        }
        if (invalid) {
            for (int i = 0; i < plugin.getFileCount(); i++) {
                SpecFileEntry entry = plugin.getFile(i);
                if (notifier != null)
                    notifier.updateProgressBarMax(entry.getSize());
            }
        }
        return invalid;
    }

    private void copyFiles(IPluginSpec plugin, URL fromPluginDir, File[] updates)
        throws UpdateException, UpdateAuthorizationFailureException {
        String fromPlugin = assignWithTrailingSlash(fromPluginDir.toString());
        for (int i = 0; i < plugin.getFileCount(); i++) {
            SpecFileEntry entry = plugin.getFile(i);
            if(entry.getUpdate())
	            copyURL(getSourceURL(plugin, fromPlugin, entry), updates[i]);
        }
    }

    private URL getSourceURL(IPluginSpec plugin, String fromPlugin, SpecFileEntry entry)
        throws UpdateException {
        buf.setLength(0);
        buf.append(fromPlugin);
        buf.append(plugin.getSourceDirectory());
        buf.append("/");
        buf.append(entry.getFileName());
        try {
            return MUMUtils.newURL(buf.substring(0, buf.length()));
        } catch (MalformedURLException e) {
            throw new UpdateException(e);
        }
    }

    private File[] getUpdates(IPluginSpec plugin, File update) {
        String toPlugin = assignWithTrailingSlash(update.toString());
        File[] updates = new File[plugin.getFileCount()];
        for (int i = 0; i < plugin.getFileCount(); i++) {
            SpecFileEntry entry = plugin.getFile(i);
            if(entry.getUpdate())
            	updates[i] = getTargetFile(plugin, toPlugin, entry);
            else
            	updates[i] = null;
        }
        return updates;
    }

    private File[] getResults(IPluginSpec plugin, File toPluginDir) {
        String finalToPlugin = assignWithTrailingSlash(toPluginDir.toString());
        File[] results = new File[plugin.getFileCount()];
        for (int i = 0; i < plugin.getFileCount(); i++) {
            SpecFileEntry entry = plugin.getFile(i);
           	results[i] = getTargetFile(plugin, finalToPlugin, entry);
        }
        return results;
    }

    private void cleanUp(
        IPluginSpec plugin,
        File pluginDir,
        File finalPluginDir,
        File[] results,
        File[] updates)
        throws UpdateException {
        boolean bad = false;
        boolean erase = false;
        File[] exclude = new File[results.length];
        for (int i = 0; i < results.length; i++) {
            if(updates[i] == null) {
            	exclude[i] = results[i];
            	results[i] = null;
            }
            else
            	exclude[i] = null;
        }
        MUMUtils.deleteTree(finalPluginDir, exclude);
        finalPluginDir.mkdirs();
        for (int i = 0; i < results.length; i++) {
        	if(updates[i] == null) continue;
            if (!updates[i].renameTo(results[i])) {
                bad = true;
                updates[i].delete();
            } else
                erase = true;
        }
        pluginDir.delete();
        if (bad) {
            if (erase)
                MUMUtils.deleteTree(finalPluginDir);
            UpdateManagerFactory.newUpdateNotifier().fatalErrorMessage(
                ERROR_MOVING_PLUGIN_FILES + plugin.toString());
            throw new SilentAbortException();
        }
    }

    private File getTargetFile(IPluginSpec plugin, String toPlugin, SpecFileEntry entry) {
        buf.setLength(0);
        buf.append(toPlugin);
        buf.append(plugin.getDirectory());
        buf.append("/");
        buf.append(entry.getFileName());
        return new File(buf.substring(0, buf.length()));
    }

    private String assignWithTrailingSlash(String txt) {
        buf.setLength(0);
        buf.append(txt);
        if (buf.charAt(buf.length() - 1) != '/')
            buf.append("/");
        return buf.substring(0, buf.length());
    }

    private void copyURL(URL from, File to)
        throws UpdateException, UpdateAuthorizationFailureException {
        try {
            if (to.exists())
                to.delete();
            to.getParentFile().mkdirs();
            in = new BufferedInputStream(from.openConnection().getInputStream());
            out = new BufferedOutputStream(new FileOutputStream(to));
            int bytesRead;
            while ((bytesRead = in.read(buffer)) != -1) {
                out.write(buffer, 0, bytesRead);
                notify(bytesRead);
                if ((notifier != null) && (notifier.isCanceled())) {
                    close(true);
                    throw new SilentAbortException();
                }
            }

        } catch (IOException e) {
            try {
                close(false);
                if (to.exists()) {
                    to.delete();
                }
            } catch (Exception ex) {
            }
            throw new UpdateAuthorizationFailureException(IMUMConstants.WRITE_FAILURE_ERROR, e);
        }
        close(true);
    }

    private void close(boolean throwException) throws UpdateException {
        Exception ex = null;
        if (in != null) {
            try {
                in.close();
            } catch (IOException e) {
                ex = e;
            }
            in = null;
        }
        if (out != null) {
            try {
                out.flush();
                out.close();
            } catch (IOException e) {
                ex = e;
            }
            out = null;
        }
        if (throwException && (ex != null))
            throw new UpdateException(ex.getMessage(), ex);
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUpdater#assignUpdateListener(gov.va.med.hds.cd.mum.update.IUpdateListener)
     */
    public void assignUpdateListener(IUpdateListener listener) {
        this.listener = listener;
    }

    private void notify(IPluginSpec spec) {
        if (notifier != null) {
        	StringBuffer buf = new StringBuffer();
        	boolean parens = false;
        	if(spec.getName() != null) {
        		buf.append(spec.getName());
        		parens = true;
        	}
        	if(parens) {
        		buf.append(notifier.PROGRESS_MESSAGE_SEPERATOR);
        		buf.append("(");
        	}
        	buf.append(spec.toString());
			if(parens)
				buf.append(")");
            notifier.updateProgressMessage(buf.toString());
        }
    }

    private void notify(long size) {
        if (notifier != null)
            notifier.updateProgressBar(size);
    }

    private int MAX_BUF_SIZE = 1024 * 16;
    private byte[] buffer = new byte[MAX_BUF_SIZE];
    private InputStream in;
    private OutputStream out;
    private IStatusNotifier notifier;
    private IUpdateListener listener;
    private StringBuffer buf;
    private ChecksumCalculator calc;
    private boolean updating;
    private File updateDirectory;
}
