package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.utils.ISavableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.File;
import java.io.IOException;

/**
 * Tracks plugin usages and deletes unused plugins
 */
public class PluginUsageTracker implements IPluginUsageTracker, IUpdateConstants {

    private File usageFile;
    private ISavableTreeData doc;
    private int root;
    private boolean saveNeeded;

    /**
     * Constructor for PluginUsageTracker.
     */
    protected PluginUsageTracker(File root) {
        super();
        usageFile = new File(root, IMUMConstants.PLUGIN_USAGE_FILENAME);
        saveNeeded = false;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUsageTracker#trackPluginUsage(gov.va.med.hds.cd.mum.boot.utils.ListMap)
     */
    public void trackPluginUsage(ListMap plugins) {
        int today = MUMUtils.getDaysSince1970();
        for (int i = 0; i < plugins.size(); i++) {
            IPluginSpec spec = (IPluginSpec) plugins.get(i);
            try {
                assignDirectory(spec.getDirectory());
            } catch (TreeDataException e) {
                return;
            }
            if (getDate() < today)
                setDate(today);
        }
        saveIfNeeded();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IPluginUsageTracker#deleteUnusedPlugins(gov.va.med.hds.cd.mum.boot.utils.ListMap)
     */
    public void deleteUnusedPlugins(ListMap plugins) {
        int lastValidDate =
            MUMUtils.getDaysSince1970() - IMUMConstants.PLUGIN_USAGE_DAYS_BEFORE_DELETE;
        for (int i = 0; i < plugins.size(); i++) {
            IExistingPlugin plugin = (IExistingPlugin) plugins.get(i);
            if (plugin.getFileCount() > 0) {
                try {
                    assignDirectory(plugin.getDirectory());
                } catch (TreeDataException e) {
                    return;
                }
                if (getDate() < lastValidDate) {
                    deleteDirectory(plugin.getFile(0).getFile().getParentFile());
                    if (xml >= 0) {
                        doc.removeChild(root, xml);
                        xml = -1;
                        saveNeeded = true;
                    }
                }
            }
        }
        saveIfNeeded();
    }

    private void saveIfNeeded() {
        if (saveNeeded) {
            try {
                doc.save(usageFile);
            } catch (IOException e) {
            }
            saveNeeded = false;
        }
    }

    private void deleteDirectory(File file) {
        if (file.exists()) {
            if (file.isDirectory()) {
                File[] files = file.listFiles();
                for (int i = 0; i < files.length; i++)
                    deleteDirectory(files[i]);
            }
        }
        file.delete();
    }

    private int getElement(String directory) throws TreeDataException {
        ensureXML();
        int result = findExistingElement(directory);
        if (result < 0)
            result = createNewElement(directory);
        return result;
    }

    private int createNewElement(String directory) {
        int result;
        result = doc.createChild(root, IPluginConstants.PLUGIN_ELEMENT);
        doc.setAttribute(result, IMUMConstants.DIRECTORY_ATTRIBUTE, directory);
        doc.setAttribute(
            result,
            IMUMConstants.LAST_USED_ATTRIBUTE,
            Integer.toString(MUMUtils.getDaysSince1970()));
        saveNeeded = true;
        return result;
    }

    private int findExistingElement(String directory) {
        for (int i = 0; i < doc.getChildCount(root); i++) {
            int child = doc.getChildID(root, i);
            if ((IPluginConstants.PLUGIN_ELEMENT.equals(doc.getName(child)))
                && (directory.equals(doc.getAttribute(child, IMUMConstants.DIRECTORY_ATTRIBUTE))))
                return child;
        }
        return -1;
    }

    private void ensureXML() throws TreeDataException {
        if (doc == null) {
            boolean build;
            if (usageFile.exists()) {
                build = false;
                try {
                    doc = TreeDataFactory.getSimpleXMLTreeDataInstance();
                    doc.load(usageFile);
                    root = doc.getRootID();
                } catch (Exception e) {
                    build = true;
                    usageFile.delete();
                }
            } else
                build = true;
            if (build) {
                usageFile.getParentFile().mkdirs();
                doc = TreeDataFactory.getSimpleXMLTreeDataInstance();
                root = doc.createRoot(IMUMConstants.PLUGINS_ELEMENT);
                saveNeeded = true;
            }
        }
    }

    private int xml;

    private void assignDirectory(String directory) throws TreeDataException {
        xml = getElement(directory);
    }

    private int getDate() {
        if (xml < 0)
            return 0;
        else
            return Integer.parseInt(doc.getAttribute(xml, IMUMConstants.LAST_USED_ATTRIBUTE));
    }

    private void setDate(int value) {
        if (xml >= 0) {
            doc.setAttribute(xml, IMUMConstants.LAST_USED_ATTRIBUTE, Integer.toString(value));
            saveNeeded = true;
        }
    }
}
