package gov.va.med.hds.cd.mum.update;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.InvalidConfigurationException;
import gov.va.med.hds.cd.mum.boot.IBootConstants;
import gov.va.med.hds.cd.mum.boot.IPluginConstants;
import gov.va.med.hds.cd.mum.boot.IStartupInfo;
import gov.va.med.hds.cd.mum.boot.IStatusNotifier;
import gov.va.med.hds.cd.mum.boot.SilentAbortException;
import gov.va.med.hds.cd.mum.boot.UpdateAuthorizationFailureException;
import gov.va.med.hds.cd.mum.boot.utils.ListMap;
import gov.va.med.hds.cd.mum.utils.MUMUtils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JOptionPane;

/**
 * Manages the update of plugins to the workstation.
 */
public class UpdateManager implements IUpdateManager, IUpdateConstants, IBootConstants {

    /**
     * Constructor for UpdateManager 
     */
    protected UpdateManager() {
        super();
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IUpdateManager#update(gov.va.med.hds.cd.mum.boot.IStartupInfo)
     */
    public void update(IStartupInfo info)
        throws
            FileNotFoundException,
            InvalidPluginDirectoryException,
            UpdateException,
            InvalidConfigurationException,
            UpdateAuthorizationFailureException {
        init();
        queryPlugins(info);
        verifyCorrectStartup(info);
        updatePlugins(info);
        info.setSettings(reader.getSettings());
    }

    private void verifyCorrectStartup(IStartupInfo info)
        throws InvalidConfigurationException, UpdateAuthorizationFailureException {
        if (!currentMatchesNeededStartup()) {
            File file = new File(info.getStartupDirectory(), FORCE_COPY_FILENAME);
            try {
                file.createNewFile();
            } catch (IOException e) {
                throw new UpdateAuthorizationFailureException(IMUMConstants.WRITE_FAILURE_ERROR, e);
            }
            if (info.getSplashScreen() != null)
                info.getSplashScreen().close();
            JOptionPane.showMessageDialog(
                null,
                INVALID_STARTUP,
                INVALID_STARTUP_TITLE,
                JOptionPane.ERROR_MESSAGE);
            throw new SilentAbortException();
        }

    }

    private boolean currentMatchesNeededStartup() throws InvalidConfigurationException {
        boolean matches = false;
        boolean found = false;
        IPluginSpec wanted = getStartupSpec();
        List possible = current.get(wanted.getID());
        for (int i = 0; i < possible.size(); i++) {
            IExistingPlugin plugin = (IExistingPlugin) possible.get(i);
            if (plugin.getDirectory().equals(wanted.getDirectory())) {
                matches = (plugin.getFileCount() == wanted.getFileCount());
                if (matches) {
                    for (int j = 0; j < plugin.getFileCount(); j++) {
                        matches = false;
                        ExistingFileEntry entry = plugin.getFile(j);
                        for (int k = 0; k < wanted.getFileCount(); k++) {
                            SpecFileEntry wantedFile = wanted.getFile(k);
                            if (entry.getFile().getName().equals(wantedFile.getFileName())) {
                                matches = (entry.getCRC() == wantedFile.getCRC());
                                if (matches)
                                    matches = (entry.getSize() == wantedFile.getSize());
                                break;
                            }
                        }
                        if (!matches)
                            break;
                    }
                }
                found = true;
                break;
            }
        }
        if (!found)
            throw new InvalidConfigurationException("Unexpected Startup Failure");
        return matches;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IUpdateManager#getConfigFileURL(gov.va.med.hds.cd.mum.boot.IStartupInfo)
     */
    public URL getConfigFileURL(IStartupInfo info) throws InvalidConfigurationException {
        if (configFileURL == null) {
            String configFile = getConfigFileName(info);
            try {
                configFileURL = MUMUtils.newURL(configFile);
            } catch (MalformedURLException e) {
                throw new InvalidConfigurationException(BAD_CONFIG_SPECIFIED, e);
            }
        }
        return configFileURL;
    }

    private String getConfigFileName(IStartupInfo info) throws InvalidConfigurationException {
        String configFile = getConfigFromArgs(info);
        if (configFile == null)
            throw new InvalidConfigurationException(NO_CONFIG_SPECIFIED);
        return configFile;
    }


    /**
     * @see gov.va.med.hds.cd.mum.update.IUpdateManager#getSpecs()
     */
    public ListMap getSpecs() {
        return target;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IUpdateManager#getCurrentPlugins()
     */
    public ListMap getCurrentPlugins() {
        return current;
    }

    /**
     * @see gov.va.med.hds.cd.mum.update.IUpdateManager#getStartupSpec()
     */
    public IPluginSpec getStartupSpec() {
        return reader.getStartupPlugin();
    }

    private void queryPlugins(IStartupInfo info)
        throws InvalidConfigurationException, FileNotFoundException, InvalidPluginDirectoryException {
        URL url = getConfigFileURL(info);
        assignRootURL(url);
        assignDestFile(info);
        reader.readConfiguration(url, target);
        updateSplash(info, reader);
        findRequiredPlugins(target);
        locator.locatePlugins(destFile, current);
    }

    private void updateSplash(IStartupInfo info, IConfigReader reader) {
        if (((info.getSplashScreen() != null)
            && (info.getSplashScreen() instanceof IStatusNotifier))) {
            String version = null;
            int root = reader.getSettings().data().getRootID();
            if (root >= 0) {
                int desktop = reader.getSettings().data().findChildID(root, DESKTOP_ELEMENT);
                if (desktop >= 0)
                    version =
                        reader.getSettings().data().getAttribute(
                            desktop,
                            IPluginConstants.VERSION_ATTRIBUTE);
            }
            if (version == null)
                version = "Unknown";
            ((IStatusNotifier) info.getSplashScreen()).setVersion("Version: " + version);
        }
    }

    /**
     * Finds plugins that match the actual plugin dependencies.  Used internally, this method is made
     * public for testing purposes only
     * @param map
     * @throws InvalidConfigurationException
     */
    public void findRequiredPlugins(ListMap map) throws InvalidConfigurationException {
        for (int i = 0; i < map.size(); i++) {
            IPluginSpec spec = (IPluginSpec) map.get(i);
            List keys = spec.requiredPlugins().getKeys();
            for (int j = 0; j < keys.size(); j++) {
                String key = (String) keys.get(j);
                String id = key.substring(0, key.indexOf(IMUMConstants.VERSION_SEPARATOR));
                List list = map.get(id);
                if (list.size() < 1)
                    throw new InvalidConfigurationException(MISSING_PREREQUISITES_ERROR);
                boolean missing = true;
                for (int k = 0; k < list.size(); k++) {
                    IPluginSpec spec2 = (IPluginSpec) list.get(k);
                    if (key.equals(spec2.getDirectory())) {
                        spec.requiredPlugins().update(key, spec2);
                        missing = false;
                        break;
                    }
                }
                if (missing)
                    throw new InvalidConfigurationException(MISSING_PREREQUISITES_ERROR);
            }
        }
    }

    private String getConfigFromArgs(IStartupInfo info) {
        if (info.getArgs() != null) {
            for (int i = 0; i < info.getArgs().length; i++) {
                if (!info.getArgs()[i].startsWith(OPTION_PREFIX)) {
                    return info.getArgs()[i];
                }
            }
        }
        return null;
    }

    private void assignRootURL(URL url) throws InvalidConfigurationException {
        StringBuffer buf = new StringBuffer(url.toString());
        int idx = buf.indexOf("!");
        if (idx >= 0)
            buf.setLength(idx);
        idx = buf.lastIndexOf("/");
        if (idx >= 0)
            buf.setLength(idx);
        if (buf.charAt(buf.length() - 1) != '/')
            buf.append('/');
        buf.append(PLUGINS_DIRECTORY);
        buf.append('/');
        try {
            rootURL = MUMUtils.newURL(buf.toString());
        } catch (MalformedURLException e) {
            throw new InvalidConfigurationException(e);
        }
    }

    private void assignDestFile(IStartupInfo info) {
        StringBuffer buf = new StringBuffer(info.getRootDirectory().getPath());
        if (buf.charAt(buf.length() - 1) != File.separatorChar)
            buf.append(File.separatorChar);
        buf.append(PLUGINS_DIRECTORY);
        buf.append(File.separatorChar);
        destFile = new File(buf.toString());
    }

    private void beginUpdate(IStartupInfo info, String message, long size) throws UpdateException {
        try {
            updater.beginUpdate(info.getRootDirectory(), message, getNotifier(info), size);
        } catch (UnauthorizedUpdateException e) {
            UpdateManagerFactory.newUpdateNotifier().fatalErrorMessage(e.getMessage());
            throw new SilentAbortException(e);
        }
    }

    private long getSizeOfUpdate(IPluginSpec spec) {
        long result = 0;
        for (int i = 0; i < spec.getFileCount(); i++) {
            if (spec.getFile(i).getUpdate())
                result += spec.getFile(i).getSize();
        }
        return result;
    }

    private IStatusNotifier getNotifier(IStartupInfo info) {
        if ((info.getSplashScreen() != null)
            && (info.getSplashScreen() instanceof IStatusNotifier))
            return (IStatusNotifier) info.getSplashScreen();
        else
            return null;
    }

    private void updatePlugins(IStartupInfo info)
        throws InvalidConfigurationException, UpdateException, UpdateAuthorizationFailureException {
        boolean add, found;
        long size = 0;
        ArrayList list = new ArrayList();
        for (int i = 0; i < target.size(); i++) {
            IPluginSpec spec = (IPluginSpec) target.get(i);
            add = true;
            found = false;
            List possible = current.get(spec.getID());
            for (int j = 0; j < possible.size(); j++) {
                IExistingPlugin plugin = (IExistingPlugin) possible.get(j);
                if (spec.samePlugin(plugin)) {
                    spec.assignNeededUpdates(plugin);
                    add = spec.hasUpdates();
                    found = true;
                    break;
                }
            }
            if (!found)
                spec.assignNeededUpdates(null);
            if (add) {
                size += getSizeOfUpdate(spec);
                list.add(spec);
            }
        }
        if (list.size() > 0) {
            beginUpdate(info, PLUGIN_UPDATE_MESSAGE, size);
            try {
                for (int i = 0; i < list.size(); i++)
                    updater.update(rootURL, destFile, (IPluginSpec) list.get(i));
            } finally {
                updater.endUpdate(getNotifier(info));
            }
        }
    }

    /**
     * Used to facilitate reuse for multiple configurations during unit testing.
     */
    public void resetConfigURL() {
        configFileURL = null;
    }

    private void init() {
        reader = ConfigReaderFactory.newConfigReader();
        locator = PluginClassFactory.newPluginLocator(true);
        updater = PluginClassFactory.newPluginUpdater();
        if (current == null)
            current = new ListMap();
        else
            current.clear();
        if (target == null)
            target = new ListMap();
        else
            target.clear();
    }

    private IConfigReader reader;
    private IPluginLocator locator;
    private IPluginUpdater updater;
    private ListMap current;
    private ListMap target;
    private URL rootURL;
    private File destFile;
    private URL configFileURL;
}