package gov.va.med.hds.cd.mum.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Map;

import gov.va.med.hds.cd.mum.IMUMConstants;

/**
 * This class is used to manage a cache of check sums for existing plugins, keyed by absolute file
 * path, size and time stamp.  This speeds up startup in that existing plugins do not need to
 * calculate check sums unless they do not have a matching entry in the cache.
 */
public class ChecksumCache implements IMUMConstants {

	public static final int NO_CHECKSUM = -1;

	private static File cacheFile;
	private static Map cache;
	private static boolean updateNeeded;

	/**
	 * Enables the cache 
	 * @param rootDirectory - the directory where the configuration/cache file should be stored
	 */
	public static void enable(File rootDirectory) {
		cacheFile = new File(rootDirectory, CRC_CACHE_FILENAME);
		cache = null;
		updateNeeded = false;
	}

	/**
	 * Disables the cache
	 */
	public static void disable() {
		cacheFile = null;
		cache = null;
		updateNeeded = false;
	}

	/**
	 * Returns true if the cache is enabled
	 */
	public static boolean isEnabled() {
		return (cacheFile != null);
	}

	/**
	 * Returns the checksum of the passed in file from the cache.  If a matching entry is not found
	 * in the cache, a result of NO_CHECKSUM (-1) is returned. 
	 * @param file
	 */
	public static long getCachedChecksum(File file) {
		if (cacheFile == null)
			return NO_CHECKSUM;
		if (cache == null) {
			loadCache();
			if (cache == null)
				return NO_CHECKSUM;
		}
		ChecksumCacheInfo fileInfo = getFileInfo(file.toString());
		if (fileInfo != null && fileInfo.inSync(file))
			return fileInfo.crc;
		else
			return NO_CHECKSUM;
	}

	private static ChecksumCacheInfo getFileInfo(String string) {
		return (ChecksumCacheInfo) cache.get(string);
	}

	private static void loadCache() {
		try {
			ObjectInputStream in =
				new ObjectInputStream(new FileInputStream(cacheFile));
			cache = (Map) in.readObject();
			in.close();
			updateNeeded = false;

		} catch (Exception e) {
			cache = new HashMap();
			updateNeeded = true;
		}
	}

	/**
	 * Updates the cache with the specified value
	 * @param file
	 * @param value
	 */
	public static void updateCache(File file, long value) {
		if (cacheFile == null)
			return;

		if (cache == null)
			cache = new HashMap();

		String name = file.toString();
		cache.put(name, new ChecksumCacheInfo(file, value));
		updateNeeded = true;
	}

	/**
	 * Saves the cache file to the disk, if any changes have been made 
	 */
	public static void saveCache() {
		if ((updateNeeded) && (cacheFile != null) && (cache != null)) {
			try {
                cacheFile.getParentFile().mkdirs();
				ObjectOutputStream out =
					new ObjectOutputStream(new FileOutputStream(cacheFile));
				out.writeObject(cache);
				out.close();
				updateNeeded = false;
			} catch (IOException e) {
			}
		}
	}

}
