package gov.va.med.hds.cd.mum.utils;

import gov.va.med.hds.cd.mum.IMUMConstants;
import gov.va.med.hds.cd.mum.boot.IBootConstants;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * Used to calculate check sum values of plugin and configuration files.
 */
public class ChecksumCalculator implements IMUMConstants {

    private MUMChecksum chksum;
    private byte[] buffer = new byte[1024 * 16];

    private static boolean jarTestMode = false;

    /**
     * Turns on or off a different jar check sum calculator algorythm that
     * produces the same check sum even when the jar files are slightly different. For
     * use ONLY in unit testing.
     * @param enabled
     */
    public static void setJarTestMode(boolean enabled) {
        jarTestMode = enabled;
    }

    /**
     * Constructor for ChecksumCalculator 
     */
    public ChecksumCalculator() {
        super();
        chksum = new MUMChecksum();
        reset();
    }

    /**
     * Resets the current check sum value being calculated.
     */
    public void reset() {
        chksum.reset();
    }

    /**
     * Returns the current check sum value.
     * @return long
     */
    public long getValue() {
        return chksum.getValue();
    }

    public void update(String value) {
        chksum.update(value);
    }

    /**
     * Calculates the check sum value of the file.  This routine has the added
     * benefit of utilizing a checksum cache if one is available, and if no
     * cache value is available updating the cache as needed.
     * @param file
     */
    public long calcCachedChecksum(File file) {
        boolean useCache = ChecksumCache.isEnabled();
        long result = ChecksumCache.NO_CHECKSUM;
        if (useCache) {
            try {
                File canonical = file.getCanonicalFile();
                file = canonical;
            } catch (IOException ioe) {
            }
            result = ChecksumCache.getCachedChecksum(file);
        }
        if (result == ChecksumCache.NO_CHECKSUM) {
            try {
                calcCRC(file.toURL());
                result = getValue();
                if (useCache)
                    ChecksumCache.updateCache(file, result);
            } catch (MalformedURLException e) {
            }
        }
        return result;
    }

    /**
     * Calculates the check sum value of the file pointed to by the passed in
     * URL. Text files are calculated on a line by line basis, ignoring charage
     * return/line feeds.  Binary files calculate based on all bytes in the file.
     * @param url
     */
    public void calcCRC(URL url) {
        reset();
        if (url != null) {
            if (jarTestMode) {
                String fileName = url.getPath().toLowerCase();
                if (fileName.endsWith(IBootConstants.JAR_EXT) || fileName.endsWith(ZIP_EXT)) {
                    calcTestZipFile(url);
                    return;
                }
            }
            boolean isText = isTextURL(url);
            try {
                InputStream input = new BufferedInputStream(url.openConnection().getInputStream());
                if (isText)
                    calcTextFile(input, false, true);
                else
                    calcBinaryFile(input, true);
            } catch (IOException e) {
                reset();
            }
        }
    }

    /**
     * This method is used to calucalte the check sum value of configuration
     * files, expecting that the passed in input stream is for a
     * configuration xml file.  It is expected that the first CRC="" value
     * encountered in the xml is the chksum of the file itself.  Since this
     * chksum can not be calculated and stored within the file without changed
     * the validity of the chksum value, this method ignores the value of the
     * first chksum="x" text found, where the value between the quotation marks
     * (x) is the part ignored.  This routine calculates the chksum on a line by
     * line basis, ignoring charage return/line feeds.
     * @param stream
    	 */
    public void calcConfigFile(InputStream stream) {
        calcTextFile(stream, true, true);
    }

    private void calcTextFile(InputStream input, boolean ignoreFirstCRC, boolean close) {
        reset();
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(input));
            try {
                String line;
                boolean first = ignoreFirstCRC;
                String crcText = null;
                if (first)
                    crcText = " " + CRC_ATTRIBUTE + "=\"";
                int idx, q1, q2;
                while ((line = reader.readLine()) != null) {
                    if (first) {
                        idx = line.indexOf(crcText);
                        if (idx >= 0) {
                            q1 = line.indexOf('\"', idx + 1);
                            if (q1 > idx) {
                                q2 = line.indexOf('\"', q1 + 1);
                                if (q2 > q1) {
                                    chksum.update(line.substring(0, q1 + 1));
                                    line = line.substring(q2);
                                    first = false;
                                }
                            }
                        }
                    }
                    chksum.update(line);
                }
            } finally {
                if (close)
                    reader.close();
            }
        } catch (IOException e) {
        }
    }

    private boolean isTextURL(URL url) {
        String fileName = url.getPath().toLowerCase();
        boolean unknown = true;
        boolean isText = false;
        for (int i = 0; i < TEXT_EXTENSIONS.length; i++) {
            if (fileName.endsWith(TEXT_EXTENSIONS[i])) {
                unknown = false;
                isText = true;
                break;
            }
        }
        if (unknown) {
            for (int i = 0; i < BINARY_EXTENSIONS.length; i++) {
                if (fileName.endsWith(BINARY_EXTENSIONS[i])) {
                    unknown = false;
                    isText = false;
                    break;
                }
            }
        }
        if (unknown)
            isText = streamIsText(url);
        return isText;
    }

    private static final byte smallestASCIICode = 32;
    private static final byte largestASCIICode = 126;
    private static final byte[] allowedCodes = { 10, 13, 9 };
    private boolean streamIsText(URL url) {
        boolean result = true;
        try {
            InputStream input = new BufferedInputStream(url.openConnection().getInputStream());
            int bytesRead;
            try {
                bytesRead = input.read(buffer);
            } finally {
                input.close();
            }
            for (int i = 0; i < bytesRead; i++) {
                if (isBinaryValue(buffer[i])) {
                    result = false;
                    break;
                }
            }
        } catch (IOException e) {
            result = false;
        }
        return result;
    }

    private boolean isBinaryValue(byte value) {
        if ((value < smallestASCIICode) || (value > largestASCIICode)) {
            for (int i = 0; i < allowedCodes.length; i++) {
                if (value == allowedCodes[i])
                    return false;
            }
            return true;
        }
        return false;
    }

    /**
     * This method is used to calculate the checksum of binary files, like jars and dlls.
     * @param input
     * @param close
     */
    public void calcBinaryFile(InputStream input, boolean close) {
        reset();
        try {
            try {
                int bytesRead;
                while ((bytesRead = input.read(buffer)) != -1) {
                    chksum.update(buffer, 0, bytesRead);
                }
            } finally {
                if (close)
                    input.close();
            }
        } catch (IOException e) {
            reset();
        }
    }

    private void calcTestZipFile(URL url) {
        ArrayList list = new ArrayList();
        try {
            URLConnection conn = url.openConnection();
            conn.connect();
            InputStream in = conn.getInputStream();
            ZipInputStream zipStream = new ZipInputStream(in);
            try {
                ZipEntry entry;
                StringBuffer buf = new StringBuffer();
                while ((entry = zipStream.getNextEntry()) != null) {
                    String name = entry.getName().toLowerCase();
                    boolean binary = true;
                    for (int i = 0; i < TEXT_EXTENSIONS.length; i++) {
                        if (name.endsWith(TEXT_EXTENSIONS[i])) {
                            binary = false;
                            break;
                        }
                    }
                    if (binary)
                        calcBinaryFile(zipStream, false);
                    else
                        calcTextFile(zipStream, false, false);
                    buf.setLength(0);
                    buf.append(entry.getName());
                    buf.append("*");
                    buf.append(Long.toString(chksum.getValue()));
                    list.add(buf.substring(0, buf.length()));
                }
            } finally {
                zipStream.close();
                in.close();
            }
            Collections.sort(list);
            chksum.reset();
            for (int i = 0; i < list.size(); i++) {
                chksum.update((String) list.get(i));
            }
        } catch (IOException e) {
        }
    }

}
