package gov.va.med.hds.cd.print.internal;

import gov.va.med.hds.cd.mum.boot.ExceptionDialog;
import gov.va.med.hds.cd.print.IPrint;
import gov.va.med.hds.cd.print.IPrintPreviewable;
import gov.va.med.hds.cd.print.IPrintable;
import gov.va.med.hds.cd.print.NoPrintServicesException;
import gov.va.med.hds.cd.print.PrintDialogCancelledException;
import gov.va.med.hds.cd.print.PrintFailedException;
import gov.va.med.hds.cd.print.PrintPluginException;
import gov.va.med.hds.cd.print.PrintPreviewFailedException;
import gov.va.med.hds.cd.prism.IDesktop;
import gov.va.med.hds.cd.prism.PrismUtilities;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPlugin;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;

import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.PrinterJob;

import javax.print.Doc;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintException;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.ServiceUI;
import javax.print.SimpleDoc;
import javax.print.attribute.Attribute;
import javax.print.attribute.DocAttributeSet;
import javax.print.attribute.HashDocAttributeSet;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSize;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.event.PrintJobAdapter;
import javax.print.event.PrintJobEvent;
import javax.print.event.PrintJobListener;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

/**
 * Implementation of the Print plugin
 */
public class Print implements IPrint, IPlugin {
	private static final String PRINT_ERROR_WINDOW_NAME = "Printing Note";
	private static final String DEFAULT_JOB_NAME = "Care Management";
	private static final String DEFAULT_PRINT_PREVIEW_WINDOW_NAME = "Print Preview";
	private static final int PRINT_DIALOG_WINDOW_WIDTH = 426;
	private static final int PRINT_DIALOG_WINDOW_HEIGHT = 383;

	private static final float DEFAULT_PRINTABLE_AREA_X = 1.0f;
	private static final float DEFAULT_PRINTABLE_AREA_Y = 1.0f;
	private static final float DEFAULT_PRINTABLE_AREA_WIDTH = 6.5f;
	private static final float DEFAULT_PRINTABLE_AREA_HEIGHT = 9.0f;

	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#startup(gov.va.med.hds.cd.runtime.IRuntimeContext, gov.va.med.hds.cd.runtime.IPluginDescriptor)
	 */
	public void startup(IRuntimeContext runtime, IPluginDescriptor plugin) {
		this.runtime = runtime;
		initDefaultAttributes();
	}
	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#shutdown()
	 */
	public void shutdown() throws Exception {
	}

	/**
	 * @see gov.va.med.hds.cd.print.IPrint#print(java.awt.print.Printable)
	 */
	public void print(Object printable, DocFlavor docFlavor) throws PrintFailedException {
		try {
			displayPrintDialog(docFlavor);
		} catch (NoPrintServicesException e) {
			ExceptionDialog.showExceptionDialog("Print", "No printers are installed. Cannot print.", e);
		} catch (PrintDialogCancelledException e){
			return;
		}
		
		DocPrintJob job = printService.createPrintJob();

		PrintJobListener pjlistener = new PrintJobAdapter() {
			public void printJobFailed(PrintJobEvent e) {
				notifyUserError("Print Job FAILED");
				runtime.getLogger(this.getClass()).error("Print Job FAILED");
			}
			public void printJobRequiresAttention(PrintJobEvent e) {
				runtime.getLogger(this.getClass()).error("Print job REQUIRES ATTENTION.");
			}
		};
		job.addPrintJobListener(pjlistener);

		DocAttributeSet das = new HashDocAttributeSet();
		Doc doc = new SimpleDoc(printable, docFlavor, das);
		try {
			job.print(doc, requestAttributes);
		} catch (PrintException e) {
			notifyUserError("Print Job Failed");
			runtime.getLogger(this.getClass()).error("Print Job Failed", e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.print.IPrint#printPreview(gov.va.med.hds.cd.print.IPrintable)
	 */
	public void printPreview(Object printable, DocFlavor docFlavor) throws PrintPreviewFailedException {
		validatePrintableAndFlavor(printable, docFlavor);
		IPrintPreviewable preview;
		try {
			preview = createPrintPreviewer();
		} catch (PrintPluginException e) {
			throw new PrintPreviewFailedException("Couldn't create a print previewer", e);
		}

		PageFormat pageFormat = new PageFormat();
		Paper paper = new Paper();
		setPaperImageableArea(paper);
		setPaperSize(paper);
		pageFormat.setPaper(paper);

		preview.init(runtime, (IPrintable) printable, pageFormat, DEFAULT_PRINT_PREVIEW_WINDOW_NAME);
	}

	/**
	 * @see gov.va.med.hds.cd.print.IPrint#pageSetup()
	 */
	public void pageSetup(DocFlavor docFlavor) {
		PrinterJob job = PrinterJob.getPrinterJob();
		pageFormat = job.pageDialog(requestAttributes);
	}
	private void initDefaultAttributes() {
		requestAttributes = new HashPrintRequestAttributeSet();	
	}

	private void validatePrintableAndFlavor(Object printable, DocFlavor docFlavor) throws PrintPreviewFailedException {
		if(docFlavor != DocFlavor.SERVICE_FORMATTED.PRINTABLE || !(printable instanceof IPrintable))
			throw new PrintPreviewFailedException("DocFlavor "+docFlavor.toString()+" not supported");
	}

	private void setPaperSize(Paper paper) {
		MediaSizeName media = (MediaSizeName) requestAttributes.get(MediaSizeName.NA_8X10.getCategory());
		if (media == null)
			media = MediaSizeName.NA_LETTER;
		
		MediaSize size = MediaSize.getMediaSizeForName(media);
		
		float x = size.getX(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
		float y = size.getY(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
		OrientationRequested orientation = (OrientationRequested)requestAttributes.get(OrientationRequested.PORTRAIT.getCategory());
		if(orientation!=null){
			if(orientation.equals(OrientationRequested.PORTRAIT) || orientation.equals(OrientationRequested.REVERSE_PORTRAIT)){
				x = size.getX(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
				y = size.getY(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
			}else if(orientation.equals(OrientationRequested.LANDSCAPE) || orientation.equals(OrientationRequested.REVERSE_LANDSCAPE)){			
				y = size.getX(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
				x = size.getY(MediaSize.INCH) * DEFAULT_DOTS_PER_INCH;
			}
		}
		paper.setSize(x, y);
	}
	private void setPaperImageableArea(Paper paper) {
		MediaPrintableArea mpa = new MediaPrintableArea(1,1,1,1,MediaPrintableArea.INCH);
		
		mpa = (MediaPrintableArea) requestAttributes.get(mpa.getCategory());
		if (mpa == null)
			mpa = new MediaPrintableArea(DEFAULT_PRINTABLE_AREA_X, 
														DEFAULT_PRINTABLE_AREA_Y, 
														DEFAULT_PRINTABLE_AREA_WIDTH, 
														DEFAULT_PRINTABLE_AREA_HEIGHT, 
														MediaPrintableArea.INCH);
		
		double x = mpa.getX(MediaPrintableArea.INCH) * DEFAULT_DOTS_PER_INCH;
		double y = mpa.getY(MediaPrintableArea.INCH) * DEFAULT_DOTS_PER_INCH;
		double width = mpa.getWidth(MediaPrintableArea.INCH) * DEFAULT_DOTS_PER_INCH;
		double height = mpa.getHeight(MediaPrintableArea.INCH) * DEFAULT_DOTS_PER_INCH;
		
		if(pageFormat != null){
			x = pageFormat.getImageableX();
			y = pageFormat.getImageableY();
		}
		OrientationRequested orientation = (OrientationRequested)requestAttributes.get(OrientationRequested.PORTRAIT.getCategory());
		if(orientation!=null){
			if(orientation.equals(OrientationRequested.PORTRAIT) || orientation.equals(OrientationRequested.REVERSE_PORTRAIT))
				paper.setImageableArea(x, y, width, height);
			else if(orientation.equals(OrientationRequested.LANDSCAPE) || orientation.equals(OrientationRequested.REVERSE_LANDSCAPE))			
				paper.setImageableArea(x, y, height, width);
		}else{
			paper.setImageableArea(x, y, width, height);
		}
	}
	private void displayPrintDialog(DocFlavor docFlavor) throws NoPrintServicesException, PrintDialogCancelledException{
		PrintService[] services = PrintServiceLookup.lookupPrintServices(docFlavor, null);
		if(services.length==0)
			throw new NoPrintServicesException();

		printService = ServiceUI.printDialog(null, calculatePrintDialogOriginX(), calculatePrintDialogOriginY(), services, printService, docFlavor, requestAttributes);

		if (printService == null)
			throw new PrintDialogCancelledException();
	}

	private int calculatePrintDialogOriginX() {
		IDesktop desktop = PrismUtilities.getDesktop(runtime);
		JFrame rootFrame = desktop.getActiveWindow().getFrame();
		return rootFrame.getX() + ((rootFrame.getWidth() - PRINT_DIALOG_WINDOW_WIDTH) / 2);
	}
	private int calculatePrintDialogOriginY() {
		IDesktop desktop = PrismUtilities.getDesktop(runtime);
		JFrame rootFrame = desktop.getActiveWindow().getFrame();
		return rootFrame.getY() + ((rootFrame.getHeight() - PRINT_DIALOG_WINDOW_HEIGHT) / 2);
	}

	private void printAttributes(String descriptor, Attribute[] attrs) {
		System.out.println("-----------" + descriptor + "---------");
		for (int i = 0; i < attrs.length; i++) {
			System.out.println("Attribute[" + i + "]= " + attrs[i].getName());
		}
		System.out.println("-------------------------------------END----------------------------------");
	}
	private IPrintPreviewable createPrintPreviewer() throws PrintPluginException {
		IPluginRegistry pluginRegistry = runtime.getPluginRegistry();
		IConfigurationElement[] pluginExtensionConfigs = pluginRegistry.getConfigurationElementsFor(PRINT_PREVIEW_EXTENSION_POINT_NAME);

		if (pluginExtensionConfigs.length != 1)
			throw new PrintPluginException(
				"The " + getClass().getName() + " plugin requires one and only one " + PRINT_PREVIEW_EXTENSION_POINT_NAME + " extension.\n There were " + pluginExtensionConfigs.length + " defined");

		IConfigurationElement pluginConfig = pluginExtensionConfigs[0];

		if (!pluginConfig.getName().equals(PRINT_PREVIEW_CONFIG_TAG_NAME))
			throw new PrintPluginException("The extension for " + PRINT_PREVIEW_EXTENSION_POINT_NAME + " must have a " + PRINT_PREVIEW_CONFIG_TAG_NAME + " tag");

		try {
			return (IPrintPreviewable) pluginConfig.createExecutableExtension(PRINT_PREVIEW_CLASS_TAG_NAME);
		} catch (Exception e) {
			throw new PrintPluginException("Couldn't instantiate the class " + pluginConfig.getAttribute(PRINT_PREVIEW_CLASS_TAG_NAME), e);
		}
	}
	private void notifyUserError(String error) {
		JOptionPane.showMessageDialog(null, error, PRINT_ERROR_WINDOW_NAME, JOptionPane.INFORMATION_MESSAGE);
	}
	private void setPrintable(IPrintable printable, PrinterJob job) {
		job.setPrintable(printable);
	}
	private IRuntimeContext runtime;
	private IPluginDescriptor plugin;
	private PrintService printService = null;
	private PageFormat pageFormat = null;
	private PrintRequestAttributeSet requestAttributes;
}
