package gov.va.med.hds.cd.print.printables;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.print.ConstructionException;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.security.ISecurity;
import gov.va.med.hds.cd.security.IUserInfo;
import gov.va.med.hds.util.dao.ISiteDAO;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.text.SimpleDateFormat;

/**
 * Provides a printable header/footer printable implementation
 */
public class HeaderFooter extends DefaultPrintable {
	
	private static final double VERTICAL_TEXT_INSET_TOP = 3.0;
	private static final double VERTICAL_TEXT_INSET_BOTTOM = 3.0;
	private static final double VERTICAL_TEXT_LINE_SPACING = 3.0;
	private static final double HORIZONTAL_TEXT_INSET_RIGHT = 5.0;
	private static final double HORIZONTAL_TEXT_INSET_LEFT = 5.0;
	
	private static final String HEADER_LEFT_LABEL_FONT = "Arial Narrow";
	private static final int HEADER_LEFT_LABEL_STYLE = Font.BOLD;
	private static final int HEADER_LEFT_LABEL_FONT_SIZE = 14;
	
	private static final String HEADER_RIGHT_LABEL_FONT = HEADER_LEFT_LABEL_FONT;
	private static final int HEADER_RIGHT_LABEL_STYLE = Font.ITALIC;
	private static final int HEADER_RIGHT_LABEL_FONT_SIZE = 14;
	
	private static final String FOOTER_CONTENT_FONT = HEADER_LEFT_LABEL_FONT;
	private static final int FOOTER_CONTENT_FONT_STYLE = Font.PLAIN;
	private static final int FOOTER_CONTENT_FONT_SIZE = 10;
	
	private static final String FOOTER_LABEL_FONT = FOOTER_CONTENT_FONT;
	private static final int FOOTER_LABEL_FONT_STYLE = Font.BOLD;
	private static final int FOOTER_LABEL_FONT_SIZE = FOOTER_CONTENT_FONT_SIZE;
	
	private static final String PRINTED_DATE_LABEL = "Printed On: ";
	private static final String PRINTED_BY_LABEL = "By: ";
	private static final String PAGE_NUMBER_LABEL = "Page ";
	private static final String MAX_PAGE_NUMBER_HOLDER = "999";
	private static final String PAGE_NUMBER_FORMAT = "###";
	private static final double PAGE_NUMBER_SPACER_WIDTH = 20.0;	
	
	private String headerRightLabel;
	private String headerLeftLabel;
	private IUserInfo userInfo;
	private double headerHeight = 0;
	private double footerHeight = 0;
	private String divisionName = null;
	private String integrationName = null;
	private String userName;
	private String printDateTime;
	/**
	 * Constructor
	 * 
	 * @param runtime
	 * @param headerLeftLabel
	 * @param headerRightLabel
	 * @throws ConstructionException
	 */
	public HeaderFooter(IRuntimeContext runtime, String headerLeftLabel, String headerRightLabel) throws ConstructionException{
		super(runtime);
		
		this.headerLeftLabel = headerLeftLabel;
		this.headerRightLabel = headerRightLabel;
		init();
	}
	/* (non-Javadoc)
	 * @see java.awt.print.Printable#print(java.awt.Graphics, java.awt.print.PageFormat, int)
	 */
	public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
		Graphics2D g2d = (Graphics2D)graphics;
		drawHeader(g2d,pageFormat,pageIndex);
		drawFooter(g2d,pageFormat,pageIndex);
		return Printable.PAGE_EXISTS;
	}
	
	/**
	 * If called after the print method is called the value in pixels of the rendered footer height
	 * is returned otherwise zero is returned.
	 * @return
	 */
	public double getJustRenderedFooterHeight() {
		return footerHeight;
	}
	/**
	 * If called after the print method is called the value in pixels of the rendered header height
	 * is returned otherwise zero is returned.
	 * @return
	 */
	public double getJustRenderedHeaderHeight() {
		return headerHeight;
	}
	private void init() throws ConstructionException{
		try {
			this.userInfo = ((ISecurity)getRuntime().getPluginInstance(ISecurity.PLUGIN_ID)).getUserInfo();
			ISiteDAO siteDao = (ISiteDAO)((IDaoRegistry)getRuntime().getPluginInstance(IDaoRegistry.PLUGIN_ID)).getDAO(ISiteDAO.class.getName());
			divisionName = siteDao.getDivisionName();
			integrationName = siteDao.getIntegrationName();
			userName = buildUserName();
			printDateTime = buildPrintDateTime();
		} catch (Exception e){
			throw new ConstructionException(e);
		}
	}
	
	private String buildPrintDateTime() {
		SimpleDateFormat dateFormat = new SimpleDateFormat("MM/dd/yy, hh:mma");
		return dateFormat.format(getPrintDate());
	}
	private void drawHeader(Graphics2D g2d, PageFormat pageFormat, int pageIndex) {
		headerHeight = 0;
		Graphics2D graphics = (Graphics2D)g2d.create();
		graphics.translate(pageFormat.getImageableX(),pageFormat.getImageableY());

		Font headerLeftFont = new Font(HEADER_LEFT_LABEL_FONT,HEADER_LEFT_LABEL_STYLE,HEADER_LEFT_LABEL_FONT_SIZE);
		Rectangle2D headerLeftLabelBounds = headerLeftFont.getStringBounds(headerLeftLabel,graphics.getFontRenderContext());
		Font headerRightFont = new Font(HEADER_RIGHT_LABEL_FONT,HEADER_RIGHT_LABEL_STYLE,HEADER_RIGHT_LABEL_FONT_SIZE);
		Rectangle2D headerRightLabelBounds = headerRightFont.getStringBounds(headerRightLabel,graphics.getFontRenderContext());
		
		double width = pageFormat.getImageableWidth();
		headerHeight=Math.max(headerLeftLabelBounds.getHeight(),headerRightLabelBounds.getHeight())+VERTICAL_TEXT_INSET_TOP+VERTICAL_TEXT_INSET_BOTTOM;
		drawBox(0.0f,0.0f,width,headerHeight,graphics);
		
		graphics.setFont(headerLeftFont);
		double textY = Math.max(headerLeftLabelBounds.getHeight(),headerRightLabelBounds.getHeight());
		graphics.drawString(headerLeftLabel,(float)HORIZONTAL_TEXT_INSET_LEFT,(float)textY);
		
		graphics.setFont(headerRightFont);
		double rightX = width-HORIZONTAL_TEXT_INSET_RIGHT-headerRightLabelBounds.getWidth();
		graphics.drawString(headerRightLabel,(float)rightX,(float)textY);
	}
	private void drawBox(double originX, double originY, double width, double height, Graphics2D graphics) {
		graphics.setColor(Color.BLACK);
		graphics.drawRect((int)originX,(int)originY,(int)width,(int)height);
	}
	
	private void drawFooter(Graphics2D g2d, PageFormat pageFormat, int pageIndex) throws PrinterException {
		footerHeight = 0;
		Graphics2D graphics = (Graphics2D)g2d.create();
		graphics.translate(pageFormat.getImageableX(),pageFormat.getImageableY());
		
		Font footerContentFont = new Font(FOOTER_CONTENT_FONT,FOOTER_CONTENT_FONT_STYLE,FOOTER_CONTENT_FONT_SIZE);
		Font footerLabelFont = new Font(FOOTER_LABEL_FONT,FOOTER_LABEL_FONT_STYLE,FOOTER_LABEL_FONT_SIZE);
		
		String dateTime = buildPrintDateTime();
		String printedBy = buildUserName();
		String pageNumber = Integer.toString(pageIndex+1);
		Rectangle2D divisionNameBounds = footerContentFont.getStringBounds(divisionName,graphics.getFontRenderContext());
		Rectangle2D interationNameBounds = footerContentFont.getStringBounds(integrationName,graphics.getFontRenderContext());
		Rectangle2D printedDateLabelBounds = footerLabelFont.getStringBounds(PRINTED_DATE_LABEL,graphics.getFontRenderContext());
		Rectangle2D printedByLabelBounds = footerLabelFont.getStringBounds(PRINTED_BY_LABEL,graphics.getFontRenderContext());
		Rectangle2D printedDateBounds = footerContentFont.getStringBounds(dateTime,graphics.getFontRenderContext());
		Rectangle2D printedByBounds = footerContentFont.getStringBounds(printedBy,graphics.getFontRenderContext());
		Rectangle2D maxPageNumberBounds = footerLabelFont.getStringBounds(MAX_PAGE_NUMBER_HOLDER,graphics.getFontRenderContext());
		Rectangle2D pageNumberLabelBounds = footerLabelFont.getStringBounds(PAGE_NUMBER_LABEL,graphics.getFontRenderContext());
		Rectangle2D pageNumberBounds = footerLabelFont.getStringBounds(pageNumber,graphics.getFontRenderContext());

		double width = pageFormat.getImageableWidth();
		double textHeight =Math.max(printedDateLabelBounds.getHeight(),printedDateBounds.getHeight());
		footerHeight = (VERTICAL_TEXT_INSET_TOP+VERTICAL_TEXT_INSET_BOTTOM+textHeight*2.0f+VERTICAL_TEXT_LINE_SPACING);
		double footerY = (pageFormat.getImageableHeight()-footerHeight);
		drawBox(0.0f,footerY,width,footerHeight,graphics);
		
		graphics.setFont(footerContentFont);
		double textLine1Y = footerY+textHeight;
		double textLine2Y = textLine1Y+textHeight+VERTICAL_TEXT_LINE_SPACING;
		
		graphics.drawString(divisionName,(float)HORIZONTAL_TEXT_INSET_LEFT,(float)textLine1Y);
		graphics.drawString(integrationName,(float)HORIZONTAL_TEXT_INSET_LEFT,(float)textLine2Y);
		
		double pageNumberX = width - HORIZONTAL_TEXT_INSET_RIGHT - maxPageNumberBounds.getWidth();
		double pageNumberLabelX = pageNumberX - pageNumberLabelBounds.getWidth();
		double dateTimeX = pageNumberLabelX - PAGE_NUMBER_SPACER_WIDTH - printedDateBounds.getWidth();
		double printedLabelX = dateTimeX - printedDateLabelBounds.getWidth();
		
		graphics.setFont(footerLabelFont);
		graphics.drawString(PRINTED_DATE_LABEL,(float)printedLabelX,(float)textLine1Y);
		graphics.drawString(PRINTED_BY_LABEL,(float)printedLabelX,(float)textLine2Y);
		
		graphics.setFont(footerContentFont);
		graphics.drawString(dateTime,(float)dateTimeX,(float)textLine1Y);
		graphics.drawString(printedBy,(float)dateTimeX,(float)textLine2Y);
		
		graphics.setFont(footerLabelFont);
		graphics.drawString(PAGE_NUMBER_LABEL,(float)pageNumberLabelX,(float)textLine2Y);
		pageNumberX = width - HORIZONTAL_TEXT_INSET_RIGHT - pageNumberBounds.getWidth();
		graphics.drawString(pageNumber,(float)pageNumberX,(float)textLine2Y);
	}
	private String buildUserName() {
		return (userInfo.getPrefix().equals("") ? "" : userInfo.getPrefix() + " ") + userInfo.getFirstName() + " " + userInfo.getLastName();
	}

}
