package gov.va.med.hds.cd.print.internal;
import gov.va.med.hds.cd.print.IPrint;
import gov.va.med.hds.cd.print.IPrintPreviewable;
import gov.va.med.hds.cd.print.IPrintable;
import gov.va.med.hds.cd.prism.IDesktop;
import gov.va.med.hds.cd.prism.PrismUtilities;
import gov.va.med.hds.cd.runtime.IRuntimeContext;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.net.URL;

import javax.print.DocFlavor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.border.EmptyBorder;
import javax.swing.border.MatteBorder;

/**
 * Default PrintPreview implementation that can handle Printable MIME types only
 */
public class PrintPreview extends JDialog implements IPrintPreviewable {
	private static final int DEFAULT_SCALE = 50;
	private static final int DEFAULT_SCALE_INDEX = 0;
	private static final String CLOSE_BUTTON_NAME = "Close";
	private static final String PRINT_BUTTON_NAME = "Print";
	private static final int ARROW_BUTTON_WIDTH = 20;
	private static final int ARROW_BUTTON_HEIGHT = 15;
	private static final String PREV_BUTTON_NAME = "Prev";
	private static final String NEXT_BUTTON_NAME = "Next";
	private static final String PRINT_ICON_IMAGE_PATH = "images/print.gif";
	private static final int DEFAULT_WINDOW_HEIGHT = 400;
	private static final int DEFAULT_WINDOW_WIDTH = 640;
	private PageFormat pageFormat;
	protected int virtualPageWidth;
	protected int virtualPageHeight;
	protected IPrintable targetPrintable;
	protected JComboBox scaleComboBox;
	protected PreviewContainer previewContainer;
	private IRuntimeContext runtime;
	private JToolBar toolBar;
	private JButton prevButton;
	private JButton nextButton;
	private JLabel currPageLabel;
	private int currPageIndex = 0;

	private int actualPageWidth;
	private int actualPageHeight;
	/**
	 * Default constructor
	 */
	public PrintPreview() {
		super();
	}
	/**
	 * Constructor
	 * 
	 * @param runtime - system runtime
	 * @param printable - object to be printed
	 * @param printerJob - printer job from which to get viewing attributes
	 * @param pageFormat - desired page format
	 * @param title - title of Print Preview Window
	 */
	public void init(IRuntimeContext runtime, IPrintable printable, PageFormat pageFormat, String title) {
		setTitle(title);
		this.runtime = runtime;
		this.pageFormat = pageFormat;
		this.setModal(true);
		targetPrintable = printable;
		setupWindow();

		previewContainer = new PreviewContainer();

		if (pageFormat.getHeight() == 0 || pageFormat.getWidth() == 0) {
			runtime.getLogger(PrintPreview.class.getName()).error("Unable to determine default page size");
			return;
		}
		virtualPageWidth = (int) (pageFormat.getWidth());
		virtualPageHeight = (int) (pageFormat.getHeight());
		int scale = DEFAULT_SCALE;
		actualPageWidth = (int) (virtualPageWidth * scale / 100);
		actualPageHeight = (int) (virtualPageHeight * scale / 100);
		setWindowSize(actualPageWidth, actualPageHeight);
		currPageIndex = -1;
		try {
			goNext();
		} catch (PrinterException e) {
			runtime.getLogger(PrintPreview.class.getName()).error("Print preview error", e);
		}

		JScrollPane ps = new JScrollPane(previewContainer);
		getContentPane().add(ps, BorderLayout.CENTER);

		setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		centerWindow();
		setVisible(true);
	}
	private void setupWindow() {
		toolBar = new JToolBar();
		setSize(DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT);
		ActionListener lst;
		setupPrevButton(toolBar);
		setupCurrPageLabel(toolBar);
		setupNextButton(toolBar);
		toolBar.addSeparator();
		setupScalesComboBox();
		toolBar.add(scaleComboBox);
		toolBar.addSeparator();
		setupPrintButton(toolBar);
		setupCloseButton(toolBar);
		getContentPane().add(toolBar, BorderLayout.NORTH);
	}
	private boolean printPageExists(int index, Graphics g, PageFormat format) throws PrinterException{
		if (targetPrintable.print(g.create(), format, index) != Printable.PAGE_EXISTS)
			return false;
		return true;
	}
	private void renderPage(Graphics g, PageFormat format, int index, BufferedImage img) throws PrinterException{
		if (targetPrintable.print(g, format, index) != Printable.PAGE_EXISTS){
			System.out.println("Failed to print page "+index);
			return;
		}
		PagePreview pp = new PagePreview(actualPageWidth, actualPageHeight, img);
		previewContainer.removeAll();
		previewContainer.add(pp);
	}
	private void goNext() throws PrinterException {
		BufferedImage img = new BufferedImage(virtualPageWidth, virtualPageHeight, BufferedImage.TYPE_INT_RGB);
		Graphics g = img.getGraphics();
		g.setColor(Color.white);
		g.fillRect(0, 0, virtualPageWidth, virtualPageHeight);
		if (printPageExists(currPageIndex + 2, g, pageFormat)){
			nextButton.setEnabled(true);
			g.fillRect(0, 0, virtualPageWidth, virtualPageHeight);
		}
		else
			nextButton.setEnabled(false);

		renderPage(g,pageFormat,currPageIndex+1,img);

		currPageIndex++;
		if (currPageIndex > 0)
			prevButton.setEnabled(true);

		setCurrPageNumber();
	}
	private void goPrev() throws PrinterException {
		currPageIndex--;
		
		BufferedImage img = new BufferedImage(virtualPageWidth, virtualPageHeight, BufferedImage.TYPE_INT_RGB);
		Graphics g = img.getGraphics();
		g.setColor(Color.white);
		g.fillRect(0, 0, virtualPageWidth, virtualPageHeight);
		renderPage(g,pageFormat,currPageIndex,img);
		
		if(currPageIndex==0)
			prevButton.setEnabled(false);

		setCurrPageNumber();
		nextButton.setEnabled(true);
	}
	private void setCurrPageNumber(){
		currPageLabel.setText(new Integer(currPageIndex+1).toString());
	}
	private void setupNextButton(JToolBar toolBar) {
		ActionListener lst;
		nextButton = new JButton(NEXT_BUTTON_NAME);
		lst = new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				try {
					goNext();
				} catch (PrinterException e) {
					runtime.getLogger(PrintPreview.class.getName()).error("Print preview error", e);
				}
			}
		};
		nextButton.addActionListener(lst);
		nextButton.setAlignmentY(0.5f);
		nextButton.setMargin(new Insets(4, 6, 4, 6));
		nextButton.setPreferredSize(new Dimension(ARROW_BUTTON_WIDTH,ARROW_BUTTON_HEIGHT));
		nextButton.setEnabled(false);
		toolBar.add(nextButton);
	}
	private void setupCurrPageLabel(JToolBar toolBar) {
		currPageLabel = new JLabel("1");
		currPageLabel.setBorder(new EmptyBorder(4,6,4,6));
		currPageLabel.setSize(ARROW_BUTTON_WIDTH,currPageLabel.getHeight());
		toolBar.add(currPageLabel);
	}
	private void setupPrevButton(JToolBar toolBar) {
		ActionListener lst;
		prevButton = new JButton(PREV_BUTTON_NAME);
		lst = new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				try {
					goPrev();
				} catch (PrinterException e) {
					runtime.getLogger(PrintPreview.class.getName()).error("Print preview error", e);
				}
			}
		};
		prevButton.addActionListener(lst);
		prevButton.setAlignmentY(0.5f);
		prevButton.setMargin(new Insets(4, 6, 4, 6));
		prevButton.setPreferredSize(new Dimension(ARROW_BUTTON_WIDTH,ARROW_BUTTON_HEIGHT));
		prevButton.setEnabled(false);
		toolBar.add(prevButton);
	}
	private void setWindowSize(int contentWidth, int contentHeight) {
		int w = contentWidth + 50;
		int h = contentHeight + toolBar.getHeight() + 100;
		setSize(w, h);
	}

	private void centerWindow() {
		IDesktop desktop = PrismUtilities.getDesktop(runtime);
		JFrame rootFrame = desktop.getActiveWindow().getFrame();
		int x = rootFrame.getX() + ((rootFrame.getWidth() - getWidth()) / 2);
		int y = rootFrame.getY() + ((rootFrame.getHeight() - getHeight()) / 2);
		setLocation(x, y);
	}

	private void setupScalesComboBox() {
		ActionListener lst;
		String[] scales = { "50 %", "75 %", "100 %", "150 %", "200 %" };
		scaleComboBox = new JComboBox(scales);
		lst = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				Thread runner = new Thread() {
					public void run() {
						String str = scaleComboBox.getSelectedItem().toString();
						if (str.endsWith("%"))
							str = str.substring(0, str.length() - 1);
						str = str.trim();
						int scale = 0;
						try {
							scale = Integer.parseInt(str);
						} catch (NumberFormatException ex) {
							return;
						}
						actualPageWidth = (int) (virtualPageWidth * scale / 100);
						actualPageHeight = (int) (virtualPageHeight * scale / 100);

						if (previewContainer != null) {
							Component[] comps = previewContainer.getComponents();
							for (int k = 0; k < comps.length; k++) {
								if (!(comps[k] instanceof PagePreview))
									continue;
								PagePreview pp = (PagePreview) comps[k];
								pp.setScaledSize(actualPageWidth, actualPageHeight);
							}
							previewContainer.doLayout();
							if (previewContainer.getParent() != null)
								if (previewContainer.getParent().getParent() != null)
									previewContainer.getParent().getParent().validate();
								else
									previewContainer.getParent().validate();
							else
								previewContainer.validate();
						}
					}
				};
				runner.start();
			}
		};
		scaleComboBox.addActionListener(lst);
		scaleComboBox.setMaximumSize(scaleComboBox.getPreferredSize());
		scaleComboBox.setEditable(true);
		scaleComboBox.setSelectedIndex(DEFAULT_SCALE_INDEX);
	}

	private void setupCloseButton(JToolBar toolBar) {
		ActionListener lst;
		JButton closeButton = new JButton(CLOSE_BUTTON_NAME);
		lst = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				dispose();
			}
		};
		closeButton.addActionListener(lst);
		closeButton.setAlignmentY(0.5f);
		closeButton.setMargin(new Insets(4, 6, 4, 6));
		toolBar.add(closeButton);

	}

	private void setupPrintButton(JToolBar toolBar) {
		URL url = PrintPreview.class.getResource(PRINT_ICON_IMAGE_PATH);

		JButton printButton = new JButton(PRINT_BUTTON_NAME, new ImageIcon(url));
		ActionListener lst = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					IPrint print = (IPrint) runtime.getPluginInstance(IPrint.PLUGIN_ID);
					print.print(targetPrintable, DocFlavor.SERVICE_FORMATTED.PRINTABLE);
					dispose();
				} catch (Exception ex) {
					runtime.getLogger(PrintPreview.class.getName()).error("Printing failed from Print Preview", ex);
				}
			}
		};
		printButton.addActionListener(lst);
		printButton.setAlignmentY(0.5f);
		printButton.setMargin(new Insets(4, 6, 4, 6));
		toolBar.add(printButton);
	}

	class PreviewContainer extends JPanel {
		protected int H_GAP = 16;
		protected int V_GAP = 10;

		public Dimension getPreferredSize() {
			int n = getComponentCount();
			if (n == 0)
				return new Dimension(H_GAP, V_GAP);
			Component comp = getComponent(0);
			Dimension dc = comp.getPreferredSize();
			int w = dc.width;
			int h = dc.height;

			Dimension dp = getParent().getSize();
			int nCol = Math.max((dp.width - H_GAP) / (w + H_GAP), 1);
			int nRow = n / nCol;
			if (nRow * nCol < n)
				nRow++;

			int ww = nCol * (w + H_GAP) + H_GAP;
			int hh = nRow * (h + V_GAP) + V_GAP;
			Insets ins = getInsets();
			return new Dimension(ww + ins.left + ins.right, hh + ins.top + ins.bottom);
		}

		public Dimension getMaximumSize() {
			return getPreferredSize();
		}

		public Dimension getMinimumSize() {
			return getPreferredSize();
		}

		public void doLayout() {
			Insets ins = getInsets();
			int x = ins.left + H_GAP;
			int y = ins.top + V_GAP;

			int n = getComponentCount();
			if (n == 0)
				return;
			Component comp = getComponent(0);
			Dimension dc = comp.getPreferredSize();
			int w = dc.width;
			int h = dc.height;

			Dimension dp = getParent().getSize();
			int nCol = Math.max((dp.width - H_GAP) / (w + H_GAP), 1);
			int nRow = n / nCol;
			if (nRow * nCol < n)
				nRow++;

			int index = 0;
			for (int k = 0; k < nRow; k++) {
				for (int m = 0; m < nCol; m++) {
					if (index >= n)
						return;
					comp = getComponent(index++);
					comp.setBounds(x, y, w, h);
					x += w + H_GAP;
				}
				y += h + V_GAP;
				x = ins.left + H_GAP;
			}
		}
	}

	class PagePreview extends JPanel {
		protected int m_w;
		protected int m_h;
		protected Image m_source;
		protected Image m_img;

		public PagePreview(int w, int h, Image source) {
			m_w = w;
			m_h = h;
			m_source = source;
			m_img = m_source.getScaledInstance(m_w, m_h, Image.SCALE_SMOOTH);
			m_img.flush();
			setBackground(Color.white);
			setBorder(new MatteBorder(1, 1, 2, 2, Color.black));
		}

		public void setScaledSize(int w, int h) {
			m_w = w;
			m_h = h;
			m_img = m_source.getScaledInstance(m_w, m_h, Image.SCALE_SMOOTH);
			repaint();
		}

		public Dimension getPreferredSize() {
			Insets ins = getInsets();
			return new Dimension(m_w + ins.left + ins.right, m_h + ins.top + ins.bottom);
		}

		public Dimension getMaximumSize() {
			return getPreferredSize();
		}

		public Dimension getMinimumSize() {
			return getPreferredSize();
		}

		public void paint(Graphics g) {
			g.setColor(getBackground());
			g.fillRect(0, 0, getWidth(), getHeight());
			g.drawImage(m_img, 0, 0, this);
			paintBorder(g);
		}
	}
}
