package gov.va.med.hds.cd.print.docs.cookbook;

import gov.va.med.hds.cd.print.IPrint;
import gov.va.med.hds.cd.print.IPrintActionEnablingListener;
import gov.va.med.hds.cd.print.IPrintActionHandler;
import gov.va.med.hds.cd.print.IPrintActionHandlerProvider;
import gov.va.med.hds.cd.print.PrintActionEnablingEvent;
import gov.va.med.hds.cd.print.PrintHandlerException;
import gov.va.med.hds.cd.runtime.IRuntimeContext;

import java.awt.Component;

import javax.print.DocFlavor;

/**
 * Cookbook example of a visual component that can enable/disable Print, 
 * Page Setup and Print Preview  File menu items. 
 * It also handles the events associated with those actions.
 * 
 * This is only a code fragment that should be used as an example.
 * It makes no sense without some real visual representation in 
 * this case.
 * 
 * This component must be added to the AWT component tree for a perspective
 * in order to get the IPrintActionHandlerProvider events.
 * 
 * The actions backing the Print, Page Setup and Print Preview menus
 * watch for prespective activate events. When a perspective is opened
 * the previous perspective component tree is searched for all implementors
 * of the IPrintActionHandlerProvider interface and the corresponding 
 * removeXXXActionEnablingListener method is called on that instance.
 * The component tree of the newly activated perspective is then searched
 * for implementors of the IPrintActionHandlerProvider and 
 * the corresponding addXXXActionEnablingListener method is called on that
 * instance.
 */
public class SimplePrintEnabler extends Component implements IPrintActionHandlerProvider{
	private IRuntimeContext runtime;
	/**
	 * Constructor
	 * 
	 * @param runtime Runtime context associated with the HealtheVet Desktop framework
	 */
	public SimplePrintEnabler(IRuntimeContext runtime){
		this.runtime = runtime;
	}
	
	/**
	 * Removes an action enabling listener for the print action
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#removePrintActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void removePrintActionEnablingListener(IPrintActionEnablingListener listener) {
		if(printEnablingListener == listener)
			printEnablingListener = null;
		// The print module will disable the print menu when the perspective containing this component is deactivated
		// If this object wants to take control of disabling the print menu it should do so here
		// with a call like the following:
		// 		listener.enable(new PrintActionEnablingEvent(this,false,null));
	}

	/**
	 * Adds an action enabling listener for the print action to this.
	 * This then enables the print action which enables the print menu item.
	 * An IPrintActionHandler is sent along with the event to enable the menu item.
	 * The IPrintActionHandler instance then connects to the IPrint plugin and prints
	 * a document when the doAction event is fired.
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#addPrintActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void addPrintActionEnablingListener(IPrintActionEnablingListener listener) {
		printEnablingListener = listener;
		printEnablingListener.enable(new PrintActionEnablingEvent(this, true, new IPrintActionHandler(){
			public void doAction() throws PrintHandlerException {
				try {
					IPrint print = (IPrint)runtime.getPluginInstance(IPrint.PLUGIN_ID);
					print.print(new SimplePrintable(),DocFlavor.SERVICE_FORMATTED.PRINTABLE);
				} catch (Exception e) {
					throw new PrintHandlerException(e);
				}
			}
		}));
	}

	/**
	 * Removes an action enabling listener for the print preview action
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#removePrintPreviewActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void removePrintPreviewActionEnablingListener(IPrintActionEnablingListener listener) {
		if(printPreviewEnablingListener == listener)
			printPreviewEnablingListener = null;
		// The print module will disable the print preview menu when the perspective containing this component is deactivated
		// If this object wants to take control of disabling the print preview menu it should do so here
		// with a call like the following:
		// 		listener.enable(new PrintActionEnablingEvent(this,false,null));
	}

	/**
	 * Adds an action enabling listener for the print preview action to this.
	 * This then enables the print preview action which enables the print preview menu item.
	 * An IPrintActionHandler is sent along with the event to enable the menu item.
	 * The IPrintActionHandler instance then connects to the IPrint plugin and does a printPreview
	 * on a document when the doAction event is fired.
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#addPrintPreviewActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void addPrintPreviewActionEnablingListener(IPrintActionEnablingListener listener) {
		printPreviewEnablingListener = listener;
		printPreviewEnablingListener.enable(new PrintActionEnablingEvent(this, true, new IPrintActionHandler(){
			public void doAction() throws PrintHandlerException {
				try {
					IPrint print = (IPrint)runtime.getPluginInstance(IPrint.PLUGIN_ID);
					print.printPreview(new SimplePrintable(),DocFlavor.SERVICE_FORMATTED.PRINTABLE);
				} catch (Exception e) {
					throw new PrintHandlerException(e);
				}
			}
		}));
	}

	/**
	 * Removes an action enabling listener for the Page Setup action
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#removePageSetupActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void removePageSetupActionEnablingListener(IPrintActionEnablingListener listener) {
		if(pageSetupEnablingListener == listener)
			pageSetupEnablingListener = null;
		// The print module will disable the page setup menu when the perspective containing this component is deactivated
		// If this object wants to take control of disabling the page setup menu it should do so here
		// with a call like the following:
		// 		listener.enable(new PrintActionEnablingEvent(this,false,null));
	}

	/**
	 * Adds an action enabling listener for the page setup action to this.
	 * This then enables the page setup action which enables the page setup menu item.
	 * An IPrintActionHandler is sent along with the event to enable the menu item.
	 * The IPrintActionHandler instance then connects to the IPrint plugin and does a pageSetup
	 * on a document when the doAction event is fired.
	 * @see gov.va.med.hds.cd.print.IPrintActionHandlerProvider#addPageSetupActionEnablingListener(gov.va.med.hds.cd.print.IPrintActionEnablingListener)
	 */
	public void addPageSetupActionEnablingListener(IPrintActionEnablingListener listener) {
		pageSetupEnablingListener = listener;
		pageSetupEnablingListener.enable(new PrintActionEnablingEvent(this, true, new IPrintActionHandler(){
			public void doAction() throws PrintHandlerException {
				try {
					IPrint print = (IPrint)runtime.getPluginInstance(IPrint.PLUGIN_ID);
					print.pageSetup(DocFlavor.SERVICE_FORMATTED.PRINTABLE);
				} catch (Exception e) {
					throw new PrintHandlerException(e);
				}
			}
		}));
	}
	private IPrintActionEnablingListener printEnablingListener = null;
	private IPrintActionEnablingListener pageSetupEnablingListener = null;
	private IPrintActionEnablingListener printPreviewEnablingListener = null;
}
