package gov.va.med.hds.cd.prism.internal;

import gov.va.med.hds.cd.config.ConfigurationException;
import gov.va.med.hds.cd.config.ConfigurationReadyEvent;
import gov.va.med.hds.cd.config.IConfiguration;
import gov.va.med.hds.cd.config.IConfigurationReadyListener;
import gov.va.med.hds.cd.config.IConfigurator;
import gov.va.med.hds.cd.config.IParameter;
import gov.va.med.hds.cd.config.NonExistentParameterException;
import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.prism.IPrismConstants;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.util.ExceptionStackTrace;

public class PrismParameters implements IPrismParameters {

	private ILogger logger;
	private IRuntimeContext runtimeContext;
	private static final String COUNTDOWN_KEY = "Countdown Seconds upon Timeout";
	private static final String TIMEOUT_KEY = "Timeout for GUI chart";
	private static final String DEFAULT_PERSPECTIVE_KEY = "Default Perspective";
	private static final String ADA_SECTION_508_NOTICE = "ADA Section 508 Notice Acknowledged";

	private static final int DEFAULT_TIMEOUT_SECONDS = 240;
	private static final int DEFAULT_COUNTDOWN_SECONDS = 20;

	private IConfiguration configuration = null;
	private static final int CONFIGURATION_READY_TIMEOUT_MILLISECONDS = 2000;

	private int timeoutSeconds = DEFAULT_TIMEOUT_SECONDS;
	private int countdownSeconds = DEFAULT_COUNTDOWN_SECONDS;
	private boolean ada508Notice = false;
	private String defaultPerspective = "";

	public PrismParameters(IRuntimeContext runtimeContext) {
		this.runtimeContext = runtimeContext;
		this.logger = runtimeContext.getLogger(PrismParameters.class);
		try {
			IConfigurator configurator =
				(IConfigurator) runtimeContext.getPluginInstance(IConfigurator.PLUGIN_ID);
			configurator.getConfiguration(
				IPrismConstants.PLUGIN_ID,
				new ConfigurationReadyListener());
			synchronized (PrismParameters.this) {
				PrismParameters.this.wait(CONFIGURATION_READY_TIMEOUT_MILLISECONDS);
			}
			logger.warning("Call to get prism configuration parameters timed out.");
		} catch (Exception e) {
			logger.warning(
				"Unable to get prism configuration parameters: "
					+ ExceptionStackTrace.getStackTrace(e));
		}
	}

	public PrismParameters(IConfiguration configuration, IRuntimeContext runtimeContext) {
		this.configuration = configuration;
		this.runtimeContext = runtimeContext;
		this.logger = runtimeContext.getLogger(PrismParameters.class);
	}

	public int getCountdownSeconds() {
		if (!hasConfiguration())
			return countdownSeconds;
		String value = getParameterValue(COUNTDOWN_KEY);
		if ((value != null) && (!value.equals(""))) {
			countdownSeconds = Integer.parseInt(value);
		} else {
			logger.warning(
				"No System Level Parameter defined for ORWOR TIMEOUT COUNTDOWN. Use option XPAR EDIT to correct.");
		}
		return countdownSeconds;
	}

	public int getTimeoutSeconds() {
		if (!hasConfiguration())
			return timeoutSeconds;
		String value = getParameterValue(TIMEOUT_KEY);
		if (value != null)
			timeoutSeconds = Integer.parseInt(value);
		return timeoutSeconds;
	}

	public String getDefaultPerspective() {
		if (!hasConfiguration())
			return "";
		String value = getParameterValue(DEFAULT_PERSPECTIVE_KEY);
		if (value == null)
			return "";
		return value;
	}

	public String[] getPerspectiveIds() {
		try {
			IPerspectiveListDAO dao = getPerspectiveListDAO();
			String[] ids = dao.getAvailablePerspectiveIds();
			return ids;
		} catch (Exception e) {
			runtimeContext.getLogger(getClass()).error("Error fetching perspective list", e);
			return new String[0];
		}
	}

	private IPerspectiveListDAO getPerspectiveListDAO() throws Exception {
		return (IPerspectiveListDAO) getDaoRegistry().getDAO(IPerspectiveListDAO.class.getName());
	}

	private IDaoRegistry getDaoRegistry() throws Exception {
		return (IDaoRegistry) runtimeContext.getPluginInstance(IDaoRegistry.PLUGIN_ID);
	}

	private IConfiguration getConfiguration() {
		return configuration;
	}

	private boolean hasConfiguration() {
		return getConfiguration() != null;
	}

	private String getParameterValue(String parameterKey) {
		try {
			IParameter param = configuration.getParameter(parameterKey);
			return param.getValue();
		} catch (NonExistentParameterException e) {
			return null;
		}
	}

	private class ConfigurationReadyListener implements IConfigurationReadyListener {

		public void configurationCallFailed(ConfigurationException ce) {
			logger.warning("Call to get prism configuration parameters failed", ce);
			synchronized (PrismParameters.this) {
				PrismParameters.this.notify();
			}
		}
		public void configurationReady(ConfigurationReadyEvent cre) {
			configuration = cre.getConfiguration();
			synchronized (PrismParameters.this) {
				PrismParameters.this.notify();
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.prism.internal.IPrismParameters#getSection508Notice()
	 */
	public boolean getSection508Notice() {
		if (!hasConfiguration())
			return false;
		String value = getParameterValue(ADA_SECTION_508_NOTICE);
		if (value != null)
			ada508Notice = value.equalsIgnoreCase("true");
		return ada508Notice;
	}
	
	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.prism.internal.IPrismParameters#setSection508Notice()
	 */
	public void setSection508Notice() {
		try {
			configuration.putParameter(ADA_SECTION_508_NOTICE, "true");
		} catch (Exception e) {
			ILogger logger = runtimeContext.getLogger(getClass().getName());
			logger.warning("Unable to load 508 notice: ", e);
		}
	}

}
