package gov.va.med.hds.cd.prism.internal;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import gov.va.med.hds.cd.prism.ISelection;

/**
 * A concrete implementation of the <code>IStructuredSelection</code> interface,
 * utilizing a simple Object array.
 * <p>
 * This class is not intended to be subclassed.
 * </p>
 */
public class Selection implements ISelection {

	/**
	 * The element that make up this selection.
	 */
	private Object[] elements;

	/**
	 * The canonical empty selection. This selection should be used instead of
	 * <code>null</code>.
	 */
	public static final Selection EMPTY = new Selection();
	/**
	 * Creates a new empty selection.  
	 * See also the static field <code>EMPTY</code> which contains an empty selection singleton.
	 *
	 * @see #EMPTY
	 */
	public Selection() {
	}
	/**
	 * Creates a structured selection from the given elements.
	 *
	 * @param elements an array of elements
	 */
	public Selection(Object[] elements) {
		this.elements = new Object[elements.length];
		System.arraycopy(elements, 0, this.elements, 0, elements.length);
	}
	/**
	 * Creates a selection containing a single object.
	 * The object must not be <code>null</code>.
	 *
	 * @param element the element
	 */
	public Selection(Object element) {
		elements = new Object[] { element };
	}
	/**
	 * Creates a selection from the given <code>List</code>.
	 */
	public Selection(List elements) {
		this.elements = elements.toArray();
	}
	
	/**
	 * Returns whether this selection is equal to given object.
	 * Two selections are equal iff they contain the same elements
	 * in the same order.
	 *
	 * @param o the other object
	 * @return <code>true</code> if they are equal, and <code>false</code> otherwise
	 */
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		//null and other classes
		if (!(o instanceof Selection)) {
			return false;
		}
		Selection s2 = (Selection) o;

		//null elements
		if (elements == null) {
			return s2.elements == null;
		}
		if (s2.elements == null) {
			return false;
		}

		//size
		int myLen = elements.length;
		if (myLen != s2.elements.length) {
			return false;
		}
		//element comparison
		for (int i = 0; i < myLen; i++) {
			if (!elements[i].equals(s2.elements[i])) {
				return false;
			}
		}
		return true;
	}
	
	public Object getFirstElement() {
		return isEmpty() ? null : elements[0];
	}
	
	public boolean isEmpty() {
		return elements == null || elements.length == 0;
	}
	
	public Iterator iterator() {
		return Arrays
			.asList(elements == null ? new Object[0] : elements)
			.iterator();
	}
	
	public int size() {
		return elements == null ? 0 : elements.length;
	}

	public Object[] toArray() {
		return elements == null ? new Object[0] : (Object[]) elements.clone();
	}
	
	public List toList() {
		return Arrays.asList(elements == null ? new Object[0] : elements);
	}
}
