package gov.va.med.hds.cd.prism.internal;

import java.text.MessageFormat;

import gov.va.med.hds.cd.prism.PrismUtilities;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveListener;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveReference;
import gov.va.med.hds.cd.prism.perspective.PerspectiveAdapter;
import gov.va.med.hds.cd.prism.perspective.PerspectiveEvent;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentReference;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;
import gov.va.med.hds.cd.workspace.IWorkspace;
import gov.va.med.hds.cd.workspace.IWorkspaceConstants;
import gov.va.med.hds.cd.workspace.IWorkspaceScope;
import gov.va.med.hds.cd.workspace.ListenerOfEventDescriptor;
import gov.va.med.hds.cd.workspace.SourceOfEventDescriptor;

public class VisualComponentEventRegistrar
	extends PerspectiveAdapter
	implements IPerspectiveListener {

	private ILogger log;
	private IWorkspace workspace;

	public VisualComponentEventRegistrar(
		ILoggingCapable logging,
		IWorkspace caster) {
		this.log = logging.getLogger(getClass());
		this.workspace = caster;
	}

	public void perspectiveClosing(PerspectiveEvent e) {
		IPerspectiveReference p = e.getPerspective();
		IVisualComponentReference[] components = p.getVisualComponents();
		for (int i = 0; i < components.length; i++) {
			disconnectSources(components[i], e);
			disconnectListeners(components[i], e);
		}
	}

	public void perspectiveOpened(PerspectiveEvent e) {
		IPerspectiveReference p = e.getPerspective();
		IVisualComponentReference[] components = p.getVisualComponents();
		for (int i = 0; i < components.length; i++) {
			connectSources(components[i], e);
			connectListeners(components[i], e);
		}
	}

	private void connectSources(
		IVisualComponentReference component,
		PerspectiveEvent e) {
		Object source = component.getVisualComponent();
		SourceOfEventDescriptor[] events =
			component.getDescriptor().getSourceOfEventDescriptors();
		for (int j = 0; j < events.length; j++) {
			IWorkspaceScope s = getScope(e, events[j].getScope());
			String eventClassName = events[j].getEventClassName();
			try {
				Class eventClass =
					source.getClass().getClassLoader().loadClass(
						eventClassName);
				s.getMulticaster().addSource(eventClass, source);
			} catch (Exception ex) {
				Object[] args =
					new Object[] {
						component.getDescriptor(),
						eventClassName,
						s.getId()};
				log.error(
					MessageFormat.format(
						"unable to connect visual component ''{0}'' as source of ''{1}'' events in workspace scope ''{2}''",
						args),
					ex);
			}
		}
	}

	private void connectListeners(
		IVisualComponentReference component,
		PerspectiveEvent e) {
		Object listener = component.getVisualComponent();
		ListenerOfEventDescriptor[] events =
			component.getDescriptor().getListenerOfEventDescriptors();
		for (int j = 0; j < events.length; j++) {
			IWorkspaceScope s = getScope(e, events[j].getScope());
			String eventClassName = events[j].getEventClassName();
			try {
				Class eventClass =
					listener.getClass().getClassLoader().loadClass(
						eventClassName);
				s.getMulticaster().addListener(eventClass, listener);
			} catch (Exception ex) {
				Object[] args =
					new Object[] {
						component.getDescriptor().getName() + "(" + component.getDescriptor().getId() + ")",
						eventClassName,
						s.getId()};
				log.error(
					MessageFormat.format(
						"unable to connect visual component ''{0}'' as listener of ''{1}'' events in workspace scope ''{2}''",
						args),
					ex);
			}
		}
	}

	private void disconnectListeners(
		IVisualComponentReference component,
		PerspectiveEvent e) {
		Object listener = component.getVisualComponent();
		ListenerOfEventDescriptor[] events =
				component.getDescriptor().getListenerOfEventDescriptors();
		for (int j = 0; j < events.length; j++) {
			IWorkspaceScope s = getScope(e, events[j].getScope());
			String eventClassName = events[j].getEventClassName();
			try {
				Class eventClass =
					listener.getClass().getClassLoader().loadClass(
						eventClassName);
				s.getMulticaster().removeListener(eventClass, listener);
			} catch (Exception ex) {
				Object[] args =
					new Object[] {
						component.getDescriptor(),
						eventClassName,
						s.getId()};
				log.error(
					MessageFormat.format(
						"unable to disconnect visual component ''{0}'' as listener of ''{1}'' events in workspace scope ''{2}''",
						args),
					ex);
			}
		}

	}

	private void disconnectSources(
		IVisualComponentReference component,
		PerspectiveEvent e) {
		Object source = component.getVisualComponent();
		SourceOfEventDescriptor[] events =
			component.getDescriptor().getSourceOfEventDescriptors();
		for (int j = 0; j < events.length; j++) {
			IWorkspaceScope s = getScope(e, events[j].getScope());
			String eventClassName = events[j].getEventClassName();
			try {
				Class eventClass =
					source.getClass().getClassLoader().loadClass(
						eventClassName);
				s.getMulticaster().removeSource(eventClass, source);
			} catch (Exception ex) {
				Object[] args =
					new Object[] {
						component.getDescriptor(),
						eventClassName,
						s.getId()};
				log.error(
					MessageFormat.format(
						"unable to disconnect visual component ''{0}'' as source of ''{1}'' events in workspace scope ''{2}''",
						args),
					ex);
			}
		}

	}
	
	private IWorkspaceScope getScope(
		PerspectiveEvent e,
		String scope) {
		if (scope == null
			|| IWorkspaceConstants.PERSPECTIVE_SCOPE.equals(
				scope)) {
			return workspace.getScope(
				PrismUtilities.getPerspectiveScopeId(
					e.getWindow(),
					e.getPerspective().getDescriptor()));
		} else if (IWorkspaceConstants.SESSION_SCOPE.equals(scope)) {
			return workspace.getScope(IWorkspaceConstants.SESSION_SCOPE);
		} else if (IWorkspaceConstants.PATIENT_SCOPE.equals(scope)) {
			return workspace.getScope(
				PrismUtilities.getPatientScopeId(e.getWindow()));
		}
		return null;
	}
}
