package gov.va.med.hds.cd.prism.menu.internal;

import gov.va.med.hds.cd.prism.IMenuConstants;
import gov.va.med.hds.cd.prism.IPrismConstants;
import gov.va.med.hds.cd.prism.action.IActionConstants;
import gov.va.med.hds.cd.prism.action.internal.ActionRegistry;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.InvalidConfigurationElementException;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;

public class MenuTreeReader implements IPrismConstants, IMenuConstants, IActionConstants {
	private static final String MENU_CREATION_ERROR = "error creating default menus:\n";

	private ILoggingCapable logging;

	public MenuTreeReader(ILoggingCapable logging) {
		this.logging = logging;
	}

	public void addRegisteredMenus(MenuTree menuTree, IPluginRegistry pluginRegistry) {
		readRegisteredMenus(pluginRegistry, menuTree);
	}

	public MenuTree readMenuTree(IPluginRegistry pluginRegistry, ActionRegistry actions) {
		MenuTree menuTree = new MenuTree(actions);
		readRegisteredMenus(pluginRegistry, menuTree);
		return menuTree;
	}

	private void readRegisteredMenus(IPluginRegistry pluginRegistry, MenuTree menuTree) {
		IConfigurationElement[] configElements =
			pluginRegistry.getConfigurationElementsFor(UNIQUE_MENU_EXTENSION_POINT_ID);

		for (int i = 0; i < configElements.length; ++i) {
			try {
				readConfigurationElement(configElements[i], menuTree);
			} catch (Exception e) {
				logging.getLogger(getClass()).error("error reading menu configuration element", e);
				continue;
			}
		}
	}

	private void readConfigurationElement(IConfigurationElement element, MenuTree menuTree)
		throws
			InvalidConfigurationElementException,
			InvalidMenuPathException,
			ClassNotFoundException,
			ActionNotFoundException {
		if (MENU_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readMenuElement(element, menuTree);
		} else if (MENU_ITEM_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readMenuItemElement(element, menuTree);
		} else if (SLOT_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readSlotElement(element, menuTree);
		} else if (SEPARATOR_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readSeparatorElement(element, menuTree);
		} else {
			throw new InvalidConfigurationElementException(
				"\""
					+ element.getName()
					+ "\" is an unrecognized configuration element name for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}
	}

	private void readMenuElement(IConfigurationElement element, MenuTree menuTree)
		throws InvalidConfigurationElementException, InvalidMenuPathException, ClassNotFoundException {
		validateMenuElement(element);

		String id = element.getAttribute(ID_ATTRIBUTE);
		String path = element.getAttribute(PATH_ATTRIBUTE);
		String className = element.getAttribute(CLASS_ATTRIBUTE);
		String mnemonicString = element.getAttribute(MNEMONIC_ATTRIBUTE);
		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		Character mnemonic = null;
		if (mnemonicString != null && mnemonicString.length() > 0) {
			mnemonic = new Character(mnemonicString.charAt(0));
		}
		if (className != null) {
			menuTree.addMenu(contextId, id, path, element, mnemonic);
		} else {
			String name = element.getAttribute(NAME_ATTRIBUTE);
			menuTree.addMenu(contextId, id, path, name, mnemonic);
		}
	}

	private void readMenuItemElement(IConfigurationElement element, MenuTree menuTree)
		throws InvalidConfigurationElementException, InvalidMenuPathException, ActionNotFoundException {
		validateMenuItemElement(element);

		String actionId = element.getAttribute(ACTION_ATTRIBUTE);
		String path = element.getAttribute(PATH_ATTRIBUTE);
		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		menuTree.addAction(contextId, actionId, path);
	}

	private void readSlotElement(IConfigurationElement element, MenuTree menuTree)
		throws InvalidConfigurationElementException, InvalidMenuPathException {
		validateSlotElement(element);

		String id = element.getAttribute(ID_ATTRIBUTE);
		String path = element.getAttribute(PATH_ATTRIBUTE);
		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		menuTree.addSlot(contextId, id, path);
	}

	private void readSeparatorElement(IConfigurationElement element, MenuTree menuTree)
		throws InvalidConfigurationElementException, InvalidMenuPathException {
		validateSlotElement(element);

		String id = element.getAttribute(ID_ATTRIBUTE);
		String path = element.getAttribute(PATH_ATTRIBUTE);
		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		menuTree.addSeparator(contextId, id, path);
	}

	private void validateMenuElement(IConfigurationElement element)
		throws InvalidConfigurationElementException, ClassNotFoundException {
		String id = element.getAttribute(ID_ATTRIBUTE);

		if (id == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ ID_ATTRIBUTE
					+ "\" attribute in \""
					+ MENU_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String path = element.getAttribute(PATH_ATTRIBUTE);

		if (path == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ PATH_ATTRIBUTE
					+ "\" attribute in \""
					+ MENU_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String className = element.getAttribute(CLASS_ATTRIBUTE);
  		if (className == null) {
  			String name = element.getAttribute(NAME_ATTRIBUTE);
			if (name == null) {
				throw new InvalidConfigurationElementException(
					"expected \""
						+ NAME_ATTRIBUTE
						+ "\" attribute in \""
						+ MENU_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
			}
  		}

		String mnemonic = element.getAttribute(MNEMONIC_ATTRIBUTE);
		if (mnemonic != null) {
			if (mnemonic.length() != 1)
				throw new InvalidConfigurationElementException(
					"expected character value for \""
						+ MNEMONIC_ATTRIBUTE
						+ "\" attribute in \""
						+ MENU_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
		}

		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		if (contextId != null) {
			if (contextId.length() == 0)
				throw new InvalidConfigurationElementException(
					"expected value for \""
						+ CONTEXT_ATTRIBUTE
						+ "\" attribute in \""
						+ MENU_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
		}
	}

	private void validateMenuItemElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {
		String action = element.getAttribute(ACTION_ATTRIBUTE);

		if (action == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ ACTION_ATTRIBUTE
					+ "\" attribute in \""
					+ MENU_ITEM_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String path = element.getAttribute(PATH_ATTRIBUTE);

		if (path == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ PATH_ATTRIBUTE
					+ "\" attribute in \""
					+ MENU_ITEM_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		if (contextId != null) {
			if (contextId.length() == 0)
				throw new InvalidConfigurationElementException(
					"expected value for \""
						+ CONTEXT_ATTRIBUTE
						+ "\" attribute in \""
						+ MENU_ITEM_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
		}
	}

	private void validateSlotElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {
		String id = element.getAttribute(ID_ATTRIBUTE);

		if (id == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ ID_ATTRIBUTE
					+ "\" attribute in \""
					+ SLOT_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String path = element.getAttribute(PATH_ATTRIBUTE);

		if (path == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ PATH_ATTRIBUTE
					+ "\" attribute in \""
					+ SLOT_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		if (contextId != null) {
			if (contextId.length() == 0)
				throw new InvalidConfigurationElementException(
					"expected value for \""
						+ CONTEXT_ATTRIBUTE
						+ "\" attribute in \""
						+ SLOT_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
		}
	}

	private void validateSeparatorElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {
		String id = element.getAttribute(ID_ATTRIBUTE);

		if (id == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ ID_ATTRIBUTE
					+ "\" attribute in \""
					+ SEPARATOR_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}

		String path = element.getAttribute(PATH_ATTRIBUTE);

		if (path == null) {
			throw new InvalidConfigurationElementException(
				"expected \""
					+ PATH_ATTRIBUTE
					+ "\" attribute in \""
					+ SEPARATOR_CONFIGURATION_ELEMENT
					+ "\" configuration element for extension point \""
					+ UNIQUE_MENU_EXTENSION_POINT_ID
					+ "\"");
		}
		
		String contextId = element.getAttribute(CONTEXT_ATTRIBUTE);
		if (contextId != null) {
			if (contextId.length() == 0)
				throw new InvalidConfigurationElementException(
					"expected value for \""
						+ CONTEXT_ATTRIBUTE
						+ "\" attribute in \""
						+ SEPARATOR_CONFIGURATION_ELEMENT
						+ "\" configuration element for extension point \""
						+ UNIQUE_MENU_EXTENSION_POINT_ID
						+ "\"");
		}
	}
}
