package gov.va.med.hds.cd.prism.perspective.internal;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.SwingConstants;

import gov.va.med.hds.cd.prism.perspective.ILayoutNode;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveDescriptor;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveReference;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponent;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentDescriptor;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentReference;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentRegistry;
import gov.va.med.hds.cd.prism.visualcomponent.internal.VisualComponentReference;
import gov.va.med.hds.cd.spectrum.MultiSplitPane;

public class PerspectiveFactory implements IPerspectiveFactory {
	
	private IVisualComponentRegistry visualComponents;

	public PerspectiveFactory(IVisualComponentRegistry visualComponents) {
		this.visualComponents = visualComponents;
	}

	public IPerspectiveReference createPerspective(IPerspectiveDescriptor perspective) {
		final List visualComponents = new ArrayList();
		JComponent c = createPerspectiveComponent(perspective, new IVisualComponentCallback() {
			public void componentCreated(IVisualComponentReference component) {
				visualComponents.add(component);
			}
		});
		IVisualComponentReference[] children = (IVisualComponentReference[]) visualComponents.toArray(new IVisualComponentReference[visualComponents.size()]);
		return new PerspectiveReference(perspective, c, children);
	}

	private JComponent createPerspectiveComponent(IPerspectiveDescriptor perspective, IVisualComponentCallback callback) {
		return createFromNode(perspective.getLayout().getRoot(), callback);
	}

	private JComponent createFromNode(ILayoutNode node, IVisualComponentCallback callback) {
		JComponent component = null;
		if (node instanceof BoxNode) {
			component = createBox((BoxNode) node, callback);
		} else if (node instanceof ComponentNode) {
			component = createComponent((ComponentNode) node, callback);
		} else if (node instanceof SpacerNode) {
			component = createSpacer((SpacerNode) node);
		}
		if (component != null) {
			setHeight(node, component);
			setWidth(node, component);
		}
		return component;
	}

	private JComponent createBox(BoxNode boxNode, IVisualComponentCallback callback) {
		Box box = new Box(boxNode);
		ILayoutNode[][] children = boxNode.split();
		for (int i = 0; i < children.length; i++) {
			if (children[i].length == 1) {
				box.add(createFromNode(children[i][0], callback));
            } else if (children[i].length == 2) {
                // TODO: this is an iteration 1 workaround for the fact that
                // MultiSplitPane is not 508 compliant. 
                JSplitPane splitPane = createJSplitPane(boxNode);
                ILayoutNode[] nodes = children[i];
                splitPane.setLeftComponent(createFromNode(nodes[0], callback));
                splitPane.setRightComponent(createFromNode(nodes[1], callback));
                double flexLeft = nodes[0].getFlex();
                double flexRight = nodes[1].getFlex();
                double totalFlex = flexLeft + flexRight;
                if (totalFlex > 0)
                    splitPane.setResizeWeight(flexLeft / totalFlex);
                else
                    splitPane.setResizeWeight(0.5);
                box.add(splitPane);
                
			} else {
				MultiSplitPane splitPane = createMultiSplitPane(boxNode);
				ILayoutNode[] nodes = children[i];
				for (int j = 0; j < nodes.length; j++) {
					splitPane.add(createFromNode(nodes[j], callback));
				}
				double[] ratios = new double[nodes.length];
				for (int j = 0; j < ratios.length; j++) {
					ratios[j] = (double) nodes[j].getFlex();
				}
				splitPane.setRatios(ratios);
				box.add(splitPane);
			}
		}
		
		return box;
	}

	private JSplitPane createJSplitPane(BoxNode boxNode) {
        if (boxNode.getOrientation() == SwingConstants.VERTICAL) {
            return new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        } else {
            return new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        }
	}

	private MultiSplitPane createMultiSplitPane(BoxNode boxNode) {
		if (boxNode.getOrientation() == SwingConstants.VERTICAL) {
			return new MultiSplitPane(MultiSplitPane.VERTICAL_SPLIT);
		} else {
			return new MultiSplitPane(MultiSplitPane.HORIZONTAL_SPLIT);
		}
	}

	private JComponent createComponent(ComponentNode componentNode, IVisualComponentCallback callback) {
		IVisualComponentDescriptor descriptor =
			visualComponents.get(componentNode.getVisualComponentId());
		IVisualComponent component = descriptor.createVisualComponent();
		callback.componentCreated(new VisualComponentReference(descriptor, component));
		return component.getComponent();
	}

	private JComponent createSpacer(SpacerNode spacerNode) {
		return new JPanel();
	}

	private void setHeight(ILayoutNode componentNode, JComponent component) {
		if (componentNode.getHeight() != ILayoutNode.UNSPECIFIED) {
			setFixedHeight(componentNode, component);
		} else {
			setMinHeight(componentNode, component);
			setMaxHeight(componentNode, component);
		}
	}

	private void setMaxHeight(ILayoutNode componentNode, JComponent component) {
		if (componentNode.getMaxHeight() != ILayoutNode.UNSPECIFIED) {
			component.setMaximumSize(
				new Dimension(component.getMaximumSize().width, componentNode.getMaxHeight()));
		}
	}

	private void setMinHeight(ILayoutNode componentNode, JComponent component) {
		if (componentNode.getMinHeight() != ILayoutNode.UNSPECIFIED) {
			component.setMinimumSize(
				new Dimension(component.getMinimumSize().width, componentNode.getMinHeight()));
		}
	}

	private void setFixedHeight(ILayoutNode componentNode, JComponent component) {
		component.setMinimumSize(
			new Dimension(component.getMinimumSize().width, componentNode.getHeight()));
		component.setPreferredSize(
			new Dimension(component.getPreferredSize().width, componentNode.getHeight()));
		component.setMaximumSize(
			new Dimension(component.getMaximumSize().width, componentNode.getHeight()));
	}

	private void setWidth(ILayoutNode componentNode, JComponent component) {
		if (componentNode.getWidth() != ILayoutNode.UNSPECIFIED) {
			component.setMinimumSize(
				new Dimension(componentNode.getWidth(), component.getMinimumSize().height));
			component.setPreferredSize(
				new Dimension(componentNode.getWidth(), component.getPreferredSize().height));
			component.setMaximumSize(
				new Dimension(componentNode.getWidth(), component.getMaximumSize().height));
		} else {
			if (componentNode.getMinHeight() != ILayoutNode.UNSPECIFIED) {
				component.setMinimumSize(
					new Dimension(componentNode.getMinWidth(), component.getMinimumSize().height));
			}
			if (componentNode.getMaxHeight() != ILayoutNode.UNSPECIFIED) {
				component.setMaximumSize(
					new Dimension(componentNode.getMaxWidth(), component.getMaximumSize().height));
			}
		}
	}
	
	private interface IVisualComponentCallback {
		void componentCreated(IVisualComponentReference component);
	}
}
