package gov.va.med.hds.cd.prism.visualcomponent.internal;

import java.text.MessageFormat;

import gov.va.med.hds.cd.prism.IPrismConstants;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.InvalidConfigurationElementException;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;

public class VisualComponentRegistryReader implements IPrismConstants {
	private static final String COMPONENT_CREATION_WARNING =
		"unable to create visual component descriptor from configuration element:\n";

	private ILogger log;

	public VisualComponentRegistryReader(ILoggingCapable logging) {
		this.log = logging.getLogger(getClass());
	}

	public VisualComponentRegistry readComponentRegistry(IPluginRegistry pluginRegistry) {
		VisualComponentRegistry result = new VisualComponentRegistry();
		readComponentDescriptors(result, pluginRegistry);
		return result;
	}

	private void readComponentDescriptors(
		VisualComponentRegistry result,
		IPluginRegistry pluginRegistry) {
		IConfigurationElement[] elements =
			pluginRegistry.getConfigurationElementsFor(
				UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID);

		for (int i = 0; i < elements.length; i++) {
			try {
				VisualComponentDescriptor component =
					readComponentDescriptor(elements[i]);
				result.add(component);
			} catch (Exception e) {
				log.error(COMPONENT_CREATION_WARNING, e);
				continue;
			}
		}
	}

	private VisualComponentDescriptor readComponentDescriptor(IConfigurationElement element)
		throws
			InvalidConfigurationElementException,
			IllegalAccessException,
			InstantiationException {
		validateVisualComponentDescriptorElement(element);

		String id = element.getAttribute(ID_ATTRIBUTE);
		String name = element.getAttribute(NAME_ATTRIBUTE);
		String className = element.getAttribute(CLASS_ATTRIBUTE);

		VisualComponentDescriptor c = new VisualComponentDescriptor(element);
		readEventConnections(c, element.getChildren());
		return c;
	}

	private void readEventConnections(
		VisualComponentDescriptor d,
		IConfigurationElement[] children)
		throws InvalidConfigurationElementException {
		for (int i = 0; i < children.length; i++) {
			readEventConnection(d, children[i]);
		}
	}

	private void readEventConnection(
		VisualComponentDescriptor d,
		IConfigurationElement element)
		throws InvalidConfigurationElementException {
		if (SOURCE_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readSourceOfEvent(d, element);
		} else if (LISTENER_CONFIGURATION_ELEMENT.equals(element.getName())) {
			readListenerOfEvent(d, element);
		} else {
			throw new InvalidConfigurationElementException("unexpected element");
		}
	}

	private void readListenerOfEvent(
		VisualComponentDescriptor d,
		IConfigurationElement element)
		throws InvalidConfigurationElementException {
		validateListenerElement(element);
		String eventClassName = element.getAttribute(EVENT_ATTRIBUTE);
		String scope = element.getAttribute(SCOPE_ATTRIBUTE);
		d.addEventListenerDescriptor(eventClassName, scope);
	}

	private void readSourceOfEvent(
		VisualComponentDescriptor d,
		IConfigurationElement element)
		throws InvalidConfigurationElementException {
		validateSourceElement(element);
		String eventClassName = element.getAttribute(EVENT_ATTRIBUTE);
		String scope = element.getAttribute(SCOPE_ATTRIBUTE);
		d.addEventSourceDescriptor(eventClassName, scope);
	}

	private void validateVisualComponentDescriptorElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {
		String plugin =
			element
				.getDeclaringExtension()
				.getDeclaringPluginDescriptor()
				.toString();
		if (element.getAttribute(ID_ATTRIBUTE) == null) {
			Object[] params =
				new Object[] {
					ID_ATTRIBUTE,
					COMPONENT_CONFIGURATION_ELEMENT,
					UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID,
					plugin };
			throw new InvalidConfigurationElementException(
				MessageFormat.format(
					"expected '{0}' attribute in '{1}' conifuration element for extension point '{2}' contributed by plugin '{3}'",
					params));
		}
		if (element.getAttribute(NAME_ATTRIBUTE) == null) {
			Object[] params =
				new Object[] {
					NAME_ATTRIBUTE,
					COMPONENT_CONFIGURATION_ELEMENT,
					UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID,
					plugin };
			throw new InvalidConfigurationElementException(
				MessageFormat.format(
					"expected '{0}' attribute in '{1}' conifuration element for extension point '{2}' contributed by plugin '{3}'",
					params));
		}
		if (element.getAttribute(CLASS_ATTRIBUTE) == null) {
			Object[] params =
				new Object[] {
					CLASS_ATTRIBUTE,
					COMPONENT_CONFIGURATION_ELEMENT,
					UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID,
					plugin };
			throw new InvalidConfigurationElementException(
				MessageFormat.format(
					"expected '{0}' attribute in '{1}' conifuration element for extension point '{2}' contributed by plugin '{3}'",
					params));
		}
	}

	private void validateSourceElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {
		if (element.getAttribute(EVENT_ATTRIBUTE) == null) {
			String plugin =
				element
					.getDeclaringExtension()
					.getDeclaringPluginDescriptor()
					.toString();
			Object[] params =
				new Object[] {
					EVENT_ATTRIBUTE,
					SOURCE_CONFIGURATION_ELEMENT,
					UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID,
					plugin };
			throw new InvalidConfigurationElementException(
				MessageFormat.format(
					"expected '{0}' attribute in '{1}' conifuration element for extension point '{2}' contributed by plugin '{3}'",
					params));
		}

	}

	private void validateListenerElement(IConfigurationElement element)
		throws InvalidConfigurationElementException {

		if (element.getAttribute(EVENT_ATTRIBUTE) == null) {
			String plugin =
				element
					.getDeclaringExtension()
					.getDeclaringPluginDescriptor()
					.toString();
			Object[] params =
				new Object[] {
					EVENT_ATTRIBUTE,
					LISTENER_CONFIGURATION_ELEMENT,
					UNIQUE_VISUAL_COMPONENT_EXTENSION_POINT_ID,
					plugin };
			throw new InvalidConfigurationElementException(
				MessageFormat.format(
					"expected '{0}' attribute in '{1}' conifuration element for extension point '{2}' contributed by plugin '{3}'",
					params));
		}

	}
}
