package gov.va.med.hds.cd.prism.test;

import java.awt.event.ActionEvent;

import javax.swing.Action;

import junit.framework.TestCase;

import gov.va.med.hds.cd.prism.IPrismConstants;
import gov.va.med.hds.cd.prism.action.AbstractAction;
import gov.va.med.hds.cd.prism.action.IActionDescriptor;
import gov.va.med.hds.cd.prism.action.internal.ActionDescriptor;
import gov.va.med.hds.cd.prism.action.internal.ActionRegistry;
import gov.va.med.hds.cd.prism.menu.internal.ActionNode;
import gov.va.med.hds.cd.prism.menu.internal.ActionNotFoundException;
import gov.va.med.hds.cd.prism.menu.internal.InvalidMenuPathException;
import gov.va.med.hds.cd.prism.menu.internal.MenuNode;
import gov.va.med.hds.cd.prism.menu.internal.MenuTree;
import gov.va.med.hds.cd.prism.menu.internal.MenuTreeNode;
import gov.va.med.hds.cd.prism.menu.internal.ParentNode;
import gov.va.med.hds.cd.prism.menu.internal.SeparatorNode;
import gov.va.med.hds.cd.prism.menu.internal.SlotNode;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.internal.ConfigurationElement;
import gov.va.med.hds.cd.runtime.internal.Extension;
import gov.va.med.hds.cd.spectrum.LookAndFeelMenu;

public class MenuTreeTest extends TestCase {

	protected IRuntimeContext runtime;
	protected ActionRegistry actions;
	protected MenuTree menuTree;

	public MenuTreeTest(String name) {
		super(name);
	}

	protected void setUp() throws Exception {
		super.setUp();
		runtime = new MockRuntime();
		actions = new ActionRegistry();
		actions.put(createActionDescriptor("actionId1"));
		actions.put(createActionDescriptor("actionId2"));
		actions.put(createActionDescriptor("actionId3"));
		actions.put(createActionDescriptor("actionId4"));
		actions.put(createActionDescriptor("actionId5"));

		menuTree = new MenuTree(actions);
	}

	protected IConfigurationElement createMenuConfigurationElement(Class c) {
		return new MockMenuConfigurationElement(c);
	}

	private IActionDescriptor createActionDescriptor(String id) {
		ActionDescriptor a = new ActionDescriptor(createActionConfigurationElement(id, TestAction.class));
		return a;
	}

	private IConfigurationElement createActionConfigurationElement(String id, Class c) {
		Extension ext = new Extension(new MockPluginDescriptor(runtime), "a.test.extension.point");
		ConfigurationElement element = new ConfigurationElement(IPrismConstants.ACTION_CONFIGURATION_ELEMENT);
		element.setDeclaringExtension(ext);
		element.addAttribute(IPrismConstants.ID_ATTRIBUTE, id);
		element.addAttribute(IPrismConstants.CLASS_ATTRIBUTE, c.getName());
		element.addAttribute(IPrismConstants.NAME_ATTRIBUTE, id);
		return element;
	}

	public void testRootMenus() {
		try {
			menuTree.addMenu("file", "/", "File", new Character('F'));
			menuTree.addMenu("edit", "/", "Edit");
			menuTree.addMenu("help", "/", "Help");
			menuTree.addMenu("lnf", "/", createMenuConfigurationElement(LookAndFeelMenu.class));

			assertEquals(4, menuTree.getRoot().getChildren().size());

			MenuTreeNode node = menuTree.getNode("/file");
			assertNotNull(node);
			assertTrue(node instanceof MenuNode);
			assertEquals("file", node.getId());
			assertEquals("File", ((MenuNode) node).getName());
			assertEquals('F', ((MenuNode) node).getMnemonic().charValue());
			assertFalse(((MenuNode) node).hasCustomMenuClass());

			node = menuTree.getNode("edit");
			assertNotNull(node);
			assertTrue(node instanceof MenuNode);
			assertEquals("edit", node.getId());
			assertEquals("Edit", ((MenuNode) node).getName());
			assertFalse(((MenuNode) node).hasCustomMenuClass());

			node = menuTree.getNode("help/");
			assertNotNull(node);
			assertTrue(node instanceof MenuNode);
			assertEquals("help", node.getId());
			assertEquals("Help", ((MenuNode) node).getName());
			assertFalse(((MenuNode) node).hasCustomMenuClass());

			node = menuTree.getNode("lnf");
			assertNotNull(node);
			assertTrue(node instanceof MenuNode);
			assertEquals("lnf", node.getId());
			assertTrue(((MenuNode) node).hasCustomMenuClass());
			assertNull(((MenuNode) node).getName());
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		}
	}

	public void testAddAction() {
		try {
			menuTree.addMenu("file", "/", "File");
			menuTree.addAction("actionId1", "file");
			menuTree.addAction("actionId2", "/file");
			menuTree.addAction("actionId3", "file/");

			assertEquals(3, ((ParentNode) menuTree.getNode("file")).getChildren().size());
			MenuTreeNode node = menuTree.getNode("file/actionId1");
			assertNotNull(node);
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId1", node.getId());
			node = menuTree.getNode("file/actionId2");
			assertNotNull(node);
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId2", node.getId());
			node = menuTree.getNode("file/actionId3");
			assertNotNull(node);
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId3", node.getId());
		} catch (Exception e) {
			fail(e.toString());
		}
	}

	public void testSubMenu() {
		try {
			menuTree.addMenu("base", "/", "Base");
			menuTree.addMenu("sub1", "base", "Sub1");
			menuTree.addMenu("sub2", "base/sub1", "Sub2");
			menuTree.addMenu("sub3", "base/sub1/sub2", "Sub3");
			menuTree.addMenu("sub4", "base/sub1/sub2/sub3", "Sub4");

			assertNotNull(menuTree.getNode("/base"));
			assertNotNull(menuTree.getNode("/base/sub1"));
			assertNotNull(menuTree.getNode("/base/sub1/sub2"));
			assertNotNull(menuTree.getNode("base/sub1/sub2/sub3/"));
			assertNotNull(menuTree.getNode("base/sub1/sub2/sub3/sub4"));
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		}
	}

	public void testSlot() {
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addSlot("slot1", "/");
			menuTree.addMenu("base2", "/", "Base2");
			menuTree.addMenu("sub1", "/slot1", "Sub1");
			menuTree.addMenu("sub2", "/slot1", "Sub2");
			menuTree.addSlot("slot2", "base1");
			menuTree.addSlot("slot3", "base2");
			menuTree.addMenu("sub3", "/base1/slot2", "Sub3");
			menuTree.addMenu("sub4", "/base2/slot3", "Sub4");
			menuTree.addAction("actionId1", "base1/slot2");
			menuTree.addAction("actionId5", "/base2/slot3");
			ParentNode node = (ParentNode) menuTree.getNode("slot1");
			assertNotNull(node);
			assertTrue(node instanceof SlotNode);
			assertEquals(2, node.getChildren().size());
			node = (ParentNode) menuTree.getNode("base1/slot2");
			assertNotNull(node);
			assertTrue(node instanceof SlotNode);
			assertEquals(2, node.getChildren().size());
			node = (ParentNode) menuTree.getNode("base1/slot3");
			assertNotNull(node);
			assertTrue(node instanceof SlotNode);
			assertEquals(2, node.getChildren().size());
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		}
	}

	public void testSeparator() {
		try {
			menuTree.addMenu("file", "/", "File");
			menuTree.addAction("actionId1", "file");
			menuTree.addSeparator("separator1", "file");
			menuTree.addAction("actionId2", "/file");

			assertEquals(3, ((ParentNode) menuTree.getNode("file")).getChildren().size());
			MenuTreeNode node = menuTree.getNode("file/actionId1");
			assertNotNull(node);
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId1", node.getId());
			node = menuTree.getNode("file/separator1");
			assertNotNull(node);
			assertTrue(node instanceof SeparatorNode);
			assertEquals("separator1", node.getId());
			node = menuTree.getNode("file/actionId2");
			assertNotNull(node);
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId2", node.getId());
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		}
	}

	public void testSlice() throws InvalidMenuPathException, ActionNotFoundException {
		menuTree.addMenu("file", "/", "File");
		menuTree.addAction("slice1", "actionId1", "file");
		menuTree.addSeparator("separator1", "file");
		menuTree.addMenu("slice2", "menu1", "file", "SubMenu");
		menuTree.addAction("slice2", "actionId2", "file/menu1");
		menuTree.addAction("actionId3", "/file");

		assertEquals(4, ((ParentNode) menuTree.getNode("file")).getChildren().size());
		MenuTreeNode node;
		try {
			node = menuTree.getNode("file/actionId1");
			fail("expected invalid menu path exception");
		} catch (InvalidMenuPathException e) {
		}
		node = menuTree.getNode("file/actionId1", "slice1");
		assertNotNull(node);
		assertTrue(node instanceof ActionNode);
		assertEquals("actionId1", node.getId());
		node = menuTree.getNode("file/separator1");
		assertNotNull(node);
		assertTrue(node instanceof SeparatorNode);
		assertEquals("separator1", node.getId());
		try {
			node = menuTree.getNode("file/menu1/actionId2");
			fail("expected invalid menu path exception");
		} catch (InvalidMenuPathException e) {
		}
		node = menuTree.getNode("file/menu1/actionId2", "slice2");
		assertNotNull(node);
		assertTrue(node instanceof ActionNode);
		assertEquals("actionId2", node.getId());
	}

	public void testAddMenuFailure() {
		try {
			menuTree.addMenu("base1", "", "Base1");
			fail("expected invalid menu path exception");
		} catch (InvalidMenuPathException e) {
		}

		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addMenu("sub1", "/base1/invalidpath1/invalidpath2", "Sub1");
			fail("expected invalid menu path exception");
		} catch (InvalidMenuPathException e) {
		}

		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("actionId1", "base1");
			menuTree.addMenu("sub1", "base1/actionId1", "Sub1");
			fail("expected invalid menu path exception");
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		} catch (InvalidMenuPathException e) {
		}
	}

	public void testAddActionFailure() {
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("actionId1", "base1");
			menuTree.addAction("actionId2", "base1/actionId1");
			fail("expected invalid menu path exception");
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		} catch (InvalidMenuPathException e) {
		}
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("non-existentaction", "base1");
			fail("expected invalid action id path exception");
		} catch (InvalidMenuPathException e) {
			fail("expected invalid menu path exception");
		} catch (ActionNotFoundException e) {
		}
	}

	public void testAddSlotFailure() {
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("actionId1", "base1");
			menuTree.addSlot("slot1", "base1/actionId1");
			fail("expected invalid menu path exception");
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		} catch (InvalidMenuPathException e) {
		}
	}

	public void testAddSeparatorFailure() {
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("actionId1", "base1");
			menuTree.addSeparator("separator1", "base1/actionId1");
			fail("expected invalid menu path exception");
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		} catch (InvalidMenuPathException e) {
		}
	}

	public void testClone() {
		try {
			menuTree.addMenu("base1", "/", "Base1");
			menuTree.addAction("actionId1", "base1");
			menuTree.addSeparator("separator1", "base1");
			menuTree.addAction("actionId2", "base1");
			menuTree.addMenu("sub1", "base1", "SubMenu1");
			menuTree.addAction("actionId3", "base1/sub1");
		} catch (ActionNotFoundException e) {
			fail(e.toString());
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		}

		MenuTree clonedTree = (MenuTree) menuTree.clone();
		try {
			MenuTreeNode node = clonedTree.getNode("base1");
			assertTrue(node instanceof MenuNode);
			assertEquals("base1", node.getId());
			assertEquals("Base1", ((MenuNode) node).getName());
			assertNotSame(node, menuTree.getNode("base1"));

			node = clonedTree.getNode("base1/actionId1");
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId1", node.getId());
			assertNotSame(node, menuTree.getNode("base1/actionId1"));

			node = clonedTree.getNode("base1/separator1");
			assertTrue(node instanceof SeparatorNode);
			assertEquals("separator1", node.getId());
			assertNotSame(node, menuTree.getNode("base1/separator1"));

			node = clonedTree.getNode("base1/actionId2");
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId2", node.getId());
			assertNotSame(node, menuTree.getNode("base1/actionId2"));

			node = clonedTree.getNode("base1/sub1");
			assertTrue(node instanceof MenuNode);
			assertEquals("sub1", node.getId());
			assertEquals("SubMenu1", ((MenuNode) node).getName());
			assertNotSame(node, menuTree.getNode("base1/sub1"));

			node = clonedTree.getNode("base1/sub1/actionId3");
			assertTrue(node instanceof ActionNode);
			assertEquals("actionId3", node.getId());
			assertNotSame(node, menuTree.getNode("base1/sub1/actionId3"));
		} catch (InvalidMenuPathException e) {
			fail(e.toString());
		}
	}

	public static class TestAction extends AbstractAction {

		public TestAction() {
			super();
		}

		public void actionPerformed(ActionEvent e) {
		}
		
		
		/* @see gov.va.med.hds.cd.runtime.IExecutableExtension#setInitializationData(gov.va.med.hds.cd.runtime.IRuntimeContext, gov.va.med.hds.cd.runtime.IConfigurationElement, java.lang.String, java.lang.Object) */
		public void setInitializationData(
			IRuntimeContext runtime,
			IConfigurationElement config,
			String propertyName,
			Object data)
			throws Exception {
			super.setInitializationData(runtime, config, propertyName, data);
			putValue(Action.NAME, getDescriptor().getId());
		}

	}
	
	private static class MockMenuConfigurationElement implements IConfigurationElement {

		private Class menuClass;

		public MockMenuConfigurationElement(Class c) {
			menuClass = c;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#createExecutableExtension(java.lang.String) */
		public Object createExecutableExtension(String propertyName) throws Exception {
			return createExecutableExtension(propertyName, null);
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#createExecutableExtension(java.lang.String, java.lang.Object) */
		public Object createExecutableExtension(String propertyName, Object data) throws Exception {
			return menuClass.newInstance();
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getAttribute(java.lang.String) */
		public String getAttribute(String name) {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getAttributeAsIs(java.lang.String) */
		public String getAttributeAsIs(String name) {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getAttributeNames() */
		public String[] getAttributeNames() {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getChildren() */
		public IConfigurationElement[] getChildren() {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getChildren(java.lang.String) */
		public IConfigurationElement[] getChildren(String name) {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getDeclaringExtension() */
		public IExtension getDeclaringExtension() {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getName() */
		public String getName() {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getValue() */
		public String getValue() {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IConfigurationElement#getValueAsIs() */
		public String getValueAsIs() {
			return null;
		}
		
	}
}
