package gov.va.med.hds.cd.prism;

import gov.va.med.hds.cd.prism.action.IActionRegistry;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveEventSource;
import gov.va.med.hds.cd.prism.perspective.IPerspectiveRegistry;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentEventSource;
import gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;

/**
 * An <code>IDesktop</code> is the root object for the Health<i>e</i>Vet Desktop
 * user interface. (It is sometimes also referred to as "Prism", or just
 * desktop)
 *
 * An <code>IDesktop</code> has one or more main windows which present to the end user 
 * information based on some underlying model.
 * 
 * An <code>IDesktop</code> usually starts with a single open window, and automatically
 * closes when its last window closes.
 *
 * Each dekstop window has a collection of perspectives; the active perspective is the one that
 * is being presented to the end user; at most one perspective is active in a window at a time.
 *
 * Each perspective has a collection of visual components. A perspective's components are 
 * arranged (tiled or stacked) for presentation on the screen. The arrangement is not fixed; 
 * Perspectives can be customized so that users can have arrangements of components as they see
 * fit. A perspective is a template for a page, capturing a collection of parts and their 
 * arrangement.
 *
 * The runtime creates an <code>IDesktop</code> when the Prism plug-in is activated; 
 * since this happens at most once during the life of the running platform, there is only one 
 * <code>IDesktop</code> instance. Due to its singular nature, it is commonly referred
 * to as the desktop.
 *
 * This interface is not intended to be implemented by clients.
 */
public interface IDesktop extends IDesktopWindowEventSource {
	
	/**
	 * Adds a listener to the desktop that's notified of desktop lifecycle
	 * events
	 * @param l the <code>IDesktopListener</code> to add.
	 */
	public void addDesktopListener(IDesktopListener l);
	
	/**
	 * Removes a listener to the desktop that's notified of desktop lifecycle
	 * events
	 * @param l the <code>IDesktopListener</code> to remove.
	 */
	public void removeDesktopListener(IDesktopListener l);
	
	/**
	 * Gets the runtime context of the desktop.
	 * 
	 * @return a reference to the instance of <code>IRuntimeContext</code> associated with this
	 * desktop.
	 */
	IRuntimeContext getRuntime();

	/**
	 * Gets the perspective registry of the desktop.
	 * 
	 * @return a reference to the instance of <code>IPerspectiveRegistry</code> managed by this
	 * desktop.
	 */
	IPerspectiveRegistry getPerspectiveRegistry();

	/**
	 * Gets the visual component registry of the desktop.
	 * 
	 * @return a reference to the instance of <code>IVisualComponentRegistry</code> managed by this
	 * desktop.
	 */
	IVisualComponentRegistry getVisualComponentRegistry();

	/**
	 * Gets the action registry of the desktop.
	 * 
	 * @return a reference to the instance of <code>IActionRegistry</code> managed by this
	 * desktop.
	 */
	IActionRegistry getActionRegistry();

	/**
	 * Gets the desktop-wide perspective event source.
	 * 
	 * Clients interested in perspective status change events across the entire desktop
	 * can register listeners to the object returned by this method.
	 * 
	 * @return a reference to a <code>IPerspectiveEventSource</code> for this desktop.
	 * 
	 * @see gov.va.med.hds.cd.prism.perspective.PerspectiveEvent
	 * @see gov.va.med.hds.cd.prism.perspective.IPerspectiveListener
	 * @see gov.va.med.hds.cd.prism.perspective.PerspectiveAdapter
	 */
	IPerspectiveEventSource getPerspectiveEventSource();
	
	/**
	 * Gets the desktop-wide visual component event source.
	 * 
	 * Clients interested in visual component status change events across the entire desktop
	 * can register listeners to the object returned by this method.
	 * 
	 * @return a reference to a <code>IVisualComponentEventSource</code> for this desktop.
	 * 
	 * @see gov.va.med.hds.cd.prism.visualcomponent.VisualComponentEvent
	 * @see gov.va.med.hds.cd.prism.visualcomponent.IVisualComponentListener
	 */
	IVisualComponentEventSource getVisualComponentEventSource();
	
	/**
	 * Returns a list of the open main windows associated with this desktop. Note that
	 * wizards and dialogs are not included in this list since they are not considered main
	 * windows. 
	 * 
	 * @return a list of open windows
	 */
	IDesktopWindow[] getWindows();
	
	/**
	 * Creates and opens a new desktop window with one perspective. The perspective is defined
	 * by the default perspective ID. The new window and new perspective become active. 
	 * 
	 * The desktop also defines a number of menu items to activate or open each registered 
	 * perspective. A complete list of these perspectives is available from the perspective
	 * registry found on <code>IDesktop</code>.
	 * 
	 * @param perspectiveID the id of the perspective to initially make active.
	 * @return the new workbench window 
	 * 
	 * @see IDesktop.getPerspectiveRegistry()
	 */
	IDesktopWindow openWindow(String perspectiveID);

	/**
	 * Closes this desktop and all its open windows.
	 * 
	 * @throws DesktopCloseVetoException when the attempt to close the 
     *            desktop is vetoed by one of the desktop windows or a desktop listener.
	 */
	void close() throws DesktopCloseVetoException;
}
