/*
 * Initially Created on Apr 24, 2003 by vhaislmonsos
 *
 */
package gov.va.med.hds.querytool;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.querytool.dao.IPredefinedQueryListDAO;
import gov.va.med.hds.querytool.dao.IQueryDAO;
import gov.va.med.hds.querytool.dao.IUserQueryListDAO;
import gov.va.med.hds.querytool.dao.QueryDAOException;

import javax.swing.event.EventListenerList;

/**
 * Implements the ICombinedQueryCollection to provide a query collection model.
 */
public class CombinedQueryCollection implements ICombinedQueryCollection {
	private IQueryList userQueries = new AbstractQueryList();
	private IQueryList predefQueries = new AbstractQueryList();
	private IUserQueryListDAO userListDAO;
	private IPredefinedQueryListDAO predefListDAO;
	private IQueryDAO queryDAO;
	private EventListenerList listenerList = new EventListenerList();
	private IQueryListListener userListListener = new IQueryListListener() {
		public void listChanged(IQueryListEvent event) {
			fireUserItemsUpdate(event);
		}
	};
	private IQueryListListener predefListListener = new IQueryListListener() {
		public void listChanged(IQueryListEvent event) {
			firePredefItemsUpdate(event);
		}
	};

	public CombinedQueryCollection() {
		super();
		userQueries.addListener(userListListener);
		predefQueries.addListener(predefListListener);
	}

	/**
	 * Informs listeners of a predefined report collection update.
	 * @param event
	 */
	protected void firePredefItemsUpdate(IQueryListEvent event) {
		if (listenerList.getListenerCount()>0) {
			IQueryCollectionEvent e = new QueryCollectionEvent(this);
			if (event.getEventID()==IQueryListEvent.INTERVAL_ADDED ||
				event.getEventID()==IQueryListEvent.INTERVAL_CHANGED) {
				e.setEventID(IQueryCollectionEvent.PREDEFINED_ITEMS_ADDED);
			}
			else {
				e.setEventID(IQueryCollectionEvent.PREDEFINED_ITEMS_REMOVED);
			}
			fireChangeEvent(e);
		}
	}

	/**
	 * Informs listeners of a user report collection update.
	 * @param event
	 */
	protected void fireUserItemsUpdate(IQueryListEvent event) {
		if (listenerList.getListenerCount()>0) {
			IQueryCollectionEvent e = new QueryCollectionEvent(this);
			if (event.getEventID()==IQueryListEvent.INTERVAL_ADDED ||
				event.getEventID()==IQueryListEvent.INTERVAL_CHANGED) {
				e.setEventID(IQueryCollectionEvent.USER_ITEMS_ADDED);
			}
			else {
				e.setEventID(IQueryCollectionEvent.USER_ITEMS_REMOVED);
			}
			fireChangeEvent(e);
		}
	}
	
	private void fireChangeEvent(IQueryCollectionEvent event) {
		if (listenerList.getListenerCount()>0) {
			IQueryCollectionListener[] listeners = (IQueryCollectionListener[])listenerList.getListeners(IQueryCollectionListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].collectionChanged(event);
			}
		}
	}

	public void addChangeListener(IQueryCollectionListener l) {
		listenerList.add(IQueryCollectionListener.class, l);
	}

	public boolean containsDisplayName(String displayName) {
		if (predefQueries.indexOf(displayName)>-1) {
			return true;
		}
		if (userQueries.indexOf(displayName)>-1) {
			return true;
		}
		return false;
	}

	public ISoftListItem getUserQueryItem(int index) {
		return userQueries.getItemAt(index);
	}

	public ISoftListItem getPredefinedQueryItem(int index) {
		return predefQueries.getItemAt(index);
	}

	public boolean isPredefinedItem(ISoftListItem item) {
		return predefQueries.indexOf(item)>-1;
	}

	public boolean isUserQuery(ISoftListItem item) {
		return userQueries.indexOf(item)>-1;
	}

	public void reloadChangedQuery(ISoftListItem item)
		throws QueryCollectionException {
		if (item.getID().equals(QueryID.nullID)) return;
		if (item.getID().equals(QueryID.unsavedID)) return;
		if (item.getItem()==null) return;
	
		IQuery query = (IQuery)item.getItem();
		if (query.isChanging() || (!query.isQueryChanged())) return;

		try {
			loadQuery(item);
		}
		catch (QueryDAOException e) {
			throw new QueryCollectionException("Error reloading " + query.getDisplayName());
		}		
	}

	public void removeChangeListener(IQueryCollectionListener l) {
		listenerList.remove(IQueryCollectionListener.class, l);
	}

	public void removeUserItem(ISoftListItem item) throws QueryCollectionException {
		try {
			IQueryID queryID = (IQueryID)item.getID();
			if (QueryID.isValid(queryID)) {
				queryDAO.delete(queryID);
			}
			userQueries.removeElement(item);
		}
		catch (Exception e) {
			throw new QueryCollectionException("Error removing query:"+item.getDisplayName());
		}
	}

	public void setRuntime(IRuntimeContext runtime) throws QueryCollectionException {
		if (predefListDAO==null || userListDAO==null) {
			try {
				IDaoRegistry daoRegistry = (IDaoRegistry)runtime.getPluginInstance(IDaoRegistry.PLUGIN_ID);
				predefListDAO = (IPredefinedQueryListDAO)daoRegistry.getDAO(IPredefinedQueryListDAO.class.getName());
				userListDAO = (IUserQueryListDAO)daoRegistry.getDAO(IUserQueryListDAO.class.getName());
			}
			catch (Exception e) {
				throw new QueryCollectionException("Error in setRuntime");
			}
		}
	}

	public int getSize() {
		return userQueries.getSize()+predefQueries.getSize();
	}

	public int getUserDefinedQuerySize() {
		return userQueries.getSize();
	}

	public int getPredefinedQuerySize() {
		return predefQueries.getSize();
	}

	public void setQueryDAO(IQueryDAO queryDAO) {
		this.queryDAO = queryDAO;
	}

	private void loadQuery(ISoftListItem item) throws QueryDAOException {
		if (!QueryID.isValid((IQueryID)item.getID())) {
			throw new QueryDAOException("Error: requesting null query ID from the database.");
		}
		IQuery query = queryDAO.getAsync((IQueryID)item.getID(), item.getDisplayName());
		query.init();
		item.setItem(query);
	}

	public boolean isQueryItem(Object item) {
		return item instanceof ISoftListItem && ((predefQueries.indexOf((ISoftListItem)item)>-1) || (userQueries.indexOf((ISoftListItem)item)>-1));
	}

	public IQuery getQuery(Object item) throws QueryCollectionException {
		if (!(item instanceof ISoftListItem)) {
			throw new QueryCollectionException("Invalid item used to try to get a query.");
		}
		
		ISoftListItem listItem = (ISoftListItem)item;
		if (QueryID.isValid((IQueryID)listItem.getID()) && listItem.getItem()==null) {
			try {
				loadQuery(listItem);
			}
			catch (QueryDAOException e) {
				throw new QueryCollectionException();
			}
		}
		
		return (IQuery)listItem.getItem();
	}

	public void saveQuery(IQuery query) throws QueryCollectionException {
		if (!QueryID.isValid(query.getQueryID()) || 
			((userQueries.indexOf(query)>-1) && (query.isQueryChanged())) ) {
			try {
				queryDAO.save(query);
			}
			catch (QueryDAOException e) {
				throw new QueryCollectionException("Error in saveQuery:"+query.getDisplayName());
			}
		}
				
	}

	public void refresh() throws QueryCollectionException {
		try {
			userListDAO.refresh(userQueries);
			predefListDAO.refresh(predefQueries);
		}
		catch (Exception e) {
			throw new QueryCollectionException("Error in refresh.");
		}
	}

	public void addQueryWithoutSave(IQuery query) throws QueryCollectionException {
		try {
			userQueries.addElement(query);
		}
		catch (QueryListException e) {
			throw new QueryCollectionException("Error adding without save:"+query.getDisplayName());
		}
	}

	public boolean isPredefinedQuery(IQuery query) {
		return predefQueries.indexOf(query)>-1;
	}

	public void removeUserQuery(IQuery query) throws QueryCollectionException {
		int index = userQueries.indexOf(query);
		if (index>-1) {
			ISoftListItem item = userQueries.getItemAt(index);
			removeUserItem(item);
		}
	}
	
}
