package gov.va.med.hds.querytool;

import gov.va.med.hds.util.mumps.NameValuePair;

import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.EventListener;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.event.EventListenerList;

/**
 * Criteria set for a query definition.
 */
public class Constraint implements IConstraint {
	private EventListenerList eventListeners = new EventListenerList();
	private HashMap relationshipTree = new HashMap();
	private Map childMap = new HashMap();
	private int changeCount;
	private PropertyChangeListener valueListener = new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent evt) {
			processValueChange(evt);
		}
	};

	private class ConstraintIterator implements Iterator {
		private int count = changeCount;
		private LinkedList rootList;
		private List array;
		private int pos = -2;
		private IValue con;
		private IValue oldCon;

		public ConstraintIterator() {
			super();
			rootList = new LinkedList();
			rootList.add(IConstraint.ROOT);
			getNext();
		}

		private void getNext() {
			checkChanges();
			if (rootList.size()<1) {
				con = null;
			}
			else {
				if (array!=null) {
					pos++;
					if (pos<array.size()) {
						con = (IValue)array.get(pos);
					}
					else {
						array = null;
						getNextNode();
						getNext();
					}
				}
				else {
					if (childMap.containsKey(rootList.getLast())) {
						array = (List)childMap.get(rootList.getLast());
						pos = -1;
						getNext();
					}
					else {
						getNextNode();
						getNext();
					}
				}
			}
		}

		private void getNextNode() {
			checkChanges();
			if (relationshipTree.containsKey(rootList.getLast())) {
				List list = (List)relationshipTree.get(rootList.getLast());
				rootList.add(list.get(0));
			}
			else {
				while (true) {
					Object endItem = rootList.removeLast();
					if (rootList.size()<1) return;
					Object item = rootList.getLast();
					List list = (List)relationshipTree.get(item);
					int index = list.indexOf(endItem) + 1;
					if (index < list.size()) {
						rootList.add(list.get(index));
						break;
					}
				}
			}
		}

		public boolean hasNext() {
			checkChanges();
			return con!=null;
		}

		public Object next() {
			checkChanges();
			oldCon = con;
			getNext();
			return oldCon;
		}

		public void remove() {
			throw new RuntimeException(new UnsupportedOperationException("ConstraintIterator does not support the remove method."));
		}
		
		private void checkChanges() {
			if (count!=changeCount) {
				throw new RuntimeException(new ConcurrentModificationException("ConstraintIterator failed due to the underlying model changing."));
			}
		}
	}
	
	public void addDataNeededListener(IParmDataNeededListener d) {
		eventListeners.add(IParmDataNeededListener.class,d);
		changeCount++;
	}

	private void processValueChange(PropertyChangeEvent e) {
		firePropertyChange(PARM_CHANGE, e.getOldValue(), e.getNewValue());
	}

	public void removeDataNeededListener(IParmDataNeededListener d) {
		eventListeners.remove(IParmDataNeededListener.class,d);
		changeCount++;
	}

	public Constraint() {
		super();
	}
	
	public void addValue (IValue value) {
		if (!childMap.containsKey(value.getID())) {
			childMap.put(value.getID(),new ArrayList());
		}
		List list = (List)childMap.get(value.getID());
		list.add(value);
		changeCount++;
		value.addPropertyChangeListener(valueListener);
		firePropertyChange(IConstraint.PARM_ADD, null, value);
	}

	public void getSetTags(List list) {
		list.addAll(childMap.keySet());
	}

	public List clearTag(Object tag) {
		List result = new ArrayList();
		clearTags(tag,result);
		changeCount++;
		firePropertyChange(IConstraint.PARM_REMOVE, result, null);
		return result;
	}
	
	private void clearTags(Object tag, List removedList) {
		if (childMap.containsKey(tag)) {
			List itemList = (List)childMap.get(tag);
			if ((itemList!=null) && (itemList.size()>0)) {
				for (Iterator i = itemList.iterator(); i.hasNext();) {
					((IValue)i.next()).clearListeners();
				}
			}
			removedList.add(tag);
			childMap.remove(tag);
		}
		List list = (List)relationshipTree.get(tag);
		if (list!=null) {
			int size = list.size();
			if (size>0) {
				for (int i = 0; i < size; i++) {
					clearTags(list.get(i),removedList);
				}
			}
		}
		changeCount++;
	}
	
	public void setTag(Object tag) {
		if (!childMap.containsKey(tag)) {
			fireDataNeeded(new ActionEvent(this,ActionEvent.ACTION_PERFORMED,"parm."+tag.toString()));
		}
	}

	public void fireDataNeeded(ActionEvent e) {
		if (eventListeners.getListenerCount(IParmDataNeededListener.class)>0) {
			EventListener[] s = eventListeners.getListeners(IParmDataNeededListener.class);
			for (int i = 0; i < s.length; i++) {
				IParmDataNeededListener listener = (IParmDataNeededListener)s[i];
				listener.actionPerformed(e);
			}
		}
	}
	
	public void addRelationship(String text) {
		NameValuePair item = new NameValuePair(text);
		String parent = item.getLeft();
		String child  = item.getRight();
		if (!relationshipTree.containsKey(parent)) {
			List list = new ArrayList();
			relationshipTree.put(parent,list);
		}
		List items = (List)relationshipTree.get(parent);
		items.add(child);
		firePropertyChange(IConstraint.PARM_ADD, null, child);
		changeCount++;
	}

	public void clearAll() {
		for (Iterator i = childMap.values().iterator(); i.hasNext();) {
			List list = (List)i.next();
			if ((list!=null) && (list.size()>0)) {
				for (Iterator j = list.iterator(); j.hasNext();) {
					((IValue)j.next()).clearListeners();
				}
			}
		}
		childMap.clear();
		firePropertyChange(IConstraint.PARM_REMOVE, null, null);
		changeCount++;
	}

	public void clearRelationships() {
		relationshipTree.clear();
		firePropertyChange(IConstraint.PARM_REMOVE, null, null);
		changeCount++;
	}

	public boolean isValueSetComplete() {
		boolean result = true;
		for (Iterator i = childMap.values().iterator(); i.hasNext()&&result;) {
			List items = (List) i.next();
			if ((items!=null) || (items.size()>1)) {
				for (int j = 0; j < items.size(); j++) {
					IValue item = (IValue)items.get(j);
					if (!item.isReadOnly()) {
						if ((item.getValues()==null) || (item.getValues().size()<1)) {
							result = false;
							break;
						}
					}
				}
			}
		}
		return result;
	}

	public List getCriterion(Object tag) {
		return (List)childMap.get(tag);
	}

	public Iterator iterator() {
		return new ConstraintIterator();
	}
	
	public List getDisplayFieldTypes() {
		List list = new ArrayList();
		for (Iterator i = iterator(); i.hasNext();) {
			IValue value = (IValue)i.next();
			String constraintName = null;
			if (value.getDisplayFieldCategory()!=null) {
				constraintName = value.getDisplayFieldCategory().toString();
				if (!list.contains(constraintName)) {
					list.add(constraintName);
				}
			}
		}
		return list;
	}

	public void addPropertyChangeListener(PropertyChangeListener listener) {
		eventListeners.add(PropertyChangeListener.class, listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener) {
		eventListeners.remove(PropertyChangeListener.class, listener);
	}
	
	private void firePropertyChange(String changeID, Object oldValue, Object newValue) {
		if (eventListeners.getListenerCount(PropertyChangeListener.class)>0) {
			PropertyChangeEvent event = new PropertyChangeEvent(this, changeID, oldValue, newValue);
			PropertyChangeListener[] listeners = (PropertyChangeListener[])eventListeners.getListeners(PropertyChangeListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].propertyChange(event);
			}
		}
	}

}
