package gov.va.med.hds.querytool;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.EventListenerList;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

/**
 * Contains a list of fields.  See the IDisplayFieldList interface for a 
 * description
 */
public class DisplayFieldList implements IDisplayFieldList {
	private List fields = new ArrayList();
	private EventListenerList propertyListeners = new EventListenerList();
	private boolean changing;
	private PropertyChangeListener itemListener = new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent e) {
			processItemChange(e);
		}
	};

	public boolean add(IDisplayField displayField) {
		if (!fields.contains(displayField)) {
			int startSize = fields.size();
			fields.add(displayField);
			displayField.addPropertyChangeListener(itemListener);
			if (!changing && (startSize != fields.size())) {
				fireListChanged(ListDataEvent.INTERVAL_ADDED,startSize,fields.size()-1);
			}
			return true;
		}
		else {
			return false;
		}
	}

	protected void processItemChange(PropertyChangeEvent e) {
		int pos = fields.indexOf(e.getSource());
		fireListChanged(ListDataEvent.CONTENTS_CHANGED, pos, pos);
	}

	public void addListDataListener(ListDataListener l) {
		propertyListeners.add(ListDataListener.class, l);
	}

	public IDisplayField get(int index) {
		if (index<fields.size() && index>-1) {
			return (IDisplayField)fields.get(index);
		}
		return null;
	}

	public int size() {
		return fields.size();
	}

	public void removeListDataListener(ListDataListener l) {
		propertyListeners.remove(ListDataListener.class, l);
	}

	public int contains(IDisplayField displayField) {
		return fields.indexOf(displayField);
	}

	/**
	 * Signals a list change event to the registered IDisplayFieldList model 
	 * listeners.
	 * @param eventType
	 * @param startPos
	 * @param endPos
	 */
	protected void fireListChanged(int eventType, int startPos, int endPos) {
		if (propertyListeners.getListenerCount() > 0) {
			ListDataListener[] listeners = (ListDataListener[]) propertyListeners.getListeners(ListDataListener.class);
			ListDataEvent e = new ListDataEvent(this, eventType, startPos, endPos);
			for (int i = 0; i < listeners.length; i++) {
				switch (eventType) {
					case ListDataEvent.INTERVAL_ADDED :
						{
							listeners[i].intervalAdded(e);
							break;
						}
					case ListDataEvent.INTERVAL_REMOVED :
						{
							listeners[i].intervalRemoved(e);
							break;
						}
					case ListDataEvent.CONTENTS_CHANGED :
						{
							listeners[i].contentsChanged(e);
							break;
						}
				}

			}
		}
	}

	public void setChangingModel(boolean changing) {
		boolean oldChange = this.changing;
		this.changing = changing;
		if (oldChange && !changing) {
			fireListChanged(ListDataEvent.CONTENTS_CHANGED,0,fields.size()-1);
		}
	}

	public void clearAll() {
		int oldSize = fields.size();
		fields.clear();
		if (oldSize>0) {
			fireListChanged(ListDataEvent.INTERVAL_REMOVED,0,oldSize-1);
		}
	}

	public boolean remove(IDisplayField field) {
		int pos;
		if ((pos=fields.indexOf(field))>-1) {
			removeAt(pos);
		}
		return pos>-1;
	}

	public IDisplayField removeAt(int position) {
		if (position>fields.size()-1 || position<0) return null;
		IDisplayField field = null;
		if (position > -1 && position < fields.size()) {
			field = (IDisplayField)fields.remove(position);
			field.removePropertyChangeListener(itemListener);
			fireListChanged(ListDataEvent.INTERVAL_REMOVED,position,position);
		}
		return field;
	}

	public int indexOf(IDisplayField item) {
		return this.fields.indexOf(item);
	}

	public void moveUp(int index) {
		if ((fields.size()>1) && (index>0) && (index<fields.size())) {
			IDisplayField field = (IDisplayField)fields.remove(index);
			fields.add(--index,field);
			fireListChanged(ListDataEvent.CONTENTS_CHANGED,index+1,index);
		}
	}

	public void moveDown(int index) {
		if ((fields.size()>1) && (index<fields.size()-1) && (index>-1)) {
			IDisplayField field = (IDisplayField)fields.remove(index);
			fields.add(++index,field);
			fireListChanged(ListDataEvent.CONTENTS_CHANGED,index-1,index);
		}
	}

}
