package gov.va.med.hds.querytool;

import gov.va.med.hds.querytool.types.Empty;

import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.EventListener;
import java.util.EventObject;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.event.EventListenerList;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

/**
 * IQuery implementation for a query definition.  A query contains parameters 
 * (constraints), display fields, and a query name, ID, and owner.
 */
public class Query implements IQuery {
	private IQueryID queryID;
	private String displayName = "";
	private String name = "";
	private boolean queryChanged = false;
	private Object owner;
	private IConstraint constraints = new Constraint();
	private IDisplayFieldList displayFields;
	private EventListenerList eList = new EventListenerList();
	private boolean hydrateRequested = false;
	private boolean changing = false;
	private PropertyChangeListener valueChangeListener = new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent e) {
			processParameterChange(e);
		}
	};
	private ListDataListener displayFieldsListener = new ListDataListener() {
		public void intervalAdded(ListDataEvent e) {
			firePropertyChanged(IQuery.DISP_FIELD_ADD, null, null);
		}

		public void intervalRemoved(ListDataEvent e) {
			firePropertyChanged(IQuery.DISP_FIELD_REMOVE, null, null);
		}

		public void contentsChanged(ListDataEvent e) {
			firePropertyChanged(IQuery.DISP, null, null);
		}
	};
	
	/**
	 * Hide the default contructor.
	 */
	private Query() {
	}

	private void processParameterChange(PropertyChangeEvent e) {
		String propertyName = null;
		if (e.getPropertyName().equalsIgnoreCase(IConstraint.PARM_REMOVE_ALL)) {
			displayFields.clearAll();
			propertyName = IQuery.PARM_CLEARALL;
		}
		else if (e.getPropertyName().equalsIgnoreCase(IConstraint.PARM_REMOVE)) {
			List types = constraints.getDisplayFieldTypes();
			List removeList = new LinkedList();
			for (int i = 0; i < displayFields.size(); i++) {
				IDisplayField f = displayFields.get(i);
				if (!types.contains(f.getCategory())) {
					removeList.add(f);
				}
			}
			for (Iterator i = removeList.iterator(); i.hasNext();) {
				displayFields.remove((IDisplayField)i.next());
			}
			propertyName = IQuery.PARM_CLEAR;
		}
		else if (e.getPropertyName().equalsIgnoreCase(IConstraint.PARM_ADD)) {
			propertyName = IQuery.PARM_ADD;
		}
		else if (e.getPropertyName().equalsIgnoreCase(IConstraint.PARM_CHANGE)) {
			propertyName = IQuery.PARM;
		}
		firePropertyChanged(propertyName, e.getOldValue(), e.getNewValue());
	}

	/**
	 * Create a new query with an IQueryID and an owner.
	 * @param queryID
	 * @param user
	 */
	public Query(IQueryID queryID, Object owner) {
		Empty empty = new Empty();
		constraints.addValue(empty);
		constraints.addPropertyChangeListener(valueChangeListener);
		displayFields = new DisplayFieldList();
		displayFields.addListDataListener(displayFieldsListener);
		constraints.addDataNeededListener(new IParmDataNeededListener() {
			public void actionPerformed(ActionEvent e) {
				fireDataNeeded(e.getActionCommand());
			}
		});
		this.queryID = queryID;
		this.owner = owner;
	}

	public Object getUserContext() {
		return owner;
	}

	public IQueryID getQueryID() {
		return queryID;
	}

	public String getDisplayName() {
		return displayName;
	}

	private void fireDataNeeded(String action) {
		if (eList.getListenerCount(IDataNeededListener.class)>0) {
			EventListener[] s = eList.getListeners(IDataNeededListener.class);
			for (int i = 0; i < s.length; i++) {
				IDataNeededListener listener = (IDataNeededListener)s[i];
				listener.actionPerformed(new ActionEvent(this,ActionEvent.ACTION_PERFORMED,action));
			}
		}
	}

	public void setDisplayName(String displayName) {
		String oldName = this.displayName;
		this.displayName = displayName;
		updateChange();
		firePropertyChanged(QUERY_DISPLAYNAME,oldName,this.displayName);
	}

	public void setQueryID(IQueryID queryID) {
		IQueryID oldID = this.queryID;
		this.queryID = queryID;
		updateChange();
		firePropertyChanged(QUERY_ID,oldID,queryID);
	}

	public void addPropertyChangeListener(PropertyChangeListener changeListener) {
		eList.add(PropertyChangeListener.class, changeListener);
	}
	
	public void addDataNeededListener(IDataNeededListener d) {
		eList.add(IDataNeededListener.class,d);
	}

	private void firePropertyChanged(String propertyName, Object oldVal, Object newVal) {
		if (changing) return;
		queryChanged = true;
		if (eList.getListenerCount(PropertyChangeListener.class) > 0) {
			PropertyChangeListener[] listeners = (PropertyChangeListener[]) eList.getListeners(PropertyChangeListener.class);
			PropertyChangeEvent pce = new PropertyChangeEvent(this, propertyName, oldVal, newVal);
			for (int i=0;i < listeners.length;i++) {
				listeners[i].propertyChange(pce);
			}
		}
	}

	public void removePropertyChangeListener(PropertyChangeListener changeListener) {
		eList.remove(PropertyChangeListener.class, changeListener);
	}
	
	public void removeDataNeededListener(IDataNeededListener d) {
		eList.remove(IDataNeededListener.class,d);
	}

	public boolean isQueryChanged() {
		return queryChanged;
	}

	public void setChanged(boolean queryChanged) {
		this.queryChanged = queryChanged;
	}

	public String getName() {
		if (!hydrateRequested) {
			hydrateRequested = true;
			fireDataNeeded(IQuery.QUERY_ALL);
		}
		return name;
	}

	public void setName(String name) {
		String oldName = this.name;
		this.name = name;
		updateChange();
		firePropertyChanged(QUERY_NAME,oldName,this.name);
	}

	public boolean isHydrateRequested() {
		return hydrateRequested;
	}

	public boolean isUnsavedQueryID() {
		return (queryID.equals(QueryID.unsavedID));
	}

	public boolean isNullQueryID() {
		return (queryID.equals(QueryID.nullID));
	}

	public String toString() {
		return displayName;
	}

	public void setChanging(boolean isChanging) {
		boolean oldChanging = this.changing;
		this.changing = isChanging;
		if (oldChanging && !isChanging) {
			firePropertyChanged(QUERY_ALL,null,null);
		}
	}

	public boolean isChanging() {
		return changing;
	}

	private void updateChange() {
		if (!changing) {
			queryChanged = true;
		}
	}

	public IDisplayFieldList getDisplayFields() {
		return displayFields;
	}

	public void init() {
		if (!hydrateRequested) {
			hydrateRequested = true;
			fireDataNeeded(IQuery.QUERY_ALL);
		}
	}

	public void reset() {
		changing = true;
		displayFields.clearAll();
		constraints.clearAll();
		constraints.clearRelationships();
		hydrateRequested = false;
		queryChanged = false;
		this.name = "";
		changing = false;
	}

	public boolean isValueSetComplete() {
		return constraints.isValueSetComplete();
	}

	public List getDisplayFieldTypes() {
		return constraints.getDisplayFieldTypes();
	}
	
	public void addFinalizeListener(IFinalizeListener listener) {
		eList.add(IFinalizeListener.class, listener);
	}
	
	public void removeFinalizeListener(IFinalizeListener listener) {
		eList.remove(IFinalizeListener.class, listener);
	}

	protected void finalize() throws Throwable {
		if (eList.getListenerCount(IFinalizeListener.class)>0) {
			EventObject e = new EventObject(this);
			IFinalizeListener[] listeners = (IFinalizeListener[])eList.getListeners(IFinalizeListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].preprocessFinalization(e);
			}
		}
		super.finalize();
	}

	public void setHydrateRequested(boolean hydrateRequested) {
		this.hydrateRequested = hydrateRequested;
	}

	public boolean equals(Object obj) {
		return obj instanceof Query && ((Query)obj).getQueryID().equals(this.getQueryID()) && ((Query)obj).getDisplayName().equalsIgnoreCase(this.getDisplayName());
	}

	public IConstraint getConstraints() {
		return constraints;
	}

}
