package gov.va.med.hds.querytool;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.prism.DesktopCloseVetoException;
import gov.va.med.hds.cd.prism.DesktopEvent;
import gov.va.med.hds.cd.prism.IDesktop;
import gov.va.med.hds.cd.prism.IDesktopListener;
import gov.va.med.hds.cd.prism.IPrismConstants;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.querytool.actions.QueryToolActionFactory;
import gov.va.med.hds.querytool.dao.IQueryDAO;
import gov.va.med.hds.querytool.dao.QueryDAOException;
import gov.va.med.hds.querytool.state.StateFactory;
import gov.va.med.hds.querytool.types.ITypeFactory;
import gov.va.med.hds.querytool.types.TypeFactory;
import gov.va.med.hds.querytool.ui.IQueryActionEvent;
import gov.va.med.hds.querytool.ui.IResultsTable;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Map;

import javax.swing.Action;
import javax.swing.event.EventListenerList;

/**
 * Main query tool controller.
 */
public class QueryTool implements IQueryTool {
	private IRuntimeContext runtime;
	private ILogger logger;
	private IPluginDescriptor plugin;
	private EventListenerList eventListeners = new EventListenerList();
	private ITypeFactory typeFactory = new TypeFactory();
	private int maxRows = Integer.MAX_VALUE;
	private Map actionMap;
	private IQueryToolState currentState;
	private IRunQuery runQuery;
	private IQuery currentQuery;
	private ICombinedQueryCollection queryCollection;
	private IQueryDAO queryDAO;
	private Map states;
	private Map eventToStateMap;
	private IResultsTable resultsTable;
	private Map authorizedActionsMap;
	private IAvailableColumnsList availableColumnList = new AvailableColumnsList();
	private PropertyChangeListener queryListener = new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent evt) {
			dispatchStateEvent(currentState);
		}
	};
	private IRunListener runListener = new IRunListener() {
		public void runEventRaised(IRunEvent event) {
			switch (event.getEventID()) {
				case IRunEvent.REPORT_EXCEEDS_MAX_LINES: {
					setMessage("Report was stopped because it exceeded the maximum number of report lines (" + Integer.toString(maxRows) + ").");
					IQueryAction action = (IQueryAction)getQueryAction(IQueryTool.RUN_COMPLETE_MESSAGE_ACTION);
					action.actionPerformed(new QueryActionEvent(this, IQueryActionEvent.RUN_COMPLETE_WITH_MESSAGE));
					break;
				}
				
				case IRunEvent.REPORT_FAILURE: {
					setMessage("There was a report execution error.\nSee the log for details.");
					currentState.setException(runQuery.getException());
					IQueryAction action = (IQueryAction)getQueryAction(IQueryTool.RUN_COMPLETE_ERROR_ACTION);
					action.actionPerformed(new QueryActionEvent(this, IQueryActionEvent.RUN_COMPLETE_WITH_ERROR));
					break;
				}
				
				case IRunEvent.REPORT_SUCCESS: {
					if (runQuery.getRowCount()<1) {
						setMessage("The report had no results.");
						IQueryAction action = (IQueryAction)getQueryAction(IQueryTool.RUN_COMPLETE_MESSAGE_ACTION);
						action.actionPerformed(new QueryActionEvent(this, IQueryActionEvent.RUN_COMPLETE_WITH_MESSAGE));
					}
					else {
						IQueryAction action = (IQueryAction)getQueryAction(IQueryTool.RUN_COMPLETE_ACTION);
						action.actionPerformed(new QueryActionEvent(QueryTool.this, IQueryActionEvent.RUN_COMPLETE));
					}
					break;
				}
				
				case IRunEvent.REPORT_SUCCESS_NON_SENSITIVE: {
					IQueryAction action = (IQueryAction)getQueryAction(IQueryTool.RUN_COMPLETE_NON_SENSITIVE_ACTION);
					action.actionPerformed(new QueryActionEvent(QueryTool.this, IQueryActionEvent.RUN_COMPLETE_NON_SENSITIVE));
					break;
				}
				
			}
		}
	};
	
	/**
	 * Constructor for QueryTool.
	 */
	public QueryTool() {
		super();
	}

	public void startup(IRuntimeContext runtime, IPluginDescriptor plugin) throws Exception {
		this.runtime = runtime;
		this.plugin = plugin;
		logger = runtime.getLogger(this.getClass());
		actionMap = QueryToolActionFactory.buildActionSet(this, logger);
		typeFactory.setRuntime(runtime);
		getQueryDAO();
		setupQueryCollection();
		availableColumnList.setRuntime(runtime);
		states = StateFactory.getStates(runtime);
		eventToStateMap = EventToStateFactory.getEventsToStatesMap();
		dispatchStateEvent(getState(IQueryTool.SELECT_AND_EDIT_STATE));
		authorizedActionsMap = QueryToolActionFactory.getAuthorizedActions(runtime);
		try {
			IDesktop desktop = (IDesktop)runtime.getPluginInstance(IPrismConstants.PLUGIN_ID);
			desktop.addDesktopListener(new IDesktopListener() {
				public void desktopOpening(DesktopEvent e) {}
				public void desktopClosed(DesktopEvent e) {}
				
				public void desktopClosing(DesktopEvent e) throws DesktopCloseVetoException {
					clearExecutedQuery();
				}

			});
		}
		catch (Exception e) {
			logger.info("Could not get an instance of Prism.Desktop.", e);
		}
	}
	
	private void setupQueryCollection() {
		if (queryCollection==null) {
			queryCollection = new CombinedQueryCollection();
			try {
				queryCollection.setRuntime(runtime);
			}
			catch (QueryCollectionException e) {
				logger.error("Error initializing the query collection.", e);
			}
			queryCollection.setQueryDAO(queryDAO);
		}
		try {
			queryCollection.refresh();
		}
		catch (QueryCollectionException e) {
			logger.error("Error refreshing the query collection.", e);
		}
	}

	public void shutdown() throws Exception {
	}

	public IQuery getCurrentQuery() {
		return currentQuery;
	}

	public void setCurrentQuery(IQuery currentQuery) {
		if (this.currentQuery!=currentQuery) {
			if (this.currentQuery!=null) {
				this.currentQuery.removePropertyChangeListener(queryListener);
			}
			this.currentQuery = currentQuery;
			if (currentQuery!=null) {
				currentQuery.addPropertyChangeListener(queryListener);
			}
		}
		if (currentState!=null) {
			currentState.setCurrentQuery(currentQuery);
		}
		
		dispatchEvent(new QueryToolEvent(this, IQueryToolEvent.CURRENT_QUERY_SET, currentQuery));
	}

	public ICombinedQueryCollection getQueryCollection() {
		return currentState.getQueryCollection();
	}

	public void addQueryToolListener(IQueryToolListener listener) {
		eventListeners.add(IQueryToolListener.class, listener);
	}

	public void removeQueryToolListener(IQueryToolListener listener) {
		eventListeners.remove(IQueryToolListener.class, listener);
	}
	
	public void dispatchEvent(IQueryToolEvent e) {
		IQueryToolState posibleNewState = getState(eventToStateMap.get(Integer.toString(e.getEventID())));
		if (posibleNewState!=null) {
			dispatchStateEvent(posibleNewState);
		}
		if (eventListeners.getListenerCount(IQueryToolListener.class)>0) {
			IQueryToolListener[] listeners = (IQueryToolListener[])eventListeners.getListeners(IQueryToolListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].actionOccured(e);
			}
		}
	}
	
	public IRunQuery getRunModel() {
		return currentState.getRunModel();
	}

	public String getMessage() {
		return currentState.getMessage();
	}

	public Throwable getError() {
		return currentState.getException();
	}

	public void registerAction(Object actionKey, IQueryAction action) {
		actionMap.put(actionKey, action);
	}

	public Action getQueryAction(Object actionKey) {
		return (Action)actionMap.get(actionKey);
	}

	public IQuery getNewQuery(String displayName) {
		String text = "Error getting a new query.";
		if (getQueryDAO()!=null) {
			try {
				IQuery newQuery = getQueryDAO().getNew(displayName);
				while (newQuery.isChanging()) {}
				return newQuery;
			}
			catch (QueryDAOException e) {
				setException(e);
			}
		}
		setMessage(text);
		return null;
	}
	
	private void setException(Throwable exception) {
		currentState.setException(exception);
	}

	private IQueryDAO getQueryDAO() {
		if (queryDAO==null) {
			try {
				IDaoRegistry reg = (IDaoRegistry)runtime.getPluginInstance(IDaoRegistry.PLUGIN_ID);
				queryDAO = (IQueryDAO)reg.getDAO(IQueryDAO.class.getName());
			}
			catch (Exception e) {
				logger.error("Error setting up the query data access object.", e);
			}
		}
		return queryDAO;
	}

	public ITypeFactory getTypeFactory() {
		return typeFactory;
	}

	public IResultsTable getResultsModel() {
		return currentState.getResults();
	}

	public void addStateListener(IStateListener listener) {
		eventListeners.add(IStateListener.class, listener);
	}
	
	public void removeStateListener(IStateListener listener) {
		eventListeners.remove(IStateListener.class, listener);
	}

	public void setupRunModel() {
		if (runQuery!=null) {
			runQuery.removeRunListener(runListener);
		}
		runQuery = new RunQuery();
		runQuery.addRunListener(runListener);
		runQuery.setRuntime(runtime);
		runQuery.setQuery(getCurrentQuery());
		runQuery.setMaxRows(maxRows);
		if (currentState!=null) {
			currentState.setRunModel(runQuery);
		}
	}

	public boolean isValidQueryName(String string) {
		if (queryCollection!=null && queryCollection.containsDisplayName(string)) {
			setMessage("Query name is already a defined query; query not created.");
			return false;
		}
		setMessage(null);
		return true;
	}
	
	public void setMessage(String string) {
		if (currentState!=null) {
			currentState.setMessage(string);
		}
	}

	public void dispatchStateEvent(IQueryToolState state) {
		IQueryToolState oldState = currentState;
		String message = null;
		Throwable exception = null;
		if (oldState!=null) {
			message = oldState.getMessage();
			exception = oldState.getException();
		}
		currentState = state;
		currentState.setCurrentQuery(currentQuery);
		currentState.setException(exception);
		currentState.setMessage(message);
		currentState.setQueryCollection(queryCollection);
		currentState.setQueryDAO(queryDAO);
		currentState.setRunModel(runQuery);
		currentState.setResultsTable(resultsTable);
		currentState.setAvailableColumnList(availableColumnList);
		currentState.reset();
		if (eventListeners.getListenerCount(IStateListener.class)>0) {
			IStateEvent event = new StateEvent(this, oldState, currentState);
			IStateListener[] listeners = (IStateListener[])eventListeners.getListeners(IStateListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].stateChanged(event);
			}
		}
	}

	public IQueryToolState getState(Object stateKey) {
		return (IQueryToolState)states.get(stateKey);
	}

	public boolean isActionAuthorized(Object actionKey) {
		if (authorizedActionsMap!=null && authorizedActionsMap.get(actionKey)!=null) {
			return ((Boolean)authorizedActionsMap.get(actionKey)).booleanValue();
		}
		return true;
	}
	
	public boolean setupResultsTable() {
		if (resultsTable!=null) {
			resultsTable.clearModels();
			resultsTable = null;
		}
		try {
			resultsTable = new ResultsTable(currentQuery, runQuery.getCompleteQueryDataDAO(), runQuery.getRowCount(), 300);
			resultsTable.setRuntime(runtime);
		}
		catch (QueryDAOException e) {
			logger.error("Error creating the results table.", e);
		}
		
		if (currentState!=null) {
			currentState.setResultsTable(resultsTable);
		}
		
		return (resultsTable!=null);
	}
	
	public boolean isNameCorrectLength(String name) {
		if (name==null || name.length()<2) {
			setMessage("Query name was too short; please enter a name longer than 1 character.");
			return false;
		}
		
		if (name.length()>80) {
			setMessage("Query name was too long; please enter a name shorter than 81 characters.");
			return false;
		}
		
		return true;
	}
	
	private void clearExecutedQuery() {
		if (runQuery!=null) {
			runQuery.clear();
		}
	}

	public IAvailableColumnsList getAvailableColumnsModel() {
		return currentState.getAvailableColumnList();
	}
	
}
