package gov.va.med.hds.querytool;

import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.clinicaldata.PatientDemographics;
import gov.va.med.hds.querytool.dao.ICompleteQueryDataDAO;
import gov.va.med.hds.querytool.dao.QueryDAOException;
import gov.va.med.hds.querytool.ui.IResultsTable;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import javax.swing.table.AbstractTableModel;

/**
 * Implements a TableModel to provide data from a completed report.
 */
public class ResultsTable extends AbstractTableModel implements IResultsTable {
	private int rowSize;
	private int columnSize;
	private int capacity = 300;
	private Map rowBuffer = new HashMap(capacity);
	private LinkedList keyList = new LinkedList();
	private ICompleteQueryDataDAO runCompleteDAO;
	private final int READ_AHEAD = 20;
	private IQuery query;
	private IRuntimeContext runtime;
	private int sortColumn;
	private int sortDirection;

	private ResultsTable() {
		super();
	}

	/**
	 * Constructor ResultsTable.
	 * @param query
	 * @param runDAO
	 * @param rowSize
	 * @param bufferCapacity
	 */
	public ResultsTable(IQuery query, ICompleteQueryDataDAO runCompleteDAO, int rowSize, int bufferCapacity) throws QueryDAOException {
		this();
		this.columnSize = query.getDisplayFields().size();
		this.query = query;
		this.rowSize = rowSize;
		if ((bufferCapacity>0) && (bufferCapacity!=this.capacity)) {
			this.capacity = bufferCapacity;
			rowBuffer = new HashMap(this.capacity);
		}
		this.runCompleteDAO = runCompleteDAO;
		init();
	}

	private void init() throws QueryDAOException {
		setSortColumn(query.getDisplayFields().get(0),true);
	}

	public int getRowCount() {
		return rowSize;
	}

	public int getColumnCount() {
		return columnSize;
	}

	public Object getValueAt(int rowIndex, int columnIndex) {
		if (runCompleteDAO==null) return "";
		Object result = new String("");
		int startIndex = rowIndex;
		int endIndex = rowIndex;
		String key = Integer.toString(rowIndex);
		if (!rowBuffer.containsKey(key)) {
			if (nextRowHasData(startIndex)) {
				endIndex = startIndex - READ_AHEAD;
			}
			else if (previousRowHasData(startIndex)) {
				endIndex = startIndex + READ_AHEAD;
			}
			
			if (endIndex>rowSize-1) {
				endIndex = rowSize-1;
			}
			else if (endIndex<0) {
				endIndex = 0;
			}
			
			if (endIndex<startIndex) {
				int tempIndex = startIndex;
				startIndex = endIndex;
				endIndex = tempIndex;
			}
			
			IReportRow[] rows = new IReportRow[0];
			try {
				rows = runCompleteDAO.getReportRow(startIndex,endIndex);
			}
			catch (QueryDAOException e) {
				runtime.getLogger(this.getClass()).error("Error retrieving result table data.", e);
			}
			String newKey;
			for (int i = 0; i < rows.length; i++) {
				newKey = Integer.toString(rowIndex+i);
				rowBuffer.put(newKey,rows[i]);
				keyList.add(newKey);
			}
			
			removeExcessOverCapacity();
		}
		IReportRow row = (IReportRow)rowBuffer.get(key);
		if (row!=null) {
			result = row.getItem(columnIndex);
		}
		return result;
	}

	private void removeExcessOverCapacity() {
		while (rowBuffer.size()>capacity) {
			rowBuffer.remove(keyList.get(0));
			keyList.remove(0);
		}
	}
	
	private boolean nextRowHasData(int index) {
		return (rowBuffer.containsKey(Integer.toString(index+1)));
	}

	private boolean previousRowHasData(int index) {
		return (rowBuffer.containsKey(Integer.toString(index-1)));
	}

	private IReportRow getRowKey(int index) {
		String key = Integer.toString(index);
		if (!rowBuffer.containsKey(key)) {
			getValueAt(index,0);
		}
		return (IReportRow) rowBuffer.get(key);
	}

	public void reset() {
		rowBuffer.clear();
		keyList.clear();
		super.fireTableDataChanged();
	}

	public void setSortColumn(IDisplayField field, boolean ascending) throws QueryDAOException {
		reset();
		if (runCompleteDAO!=null) {
			runCompleteDAO.setSortField(field,ascending);
		}
	}

	public void setDAOModel(ICompleteQueryDataDAO completeQueryDataDAO) {
		this.runCompleteDAO = completeQueryDataDAO;
	}

	public void getDetailRequest(int row, IDetailReceivable detailReceivable) {
		IReportRow key = getRowKey(row);
		if (runCompleteDAO==null) return;
		try {
			runCompleteDAO.detailResultsRequest(key, detailReceivable);
		}
		catch (QueryDAOException e) {
			detailReceivable.setDetail("Error retrieving data; see log for details.");
			runtime.getLogger(this.getClass()).error("Error getting detail data: key="+key.toString(), e);
		}
	}

	public void getPatientDetailRequest(int row, IPatDetailReceivable patDetailReceivable) {
		IReportRow key = getRowKey(row);
		if (runCompleteDAO==null) return;
		try {
			runCompleteDAO.patientInfoRequest(key, patDetailReceivable);
		}
		catch (QueryDAOException e) {
			patDetailReceivable.setPatient(new PatientDemographics("", "*Error* (See log)", "", ""));
			runtime.getLogger(this.getClass()).error("Error getting patient detail data: key="+key.toString(), e);
		}
	}

	public void setRuntime(IRuntimeContext runtime) {
		this.runtime = runtime;
	}

	public void setSortColumn(int columnPosition, int direction) throws QueryDAOException {
		sortColumn = columnPosition;
		sortDirection = direction;
		setSortColumn(query.getDisplayFields().get(columnPosition), direction==IResultsTable.ASCENDING);
	}

	public int getSortColumnPosition() {
		return sortColumn;
	}

	public int getSortDirection() {
		return sortDirection;
	}

	public void clearModels() {
		query = null;
		runCompleteDAO = null;
	}

}
