/*
 * Initially Created on Apr 16, 2003 by vhaislmonsos
 *
 */
package gov.va.med.hds.querytool;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.querytool.dao.ICompleteQueryDataDAO;
import gov.va.med.hds.querytool.dao.IMainRunDAO;
import gov.va.med.hds.querytool.dao.IRunDAO;
import gov.va.med.hds.querytool.dao.IRunSetupDAO;
import gov.va.med.hds.querytool.dao.ISensitiveDataDAO;
import gov.va.med.hds.querytool.dao.ISensitiveRunDAO;
import gov.va.med.hds.querytool.dao.IShutDownQueryDAO;
import gov.va.med.hds.querytool.dao.QueryDAOException;
import gov.va.med.hds.util.dao.IServerDateDAO;

import java.util.Date;

import javax.swing.event.EventListenerList;

/**
 * Implements the IRunQuery interface to manage the setup and run thread for main
 * patient iterations and sensitive patient iterations.
 */
public class RunQuery implements IRunQuery {
	private int maxRows = Integer.MAX_VALUE;
	private IQuery query;
	private int rowCount;
	private int patientCount;
	private int screenedCount;
	private ExecutionException exception;
	private EventListenerList listenerList = new EventListenerList();
	private Thread runThread;
	private int runEvent;
	private IRuntimeContext runtime;
	private IRunSetupDAO runSetupDAO;
	private IMainRunDAO mainRunDAO;
	private IShutDownQueryDAO shutDownQueryDAO;
	private ISensitiveRunDAO sensitiveRunDAO;
	private ISensitiveDataDAO sensitiveDataDAO;
	private ISensitiveDataModel sensitiveDataModel;
	private ICompleteQueryDataDAO completeQueryDataDAO;
	private IServerDateDAO serverDateDAO;
	private IDaoRegistry daoRegistry;
	private Date startDate = new Date();

	public RunQuery() {
		super();
	}

	public void setQuery(IQuery query) {
		this.query = query;
	}

	public int getRowCount() {
		return rowCount;
	}

	public int getScreenedCount() {
		return screenedCount;
	}

	public int getPatientCount() {
		return patientCount;
	}

	public boolean isRunning() {
		return runThread!=null && runThread.isAlive();
	}

	public void runQuery() {
		patientCount = 0;
		rowCount = 0;
		screenedCount = 0;
		exception = null;
		runEvent = IRunEvent.NO_EVENT;
		runThread = new Thread(new Runnable() {
			public void run() {
				try {
					if (!setupDAOs()) {
						return;
					}
					startDate = serverDateDAO.getDate();
					runSetupDAO.setQuery(query);
					mainRunDAO.setInitialIterator(runSetupDAO.getIterator());
					runIterations(mainRunDAO);
					if (runEvent!=IRunEvent.REPORT_SUCCESS) {
						fireAction(runEvent);
						return;
					}
					if (sensitiveDataDAO.getSize()>0) {
						sensitiveDataModel = new SensitiveDataModel(runtime, sensitiveDataDAO);
						runEvent = IRunEvent.NO_EVENT;
						fireAction(IRunEvent.REPORT_SUCCESS_NON_SENSITIVE);
						synchronized (runThread) {
							runThread.wait();
						}
						sensitiveDataModel = null;
					}
					if (sensitiveDataDAO.getSize()>0 && runEvent==IRunEvent.NO_EVENT) {
						sensitiveRunDAO.setInitialIterator("");
						runIterations(sensitiveRunDAO);
					}
					if (runEvent==IRunEvent.NO_EVENT) {
						runEvent = IRunEvent.REPORT_SUCCESS;
					}
					fireAction(runEvent);
				}
				catch (Exception e) {
					exception = new ExecutionException(e);
					fireAction(IRunEvent.REPORT_FAILURE);
				}
			}

		},"Run_Query");
		runThread.start();
	}

	private boolean setupDAOs() {
		if (daoRegistry==null) {
			try {
				daoRegistry = (IDaoRegistry)runtime.getPluginInstance(IDaoRegistry.PLUGIN_ID);
				
				runSetupDAO = (IRunSetupDAO)daoRegistry.getDAO(IRunSetupDAO.class.getName());
				mainRunDAO = (IMainRunDAO)daoRegistry.getDAO(IMainRunDAO.class.getName());
				sensitiveDataDAO = (ISensitiveDataDAO)daoRegistry.getDAO(ISensitiveDataDAO.class.getName());
				sensitiveRunDAO = (ISensitiveRunDAO)daoRegistry.getDAO(ISensitiveRunDAO.class.getName());
				completeQueryDataDAO = (ICompleteQueryDataDAO)daoRegistry.getDAO(ICompleteQueryDataDAO.class.getName());
				completeQueryDataDAO.setQuery(query);
				shutDownQueryDAO = (IShutDownQueryDAO)daoRegistry.getDAO(IShutDownQueryDAO.class.getName());
				serverDateDAO = (IServerDateDAO)daoRegistry.getDAO(IServerDateDAO.class.getName());
			}
			catch (Exception e) {
				exception = new ExecutionException("Error setting up the data access objects to run query.",e);
				fireAction(IRunEvent.REPORT_FAILURE);
				return false;
			}
		}
		return true;
	}

	public void stopQuery() {
		runEvent = IRunEvent.REPORT_SUCCESS;
		startThread();
	}
	
	private void startThread() {
		synchronized (runThread) {
			runThread.notify();
		}
	}

	public ExecutionException getException() {
		return exception;
	}
	
	protected void fireAction(int eventID) {
		if (listenerList.getListenerCount()>0) {
			IRunEvent event = new RunEvent(this,eventID);
			IRunListener[] listeners = (IRunListener[])listenerList.getListeners(IRunListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].runEventRaised(event);
			}
		}
	}
	
	private void runIterations(IRunDAO runDAO) throws QueryDAOException {
		if (runDAO.isReportFinished()) {
			runEvent = IRunEvent.REPORT_SUCCESS;
		}
		while (runEvent==IRunEvent.NO_EVENT) {
			runDAO.getNextIteration();
			if (runDAO.isPatientSelected()) {
				patientCount++;
			}
			else {
				screenedCount++;
			}
			rowCount += runDAO.getSize();
			if (rowCount>maxRows) {
				runEvent = IRunEvent.REPORT_EXCEEDS_MAX_LINES;
			}
			if (runDAO.isReportFinished()) {
				runEvent = IRunEvent.REPORT_SUCCESS;
			}
		}
	}

	public void setMaxRows(int maxRows) {
		this.maxRows = maxRows;
	}

	public void addRunListener(IRunListener listener) {
		listenerList.add(IRunListener.class, listener);
	}

	public void removeRunListener(IRunListener listener) {
		listenerList.remove(IRunListener.class, listener);
	}

	public void resumeFromPause() {
		startThread();
	}

	public void setRuntime(IRuntimeContext runtime) {
		this.runtime = runtime;
	}

	public ICompleteQueryDataDAO getCompleteQueryDataDAO() {
		return completeQueryDataDAO;
	}

	public ISensitiveDataModel getSensitiveDataModel() {
		return sensitiveDataModel;
	}

	public Date getStartDate() {
		return startDate;
	}

	public void clear() {
		runEvent = IRunEvent.REPORT_CANCELLED;
		startThread();
		while (runThread!=null && runThread.isAlive()) {}
		if (shutDownQueryDAO!=null) {
			try {
				shutDownQueryDAO.getResult();
			}
			catch (QueryDAOException e) {
				runtime.getLogger(this.getClass()).error("Error shutting down/clearing query on the server.", e);
			}
		}
	}
}
