package gov.va.med.hds.querytool.m;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.MumpsException;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.querytool.CriteriaItem;
import gov.va.med.hds.querytool.DateRange;
import gov.va.med.hds.querytool.ICriteriaItem;
import gov.va.med.hds.querytool.IQuery;
import gov.va.med.hds.querytool.IValue;
import gov.va.med.hds.querytool.dao.ConstraintDAOException;
import gov.va.med.hds.querytool.dao.IConstraintDAO;
import gov.va.med.hds.querytool.types.ITypeFactory;
import gov.va.med.hds.querytool.types.TypeFactory;
import gov.va.med.hds.util.ExceptionStackTrace;
import gov.va.med.hds.util.dao.IServerDateDAO;
import gov.va.med.hds.util.date.RelativeDate;
import gov.va.med.hds.util.mumps.NameValuePair;
import gov.va.med.hds.util.mumps.StringUtil;

import java.util.Arrays;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * M implementation of the IConstraintDAO.
 */
public class MConstraintDAO extends AbstractMDAO implements IConstraintDAO {
	private static final String REGEX_LINE_SPLT = "\\r\\n|\\n";
	private static final String CTPLST_RPC_KEY = "CTPLST";
	private static final String DFLDMAP_RPC_KEY = "DFLDMAP";
	private String[] constraintTree;
	private ICriteriaItem priorItem;
	private ITypeFactory typeFactory;
	private GregorianCalendar cal = new GregorianCalendar();
	private Date referenceDate;
	private static final String TAG_DEFN_RPC_KEY = "TAGDEF";
	private Map displayFieldCategoryMap = new HashMap();

	public MConstraintDAO() {
		super();
	}

	/**
	 * @see gov.va.med.hds.querytool.IConstraintDAO#get()
	 */
	public void get(IQuery query, String[] constraintData) throws ConstraintDAOException {
		query.getConstraints().clearAll();
		for (int i = 0; i < constraintTree.length; i++) {
			query.getConstraints().addRelationship(constraintTree[i]);
		}
		processLines(query, constraintData, null);
	}

	private void processLines(IQuery query, String[] constraintData, String defaultTag)
		throws ConstraintDAOException {
		priorItem = null;
		if (constraintData != null) {
			CriteriaItem cItem = new CriteriaItem();
			cItem.setTag(defaultTag);
			boolean hasValue = false;
			for (int i = 0; i < constraintData.length; i++) {
				NameValuePair line = new NameValuePair(constraintData[i]);
				String left = line.getLeft();
				String right = line.getRight();
				if (left.compareTo("QueryText") == 0) {
					if (hasValue) {
						saveItem(cItem, query);
						cItem = new CriteriaItem();
						cItem.setTag(defaultTag);
					} else {
						hasValue = true;
					}
					cItem.setOrder(Integer.parseInt(right));
				} else if (left.compareTo("CriterionTag") == 0) {
					cItem.setTag(right);
				} else if (left.compareTo("Caption") == 0) {
					cItem.setCaption(right);
				} else if (left.compareTo("ConstraintName") == 0) {
					cItem.setConstraintName(right);
				} else if (left.compareTo("EditorID") == 0) {
					cItem.setEditorID(right);
				} else if (left.compareTo("Value") == 0) {
					cItem.getValues().add(right);
				}
			}
			if (hasValue) {
				saveItem(cItem, query);
			}
		}
	}

	private void saveItem(ICriteriaItem cItem, IQuery query) {
		//if cItem is a header value
		if (!isValue(cItem)) {
			if (priorItem != null) {
				saveValue(priorItem, null, priorItem.getTag(), query);
			}
			priorItem = cItem;
			cItem = null;
		}
		//	if the previous is a header, save previous as a value; make cItem the previous
		//}

		if (cItem != null) {
			if ((priorItem != null) && (cItem.getTag().compareTo(priorItem.getTag()) == 0)) {
				saveValue(priorItem, cItem, cItem.getTag(), query);
				priorItem = null;
				cItem = null;
			}
			if (priorItem != null) {
				saveValue(priorItem, null, priorItem.getTag(), query);
				priorItem = null;
			}
			if (cItem != null) {
				saveValue(null, cItem, cItem.getTag(), query);
				cItem = null;
			}
		}

	}
	
	private String reformatDateString(String dateString) {
		if (dateString.length()<1) {
			return dateString;
		}
		if (dateString.indexOf('T') < 0) {
			if (dateString.equals("0")) {
				return "T";
			}
			else {
				cal.setTime(StringUtil.fmDateToDate(dateString));
				return RelativeDate.getRelativeDateString(referenceDate,cal.getTime());
			}
		}
		return dateString;
	}

	private void saveValue(ICriteriaItem item1, ICriteriaItem item2, String tag, IQuery query) {

		IValue value = null;

		//Translate M Dates into a more generic form.
		if (item2 != null) {
			if (item2.getEditorID().equals("2")) {
				for (int i = 0; i < item2.getValues().size(); i++) {
					String text = (String) item2.getValues().get(i);
					if (StringUtil.length(text, ':') > 1) {
						String piece = StringUtil.piece(text, ':', 1);
						String newPiece = reformatDateString(piece);
						if (!piece.equalsIgnoreCase(newPiece)) {
							text = StringUtil.setPiece(text, ':', 1, newPiece);
						}
						piece = StringUtil.piece(text, ':', 2);
						newPiece = reformatDateString(piece);
						if (!piece.equalsIgnoreCase(newPiece)) {
							text = StringUtil.setPiece(text, ':', 2, newPiece);
						}
					}
					item2.getValues().remove(i);
					item2.getValues().add(i, new DateRange(referenceDate, text));
				}
			}
		}

		if ((item1 != null) && (item2 != null)) {
			value = typeFactory.getValue(item2.getEditorID());
			value.setDisplayText(item1.getCaption());
			value.setDefaultDisplayText(item2.getCaption());
			value.setValues(item2.getValues());
			value.setConstraintName(item2.getConstraintName());
			item1 = null;
			item2 = null;
		}

		if (item1 != null) {
			value = typeFactory.getValue("12");
			value.setDefaultDisplayText(item1.getCaption());
		}

		if (item2 != null) {
			value = typeFactory.getValue(item2.getEditorID());
			value.setDefaultDisplayText(item2.getCaption());
			value.setValues(item2.getValues());
			value.setConstraintName(item2.getConstraintName());
		}

		value.setID(tag);
		//		query.createConstraintValue(tag,value);
		if (value.getConstraintName()!=null && value.getConstraintName().length()>0) {
			value.setDisplayFieldCategory(displayFieldCategoryMap.get(value.getConstraintName().substring(0, 1)));
		}
		query.getConstraints().addValue(value);

	}

	private boolean isValue(ICriteriaItem item) {
		return (!item.getEditorID().equals(ICriteriaItem.NULL_EDITOR));
	}

	private void getConstraintTree() throws ConstraintDAOException {
		String[] results;
		String brokerResult;

		try {
			brokerResult = mumps.callRPC(CTPLST_RPC_KEY);
		} catch (Exception e) {
			throw new ConstraintDAOException(e);
		}

		results = brokerResult.split(REGEX_LINE_SPLT);
		Arrays.sort(results);
		constraintTree = results;
	}
	
	private void getDisplayFieldCategoryMap() throws ConstraintDAOException {
		String[] results;
		String mResults;
		
		try {
			mResults = mumps.callRPC(DFLDMAP_RPC_KEY, 10000);
		}
		catch (MumpsException e) {
			throw new ConstraintDAOException("The call '"+DFLDMAP_RPC_KEY+"' failed.",e);
		}
		
		results = mResults.split(REGEX_LINE_SPLT);
		for (int i = 0; i < results.length; i++) {
			displayFieldCategoryMap.put(StringUtil.piece(results[i],'=',1), StringUtil.piece(results[i],'=',2));
		}
	}

	public void setInitializationData(
		IRuntimeContext runtime,
		IConfigurationElement config,
		String propertyName, Object data)
		throws Exception {
		super.setInitializationData(runtime, config, propertyName, data);
		
		getConstraintTree();
		getDisplayFieldCategoryMap();
		IDaoRegistry daoRegistry = (IDaoRegistry)runtime.getPluginInstance(IDaoRegistry.PLUGIN_ID);
		IServerDateDAO server = (IServerDateDAO)daoRegistry.getDAO(IServerDateDAO.class.getName());
		referenceDate = server.getDate();
		typeFactory = new TypeFactory(runtime);
	}

	public synchronized void getTag(final IQuery query, final String tag)
		throws ConstraintDAOException {
		Thread getTagThread = new Thread() {
			public void run() {
				processTag(query, tag);
			}
		};
		getTagThread.setName("Querytool_Tag_Request");
		getTagThread.start();
	}

	private void processTag(IQuery query, String tag) {
		List parms = new LinkedList();
		parms.add(tag);
		String result = null;
		try {
			result = mumps.callRPC(TAG_DEFN_RPC_KEY, parms);
		} catch (Exception e) {
			logErrorMessage("Error calling " + TAG_DEFN_RPC_KEY+"\n "+ExceptionStackTrace.getStackTrace(e));
		}
		if (result != null && result.length() > 0) {
			String[] results = result.split(REGEX_LINE_SPLT);
			try {
				processLines(query, results, tag);
			} catch (ConstraintDAOException e) {
				logErrorMessage(
					"Error processing Constraint data from the database:" + e.getMessage());
			}
		}
	}

	private void logErrorMessage(String message) {
		ILogger logger = runtime.getLogger(IConstraintDAO.class);
		logger.error(message);
	}

}
