package gov.va.med.hds.querytool.m;

import gov.va.med.hds.cd.daoregistry.IDaoRegistry;
import gov.va.med.hds.cd.m.IMumpsUserInfo;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.security.ISecurity;
import gov.va.med.hds.querytool.ICriteriaItem;
import gov.va.med.hds.querytool.IDataNeededListener;
import gov.va.med.hds.querytool.IDisplayField;
import gov.va.med.hds.querytool.IQuery;
import gov.va.med.hds.querytool.IQueryID;
import gov.va.med.hds.querytool.IValue;
import gov.va.med.hds.querytool.Query;
import gov.va.med.hds.querytool.QueryID;
import gov.va.med.hds.querytool.dao.ConstraintDAOException;
import gov.va.med.hds.querytool.dao.IConstraintDAO;
import gov.va.med.hds.querytool.dao.IDisplayFieldListDAO;
import gov.va.med.hds.querytool.dao.IQueryDAO;
import gov.va.med.hds.querytool.dao.QueryDAOException;
import gov.va.med.hds.util.mumps.StringUtil;

import java.awt.event.ActionEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * M implementation of a Query DAO.
 */
public class MQueryDAO extends AbstractMDAO implements IQueryDAO {
	private static final String DELETE_RPC_KEY = "ORRHCR DELETE";
	private static final String SAVDEF_RPC_KEY = "SAVDEF";
	private static final String REPORT_DEFN_RPC_KEY = "RPTDEF";
	private IDaoRegistry daoRegistry;
	private IDisplayFieldListDAO displayFieldListDAO;
	private IConstraintDAO constraintDAO;
	private IDataNeededListener dataNeededListener =
		new IDataNeededListener() {
		public void actionPerformed(final ActionEvent e) {
			Thread request = new Thread() {
				public void run() {
					fireDataRequest(e);
				}
			};
			request.setName("Querytool data request");
			request.start();
		}
	};
	private Map queryLines = new HashMap();
	private Map parmLines = new HashMap();
	private Map dispLines = new HashMap();
	private IMumpsUserInfo userInfo;
	private final int MAX_STRING_LENGTH = 75;

	public MQueryDAO() {
		super();
	}

	public IQuery getAsync(IQueryID queryID, String displayText) throws QueryDAOException {
		IQuery q = null;
		q = new Query(queryID, getMumpsUserInfo(runtime).getDuz());
		q.setDisplayName(displayText);
		q.setChanged(false);
		setQuery(q);
		return q;
	}

	private IMumpsUserInfo getMumpsUserInfo(IRuntimeContext runtime)
		throws QueryDAOException {
		if (userInfo == null) {
			try {
				ISecurity security = null;
				security =
					(ISecurity)runtime.getPluginInstance(ISecurity.PLUGIN_ID);
				Set set =
					(Set)security.getSubject().getPrincipals(
						IMumpsUserInfo.class);
				userInfo = (IMumpsUserInfo)set.iterator().next();
			}
			catch (Exception e) {
				sendErrorMessage("Unable to get the Mumps user context.", e);
			}
		}
		return userInfo;
	}

	public IQuery getSync(IQueryID queryID) throws QueryDAOException {
		IQuery query = new Query(queryID, getMumpsUserInfo(runtime).getDuz());
		query.setChanging(true);
		ActionEvent event = new ActionEvent(query, 0, IQuery.QUERY);
		fireQueryDataRequest(event);
		query.setChanging(false);
		query.setChanged(false);
		setQuery(query);
		return query;
	}

	public void save(IQuery query) throws QueryDAOException {
		List parms = new LinkedList();
		parms.add(getDefnStrings(query));
		String result = null;
		try {
			result = mumps.callRPC(SAVDEF_RPC_KEY, parms);
		}
		catch (Exception e) {
			throw new QueryDAOException("Error saving query.");
		}
		String id = StringUtil.piece(result,'^',1);
		String message = StringUtil.piece(result,'^',2);
		if (id.equals("0")) {
			sendErrorMessage("The query save RPC returned an exception.",new Exception(message));
		}
		else {
			if (query.isNullQueryID() || query.isUnsavedQueryID()) {
				query.setQueryID(new QueryID(id));
			}
		}
	}

	public boolean delete(IQueryID queryID) {
		List parms = new ArrayList();
		parms.add(queryID.toString());
		try {
			mumps.callRPC(DELETE_RPC_KEY, parms, 3000);
		}
		catch (Exception e) {
			sendErrorMessage("Error deleting a query.",e);
			return false;
		}
		return true;
	}

	public void setInitializationData(
			IRuntimeContext runtime,
			IConfigurationElement config,
			String propertyName, Object data) throws Exception {

		super.setInitializationData(runtime, config, propertyName, data);

		daoRegistry = (IDaoRegistry)runtime.getPluginInstance(IDaoRegistry.PLUGIN_ID);
		constraintDAO = (IConstraintDAO)daoRegistry.getDAO(IConstraintDAO.class.getName());
		displayFieldListDAO = (IDisplayFieldListDAO)daoRegistry.getDAO(IDisplayFieldListDAO.class.getName());
	}

	public void clearQuery(IQuery query) {
		query.removeDataNeededListener(dataNeededListener);
	}

	public void setQuery(IQuery query) {
		query.addDataNeededListener(dataNeededListener);
	}

	private synchronized void fireQueryDataRequest(final ActionEvent event) {
		IQuery q = (IQuery)event.getSource();
		final String key;
		if (!QueryID.isValid(q.getQueryID())) {
			key = "22";
		}
		else {
			key = q.getQueryID().toString();
		}
		
		if ((!queryLines.containsKey(key))
			|| (!parmLines.containsKey(key))
			|| (!dispLines.containsKey(key))) {
			List parms = new LinkedList();
			parms.add(key);
			String result = null;
			try {
				result = mumps.callRPC(REPORT_DEFN_RPC_KEY, parms);
			}
			catch (Exception e) {
				sendErrorMessage("Error retrieving report def'n: " + REPORT_DEFN_RPC_KEY, e);
			}
			try {
				processLines(key, result);
			}
			catch (QueryDAOException e1) {
				sendErrorMessage("Error processing query lines from database.",e1);
			}
		}
		
		if ((queryLines.containsKey(key))
			&& (parmLines.containsKey(key))
			&& (dispLines.containsKey(key))) {
			q.setChanging(true);
			String[] lines = (String[])queryLines.remove(key);
			q.setDisplayName(lines[1].substring(12));
			q.setName(lines[0].substring(5));

			lines = (String[])parmLines.remove(key);
			try {
				constraintDAO.get(q, lines);
			}
			catch (ConstraintDAOException e) {
				sendErrorMessage("Error processing query constraints from the database.", e);
			}

			lines = (String[])dispLines.remove(key);
			try {
				displayFieldListDAO.get(q, lines);
			}
			catch (QueryDAOException e) {
				sendErrorMessage("Error processing display fields from the database.", e);
			}
			q.setChanging(false);
		}
		else {
			sendErrorMessage("Query definition data was incomplete.", new QueryDAOException("Invalid data trying to update query " + q.getDisplayName()));
		}
	}

	private void processLines(String key, String linesToProcess) throws QueryDAOException {
		List parmList = new LinkedList();
		List qList = new LinkedList();
		List dispList = new LinkedList();
		String line = null;
		BufferedReader reader = new BufferedReader(new StringReader(linesToProcess));
		try {
			while ((line=reader.readLine())!=null) {
				if (line.startsWith("Name=")) {
					qList.add(line);
				}
				else
					if (line.startsWith("Displ")) {
						qList.add(line);
					}
					else
						if (line.startsWith("Colum")) {
							dispList.add(line);
						}
						else {
							parmList.add(line);
						}
			}
		}
		catch (IOException e) {
			throw new QueryDAOException();
		}
		queryLines.put(key, (String[])qList.toArray(new String[0]));
		parmLines.put(key, parmList.toArray(new String[0]));
		dispLines.put(key, dispList.toArray(new String[0]));
	}

	private void fireParmDataRequest(final ActionEvent e)
		throws ConstraintDAOException {
		constraintDAO.getTag(
			(IQuery)e.getSource(),
			StringUtil.piece(e.getActionCommand(), '.', 2));
	}

	private void fireDispDataRequest(final ActionEvent e) {
		throw new RuntimeException("Display data request should not run.");
	}

	private void fireDataRequest(ActionEvent e) {
		String object = StringUtil.piece(e.getActionCommand(), '.', 1);
		if (object.equals(IQuery.QUERY)) {
			fireQueryDataRequest(e);
		}
		else
			if (object.equals(IQuery.PARM)) {
				try {
					fireParmDataRequest(e);
				}
				catch (ConstraintDAOException ex) {
					sendErrorMessage("Query constraint data request failure.", ex);
				}
			}
			else
				if (object.equals(IQuery.DISP)) {
					fireDispDataRequest(e);
				}
				else {
					sendErrorMessage("Data request not registered.", new QueryDAOException("Invalid call for display field data from database."));
				}
	}

	public List getDefnStrings(IQuery query) {
		List lines = new ArrayList();
		Count count = new Count();
		count.num = 1;
		synchronized (query) {
			lines.add("IEN=" + query.getQueryID().toString());
			lines.add("Name=" + query.getName());
			lines.add("DisplayText=" + query.getDisplayName());
			processContraints(query, lines, count);
			int fieldSize = query.getDisplayFields().size();
			for (int i = 0; i < fieldSize; i++) {
				IDisplayField field = query.getDisplayFields().get(i);
				lines.add("Column=" + Integer.toString(i));
				lines.add("ColumnName=" + field.getID());
				lines.add("ColumnHeader=" + field.getDisplayText());
				lines.add("ColumnWidth=" + Integer.toString(field.getColumnWidth()));
			}
		}
		return lines;
	}

	private class Count {
		public int num;
		public String toString() {
			return Integer.toString(num);
		}
	}

	private void processContraints(IQuery query, List lines, Count count) {
		for (Iterator i = query.getConstraints().iterator(); i.hasNext();) {
			IValue item = (IValue)i.next();
			lines.add("QueryText=" + count.toString());
			count.num++;
			lines.add("CriterionTag=" + item.getID());
			if ((item.getDisplayText() == null)
				|| (item.getDisplayText().length() < 1)) {
				lines.add("Caption=" + limitStringLength(item.getDefaultDisplayText(), MAX_STRING_LENGTH));
			}
			else {
				lines.add("Caption=" + limitStringLength(item.getDisplayText(),MAX_STRING_LENGTH));
				lines.add("ConstraintName=");
				lines.add("EditorID=" + ICriteriaItem.NULL_EDITOR);
				
				lines.add("QueryText=" + count.toString());
				count.num++;
				lines.add("CriterionTag=" + item.getID());
				lines.add("Caption=" + limitStringLength(item.getDefaultDisplayText(),MAX_STRING_LENGTH));
			}
			lines.add("ConstraintName=" + item.getConstraintName());
			lines.add("EditorID=" + item.getEditorID().toString());
			if ((item.getValues() != null)
				&& (item.getValues().size() > 0)) {
				for (Iterator k = item.getValues().iterator();
					k.hasNext();
					) {
					lines.add("Value=" + k.next().toString());
				}
			}
		}
	}

	private String limitStringLength(String inputString, int length) {
		if (inputString.length()>length) {
			return inputString.substring(0, length) + "...";
		}
		else {
			return inputString;
		}
	}

	public List getReportStrings(IQuery query) {
		List list = new ArrayList();
		synchronized (query) {
			getConstraintStrings(query, list);
			int size = query.getDisplayFields().size();
			for (int i = 0; i < size; i++) {
				list.add("Report.Column=" + query.getDisplayFields().get(i).getID());
			}
		}
		return list;
	}

	private void getConstraintStrings(IQuery query, List list) {
		for (Iterator i = query.getConstraints().iterator(); i.hasNext();) {
			IValue item = (IValue)i.next();
			if ((item.getValues() != null)
				&& (item.getValues().size() > 0)) {
				for (Iterator v = item.getValues().iterator();
					v.hasNext();
					) {
					list.add(
						item.getConstraintName()
							+ "="
							+ v.next().toString());
				}
			}
		}
	}
	
	private void sendErrorMessage(String message, Throwable t) {
		runtime.getLogger(MQueryDAO.class).fatal(message, t);
	}

	public IQuery getNew(String displayText) throws QueryDAOException {
		IQuery newQuery = null;
		try {
			newQuery = new Query(QueryID.nullID, getMumpsUserInfo(runtime).getDuz());
		}
		catch (QueryDAOException e1) {
			throw new QueryDAOException("New Query: Runtime error retrieving the user information.");
		}
		ActionEvent event = new ActionEvent(newQuery, -1, IQuery.QUERY);
		newQuery.setHydrateRequested(true);
		fireQueryDataRequest(event);
		newQuery.setQueryID(QueryID.nullID);
		if (displayText!=null) {
			newQuery.setDisplayName(displayText);
		}
		setQuery(newQuery);
		return newQuery;
	}

}
