package gov.va.med.hds.querytool.types;

import gov.va.med.hds.querytool.IValue;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.EventListener;
import java.util.List;

import javax.swing.event.EventListenerList;

/**
 * Abstract type value.
 */
public abstract class AbstractType implements IValue {
	private List values = new ArrayList();
	private String id;
	private String defaultDisplayText;
	private String displayText = new String();
	private boolean isReadOnly = false;
	private EventListenerList propertyChangeListeners = new EventListenerList();
	private String constraintName = "";
	private Object lock;
	private Object editorID;
	private Object displayFieldCategory;

	public AbstractType() {
		super();
		lock = this;
		defaultDisplayText = getEmptyDisplayText();
		id = "Blank ID";
	}
	
	public AbstractType(String id) {
		this();
		this.id = id;
	}

	public List getValues() {
		synchronized (lock) {
			return values;
		}
	}

	public void setValues(List list) {
		synchronized (lock) {
			values.clear();
			values.addAll(list);
			firePropertyChanged("VALUES",null,null);
		}
	}

	public String getID() {
		synchronized (lock) {
			return id;
		}
	}

	public boolean isValue() {
		return true;
	}

	public void setID(String id) {
		String oldText = this.id;
		synchronized (lock) {
			this.id = id;
			firePropertyChanged("ID",oldText,this.id);
		}
	}

	public String getDefaultDisplayText() {
		synchronized (lock) {
			return defaultDisplayText;
		}
	}

	public void setDefaultDisplayText(String defaultDisplayText) {
		String oldText = this.defaultDisplayText;
		synchronized (lock) {
			this.defaultDisplayText = defaultDisplayText;
			firePropertyChanged("DEFAULT_DISPLAY",oldText,this.defaultDisplayText);
		}
	}

	public String getDisplayText() {
		synchronized (lock) {
			return displayText;
		}
	}

	public void setDisplayText(String displayText) {
		String oldText = this.displayText;
		synchronized (lock) {
			this.displayText = displayText;
			firePropertyChanged("DISPLAY",oldText,this.displayText);
		}
	}

	public String getFullDisplayText() {
		synchronized (lock) {
			String result = new String();
			if (displayText != null) {
				result = result + displayText;
			}
	
			String valText = getValueText();
	
			if ((valText.length() > 0) && (result.length() > 0)) {
				result = result + " ";
			}
	
			if (valText.length() > 0) {
				result = result + valText;
			}
	
			return result;
		}
	}

	private String getValueText() {
		synchronized (lock) {
			String result = new String();

			if (defaultDisplayText != null) {
				result += defaultDisplayText;
			}
	
			return result;
		}
	}

	public boolean isReadOnly() {
		synchronized (lock) {
			return isReadOnly;
		}
	}

	public void setIsReadOnly(boolean isReadOnly) {
		synchronized (lock) {
			this.isReadOnly = isReadOnly;
		}
	}

	public void addPropertyChangeListener(PropertyChangeListener e) {
		synchronized (lock) {
			propertyChangeListeners.add(PropertyChangeListener.class, e);
		}
	}

	public void removePropertyChangeListener(PropertyChangeListener e) {
		synchronized (lock) {
			propertyChangeListeners.remove(PropertyChangeListener.class, e);
		}
	}

	protected void firePropertyChanged(String changed, String oldItem, String newItem) {
		synchronized (lock) {
			if (propertyChangeListeners.getListenerCount()>0) {
				EventListener[] listeners = propertyChangeListeners.getListeners(PropertyChangeListener.class);
				PropertyChangeEvent pce = new PropertyChangeEvent(this, changed, oldItem, newItem);
				for (int i=0; i < listeners.length; i++) {
					PropertyChangeListener listener = (PropertyChangeListener)listeners[i];
					listener.propertyChange(pce);
				}
			}
		}
	}
	
	public abstract String getEmptyDisplayText();
	
	public String getConstraintName() {
		synchronized (lock) {
			return constraintName;
		}
	}

	public void setConstraintName(String constraintName) {
		String oldText = this.constraintName;
		synchronized (lock) {
			this.constraintName = constraintName;
			firePropertyChanged("constraintName",oldText,this.constraintName);
		}
	}

	public void setLock(Object lock) {
		synchronized (this.lock) {
			this.lock = lock;
		}
	}
	
	protected void setDefaultDisplayTextNoEvent(String text) {
		synchronized (this.lock) {
			this.defaultDisplayText = text;
		}
	}

	public void clearListeners() {
		synchronized (this.lock) {
			Object[] items = new Object[propertyChangeListeners.getListenerList().length];
			System.arraycopy(propertyChangeListeners.getListenerList(), 0, items, 0, propertyChangeListeners.getListenerList().length);
			for (int i = 0; i < items.length-1; i +=2) {
				Class listenerClass = (Class)items[i];
				EventListener listener = (EventListener)items[i+1];
				propertyChangeListeners.remove(listenerClass, listener);
			}
		}
	}
	
	public Object getEditorID() {
		return editorID;
	}

	public void setEditorID(Object editorID) {
		this.editorID = editorID;
	}

	public Object getDisplayFieldCategory() {
		return displayFieldCategory;
	}

	public void setDisplayFieldCategory(Object displayFieldCategory) {
		this.displayFieldCategory = displayFieldCategory;
	}

}
