package gov.va.med.hds.querytool.ui;

import gov.va.med.hds.querytool.IAvailableColumnsList;
import gov.va.med.hds.querytool.IDisplayField;
import gov.va.med.hds.querytool.IQuery;
import gov.va.med.hds.querytool.dao.QueryDAOException;

import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

/**
 * Resolves the user interface events to query and UI updates.
 */
public class QueryColumnPageManager {
	private IQuery query;
	private IAvailableColumnsList availableColumnsList;
	private QueryColumnPage queryColumnPage;
	private ListDataListener queryListListener = new ListDataListener() {
		public void intervalAdded(ListDataEvent e) {
			fireItemAdded(e);
		}

		public void intervalRemoved(ListDataEvent e) {
			fireItemRemoved(e);
		}

		public void contentsChanged(ListDataEvent e) {
			fireContentsChanged(e);
		}
	};
	private IQueryPageListener pageListener = new IQueryPageListener() {
		public void pageEvent(IQueryPageEvent e) {
			fireUIEvent(e);
		}
	};
	private IColumnListModel columnListModel = new ColumnListModel();
	private AvailableColumnList availableColumnList = new AvailableColumnList();
	private SampleTableModel sampleTableModel = new SampleTableModel();

	/**
	 * Method setQueryModel.
	 * @param query
	 */
	public void setQueryModel(IQuery query) {
		if (this.query!=null) {
			this.query.getDisplayFields().removeListDataListener(queryListListener);
		}
		this.query = query;
		if (query!=null) {
			query.getDisplayFields().addListDataListener(queryListListener);
		}
		syncColumnModel();
	}
	
	/**
	 * Method setAvailableColumnsModel.
	 * @param availableList
	 */
	public void setAvailableColumnsModel(IAvailableColumnsList availableColumnsList) {
		this.availableColumnsList = availableColumnsList;
		sampleTableModel.setAvailableColumnsModel(availableColumnsList);
		availableColumnList.setAvailableColumnModel(availableColumnsList);
	}
	
	/**
	 * Method setPageModel.
	 * @param queryColumnPage
	 */
	public void setPageModel(QueryColumnPage queryColumnPage) throws QueryDAOException {
		this.queryColumnPage = queryColumnPage;
		
		queryColumnPage.setInputListModel(availableColumnList);
		queryColumnPage.setTableModel(sampleTableModel);
		queryColumnPage.setTableColumnModel(columnListModel);
		queryColumnPage.setOutputListModel(columnListModel);
		queryColumnPage.setOutputListModel(columnListModel);

		queryColumnPage.addPageListener(pageListener);
	}
	
	private void syncColumnModel() {
		clearAll(columnListModel);
		for (int i = 0; i < query.getDisplayFields().size(); i++) {
			IDisplayField f = query.getDisplayFields().get(i);
			addColumnFromField(availableColumnsList.indexOf(f), f);
		}
	}

	private void fireContentsChanged(ListDataEvent e) {
		this.queryColumnPage.setOutSelected(e.getIndex1());
		if (columnListModel.getColumn(e.getIndex0()).getIdentifier().equals(query.getDisplayFields().get(e.getIndex0()))) {
			return;
		}
		columnListModel.moveColumn(e.getIndex0(), e.getIndex1());
	}

	private void fireItemRemoved(ListDataEvent e) {
		for (int i = 0; i < e.getIndex1()-e.getIndex0()+1; i++) {
			TableColumn col = columnListModel.getColumn(e.getIndex0());
			columnListModel.removeColumn(col);
		}
	}

	private void fireItemAdded(ListDataEvent e) {
		for (int i = e.getIndex0(); i < e.getIndex1()+1; i++) {
			IDisplayField f = (IDisplayField)query.getDisplayFields().get(i);
			int dataPos = availableColumnsList.indexOf(f);
			addColumnFromField(dataPos, f);
		}
	}

	private void addColumnFromField(int dataPosition, IDisplayField f) {
		IDisplayField refField = availableColumnsList.getMutable(dataPosition);
		if (refField==null) return;
		TableColumn col = new TableColumn();
		col.setHeaderValue(f.getDisplayText());
		col.setIdentifier(refField);
		col.setModelIndex(dataPosition);
		col.setPreferredWidth(f.getColumnWidth());
		col.setWidth(f.getColumnWidth());
		col.setResizable(true);
		columnListModel.addColumn(col);
	}

	private void clearAll(TableColumnModel columnModel) {
		while (columnModel.getColumnCount()>0) {
			TableColumn col = columnModel.getColumn(0);
			columnModel.removeColumn(col);
		}
	}

	private void fireUIEvent(IQueryPageEvent e) {
		switch (e.getEventID()) {
			case IQueryPageEvent.CLEAR_ALL: {
				query.getDisplayFields().clearAll();
				break;
			}
			case IQueryPageEvent.SELECT_ALL: {
				addAll();
				break;
			}
			case IQueryPageEvent.MOVE_DOWN: {
				query.getDisplayFields().moveDown(queryColumnPage.getOutSelectionModel().getAnchorSelectionIndex());
				break;
			}
			case IQueryPageEvent.MOVE_UP: {
				query.getDisplayFields().moveUp(queryColumnPage.getOutSelectionModel().getAnchorSelectionIndex());
				break;
			}
			case IQueryPageEvent.MOVE_COLUMN: {
				updateQuery(e.getFromIndex(),e.getToIndex());
				break;
			}
			case IQueryPageEvent.REMOVE: {
				if (e.getIndexArray()!=null && e.getIndexArray().length>0) {
					for (int i = e.getIndexArray().length-1; i > -1; i--) {
						removeItems(e.getIndexArray()[i],e.getIndexArray()[i]);
					}
				}
				break;
			}
			case IQueryPageEvent.SELECT: {
				if (e.getIndexArray()!=null && e.getIndexArray().length>0) {
					for (int i = 0; i < e.getIndexArray().length; i++) {
						addItems(e.getIndexArray()[i],e.getIndexArray()[i]);
					}
				}
				break;
			}
			case IQueryPageEvent.COLUMN_RESIZE: {
				resizeRequest();
				break;
			}
		}
	}
	
	private void resizeRequest() {
		for (int i = 0; i < query.getDisplayFields().size(); i++) {
			IDisplayField field = query.getDisplayFields().get(i);
			TableColumn column = columnListModel.getColumn(i);
			if (field.getColumnWidth()!=column.getWidth()) {
				field.setColumnWidth(column.getWidth());
			}
		}
	}

	private void updateQuery(int fromIndex, int toIndex) {
		if (columnListModel.getColumn(fromIndex).getIdentifier().equals(query.getDisplayFields().get(fromIndex))) {
			return;
		}
		if (fromIndex-toIndex>0) {
			for (int i = fromIndex; i > toIndex; i--) {
				query.getDisplayFields().moveUp(i);
			}
		}
		else if (toIndex-fromIndex>0) {
			for (int i = fromIndex; i < toIndex; i++) {
				query.getDisplayFields().moveDown(i);
			}
		}
	}

	private void addAll() {
		addItems(0,availableColumnsList.size()-1);
	}
	
	private void addItems(int fromIndex, int toIndex) {
		for (int i = fromIndex; i < toIndex+1; i++) {
			query.getDisplayFields().add(availableColumnsList.getMutable(i));
		}
	}
	
	private void removeItems(int fromIndex, int toIndex) {
		for (int i = toIndex; i > fromIndex-1; i--) {
			query.getDisplayFields().removeAt(fromIndex);
		}
	}
	
	/**
	 * Clears all models.
	 */
	public void clearModels() {
		if (columnListModel!=null) {
			clearAll(columnListModel);
		}
		if (query!=null) {
			query.getDisplayFields().removeListDataListener(queryListListener);
			query = null;
		}
		if (queryColumnPage!=null) {
			queryColumnPage.removePageListener(pageListener);
			queryColumnPage.clearModels();
			queryColumnPage = null;
		}
	}
	
}
