package gov.va.med.hds.querytool.ui.printables;

import gov.va.med.hds.cd.print.ConstructionException;
import gov.va.med.hds.cd.print.printables.DefaultPrintable;
import gov.va.med.hds.cd.print.printables.HeaderFooter;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.spectrum.linkpanel.IHTMLSelectionModel;
import gov.va.med.hds.querytool.IQuery;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.util.ArrayList;

import javax.swing.JEditorPane;
import javax.swing.JTable;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.text.View;

/**
 * The printable for a query result table.
 * Provides all of the graphics 3D based printing for the query result table.
 */
public class QueryResultTablePrintable extends DefaultPrintable {
	private static final int CONSTRAINT_BUFFER = 20;
	private static final String REPORT_NAME_SUFFIX = " Report";
	private static final double TABLE_SCALE_X = 0.7;
	private static final double TABLE_SCALE_Y = 0.7;
	private static final double TABLE_SCALE_INV_X = 1.0/TABLE_SCALE_X;
	private static final double TABLE_SCALE_INV_Y = 1.0/TABLE_SCALE_Y;
	private static final int DEFAULT_CONSTRAINT_FONT_SIZE = 8;
	private static final int DEFAULT_CONSTRAINT_FONT_STYLE = Font.PLAIN;
	private static final String DEFAULT_CONSTRAINTS_FONT = "MS Sans Serif";
	
	private static final String CONSTRAINTS_LABEL_FONT = "Arial Narrow";
	private static final int CONSTRAINTS_LABEL_STYLE = Font.BOLD;
	private static final int QUERY_DATE_CONTENT_STYLE = DEFAULT_CONSTRAINT_FONT_STYLE;
	private static final int CONSTRAINTS_LABEL_FONT_SIZE = 12;
	private static final String CONSTRAINTS_LABEL_TEXT = "Constraints";
	private static final String QUERY_DATE_LABEL_TEXT = "Query Executed On: ";
	private static final int HORIZONTAL_TEXT_INSET_RIGHT = 5;
	private static final int VERTICAL_TEXT_INSET_TOP = 5;
	private static final int CONSTRAINTS_DIVIDER = DEFAULT_CONSTRAINT_FONT_SIZE;
	private static final int CONSTRAINTS_RIGHT_MARGIN_FUDGE_FACTOR = DEFAULT_CONSTRAINT_FONT_SIZE;

	private static final int ROW_HEADER_TOP_INSET = 2;
	private static final int ROW_HEADER_BOTTOM_INSET = 4;
	private static final int ROW_HEADER_RIGHT_INSET = 5;
	private static final int ROW_HEADER_LEFT_INSET = 5;
	private static final int HEADER_FONT_STYLE = Font.BOLD;
	private static final float HEADER_LINE_WIDTH = 1.0f;
	private static final float HEADER_BOTTOM_LINE_WIDTH = 2.0f;

	private static final int CELL_TOP_INSET = 2;
	private static final int CELL_BOTTOM_INSET = 6;
	private static final int CELL_RIGHT_INSET = 5;
	private static final int CELL_LEFT_INSET = 5;
	private static final int CELL_FONT_STYLE = Font.PLAIN;
	private static final float CELL_LINE_WIDTH = 1.0f;

	private static final int NON_HEADER_PAGE_TABLE_OFFSET_TOP = 5;
	private static final int TABLE_OFFSET_BOTTOM = 5;

	private IQuery query;
	private String executionDate;
	private HeaderFooter headerFooter;
	private IHTMLSelectionModel constraints;
	private JTable table;
	private static final String HEADER_LEFT_LABEL = "Care Management";
	private double constraintsCalculatedHeight = 0;
	private int pageCount = 0;
	private int headerPageCount = 0;
	private double topPageRowHeight = 0;
	private ArrayList tablePageDescriptors = new ArrayList(); 

	/**
	 * Constructor
	 * @param name - name to call the print job
	 */
	public QueryResultTablePrintable(IRuntimeContext runtime, JTable table, IHTMLSelectionModel queryConstraints, String executionDate, IQuery query) throws ConstructionException {
		super(runtime);
		this.table = table;
		this.constraints = queryConstraints;
		this.executionDate = executionDate;
		this.query = query;
		headerFooter = new HeaderFooter(runtime, HEADER_LEFT_LABEL, query.getDisplayName()+REPORT_NAME_SUFFIX);
	}
	/** 
	 * @see java.awt.print.Printable#print(java.awt.Graphics, java.awt.print.PageFormat, int)
	 */
	public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
		if (pageIndex == 0) {	
			processFirstPage(graphics,pageFormat);		
		}else if(pageIndex >= pageCount){
			if(pageCount==0)
				processFirstPage(graphics.create(),pageFormat);
			if(pageIndex >= pageCount)
				return Printable.NO_SUCH_PAGE;
			else
				processNonFirstPage(graphics,pageFormat,pageIndex);			
		}else{
			processNonFirstPage(graphics,pageFormat,pageIndex);
		}
		
		return Printable.PAGE_EXISTS;
	}
	
	private void processFirstPage(Graphics graphics, PageFormat pageFormat) throws PrinterException{
		init();
		Graphics2D g2d = printHeader(graphics, pageFormat, 0);

		double headHeight = headerFooter.getJustRenderedHeaderHeight();
		double footHeight = headerFooter.getJustRenderedFooterHeight();
		double height = pageFormat.getImageableHeight() - headHeight - footHeight - TABLE_OFFSET_BOTTOM;
		double width = pageFormat.getImageableWidth();
		double constraintHeight = height-CONSTRAINT_BUFFER;
		Graphics2D constraintg2d = (Graphics2D) g2d.create(0, (int)(headHeight+1.0), (int)width, (int)constraintHeight);
		constraintsCalculatedHeight = printContraints(constraintg2d, width, constraintHeight);
		if(table.getColumnCount() > 0){
			topPageRowHeight = height - constraintsCalculatedHeight;
			calculatePagination(g2d,scaleY(height),scaleX(width-1),scaleY(topPageRowHeight));
			Graphics2D tableg2d = (Graphics2D) g2d.create(0, (int) (headHeight + constraintsCalculatedHeight + 1.0), (int)scaleX(width), (int)scaleY(topPageRowHeight));
			tableg2d.scale(TABLE_SCALE_X,TABLE_SCALE_Y);
			printTable(0,tableg2d, scaleX(width), scaleY(topPageRowHeight));
		}
	}
	private void processNonFirstPage(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException{
		Graphics2D g2d = printHeader(graphics, pageFormat, pageIndex);

		double headHeight = headerFooter.getJustRenderedHeaderHeight();
		double footHeight = headerFooter.getJustRenderedFooterHeight();
		double height = pageFormat.getImageableHeight() - headHeight - footHeight - TABLE_OFFSET_BOTTOM;
		double width = pageFormat.getImageableWidth();
			
		if(pageIndex >= headerPageCount){
			Graphics2D tableg2d = (Graphics2D) g2d.create(0, (int) (headHeight + 1.0 + NON_HEADER_PAGE_TABLE_OFFSET_TOP), (int)scaleX(width), (int)scaleY(height));
			tableg2d.scale(TABLE_SCALE_X,TABLE_SCALE_Y);
			printTable(pageIndex,tableg2d, scaleX(width), scaleY(height));
		}else{
			height = topPageRowHeight;
			Graphics2D tableg2d = (Graphics2D) g2d.create(0, (int) (headHeight + constraintsCalculatedHeight + 1.0), (int)scaleX(width), (int)scaleY(height));
			tableg2d.scale(TABLE_SCALE_X,TABLE_SCALE_Y);
			printTable(pageIndex,tableg2d, scaleX(width), scaleY(height));				
		}		
	}
	private void init() {
		constraintsCalculatedHeight = 0;
		pageCount = 0;
		headerPageCount = 0;
		topPageRowHeight = 0;
		tablePageDescriptors.clear();
	}
	private double scaleX(double val){
		return val*TABLE_SCALE_INV_X;
	}
	private double scaleY(double val){
		return val*TABLE_SCALE_INV_Y;
	}
	private Graphics2D printHeader(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
		Graphics2D g2d = (Graphics2D) graphics.create();
		headerFooter.print(g2d, pageFormat, pageIndex);
		g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
		return g2d;
	}
	
	private void calculatePagination(Graphics2D graphics, double height, double width, double topPageRowHeight) {
		int colCount = table.getColumnCount();
		TableColumnModel colModel = table.getColumnModel();
		HeaderRenderer header = new HeaderRenderer(table,graphics);
		CellRenderer cell = new CellRenderer(table,graphics);
		
		double allCellArea = Math.max(0.0,topPageRowHeight - header.getHeight());
		int maxRowCount = (int)Math.floor(allCellArea/cell.getHeight());
		int startRow = 0;
		double accumWidth = 0.0;
		TablePageDescriptor desc; 
		int startCol=0,endCol;
		int rowsCalculated = Math.min(maxRowCount,table.getRowCount());
		
		// Add from left to right page columns
		for(int i=0;i<colCount;i++){
			if(accumWidth+colModel.getColumn(i).getWidth()>width){
				endCol = i-1;
				tablePageDescriptors.add(new TablePageDescriptor(true,startCol,endCol,startRow,rowsCalculated-1));
				startCol = i;
				endCol = i;
				accumWidth = 0;
			}
			accumWidth += colModel.getColumn(i).getWidth();
		}
		// Add last page column
		tablePageDescriptors.add(new TablePageDescriptor(true,startCol,colCount-1,startRow,rowsCalculated-1));
		
		headerPageCount = tablePageDescriptors.size();
		Object[] headerDescriptors = tablePageDescriptors.toArray();
		
		// Now create pages from rows
		int totalRows = table.getRowCount();
		while(rowsCalculated < totalRows){
			startRow = rowsCalculated;
			maxRowCount = (int)Math.floor(height/cell.getHeight());
			int newRowsCalculated = Math.min(maxRowCount,table.getRowCount()-rowsCalculated);
			for(int j=0;j<headerPageCount;j++){
				desc = (TablePageDescriptor)headerDescriptors[j];
				tablePageDescriptors.add(new TablePageDescriptor(false,desc.startColumn,desc.endColumn,startRow,startRow+newRowsCalculated-1));
			}
			rowsCalculated += newRowsCalculated;
		}
		pageCount = tablePageDescriptors.size();
	}
	private double printTable(int pageIndex,Graphics2D graphics, double width, double height) throws PrinterException {		
		TablePageDescriptor desc = (TablePageDescriptor)tablePageDescriptors.get(pageIndex);
		double startX = 0;
		double startY = 0;
		int borderFlags = 0x0;
		if(desc.hasHeader){			
			HeaderRenderer header = new HeaderRenderer(table,graphics);
			for(int i=desc.startColumn;i<=desc.endColumn;i++){
				borderFlags = LINE_TOP|LINE_BOTTOM|LINE_RIGHT;
				if(i==desc.startColumn)
					borderFlags |= LINE_LEFT;
				startX += header.draw(graphics,(int)startX,(int)startY,i,borderFlags);
			}
			startY += header.getHeight()+1;
		}
		CellRenderer cell = new CellRenderer(table,graphics);
		for(int i=desc.startRow;i<=desc.endRow;i++){
			startX = 0;
			for(int j=desc.startColumn;j<=desc.endColumn;j++){
				borderFlags = LINE_BOTTOM|LINE_RIGHT;
				if(j==desc.startColumn)
					borderFlags |= LINE_LEFT;
				if(i==desc.startRow)
					borderFlags |= LINE_TOP;
				try{
					startX += cell.draw(graphics,(int)startX,(int)startY,i,j,borderFlags);
				}catch(Exception e){
					throw new PrinterException(e.getMessage());
				}
			}
			startY += cell.getHeight();
		}
		return startY;
	}
	private static final int LINE_TOP = 0x01;
	private static final int LINE_BOTTOM = 0x02;
	private static final int LINE_LEFT = 0x04;
	private static final int LINE_RIGHT = 0x08;

	
	private double printContraints(Graphics2D graphics, double maxWidth, double maxHeight) {
		if (constraints == null)
			return 0;
		double constraintsHeight = 0;

		constraintsHeight = drawConstraintsLabels(graphics, maxWidth, constraintsHeight);
		double width = maxWidth - CONSTRAINTS_RIGHT_MARGIN_FUDGE_FACTOR;
		maxHeight = maxHeight - constraintsHeight;
		Graphics2D g2d = (Graphics2D) graphics.create(0, (int) (constraintsHeight + 1.0), (int) width, (int) (maxHeight));
		constraintsHeight += drawConstraintsContent(g2d, width, maxHeight);
		
		return constraintsHeight;
	}
	private double drawConstraintsContent(Graphics2D graphics, double suggestedWidth, double maxHeight) {
		Font font = new Font(DEFAULT_CONSTRAINTS_FONT,DEFAULT_CONSTRAINT_FONT_STYLE,DEFAULT_CONSTRAINT_FONT_SIZE);
		JEditorPane pane = new JEditorPane("text/html", constraints.getHTML(font));
		pane.setSize((int) suggestedWidth, (int)maxHeight);
		pane.validate();
		View view = pane.getUI().getRootView(pane);
		Rectangle allocation = new Rectangle(0,0,(int)suggestedWidth,(int)pane.getPreferredSize().getHeight());
		view.paint(graphics,allocation);		
		return (float)Math.min(pane.getPreferredSize().getHeight(),maxHeight);
	}
	private double drawConstraintsLabels(Graphics2D graphics, double maxWidth, double constraintsHeight) {
		Font constraintLabelFont = new Font(CONSTRAINTS_LABEL_FONT, CONSTRAINTS_LABEL_STYLE, CONSTRAINTS_LABEL_FONT_SIZE);
		Font queryDateFont = new Font(CONSTRAINTS_LABEL_FONT, QUERY_DATE_CONTENT_STYLE, CONSTRAINTS_LABEL_FONT_SIZE);
		String constraintsLabel = query.getDisplayName() + " " + CONSTRAINTS_LABEL_TEXT;
		Rectangle2D constraintLabelBounds = constraintLabelFont.getStringBounds(constraintsLabel, graphics.getFontRenderContext());
		Rectangle2D queryDateLabelBounds = constraintLabelFont.getStringBounds(QUERY_DATE_LABEL_TEXT, graphics.getFontRenderContext());
		Rectangle2D queryDateBounds = queryDateFont.getStringBounds(executionDate, graphics.getFontRenderContext());

		graphics.setColor(Color.BLACK);
		graphics.setFont(constraintLabelFont);
		float textY = (float) (Math.max(constraintLabelBounds.getHeight(), queryDateBounds.getHeight()) + VERTICAL_TEXT_INSET_TOP);
		float textX = (float) (maxWidth - HORIZONTAL_TEXT_INSET_RIGHT - queryDateBounds.getWidth() - queryDateLabelBounds.getWidth());
		graphics.drawString(constraintsLabel, 0, textY);
		graphics.drawString(QUERY_DATE_LABEL_TEXT, textX, textY);
		graphics.setFont(queryDateFont);
		textX = (float) (maxWidth - HORIZONTAL_TEXT_INSET_RIGHT - queryDateBounds.getWidth());
		graphics.drawString(executionDate, textX, textY);
		constraintsHeight += textY;
		constraintsHeight += CONSTRAINTS_DIVIDER;
		return constraintsHeight;
	}
	private class TablePageDescriptor{
		public TablePageDescriptor(boolean hasHeader, int startColumn, int endColumn, int startRow, int endRow){
			this.hasHeader = hasHeader;
			this.startColumn = startColumn;
			this.endColumn = endColumn;
			this.startRow = startRow;
			this.endRow = endRow;
		}
		boolean hasHeader = false;
		int startColumn = 0;
		int endColumn = 0;
		int startRow = 0;
		int endRow = 0;
	}
	private class HeaderRenderer{
		public HeaderRenderer(JTable table, Graphics2D graphics){
			this.table = table;
			calculateHeight(graphics);
		}
		
		public double draw(Graphics2D graphics, int x, int y, int col, int borders){
			TableColumn column = table.getColumnModel().getColumn(col);
			graphics.setColor(Color.BLACK);
			width = column.getWidth();

			graphics.setStroke(new BasicStroke(HEADER_LINE_WIDTH));
			int yy = (int)Math.ceil(y+height);
			int xx = (int)Math.ceil(x+width);
			if((borders & LINE_LEFT) > 0){
				graphics.drawLine(x,y,x,yy);
			}
			if((borders & LINE_TOP) > 0){
				graphics.drawLine(x,y,xx,y);
			}
			if((borders & LINE_RIGHT) > 0){
				graphics.drawLine(xx,y,xx,yy);
			}
			graphics.setStroke(new BasicStroke(HEADER_BOTTOM_LINE_WIDTH));
			if((borders & LINE_BOTTOM) > 0){
				graphics.drawLine(x,yy,xx-1,yy);
			}			
			Rectangle savedClipBounds = graphics.getClipBounds();
			double clipx = Math.ceil(x+ROW_HEADER_LEFT_INSET);
			double clipy = Math.ceil(y+ROW_HEADER_TOP_INSET);
			double clipWidth = Math.ceil(width-ROW_HEADER_LEFT_INSET-ROW_HEADER_RIGHT_INSET);
			double clipHeight = Math.ceil(height-ROW_HEADER_BOTTOM_INSET);
			
			graphics.setClip((int)clipx,(int)clipy,(int)clipWidth,(int)clipHeight);			
			graphics.setFont(font);
			int textX = (int)Math.ceil(x+ROW_HEADER_LEFT_INSET);
			int textY = (int)Math.ceil(y+ROW_HEADER_TOP_INSET+fontHeight);
			graphics.drawString(column.getHeaderValue().toString(),textX,textY);

			graphics.setClip(savedClipBounds);
			return width;
		}
		public double getHeight(){
			return height;
		}
		public double getWidth(){
			return width;
		}
		private void calculateHeight(Graphics2D graphics) {
			font = new Font(table.getFont().getFamily(),HEADER_FONT_STYLE,table.getFont().getSize());
			Rectangle2D rect = font.getMaxCharBounds(graphics.getFontRenderContext());
			fontHeight = rect.getHeight();
			height = fontHeight+ROW_HEADER_TOP_INSET+ROW_HEADER_BOTTOM_INSET;
		}
		private JTable table;
		private double height=0;
		private double fontHeight=0;
		private double width=0;
		private Font font;
	}
	private class CellRenderer{
		public CellRenderer(JTable table, Graphics2D graphics){
			this.table = table;
			calculateHeight(graphics);
		}
		public double draw(Graphics2D graphics, int x, int y, int row, int col, int borders){
			TableColumn column = table.getColumnModel().getColumn(col);
			graphics.setColor(Color.BLACK);
			width = column.getWidth();
			graphics.setStroke(new BasicStroke(CELL_LINE_WIDTH));
			int yy = (int)Math.ceil(y+height);
			int xx = (int)Math.ceil(x+width);
			if((borders & LINE_LEFT) > 0){
				graphics.drawLine(x,y,x,yy);
			}
			if((borders & LINE_TOP) > 0){
				graphics.drawLine(x,y,xx,y);
			}
			if((borders & LINE_RIGHT) > 0){
				graphics.drawLine(xx,y,xx,yy);
			}
			if((borders & LINE_BOTTOM) > 0){
				graphics.drawLine(x,yy,xx,yy);
			}
			Rectangle savedClipBounds = graphics.getClipBounds();
			double clipx = Math.ceil(x+CELL_LEFT_INSET);
			double clipy = Math.ceil(y+CELL_TOP_INSET);
			double clipWidth = Math.ceil(width-CELL_LEFT_INSET-CELL_RIGHT_INSET);
			double clipHeight = Math.ceil(height);
			
			graphics.setClip((int)clipx,(int)clipy,(int)clipWidth,(int)clipHeight);			
			graphics.setFont(font);
			int textX = (int)Math.ceil(x+ROW_HEADER_LEFT_INSET);
			int textY = (int)Math.ceil(y+ROW_HEADER_TOP_INSET+fontHeight);
			String cellValue = table.getModel().getValueAt(row,column.getModelIndex()).toString();
			graphics.drawString(cellValue,textX,textY);

			graphics.setClip(savedClipBounds);

			return width;	
		}
		public double getHeight(){
			return height;
		}
		public double getWidth(){
			return width;
		}
		private void calculateHeight(Graphics2D graphics) {
			font = new Font(table.getFont().getFamily(),CELL_FONT_STYLE,table.getFont().getSize());
			Rectangle2D rect = font.getMaxCharBounds(graphics.getFontRenderContext());
			fontHeight = rect.getHeight();
			height = fontHeight+CELL_TOP_INSET+CELL_BOTTOM_INSET;
		}
		private JTable table;
		private double height=0;
		private double width=0;
		private double fontHeight=0;
		private Font font;
	}

}
