package gov.va.med.hds.querytool;

import java.util.List;

/**
 * Interface for managing a list of queries (contained in ISoftListItems.)
 * 
 * The actual query references are stored in soft references, which could be cleaned up
 * by the JVM at any time.  If a query is a new query or changed, a hard reference is 
 * kept to keep the JVM from cleaning up the object and the reference.
 * 
 * If an actual query object is not stored by the ISoftListItem, an ID and displayName
 * is stored in its place.
 */
public interface IQueryList {
	
	/**
	 * Creates an ISoftListItem for the given query and stores the item in the list.
	 * Inserts the item in the list in the toString alphabetical order position. 
	 * @param query
	 */
	void addElement(IQuery query) throws QueryListException;
	
	/**
	 * Creates an ISoftListItem for the given queryID and displayName.  Inserts the 
	 * item in the list in the toString alphabetical order position. 
	 * @param queryID
	 * @param displayName
	 */
	void addElement(IQueryID queryID, String displayName) throws QueryListException;
	
	/**
	 * Removes the ISoftListItem at the specified index.
	 * @param index
	 * @throws QueryListException
	 */
	void removeElementAt(int index) throws QueryListException;
	
	/**
	 * Removes the item that has the queryID specified by the given query from the
	 * list.  If the item does not exists, method does nothing.
	 * @param query
	 */
	void removeElement(IQuery query) throws QueryListException;
	
	/**
	 * Removes the item that matches the specified queryID.  If the item does not
	 * exist, this method does nothing. 
	 * @param queryID
	 * @throws QueryListException
	 */
	void removeElement(IQueryID queryID) throws QueryListException;
	
	/**
	 * Removes the specified soft list item.  If the item does not exist, this routine
	 * does nothing.
	 * @param listItem
	 * @throws QueryListException
	 */
	void removeElement(ISoftListItem listItem) throws QueryListException;
	
	/**
	 * Returns the query list as an immutable list type.
	 * @return list
	 */
	List getAsList();

	/**
	 * Returns the index of the item that contains the given queryID in the list.
	 * If the item does not exist, returns -1.
	 * @param queryID
	 * @return int
	 */
	int indexOf(IQueryID queryID);
	
	/**
	 * Returns the index of the item that has a queryID that matches the queryID in
	 * the given query.  If it does not exist, it returns a -1.
	 * @param query
	 * @return int
	 */
	int indexOf(IQuery query);

	/**
	 * Returns the index of the item that has a matching displayName.  Returns -1
	 * otherwise.
	 * @param displayName
	 * @return int
	 */
	int indexOf(String displayName);

	/**
	 * Returns the size of the list.
	 * @return int
	 */
	int getSize();
	
	/**
	 * Returns the query from the ISoftListItem at the given index.  If the item does
	 * not have a reference of the query, returns null.
	 * @param index
	 * @return query
	 */
	IQuery getQueryAt(int index);
	
	/**
	 * Returns the ISoftListItem at the specified index.  If the index is out of range,
	 * this method return null.
	 * @param index
	 * @return
	 */
	ISoftListItem getItemAt(int index);
	
	/**
	 * Adds a query list listener.
	 * @param listener
	 */
	void addListener(IQueryListListener listener);
	
	/**
	 * Removes a registered query list listener.
	 * @param listener
	 */
	void removeListener(IQueryListListener listener);

	/**
	 * Returns the index of the given item.
	 * @param item
	 * @return index
	 */
	int indexOf(ISoftListItem item);
	
}
