/*
 * Initially Created on Apr 16, 2003 by vhaislmonsos
 *
 */
package gov.va.med.hds.querytool;

import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.querytool.dao.ICompleteQueryDataDAO;

import java.util.Date;

/**
 * Encapsulates the execution of a query with an IRunDAO and its own execution thread.
 */
public interface IRunQuery {
	
	/**
	 * Sets the query to be executed.
	 */
	void setQuery(IQuery query);
	
	/**
	 * Sets the runtime so the implementing class can get the necessary DAOs to execute 
	 * the query.
	 * @param runtime
	 */
	void setRuntime(IRuntimeContext runtime);
	
	/**
	 * Returns the count of detail rows in the report.  This is the number of detail
	 * items for all the patients found.
	 * @return int
	 */
	int getRowCount();
	
	/**
	 * Returns the number of patients excluded because of report parameters.
	 * @return int
	 */
	int getScreenedCount();
	
	/**
	 * The number of patients whose detail rows make up the report.
	 * @return int
	 */
	int getPatientCount();
	
	/**
	 * Returns the running state of the thread managed by the implementing class.  If true, the run 
	 * thread is currently executing; if false, the run thread has not been created or has been abandoned.
	 * @return boolean
	 */
	boolean isRunning();
	
	/**
	 * When all of the set methods have been called, tells this class to start a run
	 * thread and execute the query.  If a query is currently running, this call is 
	 * ignored.  This thread remains in existance until the implementing class raises certain
	 * events (see the isRunning method.)  The implementing class should use the stopQuery 
	 * method to signal the thread to shut down.
	 */
	void runQuery();
	
	/**
	 * Permanently stops a running query.  A subsequent call to runQuery restarts the 
	 * query execution from the beginning, and NOT from the point after the stopQuery
	 * call.
	 */
	void stopQuery();
	
	/**
	 * Adds a listener for events.
	 * @param listener
	 */
	void addRunListener(IRunListener listener);
	
	/**
	 * Removes the specified listener.
	 * @param listener
	 */
	void removeRunListener(IRunListener listener);
	
	/**
	 * If a report failure event is raised, this method can be called to get the
	 * exception class for the failure.
	 * @return exception
	 */
	ExecutionException getException();

	/**
	 * Sets the number of display rows that the implementation will retrieve until
	 * the number is exceeded.  After that, the query will stop and issue a max
	 * reached event.
	 * @param mawRows
	 */
	void setMaxRows(int maxRows);
	
	/**
	 * Restarts the query execution when paused for sensitive patients.
	 */
	void resumeFromPause();
	
	/**
	 * Returns the DAO that manages the setup of the sensitive patient iterator.  If the execution state
	 * of the query is not at the point where this is available, this method returns null.
	 * @return sensitiveDataDAO
	 */
	ISensitiveDataModel getSensitiveDataModel();
	
	/**
	 * Returns the DAO that can be used to get report results from the database.  Returns null if the
	 * report does not contain data.
	 * @return completeQueryDataDAO
	 */
	ICompleteQueryDataDAO getCompleteQueryDataDAO();
	
	/**
	 * Returns the Date that the current report was started.
	 * @return startDateTime
	 */
	Date getStartDate();

	/**
	 * Clears the model from the currently executing query, or clears the completed query.  A query will not
	 * be executing nor will results be available after this call is made.
	 */
	void clear();
	
}
