package gov.va.med.hds.cd.runtime.internal;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.event.EventListenerList;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IPlugin;
import gov.va.med.hds.cd.runtime.IPluginActivationListener;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.PluginActivatingException;
import gov.va.med.hds.cd.runtime.PluginEvent;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginShutdownException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;
import gov.va.med.hds.cd.runtime.logging.ILogger;

/**
 * The ActivePluginManager is responsible for tracking which plugins
 * are currently active.  If a plugin descriptor specficies a class
 * that implements <code>IPlugin</code>, <code>getPluginInstance()</code>
 * will lazily instantiate the specified class, call <code>startup(), and
 * then make that instance available to future clients.  The specified
 * plugin is then considered to be "activated."
 * <p>
 * When a plugin instance is activated, all of its prerequisites are also
 * activated recursively.
 *
 * @see gov.va.med.hds.cd.runtime.IPlugin
 */
public class ActivePluginManager {
	private static final String NO_CLASS_SPECIFIED =
		"this plugin has no specified class implementing '"
			+ IPlugin.class.getName()
			+ "'";

	private EventListenerList listenerList = new EventListenerList();
	private IRuntimeContext runtime;
	private ILogger log;
	private IPluginDescriptor[] plugins;

	private PluginDescriptorMap activePlugins;
	private PluginDescriptorMap activatingPlugins;

	public ActivePluginManager(IRuntimeContext runtime) {
		this.runtime = runtime;
		this.log = runtime.getLogger(getClass());
		this.activePlugins = new PluginDescriptorMap();
		this.activatingPlugins = new PluginDescriptorMap();
		plugins = runtime.getPluginRegistry().getPluginDescriptors();
	}

	public void addPluginActivationListener(IPluginActivationListener l) {
		listenerList.add(IPluginActivationListener.class, l);
	}

	public void removePluginActivationListener(IPluginActivationListener l) {
		listenerList.remove(IPluginActivationListener.class, l);
	}

	public Object getPluginInstance(
		String pluginId,
		Version version,
		VersionMatchCondition matchCondition)
		throws PluginStartupException, PluginNotFoundException {
		log.debug(
			"entering getPluginInstance("
				+ pluginId
				+ ","
				+ version
				+ ","
				+ matchCondition
				+ ")");
		IPluginDescriptor plugin =
			findMatchingPlugin(pluginId, version, matchCondition);
		if (plugin == null)
			throw new PluginNotFoundException(
				pluginId,
				version,
				matchCondition);
		if (isPluginActivating(plugin))
			throw new PluginActivatingException(plugin);
		if (!isPluginActive(plugin))
			activatePlugin(plugin);
		Object pluginInstance = activePlugins.get(plugin);
		if (pluginInstance == null)
			throw new PluginStartupException(plugin, NO_CLASS_SPECIFIED);
		log.debug(
			"exiting getPluginInstance("
				+ pluginId
				+ ","
				+ version
				+ ","
				+ matchCondition
				+ ")");
		return pluginInstance;
	}

	public boolean isPluginActive(
		String pluginId,
		Version version,
		VersionMatchCondition matchCondition) {
		return activePlugins.contains(pluginId, version, matchCondition);
	}

	/**
	 * Calls <code>shutdown()</code> on all active plugins, in the order
	 * implied by the dependency tree of the plugin descriptors, and 
	 * releases references to the managed instances of those plugins.
	 * 
	 * @see gov.va.med.hds.cd.runtime.IPlugin
	 */
	public void shutdown() {
		while (!activePlugins.isEmpty()) {
			List pluginsToShutdown = new LinkedList();
			for (Iterator i = activePlugins.keySet().iterator();
				i.hasNext();
				) {
				IPluginDescriptor plugin = (IPluginDescriptor) i.next();
				if (!hasDependentActivePlugins(plugin)) {
					pluginsToShutdown.add(plugin);
				}
			}

			for (Iterator i = pluginsToShutdown.iterator(); i.hasNext();) {
				try {
					IPluginDescriptor plugin = (IPluginDescriptor) i.next();
					shutdownPlugin(plugin);
					activePlugins.remove(plugin);
					log.info("plugin " + plugin + " deactivated");
					firePluginDeactivated(plugin);
				} catch (PluginShutdownException e) {
					log.error("Error deactivating plugin", e);
				}
			}
		}
		listenerList = null;
	}

	protected void firePluginActivated(IPluginDescriptor p) {
		final PluginEvent e = new PluginEvent(runtime, p);
		IPluginActivationListener[] listeners =
			(IPluginActivationListener[]) listenerList.getListeners(
				IPluginActivationListener.class);
		for (int i = 0; i < listeners.length; i++) {
			listeners[i].pluginActivated(e);
		}
	}

	protected void firePluginDeactivated(IPluginDescriptor p) {
		final PluginEvent e = new PluginEvent(runtime, p);
		IPluginActivationListener[] listeners =
			(IPluginActivationListener[]) listenerList.getListeners(
				IPluginActivationListener.class);
		for (int i = 0; i < listeners.length; i++) {
			listeners[i].pluginDeactivated(e);
		}
	}

	private boolean isPluginActivating(
		String pluginId,
		Version version,
		VersionMatchCondition matchCondition) {
		if (RuntimeUtilities.isRuntimePlugin(pluginId))
			return false;
		return activatingPlugins.contains(pluginId, version, matchCondition);
	}

	private boolean isPluginActive(IPluginDescriptor plugin) {
		return activePlugins.contains(plugin);
	}

	private boolean isPluginActivating(IPluginDescriptor plugin) {
		return activatingPlugins.contains(plugin);
	}

	private IPluginDescriptor findMatchingPlugin(
		String pluginId,
		Version version,
		VersionMatchCondition matchCondition) {
		return runtime.getPluginRegistry().getPluginDescriptor(
			pluginId,
			version,
			matchCondition);
	}

	private void activatePlugin(IPluginDescriptor pluginDescriptor)
		throws PluginNotFoundException, PluginStartupException {
		log.debug("entering activatePlugin(" + pluginDescriptor + ")");
		try {
			addActivatingPlugin(pluginDescriptor);
			activatePrerequisites(pluginDescriptor);
			instantiatePlugin(pluginDescriptor);
		} catch (PluginNotFoundException e) {
			throw e;
		} catch (PluginStartupException e) {
			throw e;
		} finally {
			removeActivatingPlugin(pluginDescriptor);
			log.debug("exiting activatePlugin(" + pluginDescriptor + ")");
		}
	}

	private void instantiatePlugin(IPluginDescriptor pluginDescriptor)
		throws PluginStartupException {
		log.debug("entering instantiatePlugin(" + pluginDescriptor + ")");
		if (hasNoPluginClass(pluginDescriptor)) {
			addActivePlugin(pluginDescriptor, null);
		} else {
			try {
				Class c = pluginDescriptor.getPluginClassLoader().loadClass(pluginDescriptor.getClassName());
				IPlugin pluginInstance = (IPlugin) c.newInstance();
				startupPlugin(pluginDescriptor, pluginInstance);
				addActivePlugin(pluginDescriptor, pluginInstance);
			} catch (ClassNotFoundException e) {
				throw new PluginStartupException(pluginDescriptor, e);
			} catch (InstantiationException e) {
				throw new PluginStartupException(pluginDescriptor, e);
			} catch (IllegalAccessException e) {
				throw new PluginStartupException(pluginDescriptor, e);
			}
		}
		log.debug("exiting instantiatePlugin(" + pluginDescriptor + ")");
	}

	private boolean hasNoPluginClass(IPluginDescriptor pluginDescriptor) {
		return pluginDescriptor.getClassName() == null
			|| pluginDescriptor.getClassName().equals("");
	}

	private void addActivePlugin(
		IPluginDescriptor descriptor,
		Object instance) {
		activePlugins.put(descriptor, instance);
		log.info("plugin " + descriptor + " is active");
		firePluginActivated(descriptor);
	}

	private void addActivatingPlugin(IPluginDescriptor plugin) {
		activatingPlugins.put(plugin, plugin);
		log.info("plugin " + plugin + " is activating");
	}

	private void removeActivatingPlugin(IPluginDescriptor plugin) {
		activatingPlugins.remove(plugin);
		log.debug("plugin " + plugin + " is no longer activating");
	}

	private void activatePrerequisites(IPluginDescriptor pluginDescriptor)
		throws PluginNotFoundException, PluginStartupException {
		log.debug("entering activatePrerequisites(" + pluginDescriptor + ")");
		IPluginPrerequisite[] prereqs =
			pluginDescriptor.getPluginPrerequisites();
		for (int i = 0; i < prereqs.length; i++) {
			IPluginDescriptor prereq =
				findMatchingPlugin(
					prereqs[i].getUniqueIdentifier(),
					prereqs[i].getVersion(),
					RuntimeUtilities.getPrerequisiteMatchCondition(prereqs[i]));
			if (prereq == null
				|| RuntimeUtilities.isRuntimePlugin(prereq)
				|| isPluginActive(prereq)
				|| isPluginActivating(prereq))
				continue;
			try {
				activatePlugin(prereq);
			} catch (PluginStartupException e) {
				throw new PluginStartupException(pluginDescriptor, e);
			}
		}
		log.debug("exiting activatePrerequisites(" + pluginDescriptor + ")");
	}

	private void startupPlugin(
		IPluginDescriptor pluginDescriptor,
		IPlugin plugin)
		throws PluginStartupException {
		try {
			log.debug("plugin " + pluginDescriptor + " is starting up");
			plugin.startup(runtime, pluginDescriptor);
		} catch (Exception e) {
			throw new PluginStartupException(pluginDescriptor, e);
		}
	}

	private void shutdownPlugin(IPluginDescriptor pluginDescriptor)
		throws PluginShutdownException {
		IPlugin plugin = (IPlugin) activePlugins.get(pluginDescriptor);
		if (plugin != null) {
			try {
				log.debug("plugin " + pluginDescriptor + " is shutting down");
				plugin.shutdown();
			} catch (Exception e) {
				throw new PluginShutdownException(pluginDescriptor, e);
			}
		}
	}

	private boolean hasDependentActivePlugins(IPluginDescriptor plugin) {
		IPluginDescriptor[] dependants = getDependentPlugins(plugin);
		for (int i = 0; i < dependants.length; i++) {
			if (isPluginActive(dependants[i])) {
				return true;
			}
		}

		return false;
	}

	private IPluginDescriptor[] getDependentPlugins(IPluginDescriptor plugin) {
		List dependants = new LinkedList();
		for (int i = 0; i < plugins.length; i++) {
			IPluginPrerequisite[] prereqs = plugins[i].getPluginPrerequisites();
			for (int j = 0; j < prereqs.length; j++) {
				if (prereqs[j]
					.getUniqueIdentifier()
					.equals(plugin.getUniqueIdentifier())) {
					dependants.add(plugins[i]);
				}
			}
		}

		return (IPluginDescriptor[]) dependants.toArray(
			new IPluginDescriptor[dependants.size()]);
	}
}
