package gov.va.med.hds.cd.runtime.internal;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.runtime.IRuntimeConstants;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidExtensionException;
import gov.va.med.hds.cd.runtime.InvalidExtensionPointException;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.InvalidPrerequisiteException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;
import gov.va.med.hds.util.ExceptionStackTrace;

public class DOMPluginReader implements IPluginReader, IRuntimeConstants {

	private static final Logger log = Logger.getLogger(DOMPluginReader.class.getName());

	private DocumentBuilder builder;
	private IRuntimeContext runtime;
	private DOMConfigurationElementReader configElementReader;

	public DOMPluginReader(IRuntimeContext runtime) {
		configElementReader = new DOMConfigurationElementReader(runtime);
		this.runtime = runtime;
		this.builder = createBuilder();
	}

	private DocumentBuilder createBuilder() {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);
			DocumentBuilder builder = factory.newDocumentBuilder();
			builder.setErrorHandler(new BuilderErrorHandler());
			return builder;
		} catch (FactoryConfigurationError e) {
			log.severe(ExceptionStackTrace.getStackTrace(e));
		} catch (ParserConfigurationException e) {
			log.severe(ExceptionStackTrace.getStackTrace(e));
		}
		return null;
	}

	public PluginDescriptor readPlugin(InputStream inputStream, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		return readPlugin(new InputSource(inputStream), classLoader);
	}

	public PluginDescriptor readPlugin(Reader reader, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		return readPlugin(new InputSource(reader), classLoader);
	}

	public PluginDescriptor readPlugin(InputSource inputSource, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		try {
			Document d = builder.parse(inputSource);
			Element pluginElement = (Element) d.getElementsByTagName(PLUGIN_ELEMENT).item(0);
			return readPlugin(pluginElement, classLoader);
		} catch (SAXException e) {
			throw new InvalidPluginDescriptorException(e);
		} catch (IOException e) {
			log.severe(ExceptionStackTrace.getStackTrace(e));
		}
		return null;
	}

	public PluginDescriptor readPlugin(Element pluginElement, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		String name = pluginElement.getAttribute(NAME_ATTRIBUTE);
		String id = pluginElement.getAttribute(ID_ATTRIBUTE);
		Version version = null;
		try {
			version = new Version(pluginElement.getAttribute(VERSION_ATTRIBUTE));
		} catch (VersionFormatException e) {
			throw new InvalidPluginDescriptorException(e);
		}
		String className = pluginElement.getAttribute(CLASS_ATTRIBUTE);
		String providerName = pluginElement.getAttribute(PROVIDER_ATTRIBUTE);

		PluginValidation.validatePluginAttributes(id, name, version);

		PluginDescriptor plugin = new PluginDescriptor(runtime, classLoader, id, name, version, className, providerName);

		readPrerequisites(pluginElement, plugin);
		readExtensionPoints(pluginElement, plugin);
		readExtensions(pluginElement, plugin);

		return plugin;
	}

	private void readExtensions(Element pluginElement, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		Element element = (Element) pluginElement.getElementsByTagName(EXTENSIONS_ELEMENT).item(0);
		if (element == null) {
			return;
		}

		NodeList extensions = element.getElementsByTagName(EXTENSION_ELEMENT);
		if (extensions.getLength() == 0) {
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ EXTENSION_ELEMENT
					+ "' child element in '"
					+ EXTENSIONS_ELEMENT
					+ "' element");
		}

		for (int i = 0; i < extensions.getLength(); ++i) {
			try {
				readAndAddExtension((Element) extensions.item(i), plugin);
			} catch (InvalidExtensionException e) {
				throw new InvalidPluginDescriptorException(e);
			}
		}
	}

	private void readAndAddExtension(Element extensionElement, PluginDescriptor plugin)
		throws InvalidExtensionException {

		String extensionPointId = extensionElement.getAttribute(EXTENDS_ATTRIBUTE);
		String name = extensionElement.getAttribute(NAME_ATTRIBUTE);
		String id = extensionElement.getAttribute(ID_ATTRIBUTE);

		PluginValidation.validateExtensionAttributes(extensionPointId);

		ConfigurationElement[] congifurationElements = readConfigurationElements(extensionElement);
		plugin.addExtension(extensionPointId, id, name, congifurationElements);
	}

	private ConfigurationElement[] readConfigurationElements(Element extensionElement) {
		return configElementReader.readConfigurationElements(extensionElement);
	}

	private void readPrerequisites(Element pluginElement, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		Element element = (Element) pluginElement.getElementsByTagName(PREREQUISITES_ELEMENT).item(0);
		if (element == null) {
			return;
		}

		NodeList prereqs = element.getElementsByTagName(PREREQUISITE_ELEMENT);
		if (prereqs.getLength() == 0) {
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ PREREQUISITE_ELEMENT
					+ "' child element in '"
					+ PREREQUISITES_ELEMENT
					+ "' element");
		}

		for (int i = 0; i < prereqs.getLength(); ++i) {
			try {
				readAndAddPrerequisite((Element) prereqs.item(i), plugin);
			} catch (InvalidPrerequisiteException e) {
				throw new InvalidPluginDescriptorException(e);
			}
		}
	}

	private void readAndAddPrerequisite(Element element, PluginDescriptor plugin)
		throws InvalidPrerequisiteException {
		String id = element.getAttribute(PLUGIN_ELEMENT);
		Version version = null;
		try {
			version = new Version(element.getAttribute(VERSION_ATTRIBUTE));
		} catch (VersionFormatException e) {
			throw new InvalidPrerequisiteException(e);
		}
		String match = element.getAttributeNode(MATCH_ATTRIBUTE) != null ? element.getAttribute(MATCH_ATTRIBUTE) : null;	
		
		PluginValidation.validatePrerequisiteAttributes(id, version, match);
		
		plugin.addPrerequisite(id, version, VersionMatchCondition.valueOf(match));
	}

	private void readExtensionPoints(Element pluginElement, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		Element element = (Element) pluginElement.getElementsByTagName(EXTENSION_POINTS_ELEMENT).item(0);
		if (element == null) {
			return;
		}

		NodeList extensionPoints = element.getElementsByTagName(EXTENSION_POINT_ELEMENT);
		if (extensionPoints.getLength() == 0) {
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ EXTENSION_POINT_ELEMENT
					+ "' child element in '"
					+ EXTENSION_POINTS_ELEMENT
					+ "' element");
		}

		for (int i = 0; i < extensionPoints.getLength(); ++i) {
			try {
				readAndAddExtensionPoint((Element) extensionPoints.item(i), plugin);
			} catch (InvalidExtensionPointException e) {
				throw new InvalidPluginDescriptorException(e);
			}
		}
	}

	private void readAndAddExtensionPoint(Element element, PluginDescriptor plugin)
		throws InvalidExtensionPointException {
		String id = element.getAttribute(ID_ATTRIBUTE);
		String name = element.getAttribute(NAME_ATTRIBUTE);
		String schemaReference = element.getAttribute(SCHEMA_ATTRIBUTE);
		
		PluginValidation.validateExtensionPointAttributes(id, name);
		
		plugin.addExtensionPoint(
			id,
			name,
			schemaReference);
	}

	private static class BuilderErrorHandler implements ErrorHandler {
		public void warning(SAXParseException exception) throws SAXException {
		}
		public void error(SAXParseException exception) throws SAXException {
			throw exception;
		}
		public void fatalError(SAXParseException exception) throws SAXException {
			throw exception;
		}
	}	
}
