package gov.va.med.hds.cd.runtime.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidExtensionException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;

public class PluginDescriptor implements IPluginDescriptor {

	private IRuntimeContext runtime = null;
	private ClassLoader classLoader = null;

	private String name;
	private String id;
	private Version version;
	private String className;
	private String providerName;

	private List prerequisites = new ArrayList();
	private Map extensionPoints = new HashMap();
	private Map extensionsByExtensionPoint = new HashMap();
	private Map extensionsByExtensionId = new HashMap();

	public PluginDescriptor(
		IRuntimeContext runtime,
		ClassLoader classLoader,
		String id,
		String name,
		Version version) {
		this(runtime, classLoader, id, name, version, null, null);
	}

	public PluginDescriptor(
		IRuntimeContext runtime,
		ClassLoader classLoader,
		String id,
		String name,
		Version version,
		String className) {
		this(runtime, classLoader, id, name, version, className, null);
	}

	public PluginDescriptor(
		IRuntimeContext runtime,
		ClassLoader classLoader,
		String id,
		String name,
		Version version,
		String className,
		String providerName) {
		this.runtime = runtime;
		this.classLoader = classLoader;
		this.name = name;
		this.id = id;
		this.version = version;
		this.className = "".equals(className) ? null : className;
		this.providerName = (providerName != null ? providerName : "");
	}

	public IRuntimeContext getRuntimeContext() {
		return runtime;
	}

	public ClassLoader getPluginClassLoader() {
		return classLoader;
	}

	public String getLabel() {
		return name;
	}

	public String getProviderName() {
		return providerName;
	}

	public String getUniqueIdentifier() {
		return id;
	}

	public Version getVersion() {
		return version;
	}

	public String getClassName() {
		return className;
	}

	public IPluginPrerequisite[] getPluginPrerequisites() {
		IPluginPrerequisite[] returnVal = new IPluginPrerequisite[prerequisites.size()];
		prerequisites.toArray(returnVal);
		return returnVal;
	}

	public IExtension getExtension(String extensionId) {
		return (IExtension) extensionsByExtensionId.get(extensionId);
	}

	public IExtension[] getExtensions() {
		List returnVal = new LinkedList();
		for (Iterator i = extensionsByExtensionPoint.values().iterator(); i.hasNext();) {
			List l = (List) i.next();
			returnVal.addAll(l);
		}
		return (IExtension[]) returnVal.toArray(new IExtension[returnVal.size()]);
	}

	public IExtensionPoint getExtensionPoint(String extensionPointId) {
		return (IExtensionPoint) extensionPoints.get(extensionPointId);
	}

	public IExtensionPoint[] getExtensionPoints() {
		IExtensionPoint[] returnVal = new IExtensionPoint[extensionPoints.size()];
		extensionPoints.values().toArray(returnVal);
		return returnVal;
	}

	public void addPrerequisite(String prerequisiteId, Version prerequisiteVersion) {
		prerequisites.add(new PluginPrerequisite(prerequisiteId, prerequisiteVersion));
	}

	public void addPrerequisite(
		String prerequisiteId,
		Version prerequisiteVersion,
		VersionMatchCondition matchCondition) {
		if (matchCondition == null) addPrerequisite(prerequisiteId, prerequisiteVersion);
		else prerequisites.add(new PluginPrerequisite(prerequisiteId, prerequisiteVersion, matchCondition));
	}

	public void addExtensionPoint(String id, String name) {
		extensionPoints.put(id, new ExtensionPoint(this, id, name));
	}

	public void addExtensionPoint(String id, String name, String schemaReference) {
		extensionPoints.put(id, new ExtensionPoint(this, id, name, schemaReference));
	}

	public void addExtension(String extensionPointId, ConfigurationElement[] configurationElements)
		throws InvalidExtensionException {
		addExtension(extensionPointId, null, null, configurationElements);
	}

	public void addExtension(
		String extensionPointId,
		String extensionId,
		String extensionName,
		ConfigurationElement[] configurationElements)
		throws InvalidExtensionException {
		if (extensionPointId == null)
			throw new InvalidExtensionException(
				new IllegalArgumentException("extensionPointId must not be null"));
		if (extensionPointId.indexOf('.') == -1)
			throw new InvalidExtensionException(
				"extension point id \"" + extensionPointId + "\" is invalid");
		if (!hasPrerequisite(extensionPointId))
			throw new InvalidExtensionException(
				"plugin \""
					+ id
					+ "\" declared extension to extension point id \""
					+ extensionPointId
					+ "\" without declaring a prerequisite for the required plugin");
		Extension extension = new Extension(this, extensionPointId, extensionId, extensionName);
		for (int i = 0; i < configurationElements.length; ++i) {
			extension.addConfigurationElement(configurationElements[i]);
		}
		List l = (List) extensionsByExtensionPoint.get(extensionPointId);
		if (l == null) {
			l = new LinkedList();
			extensionsByExtensionPoint.put(extensionPointId, l);
		}
		l.add(extension);

		if (extension.getSimpleIdentifier() != null
			&& !extension.getSimpleIdentifier().equals("")) {
			extensionsByExtensionId.put(extension.getSimpleIdentifier(), extension);
		}
	}

	private boolean hasPrerequisite(String extensionPointId) {
		String pluginId = extensionPointId.substring(0, extensionPointId.lastIndexOf('.'));
		if (pluginId.equals(getUniqueIdentifier()))
			return true;

		for (Iterator i = prerequisites.iterator(); i.hasNext();) {
			IPluginPrerequisite prereq = (IPluginPrerequisite) i.next();
			if (prereq.getUniqueIdentifier().equals(pluginId))
				return true;
		}
		return false;
	}
	
	public String toString() {
		return getUniqueIdentifier() + " v" + getVersion() + " (" + getLabel() +")";
	}
}
