package gov.va.med.hds.cd.runtime.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.VersionComparator;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;

/**
 * The default implementation of <code>IPluginRegistry</code>.
 *
 * @see gov.va.med.hds.cd.runtime.IPluginRegistry
 * @see gov.va.med.hds.cd.runtime.RuntimeDescriptor
 */
public class PluginRegistry implements IPluginRegistry {

	private static final VersionComparator VERSION_COMPARATOR =
		new VersionComparator();

	private Map extensionPointLinks = new HashMap(30);
	private List pendingPlugins = new LinkedList();

	private Map pluginMap = new HashMap(30);
	private IRuntimeContext runtime = null;

	/**
	 * Creates a new plugin registry which contains only 
	 * the descriptor of the runtime plugin.
	 * 
	 * @see RuntimeDescriptor
	 */
	public PluginRegistry(IRuntimeContext runtime) {
		setRuntimeContext(runtime);
		addPlugin(new RuntimeDescriptor(runtime, getClass().getClassLoader()));
		addPlugin(new RuntimeCoreDescriptor(runtime, getClass().getClassLoader()));
	}

	/**
	 * Adds the specified plugin to this registry.  An existing plugin
	 * with the same unique id and version is replaced by the new
	 * value.  Plugins whose dependencies are unmet at the time of the
	 * add are placed in a pending or unresolved state.  As more plugins
	 * are added, if an unresolved plugin's dependencies are met, it is
	 * moved out of the unresolved state into the registry itself.
	 *
	 * @param plugin the plugin descriptor to add
	 */
	public void add(IPluginDescriptor plugin) throws InvalidPluginDescriptorException {
		validatePluginDescriptor(plugin);

		if (!(containsAllPrerequisites(plugin) && containsAllRequiredExtensionPoints(plugin))) {
			pendingPlugins.add(plugin);
			return;
		}

		addPlugin(plugin);

		resolvePendingPlugins();
	}

	/**
	 * Adds all of plugin descriptors in the specified registry to this registry.
	 * 
	 * @param registry the registry to merge with this one.
	 * 
	 * @see PluginRegistry#add(IPluginDescriptor)
	 */
	public void add(IPluginRegistry registry) throws InvalidPluginDescriptorException {
		IPluginDescriptor[] plugins = registry.getPluginDescriptors();
		for (int i = 0; i < plugins.length; ++i) {
			add(plugins[i]);
		}
	}

	private void addPlugin(IPluginDescriptor plugin) {
		SortedMap pluginsByVersion = getPluginsByVersion(plugin.getUniqueIdentifier());
		IPluginDescriptor duplicate = findDuplicate(pluginsByVersion.values(), plugin);
		if (duplicate != null) {
			pluginsByVersion.remove(duplicate.getVersion());
		}

		resolveExtensionPointLinks(plugin);

		pluginsByVersion.put(plugin.getVersion(), plugin);
	}

	private boolean containsAllPrerequisites(IPluginDescriptor plugin) {
		IPluginPrerequisite[] prereqs = plugin.getPluginPrerequisites();
		for (int i = 0; i < prereqs.length; ++i) {
			if (!containsPrerequisite(prereqs[i]))
				return false;
		}
		return true;
	}

	private boolean containsAllRequiredExtensionPoints(IPluginDescriptor plugin) {
		IExtension[] extensions = plugin.getExtensions();
		for (int i = 0; i < extensions.length; i++) {
			if (isExtensionToSelf(extensions[i].getExtensionPointUniqueIdentifier(), plugin))
				continue;
			if (!containsExtensionPoint(extensions[i].getExtensionPointUniqueIdentifier())) {
				return false;
			}
		}
		return true;
	}

	private boolean containsExtensionPoint(String extensionPointUniqueId) {
		return getExtensionPoint(extensionPointUniqueId) != null;
	}

	private boolean containsPrerequisite(IPluginPrerequisite prereq) {
		IPluginDescriptor[] plugins = getPluginDescriptors(prereq.getUniqueIdentifier());
		for (int i = 0; i < plugins.length; i++) {
			if (prereq.isMatchedAsGreaterOrEqual()
				&& plugins[i].getVersion().isGreaterOrEqualTo(prereq.getVersion()))
				return true;
			else if (
				prereq.isMatchedAsCompatible()
					&& plugins[i].getVersion().isCompatibleWith(prereq.getVersion()))
				return true;
			else if (
				prereq.isMatchedAsEquivalent()
					&& plugins[i].getVersion().isEquivalentTo(prereq.getVersion()))
				return true;
			else if (
				prereq.isMatchedAsPerfect()
					&& plugins[i].getVersion().isPerfect(prereq.getVersion()))
				return true;
		}
		return false;
	}

	private IPluginDescriptor findDuplicate(Collection plugins, IPluginDescriptor plugin) {
		for (Iterator i = plugins.iterator(); i.hasNext();) {
			IPluginDescriptor d = (IPluginDescriptor) i.next();
			if (d.getVersion().equals(plugin.getVersion()))
				return d;
		}
		return null;
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getConfigurationElementsFor(String)
	 */
	public IConfigurationElement[] getConfigurationElementsFor(String extensionPointId) {
		List configurationElements = new LinkedList();
		List extensions = (List) extensionPointLinks.get(extensionPointId);
		if (extensions == null)
			return new IConfigurationElement[0];

		for (Iterator i = extensions.iterator(); i.hasNext();) {
			IExtension extension = (IExtension) i.next();
			IConfigurationElement[] cfgElmnts = extension.getConfigurationElements();
			for (int j = 0; j < cfgElmnts.length; ++j) {
				configurationElements.add(cfgElmnts[j]);
			}
		}
		return (IConfigurationElement[]) configurationElements.toArray(
			new IConfigurationElement[configurationElements.size()]);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getConfigurationElementsFor(String, String)
	 */
	public IConfigurationElement[] getConfigurationElementsFor(
		String pluginId,
		String extensionPointName) {
		return getConfigurationElementsFor(pluginId + "." + extensionPointName);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getConfigurationElementsFor(String, String, String)
	 */
	public IConfigurationElement[] getConfigurationElementsFor(
		String pluginId,
		String extensionPointName,
		String extensionId) {
		IExtension extension = getExtension(pluginId, extensionPointName, extensionId);
		if (extension == null)
			return new IConfigurationElement[0];
		return extension.getConfigurationElements();
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtension(String, String)
	 */
	public IExtension getExtension(String extensionPointId, String extensionId) {
		if (!containsExtensionPoint(extensionPointId))
			return null;
		List extensions = (List) extensionPointLinks.get(extensionPointId);
		if (extensions == null)
			return null;

		for (Iterator i = extensions.iterator(); i.hasNext();) {
			IExtension extension = (IExtension) i.next();
			if (extension.getSimpleIdentifier().equals(extensionId))
				return extension;
		}

		return null;
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtension(String, String, String)
	 */
	public IExtension getExtension(String pluginId, String extensionPointId, String extensionId) {
		return getExtension(pluginId + "." + extensionPointId, extensionId);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtensionPoint(String)
	 */
	public IExtensionPoint getExtensionPoint(String extensionPointUniqueId) {
		int lastdot = extensionPointUniqueId.lastIndexOf('.');
		if (lastdot == -1)
			return null;
		return getExtensionPoint(
			extensionPointUniqueId.substring(0, lastdot),
			extensionPointUniqueId.substring(lastdot + 1));
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtensionPoint(String, String)
	 */
	public IExtensionPoint getExtensionPoint(String pluginId, String extensionPointId) {
		IPluginDescriptor descriptor = getPluginDescriptor(pluginId);
		if (descriptor == null)
			return null;
		return descriptor.getExtensionPoint(extensionPointId);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtensionPoints()
	 */
	public IExtensionPoint[] getExtensionPoints() {
		List result = new ArrayList();
		for (Iterator i = pluginMap.values().iterator(); i.hasNext();) {
			Collection plugins = ((Map) i.next()).values();
			for (Iterator j = plugins.iterator(); j.hasNext();) {
				IExtensionPoint[] pointList = ((IPluginDescriptor) j.next()).getExtensionPoints();
				if (pointList != null) {
					for (int k = 0; k < pointList.length; k++)
						result.add(pointList[k]);
				}
			}
		}
		return (IExtensionPoint[]) result.toArray(new IExtensionPoint[result.size()]);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtensions(String)
	 */
	public IExtension[] getExtensions(String extensionPointUniqueId) {
		List extensions = (List) extensionPointLinks.get(extensionPointUniqueId);
		if (extensions == null)
			return new IExtension[0];
		return (IExtension[]) extensions.toArray(new IExtension[extensions.size()]);
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getExtensions(String, String)
	 */
	public IExtension[] getExtensions(String pluginId, String extensionPointId) {
		return getExtensions(pluginId + "." + extensionPointId);
	}

	/**
	 * Obtain a list of missing prerequisites for a given plugin.
	 * 
	 * @param the plugin to obtain missing prerequisites for.
	 * @return an array of plugin prerequisites.
	 */
	public IPluginPrerequisite[] getMissingPrerequisites(IPluginDescriptor plugin) {
		List missingPrereqs = new LinkedList();
		IPluginPrerequisite[] prereqs = plugin.getPluginPrerequisites();
		for (int i = 0; i < prereqs.length; i++) {
			if (!containsPrerequisite(prereqs[i]))
				missingPrereqs.add(prereqs[i]);
		}
		return (IPluginPrerequisite[]) missingPrereqs.toArray(
			new IPluginPrerequisite[missingPrereqs.size()]);
	}

	/*
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getPluginDescriptor(String)
	 */
	public IPluginDescriptor getPluginDescriptor(String pluginId) {
		SortedMap pluginsByVersion = getPluginsByVersion(pluginId);
		if (pluginsByVersion == null) return null;
		if (pluginsByVersion.isEmpty()) return null;
		return (IPluginDescriptor) pluginsByVersion.get(pluginsByVersion.lastKey());
	}

	/*
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getPluginDescriptor(String, Version)
	 */
	public IPluginDescriptor getPluginDescriptor(String pluginId, Version version) {
		SortedMap pluginsByVersion = getPluginsByVersion(pluginId);
		return (IPluginDescriptor) pluginsByVersion.get(version);
	}

	/*
	 * @see gov.va.med.hds.cd.runtime.IPluginRegistry#getPluginDescriptor(String, Version, VersionMatchCondition)
	 */
	public IPluginDescriptor getPluginDescriptor(
		String pluginId,
		Version version,
		VersionMatchCondition matchCondition) {
		SortedMap plugins = getPluginsByVersion(pluginId);
		if (plugins.isEmpty()) return null;
		if (matchCondition.equals(VersionMatchCondition.GREATER_OR_EQUAL)) return (IPluginDescriptor) plugins.get(plugins.lastKey());
		if (matchCondition.equals(VersionMatchCondition.PERFECT)) return (IPluginDescriptor) plugins.get(version);
		SortedMap matchingVersions = getPluginsByMatchingVersion(plugins, version, matchCondition);
		return (IPluginDescriptor)  matchingVersions.get(matchingVersions.lastKey());
	}

	private SortedMap getPluginsByMatchingVersion(SortedMap plugins, Version v, VersionMatchCondition matchCondition) {
		if (matchCondition.equals(VersionMatchCondition.COMPATIBLE)) {
			Version highestCompatibleVersionExclusive = new Version(v.getMajorComponent() + 1, v.getMinorComponent(), v.getServiceComponent(), v.getQualifierComponent());
			return plugins.subMap(v, highestCompatibleVersionExclusive);
		}
		if (matchCondition.equals(VersionMatchCondition.EQUIVALENT)) {
			Version highestEquivalentVersionExclusive = new Version(v.getMajorComponent(), v.getMinorComponent() + 1, v.getServiceComponent(), v.getQualifierComponent());
			return plugins.subMap(v, highestEquivalentVersionExclusive);
		}
		if (matchCondition.equals(VersionMatchCondition.PERFECT)) {
			Version highestPerfectVersionExclusive = new Version(v.getMajorComponent(), v.getMinorComponent(), v.getServiceComponent() + 1, v.getQualifierComponent());
			return plugins.subMap(v, highestPerfectVersionExclusive);
		}
		if (matchCondition.equals(VersionMatchCondition.GREATER_OR_EQUAL)) {
			return plugins.tailMap(v);
		}
		return null;
	}

	/**
	 * Returns the list of plugins managed by this registry.
	 *
	 * @return the plugins in this registry
	 */
	public IPluginDescriptor[] getPluginDescriptors() {
		List result = new LinkedList();
		for (Iterator i = pluginMap.values().iterator(); i.hasNext();) {
			result.addAll(((Map) i.next()).values());
		}
		return (IPluginDescriptor[]) result.toArray(new IPluginDescriptor[result.size()]);
	}

	/**
	 * Returns all versions of the identified plugin descriptor
	 * known to this plugin registry.
	 * Returns an empty array if there are no plugins
	 * with the specified identifier.
	 *
	 * @param id the unique identifier of the plugin (e.g. <code>"gov.med.hds.cd.ui.Prism"</code>).
	 * @return the plugin descriptors known to this plugin registry
	 */
	public IPluginDescriptor[] getPluginDescriptors(String id) {
		SortedMap plugins = (SortedMap) pluginMap.get(id);
		return plugins == null
			? new IPluginDescriptor[0]
			: (IPluginDescriptor[]) plugins.values().toArray(new IPluginDescriptor[plugins.size()]);
	}

	private SortedMap getPluginsByVersion(String id) {
		SortedMap plugins = (SortedMap) pluginMap.get(id);
		if (plugins == null) {
			plugins = new TreeMap(VERSION_COMPARATOR);
			pluginMap.put(id, plugins);
		}
		return plugins;
	}

	public IExtension[] getUnresolvedExtensions(IPluginDescriptor plugin) {
		List extensionsToMissingExtensionPoints = new LinkedList();
		IExtension[] extensions = plugin.getExtensions();
		for (int i = 0; i < extensions.length; i++) {
			if (isExtensionToSelf(extensions[i].getExtensionPointUniqueIdentifier(), plugin))
				continue;
			if (!containsExtensionPoint(extensions[i].getExtensionPointUniqueIdentifier())) {
				extensionsToMissingExtensionPoints.add(extensions[i]);
			}
		}
		return (IExtension[]) extensionsToMissingExtensionPoints.toArray(
			new IExtension[extensionsToMissingExtensionPoints.size()]);
	}

	/**
	 * Obtain a list of unresolved plugins.
	 */
	public IPluginDescriptor[] getUnresolvedPluginDescriptors() {
		resolvePendingPlugins();
		return (IPluginDescriptor[]) pendingPlugins.toArray(
			new IPluginDescriptor[pendingPlugins.size()]);
	}

	private boolean isExtensionToSelf(String extensionPointUniqueId, IPluginDescriptor plugin) {
		String extensionPointId =
			extensionPointUniqueId.substring(extensionPointUniqueId.lastIndexOf('.') + 1);
		return plugin.getExtensionPoint(extensionPointId) != null;
	}

	/**
	 * Returns whether or not this registry is resolved.  The registry
	 * is resolved when all plugin descriptors' dependencies have been met.
	 * 
	 * @return <code>true</code> when there are no unresolved plugins
	 */
	public boolean isResolved() {
		resolvePendingPlugins();
		return pendingPlugins.isEmpty();
	}

	/**
	 * Removes the plugin with id and version if it exists in this registry.
	 * This method has no effect if a plugin with the given id and version 
	 * cannot be found.
	 *
	 * @param pluginId the unique identifier of the plugin to remove
	 * @param version the version of the plugin to remove
	 */
	public void removePlugin(String pluginId, String version) {
		// TODO_SB: NOT IMPLEMENTED
	}
	/**
	 * Removes all versions of the given plugin from this registry.
	 * This method has no effect if such a plugin cannot be found.
	 *
	 * @param pluginId the unique identifier of the plugins to remove
	 */
	public void removePlugins(String pluginId) {
		pluginMap.remove(pluginId);
		// TODO_SB: handle removal correctly (update links and dependencies)
	}

	private void resolveExtensionPointLinks(IPluginDescriptor plugin) {
		IExtension[] extensions = plugin.getExtensions();
		for (int i = 0; i < extensions.length; ++i) {
			List extensionLinks =
				(List) extensionPointLinks.get(extensions[i].getExtensionPointUniqueIdentifier());
			if (extensionLinks == null) {
				extensionLinks = new LinkedList();
				extensionPointLinks.put(
					extensions[i].getExtensionPointUniqueIdentifier(),
					extensionLinks);
			}
			extensionLinks.add(extensions[i]);
		}
	}

	private void resolvePendingPlugins() {
		List resolved = new LinkedList();
		for (Iterator i = pendingPlugins.iterator(); i.hasNext();) {
			IPluginDescriptor plugin = (IPluginDescriptor) i.next();
			if (containsAllPrerequisites(plugin) && containsAllRequiredExtensionPoints(plugin)) {
				resolved.add(plugin);
			}
		}
		pendingPlugins.removeAll(resolved);
		for (Iterator i = resolved.iterator(); i.hasNext();) {
			IPluginDescriptor plugin = (IPluginDescriptor) i.next();
			try {
				add(plugin);
			} catch (InvalidPluginDescriptorException e) {
			}
		}
	}
	/**
	 * Sets the runtime context to which this plugin registry belongs.
	 * @param runtimeContext
	 */
	private void setRuntimeContext(IRuntimeContext runtimeContext) {
		this.runtime = runtimeContext;
		// TODO_SB: should this reset the runtime context of any already-added
		// plugin descriptors?
	}

	private void validatePluginDescriptor(IPluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		if (RuntimeUtilities.isRuntimePlugin(plugin))
			throw new InvalidPluginDescriptorException(
				"plugin '" + plugin.getUniqueIdentifier() + "' is reserved");
		else if (plugin.getUniqueIdentifier() == null || plugin.getUniqueIdentifier().equals(""))
			throw new InvalidPluginDescriptorException("expected unique identifier in plugin descriptor");
		else if (plugin.getLabel() == null || plugin.getLabel().equals(""))
			throw new InvalidPluginDescriptorException("expected displayable name in plugin descriptor");
		else if (plugin.getVersion() == null)
			throw new InvalidPluginDescriptorException("expected valid version identifier in plugin descriptor");
	}
}
