package gov.va.med.hds.cd.runtime.internal;

import jargs.gnu.CmdLineParser;
import jargs.gnu.CmdLineParser.IllegalOptionValueException;
import jargs.gnu.CmdLineParser.UnknownOptionException;

import gov.va.med.hds.cd.mum.IApplication;
import gov.va.med.hds.cd.mum.IPluginSource;
import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.boot.ExceptionDialog;
import gov.va.med.hds.cd.mum.boot.ISplashScreen;
import gov.va.med.hds.cd.mum.boot.IStatusNotifier;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IMainPlugin;
import gov.va.med.hds.cd.runtime.IPluginActivationListener;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntime;
import gov.va.med.hds.cd.runtime.IRuntimeConstants;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidMainPluginException;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;
import gov.va.med.hds.cd.runtime.logging.Level;
import gov.va.med.hds.cd.runtime.logging.internal.Log4JLogging;
import gov.va.med.hds.cd.runtime.logging.log4j.internal.ConfigurationElementLogConfigurator;
import gov.va.med.hds.cd.runtime.logging.log4j.internal.LogOutput;

public class Runtime implements IRuntime, IRuntimeContext, IRuntimeConstants, IApplication {

	private static class RuntimeCmdLineParser extends CmdLineParser {

		public static final Option MAIN_PLUGIN = new CmdLineParser.Option.StringOption('m', "main");

		public RuntimeCmdLineParser() {
			super();
			addOption(MAIN_PLUGIN);
		}

		private String getMainPluginId() {
			return (String) getOptionValue(MAIN_PLUGIN);
		}
	}

	private static final String LAUNCHING_MAIN_PLUGIN = "Starting Main Plugin";
	private static final String READING_REGISTRY = "Reading Plugin Registry";
	private static final String READING_SETTINGS = "Reading Desktop Settings";

	public static final String DEFAULT_MAIN_PLUGIN = "gov.va.med.hds.cd.session";

	private static final Version VERSION_ZERO = new Version(0, 0, 0);

	private ActivePluginManager activePluginManager;

	private RuntimeCmdLineParser cmdLineParser = new RuntimeCmdLineParser();
	private String mainPluginId = DEFAULT_MAIN_PLUGIN;

	protected PluginRegistry registry;
	private ConfigurationElement[] settings = new ConfigurationElement[0];

	private ILoggingCapable logging = new Log4JLogging();
	private ILogger log;

	public Runtime() {
		this.registry = new PluginRegistry(this);
		this.activePluginManager = new ActivePluginManager(this);
		this.log = getLogger(getClass());
	}

	protected void cleanUp() {
		if (activePluginManager != null)
			activePluginManager.shutdown();
		activePluginManager = null;
		registry = null;
	}

	public String getMainPluginId() {
		return mainPluginId;
	}

	public synchronized Object getPluginInstance(String pluginId)
		throws PluginStartupException, PluginNotFoundException {
		if (PLUGIN_ID.equals(pluginId)) return this;
		return activePluginManager.getPluginInstance(pluginId, VERSION_ZERO, VersionMatchCondition.GREATER_OR_EQUAL);
	}

	public synchronized Object getPluginInstance(String pluginId, Version version)
		throws PluginStartupException, PluginNotFoundException {
		if (PLUGIN_ID.equals(pluginId)) return this;
		return activePluginManager.getPluginInstance(pluginId, version, VersionMatchCondition.PERFECT);
	}

	public IPluginRegistry getPluginRegistry() {
		return registry;
	}

	public IConfigurationElement[] getSettings() {
		return settings;
	}

	public ILogger getLogger(Class c) {
		return logging.getLogger(c);
	}

	public ILogger getLogger(String name) {
		return logging.getLogger(name);
	}

	public ILogger getRootLogger() {
		return logging.getRootLogger();
	}

	public boolean hasValidMainPlugin() {
		try {
			validateMainPluginId();
			return true;
		} catch (InvalidMainPluginException e) {
			return false;
		}
	}

	public boolean isPluginActive(String pluginId) {
		return activePluginManager.isPluginActive(pluginId, VERSION_ZERO, VersionMatchCondition.GREATER_OR_EQUAL);
	}
	
	public boolean isPluginActive(String pluginId, Version version) {
		return activePluginManager.isPluginActive(pluginId, version, VersionMatchCondition.PERFECT);
	}

	public void addPluginActivationListener(IPluginActivationListener l) {
		activePluginManager.addPluginActivationListener(l);
	}

	public void removePluginActivationListener(IPluginActivationListener l) {
		activePluginManager.removePluginActivationListener(l);
	}

	protected void launchMainPlugin(ISplashScreen splash) {
		try {
			IMainPlugin main = null;
			try {
				main = (IMainPlugin) getPluginInstance(getMainPluginId());
				validateMainPluginId();
			} catch (ClassCastException e) {
				showErrorDialog(
					"The specified main plugin '"
						+ getMainPluginId()
						+ "' does not implement '"
						+ IMainPlugin.class.getName()
						+ "'.", e);
				shutdown();
			}
			if (main != null)
				main.run(this, splash, cmdLineParser.getRemainingArgs());
		} catch (PluginNotFoundException e) {
			logUnresolvedPlugins(Level.FATAL);
			showErrorDialog(
				"The specified main plugin '"
					+ getMainPluginId()
					+ "' was not found.  It is likely that it has unresolved dependencies.", e);
			shutdown();
		} catch (InvalidMainPluginException e) {
			String message = "The specified main plugin '"
			+ getMainPluginId()
			+ "' does not extend the extension point '"
			+ UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID
			+ "'";
			log.fatal(message, e);
			showErrorDialog(message, e);
			shutdown();
		} catch (PluginStartupException e) {
			String message = "A fatal error occurred starting up main plugin '"
			+ getMainPluginId()
			+ "'";
			log.fatal(message, e);
			showErrorDialog(message, e);
			shutdown();
		}
	}

	private void parseArguments(String[] args) {
		try {
			cmdLineParser.parse(args);
		} catch (IllegalOptionValueException e) {
			//TODO_SB: not implemented
		} catch (UnknownOptionException e) {
			//TODO_SB: not implemented
		}
		if (cmdLineParser.getMainPluginId() != null && !cmdLineParser.getMainPluginId().equals(""))
			mainPluginId = cmdLineParser.getMainPluginId();
	}

	private void parseSettings() {
		IConfigurationElement[] settings = getSettings();
		IConfigurationElement[] runtimeSettings = null;
		for (int i = 0; i < settings.length; i++) {
			if ("runtime".equals(settings[i].getName())) {
				runtimeSettings = settings[i].getChildren();
			} else if ("logging".equals(settings[i].getName())) {
				ConfigurationElementLogConfigurator.configure(settings[i]);
			}
		}
		if (runtimeSettings == null) return;
		for (int i = 0; i < runtimeSettings.length; i++) {
			if (MAIN_PLUGIN_EXTENSION_POINT_ID.equals(runtimeSettings[i].getName())) {
				String mainPluginId = settings[i].getValue();
				if (mainPluginId != null && !"".equals(mainPluginId)) {
					setMainPluginId(mainPluginId);
				}
				break;
			}
		}
	}

	protected void readPluginDescriptors(IPluginSource[] plugins) {
		if (plugins == null || plugins.length == 0)
			return;

		TreeDataPluginReader reader = new TreeDataPluginReader(this);
		for (int i = 0; i < plugins.length; i++) {
			try {
				PluginDescriptor plugin =
					reader.readPlugin(plugins[i].getManifest(), plugins[i].getClassLoader());
				if (!RuntimeUtilities.isRuntimePlugin(plugin))
					registry.add(plugin);
			} catch (InvalidPluginDescriptorException e) {
				log.error("error reading plugin descriptor", e);
			}
		}

		registerURLHandlers();
	}

	protected void registerURLHandlers() {
		RuntimeURLHandlerFactory.registerAll(this, log);
	}

	protected void readSettings(ITreeData settingsElement) {
		if (settingsElement == null)
			return;
		TreeDataConfigurationElementReader reader = new TreeDataConfigurationElementReader(this);
		this.settings = reader.readConfigurationElements(settingsElement);
	}

	public void run() {
		run(new String[0]);
	}

	public void run(String[] args) {
		run(args, null, null, null);
	}

	public void run(
		String[] args,
		ITreeData settings,
		IPluginSource[] plugins,
		ISplashScreen splash) {
		LogOutput.install();
		updateSplash(splash, READING_SETTINGS);
		readSettings(args, settings);
		updateSplash(splash, READING_REGISTRY);
		readRegistry(plugins);
		updateSplash(splash, LAUNCHING_MAIN_PLUGIN);
		launchMainPlugin(splash);
		LogOutput.uninstall();
	}

	private void readSettings(String[] args, ITreeData settings) {
		readSettings(settings);
		parseSettings();
		parseArguments(args);
	}

	private void readRegistry(IPluginSource[] plugins) {
		readPluginDescriptors(plugins);
		logUnresolvedPlugins(Level.ERROR);
	}

	public void setMainPluginId(String mainPluginId) {
		this.mainPluginId = mainPluginId;
	}

	public synchronized void shutdown() {
		cleanUp();
		System.exit(0);
	}

	protected void validateMainPluginId() throws InvalidMainPluginException {
		IExtension[] extensions =
			getPluginRegistry().getExtensions(UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID);
		for (int i = 0; i < extensions.length; i++) {
			if (extensions[i]
				.getDeclaringPluginDescriptor()
				.getUniqueIdentifier()
				.equals(getMainPluginId()))
				return;
		}
		throw new InvalidMainPluginException(getMainPluginId());
	}

	protected void logUnresolvedPlugins(Level logLevel) {
		if (registry.isResolved())
			return;

		StringBuffer message = new StringBuffer();
		IPluginDescriptor[] unresolvedPlugins = registry.getUnresolvedPluginDescriptors();
		for (int i = 0; i < unresolvedPlugins.length; i++) {
			IPluginPrerequisite[] prereqs = registry.getMissingPrerequisites(unresolvedPlugins[i]);
			for (int j = 0; j < prereqs.length; j++) {
				logUnresolvedPluginWithUnmetDependency(logLevel, unresolvedPlugins[i], prereqs[j]);
			}
			IExtension[] unresolvedExtensions =
				registry.getUnresolvedExtensions(unresolvedPlugins[i]);
			for (int j = 0; j < unresolvedExtensions.length; j++) {
				logUnresolvedPluginExtendingMissingExtensionPoint(logLevel, unresolvedPlugins[i], unresolvedExtensions[j]);
			}
		}
	}

	private void logUnresolvedPluginWithUnmetDependency(Level logLevel, IPluginDescriptor plugin, IPluginPrerequisite prereq) {
		StringBuffer message = new StringBuffer();
		message.append("plugin '");
		message.append(plugin.getUniqueIdentifier());
		message.append(" v" + plugin.getVersion());
		message.append("' has unmet dependency on '");
		message.append(prereq.getUniqueIdentifier());
		message.append(" v" + prereq.getVersion());
		message.append("'");
		log.log(logLevel, message.toString());
	}

	private void logUnresolvedPluginExtendingMissingExtensionPoint(Level logLevel, IPluginDescriptor plugin, IExtension extension) {
		StringBuffer message = new StringBuffer();
		message.append("plugin '");
		message.append(plugin.getUniqueIdentifier());
		message.append(" v" + plugin.getVersion());
		message.append("' extends missing extension point '");
		message.append(extension.getExtensionPointUniqueIdentifier());
		message.append("'");
		log.log(logLevel, message.toString());
	}

	private void updateSplash(ISplashScreen splash, String message) {
		getLogger(getClass()).info(message);
		if (!(splash instanceof IStatusNotifier))
			return;
		IStatusNotifier notifier = (IStatusNotifier) splash;
		notifier.displayMessage(message);
	}

	private void showErrorDialog(String message, Exception e) {
		ExceptionDialog.showExceptionDialog("HealtheVet Desktop Runtime Error", message, e);
	}
}
