package gov.va.med.hds.cd.runtime.internal;

import java.io.InputStream;
import java.io.Reader;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.mum.VersionFormatException;
import gov.va.med.hds.cd.mum.boot.utils.ILoadableTreeData;
import gov.va.med.hds.cd.mum.boot.utils.ITreeData;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataException;
import gov.va.med.hds.cd.mum.boot.utils.TreeDataFactory;
import gov.va.med.hds.cd.runtime.IRuntimeConstants;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidExtensionException;
import gov.va.med.hds.cd.runtime.InvalidExtensionPointException;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.InvalidPrerequisiteException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;

/**
 * enclosing_type
 */
public class TreeDataPluginReader implements IPluginReader, IRuntimeConstants {

	/**
	 * Constructor for TreeDataPluginReader.
	 */
	public TreeDataPluginReader() {
		super();
	}

	private IRuntimeContext runtime;
	private TreeDataConfigurationElementReader configElementReader;

	public TreeDataPluginReader(IRuntimeContext runtime) {
		configElementReader = new TreeDataConfigurationElementReader(runtime);
		this.runtime = runtime;
	}

	public PluginDescriptor readPlugin(ITreeData xml, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		int root = xml.getRootID();
		String name = xml.getAttribute(root, NAME_ATTRIBUTE);
		String id = xml.getAttribute(root, ID_ATTRIBUTE);
		Version version = null;
		try {
			version = new Version(xml.getAttribute(root, VERSION_ATTRIBUTE));
		} catch (VersionFormatException e) {
			throw new InvalidPluginDescriptorException(e);
		}
		String className = xml.getAttribute(root, CLASS_ATTRIBUTE);
		String providerName = xml.getAttribute(root, PROVIDER_ATTRIBUTE);

		PluginValidation.validatePluginAttributes(id, name, version);

		PluginDescriptor plugin =
			new PluginDescriptor(runtime, classLoader, id, name, version, className, providerName);

		readPrerequisites(xml, root, plugin);
		readExtensionPoints(xml, root, plugin);
		readExtensions(xml, root, plugin);

		return plugin;
	}

	private void readExtensions(ITreeData xml, int root, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		int element = xml.findChildID(root, EXTENSIONS_ELEMENT);
		if (element < 0)
			return;

		boolean noExtensions = true;
		for (int i = 0; i < xml.getChildCount(element); i++) {
			int child = xml.getChildID(element, i);
			if (EXTENSION_ELEMENT.equals(xml.getName(child))) {
				noExtensions = false;
				try {
					readAndAddExtension(xml, child, plugin);
				} catch (InvalidExtensionException e) {
					throw new InvalidPluginDescriptorException(e);
				}
			}
		}
		if (noExtensions)
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ EXTENSION_ELEMENT
					+ "' child element in '"
					+ EXTENSIONS_ELEMENT
					+ "' element");
	}

	private void readAndAddExtension(ITreeData xml, int extensionElement, PluginDescriptor plugin)
		throws InvalidExtensionException {

		String extensionPointId = xml.getAttribute(extensionElement, EXTENDS_ATTRIBUTE);
		String name = xml.getAttribute(extensionElement, NAME_ATTRIBUTE);
		String id = xml.getAttribute(extensionElement, ID_ATTRIBUTE);

		PluginValidation.validateExtensionAttributes(extensionPointId);

		ConfigurationElement[] congifurationElements =
			readConfigurationElements(xml, extensionElement);
		plugin.addExtension(extensionPointId, id, name, congifurationElements);
	}

	private ConfigurationElement[] readConfigurationElements(ITreeData xml, int extensionElement) {
		return configElementReader.readConfigurationElements(xml, extensionElement);
	}

	private void readPrerequisites(ITreeData xml, int pluginElement, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		int element = xml.findChildID(pluginElement, PREREQUISITES_ELEMENT);
		if (element < 0) {
			return;
		}

		boolean noPrereqs = true;
		for (int i = 0; i < xml.getChildCount(element); i++) {
			int child = xml.getChildID(element, i);
			if (PREREQUISITE_ELEMENT.equals(xml.getName(child))) {
				noPrereqs = false;
				try {
					readAndAddPrerequisite(xml, child, plugin);
				} catch (InvalidPrerequisiteException e) {
					throw new InvalidPluginDescriptorException(e);
				}
			}
		}
		if (noPrereqs)
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ PREREQUISITE_ELEMENT
					+ "' child element in '"
					+ PREREQUISITES_ELEMENT
					+ "' element");
	}

	private void readAndAddPrerequisite(ITreeData xml, int element, PluginDescriptor plugin)
		throws InvalidPrerequisiteException {
		String id = xml.getAttribute(element, PLUGIN_ELEMENT);
		Version version = null;
		try {
			version = new Version(xml.getAttribute(element, VERSION_ATTRIBUTE));
		} catch (VersionFormatException e) {
			throw new InvalidPrerequisiteException(e);
		}
		String match = xml.getAttribute(element, MATCH_ATTRIBUTE);
		PluginValidation.validatePrerequisiteAttributes(id, version, match);
		plugin.addPrerequisite(id, version, VersionMatchCondition.valueOf(match));
	}

	private void readExtensionPoints(ITreeData xml, int pluginElement, PluginDescriptor plugin)
		throws InvalidPluginDescriptorException {
		int element = xml.findChildID(pluginElement, EXTENSION_POINTS_ELEMENT);
		if (element < 0) {
			return;
		}

		boolean noExtensions = true;
		for (int i = 0; i < xml.getChildCount(element); i++) {
			int child = xml.getChildID(element, i);
			if (EXTENSION_POINT_ELEMENT.equals(xml.getName(child))) {
				noExtensions = false;
				try {
					readAndAddExtensionPoint(xml, child, plugin);
				} catch (InvalidExtensionPointException e) {
					throw new InvalidPluginDescriptorException(e);
				}
			}
		}
		if (noExtensions)
			throw new InvalidPluginDescriptorException(
				"expected '"
					+ EXTENSION_POINT_ELEMENT
					+ "' child element in '"
					+ EXTENSION_POINTS_ELEMENT
					+ "' element");
	}

	private void readAndAddExtensionPoint(ITreeData xml, int element, PluginDescriptor plugin)
		throws InvalidExtensionPointException {
		String id = xml.getAttribute(element, ID_ATTRIBUTE);
		String name = xml.getAttribute(element, NAME_ATTRIBUTE);
		String schemaReference = xml.getAttribute(element, SCHEMA_ATTRIBUTE);
		PluginValidation.validateExtensionPointAttributes(id, name);
		plugin.addExtensionPoint(id, name, schemaReference);
	}
	/**
	 * @see gov.va.med.hds.cd.runtime.internal.IPluginReader#readPlugin(java.io.InputStream, java.lang.ClassLoader)
	 */
	public PluginDescriptor readPlugin(InputStream inputStream, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		try {
			ILoadableTreeData xml = TreeDataFactory.newSAXTreeData();
			xml.load(inputStream);
			return readPlugin(xml, classLoader);
		} catch (TreeDataException e) {
			throw new InvalidPluginDescriptorException(e);
		}
	}

	/**
	 * @see gov.va.med.hds.cd.runtime.internal.IPluginReader#readPlugin(java.io.Reader, java.lang.ClassLoader)
	 */
	public PluginDescriptor readPlugin(Reader reader, ClassLoader classLoader)
		throws InvalidPluginDescriptorException {
		try {
			ILoadableTreeData xml = TreeDataFactory.newSAXTreeData();
			xml.load(reader);
			return readPlugin(xml, classLoader);
		} catch (TreeDataException e) {
			throw new InvalidPluginDescriptorException(e);
		}
	}
}
