package gov.va.med.hds.cd.runtime.logging.log4j.internal;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.Writer;

import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.PlainDocument;

import org.apache.log4j.Logger;
import org.apache.log4j.TTCCLayout;
import org.apache.log4j.WriterAppender;

import gov.va.med.hds.cd.mum.boot.plaf.desktop.DesktopLookAndFeel;

/** This class captures output sent to <code>System.out</code> and
 * <code>System.err</code>, and can display this output in a dialog box
 * at the user's request.  
 */
public class LogOutput {

	private static final boolean PANIC_INITIALLY_ENABLED = true;
	private static LogOutputCollector l = null;

	/** Install a console output object to trap program output.
	 */
	public static void install() {
		if (isInstalled()) {
			return;
		}
		
		l = new LogOutputCollector(PANIC_INITIALLY_ENABLED);
		Logger.getRoot().addAppender(l);
	}

	public static void uninstall() {
		if (!isInstalled()) return;
		Logger.getRoot().removeAppender(l);
		l = null;
	}

	/** Turn panic output mode on or off.
	 * 
	 * When operating in panic mode, any data sent to System.err will
	 * trigger automatic display of a GUI.  When panic mode is off, the
	 * GUI will only be displayed in response to explicit calls to 
	 * showWindow().
	 */
	public static void setPanicEnabled(boolean panicEnabled) {
		install();
		dispatchEvent(System.err, panicEnabled ? ENABLE_PANIC : CANCEL_PANIC);
	}

	/** Display a dialog containing the accumulated console output.
	 */
	public static void showWindow() {
		install();
		dispatchEvent(System.err, SHOW_GUI);
	}

	private static final String CANCEL_PANIC = "CANCEL_PANIC";
	private static final String ENABLE_PANIC = "ENABLE_PANIC";
	private static final String SHOW_GUI = "SHOW_GUI";

	private static void dispatchEvent(Object listener, String command) {
		if (listener instanceof ActionListener)
			((ActionListener) listener).actionPerformed(
				new ActionEvent(command, 0, command));
	}

	private static boolean isInstalled() {
		return l != null;
	}

	private static class LogOutputCollector extends WriterAppender {

		/** A document model to hold the consolidated output */
		private Document document = new PlainDocument();

		/** A user interface for displaying consolidated output */
		private LogOutputFrame gui = null;

		private volatile boolean panic;

		public LogOutputCollector(boolean panic) {
			super();
			setLayout(new TTCCLayout());
			setWriter(new DocumentWriter());
			this.panic = panic;
		}

		public synchronized void showWindow() {
			try {
				if (gui == null)
					gui = new LogOutputFrame(document);
				gui.setVisible(true);
				gui.toFront();
			} catch (Throwable t) {
				// don't print anything!  If we're in panic mode, it will
				// cause a StackOverflowError.
				setPanicEnabled(false);
			}
		}

		public final void actionPerformed(ActionEvent e) {
			String command = e.getActionCommand();
			if (CANCEL_PANIC.equals(command))
				panic = false;
			else if (ENABLE_PANIC.equals(command))
				panic = true;
			else if (SHOW_GUI.equals(command))
				showGUI();
		}

		private final void showGUI() {
			showWindow();
		}

		private final void maybePanic() {
			if (panic)
				showGUI();
		}

		private class DocumentWriter extends Writer {

			private synchronized void append(String s) {
				try {
					s = s.replaceAll("\r", "");
					int docLength = document.getLength();
					document.insertString(docLength, s, null);
				} catch (BadLocationException e) {
				}
			}

			public void write(char[] cbuf, int off, int len)
				throws IOException {
				append(new String(cbuf, off, len));
				maybePanic();
			}

			public void flush() throws IOException {
				// NOOP
			}

			public void close() throws IOException {
				// NOOP
			}
		}
	}

	/** A user interface which can display log output to the user.
	 */
	private static final class LogOutputFrame extends JFrame {

		private static final String TITLE = "HealtheVet Desktop Log Output";

		private JTextArea textArea;

		public LogOutputFrame(Document document) {
			super(TITLE);
			setIconImage(DesktopLookAndFeel.getDesktopIconImage());
			setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);

			JMenuBar menubar = new JMenuBar();
			setJMenuBar(menubar);

			JMenuItem item = new JMenuItem("Copy");
			item.setMnemonic('C');
			item.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					textArea.selectAll();
					textArea.copy();
				}
			});
			menubar.add(item);

			item = new JMenuItem("Clear");
			item.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					textArea.setText(null);
				}
			});
			menubar.add(item);

			//			item = new JMenuItem("Close");
			//			item.addActionListener(new ActionListener() {
			//				public void actionPerformed(ActionEvent e) {
			//					dispose();
			//				}
			//			});
			//			menubar.add(item);

			textArea = new JTextArea(document);
			textArea.setEditable(false);
			textArea.setCaretColor(textArea.getBackground());
			getContentPane().add(new JScrollPane(textArea));
			setSize(new Dimension(600, 300));
		}

	}
}
