package gov.va.med.hds.cd.runtime.test;

import java.util.Stack;

import junit.framework.TestCase;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginActivationListener;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;
import gov.va.med.hds.cd.runtime.internal.ActivePluginManager;
import gov.va.med.hds.cd.runtime.internal.PluginPrerequisite;
import gov.va.med.hds.cd.runtime.internal.PluginRegistry;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;
import gov.va.med.hds.cd.runtime.logging.internal.Log4JLogging;
import gov.va.med.hds.util.ExceptionStackTrace;

public class ActivePluginManagerTest extends TestCase {

	public static final Version TEST_VERSION_2 = new Version("2.1.3.test");
	public static final Version TEST_VERSION_1 = new Version("1.0.0.test");
	public static final String TEST_PLUGIN_ID_1 = "someId1";
	public static final String TEST_PLUGIN_ID_2 = "someId2";
	public static final String TEST_PLUGIN_ID_3 = "someId3";

	private MockRuntime runtime;
	private PluginRegistry registry = new PluginRegistry(null);

	public Stack shutdownStack = new Stack();

	public ActivePluginManagerTest(String name) {
		super(name);
	}

	protected void createMockRegistry1() {
		MockPluginDescriptor mockDescriptor1 =
			createMockDescriptor(
				TEST_PLUGIN_ID_1,
				TEST_VERSION_1,
				new PluginPrerequisite[] {
					 new PluginPrerequisite(TEST_PLUGIN_ID_2, TEST_VERSION_1)});
		MockPluginDescriptor mockDescriptor2 =
			createMockDescriptor(
				TEST_PLUGIN_ID_2,
				TEST_VERSION_1,
				new PluginPrerequisite[] {
					 new PluginPrerequisite(TEST_PLUGIN_ID_3, TEST_VERSION_1)});
		MockPluginDescriptor mockDescriptor3 =
			createMockDescriptor(TEST_PLUGIN_ID_3, TEST_VERSION_1, new PluginPrerequisite[0]);

		try {
			registry.add(mockDescriptor1);
			registry.add(mockDescriptor2);
			registry.add(mockDescriptor3);
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
	}

	protected void createMockRegistry2() {
		MockPluginDescriptor mockDescriptor1 =
			createMockDescriptor(TEST_PLUGIN_ID_1, TEST_VERSION_1, new PluginPrerequisite[0]);
		MockPluginDescriptor mockDescriptor1v2 =
			createMockDescriptor(TEST_PLUGIN_ID_1, TEST_VERSION_2, new PluginPrerequisite[0]);

		try {
			registry.add(mockDescriptor1);
			registry.add(mockDescriptor1v2);
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
	}

	protected MockPluginDescriptor createMockDescriptor(
		String pluginId,
		Version version,
		IPluginPrerequisite[] prereqs) {
		MockPluginDescriptor result = new MockPluginDescriptor();
		for (int i = 0; i < 30; ++i) {
			result.setupGetUniqueIdentifier(pluginId);
			result.setupGetVersion(version);
			result.setupGetLabel("a plugin");
			if (TEST_PLUGIN_ID_2.equals(pluginId)) {
				result.setupGetClassName("");
			} else {
				result.setupGetClassName(ShutdownMockPlugin.class.getName());
			}
			result.setupGetPluginPrerequisites(prereqs);
			result.setupGetExtensions(new IExtension[0]);
			result.setupGetExtensionPoints(new IExtensionPoint[0]);
		}
		return result;
	}

	public void testActivateMultipleVersions()
		throws PluginStartupException, PluginNotFoundException {
		createMockRegistry2();
		runtime = new MockRuntime(registry);

		ActivePluginManager activePlugins = new ActivePluginManager(runtime);
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_2, VersionMatchCondition.PERFECT));
		activePlugins.getPluginInstance(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT);
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_2, VersionMatchCondition.PERFECT));
	}

	public void testStartupSequence() {
		createMockRegistry1();
		runtime = new MockRuntime(registry);

		ActivePluginManager activePlugins = new ActivePluginManager(runtime);
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_2, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		Object pluginInstance = null;
		try {
			pluginInstance = activePlugins.getPluginInstance(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT);
		} catch (PluginStartupException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		} catch (PluginNotFoundException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		}
		assertNotNull(pluginInstance);
		assertTrue(pluginInstance instanceof ShutdownMockPlugin);
		((MockPlugin) pluginInstance).verify();
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_2, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		try {
			pluginInstance = activePlugins.getPluginInstance(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT);
		} catch (PluginStartupException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		} catch (PluginNotFoundException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		}
		assertNotNull(pluginInstance);
		assertTrue(pluginInstance instanceof ShutdownMockPlugin);
		((MockPlugin) pluginInstance).verify();
	}

	public void testShutdownSequence() {
		createMockRegistry1();
		runtime = new MockRuntime(registry);

		ActivePluginManager activePlugins = new ActivePluginManager(runtime);
		try {
			ShutdownMockPlugin plugin =
				(ShutdownMockPlugin) activePlugins.getPluginInstance(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT);
			plugin.setTest(this);
			plugin = (ShutdownMockPlugin) activePlugins.getPluginInstance(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT);
			plugin.setTest(this);
		} catch (PluginStartupException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		} catch (PluginNotFoundException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		}
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_2, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertTrue(activePlugins.isPluginActive(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT));

		activePlugins.shutdown();

		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_1, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_2, TEST_VERSION_1, VersionMatchCondition.PERFECT));
		assertFalse(activePlugins.isPluginActive(TEST_PLUGIN_ID_3, TEST_VERSION_1, VersionMatchCondition.PERFECT));

		assertEquals(TEST_PLUGIN_ID_3, shutdownStack.pop());
		assertEquals(TEST_PLUGIN_ID_1, shutdownStack.pop());
	}

	public static class ShutdownMockPlugin extends MockPlugin {

		private ActivePluginManagerTest test;
		private IPluginDescriptor plugin;

		public void setTest(ActivePluginManagerTest test) {
			this.test = test;
		}

		public void startup(IRuntimeContext runtime, IPluginDescriptor plugin) throws Exception {
			super.startup(runtime, plugin);
			this.plugin = plugin;
		}

		public void shutdown() throws Exception {
			test.shutdownStack.push(plugin.getUniqueIdentifier());
			super.shutdown();
		}
	}

	private static class MockRuntime implements IRuntimeContext {
		private ILoggingCapable logging = new Log4JLogging();
		private IPluginRegistry registry;

		public MockRuntime(IPluginRegistry registry) {
			this.registry = registry;
		}

		public IConfigurationElement[] getSettings() {
			return null;
		}

		public IPluginRegistry getPluginRegistry() {
			return registry;
		}

		public boolean isPluginActive(String pluginId) {
			return false;
		}

		public Object getPluginInstance(String pluginId)
			throws PluginStartupException, PluginNotFoundException {
			return null;
		}

		public Object getPluginInstance(String pluginId, Version version)
			throws PluginStartupException, PluginNotFoundException {
			return null;
		}

		public ILogger getLogger(Class c) {
			return logging.getLogger(c);
		}

		public ILogger getLogger(String name) {
			return logging.getLogger(name);
		}

		public ILogger getRootLogger() {
			return logging.getRootLogger();
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#isPluginActive(java.lang.String, gov.va.med.hds.cd.mum.Version) */
		public boolean isPluginActive(String pluginId, Version version) {
			// TODO_SB: Auto-generated method stub
			return false;
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#isPluginActivating(java.lang.String, gov.va.med.hds.cd.mum.Version) */
		public boolean isPluginActivating(String pluginId, Version version) {
			// TODO_SB: Auto-generated method stub
			return false;
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#addPluginActivationListener(gov.va.med.hds.cd.runtime.IPluginActivationListener) */
		public void addPluginActivationListener(IPluginActivationListener l) {
			// TODO_SB: Auto-generated method stub
			
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#removePluginActivationListener(gov.va.med.hds.cd.runtime.IPluginActivationListener) */
		public void removePluginActivationListener(IPluginActivationListener l) {
			// TODO_SB: Auto-generated method stub
			
		}
	}
}
