package gov.va.med.hds.cd.runtime.test;

import junit.framework.TestCase;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginActivationListener;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.InvalidExtensionException;
import gov.va.med.hds.cd.runtime.PluginNotFoundException;
import gov.va.med.hds.cd.runtime.PluginStartupException;
import gov.va.med.hds.cd.runtime.internal.ConfigurationElement;
import gov.va.med.hds.cd.runtime.internal.PluginDescriptor;
import gov.va.med.hds.cd.runtime.logging.ILogger;

public class PluginDescriptorTest extends TestCase {

	private MockRuntime runtime = new MockRuntime();

	public PluginDescriptorTest(String name) {
		super(name);
	}

	public void testAddExtensionPoint() {
		PluginDescriptor plugin = new PluginDescriptor(runtime, getClass().getClassLoader(), "test", "test Plugin", new Version("0.0.1.test"));
		plugin.addExtensionPoint("testExtensionPoint", "A displayable name for this extension point");
		IExtensionPoint point = plugin.getExtensionPoint("testExtensionPoint");
		assertNotNull(point);
		assertEquals("testExtensionPoint", point.getSimpleIdentifier());
		assertEquals("test.testExtensionPoint", point.getUniqueIdentifier());
//		assertEquals("test Plugin", point.getLabel());
		assertNull(point.getSchemaReference());
		assertEquals(plugin, point.getDeclaringComponentDescriptor());
		
		assertEquals(point, plugin.getExtensionPoints()[0]);
	}

	public void testAddExtensionWithoutId() {
		PluginDescriptor plugin = new PluginDescriptor(runtime, getClass().getClassLoader(), "test", "test Plugin", new Version("0.0.1.test"));
		plugin.addPrerequisite("testPlugin", new Version("0.3.2.test"));
		
		try {
			plugin.addExtension("testPlugin.testExtensionPoint", new ConfigurationElement[0]);
		} catch (InvalidExtensionException e) {
			fail(e.getMessage());
		}
		
		assertNull(plugin.getExtension("testExtensionId"));
		
		IExtension extension = plugin.getExtensions()[0];
		assertNotNull(extension);
		assertEquals("testPlugin.testExtensionPoint", extension.getExtensionPointUniqueIdentifier());
		assertNull(extension.getSimpleIdentifier());
		assertNull(extension.getUniqueIdentifier());
		assertNull(extension.getLabel());
		assertEquals(plugin, extension.getDeclaringPluginDescriptor());
	}

	public void testAddExtensionWithId() {
		PluginDescriptor plugin = new PluginDescriptor(runtime, getClass().getClassLoader(), "test", "test Plugin", new Version("0.0.1.test"));	
		plugin.addPrerequisite("testPlugin", new Version("0.3.2.test"));
		try {
			plugin.addExtension("testPlugin.testExtensionPoint2", "testExtensionId", "testExtensionName", new ConfigurationElement[0]);
		} catch (InvalidExtensionException e) {
			fail(e.getMessage());
		}
		
		IExtension extension = plugin.getExtension("testExtensionId");
		assertNotNull(extension);
		assertEquals("testPlugin.testExtensionPoint2", extension.getExtensionPointUniqueIdentifier());
		assertEquals("testExtensionId", extension.getSimpleIdentifier());
		assertEquals("test.testExtensionId", extension.getUniqueIdentifier());
		assertEquals("testExtensionName", extension.getLabel());
		assertEquals(plugin, extension.getDeclaringPluginDescriptor());
		assertEquals(extension, plugin.getExtensions()[0]);
	}

	public void testAddInvalidExtension() {
		PluginDescriptor plugin = new PluginDescriptor(runtime, getClass().getClassLoader(), "test", "test Plugin", new Version("0.0.1.test"));
		try {
			plugin.addExtension("testExtensionPoint", new ConfigurationElement[0]);
			fail("expected " + InvalidExtensionException.class.getName()); 
		} catch (InvalidExtensionException e) {}
		
		try {
			plugin.addExtension("testPlugin.testExtensionPoint", new ConfigurationElement[0]);
			fail("expected " + InvalidExtensionException.class.getName()); 
		} catch (InvalidExtensionException e) {}
	}
	
	public void testAddMultipleExtensions() {
		PluginDescriptor plugin = new PluginDescriptor(runtime, getClass().getClassLoader(), "test", "test Plugin", new Version("0.0.1.test"));
		plugin.addPrerequisite("testPlugin", new Version("0.0.2.test"));
		try {
			plugin.addExtension("testPlugin.testExtensionPoint", new ConfigurationElement[0]);
			plugin.addExtension("testPlugin.testExtensionPoint", new ConfigurationElement[0]);
			plugin.addExtension("testPlugin.testExtensionPoint", new ConfigurationElement[0]);
		} catch (InvalidExtensionException e) {
			fail(e.getMessage());
		}
		IExtension[] extensions = plugin.getExtensions();
		assertEquals(3, extensions.length);
		assertEquals("testPlugin.testExtensionPoint", extensions[0].getExtensionPointUniqueIdentifier());
		assertEquals("testPlugin.testExtensionPoint", extensions[1].getExtensionPointUniqueIdentifier());
		assertEquals("testPlugin.testExtensionPoint", extensions[2].getExtensionPointUniqueIdentifier());
	}
	
	private static class MockRuntime implements IRuntimeContext {
		
		public Object getPluginInstance(String pluginId)
			throws PluginStartupException, PluginNotFoundException {
			return null;
		}

		public IPluginRegistry getPluginRegistry() {
			return null;
		}

		public boolean isPluginActive(String pluginId) {
			return false;
		}

		public IConfigurationElement[] getSettings() {
			return null;
		}

		public ILogger getLogger(Class c) {
			return null;
		}

		public ILogger getLogger(String name) {
			return null;
		}

		public ILogger getRootLogger() {
			return null;
		}

		public Object getPluginInstance(String pluginId, Version version)
			throws PluginStartupException, PluginNotFoundException {
			return null;
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#isPluginActive(java.lang.String, gov.va.med.hds.cd.mum.Version) */
		public boolean isPluginActive(String pluginId, Version version) {
			// TODO_SB: Auto-generated method stub
			return false;
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#isPluginActivating(java.lang.String, gov.va.med.hds.cd.mum.Version) */
		public boolean isPluginActivating(String pluginId, Version version) {
			// TODO_SB: Auto-generated method stub
			return false;
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#addPluginActivationListener(gov.va.med.hds.cd.runtime.IPluginActivationListener) */
		public void addPluginActivationListener(IPluginActivationListener l) {
			// TODO_SB: Auto-generated method stub
			
		}

		/* @see gov.va.med.hds.cd.runtime.IRuntimeContext#removePluginActivationListener(gov.va.med.hds.cd.runtime.IPluginActivationListener) */
		public void removePluginActivationListener(IPluginActivationListener l) {
			// TODO_SB: Auto-generated method stub
			
		}

	}
}
