package gov.va.med.hds.cd.runtime.test;

import junit.framework.TestCase;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeConstants;
import gov.va.med.hds.cd.runtime.InvalidExtensionException;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.VersionMatchCondition;
import gov.va.med.hds.cd.runtime.internal.ConfigurationElement;
import gov.va.med.hds.cd.runtime.internal.PluginDescriptor;
import gov.va.med.hds.cd.runtime.internal.PluginRegistry;
import gov.va.med.hds.cd.runtime.internal.RuntimeCoreDescriptor;
import gov.va.med.hds.cd.runtime.internal.RuntimeDescriptor;

public class PluginRegistryTest extends TestCase implements IRuntimeConstants {
	
	protected PluginRegistry registry;
	
	public PluginRegistryTest(String name) {
		super(name);
	}

	protected void setUp() {
		registry = new PluginRegistry(null);
		try {
			registry.add(createPrismDescriptor());
		} catch (InvalidPluginDescriptorException e) {
			fail();
		}
	}

	private PluginDescriptor createPrismDescriptor() {
		PluginDescriptor prism = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.prism", "Prism", new Version("0.1"), "gov.va.med.hds.cd.prism.Prism");
		prism.addPrerequisite(PLUGIN_ID, RuntimeDescriptor.VERSION);
		try {
			prism.addExtension(UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID, "StartingPoint", "The Main Entry Point!", createPrismConfigurationElements());
		} catch (InvalidExtensionException e) {}
		return prism;
	}

	private ConfigurationElement[] createPrismConfigurationElements() {
		ConfigurationElement[] result = new ConfigurationElement[1];
		result[0] = new ConfigurationElement("mainEntryPoint");
		result[0].addAttribute("class", "gov.va.med.hds.cd.ui.Prism");
		return result;
	}


	public void testExtensionPoints() {
		IExtensionPoint[] extensionPoints = registry.getExtensionPoints();
		assertEquals(2, extensionPoints.length);
		
		assertNull(registry.getExtensionPoint("some.test.extension.point.id"));
		assertNull(registry.getExtensionPoint("some.test.plugin.id", "someExtensionPointName"));
        
		IExtensionPoint point = registry.getExtensionPoint(UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertNotNull(point);
        
        int id0 = 0, id1 = 1;
        if (extensionPoints[0] != point) {
            id0 = 1; id1 = 0;
        }
        assertEquals(extensionPoints[id0], point);
		
		point = registry.getExtensionPoint(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertNotNull(point);
		assertEquals(extensionPoints[id0], point);       
        
        point = registry.getExtensionPoint(UNIQUE_HANDLER_PLUGIN_EXTENSION_POINT_ID);
        assertNotNull(point);
        assertEquals(extensionPoints[id1], point);       
        
        point = registry.getExtensionPoint(PLUGIN_ID, HANDLER_PLUGIN_EXTENSION_POINT_ID);
        assertNotNull(point);
        assertEquals(extensionPoints[id1], point);
	}
	
	public void testExtensions() {
		assertNull(registry.getExtension("some.test.extension.point.id", "some.extension.id"));
		IExtension extension = registry.getExtension(UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID, "StartingPoint");
		assertNotNull(extension);
		extension = registry.getExtension(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID, "StartingPoint");
		assertNotNull(extension);
		IExtension[] extensions = registry.getExtensions(UNIQUE_MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertEquals(1, extensions.length);
		assertEquals(extension, extensions[0]);
		extensions = registry.getExtensions(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertEquals(1, extensions.length);
		assertEquals(extension, extensions[0]);
	}
	
	public void testConfigurationElements() {
		assertEquals(0, registry.getConfigurationElementsFor("some.test.extension.point.id").length);
		assertEquals(0, registry.getConfigurationElementsFor("some.test.plugin.id", "someExtensionPointName").length);
		assertEquals(0, registry.getConfigurationElementsFor("some.test.plugin.id", "someExtensionPointName", "someExtensionName").length);
		
		IConfigurationElement[] configurationElements = registry.getConfigurationElementsFor(PLUGIN_ID + "." + MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertConfigurationElements(configurationElements);
		configurationElements = registry.getConfigurationElementsFor(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID);
		assertConfigurationElements(configurationElements);
		configurationElements = registry.getConfigurationElementsFor(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID, "StartingPoint");
		assertConfigurationElements(configurationElements);
	}

	private void assertConfigurationElements(IConfigurationElement[] configurationElements) {
		assertEquals(1, configurationElements.length);
		assertEquals("mainEntryPoint", configurationElements[0].getName());
		assertNull(configurationElements[0].getValue());
		assertEquals(0, configurationElements[0].getChildren("someChildElementName").length);
		assertEquals(1, configurationElements[0].getAttributeNames().length);
		assertEquals("class", configurationElements[0].getAttributeNames()[0]);
		assertEquals("gov.va.med.hds.cd.ui.Prism", configurationElements[0].getAttribute("class"));
		assertEquals(0, configurationElements[0].getChildren().length);
		assertEquals(registry.getExtension(PLUGIN_ID, MAIN_PLUGIN_EXTENSION_POINT_ID, "StartingPoint"), configurationElements[0].getDeclaringExtension());
	}
	
	public void testExistenceOfRuntimePlugins() {
		IPluginDescriptor plugin = registry.getPluginDescriptor(PLUGIN_ID);
		assertNotNull(plugin);
		assertEquals(VERSION, plugin.getVersion());
		
		plugin = registry.getPluginDescriptor(RuntimeCoreDescriptor.PLUGIN_ID);
		assertNotNull(plugin);
		assertEquals(RuntimeCoreDescriptor.VERSION, plugin.getVersion());
	}
	
	public void testAddRuntimePlugin() {
		IPluginDescriptor runtime = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.runtime", "Runtime", new Version("0.1.0.test"), Runtime.class.getName());
		try {
			registry.add(runtime);
			fail("expected reserved id exception");
		} catch (InvalidPluginDescriptorException e) {
			
		}
		IPluginDescriptor runtimeCore = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.runtime.core", "Runtime Core", new Version("0.1.0.test"), null);
		try {
			registry.add(runtimeCore);
			fail("expected reserved id exception");
		} catch (InvalidPluginDescriptorException e) {
		}
	}
	
	public void testAddExistingPlugin() {
		IPluginDescriptor prism = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.prism", "Prism 2", new Version("0.1"), "gov.va.med.hds.cd.prism.Prism2");
		try {
			registry.add(prism);
			prism = registry.getPluginDescriptor("gov.va.med.hds.cd.prism");
			assertEquals("Prism 2", prism.getLabel());
			assertEquals("gov.va.med.hds.cd.prism.Prism2", prism.getClassName());
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
	}
	
	public void testAddSelfExtendingPlugin() {
		PluginDescriptor m = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.test", "Self Extending Plugin", new Version("0.1"), "gov.va.med.hds.cd.test.Test");
		m.addExtensionPoint("testExtensionPoint", "Test Extension Point");
		try {
			m.addExtension("gov.va.med.hds.cd.test.testExtensionPoint", new ConfigurationElement[0]);
			registry.add(m);
		} catch (InvalidExtensionException e) {
			fail(e.getMessage());
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
		assertNotNull(registry.getPluginDescriptor("gov.va.med.hds.cd.test"));
	}
	
	public void testUnresolvedPlugin() {
		assertTrue(registry.isResolved());
		assertEquals(0, registry.getUnresolvedPluginDescriptors().length);
		PluginDescriptor m = new PluginDescriptor(null, getClass().getClassLoader(), "gov.va.med.hds.cd.test", "Plugin With Unmet Dependency", new Version("0.1"), "gov.va.med.hds.cd.test.Test");
		m.addPrerequisite("some.plugin.id", new Version("2.0.0.test"));
		try {
			registry.add(m);
		} catch (InvalidPluginDescriptorException e) {
			fail(e.toString());
		}
		assertFalse(registry.isResolved());
		assertEquals(1, registry.getUnresolvedPluginDescriptors().length);
		assertEquals(m, registry.getUnresolvedPluginDescriptors()[0]);
	}
	
	public void testMultipleExtensionsByOnePlugin() {
		PluginDescriptor m1 = new PluginDescriptor(null, getClass().getClassLoader(), "plugin1", "plugin 1", new Version("0.1.2"), MockPlugin.class.getName());
		PluginDescriptor m2 = new PluginDescriptor(null, getClass().getClassLoader(), "plugin2", "plugin 2", new Version("0.1.2"), MockPlugin.class.getName());
		m1.addExtensionPoint("testExtensionPoint", "Test");
		m2.addPrerequisite("plugin1", new Version("0.1.2"));
		try {
			m2.addExtension("plugin1.testExtensionPoint", new ConfigurationElement[] { new ConfigurationElement("Test1") });
			m2.addExtension("plugin1.testExtensionPoint", new ConfigurationElement[] { new ConfigurationElement("Test2") });
			m2.addExtension("plugin1.testExtensionPoint", new ConfigurationElement[] { new ConfigurationElement("Test3") });
		} catch (InvalidExtensionException e) { 
			fail(e.getMessage());
		}
		try {
			registry.add(m1);
			registry.add(m2);
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
		assertTrue(registry.isResolved());
		IConfigurationElement[] elements = registry.getConfigurationElementsFor("plugin1.testExtensionPoint");
		assertEquals(3, elements.length);
		assertEquals("Test1", elements[0].getName());
		assertEquals("Test2", elements[1].getName());
		assertEquals("Test3", elements[2].getName());
	}
	
	public void testGetPluginDescriptorVersionMatching() {
		PluginDescriptor m1 = new PluginDescriptor(null, getClass().getClassLoader(), "testPlugin", "A Test Plugin", new Version("0.9.0.beta"), MockPlugin.class.getName());
		PluginDescriptor m2 = new PluginDescriptor(null, getClass().getClassLoader(), "testPlugin", "A Test Plugin", new Version("1.0"), MockPlugin.class.getName());
		PluginDescriptor m3 = new PluginDescriptor(null, getClass().getClassLoader(), "testPlugin", "A Test Plugin", new Version("1.1"), MockPlugin.class.getName());
		PluginDescriptor m4 = new PluginDescriptor(null, getClass().getClassLoader(), "testPlugin", "A Test Plugin", new Version("1.1.4"), MockPlugin.class.getName());
		PluginDescriptor m5 = new PluginDescriptor(null, getClass().getClassLoader(), "testPlugin", "A Test Plugin", new Version("2.0"), MockPlugin.class.getName());
		try {
			registry.add(m1);
			registry.add(m2);
			registry.add(m3);
			registry.add(m4);
			registry.add(m5);
		} catch (InvalidPluginDescriptorException e) {
			fail(e.getMessage());
		}
		assertTrue(registry.isResolved());
		assertEquals(m5, registry.getPluginDescriptor("testPlugin"));
		assertEquals(m3, registry.getPluginDescriptor("testPlugin", new Version("1.1")));
		assertEquals(m2, registry.getPluginDescriptor("testPlugin", new Version("1.0"), VersionMatchCondition.PERFECT));
		assertEquals(m1, registry.getPluginDescriptor("testPlugin", new Version("0.9.0.beta"), VersionMatchCondition.PERFECT));
		assertEquals(m4, registry.getPluginDescriptor("testPlugin", new Version("1.1"), VersionMatchCondition.EQUIVALENT));
		assertEquals(m4, registry.getPluginDescriptor("testPlugin", new Version("1.0"), VersionMatchCondition.COMPATIBLE));
		assertEquals(m5, registry.getPluginDescriptor("testPlugin", new Version("1.0"), VersionMatchCondition.GREATER_OR_EQUAL));
	}
}
