package gov.va.med.hds.cd.runtime.test;

import java.io.StringReader;

import junit.framework.TestCase;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IExtension;
import gov.va.med.hds.cd.runtime.IExtensionPoint;
import gov.va.med.hds.cd.runtime.IPluginPrerequisite;
import gov.va.med.hds.cd.runtime.IRuntimeConstants;
import gov.va.med.hds.cd.runtime.InvalidPluginDescriptorException;
import gov.va.med.hds.cd.runtime.internal.IPluginReader;
import gov.va.med.hds.cd.runtime.internal.PluginDescriptor;
import gov.va.med.hds.util.ExceptionStackTrace;

public abstract class ValidPluginReaderTest extends TestCase implements IRuntimeConstants {
	protected static final String VALID_RUNTIME_MANIFEST_XML =
		"<?xml version=\"1.0\"?>"
			+ "<plugin name=\"Runtime\" id=\"gov.va.med.hds.cd.runtime\" version=\"0.2.3.test\" class=\"some.test.class\" provider-name=\"CPRS Re\">"
			+ "<extension-points>"
			+ "<extension-point name=\"Main Plugin\" id=\"mainPlugin\" schema=\"MainPlugin.xsd\"/>"
			+ "</extension-points>"
			+ "</plugin>";
	protected static final String VALID_PRISM_MANIFEST_XML =
		"<?xml version=\"1.0\"?>"
			+ "<plugin name=\"Prism\" id=\"gov.va.med.hds.cd.prism\" version=\"0.1\" provider-name=\"CPRS Re\">"
			+ "<prerequisites>"
			+ "<requires plugin=\"gov.va.med.hds.cd.runtime\" version=\"0.2.3.test\"/>"
			+ "<requires plugin=\"some.test.plugin1\" version=\"3232.45.324\" match=\"greaterOrEqual\"/>"
			+ "<requires plugin=\"some.test.plugin2\" version=\"3232.45.324\" match=\"compatible\"/>"
			+ "<requires plugin=\"some.test.plugin3\" version=\"3232.45.324\" match=\"equivalent\"/>"
			+ "<requires plugin=\"some.test.plugin4\" version=\"3232.45.324\" match=\"perfect\"/>"
			+ "<requires plugin=\"some.test.plugin5\" version=\"3232.45.324\" match=\"perfect\"/>"
			+ "</prerequisites>"
			+ "<extensions>"
			+ "<extension extends=\"gov.va.med.hds.cd.runtime.mainPlugin\">"
			+ "<!-- some content that is validated by MainEntryPoint.xsd -->"
			+ "<mainEntryPoint class=\"gov.va.med.hds.cd.ui.Prism\"/>"
			+ "<root>"
			+ "<child1/>"
			+ "<child2 aTestAttribute=\"theAttributeValue\">"
			+ "<child21/>"
			+ "</child2>"
			+ "<child3/>"
			+ "</root>"
			+ "<aValue>testValue</aValue>"
			+ "<aSecondValue>aSecondTestValue</aSecondValue>"
			+ "</extension>"
			+ "<extension extends=\"some.test.plugin1.testExtensionPoint\">"
			+ "<plugin>prerequisite</plugin>"
			+ "</extension>"
			+ "</extensions>"
			+ "</plugin>";

	public ValidPluginReaderTest(String name) {
		super(name);
	}

	protected abstract IPluginReader createPluginReader();

	public void testReadRuntimePlugin() {
		try {
			IPluginReader reader = createPluginReader();
			PluginDescriptor plugin =
				reader.readPlugin(
					new StringReader(VALID_RUNTIME_MANIFEST_XML),
					getClass().getClassLoader());

			// assert plugin properties
			assertEquals("gov.va.med.hds.cd.runtime", plugin.getUniqueIdentifier());
			assertEquals(new Version("0.2.3.test"), plugin.getVersion());
			assertEquals("CPRS Re", plugin.getProviderName());
			assertEquals("some.test.class", plugin.getClassName());
			assertEquals(getClass().getClassLoader(), plugin.getPluginClassLoader());
			assertTrue(plugin.getRuntimeContext() instanceof TestRuntime);

			// assert libraries
			// assert prerequisites
			assertEquals(0, plugin.getPluginPrerequisites().length);

			// assert extensionPoints
			assertEquals(1, plugin.getExtensionPoints().length);
			assertNull(plugin.getExtensionPoint("some.test.id"));

			IExtensionPoint mainPlugin = plugin.getExtensionPoint("mainPlugin");
			assertNotNull(mainPlugin);
			assertEquals(plugin, mainPlugin.getDeclaringComponentDescriptor());
			assertEquals("gov.va.med.hds.cd.runtime.mainPlugin", mainPlugin.getUniqueIdentifier());
			assertEquals("mainPlugin", mainPlugin.getSimpleIdentifier());
			assertEquals("Main Plugin", mainPlugin.getLabel());
			assertEquals("MainPlugin.xsd", mainPlugin.getSchemaReference());

			// assert extensions
			assertEquals(0, plugin.getExtensions().length);
			assertNull(plugin.getExtension("someTestName"));
		} catch (InvalidPluginDescriptorException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		}
	}

	public void testReadPrismPlugin() {
		try {
			IPluginReader reader = createPluginReader();
			PluginDescriptor plugin =
				reader.readPlugin(
					new StringReader(VALID_PRISM_MANIFEST_XML),
					getClass().getClassLoader());

			// assert plugin properties
			assertEquals("gov.va.med.hds.cd.prism", plugin.getUniqueIdentifier());
			assertEquals(0, plugin.getVersion().getMajorComponent());
			assertEquals(1, plugin.getVersion().getMinorComponent());
			assertNull(plugin.getClassName());
			assertEquals("CPRS Re", plugin.getProviderName());
			assertEquals(getClass().getClassLoader(), plugin.getPluginClassLoader());
			assertNotNull(plugin.getRuntimeContext());

			// assert prerequisites
			IPluginPrerequisite[] prereqs = plugin.getPluginPrerequisites();
			assertEquals(6, prereqs.length);
			IPluginPrerequisite prereq = getPrerequisite(prereqs, "gov.va.med.hds.cd.runtime");
			assertNotNull(prereq);
			assertEquals(new Version("0.2.3.test"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());
			prereq = getPrerequisite(prereqs, "some.test.plugin1");
			assertNotNull(prereq);
			assertEquals(new Version("3232.45.324"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());
			prereq = getPrerequisite(prereqs, "some.test.plugin2");
			assertNotNull(prereq);
			assertEquals(new Version("3232.45.324"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());
			prereq = getPrerequisite(prereqs, "some.test.plugin3");
			assertNotNull(prereq);
			assertEquals(new Version("3232.45.324"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());
			prereq = getPrerequisite(prereqs, "some.test.plugin4");
			assertNotNull(prereq);
			assertEquals(new Version("3232.45.324"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());
			prereq = getPrerequisite(prereqs, "some.test.plugin5");
			assertNotNull(prereq);
			assertEquals(new Version("3232.45.324"), prereq.getVersion());
			assertTrue(prereq.isMatchedAsGreaterOrEqual());

			// assert extensionPoints
			assertEquals(0, plugin.getExtensionPoints().length);
			assertNull(plugin.getExtensionPoint("some.test.id"));

			// assert extensions
			IExtension[] extensions = plugin.getExtensions();
			assertEquals(2, extensions.length);
			assertNull(plugin.getExtension("someTestName"));
			int id0 = 0, id1 = 1;
			if (!"gov.va.med.hds.cd.runtime.mainPlugin"
				.equals(extensions[0].getExtensionPointUniqueIdentifier())) {
				id1 = 0;
				id0 = 1;
			}
			assertEquals(plugin, extensions[id0].getDeclaringPluginDescriptor());
			assertNull(extensions[id0].getUniqueIdentifier());
			assertNull(extensions[id0].getSimpleIdentifier());
			assertNull(extensions[id0].getLabel());
			assertEquals(
				"gov.va.med.hds.cd.runtime.mainPlugin",
				extensions[id0].getExtensionPointUniqueIdentifier());

			assertEquals(plugin, extensions[id1].getDeclaringPluginDescriptor());
			assertNull(extensions[id1].getUniqueIdentifier());
			assertNull(extensions[id1].getSimpleIdentifier());
			assertNull(extensions[id1].getLabel());
			assertEquals(
				"some.test.plugin1.testExtensionPoint",
				extensions[id1].getExtensionPointUniqueIdentifier());

			// assert configuration elements
			IConfigurationElement[] configElements = extensions[id0].getConfigurationElements();
			assertEquals(4, configElements.length);
			assertEquals("mainEntryPoint", configElements[0].getName());
			assertEquals(extensions[id0], configElements[0].getDeclaringExtension());
			assertEquals("gov.va.med.hds.cd.ui.Prism", configElements[0].getAttribute("class"));
			assertEquals(0, configElements[0].getChildren().length);

			assertEquals("root", configElements[1].getName());
			assertEquals(extensions[id0], configElements[1].getDeclaringExtension());

			IConfigurationElement[] children1 = configElements[1].getChildren();
			assertEquals(3, children1.length);
			assertEquals("child1", children1[0].getName());
			assertEquals(extensions[id0], children1[0].getDeclaringExtension());
			assertEquals("child2", children1[1].getName());
			assertEquals(extensions[id0], children1[1].getDeclaringExtension());
			assertEquals("theAttributeValue", children1[1].getAttribute("aTestAttribute"));
			assertEquals("child3", children1[2].getName());
			assertEquals(extensions[id0], children1[2].getDeclaringExtension());
			assertEquals(1, children1[1].getChildren().length);
			assertEquals("child21", children1[1].getChildren()[0].getName());

			assertEquals("aValue", configElements[2].getName());
			assertEquals("testValue", configElements[2].getValue());

			assertEquals("aSecondValue", configElements[3].getName());
			assertEquals("aSecondTestValue", configElements[3].getValue());

			configElements = extensions[id1].getConfigurationElements();
			assertEquals(1, configElements.length);
			assertEquals(extensions[id1], configElements[0].getDeclaringExtension());
			assertEquals("plugin", configElements[0].getName());
			assertEquals("prerequisite", configElements[0].getValue());
		} catch (InvalidPluginDescriptorException e) {
			fail(ExceptionStackTrace.getStackTrace(e));
		}
	}

	private IPluginPrerequisite getPrerequisite(
		IPluginPrerequisite[] prereqs,
		String prerequisiteId) {
		for (int i = 0; i < prereqs.length; i++) {
			if (prereqs[i].getUniqueIdentifier().equals(prerequisiteId))
				return prereqs[i];
		}
		return null;
	}
}
