package gov.va.med.hds.cd.runtime;

/**
 * Interface for executable extension classes that require access to 
 * their configuration element, or implement an extension adapter.
 * <p>
 * Extension adapters are typically required in cases where the extension
 * implementation does not follow the interface rules specified
 * by the provider of the extension point. In these
 * cases, the role of the adapter is to map between the extension point
 * interface, and the actual extension implementation. In general, adapters
 * are used when attempting to plug-in existing Java implementations, or
 * non-Java implementations (e.g., external executables).
 * </p>
 * <p>
 * Clients may implement this interface.
 * </p>
 * 
 * @see IConfigurationElement#createExecutableExtension 
 */
public interface IExecutableExtension {

	/**
	 * This method is called by the implementation of the method
	 * <code>IConfigurationElement.createExecutableExtension</code>
	 * on a newly constructed extension, passing it its relevant configuration 
	 * information. 
	 * <p>
	 * Regular executable extensions specify their Java implementation 
	 * class name as an attribute of the configuration element for the 
	 * extension. For example
	 * <pre>
	 *     &lt;action run="com.example.BaseAction"/&gt;
	 * </pre>
	 * In the above example, this method would be called with a reference
	 * to the <code>&lt;action&gt;</code> element (first argument), and
	 * <code>"run"</code> as the name of the attribute that defined
	 * this executable extension (second argument).
	 * </p>
	 *
	 * @param runtime a handle to the runtime context
	 * @param config the configuration element used to trigger this execution. 
	 *      It can be queried by the executable extension for specific
	 *      configuration properties
	 * @param propertyName the name of an attribute of the configuration element
	 *      used on the <code>createExecutableExtension(String)</code> call. This
	 *      argument can be used in the cases where a single configuration element
	 *      is used to define multiple executable extensions.
	 * @param data 
	 * @exception Exception if error(s) detected during initialization
	 * processing
	 * @see IConfigurationElement#createExecutableExtension
	 */
	void setInitializationData(
		IRuntimeContext runtime,
		IConfigurationElement config,
		String propertyName,
		Object data)
		throws Exception;

}
