package gov.va.med.hds.cd.runtime;

/**
 * An extension declared in a plugin.
 * All information is obtained from the declaring plugin's 
 * manifest.
 * <p>
 * This interface is not intended to be implemented by clients.
 * </p>
 */
public interface IExtension {
	/**
	 * Returns all configuration elements declared by this extension.
	 * These elements are a direct reflection of the configuration 
	 * markup supplied in the manifest for the plugin that declares
	 * this extension.  Returns an empty array if this extension does
	 * not declare any configuration elements.
	 *
	 * @return the configuration elements declared by this extension 
	 */
	IConfigurationElement[] getConfigurationElements();
	/** 
	 * Returns the descriptor of the plugin that declares this extension.
	 *
	 * @return the plugin that declares this extension
	 */
	IPluginDescriptor getDeclaringPluginDescriptor();
	/**
	 * Returns the unique identifier of the extension point
	 * that this extension gets plugged into.
	 *
	 * @return the unique identifier of the relevant extension point
	 */
	String getExtensionPointUniqueIdentifier();
	/**
	 * Returns a displayable label for this extension.
	 * Returns the empty string if no label for this extension
	 * is specified in the plugin manifest file.
	 * <p> Note that any translation specified in the plugin manifest
	 * file is automatically applied.
	 * <p>
	 *
	 * @see IPluginDescriptor#getResourceString 
	 *
	 * @return a displayable string label for this extension,
	 *    possibly the empty string
	 */
	String getLabel();
	/**
	 * Returns the simple identifier of this extension, or <code>null</code>
	 * if this extension does not have an identifier.
	 * This identifier is specified in the plugin manifest as a non-empty 
	 * string containing no period characters (<code>'.'</code>) and must be
	 * unique within the defining plugin.
	 *
	 * @return the simple identifier of the extension (e.g. <code>"main"</code>)
	 *  or <code>null</code>
	 */
	String getSimpleIdentifier();
	/**
	 * Returns the unique identifier of this extension, or <code>null</code>
	 * if this extension does not have an identifier.
	 * If available, this identifier is unique within the plugin registry, and
	 * is composed of the identifier of the plugin that declared
	 * this extension and this extension's simple identifier.
	 *
	 * @return the unique identifier of the extension
	 *    (e.g. <code>"com.example.acme.main"</code>), or <code>null</code>
	 */
	String getUniqueIdentifier();
}
