
package gov.va.med.hds.cd.runtime;

/**
 * An extension point declared in a plugin.
 * Except for the list of extensions plugged in to it, the information 
 * available for an extension point is obtained from the declaring plugin's 
 * manifest.
 * <p>
 * This interface is not intended to be implemented by clients.
 * </p>
 */
public interface IExtensionPoint {
	/** 
	 * Returns the descriptor of the plugin that declares this extension point.
	 *
	 * @return the plugin that declares this extension point
	 */
	IPluginDescriptor getDeclaringComponentDescriptor();
	/**
	 * Returns a displayable label for this extension point.
	 * Returns the empty string if no label for this extension point
	 * is specified in the plugin manifest file.
	 * <p> Note that any translation specified in the plugin manifest
	 * file is automatically applied.
	 * </p>
	 *
	 * @return a displayable string label for this extension point,
	 *    possibly the empty string
	 * @see IPluginDescriptor#getResourceString 
	 */
	String getLabel();
	/**
	 * Returns reference to the extension point schema. The schema 
	 * reference is returned as a URL path relative to the plugin 
	 * installation URL. 
	 * Returns the empty string if no schema for this extension point
	 * is specified in the plugin manifest file.
	 *
	 * @return a relative URL path, or an empty string
	 * @see IPluginDescriptor#getInstallURL
	 */
	String getSchemaReference();
	/**
	 * Returns the simple identifier of this extension point.
	 * This identifier is a non-empty string containing no
	 * period characters (<code>'.'</code>) and is guaranteed
	 * to be unique within the defining plugin.
	 *
	 * @return the simple identifier of the extension point (e.g. <code>"builders"</code>)
	 */
	String getSimpleIdentifier();
	/**
	 * Returns the unique identifier of this extension point.
	 * This identifier is unique within the plugin registry, and
	 * is composed of the identifier of the plugin that declared
	 * this extension point and this extension point's simple identifier.
	 *
	 * @return the unique identifier of the extension point
	 *    (e.g. <code>"org.eclipse.core.resources.builders"</code>)
	 */
	String getUniqueIdentifier();
}
