package gov.va.med.hds.cd.runtime;

/**
 * Classes implementing </code>IPlugin<code> provide plugin implementation to
 * clients of the runtime.  <code>IPlugin</code> lets the runtime control the
 * lifecycle of plugin code.  A single instance of the registered class is
 * instantiated by the runtime and then is accessible to clients of that
 * plugin's API.  
 * <p>
 * Plugins are "started up" or "activated" lazily. That is, not until a client
 * requests access to the plugin instance.  The startup sequence is also
 * governed by what a plugin declares as its prerequisites in its manifest. A
 * plugin's prerequisites are activated before <code>startup()</code> is called
 * for that plugin.
 * </p>
 * <p>
 * Classes specified in the 'class' attribute of a plugin's manifest must
 * implement this interface.
 * </p>
 * <p>
 * Clients may implement this interface.
 * </p>
 * @see gov.va.med.hds.cd.runtime.IRuntimeContext#getPluginInstance(String)
 */
public interface IPlugin {
	/**
	 * Initializes this plugin.  
	 * 
	 * To provide the best user experience implementations of this
	 * method should remain as lightweight as possible, as they can slow down
	 * desktop launch times. (For example, it is probably not a good idea to
	 * attempt to fetch data in this method)
	 * 
	 * @param runtime the runtime context that this plugin will be a part of
	 * after startup.
	 * @param plugin the 
	 * @throws Exception if plugin instance throws an exception during startup
	 */
	void startup(IRuntimeContext runtime, IPluginDescriptor plugin) throws Exception;
	/**
	 * Shutdown this plugin.
	 * 
	 * Plug-ins are expected to free any shared resources they manage.  Plug-ins 
	 * should not store state at this time; a separate lifecycle event preceding 
	 * the shutdown notice will tell plug-ins the right time to be saving their state.
	 * 
	 * @throws Exception if plugin instance throws an exception during shutdown
	 */
	void shutdown() throws Exception;
}
