package gov.va.med.hds.cd.runtime;

import gov.va.med.hds.cd.mum.Version;

/**
 * A plugin descriptor contains information about a plugin
 * obtained from the plugin's manifest.
 * <p>
 * Plugin descriptors are platform-defined objects that exist
 * in the plugin registry independent of whether a plugin has
 * been started. In contrast, a plugin's runtime object 
 * (<code>getPlugin</code>) generally runs plugin-defined code.
 * </p>
 * <p>
 * This interface is not intended to be implemented by clients.
 * </p>
 */
public interface IPluginDescriptor {
	/**
	 * Returns the extension with the given simple identifier declared in
	 * this plugin, or <code>null</code> if there is no such extension.
	 * Since an extension might not have an identifier, some extensions
	 * can only be found via the <code>getExtensions</code> method.
	 *
	 * @param extensionName the simple identifier of the extension (e.g. <code>"main"</code>).
	 * @return the extension, or <code>null</code>
	 */
	IExtension getExtension(String extensionName);
	/**
	 * Returns the extension point with the given simple identifier
	 * declared in this plugin, or <code>null</code> if there is no such extension point.
	 *
	 * @param extensionPointId the simple identifier of the extension point (e.g. <code>"wizard"</code>).
	 * @return the extension point, or <code>null</code>
	 */
	IExtensionPoint getExtensionPoint(String extensionPointId);
	/**
	 * Returns all extension points declared by this plugin.
	 * Returns an empty array if this plugin does not declare any extension points.
	 *
	 * @return the extension points declared by this plugin
	 */
	IExtensionPoint[] getExtensionPoints();
	/**
	 * Returns all extensions declared by this plugin.
	 * Returns an empty array if this plugin does not declare any extensions.
	 *
	 * @return the extensions declared by this plugin
	 */
	IExtension[] getExtensions();
	/**
	 * Returns a displayable label for this plugin.
	 * Returns the empty string if no label for this plugin
	 * is specified in the plugin manifest file.
	 * <p> Note that any translation specified in the plugin manifest
	 * file is automatically applied.
	 * </p>
	 *
	 * @return a displayable string label for this plugin,
	 *    possibly the empty string
	 * @see #getResourceString 
	 */
	String getLabel();
	/**
	* Returns a list of plugin prerequisites required
	* for correct execution of this plugin.
	*
	* @return an array of plugin prerequisites, or an empty array
	* if no prerequisites were specified
	*/
	IPluginPrerequisite[] getPluginPrerequisites();
	/**
	 * Returns the name of the provider of this plugin.
	 * Returns the empty string if no provider name is specified in 
	 * the plugin manifest file.
	 * <p> Note that any translation specified in the plugin manifest
	 * file is automatically applied.
	 * </p>
	 *
	 * @see #getResourceString 
	 *
	 * @return the name of the provider, possibly the empty string
	 */
	String getProviderName();
	/**
	 * Returns the unique identifier of this plugin.
	 * This identifier is a non-empty string and is unique 
	 * within the plugin registry.
	 *
	 * @return the unique identifier of the plugin 
	 *		(e.g. <code>"org.eclipse.core.runtime"</code>)
	 */
	String getUniqueIdentifier();
	/**
	 * Returns the version identifier of this plugin.
	 *
	 * @return the plugin version identifier
	 */
	Version getVersion();
	/**
	 * Returns the name of a class that implements <code>IPlugin</code>.
	 * @return the name of a class that implements <code>IPlugin</code>
	 */
	String getClassName();
	/**
	 * Returns the runtime context to which this plugin descriptor belongs.
	 * @return IRuntimeContext
	 */
	IRuntimeContext getRuntimeContext();

	/**
	 * Returns the plug-in class loader used to load classes and resources for this
	 * plug-in. The class loader can be used to directly access plug-in resources
	 * and classes.
	 * 
	 * The following examples illustrate the direct use of the plug-in class
	 * loader and its effect on plug-in activation (example ignores error
	 * handling).
	 * 
	 * @returns the plug-in class loader
	 */
	ClassLoader getPluginClassLoader();
}
