
package gov.va.med.hds.cd.runtime;

import gov.va.med.hds.cd.mum.Version;

/**
 * A prerequisite entry declared by a plugin. The declaration causes
 * classes defined by the prerequisite plugin to be visible
 * to the plugin that declared the dependency.
 * <p>
 * This interface is not intended to be implemented by developers.
 * </p>
 *
 * @see IPluginDescriptor#getPluginPrerequisites
 */
public interface IPluginPrerequisite {
	/**
	 * Returns the plugin identifier of the prerequisite plugin.
	 * 
	 * @return the plugin identifier
	 */
	String getUniqueIdentifier();
	/**
	 * Returns the version identifier of the prerequisite plugin,
	 * or <code>null</code> if none.
	 * 
	 * @return the plugin version identifier, or <code>null</code> if 
	 *    none was specified
	 */
	Version getVersion();
	/**
	 * Indicates whether this prerequisite plugin is further exposed to any
	 * plugins that declare a dependency on this plugin. This allows
	 * for chaining of dependencies. For example, if plugin A depends
	 * on plugin B which depends on plugin C, the classes from C 
	 * are typically visible to B, but not to A.  A can get around this 
	 * if either B explicitly exports its dependency on C, or 
	 * A explicitly declares C as a prerequisite in addition to B.
	 * 
	 * @return <code>true</code> if this prerequisite plugin is exposed,
	 *    <code>false</code> otherwise
	 */
//	public boolean isExported();
	/**
	 * Indicates that this plugin prerequisite can be resolved
	 * against a configured plugin with an identifier that is
	 * greater than or equal to it.
	 *
	 * @return <code>true</code> if greater or equal match is allowed,
	 *   <code>false</code> otherwise.
	 */
	boolean isMatchedAsGreaterOrEqual();
	/**
	 * Indicates that this plugin prerequisite can be resolved
	 * against a configured plugin with a compatible identifier.
	 *
	 * @return <code>true</code> if compatible match is allowed,
	 *   <code>false</code> if exact match is required.
	 */
	boolean isMatchedAsCompatible();
	/**
	 * Indicates that this plugin prerequisite can only be resolved
	 * against a configured plugin with an equivalent plugin 
	 * identifier.
	 *
	 * @return <code>true</code> if only equivalent identifier match
	 * satisfies this dependency, <code>false</code> otherwise.
	 */
	boolean isMatchedAsEquivalent();
	/**
	 * Indicates that this plugin prerequisite can only be resolved
	 * against a configured plugin with a plugin identifier that
	 * is perfectly equal.
	 *
	 * @return <code>true</code> if only perfectly equal
	 * identifier match satisfies this dependency,
	 * <code>false</code> otherwise.
	 */
	boolean isMatchedAsPerfect();
}
