package gov.va.med.hds.cd.runtime;

import gov.va.med.hds.cd.mum.Version;


/**
 * The plugin registry holds the master list of all
 * discovered plugins, extension points, and extensions.
 * <p>
 * The plugin registry can be queried, by name, for
 * plugins, extension points, and extensions.
 * </p>
 * <p>
 * This interface is not intended to be implemented by clients.
 * </p>
 */
public interface IPluginRegistry {
    /**
     * Returns all configuration elements from all extensions configured
     * into the identified extension point. Returns an empty array if the extension
     * point does not exist, has no extensions configured, or none of the extensions
     * contain configuration elements.
     *
     * @param extensionPointId the unique identifier of the extension point
     *                (e.g. <code>"org.eclipse.core.resources.builders"</code>)
     * @return the configuration elements
     */
    IConfigurationElement[] getConfigurationElementsFor(String extensionPointId);

    /**
     * Returns all configuration elements from all extensions configured
     * into the identified extension point. Returns an empty array if the extension
     * point does not exist, has no extensions configured, or none of the extensions
     * contain configuration elements.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"org.eclipse.core.resources"</code>)
     * @param extensionPointName the simple identifier of the
     *                extension point (e.g. <code>"builders"</code>)
     * @return the configuration elements
     */
    IConfigurationElement[] getConfigurationElementsFor(String pluginId,
                                                               String extensionPointName);

    /**
     * Returns all configuration elements from the identified extension.
     * Returns an empty array if the extension does not exist or
     * contains no configuration elements.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"org.eclipse.core.resources"</code>)
     * @param extensionPointName the simple identifier of the
     *                extension point (e.g. <code>"builders"</code>)
     * @param extensionId the unique identifier of the extension
     *                (e.g. <code>"com.example.acme.coolbuilder</code>)
     * @return the configuration elements
     */
    IConfigurationElement[] getConfigurationElementsFor(String pluginId,
                                                               String extensionPointName,
                                                               String extensionId);

    /**
     * Returns the specified extension in this plugin registry,
     * or <code>null</code> if there is no such extension.
     * The first parameter identifies the extension point, and the second
     * parameter identifies an extension plugged in to that extension point.
     *
     * @param extensionPointId the unique identifier of the extension point
     *                (e.g. <code>"org.eclipse.core.resources.builders"</code>)
     * @param extensionId the unique identifier of the extension
     *                (e.g. <code>"com.example.acme.coolbuilder"</code>)
     * @return the extension, or <code>null</code>
     */
    IExtension getExtension(String extensionPointId, String extensionId);

    /**
     * Returns the specified extension in this plugin registry,
     * or <code>null</code> if there is no such extension.
     * The first two parameters identify the extension point, and the third
     * parameter identifies an extension plugged in to that extension point.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>)
     * @param extensionPointName the simple identifier of the
     *                extension point (e.g. <code>"perspective"</code>)
     * @param extensionId the unique identifier of the extension
     *                (e.g. <code>"com.example.acme.coolplugin"</code>)
     * @return the extension, or <code>null</code>
     */
    IExtension getExtension(String pluginId, String extensionPointName, String extensionId);

	/**
	 * Returns the extensions in this plugin registry that extend the specified
	 * extension point or an empty array if there are none.
	 * 
	 * @param extensionPointId the unique identifier of the extension point (e.
	 * g. <code>"gov.va.med.hds.cd.prism.visualComponent"</code>)
	 * @return an array of extensions, or an empty array
	 */
    IExtension[] getExtensions(String extensionPointId);

	/**
	 * Returns the extensions in this plugin registry that
	 * extend the specified extension point or an empty array if there are none.
	 * Calling this method is equivalent to calling <code>getExtensions(pluginId + "." + extensionPointName)</code>
	 * 
	 * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>)
     * @param extensionPointName the simple identifier of the
     *                extension point (e.g. <code>"perspective"</code>)
	 * @return an array of extensions, or an empty array
	 */
    IExtension[] getExtensions(String pluginId, String extensionPointName);

    /**
     * Returns the extension point with the given extension point identifier
     * in this plugin registry, or <code>null</code> if there is no such
     * extension point.
     *
     * @param extensionPointId the unique identifier of the extension point
     *    (e.g., <code>"gov.va.med.hds.cd.prism.action"</code>)
     * @return the extension point, or <code>null</code>
     */
    IExtensionPoint getExtensionPoint(String extensionPointId);

    /**
     * Returns the extension point in this plugin registry
     * with the given plugin identifier and extension point simple identifier,
     * or <code>null</code> if there is no such extension point.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>)
     * @param extensionPointName the simple identifier of the
     *                extension point (e.g. <code>"action"</code>)
     * @return the extension point, or <code>null</code>
     */
    IExtensionPoint getExtensionPoint(String pluginId, String extensionPointName);

    /**
     * Returns all extension points known to this plugin registry.
     * Returns an empty array if there are no extension points.
     *
     * @return the extension points known to this plugin registry
     */
    IExtensionPoint[] getExtensionPoints();

    /**
     * Returns the plugin descriptor with the given plugin identifier
     * in this plugin registry, or <code>null</code> if there is no such
     * plugin.  If there are multiple versions of the identified plugin,
     * one will be non-deterministically choosen and returned.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"com.example.acme"</code>).
     * @return the plugin descriptor, or <code>null</code>
     */
    IPluginDescriptor getPluginDescriptor(String pluginId);

    /**
     * Returns the plugin descriptor with the given plugin identifier
     * and version in this plugin registry, or <code>null</code> if
     * there is no such plugin.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>).
     * @param version the plugin version identifier
     * @return the plugin descriptor, or <code>null</code>
     */
    IPluginDescriptor getPluginDescriptor(String pluginId, Version version);

	/**
	 * Returns a plugin descriptor with the given plugin identifier and version that
	 * statisfies the specified match condition in this plugin registry, or <code>null</code> if
     * there is no such plugin.
     * 
	 * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>).
	 * @param version the plugin version identifier
	 * @param matchCondition 
	 * @return the plugin descriptor, or <code>null</code>
	 * 
	 * @see VersionMatchCondition
	 */
	IPluginDescriptor getPluginDescriptor(String pluginId, Version version, VersionMatchCondition matchCondition);

    /**
     * Returns all plugin descriptors known to this plugin registry.
     * Returns an empty array if there are no installed plugins.
     *
     * @return the plugin descriptors known to this plugin registry
     */
    IPluginDescriptor[] getPluginDescriptors();

    /**
     * Returns all versions of the identified plugin descriptor
     * known to this plugin registry.
     * Returns an empty array if there are no plugins
     * with the specified identifier.
     *
     * @param pluginId the unique identifier of the plugin
     *                (e.g. <code>"gov.va.med.hds.cd.prism"</code>).
     * @return the plugin descriptors known to this plugin registry
     */
    IPluginDescriptor[] getPluginDescriptors(String pluginId);
}
