package gov.va.med.hds.cd.runtime;

import gov.va.med.hds.cd.mum.Version;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;

/**
 * The <code>IRuntimeContext</code> is an object which can be used to locate
 * other objects managed by the runtime.  In particular, instances of classes
 * provided by plugins are accessible via the <code>IRuntimeContext</code>.
 * Additionally, it provides access to the plugin registry, and to global
 * settings for the application.
 * 
 * This interface is not intended to be implemented by clients.
 */
public interface IRuntimeContext extends ILoggingCapable {
	/**
	 * Returns the set of global application settings.
	 * @return IConfigurationElement
	 */
	IConfigurationElement[] getSettings();
	/**
	 * Returns the plugin registry.
	 * @return the plugin registry.
	 * @see IPluginRegistry
	 */
	IPluginRegistry	getPluginRegistry();
	/**
	 * Returns whether the plugin specified has been activated. 
	 * Invoking this method will not cause the plugin to be activated.
	 * 
	 * @param pluginId the id of the plugin
	 * @return <code>true</code> if the plugin is activated, and
	 *   <code>false</code> otherwise
	 */
	boolean isPluginActive(String pluginId);
	/**
	 * Returns whether the plugin specified has been activated. 
	 * Invoking this method will not cause the plugin to be activated.
	 * 
	 * @param pluginId the id of the plugin
	 * @param version, the version of the plugin
	 * @return <code>true</code> if the plugin is activated, and
	 *   <code>false</code> otherwise
	 */
	boolean isPluginActive(String pluginId, Version version);
	/**
	 * Returns the active instance of the latest version of a plugin with the
	 * specified id.
	 * 
	 * To get a specific version of a plugin instance call 
	 * <code>getPluginInstance(String, Version)</code>.
	 * 
	 * If the greatest versioned plugin is not yet active, it is instantiated
	 * and <code>IPlugin.startup()</code> is called on the newly created
	 * instance.  If the specified plugin has prerequisites declared in its
	 * descriptor, these are also instantiated and 
	 * <code>IPlugin.startup()</code> is invoked on each prerequisite plugin
	 * before the initial plugin.  The idea is that plugins are instantiated 
	 * and started up according to the dependency graph declared in the plugin
	 * registry.
	 * 
	 * @param pluginId the id of the plugin
	 * @return the latest version of an instance of a plugin if active, or
	 *   activation was successful, and <code>null</code> otherwise
	 * 
	 * @throws PluginStartupException - if an error occurs during plugin startup.
	 * @throws PluginActivatingException - if the requested plugin is in the 
	 * process of starting up.
	 * @throws PluginNotFoundException - if the specified pluginId cannot
	 * be found.
	 * 
	 * @see IRuntimeContext#getPluginInstance(String, Version)
	 * @see IPlugin#startup(IRuntimeContext, IPluginDescriptor)
	 */
	Object getPluginInstance(String pluginId) throws PluginStartupException, PluginNotFoundException;
	
	/**
	 * Returns the active instance of the specified plugin.
	 * 
	 * If the specified plugin is not yet active, it is instantiated
	 * and <code>IPlugin.startup()</code> is called on the newly created
	 * instance.  If the specified plugin has prerequisites declared in its
	 * descriptor, these are also instantiated and 
	 * <code>IPlugin.startup()</code> is invoked on each prerequisite plugin
	 * before the initially specified plugin.  The idea is that plugins are
	 * instantiated and started up according to the dependency graph
	 * declared in the plugin registry.
	 * 
	 * @param pluginId the id of the plugin
	 * @param version the version of the plugin
	 * @return an instance of the specified plugin if active, or
	 * activation was successful, and <code>null</code> otherwise
	 * 
	 * @throws PluginStartupException - if an error occurs during plugin startup.
	 * @throws PluginActivatingException - if the requested plugin is in the 
	 * process of starting up.
	 * @throws PluginNotFoundException - if the specified pluginId cannot
	 * be found.
	 */
	Object getPluginInstance(String pluginId, Version version) throws PluginStartupException, PluginNotFoundException;

	/**
	 * Adds a listener to the runtime that is notified each time a 
	 * plugin is activated or deactivated.
	 *
	 * @param l the <code>IPluginActivationListener</code> to be added
	 */
	void addPluginActivationListener(IPluginActivationListener l);
	
	/**
	 * Removes a listener from being notified by the runtime of plugin activation
	 * events. 
	 * 
	 * @param l the <code>IPluginActivationListener</code> to be removed
	 */
	void removePluginActivationListener(IPluginActivationListener l);
}
