package gov.va.med.hds.cd.runtime.logging;

/**
 * A <code>Logger</code> object is used to log messages for a specific
 * system or application component. Loggers are normally named, using a
 * hierarchical dot-separated namespace. Logger names can be arbitrary strings,
 * but they should normally be based on the package name or class name of the
 * logged component, such as <code>"gov.va.med.hds.cd.runtime"</code> or 
 * <code>"gov.va.med.hds.cd.prism"</code>.
 */
public interface ILogger {
	/**
	 * Log a FATAL message.
	 * 
	 * If the logger is currently enabled for the FATAL message level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param message the message to log.
	 */
	void fatal(String message);
	/**
	 * Log a FATAL exception.
	 * 
	 * If the logger is currently enabled for the FATAL message level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 * 
	 * @param t the exception to log, including its stack trace.
	 */
	void fatal(String message, Throwable t);
	/**
	 * Log an ERROR message.
	 * 
	 * If the logger is currently enabled for the ERROR message level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param message the message to log.
	 */
	void error(String message);
	/**
	 * Log an ERROR exception.
	 * 
	 * If the logger is currently enabled for the ERROR message level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 * 
	 * @param t the exception to log, including its stack trace.
	 */
	void error(String message, Throwable t);
	/**
	 * Log a WARNING message.
	 * 
	 * If the logger is currently enabled for the WARNING message level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param message the message to log.
	 */
	void warning(String message);
	/**
	 * Log an exception at the WARNING level.
	 * 
	 * If the logger is currently enabled for the WARNING message level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 * 
	 * @param t the exception to log, including its stack trace.
	 */
	void warning(String message, Throwable t);
	/**
	 * Log an INFO message.
	 * 
	 * If the logger is currently enabled for the INFO message level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param message the message to log.
	 */
	void info(String message);
	/**
	 * Log an exception at the INFO level.
	 * 
	 * If the logger is currently enabled for the INFO message level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 * 
	 * @param t the exception to log, including its stack trace.
	 */
	void info(String message, Throwable t);
	/**
	 * Log a DEBUG message.
	 * 
	 * If the logger is currently enabled for the DEBUG message level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param message the message to log.
	 */
	void debug(String message);
	/**
	 * Log an exception at the DEBUG level.
	 * 
	 * If the logger is currently enabled for the DEBUG message level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 * 
	 * @param t the exception to log, including its stack trace.
	 */
	void debug(String message, Throwable t);
	/**
	 * 
	 * @return true if output is to be sent to the logger's parent
	 */
	boolean getUseParentHandlers();
	/**
	 * Log a message at the specified level.
	 * 
	 * If the logger is currently enabled for the specified level then the
	 * given message is forwarded to all the registered output Handler objects. 
	 * 
	 * @param level the level at which to log the message.
	 * @param message the message to log.
	 */
	void log(Level level, String message);

	/**
	 * Log an exception at the specified level.
	 * 
	 * If the logger is currently enabled for the specified level then the
	 * given exception's stack trace is forwarded to all the registered output
	 * Handler objects. 
	 *
	 * @param level the level at which to log the exception.
	 * @param t the exception to log, including its stack trace.
	 */
	void log(Level level, String message, Throwable t);
	/**
	 * Add a log IHandler to receive logging messages.
	 * 
	 * By default, ILoggers also send their output to their parent logger. Typically
	 * the root ILogger is configured with a set of Handlers that essentially act as
	 * default handlers for all loggers.
	 * 
	 * @param logHandler a logging IHandler
	 */
	void addHandler(IHandler logHandler);
	/**
	 * Remove a log IHandler.
	 * 
	 * Returns silently if the given IHandler is not found. 
	 * 
	 * @param logHandler a logging IHandler
	 */
	void removeHandler(IHandler logHandler);
	/**
	 * Specify whether or not this logger should send its output to it's parent ILogger.
	 * This means that any ILogRecords will also be written to the parent's IHandlers, 
	 * and potentially to its parent, recursively up the namespace.
	 * @param useParentHandlers
	 */
	void setUseParentHandlers(boolean useParentHandlers);
}
