package gov.va.med.hds.cd.runtime.logging;

import java.io.Serializable;

/**
 * The Level class defines a set of standard logging levels that can be used to control logging output. 
 * 
 * The logging Level objects are ordered and are specified by ordered integers. Enabling logging at a 
 * given level also enables logging at all higher levels
 */
public class Level implements Serializable {
	
	private static final int FATAL_INT 	= 1000;
	private static final int ERROR_INT 	= 900;
	private static final int WARNING_INT	= 800;
	private static final int INFO_INT 	= 700;
	private static final int DEBUG_INT 	= 600;

	/**
	 * The <code>FATAL</code> level designates very severe error events
	 * that will presumably lead the application to abort.
	 */
	public static final Level FATAL = new Level("FATAL", FATAL_INT);
	
	/**
	 * The <code>ERROR</code> level designates error events that might
	 * still allow the application to continue running.
	 */
	public static final Level ERROR = new Level("ERROR", ERROR_INT);
	
	/**
	 * The <code>WARNING</code> level designates potentially harmful
	 * situations.
	 */
	public static final Level WARNING = new Level("WARNING", WARNING_INT);
	
	/**
	 * The <code>INFO</code> level designates informational messages
	 * that highlight the progress of the application at coarse-grained level.
	 */
	public static final Level INFO = new Level("INFO", INFO_INT);
	
	/**
	 * The </code>DEBUG</code> lvel designates fine-grained informational
	 * events that are most useful to debug an application.
	 */
	public static final Level DEBUG = new Level("DEBUG", DEBUG_INT);
	
	private String name;
	private int value;
	
	/**
	 * Return the non-localized string name of the Level
	 * 
	 * @return non-localized name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Return the integer value for this level. 
	 * This integer value can be used for efficient ordering comparisons between
	 * Level objects
	 * 
	 * @return the integer value for this level
	 */
	public int toInt() {
		return value;
	}
	
	/**
	 * Create a named Level with a given integer value.
	 * 
	 * @param name the name of the Level, for example "SEVERE"
	 * @param value an integer value for the level
	 */
	protected Level(String name, int value) {
		this.name = name;
		this.value = value;
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		if (!(obj instanceof Level)) { return false; }
		Level l = (Level) obj;
		return toInt() == l.toInt();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return toInt();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return getName();
	}
}
