package gov.va.med.hds.cd.security;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IPluginRegistry;
import gov.va.med.hds.cd.runtime.IRuntimeContext;

import java.security.PrivilegedAction;
import java.util.Iterator;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

/**
 * Implementation of the ISecurity interface based on the Java SDK JAAS security architecture.
 */
public class Security implements ISecurity, gov.va.med.hds.cd.runtime.IPlugin {
	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#startup(IRuntimeContext, IPluginDescriptor)
	 */
	public void startup(IRuntimeContext runtime, IPluginDescriptor pluginDescriptor)
		throws Exception {
		IPluginRegistry pluginRegistry = runtime.getPluginRegistry();
		IConfigurationElement[] pluginExtensionConfigs =
			pluginRegistry.getConfigurationElementsFor(PLUGIN_EXTENSION_NAME);

		if (pluginExtensionConfigs.length != 1)
			throw new PluginException(
				"The Security plugin requires one and only one security plugin extension.\n There were "
					+ pluginExtensionConfigs.length
					+ " defined");

		IConfigurationElement pluginConfig = pluginExtensionConfigs[0];

		if (!pluginConfig.getName().equals(PLUGIN_ELEMENT_NAME))
			throw new PluginException(
				"The extension for "
					+ PLUGIN_EXTENSION_NAME
					+ " must have a "
					+ PLUGIN_ELEMENT_NAME
					+ " tag");

		IPluginFactory factory;
		factory = (IPluginFactory) pluginConfig.createExecutableExtension(CLASS_ATTRIBUTE_NAME);
		if (factory == null)
			throw new Exception(
				"the Security plugin plugin extension requires a valid \""
					+ CLASS_ATTRIBUTE_NAME
					+ "\" name. "
					+ "Be sure the classFactory type is included in the classpath as well.");
		plugin = factory.create(runtime);

		try {
			Configuration.setConfiguration(plugin.getConfiguration());
		} catch (PluginException e) {
			throw e;
		}
		try {
			loginContext = plugin.createLoginContext();
		} catch (PluginException e) {
			throw e;
		}
	}
	/**
	 * @see gov.va.med.hds.cd.runtime.IPlugin#shutdown()
	 */
	public void shutdown() throws Exception {
		logout();
	}
	/**
	 * @see gov.va.med.hds.cd.security.ISecurity#login()
	 */
	public IUserInfo login() throws LoginException {
		loginContext.login();
		loggedIn = true;
		return getUserInfo();
	}
	/**
	 * @see gov.va.med.hds.cd.security.ISecurity#logout()
	 */
	public void logout() throws LoginException {
		if (loggedIn) {
			try {
				loginContext.logout();
			} finally {
				loggedIn = false;
			}
		}
	}
	/**
	 * @see gov.va.med.hds.cd.security.ISecurity#getSubject()
	 */
	public Subject getSubject() {
		return loginContext.getSubject();
	}
	/**
	 * @see gov.va.med.hds.cd.security.ISecurity#getUserInfo()
	 */
	public IUserInfo getUserInfo() throws CouldNotCreateUserInfoException {
		Subject subject = getSubject();
		Set principals = subject.getPrincipals(IUserInfo.class);
		if (principals.size() == 0)
			throw new CouldNotCreateUserInfoException(
				"No Principal exists in the Subject that implements " + IUserInfo.class.getName());
		if (principals.size() > 1)
			throw new CouldNotCreateUserInfoException(
				"More than one Principal in the Subject implements "
					+ IUserInfo.class.getName()
					+ ". User info is ambiguous.");
		Iterator it = principals.iterator();
		return (IUserInfo) it.next();
	}
	private IPlugin plugin = null;
	private LoginContext loginContext = null;
	private boolean loggedIn = false;

	/* (non-Javadoc)
	 * @see gov.va.med.hds.cd.security.ISecurity#doAs(PrivilegedAction)
	 */
	public void doAs(PrivilegedAction action) {
		Subject.doAs(getSubject(), action);
	}

}
