package gov.va.med.hds.cd.session.internal;

import javax.security.auth.login.LoginException;
import javax.swing.event.EventListenerList;

import gov.va.med.hds.cd.mum.boot.ExceptionDialog;
import gov.va.med.hds.cd.mum.boot.ISplashScreen;
import gov.va.med.hds.cd.mum.boot.IStatusNotifier;
import gov.va.med.hds.cd.runtime.IMainPlugin;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntime;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.security.AuthenticationFailedException;
import gov.va.med.hds.cd.security.ConnectionFailedException;
import gov.va.med.hds.cd.security.ISecurity;
import gov.va.med.hds.cd.security.OptionValidationException;
import gov.va.med.hds.cd.security.TooManyInvalidLoginAttemptsException;
import gov.va.med.hds.cd.security.UserCancelledAuthenticationException;
import gov.va.med.hds.cd.session.ISession;
import gov.va.med.hds.cd.session.ISessionListener;
import gov.va.med.hds.cd.session.ISessionPlugin;
import gov.va.med.hds.cd.session.IUserDescriptor;

public class SessionPlugin implements ISessionPlugin, IMainPlugin, ISession {

	private static final String UNABLE_TO_SIGN_OFF = "unable to sign off";
	private static final String UNABLE_TO_SIGNON = "unable to sign on";
	private static final String ERROR_DIALOG_TITLE = "VistA Sign-on Error";
	private static final String DEFAULT_MAIN_PLUGIN_ID =
		"gov.va.med.hds.cd.prism";

	private EventListenerList listenerList = new EventListenerList();
	private ILogger log;
	private IRuntimeContext runtimeContext;
	private boolean signedOn = false;
	private ISecurity securityPlugin = null;

	public void addSessionListener(ISessionListener l) {
		listenerList.add(ISessionListener.class, l);
	}

	public void removeSessionListener(ISessionListener l) {
		listenerList.remove(ISessionListener.class, l);
	}

	public IUserDescriptor getUser() {
		return new UserDescriptor();
	}

	public ISession getSession() {
		if (isSignedOn())
			return this;
		return null;
	}

	private boolean isSignedOn() {
		return signedOn;
	}

	public void run(IRuntime runtime, ISplashScreen splash, String[] args) {
		updateSplash(splash, "Signing on to VistA");
		signOn();
		if (!isSignedOn()) {
			splash.close();
			runtime.shutdown();
			return;
		}
		updateSplash(splash, "Launching Desktop");
		try {
			IMainPlugin mainPlugin =
				(IMainPlugin) runtimeContext.getPluginInstance(
					getMainPluginId());
			mainPlugin.run(runtime, splash, args);
		} catch (Exception e) {
			log.fatal("unable to get reference to main plugin ''{0}''", e);
		}
	}

	public void startup(IRuntimeContext runtime, IPluginDescriptor plugin)
		throws Exception {
		this.runtimeContext = runtime;
		this.log = runtimeContext.getLogger(getClass());
	}

	public void shutdown() throws Exception {
		if (isSignedOn())
			signOff();

	}

	private String getMainPluginId() {
		return DEFAULT_MAIN_PLUGIN_ID;
	}

	private void signOn() {
		try {
			securityPlugin =
				(ISecurity) runtimeContext.getPluginInstance(
					ISecurity.PLUGIN_ID);
		} catch (Exception e) {
			log.fatal("unable to get reference to seurity plugin", e);
			return;
		}
		try {
			securityPlugin.login();
			signedOn = true;
		} catch (AuthenticationFailedException e) {
			notifyUser(e);
			log.info(UNABLE_TO_SIGNON, e);
			signedOn = false;
		} catch (OptionValidationException e) {
			notifyUser(e);
		} catch (UserCancelledAuthenticationException e) {
			// NOOP
		} catch (TooManyInvalidLoginAttemptsException e) {
			notifyUser(e);
		} catch (ConnectionFailedException e) {
			showErrorDialog(e);
			log.fatal(UNABLE_TO_SIGNON, e);
		} catch (LoginException e) {
			showErrorDialog(e);
			log.fatal(UNABLE_TO_SIGNON, e);
		}
	}

	private void showErrorDialog(Exception e) {
		ExceptionDialog.showExceptionDialog(ERROR_DIALOG_TITLE, e.getMessage(), e);
	}

	private void notifyUser(Exception e) {
		ExceptionDialog.showExceptionDialog(ERROR_DIALOG_TITLE, e.getMessage(), e);
	}

	private void signOff() {
		if (!isSignedOn())
			return;
		try {
			ISecurity securityPlugin =
				(ISecurity) runtimeContext.getPluginInstance(
					ISecurity.PLUGIN_ID);
			securityPlugin.logout();
		} catch (Exception ex) {
			log.fatal(UNABLE_TO_SIGN_OFF, ex);
		}
	}

	private void updateSplash(ISplashScreen splash, String message) {
		log.info(message);
		if (!(splash instanceof IStatusNotifier))
			return;
		IStatusNotifier notifier = (IStatusNotifier) splash;
		notifier.displayMessage(message);
	}
}
