package gov.va.med.hds.cd.spectrum.test;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;

import javax.swing.JLabel;

import junit.framework.TestCase;

import gov.va.med.hds.cd.spectrum.date.BasicDateComboBoxEditor;
import gov.va.med.hds.cd.spectrum.date.DateComboBox;
import gov.va.med.hds.cd.spectrum.date.DefaultDateSelectionModel;
import gov.va.med.hds.cd.spectrum.date.IDateComboBoxEditor;
import gov.va.med.hds.cd.spectrum.date.IDateSelectionModel;
import gov.va.med.hds.util.CalendarMath;

public class DateComboBoxTest extends TestCase {

	private ActionEvent actionEvent;
	private PropertyChangeEvent propertyChangeEvent;

	public DateComboBoxTest(String name) {
		super(name);
	}

	private void assertInitialState(DateComboBox combo) {
		assertTrue(combo.isEditable());
		assertTrue(combo.isEnabled());
		assertFalse(combo.isPopupVisible());
		assertTrue(combo.getEditor() instanceof BasicDateComboBoxEditor);
	}

	public void testDefaultConstructor() {
		DateComboBox combo = new DateComboBox();
		assertTrue(CalendarMath.isSameDate(Calendar.getInstance(), combo.getSelectedCalendar()));
		assertInitialState(combo);
	}

	public void testCalendarConstructor() {
		Calendar initialDate = new GregorianCalendar(2005, 3, 5);
		DateComboBox combo = new DateComboBox(initialDate);
		assertEquals(initialDate, combo.getSelectedCalendar());
		assertInitialState(combo);
	}

	public void testDateConstructor() {
		Calendar initialDate = new GregorianCalendar(1997, 11, 2);
		DateComboBox combo = new DateComboBox(initialDate.getTime());
		assertEquals(initialDate, combo.getSelectedCalendar());
		assertInitialState(combo);
	}

	public void testModelConstructor() {
		Calendar initialDate = new GregorianCalendar(1903, 7, 4);
		IDateSelectionModel model = new DefaultDateSelectionModel(initialDate);
		DateComboBox combo = new DateComboBox(model);
		assertEquals(model, combo.getSelectionModel());
		assertEquals(initialDate, combo.getSelectedCalendar());
		assertInitialState(combo);
	}

	public void testActionEvent() {
		DateComboBox combo = new DateComboBox();
		combo.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				actionEvent = e;
			}
		});
		assertNull(actionEvent);
		combo.setSelectedDate(new GregorianCalendar(2005, 3, 5));
		assertNotNull(actionEvent);
	}

	public void testSelectionModelPropertyChange() {
		DateComboBox combo = new DateComboBox();
		combo.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
				propertyChangeEvent = e;
			}
		});
		assertNull(propertyChangeEvent);
		IDateSelectionModel oldModel = combo.getSelectionModel();
		IDateSelectionModel newModel = new DefaultDateSelectionModel();
		combo.setSelectionModel(newModel);
		assertNotNull(propertyChangeEvent);
		assertEquals(combo, propertyChangeEvent.getSource());
		assertEquals("selectionModel", propertyChangeEvent.getPropertyName());
		assertEquals(oldModel, propertyChangeEvent.getOldValue());
		assertEquals(newModel, propertyChangeEvent.getNewValue());
		assertEquals(newModel, combo.getSelectionModel());
	}

	public void testEditablePropertyChange() {
		DateComboBox combo = new DateComboBox();
		combo.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
				propertyChangeEvent = e;
			}
		});
		assertNull(propertyChangeEvent);
		boolean oldEditableValue = combo.isEditable();
		boolean newEditableValue = !oldEditableValue;
		combo.setEditable(newEditableValue);
		assertNotNull(propertyChangeEvent);
		assertEquals(combo, propertyChangeEvent.getSource());
		assertEquals("editable", propertyChangeEvent.getPropertyName());
		assertEquals(Boolean.valueOf(oldEditableValue), propertyChangeEvent.getOldValue());
		assertEquals(Boolean.valueOf(newEditableValue), propertyChangeEvent.getNewValue());
		assertEquals(newEditableValue, combo.isEditable());
	}

	public void testEnabledPropertyChange() {
		DateComboBox combo = new DateComboBox();
		combo.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
				propertyChangeEvent = e;
			}
		});
		assertNull(propertyChangeEvent);
		boolean oldEnabledValue = combo.isEnabled();
		boolean newEnabledValue = !oldEnabledValue;
		combo.setEnabled(newEnabledValue);
		assertNotNull(propertyChangeEvent);
		assertEquals(combo, propertyChangeEvent.getSource());
		assertEquals("enabled", propertyChangeEvent.getPropertyName());
		assertEquals(Boolean.valueOf(oldEnabledValue), propertyChangeEvent.getOldValue());
		assertEquals(Boolean.valueOf(newEnabledValue), propertyChangeEvent.getNewValue());
		assertEquals(newEnabledValue, combo.isEnabled());
	}
	
	public void testEditorUpdate() {
		DateComboBox combo = new DateComboBox();
		MockDateComboBoxEditor editor = new MockDateComboBoxEditor();
		combo.setEditor(editor);
		assertEquals(combo.getSelectedCalendar(), editor.getDate());
		combo.setSelectedDate(new GregorianCalendar(1975, Calendar.JULY, 23));
		assertEquals(combo.getSelectedCalendar(), editor.getDate());
	}
}

class MockDateComboBoxEditor implements IDateComboBoxEditor {

	private JLabel label = new JLabel();
	private Calendar date = null;

	public void addActionListener(ActionListener l) {	
	}
	
	public void removeActionListener(ActionListener l) {
	}

	public Component getEditorComponent() {
		return label;
	}

	public Calendar getDate() {
		return date;
	}

	public void setDate(Calendar date) {
		this.date = date;
	}

	
	public DateFormat getDateFormat() {
		return null;
	}


	public void setDateFormat(DateFormat format) {
	}

	public void selectAll() {
	}
}
