package gov.va.med.hds.cd.spectrum;

import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Enumeration;

import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.TableColumnModelEvent;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;


/**
 * Department of Veterans Affairs
 * OI Field Office - Salt Lake City
 * Health Data Systems
 *
 * The EditableTableHeader class allows adding editor component to
 * a column of the table header.  The editor is activated when the mouse
 * is over the header column and deactivated when mouse exits the editor 
 * component.  The value of the the header becomes the value of the toString()
 * method of the editor component.  The cells of the table header are not
 * reorderable (they cannot be moved or removed).
 *  
 * @author vhaislschras
 */
public class EditableTableHeader extends JTableHeader implements ComponentListener
{
  public static final String CELL_EDITOR_KEY = "gov.va.med.hds.TableHeaderEditor";
  protected int hasEditors = 0;
  
  /**
   * Constructor for EditableTableHeader.
   */
  public EditableTableHeader()
  {
    super();
    super.setReorderingAllowed(false);
    addComponentListener(this);
    addMouseListener(new EditableTableHeaderMouseAdapter());
  }

  /**
   * Constructor for EditableTableHeader.
   * @param cm The ColumnTableModel to use for the header and the
   * table it represents.
   */
  public EditableTableHeader(TableColumnModel cm)
  {
    super(cm);
    super.setReorderingAllowed(false);
    addComponentListener(this);
    addMouseListener(new EditableTableHeaderMouseAdapter());
  }

  /**
   * Sets the editor for the column header.
   * @param editor a JComponent to use as the header editor.  The value
   * the component's toString() method is used for the column header value
   * when the editor component is not active.
   * 
   * @param column the column to attach this editor to.
   */
  public void setHeaderEditor(JComponent editor, int column)
  {

    // Attach the editor to the column
    TableColumn tc = getColumnModel().getColumn(column);
    tc.setIdentifier(editor);
    tc.setHeaderValue(editor.toString());
    // Attach the column to the editor
    editor.putClientProperty(CELL_EDITOR_KEY, tc);

    // Add the editor to the JTableHeader container
    add(editor);
    
    // Add the editor as a listener to mouse input so
    // it can hide itself when the mouse exits
    editor.addMouseListener(new EditableTableHeaderMouseAdapter());
    
    // Size the editor to the column
    refactorEditor(editor, column);
    
    hasEditors++;
  }

  /**
   * Removes the current editor from the specified column.
   * @param column the column from which the editor is removed.
   */
  public void removeHeaderEditor(int column)
  {
    if (null != getHeaderEditor(column))
      getColumnModel().getColumn(column).setIdentifier(null);
  }
  
  /**
   * Returns the editor for the specified column's header.
   * 
   * @param column the column from which to return the editor.
   * @return If the column has an editor, the JComponent that is the editor
   * otherwise null.
   */
  public JComponent getHeaderEditor(int column)
  {
    return getHeaderEditor(getColumnModel().getColumn(column));
  }
  
  private JComponent getHeaderEditor(TableColumn tc)
  {
    Object editor = tc.getIdentifier();
    return (JComponent)(editor instanceof JComponent ? editor : null);
  }
  
  private void refactorEditors()
  {
    if (0 < hasEditors)
    {
      Enumeration tableColumns = getColumnModel().getColumns();
      while (tableColumns.hasMoreElements())
      {
        TableColumn tc = (TableColumn)tableColumns.nextElement();
        JComponent editor = getHeaderEditor(tc);
        if (null != editor)
          refactorEditor(editor, getColumnModel().getColumnIndex(editor));
      }
    }
  }

  private void refactorEditor(JComponent editor, int column)
  {
    if (isShowing())
    {
      TableColumnModel tcm = getColumnModel();
      
      // This will be used to determine the position of the editor
      Rectangle headerBounds = getBounds();
      int top = (int)headerBounds.getY();
      int height = (int)headerBounds.getHeight();
      int width = tcm.getColumn(column).getWidth() - 2;

      // Find the location of this column
      int left = 0;
      int i = -1;
      while ((++i) < column)
        left += tcm.getColumn(i).getWidth();
      
      // Set the editor to the correct coordinates
      Rectangle editorBounds = new Rectangle(left, top, width, height);
      editor.setBounds(editorBounds);
      editor.repaint(editorBounds);
    }
  }

  /**
   * Called when the EditiableTableHeader component is hidden.
   * @param e the ComponentEvent object
   * @see java.awt.event.ComponentListener#componentHidden(ComponentEvent)
   */
  public void componentHidden(ComponentEvent e)
  {
    if (0 < hasEditors)
    {
      Enumeration tableColumns = getColumnModel().getColumns();
      while (tableColumns.hasMoreElements())
      {
        TableColumn tc = (TableColumn)tableColumns.nextElement();
        JComponent editor = getHeaderEditor(tc);
        if (null != editor && editor.isShowing())
          editor.setVisible(false);
      }
    }
  }

  /**
   * Called when the EditiableTableHeader component is moved.
   * @param e the ComponentEvent object
   * @see java.awt.event.ComponentListener#componentMoved(ComponentEvent)
   */
  public void componentMoved(ComponentEvent e)
  {
    refactorEditors();
  }

  /**
   * Called when the EditiableTableHeader component is resized.
   * @param e the ComponentEvent object
   * @see java.awt.event.ComponentListener#componentResized(ComponentEvent)
   */
  public void componentResized(ComponentEvent e)
  {
    refactorEditors();
  }

  /**
   * Called when the EditiableTableHeader component is shown.
   * @param e the ComponentEvent object
   * @see java.awt.event.ComponentListener#componentShown(ComponentEvent)
   */
  public void componentShown(ComponentEvent e)
  {
    if (0 < hasEditors)
    {
      Enumeration tableColumns = getColumnModel().getColumns();
      while (tableColumns.hasMoreElements())
      {
        TableColumn tc = (TableColumn)tableColumns.nextElement();
        JComponent editor = getHeaderEditor(tc);
        if (null != editor && editor.isVisible())
          editor.setVisible(true);
      }
    }
  }

  /**
   * Called when a new column is added to the EditiableTableHeader component.
   * @param e the TableColumnModelEvent object
   * @see javax.swing.event.TableColumnModelListener#columnAdded(TableColumnModelEvent)
   */
  public void columnAdded(TableColumnModelEvent e)
  {
    super.columnAdded(e);
    refactorEditors();
  }

  /**
   * Called when a column margin is changed in the EditiableTableHeader component.
   * @param e the TableColumnModelEvent object
   * @see javax.swing.event.TableColumnModelListener#columnMarginChanged(ChangeEvent)
   */
  public void columnMarginChanged(ChangeEvent e)
  {
    super.columnMarginChanged(e);
    refactorEditors();
  }

  /**
   * Called when a new column in the EditiableTableHeader component is moved.  This should
   * never happen.
   * @param e the TableColumnModelEvent object
   * @see javax.swing.event.TableColumnModelListener#columnMoved(TableColumnModelEvent)
   */
  public void columnMoved(TableColumnModelEvent e)
  {
    super.columnMoved(e);
    refactorEditors();
  }

  /**
   * Called when a new column in the EditiableTableHeader component is removed.  This should
   * never happen.
   * @param e the TableColumnModelEvent object
   * @see javax.swing.event.TableColumnModelListener#columnRemoved(TableColumnModelEvent)
   */
  public void columnRemoved(TableColumnModelEvent e)
  {
    super.columnRemoved(e);
    if (null != getResizingColumn())
      refactorEditors();
  }

  /**
   * Overriden from JTableHeader to prevent reordering.
   * @param reorderingAllowed regardeless of this value reordering will be disallowed.
   * 
   * @see javax.swing.table.JTableHeader#setReorderingAllowed(boolean)
   */
  public void setReorderingAllowed(boolean reorderingAllowed)
  {
    // never allow this to be reordered
    super.setReorderingAllowed(false);
  }

}

class EditableTableHeaderMouseAdapter extends MouseAdapter
{
  /* (non-Javadoc)
   * @see java.awt.event.MouseListener#mouseEntered(MouseEvent)
   */
  public void mouseEntered(MouseEvent e)
  {
    super.mouseEntered(e);
    if (e.getSource() instanceof EditableTableHeader)
    {
      EditableTableHeader source = (EditableTableHeader)e.getSource();
      int colIndex = source.getColumnModel().getColumnIndexAtX(e.getX());
      JComponent editor = source.getHeaderEditor(colIndex);
      if (null != editor)
        editor.setVisible(true);
    }
    else if (e.getSource() instanceof JComponent)
    {
      JComponent source = (JComponent)e.getSource();
      source.requestFocusInWindow();
    }
  }

  /* (non-Javadoc)
   * @see java.awt.event.MouseListener#mouseExited(MouseEvent)
   */
  public void mouseExited(MouseEvent e)
  {
    super.mouseExited(e);
    if (e.getSource() instanceof JComponent && !(e.getSource() instanceof EditableTableHeader))
    {
      JComponent source = (JComponent)e.getSource();
      TableColumn tc = (TableColumn)source.getClientProperty(EditableTableHeader.CELL_EDITOR_KEY);
      if (null != tc)
        tc.setHeaderValue(source.toString());
//      if (!(source instanceof JComboBox))
//        source.setVisible(false);
//      // JComboBoxes need additional handling.
//      else
//      {
//        if (!(((JComboBox)source).isPopupVisible()))
//          source.setVisible(false);
//        // JComboBoxes normally consume low lever mouse events.
//        // If it got this far it's because the mouse event handler for
//        // the combobox was overridden and the event was re-dispatched.
//        e.consume();
//      }
    }
  }

}
