/*
        ListTree.java

        $Revision: 1.3 $
        $Date: 2001/06/29 21:26:47 $

        Copyright (c) 2000 Big Idea Interactive.  All rights reserved.
*/
package gov.va.med.hds.cd.spectrum;

import java.awt.Component;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractListModel;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListSelectionModel;
import javax.swing.Icon;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.tree.TreeModel;

public class ListTree extends JPanel {

    protected TreeModel         model;
    protected List              mLists;
    protected JList             mRootList;
    protected LSL               mSelectionListener;

    public ListTree(TreeModel model) {
        super();
        this.model = model;
        setLayout(new BoxLayout(this, BoxLayout.X_AXIS));

        mLists = new ArrayList();
        mSelectionListener = new LSL();

        addList(createListPanel(model.getRoot()));
    }

    public Object getSelectedValue() {
        return getList(mLists.size() - 1).getList().getSelectedValue();
    }

    protected void addList(ListPanel l) {
        if (l != null) {
            l.getList().addListSelectionListener(mSelectionListener);
            mLists.add(l);
            add(l);
            refresh();
        }
    }

    protected void removeList(int index) {
        removeList(getList(index));
    }

    protected void removeList(ListPanel l) {
        if (l != null) {
            remove(l);
            mLists.remove(l);
            refresh();
        }
    }

    protected ListPanel getList(int index) {
        if (mLists.size() > 0) {
            if (index >= 0 && index < mLists.size()) {
                return (ListPanel) mLists.get(index);
            }
        }

        return null;
    }

    protected int indexOf(JList l) {
        for (int i = mLists.size() - 1; i >=0; i--) {
            ListPanel p = (ListPanel) mLists.get(i);
            if (p.getList().equals(l)) {
                return i;
            }
        }

        return -1;
    }

    private void removeAfter(int index) {
        if (mLists.size() > 0) {
            if (index >= 0 && index < mLists.size()) {
                for (int i = mLists.size() - 1; i > index; i--) {
                    ListPanel l = (ListPanel) mLists.get(i);
                    remove(l);
                    mLists.remove(l);
                }
                refresh();
            }
        }
    }

    protected ListPanel createListPanel(Object node) {
        return new ListPanel(model, node);
    }

    protected void refresh() {
        revalidate();
        repaint();
    }

    public static class ListPanel extends JScrollPane {

        private static CellRenderer sRenderer;
        private JList mList;
        private TreeModel model;

        public ListPanel(TreeModel model, Object node) {
            super();
            this.model = model;
            sRenderer = new CellRenderer(model);
            mList = new JList(new ListTreeModelAdapter(model, node));
            mList.setSelectionModel(new SingleSelectionModel());
            mList.setCellRenderer(sRenderer);
            setViewportView(mList);
        }

        public JList getList() {
            return mList;
        }

        public Object getSelection() {
            return mList.getSelectedValue();
        }

        public void setNode(Object node) {
            mList.setModel(new ListTreeModelAdapter(model, node));
            revalidate();
        }

        public Dimension getMaximumSize() {
            //return new Dimension(100, Integer.MAX_VALUE);
            return new Dimension(getPreferredSize().width + 16, super.getMaximumSize().height);
        }

        public Dimension getMinimumSize() {
            //return new Dimension(100, 0);
            return new Dimension(getPreferredSize().width + 16, super.getMinimumSize().height);
        }

        protected class SingleSelectionModel extends DefaultListSelectionModel {

            protected boolean mGestureStarted = false;

            public SingleSelectionModel() {
                setSelectionMode(SINGLE_SELECTION);
            }

            public void setValueIsAdjusting(boolean isAdjusting) {
                if (isAdjusting == false) {
                    mGestureStarted = false;
                }
            }


            public void setSelectionInterval(int index0, int index1) {
                if (!mGestureStarted) {
                    int oldIndex = getMinSelectionIndex();
                    super.setSelectionInterval(index0, index1);
                    int newIndex = getMinSelectionIndex();
                    if (oldIndex != newIndex) {
                        updateSingleSelection(oldIndex, newIndex);
                    }
                }
                mGestureStarted = true;
            }

            public void updateSingleSelection(int oldIndex, int newIndex) {
            }
        }

        private static class CellRenderer extends DefaultListCellRenderer {

            transient protected Icon mLeafIcon;
            transient protected Icon mClosedIcon;
            transient protected Icon mOpenIcon;
            private TreeModel model;

            public CellRenderer(TreeModel model) {
                super();
                this.model = model;

                mLeafIcon = UIManager.getIcon("Tree.leafIcon");
                mClosedIcon = UIManager.getIcon("Tree.closedIcon");
                mOpenIcon = UIManager.getIcon("Tree.openIcon");
            }

            public Component getListCellRendererComponent(
                                                          JList list,
                                                          Object value,
                                                          int index,
                                                          boolean isSelected,
                                                          boolean cellHasFocus) {
                super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

                int numChildren = model.getChildCount(value);
                if (numChildren <= 0) {
                    //System.out.println("leaf");
                    setIcon(mLeafIcon);
                } else {
                    if (isSelected) {
                        setIcon(mOpenIcon);
                    } else {
                        setIcon(mClosedIcon);
                    }
                }

                return this;
            }
        }
    }

    class LSL implements ListSelectionListener {
        public void valueChanged(ListSelectionEvent e) {
            if (!e.getValueIsAdjusting()) {
                Object source = e.getSource();
                int listNum = indexOf((JList) source);
                if (listNum == -1) return;
                                // System.out.println("selection in list " + listNum);

                Object node = getList(listNum).getSelection();
                if (node == null) return;

                ListPanel l = getList(listNum + 1);
                if (l == null) {
                    if (model.getChildCount(node) > 0) {
                        l = createListPanel(node);
                        addList(l);
                    }
                } else {
                    if (model.getChildCount(node) > 0) {
                        l.setNode(node);
                        removeAfter(listNum + 1);
                    } else {
                        removeAfter(listNum);
                    }
                }
            }
        }
    }
}

class ListTreeModelAdapter extends AbstractListModel {

    private TreeModel model;
    private Object mNode;

    public ListTreeModelAdapter(TreeModel model, Object node) {
        super();
        this.model = model;
        this.mNode = node;
    }

    public Object getElementAt(int index) {
        return model.getChild(mNode, index);
    }

    public int getSize() {
        return model.getChildCount(mNode);
    }
}
