package gov.va.med.hds.cd.spectrum;
import java.awt.Component;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.UIManager;

import gov.va.med.hds.cd.spectrum.plaf.MultiSplitPaneUI;
import gov.va.med.hds.cd.spectrum.plaf.basic.BasicMultiSplitPaneUI;
import gov.va.med.hds.cd.spectrum.plaf.metal.MetalMultiSplitPaneUI;
import gov.va.med.hds.cd.spectrum.plaf.windows.WindowsMultiSplitPaneUI;

public class MultiSplitPane extends JComponent {

	/**
	 * @see #getUIClassID
	 * @see #readObject
	 */
	private static final String uiClassID = "MultiSplitPaneUI";

	/**
	 * Vertical split indicates the <code>Component</code>s are
	 * split along the y axis.
	 */
	public final static int VERTICAL_SPLIT = 0;

	/**
	 * Horizontal split indicates the <code>Component</code>s are
	 * split along the x axis.
	 */
	public final static int HORIZONTAL_SPLIT = 1;

	/**
	 * Used to add a <code>Component</code> that will represent a divider.
	 */
	public final static String DIVIDER = "divider";

	/**
	 * Bound property name for orientation (horizontal or vertical).
	 */
	public final static String ORIENTATION_PROPERTY = "orientation";

	/**
	 * Bound property name for border.
	 */
	public final static String DIVIDER_SIZE_PROPERTY = "dividerSize";

	public final static double DEFAULT_RATIO = -1.0;

	/**
	 * How the views are split.
	 */
	protected int orientation;

	/**
	 * Size of the dividers.
	 */
	protected int dividerSize;

	protected List components = new ArrayList();

	protected List dividers = new ArrayList();

	protected double ratios[] = new double[0];

	public MultiSplitPane(int orientation) {
		super();
		if (orientation != HORIZONTAL_SPLIT && orientation != VERTICAL_SPLIT)
			throw new IllegalArgumentException(
				"cannot create MultiSplitPane, "
					+ "orientation must be one of "
					+ "MultiSplitPane.HORIZONTAL_SPLIT "
					+ "or MultiSplitPane.VERTICAL_SPLIT");
		this.orientation = orientation;
		updateUI();
	}

	/**
	 * Sets the L&F object that renders this component.
	 *
	 * @param ui  the <code>MultiSplitPaneUI</code> L&F object
	 * @see UIDefaults#getUI
	 * @beaninfo
	 *        bound: true
	 *       hidden: true
	 *    attribute: visualUpdate true
	 *  description: The UI object that implements the Component's LookAndFeel. 
	 */
	public void setUI(MultiSplitPaneUI ui) {
		if ((MultiSplitPaneUI) this.ui != ui) {
			super.setUI(ui);
			revalidate();
		}
	}

	/**
	 * Returns the <code>SplitPaneUI</code> that is providing the
	 * current look and feel. 
	 *
	 * @return the <code>SplitPaneUI</code> object that renders this component
	 * @beaninfo
	 *       expert: true
	 *  description: The L&F object that renders this component.
	 */
	public MultiSplitPaneUI getUI() {
		return (MultiSplitPaneUI) ui;
	}

	/**
	 * Notification from the <code>UIManager</code> that the L&F has changed. 
	 * Replaces the current UI object with the latest version from the 
	 * <code>UIManager</code>.
	 *
	 * @see JComponent#updateUI
	 */
	public void updateUI() {
		if (UIManager
			.getLookAndFeel()
			.getClass()
			.equals(javax.swing.plaf.metal.MetalLookAndFeel.class)) {
			UIManager.put(uiClassID, MetalMultiSplitPaneUI.class.getName());
		} else if (
			com.sun.java.swing.plaf.windows.WindowsLookAndFeel.class.isAssignableFrom(UIManager.getLookAndFeel().getClass())) {
			UIManager.put(uiClassID, WindowsMultiSplitPaneUI.class.getName());
		} else {
			UIManager.put(uiClassID, BasicMultiSplitPaneUI.class.getName());
		}
		setUI((MultiSplitPaneUI) UIManager.getUI(this));
//		revalidate();
	}

	/**
	 * Returns the name of the L&F class that renders this component.
	 *
	 * @return the string "MultiSplitPaneUI"
	 * @see JComponent#getUIClassID
	 * @see UIDefaults#getUI
	 * @beaninfo
	 *       expert: true
	 *  description: A string that specifies the name of the L&F class.
	 */
	public String getUIClassID() {
		return uiClassID;
	}

	/**
	 * Sets the size of the divider.
	 *
	 * @param newSize an integer giving the size of the divider in pixels
	 * @beaninfo
	 *        bound: true
	 *  description: The size of the divider.
	 */
	public void setDividerSize(int newSize) {
		int oldSize = this.dividerSize;

		if (oldSize != newSize) {
			this.dividerSize = newSize;
			firePropertyChange(DIVIDER_SIZE_PROPERTY, oldSize, newSize);
		}
	}

	/**
	 * Returns the size of the divider.
	 *
	 * @return an integer giving the size of the divider in pixels
	 */
	public int getDividerSize() {
		return dividerSize;
	}

	/**
	 * Sets the orientation, or how the splitter is divided. The options
	 * are:<ul>
	 * <li>MultiSplitPane.VERTICAL_SPLIT  (above/below orientation of components)
	 * <li>MultiSplitPane.HORIZONTAL_SPLIT  (left/right orientation of components)
	 * </ul>
	 *
	 * @param orientation an integer specifying the orientation
	 * @exception IllegalArgumentException if orientation is not one of:
	 *        HORIZONTAL_SPLIT or VERTICAL_SPLIT.
	 * @beaninfo
	 *        bound: true
	 *  description: The orientation, or how the splitter is divided.
	 *         enum: HORIZONTAL_SPLIT JSplitPane.HORIZONTAL_SPLIT
	 *               VERTICAL_SPLIT   JSplitPane.VERTICAL_SPLIT
	 */
	public void setOrientation(int orientation) {
		if ((orientation != VERTICAL_SPLIT)
			&& (orientation != HORIZONTAL_SPLIT)) {
			throw new IllegalArgumentException(
				"MultiSplitPane: orientation must "
					+ "be one of "
					+ "MultiSplitPane.VERTICAL_SPLIT or "
					+ "MultiSplitPane.HORIZONTAL_SPLIT");
		}

		int oldOrientation = this.orientation;

		this.orientation = orientation;
		firePropertyChange(ORIENTATION_PROPERTY, oldOrientation, orientation);
	}

	/**
	 * Returns the orientation.
	 * 
	 * @return an integer giving the orientation
	 * @see #setOrientation
	 */
	public int getOrientation() {
		return orientation;
	}

	/** 
	 * Returns true, so that calls to <code>revalidate</code>
	 * on any descendant of this <code>MultiSplitPane</code>
	 * will cause a request to be queued that
	 * will validate the <code>MultiSplitPane</code> and all its descendants.
	 * 
	 * @return true
	 * @see JComponent#revalidate
	 * 
	 * @beaninfo
	 *    hidden: true
	 */
	public boolean isValidateRoot() {
		return true;
	}

	public void setDividerLocation(int divider, int location) {
		MultiSplitPaneUI ui = getUI();
		if (ui != null) {
			ui.setDividerLocation(this, divider, location);
		}
	}

	/**
	 * Subclassed to message the UI with <code>finishedPaintingChildren</code>
	 * after super has been messaged, as well as painting the border.
	 *
	 * @param g the <code>Graphics</code> context within which to paint
	 */
	protected void paintChildren(Graphics g) {
		super.paintChildren(g);

		MultiSplitPaneUI ui = getUI();

		if (ui != null) {
			Graphics tempG = createSwingGraphics(g);
			ui.finishedPaintingChildren(this, tempG);
			tempG.dispose();
		}
	}

	private static Graphics createSwingGraphics(Graphics g) {
		if (g == null) {
			Thread.dumpStack();
			return null;
		}
		return g.create();
	}

	protected void addImpl(Component comp, Object constraints, int index) {
		if (constraints != null && constraints.equals(DIVIDER)) {
			dividers.add(comp);
			System.out.println();
		} else {
			addDivider();
			components.add(comp);
			addRatioField();
		}
		super.addImpl(comp, constraints, index);
	}

	protected void addRatioField() {
		double[] newRatios = new double[ratios.length + 1];
		System.arraycopy(ratios, 0, newRatios, 0, ratios.length);
		newRatios[ratios.length] = DEFAULT_RATIO;
		ratios = newRatios;
	}

	private void addDivider() {
		if (components.size() >= 1) {
			MultiSplitPaneUI ui = getUI();
			if (ui != null) {
				ui.addDivider(this);
			}
		}
	}

	public Component getComponent(int n) {
		if (n >= 0 && n < components.size())
			return (Component) components.get(n);
		else
			return null;
	}

	public Component getComponentFromAll(int n) {
		return super.getComponent(n);
	}

	public int getComponentCount() {
		return components.size();
	}

	public int getAllComponentCount() {
		return super.getComponentCount();
	}

	public Component getDivider(int n) {
		if (n >= 0 && n < dividers.size())
			return (Component) dividers.get(n);
		else
			return null;
	}

	public int getDividerCount() {
		return dividers.size();
	}

	public Component[] getComponents() {
		Component[] returnVal = new Component[components.size()];
		components.toArray(returnVal);
		return returnVal;
	}

	public Component[] getAllComponents() {
		return super.getComponents();
	}

	public Component[] getDividers() {
		Component[] returnVal = new Component[dividers.size()];
		dividers.toArray(returnVal);
		return returnVal;
	}

	public double[] getRatios() {
		double[] returnVal = new double[ratios.length];
		System.arraycopy(ratios, 0, returnVal, 0, ratios.length);
		return returnVal;
	}

	public Component getComponentPrecedingDivider(int divider) {
		if (getComponentCount() == 0)
			return null;
		if (getComponentCount() == 1)
			return (Component) components.get(0);
		if (divider < 0)
			return (Component) components.get(0);
		if (divider >= dividers.size())
			return (Component) components.get(components.size() - 1);
		return (Component) components.get(divider);
	}

	public Component getComponentSucceedingDivider(int divider) {
		if (getComponentCount() == 0)
			return null;
		if (getComponentCount() == 1)
			return (Component) components.get(0);
		if (divider < 0)
			return (Component) components.get(0);
		if (divider >= dividers.size())
			return (Component) components.get(components.size() - 1);
		return (Component) components.get(divider + 1);
	}

	public void setRatios(double[] ratios) {
		if (ratios.length != this.ratios.length)
			throw new IllegalArgumentException();
		normalizeRatios(ratios);
		System.arraycopy(ratios, 0, this.ratios, 0, ratios.length);
	}

	private boolean areRatiosValid(double[] ratios) {
		double total = 0.0;
		for (int i = 0; i < ratios.length; ++i) {
			total += ratios[i];
		}
		return total == 1.0;
	}

	private void normalizeRatios(double[] ratios) {
		double length = 0.0;
		for (int i = 0; i < ratios.length; ++i) {
			length += ratios[i];
		}
		if (length > 0.0) {
			for (int i = 0; i < ratios.length; ++i) {
				ratios[i] = ratios[i] / length;
			}
		} else {
			for (int i = 0; i < ratios.length; ++i) {
				ratios[i] = 1.0 / ratios.length;
			}
		}
	}
	
	public void remove(Component comp) {
		if (components.contains(comp)) components.remove(comp);
		if (dividers.contains(comp)) dividers.remove(comp);
		super.remove(comp);
	}

	public void remove(int index) {
		super.remove(index);
	}

	public void removeAll() {
		dividers.clear();
		components.clear();
		super.removeAll();
	}
	
	public void removeDividers() {
		for (Iterator i = dividers.iterator(); i.hasNext();) {
			Component c = (Component) i.next();
			super.remove(c);
		}
		dividers.clear();
		revalidate();
	}

}
