package gov.va.med.hds.cd.spectrum.bucket;
import java.awt.Dimension;
import java.awt.event.ActionListener;

import javax.swing.Action;
import javax.swing.JComponent;
/**
 * Basic slush bucket interface.  An implementation provides default buttons and fires events when buttons
 * are pressed, and users can call the enabled methods to make the buttons available or unavailable.  Or, 
 * users can simply set the button actions and the bucket components and work with the actions and models
 * independently.
 */
public interface ISlushBucket {
	public static final int SELECT_ACTION = 1;
	public static final int SELECT_ALL_ACTION = 2;
	public static final int REMOVE_ACTION = 3;
	public static final int REMOVE_ALL_ACTION = 4;

	/**
	 * Programmatically fires one of the event actions.
	 * @param actionId
	 */
	void fireAction(int actionId);
	
	/**
	 * Sets the enabled state of the "Remove All" button.
	 * @param enabled
	 */
	void setRemoveAllButtonEnabled(boolean enabled);
	
	/**
	 * Sets the enabled state of the "Remove" button.
	 * @param enabled
	 */
	void setRemoveButtonEnabled(boolean enabled);
	
	/**
	 * Sets the enabled state of the "Select All" button.
	 * @param enabled
	 */
	void setSelectAllButtonEnabled(boolean enabled);

	/**
	 * Sets the enabled state of the "Select" button.
	 * @param enabled
	 */
	void setSelectButtonEnabled(boolean enabled);

	/**
	 * Sets the given component as the "in bucket" component.
	 * @param c
	 */
	void setInBucketComponent(JComponent c);
	
	/**
	 * Same as setInBucketComponent(JComponent c), but also sets the mnemonic for the
	 * label for this component.
	 * @param c
	 * @param mnemonic
	 */
	void setInBucketComponent(JComponent c, char mnemonic);
	
	/**
	 * Sets the given component as the "out bucket" component.
	 * @param c
	 */
	void setOutBucketComponent(JComponent c);
		
	/**
	 * Same as setOutBucketComponent(JComponent c), but also sets the mnemonic for the
	 * label for this component.
	 * @param c
	 * @param mnemonic
	 */
	void setOutBucketComponent(JComponent c, char mnemonic);
	
	/**
	 * Sets the label associated with the in bucket.
	 * @param labelText
	 */
	void setInBucketLabel(String labelText);
	
	/**
	 * Sets the label associated with the out bucket.
	 * @param labelText
	 */
	void setOutBucketLabel(String labelText);
	
	/**
	 * Sets the mnemonic for the in bucket label.
	 * @param mnemonic
	 */
	void setInBucketMnemonic(char mnemonic);
	
	/**
	 * Sets the mnemonic for the out bucket label.
	 * @param mnemonic
	 */
	void setOutBucketMnemonic(char mnemonic);
	
	/**
	 * Adds an action listener for button events.
	 * @param listener
	 */
	void addActionListener(ActionListener listener);
	
	/**
	 * Removes the listener for button events.
	 * @param listener
	 */
	void removeActionListener(ActionListener listener);
	
	/**
	 * Set the preferred dimensions for this component.
	 * @param dim
	 */
	void setPreferredBucketDimension(Dimension dim);
	
	/**
	 * Optional method that assigns an action to the "select" button.
	 * @param action
	 */
	void setSelectAction(Action action);
	
	/**
	 * Optional method that assigns an action to the "select all" button.
	 * @param action
	 */
	void setSelectAllAction(Action action);
	
	/**
	 * Optional method that assigns an action to the "remove" button.
	 * @param action
	 */
	void setRemoveAction(Action action);
	
	/**
	 * Optional method that assigns an action to the "remove all" button.
	 * @param action
	 */
	void setRemoveAllAction(Action action);
	
}