package gov.va.med.hds.cd.spectrum.date;

import java.text.SimpleDateFormat;
import java.util.Calendar;

import javax.swing.ListModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.table.AbstractTableModel;

import gov.va.med.hds.util.CalendarMath;

public class CalendarTableModel extends AbstractTableModel {

	private static final int CALENDAR_ROWS = 6;
	private static final int CALENDAR_COLUMNS = 7;

	private static String[] columnHeaders;

	private Calendar anchorDate;
	private ListModel monthModel;
	private ListModel previousMonthModel;
	private ListModel nextMonthModel;
	
	private ListDataListener monthChangeListener = new MonthChangeListener();

	/**
	 * TODO_SB: Provide description for "CalendarTableModel" constructor
	 * @param initialDate
	 */
	public CalendarTableModel(Calendar anchorDate) {
		setAnchorDate(anchorDate);
	}
	
	public void setAnchorDate(Calendar anchorDate) {
		Calendar oldAnchor = this.anchorDate;
		this.anchorDate = (Calendar) anchorDate.clone();
		if (!CalendarMath.isSameMonth(oldAnchor, anchorDate))
		setMonthModel(new MonthListModel(anchorDate));
	}
	
	public Calendar getAnchorDate() {
		return anchorDate;
	}
	
	private void configureNextMonthModel() {
		Calendar firstDayOfNextMonth = getLastDayOfMonth();
		firstDayOfNextMonth.add(Calendar.DATE, 1);
		this.nextMonthModel = new MonthListModel(firstDayOfNextMonth);
	}

	private void configurePreviousMonthModel() {
		Calendar lastDayOfPreviousMonth = getFirstDayOfMonth();
		lastDayOfPreviousMonth.add(Calendar.DATE, -1);
		this.previousMonthModel = new MonthListModel(lastDayOfPreviousMonth);
	}

	private void setMonthModel(ListModel monthModel) {
		if (this.monthModel != null) {
			this.monthModel.removeListDataListener(monthChangeListener);
		}
		this.monthModel = monthModel;
		this.monthModel.addListDataListener(monthChangeListener);
		
		configurePreviousMonthModel();
		configureNextMonthModel();
		configureColumnHeaders();
		
		fireTableDataChanged();
	}

	private ListModel getMonthModel() {
		return monthModel;
	}

	public int getRowCount() {
		return CALENDAR_ROWS;
	}
	public int getColumnCount() {
		return CALENDAR_COLUMNS;
	}

	public String getColumnName(int columnIndex) {
		return columnHeaders[columnIndex];
	}
	public Class getColumnClass(int columnIndex) {
		return Calendar.class;
	}

	public boolean isCellEditable(int rowIndex, int columnIndex) {
		return false;
	}

	public Object getValueAt(int rowIndex, int columnIndex) {
		int index = getArrayIndex(rowIndex, columnIndex);
		if (isInCurrentMonth(index)) {
			return monthModel.getElementAt(getModelIndex(index));
		} else if (isInPreviousMonth(index)) {
			return previousMonthModel.getElementAt(getPreviousMonthIndex(index));
		} else if (isInNextMonth(index)) {
			return nextMonthModel.getElementAt(getNextMonthIndex(index));
		}
		throw new ArrayIndexOutOfBoundsException();
	}
	
	private int getRowIndex(int modelIndex) {
		int arrayIndex = modelIndex + getMonthStartIndex();
		return arrayIndex / CALENDAR_COLUMNS;
	}
	
	private int getColumnIndex(int modelIndex) {
		int arrayIndex = modelIndex + getMonthStartIndex();
		return arrayIndex % CALENDAR_COLUMNS;
	}

	private boolean isToday(int rowIndex, int columnIndex) {
		return CalendarMath.isSameDate(Calendar.getInstance(), (Calendar) getValueAt(rowIndex, columnIndex));
	}

	private boolean isInCurrentMonth(int rowIndex, int columnIndex) {
		return isInCurrentMonth(getArrayIndex(rowIndex, columnIndex));
	}

	private boolean isInPreviousMonth(int rowIndex, int columnIndex) {
		return isInPreviousMonth(getArrayIndex(rowIndex, columnIndex));
	}

	private boolean isInNextMonth(int rowIndex, int columnIndex) {
		return isInNextMonth(getArrayIndex(rowIndex, columnIndex));
	}

	private boolean isInCurrentMonth(int arrayIndex) {
		return arrayIndex >= getMonthStartIndex()
			&& arrayIndex < getMonthStartIndex() + getDaysInMonth();
	}

	private boolean isInPreviousMonth(int arrayIndex) {
		return arrayIndex >= 0 && arrayIndex < getMonthStartIndex();
	}

	private boolean isInNextMonth(int arrayIndex) {
		return arrayIndex >= getMonthStartIndex() + getDaysInMonth() && arrayIndex < CALENDAR_ROWS * CALENDAR_COLUMNS;
	}

	private void configureColumnHeaders() {
		if (columnHeaders == null) {
			columnHeaders = new String[CALENDAR_COLUMNS];

			Calendar c = Calendar.getInstance();
			int firstDayOfWeek = c.getFirstDayOfWeek();
			c.set(Calendar.DATE, 1);
			while (c.get(Calendar.DAY_OF_WEEK) != firstDayOfWeek)
				c.roll(Calendar.DATE, 1);

			SimpleDateFormat formatter = new SimpleDateFormat("EEE");
			for (int i = 0; i < CALENDAR_COLUMNS; i++) {
				columnHeaders[i] = formatter.format(c.getTime());
				c.roll(Calendar.DATE, 1);
			}
		}
	}

	private int getArrayIndex(int rowIndex, int columnIndex) {
		return (rowIndex * CALENDAR_COLUMNS) + columnIndex;
	}

	private int getModelIndex(int row, int col) {
		return getModelIndex(getArrayIndex(row, col));
	}

	private int getModelIndex(int arrayIndex) {
		int modelIndex = arrayIndex - getMonthStartIndex();
		if (modelIndex < 0) return -1;
		if (modelIndex >= getMonthModel().getSize()) return -1;
		return modelIndex;
	}

	private int getNextMonthIndex(int index) {
		return index - (getMonthStartIndex() + getDaysInMonth());
	}

	private int getPreviousMonthIndex(int index) {
		return previousMonthModel.getSize() - getMonthStartIndex() + index;
	}

	private int getMonthStartIndex() {
		return getFirstWeekDayOfMonth();
	}

	private int getDaysInMonth() {
		return monthModel.getSize();
	}

	private int getFirstWeekDayOfMonth() {
		Calendar firstDayOfMonth = getFirstDayOfMonth();
		int firstWeekDayOfMonth =
			firstDayOfMonth.get(Calendar.DAY_OF_WEEK) - firstDayOfMonth.getFirstDayOfWeek();
		if (firstWeekDayOfMonth == -1)
			firstWeekDayOfMonth = Calendar.FRIDAY;
		return firstWeekDayOfMonth;
	}

	private Calendar getFirstDayOfMonth() {
		return (Calendar) ((Calendar) monthModel.getElementAt(0)).clone();
	}
	
	private Calendar getLastDayOfMonth() {
		return (Calendar) ((Calendar) monthModel.getElementAt(monthModel.getSize() - 1)).clone();
	}
	
	private class MonthChangeListener implements ListDataListener {
		public void intervalAdded(ListDataEvent e) {
			fireTableDataChanged();
		}
		
		public void intervalRemoved(ListDataEvent e) {
			fireTableDataChanged();
		}
		
		public void contentsChanged(ListDataEvent e) {
			fireTableDataChanged();
		}
	}
}