package gov.va.med.hds.cd.spectrum.date;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.HeadlessException;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.ListSelectionModel;
import javax.swing.SpinnerDateModel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.BevelBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import gov.va.med.hds.util.CalendarMath;

public class DateChooser extends JPanel {
	private static final String NEXT_MONTH = "forward.gif";
	private static final String PREVIOUS_MONTH = "backward.gif";

	private static final String NONE_SELECTED = "No date selected";

	public static final String SELECTION_MODEL_PROPERTY = "selectionModel";

	private IDateSelectionModel selectionModel;

	private JButton previousMonthButton;
	private JButton nextMonthButton;
	private JLabel monthLabel;
	private CalendarTable calendarTable;

	private SpinnerDateModel monthSpinner = new SpinnerDateModel();
	private Object selectionUpdateSource = null;

	private SimpleDateFormat monthFormat = new SimpleDateFormat("MMMM dd, yyyy");

	/**
	 * Shows a modal date-chooser dialog and blocks until the
	 * dialog is hidden.  If the user presses the "OK" button, then
	 * this method hides/disposes the dialog and returns the selected date.
	 * If the user presses the "Cancel" button or closes the dialog without
	 * pressing "OK", then this method hides/disposes the dialog and returns
	 * <code>null</code>.
	 *
	 * @param parent the parent <code>Component</code> for the dialog
	 * @param title the String containing the dialog's title
	 * @param initialDate the initial date set when the date-chooser is shown
	 * @return the selected date or <code>null</code> if the user opted out
	 * @exception HeadlessException if GraphicsEnvironment.isHeadless()
	 * returns true.
	 * @see java.awt.GraphicsEnvironment#isHeadless
	 */
	public static Calendar showDialog(Component parent, String title, Calendar initialDate) throws HeadlessException {
		final DateChooser chooser = new DateChooser(initialDate != null ? initialDate : Calendar.getInstance());
		DateTracker ok = new DateTracker(chooser);
		JDialog dialog = createDialog(parent, title, true, chooser, ok, null);
		dialog.addWindowListener(new DateChooserDialog.Closer());
		dialog.addComponentListener(new DateChooserDialog.DisposeOnClose());

		dialog.show(); // blocks until user brings dialog down...
		
		return ok.getDate();
	}

	/**
	 * Creates and returns a new dialog containing the specified
	 * <code>DateChooser</code> pane along with "OK", "Cancel", and "Reset"
	 * buttons. If the "OK" or "Cancel" buttons are pressed, the dialog is
	 * automatically hidden (but not disposed).  If the "Reset"
	 * button is pressed, the date-chooser's date will be reset to the
	 * date which was set the last time <code>show</code> was invoked on the
	 * dialog and the dialog will remain showing.
	 *
	 * @param c the parent component for the dialog
	 * @param title the title for the dialog
	 * @param modal a boolean. When true, the remainder of the program
	 *              is inactive until the dialog is closed.
	 * @param chooserPane the date-chooser to be placed inside the dialog
	 * @param okListener the ActionListener invoked when "OK" is pressed
	 * @param cancelListener the ActionListener invoked when "Cancel" is pressed
	 * @return a new dialog containing the date-chooser pane
	 * @exception HeadlessException if GraphicsEnvironment.isHeadless()
	 * returns true.
	 * @see java.awt.GraphicsEnvironment#isHeadless
	 */
	public static JDialog createDialog(Component c, String title, boolean modal,
		DateChooser chooserPane, ActionListener okListener,
		ActionListener cancelListener) throws HeadlessException {
		return new DateChooserDialog(c, title, modal, chooserPane,
												okListener, cancelListener);
	}
		
	/**
	 * Creates a date chooser pane with an initial date of today.
	 */
	public DateChooser() {
		this(Calendar.getInstance());
	}

	/**
	 * Creates a date chooser pane with the specified initial date.
	 *
	 * @param initialColor the initial color set in the chooser
	 */
	public DateChooser(Calendar initialDate) {
		this(new DefaultDateSelectionModel(initialDate));
	}

	/**
	 * Creates a date chooser pane with the specified
	 * <code>IDateSelectionModel</code>.
	 *
	 * @param model the <code>IDateSelectionModel</code> to be used
	 */
	public DateChooser(IDateSelectionModel model) {
		this.selectionModel = model;
		configure();
	}	

	/**
	 * Gets the current date value from the date chooser.
	 * By default, this delegates to the model.
	 *
	 * @return the current date value of the date chooser
	 */
	public Calendar getDate() {
		return getSelectionModel().getSelectedDate();
	}

	public void setDate(Date date) {
		Calendar c = Calendar.getInstance();
		c.setTime(date);
		setDate(c);
	}

	/**
	 * Sets the current date of the date chooser to the specified date.
	 * The <code>IDateSelectionModel</code> will fire a <code>ChangeEvent</code>
	 * @param date the date to be set in the color chooser
	 * @see JComponent#addPropertyChangeListener
	 *
	 * @beaninfo
	 *       bound: false
	 *      hidden: false
	 * description: The current color the chooser is to display.
	 */
	public void setDate(Calendar date) {
		getSelectionModel().setSelectedDate(date);
	}

	public IDateSelectionModel getSelectionModel() {
		return selectionModel;
	}
	   
	/**
	 * Sets the model containing the selected color.
	 *
	 * @param model the new <code>IDateSelectionModel</code> object
	 *
	 * @beaninfo
	 *       bound: true
	 *      hidden: true
	 * description: The model which contains the currently selected date.
	 */
	public void setSelectionModel(IDateSelectionModel model) {
		IDateSelectionModel oldModel = this.selectionModel;
		this.selectionModel = model;
		firePropertyChange(SELECTION_MODEL_PROPERTY, oldModel, model);
	}

	private int locationToIndex(Point p) {
		Point tablePoint = SwingUtilities.convertPoint(this, p, calendarTable);
		int row = calendarTable.rowAtPoint(tablePoint);
		int col = calendarTable.columnAtPoint(tablePoint);
		if (row == -1 || col == -1)
			return -1;
		return -1;
//		return calendarTable.getModelIndex(row, col);
	}

	private void configure() {
		configureMonthSpinner();
		configureMainPanel();
		configureButtonSizes();
		configureTableSelection();
		configureSelectionModel();
	}

	private void configureTableSelection() {
		calendarTable.setSelectedDate(getDate());
		ListSelectionListener l = new TableSelectionListener();
		calendarTable.getSelectionModel().addListSelectionListener(l);
		calendarTable.getColumnModel().getSelectionModel().addListSelectionListener(l);
	}

	private void configureSelectionModel() {
		getSelectionModel().addChangeListener(new SelectionListenerLabel());
		getSelectionModel().addChangeListener(new SelectionListenerButtons());
		getSelectionModel().addChangeListener(new SelectionListenerTable());
		getSelectionModel().addChangeListener(new SelectionListenerSpinner());
	}

	private void configureMainPanel() {
		setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
		setLayout(new BorderLayout());
		add(createNorthPane(), BorderLayout.NORTH);
		add(createCenterPane(), BorderLayout.CENTER);
	}

	private void configureButtonSizes() {
		int columnWidth = calendarTable.getColumnModel().getColumn(0).getMinWidth();
		columnWidth =
			columnWidth - nextMonthButton.getInsets().left - nextMonthButton.getInsets().right - 2;
		Dimension preferredSize = nextMonthButton.getPreferredSize();
		preferredSize.width = columnWidth;
		previousMonthButton.setMinimumSize(preferredSize);
		nextMonthButton.setMinimumSize(preferredSize);
		nextMonthButton.setPreferredSize(preferredSize);
		previousMonthButton.setPreferredSize(preferredSize);
	}

	private JComponent createNorthPane() {
		previousMonthButton = new JButton();
		previousMonthButton.setAction(new PreviousMonthAction());

		nextMonthButton = new JButton();
		nextMonthButton.setAction(new NextMonthAction());

		GridBagLayout layout = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		c.gridwidth = c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;

		JPanel p = new JPanel();
		p.setLayout(layout);

		c.gridy = c.gridx = 0;
		c.weightx = 1.0;
		layout.setConstraints(previousMonthButton, c);
		p.add(previousMonthButton);

		this.monthLabel = createDateLabel();
		c.gridx = 1;
		c.gridwidth = 5;
		c.weightx = 5.0;
		layout.setConstraints(monthLabel, c);
		p.add(monthLabel);

		c.gridx = 6;
		c.gridwidth = 1;
		c.weightx = 1.03;
		layout.setConstraints(nextMonthButton, c);
		p.add(nextMonthButton);
		return p;
	}

	private JComponent createCenterPane() {
		JPanel p = new JPanel();
		p.setLayout(new BorderLayout());
		calendarTable = new CalendarTable(getDate());
		calendarTable.addMouseListener(new CalendarTableMouseListener());
		calendarTable.addMouseMotionListener(new CalendarTableMouseMotionListener());
		p.add(calendarTable.getTableHeader(), BorderLayout.NORTH);
		p.add(calendarTable, BorderLayout.CENTER);
		return p;
	}

	private void configureMonthSpinner() {
		monthSpinner.setValue(getSelectionModel().getSelectedDate().getTime());
		monthSpinner.setCalendarField(Calendar.MONTH);
		monthSpinner.addChangeListener(new MonthSpinnerListener());
	}

	private JLabel createDateLabel() {
		JLabel l = new JLabel();
		l.setOpaque(true);
		l.setFont(getFont().deriveFont(Font.BOLD));
		l.setHorizontalAlignment(SwingConstants.CENTER);
		l.setText(getDateText(getDate()));
		return l;
	}

	private void setSelectionUpdateSource(Object source) {
		this.selectionUpdateSource = source;
	}

	private boolean isUpdatingSelection() {
		return this.selectionUpdateSource != null;
	}

	private Object getSelectionUpdateSource() {
		return this.selectionUpdateSource;
	}

	public void clearSelection() {
		getSelectionModel().setSelectedDate(null);
	}

	private String getDateText(Calendar date) {
		if (date == null) return NONE_SELECTED;
		return monthFormat.format(date.getTime());
	}

	private class NextMonthAction extends AbstractAction implements ListSelectionListener {

		public NextMonthAction() {
			super();
			putValue(
				Action.SMALL_ICON,
				new ImageIcon(NextMonthAction.class.getResource(NEXT_MONTH)));
		}

		public void actionPerformed(ActionEvent e) {
			monthSpinner.setValue(monthSpinner.getNextValue());
		}

		public void valueChanged(ListSelectionEvent e) {
			if (e.getValueIsAdjusting())
				return;
			setEnabled(getDate() != null);
		}
	}

	private class PreviousMonthAction extends AbstractAction implements ListSelectionListener {

		public PreviousMonthAction() {
			super();
			putValue(
				Action.SMALL_ICON,
				new ImageIcon(PreviousMonthAction.class.getResource(PREVIOUS_MONTH)));
		}

		public void actionPerformed(ActionEvent e) {
			monthSpinner.setValue(monthSpinner.getPreviousValue());
		}

		public void valueChanged(ListSelectionEvent e) {
			if (e.getValueIsAdjusting())
				return;
			setEnabled(getDate() != null);
		}
	}

	private class TableSelectionListener implements ListSelectionListener {
		public void valueChanged(ListSelectionEvent e) {
			if (e.getValueIsAdjusting()
				|| ((ListSelectionModel) e.getSource()).isSelectionEmpty()
				|| isUpdatingSelection())
				return;
				
			Calendar date = calendarTable.getSelectedDate();
			setSelectionUpdateSource(calendarTable);
			if (date == null || CalendarMath.isSameMonth(calendarTable.getAnchorDate(), date)) {
				setDate(date);
				setSelectionUpdateSource(null);
			} else {
				setSelectionUpdateSource(null);
				setDate(date);
			}
		}
	}

	private class MonthSpinnerListener implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			if (isUpdatingSelection())
				return;
			setSelectionUpdateSource(monthSpinner);
			Calendar newMonth = Calendar.getInstance();
			newMonth.setTime(monthSpinner.getDate());
			setDate(newMonth);
			setSelectionUpdateSource(null);
		}
	}

	private class SelectionListenerTable implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			if (calendarTable.equals(getSelectionUpdateSource()))
				return;
			calendarTable.setSelectedDate(getDate());
		}
	}

	private class SelectionListenerSpinner implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			if (monthSpinner.equals(getSelectionUpdateSource()))
				return;
			Calendar selection = getDate();
			if (selection != null) {
				monthSpinner.setValue(selection.getTime());
			}
		}
	}

	private class SelectionListenerLabel implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			monthLabel.setText(getDateText(getDate()));
		}
	}
	
	private class SelectionListenerButtons implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			nextMonthButton.setEnabled(getDate() != null);
			previousMonthButton.setEnabled(getDate() != null);
		}
	}
	
	private class CalendarTableMouseListener implements MouseListener {
		public void mouseClicked(MouseEvent e) {
			processMouseEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
		public void mousePressed(MouseEvent e) {
			processMouseEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
		public void mouseReleased(MouseEvent e) {
			processMouseEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
		public void mouseEntered(MouseEvent e) {
			processMouseEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
		public void mouseExited(MouseEvent e) {
			processMouseEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
	}
	
	private class CalendarTableMouseMotionListener implements MouseMotionListener {
		public void mouseDragged(MouseEvent e) {
			processMouseMotionEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
		public void mouseMoved(MouseEvent e) {
			processMouseMotionEvent(SwingUtilities.convertMouseEvent(calendarTable, e, DateChooser.this));
		}
	}
}

class DateTracker implements ActionListener, Serializable {
	private DateChooser chooser;
	private Calendar date;

	public DateTracker(DateChooser c) {
		chooser = c;
	}

	public void actionPerformed(ActionEvent e) {
		date = chooser.getDate();
	}

	public Calendar getDate() {
		return date;
	}
}
