package gov.va.med.hds.cd.spectrum.dhtml;

import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Shape;
import java.awt.event.ContainerEvent;
import java.awt.event.ContainerListener;

import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.ComponentView;
import javax.swing.text.Element;
import javax.swing.text.Position;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;
import javax.swing.text.html.BlockView;
import javax.swing.text.html.CSS;

class CollapsingDivView extends BlockView {

    private boolean isCollapsed;
    private ParentContainerListener containerListener;

    public CollapsingDivView(Element elem, int axis) {
        super(elem, axis);
        containerListener = new ParentContainerListener();
        updateCollapsedState();
    }

    // Override various methods in BlockView to implement collapsing behavior.

    public boolean isVisible() {
        return !isCollapsed;
    }
    public float getPreferredSpan(int axis) {
        return (isCollapsed ? 0 : super.getPreferredSpan(axis));
    }
    public float getMinimumSpan(int axis) {
        return (isCollapsed ? 0 : super.getMinimumSpan(axis));
    }
    public float getMaximumSpan(int axis) {
        return (isCollapsed ? 0 : super.getMaximumSpan(axis));
    }
    public void paint(Graphics g, Shape allocation) {
        if (!isCollapsed)
            super.paint(g, allocation);
    }
    public Shape modelToView(int pos, Shape a, Position.Bias b)
        throws BadLocationException {
        return (isCollapsed ? a : super.modelToView(pos, a, b));
    }
    public int getNextVisualPositionFrom(
        int pos,
        Position.Bias b,
        Shape a,
        int direction,
        Position.Bias[] biasRet)
        throws BadLocationException {

        if (isCollapsed)
            return getElement().getEndOffset();
        else
            return super.getNextVisualPositionFrom(
                pos,
                b,
                a,
                direction,
                biasRet);
    }

    /** pass parent changes to our listener */
    public void setParent(View parent) {
        containerListener.setParent(parent);
        super.setParent(parent);
    }

    /** Recalculate whether or not we should be collapsed. */
    private void updateCollapsedState() {
        AttributeSet attr = getStyleSheet().getViewAttributes(this);
        Object o = attr.getAttribute(CSS.Attribute.DISPLAY);
        isCollapsed = "none".equalsIgnoreCase(String.valueOf(o));
    }

    /** Potentially update our collapsed state when the document changes */
    public void changedUpdate(DocumentEvent e, Shape a, ViewFactory f) {
        super.changedUpdate(e, a, f);
        boolean wasCollapsed = isCollapsed;
        updateCollapsedState();
        if (isCollapsed != wasCollapsed) {
            //System.out.println("changedUpdate->setChildVisibility");
            resetChildVisibility();
            getParent().preferenceChanged(this, true, true);
        }
    }

    /** Objects of type ComponentView are not painted by our paint method; 
     * instead, they are painted by the parent container.  This means we
     * cannot short circuit their painting with our paint method.
     * Instead, we must explicitly set their visibility flag. */
    private void resetChildVisibility() {
        setChildVisibility(this, !isCollapsed);
    }

    /** Set the visibility of every ComponentView underneath <tt>v</tt>. */
    private void setChildVisibility(View v, boolean visible) {
        for (int i = v.getViewCount(); i-- > 0;) {
            View child = v.getView(i);
            if (child instanceof ComponentView) {
                Component c = ((ComponentView) child).getComponent();
                if (c != null)
                    c.setVisible(visible);
            }
            if (child != null)
                setChildVisibility(child, visible);
        }
    }

    /** ComponentView objects get added underneath us asynchronously. This 
     * object listens for newly added components and resets the visibility
     * flags.
     */
    private class ParentContainerListener implements ContainerListener {

        private Container container = null;

        public void setParent(View parent) {
            Container newContainer = null;
            if (parent != null)
                newContainer = parent.getContainer();
            if (container == newContainer)
                return;

            if (container != null)
                container.removeContainerListener(this);
            if (newContainer != null)
                newContainer.addContainerListener(this);
            container = newContainer;
        }

        public void componentAdded(ContainerEvent e) {
            if (e.getChild() != null)
                SwingUtilities.invokeLater(new VisibilityUpdater());
        }

        public void componentRemoved(ContainerEvent e) {
        }
    }

    /** A simple object for invoking a deferred visibility update */
    private class VisibilityUpdater implements Runnable {
        public VisibilityUpdater() {
            currentUpdater = this;
        }
        public void run() {
            if (currentUpdater == this)
                resetChildVisibility();
        }
    }

    private volatile Object currentUpdater = null;
}