package gov.va.med.hds.cd.spectrum.dhtml;

import java.awt.event.MouseEvent;
import java.util.Enumeration;
import java.util.EventObject;

import javax.swing.JEditorPane;
import javax.swing.text.AttributeSet;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.StyleConstants;
import javax.swing.text.View;
import javax.swing.text.ViewFactory;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.StyleSheet;

import gov.va.med.hds.cd.spectrum.dhtml.scriptlet.IScriptletEvent;
import gov.va.med.hds.cd.spectrum.dhtml.scriptlet.Registry;
import gov.va.med.hds.cd.spectrum.dhtml.scriptlet.ScriptletEnvironment;
import gov.va.med.hds.cd.spectrum.dhtml.scriptlet.ScriptletException;

public class DHTMLEditorKit
    extends HTMLEditorKit
    implements DHTMLDocument.DHTMLEventListener {

    public static final String SCRIPTLET_ENVIRONMENT_PROPERTY =
        "gov.va.med.hds.cd.spectrum.dhtml.scriptletEnvironment";

    private Registry registry;
    private Object envObj;
    private DHTMLPane dhtmlPane;
    private ViewFactory viewFactory;

    private static final boolean SUPPORT_ASYNC = false;

    public DHTMLEditorKit(Registry registry, Object envObj, DHTMLPane pane) {
        this.registry = registry;
        this.envObj = envObj;
        this.dhtmlPane = pane;
        this.viewFactory = new DHTMLViewFactory();
    }

    public ViewFactory getViewFactory() {
        return viewFactory;
    }

    public Document createDefaultDocument() {
        StyleSheet styles = getStyleSheet();
        StyleSheet ss = new StyleSheet();

        ss.addStyleSheet(styles);

        DHTMLDocument doc = new DHTMLDocument(ss);
        doc.setParser(getParser());
        if (SUPPORT_ASYNC) {
            doc.setAsynchronousLoadPriority(4);
            doc.setTokenThreshold(100);
        } else {
            doc.setAsynchronousLoadPriority(-1);
        }
        doc.addDHTMLEventListener(this);
        return doc;
    }

    public static void setAttribute(
        Element element,
        Object attr,
        Object value) {
        Document doc = element.getDocument();
        if (doc instanceof DHTMLDocument) {
            DHTMLDocument myDoc = (DHTMLDocument) doc;
            myDoc.setAttribute(element, attr, value);
        }
    }

    public static Object runScriptlet(
        String command,
        int type,
        Element sourceElement)
        throws ScriptletException {

        if (sourceElement == null)
            return null;
        else
            return runScriptlet(
                command,
                type,
                (DHTMLDocument) sourceElement.getDocument(),
                sourceElement);
    }

    public static Object runScriptlet(
        String command,
        int type,
        DHTMLDocument document,
        Element sourceElement)
        throws ScriptletException {

        if (document == null)
            return null;

        ScriptletEnvironment scriptletEnvironment =
            (ScriptletEnvironment) document.getProperty(
                SCRIPTLET_ENVIRONMENT_PROPERTY);
        if (scriptletEnvironment == null)
            return null;

        return scriptletEnvironment.exec(command, type, sourceElement);
    }

    public void documentLoaded(EventObject event) {
        // create and initialize all the scriptlets named in the document header
        DHTMLDocument doc = (DHTMLDocument) event.getSource();
        if (doc != null) {
            ScriptletEnvironment env =
                new ScriptletEnvironment(registry, doc, dhtmlPane, envObj);
            doc.putProperty(SCRIPTLET_ENVIRONMENT_PROPERTY, env);
        }
        
        // if the <body> tag has an "onload" attribute, run it as a scriptlet
        Element body =
            doc.getElement(
                doc.getDefaultRootElement(),
                StyleConstants.NameAttribute,
                HTML.Tag.BODY);
        if (body != null) { 
            AttributeSet attr = body.getAttributes();
            String onLoad = (String) attr.getAttribute("onload");
            if (onLoad != null && onLoad.length() > 0) 
                try {
                    runScriptlet(onLoad, IScriptletEvent.DOCUMENT_LOADED, doc, body);
                } catch (ScriptletException e) {
                    e.printStackTrace();
                }
        }
    }

    /**
     * Called when the kit is being installed into the
     * a JEditorPane. 
     *
     * @param c the JEditorPane
     */
    public void install(JEditorPane c) {
        c.addMouseListener(repairedLinkHandler);
        c.addMouseMotionListener(repairedLinkHandler);
        super.install(c);
    }
    
    /**
     * Called when the kit is being removed from the
     * JEditorPane.  This is used to unregister any 
     * listeners that were attached.
     *
     * @param c the JEditorPane
     */
    public void deinstall(JEditorPane c) {
        c.removeMouseListener(repairedLinkHandler);
        c.removeMouseMotionListener(repairedLinkHandler);
        super.deinstall(c);
    }

    /**
     * Creates a copy of the editor kit.
     *
     * @return the copy
     */
    public Object clone() {
        DHTMLEditorKit o = (DHTMLEditorKit) super.clone();
        if (o != null)
            o.repairedLinkHandler = new RepairedLinkController();
        return o;
    }
    

    /*
     * Returns the object in an AttributeSet matching a key
     */
    static private Object getAttribValue(AttributeSet attr, HTML.Attribute key) {
        Enumeration names = attr.getAttributeNames();
        while (names.hasMoreElements()) {
            Object nextKey = names.nextElement();
            Object nextVal = attr.getAttribute(nextKey);
            if (nextVal instanceof AttributeSet) {
                Object value = getAttribValue((AttributeSet) nextVal, key);
                if (value != null) {
                    return value;
                }
            } else if (nextKey == key) {
                return nextVal;
            }
        }
        return null;
    }
    
    private class DHTMLViewFactory extends HTMLFactory {
        public View create(Element elem) {
            AttributeSet attrs = elem.getAttributes();
            View result = createCustomView(elem, attrs);
            if (result == null)
                result = maybeReplaceView(elem, attrs, super.create(elem));
                
            return result;
        }

        /** Create a custom view for the given structural element of a document.
         * 
         * @param elem the piece of the document to build a view of 
         * @return View a custom view for the element, or <code>null</code> if this
         * element does not require a custom view
         */
        private View createCustomView(Element elem, AttributeSet attrs) {            
            Object elemName = attrs.getAttribute(StyleConstants.NameAttribute);
            Object id = attrs.getAttribute(HTML.Attribute.ID);
            

            // if this is an <INPUT> component, create a custom form view.
            if (elemName == HTML.Tag.INPUT)
                return new CustomFormView(elem);

            // if this is a <DIV> component with an ID attribute, create a
            // collapsible DIV view.
            if (elemName == HTML.Tag.DIV) {
                if (attrs.isDefined(DeferredDivView.DEFERRED_SRC_ATTRIBUTE))
                    return new DeferredDivView(elem, View.Y_AXIS);
                else if (id != null)
                    return new CollapsingDivView(elem, View.Y_AXIS);
            }

            return null;
        }

        private View maybeReplaceView(Element elem, AttributeSet attrs, View result) {
            if (result == null)
                return result;
                
            String className = result.getClass().getName();
            if (className.endsWith("HiddenTagView") || className.endsWith("CommentView")) {
                // the HiddenTagView classes have 508 compliance problems in
                // JRE1.4.  They are supposed to be completely invisible, but
                // unfortunately they can receive the keyboard focus.  See
                //  http://developer.java.sun.com/developer/bugParade/bugs/4839739.html
                // for more information.
                //
                // to fix the problem, we will replace these broken elements
                // with our own implementation of an invisible element.                 
                return new InvisibleView(elem);
            }
            
            return result;
        }
    }
    
    protected RepairedLinkController repairedLinkHandler = new RepairedLinkController();
    
    protected static class RepairedLinkController extends LinkController {

        protected void handleLinkActivationEvent(MouseEvent e) {
            super.mouseClicked(e);
        }
        
        public void mouseClicked(MouseEvent e) {
            // do nothing - work will be handled by the mouseReleased method.
        }        

        public void mouseReleased(MouseEvent e) {
            JEditorPane editor = (JEditorPane) e.getSource();
            String selection = editor.getSelectedText();
            if (selection == null || selection.length() == 0)
                handleLinkActivationEvent(e);
        }

    }

}
