package gov.va.med.hds.cd.spectrum.dhtml.scriptlet;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.text.BadLocationException;
import javax.swing.text.Element;

import gov.va.med.hds.cd.spectrum.dhtml.DHTMLDocument;
import gov.va.med.hds.cd.spectrum.dhtml.DHTMLEditorKit;

public class Replace implements IScriptlet {

    public void init(IScriptletEvent event) {
    }

    public void replaceHTML(IScriptletEvent event, String id, String html)
        throws BadLocationException {
        try {
            doReplace(event, id, null, html, false);
        } catch (IOException ioe) {
            // can't happen
        }
    }

    public void replaceURL(IScriptletEvent event, String id, String url)
        throws BadLocationException, IOException {
        doReplace(event, id, url, null, false);
    }

    public void replaceInnerHTML(
        IScriptletEvent event,
        String id,
        String html)
        throws BadLocationException {
        try {
            doReplace(event, id, null, html, true);
        } catch (IOException ioe) {
            // can't happen
        }
    }

    public void replaceInnerURL(IScriptletEvent event, String id, String url)
        throws BadLocationException, IOException {
        doReplace(event, id, url, null, true);
    }

    private void doReplace(
        IScriptletEvent event,
        String id,
        String url,
        String html,
        boolean inner)
        throws BadLocationException, IOException {

        // check entry criteria
        if (event == null || id == null || (url == null && html == null))
            throw new NullPointerException();

        DHTMLDocument document = event.getDocument();
        if (document == null)
            return;
        Element elementToReplace = document.getElement(id);
        if (elementToReplace == null)
            return;

        if (url != null) {
            URL base = event.getDocument().getBase();
            html = getHTMLBody(base, url);
        }
        
        int caretPos = event.getPane().getCaretPosition();
        if (caretPos >= elementToReplace.getStartOffset() &&
            caretPos <= elementToReplace.getEndOffset()) {
            // if the caret is currently in the section that is about to
            // be replaced, then make preparations to position it at the
            // start of the inserted HTML.
            caretPos = elementToReplace.getStartOffset();
        } else {
            caretPos = -1;
        }

        if (inner)
            document.setInnerHTML(elementToReplace, html);
        else
            document.setOuterHTML(elementToReplace, html);
        
        if (caretPos >= 0)
            event.getPane().setCaretPosition(caretPos);
        
        ScriptletEnvironment env =
            (ScriptletEnvironment) document.getProperty(
                DHTMLEditorKit.SCRIPTLET_ENVIRONMENT_PROPERTY);
        env.fireSubdocumentLoaded(elementToReplace);
    }

    public static final String getHTMLBody(URL base, String url)
        throws IOException {
        return getHTMLBody(new URL(base, url));
    }

    public static String getHTMLBody(URL url) throws IOException {
        URLConnection conn = url.openConnection();
        conn.connect();

        String encoding = conn.getContentEncoding();
        Reader in =
            (encoding == null
                ? new InputStreamReader(conn.getInputStream())
                : new InputStreamReader(conn.getInputStream(), encoding));

        StringBuffer content = new StringBuffer();
        char[] buf = new char[2048];
        int charsRead;
        while ((charsRead = in.read(buf)) != -1)
            content.append(buf, 0, charsRead);

        Matcher m = BODY_START.matcher(content);
        int start = (m.find() ? m.end() : 0);
        m = BODY_END.matcher(content);
        int end = (m.find(start) ? m.start() : content.length());
        String html = content.substring(start, end);

        if ("text/plain".equalsIgnoreCase(conn.getContentType()))
            html = "<pre>" + escapeEntities(html, false) + "</pre>";
        else {
            html = rerootURLs(html, url);
            if (html.trim().length() == 0)
            	html = "<div>&nbsp;</div>";
        }
            
        return html;
    }

    static final Pattern BODY_START =
        Pattern.compile("<body[^>]*>", Pattern.CASE_INSENSITIVE);
    static final Pattern BODY_END =
        Pattern.compile("</body>", Pattern.CASE_INSENSITIVE);

    private static final String XML_CHARS = "'<>&\"";
    private static final String HTML_CHARS = XML_CHARS.substring(1);

    private static String escapeEntities(String value, boolean xml) {
        String delim = (xml ? XML_CHARS : HTML_CHARS);
        StringTokenizer tok = new StringTokenizer(value, delim, true);
        StringBuffer result = new StringBuffer();
        String token;
        while (tok.hasMoreTokens()) {
            token = tok.nextToken();
            if ("<".equals(token))
                result.append("&lt;");
            else if (">".equals(token))
                result.append("&gt;");
            else if ("&".equals(token))
                result.append("&amp;");
            else if ("\"".equals(token))
                result.append("&quot;");
            else if ("'".equals(token))
                result.append("&apos;");
            else
                result.append(token);
        }
        return result.toString();
    }

    private static String rerootURLs(String html, URL base) {
        Pattern p = getURLPattern();
        int pos = 0, start, end;
        while (true) {
            Matcher m = p.matcher(html);
            if (m.find(pos) == false) return html;
            
            String url = m.group(URL_GROUP_POS);
            start = m.start(URL_GROUP_POS);
            end = m.end(URL_GROUP_POS);
            try {                
                URL newURL = new URL(base, url);
                url = newURL.toString();
            } catch (MalformedURLException mue) {}

            html = html.substring(0, start) + url + html.substring(end);
            pos = start + url.length();
        } 
    }

    private static Pattern getURLPattern() {
        if (URL_PATTERN == null) {
            StringBuffer pattern = new StringBuffer();
            pattern.append("<(");
            for (int i = 0;   i < TAGS.length;  i++) {
                pattern.append(TAGS[i][0]).append("[^<>]+").append(TAGS[i][1])
                       .append("|");
            }
            pattern.setLength(pattern.length()-1);
            pattern.append(")=([\"'])([^:<>\"']+)\\2");
            URL_PATTERN = Pattern.compile(pattern.toString(), Pattern.CASE_INSENSITIVE);
        }
        return URL_PATTERN;
    }

    private static final String[][] TAGS = {
        { "a", "href" },
        { "img", "src" },
        { "div", "deferredsrc" }
    };
    private static final int URL_GROUP_POS = 3;

    private static Pattern URL_PATTERN = null;

}
