package gov.va.med.hds.cd.spectrum.list;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.Vector;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.ListModel;
import javax.swing.event.EventListenerList;

/**
 * Extends JList to provide <code>AssertEvent</code>s for mouse clicks or
 * &lt;enter&gt; key presses.  Supported mouse clicks include single or double
 * clicks, and defaults to single clicks.
 */
public class AssertList extends JList {
	public static final int SINGLE_CLICK = 1;
	public static final int DOUBLE_CLICK = 2;
	private EventListenerList assertList = new EventListenerList();
	private int mouseClicks = SINGLE_CLICK;

	/**
	 * Delegates to JList constructor.
	 * @param dataModel
	 */
	public AssertList(ListModel dataModel) {
		super(dataModel);
	}

	/**
	 * Delegates to JList constructor.
	 * @param listData
	 */
	public AssertList(Object[] listData) {
		super(listData);
	}

	/**
	 * Delegates to JList constructor.
	 * @param listData
	 */
	public AssertList(Vector listData) {
		super(listData);
	}

	/**
	 * Delegates to JList constructor.
	 */
	public AssertList() {
		super();
	}

	/**
	 * @see java.awt.Component#processMouseEvent(java.awt.event.MouseEvent)
	 */
	protected void processMouseEvent(MouseEvent e) {
		super.processMouseEvent(e);
		if (validClickCount(e)==mouseClicks) {
			int[] indices = super.getSelectedIndices();
			if (indices.length>0) {
				e.consume();
				int x = -1;
				int y = -1;
				if (indices.length==1) {
					x = indices[0];
					y = x;
				}
				fireAssertEvent(new AssertEvent(this,x,y,false,AssertEvent.MOUSE_CLICK));
			}
		}
	}
	
	private int validClickCount(MouseEvent e) {
		if (mouseClicks == SINGLE_CLICK &&
			e.getID()==MouseEvent.MOUSE_PRESSED && 
			e.getButton()==MouseEvent.BUTTON1 && hasAssertListener()) {
			return SINGLE_CLICK;
		}
		else if (mouseClicks==DOUBLE_CLICK &&
					e.getID()==MouseEvent.MOUSE_RELEASED &&
					e.getButton()==MouseEvent.BUTTON1 &&
					e.getClickCount()==DOUBLE_CLICK) {
			return DOUBLE_CLICK;
		}
		return 0;
	}

	private boolean hasAssertListener() {
		return assertList.getListenerCount(IAssertListener.class)>0;
	}

	/**
	 * Notify listeners of assert events.
	 * @param e
	 */
	protected void fireAssertEvent(AssertEvent e) {
		IAssertListener[] list = (IAssertListener[])assertList.getListeners(IAssertListener.class);
		for (int i = 0; i < list.length; i++) {
			list[i].valueChanged(e);
		}
	}
	
	/**
	 * @see java.awt.Component#processKeyEvent(java.awt.event.KeyEvent)
	 */
	protected void processKeyEvent(KeyEvent e) {
		if (e.getID()==KeyEvent.KEY_PRESSED && hasAssertListener()) {
			int[] indices = super.getSelectedIndices();
			if ((e.getKeyCode()==KeyEvent.VK_ENTER) && (indices.length>0)) {
				e.consume();
				int x = -1;
				int y = -1;
				if (indices.length==1) {
					x = indices[0];
					y = x;
				}
				fireAssertEvent(new AssertEvent(this,x,y,false,AssertEvent.ENTER_PRESSED));
			}
		}
		super.processKeyEvent(e);
	}
	
	/**
	 * Adds an IAssertListener to this component.
	 * @param l
	 */
	public void addAssertListener(IAssertListener l) {
		assertList.add(IAssertListener.class, l);
	}

	/**
	 * Removes an IAssertListener.
	 * @param l
	 */
	public void removeAssertListener(IAssertListener l) {
		assertList.remove(IAssertListener.class, l);
	}
	
	public static void main(String[] args) {
		JFrame frame = new JFrame("AssertList");
		AssertList list = new AssertList();
		DefaultComboBoxModel model = new DefaultComboBoxModel();
		model.addElement(new String("item 1"));
		model.addElement(new String("line 2"));
		list.addAssertListener(new IAssertListener() {
			public void valueChanged(AssertEvent e) {
				System.out.println(e.getEventID());
				System.out.println(e.getFirstIndex());
				System.out.println(e.getLastIndex());
			}
		});
		list.setModel(model);
		list.setClickCount(DOUBLE_CLICK);
		frame.getContentPane().add(list);
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.show();
	}
	
	public void setClickCount(int mouseClickCountToAssert) {
		if (mouseClickCountToAssert==SINGLE_CLICK || mouseClickCountToAssert==DOUBLE_CLICK) {
			mouseClicks = mouseClickCountToAssert;
		}
	}

}
