package gov.va.med.hds.cd.spectrum.longlist;

import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.SwingUtilities;
import javax.swing.event.EventListenerList;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

import gov.va.med.hds.util.mumps.StringUtil;

/**
 * Implementation for IBufferModel.
 */
public class BufferModel implements IBufferModel {
	private int maxSize = 300;
	private LinkedList buffer = new LinkedList();
	private boolean boundryKnown = false;
	private boolean atTop = false;
	private boolean atBottom = false;
	private int defaultCallSize = 10;
	private EventListenerList eventListenerList = new EventListenerList();
	private boolean callInProgress = false;
	private DataRequest dataRequest;
	private Thread modelThread;
	private Object[] dbDataReceived;
	private int dbDataSizeExpected;
	private LinkedList smallList = new LinkedList();
	private Comparator comparator = new BufferItemComparator();
	private int pendingSearchPosition = -1;

	/**
	 * Default constructor for BufferModel; the capacity defaults to 300 when
	 * this constructor is used.
	 */
	public BufferModel() {
		super();
	}

	public BufferModel(int capacity) {
		this();
		this.maxSize = capacity;
	}

	public int getCapacity() {
		return maxSize;
	}

	public void addListDataListener(ListDataListener listener) {
		eventListenerList.add(ListDataListener.class, listener);
	}

	public void removeListDataListener(ListDataListener listener) {
		eventListenerList.remove(ListDataListener.class, listener);
	}

	public void addDataNeededListener(IDataNeededListener listener) {
		eventListenerList.add(IDataNeededListener.class, listener);
	}

	public void removeDataNeededListener(IDataNeededListener listener) {
		eventListenerList.remove(IDataNeededListener.class, listener);
	}
	
	public void addListPositionListener(IListPositionListener listener) {
		eventListenerList.add(IListPositionListener.class, listener);
	}

	public void removeListPositionListener(IListPositionListener listener) {
		eventListenerList.remove(IListPositionListener.class, listener);
	}

	public synchronized boolean getDataRequest(DataRequest req) {
		if (!callInProgress) {
			callInProgress = true;
			this.dataRequest = req;
			makeModelThread();
			modelThread.start();
			return true;
		}
		else {
			return false;
		}
	}

	private void makeModelThread() {
		modelThread = new Thread(new Runnable() {
			public void run() {
				try {
					processDataRequest();
				}
				catch (Exception e) {
					throw new RuntimeException(e);
				}
				clearCallInProgress();
			}
		}, "BufferModel_Request");
	}
	
	private void clearCallInProgress() {
		this.dataRequest = null;
		callInProgress = false;
	}

	private void processDataRequest() throws BufferModelException {
		int dataCallMade = NO_REQUEST;
		if ((dataRequest == null) || (dataRequest.getRequestID()==RESET_REQUEST)) {
			reset();
			return;
		}
		
		if (dataRequest.getSize()==0) {
			dataRequest.setSize(defaultCallSize);
		}
		
		dbDataReceived = null;
		
		switch (dataRequest.getRequestID()) {
			case TOP_REQUEST :
				{
					dataCallMade = getTop();
					break;
				}
			case BOTTOM_REQUEST :
				{
					dataCallMade = getBottom();
					break;
				}
			case NEXT_REQUEST :
				{
					dataCallMade = getNext();
					break;
				}
			case PREV_REQUEST :
				{
					dataCallMade = getPrev();
					break;
				}
			case SEARCH_REQUEST :
				{
					dataCallMade = getSearch();
					break;
				}
			default :
				{
				}
		}
		
		if (dataCallMade!=NO_REQUEST) {
			if (dbDataReceived==null) {
				try {
					synchronized(modelThread) {
						modelThread.wait(5000);
					}
				}
				catch (InterruptedException e) {
				}
			}
			if (dbDataReceived != null) {
				processDataReceived(dataCallMade);
			}
		}
	}
	
	private void processDataReceived(int dataType) throws BufferModelException {
		switch (dataType) {
			case TOP_REQUEST :
				{
					setTop();
					break;
				}
			case BOTTOM_REQUEST :
				{
					setBottom();
					break;
				}
			case SEARCH_REQUEST :
				{
					setSearch();
					break;
				}
			case NEXT_REQUEST :
				{
					setNext();
					break;
				}
			case PREV_REQUEST :
				{
					setPrev();
					break;
				}
			default :
				{
				}
		}
	}

	private int getBottom() throws BufferModelException {
		if ((!boundryKnown) || (boundryKnown && !atBottom)) {
			fireDataNeeded(null, REVERSE_DIRECTION, dataRequest.getSize());
			return BOTTOM_REQUEST;
		}
		else {
			if (buffer.size() < dataRequest.getSize() && buffer.size()>0) {
				fireDataNeeded(buffer.getFirst(),REVERSE_DIRECTION,dataRequest.getSize()-buffer.size());
				return PREV_REQUEST;
			}
			else {
				int oldSize = smallList.size();
				if (buffer.size() >= defaultCallSize) { 
					fillSmallList(buffer.size()-defaultCallSize,defaultCallSize,BOTTOM_REQUEST);
				}
				else {
					fillSmallList(0,buffer.size()-1,BOTTOM_REQUEST);
				}
				callContentsChanged(oldSize,BOTTOM_REQUEST);
				return NO_REQUEST;
			}
		}
	}
	
	private void setBottom() {
		resetBuffer();
		if (dbDataReceived.length < dbDataSizeExpected) {
			atTop = true;
		}
		atBottom = true;
		boundryKnown = true;
		for (int i = 0; i < dbDataReceived.length; i++) {
			buffer.addFirst(dbDataReceived[i]);
		}
		if (buffer.size() >= defaultCallSize) {
			fillSmallList(buffer.size()-defaultCallSize,defaultCallSize,dataRequest.getRequestID());
		}
		else {
			fillSmallList(0,buffer.size(),dataRequest.getRequestID());
		}
	}
	
	private void callContentsChanged(int oldSize, int type) {
		int newSize = smallList.size();
		if (newSize > oldSize) {
			fireIntervalAdded(oldSize,newSize-1,type,defaultCallSize);
		}
		else if (oldSize > newSize) {
			fireIntervalRemoved(newSize,oldSize-1,type,defaultCallSize);
		}
		else {
			if (newSize<1) newSize = 1;
			fireListContentsChanged(0,newSize-1,type,defaultCallSize);
		}
	}

	private int getTop() throws BufferModelException {
		if ((!boundryKnown) || (boundryKnown && !atTop)) {
			fireDataNeeded(null, FORWARD_DIRECTION, dataRequest.getSize());
			return TOP_REQUEST;
		}
		else {
			if (buffer.size()<dataRequest.getSize() && buffer.size()>0) {
				fireDataNeeded(buffer.getLast(),FORWARD_DIRECTION,dataRequest.getSize()-buffer.size());
				return NEXT_REQUEST;
			}
			else {
				fillSmallList(0,defaultCallSize,TOP_REQUEST);
				return NO_REQUEST;
			}
		}
	}
	
	private void fillSmallList(int start, int size, int type) {
		int beginSize = smallList.size();
		smallList.clear();
		if ((start+size)>buffer.size()) {
			size = buffer.size() - start;
		}
		List subList = buffer.subList(start,start+size);
		for (Iterator i = subList.iterator(); i.hasNext();) {
			smallList.add(i.next());
		}
		callContentsChanged(beginSize,type);
	}
	
	private void appendSmallList(int start, int size, int type) {
		int beginSize = smallList.size();
		if (start < buffer.size()) {
			if (start+size > buffer.size()) {
				size = buffer.size() - start;
			}
			List subList = buffer.subList(start, start+size);
			for (Iterator i=subList.iterator();i.hasNext();) {
				smallList.addLast(i.next());
			}
			while (smallList.size()>defaultCallSize) {
				smallList.removeFirst();
			}
		}
		callContentsChanged(beginSize,type);
	}
	
	private void prependSmallList(int start, int size, int type) {
		int beginSize = smallList.size();
		if (start>-1) {
			for (int i=0;i<size;i++) {
				if (start-i>-1) {
					smallList.addFirst(buffer.get(start-i));
				}
				else {
					break;
				}
			}
			while (smallList.size()>defaultCallSize) {
				smallList.removeLast();
			}
		}
		callContentsChanged(beginSize,type);
	}
	
	private void setTop() {
		resetBuffer();
		if (dbDataReceived.length < dbDataSizeExpected) {
			atBottom = true;
		}
		for (int i = 0; i < dbDataReceived.length; i++) {
			buffer.addLast(dbDataReceived[i]);
		}
		boundryKnown = true;
		atTop = true;
		fillSmallList(0,defaultCallSize,dataRequest.getRequestID());
	}

	/**
	 * Method dumpBuffer.
	 */
	private void resetBuffer() {
		buffer = new LinkedList();
		atBottom = false;
		atTop = false;
		boundryKnown = false;
	}

	/**
	 * Returns the atBottom.
	 * @return boolean
	 */
	public boolean isAtBottom() {
		return (boundryKnown && atBottom && (((smallList.size()==0) && (buffer.size()==0)) || (smallList.getLast()==buffer.getLast())));
	}

	public Object getElementAt(int index) {
		if (index < smallList.size()) {
			return smallList.get(index);
		}
		else {
			return null;
		}
	}
	
	public boolean isAtTop() {
		return (boundryKnown && atTop && ((smallList.size()==0 && buffer.size()==0) || (smallList.getFirst()==buffer.getFirst())));
	}

	public boolean isCallInProgress() {
		return callInProgress;
	}
	
	public int getSize() {
		return smallList.size();
	}

	private int getSearch() throws BufferModelException {
		int dataCallMade = NO_REQUEST;
		pendingSearchPosition = -1;
		Object item = dataRequest.getStartAt();
		int position = indexWithinBuffer(item);
		if ((position<0) || (position>buffer.size()-1)) {
			if (position < 0) {
				//Reverse search
				if (!boundryKnown || (boundryKnown && !atTop)) {
					resetBuffer();
					dataCallMade = SEARCH_REQUEST;
					fireDataNeeded(item, FORWARD_DIRECTION, dataRequest.getSize(), true);
				}
				else {
					fillSmallList(0,defaultCallSize,SEARCH_REQUEST);
				}
			}
			else {
				//Forward search
				if (!boundryKnown || (boundryKnown && !atBottom)) {
					resetBuffer();
					dataCallMade = SEARCH_REQUEST;
					fireDataNeeded(item, FORWARD_DIRECTION, dataRequest.getSize(), true);
				}
				else {
					if (buffer.size()<defaultCallSize) {
						fillSmallList(0,buffer.size(),SEARCH_REQUEST);
					}
					else {
						fillSmallList(buffer.size()-defaultCallSize,defaultCallSize,SEARCH_REQUEST);
					}
				}
			}
		}
		else {
			if (boundryKnown && atTop && atBottom && buffer.size()<defaultCallSize) {
				firePositionChanged(position, SEARCH_REQUEST);
				return dataCallMade;
			}
			if (((boundryKnown && !atBottom) || (!boundryKnown)) && ((position + dataRequest.getSize() + 1) > buffer.size())) {
				int diff = position + defaultCallSize - buffer.size();
				dataCallMade = NEXT_REQUEST;
				fireDataNeeded(buffer.get(buffer.size()-1), FORWARD_DIRECTION, diff);
				pendingSearchPosition = position;
			}
			else {
				fillSmallList(position,defaultCallSize,SEARCH_REQUEST);
			}
		}
		
		return dataCallMade;
	}

	private int indexWithinBuffer(Object item) {
		int result;
		if (buffer.size()<1) return 1;
		if ((result=item.toString().compareToIgnoreCase(buffer.getFirst().toString()))<0) {
			return result;
		}
		else if ((result=item.toString().compareToIgnoreCase(buffer.getLast().toString()))>0) {
			return buffer.size()-1+result;
		}
		result = Collections.binarySearch(buffer, item, comparator);
		if (result<0) {
			result = (result*-1) - 1;
			result = closestMatchIndex(item.toString(), result);
		}
		return result;
	}
	
	private int closestMatchIndex(String text, int index) {
		if ((index > 0) && (index < buffer.size())) {
			int beforeIndex = StringUtil.diffStartIgnoreCase(text, buffer.get(index-1).toString());
			int afterIndex = StringUtil.diffStartIgnoreCase(text, buffer.get(index).toString());
			if (afterIndex==-1) return index;
			if (beforeIndex==-1) return index--;
			if (beforeIndex>=afterIndex) {
				return index--;
			}
		}
		return index;
	}

	private void fireListContentsChanged(int index0, int index1, int type, int size) {
		if (SwingUtilities.isEventDispatchThread()) {
			BufferListDataEvent event = new BufferListDataEvent(this,ListDataEvent.CONTENTS_CHANGED,index0,index1,type,size);
			if (eventListenerList.getListenerCount(ListDataListener.class) > 0) {
				ListDataListener[] listeners =
					(ListDataListener[]) eventListenerList.getListeners(
						ListDataListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].contentsChanged(event);
				}
			}
			if (eventListenerList.getListenerCount(IListPositionListener.class) > 0) {
				IListPositionListener[] listeners =
					(IListPositionListener[]) eventListenerList.getListeners(
						IListPositionListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].contentsChanged(event);
				}
			}
		}
		else {
			final int x = index0;
			final int y = index1;
			final int z = type;
			final int zz = size;
			Runnable callFireListChanged = new Runnable() {
				public void run() {
					fireListContentsChanged(x,y,z,zz);
				}
			};
			SwingUtilities.invokeLater(callFireListChanged);
		}
	}
	
	private void firePositionChanged(int position, int requestType) {
		BufferListDataEvent event = new BufferListDataEvent(this,ListDataEvent.CONTENTS_CHANGED,position,position,requestType);
		if (eventListenerList.getListenerCount(IListPositionListener.class) > 0) {
			IListPositionListener[] listeners =
				(IListPositionListener[]) eventListenerList.getListeners(
					IListPositionListener.class);
			for (int i = 0; i < listeners.length; i++) {
				listeners[i].positionUpdated(event);
			}
		}
	}
	
	/**
	 * Method fireIntervalAdded is called when data is added in the specified
	 * location.
	 * @param index0
	 * @param index1
	 */
	private void fireIntervalAdded(int index0, int index1, int type, int size) {
		if (SwingUtilities.isEventDispatchThread()) {
			BufferListDataEvent event = new BufferListDataEvent(this,ListDataEvent.INTERVAL_ADDED,index0,index1,type,size);
			if (eventListenerList.getListenerCount(ListDataListener.class) > 0) {
				ListDataListener[] listeners =
					(ListDataListener[]) eventListenerList.getListeners(
						ListDataListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].intervalAdded(event);
				}
			}
			if (eventListenerList.getListenerCount(IListPositionListener.class) > 0) {
				IListPositionListener[] listeners =
					(IListPositionListener[]) eventListenerList.getListeners(
						IListPositionListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].intervalAdded(event);
				}
			}
		}
		else {
			final int x = index0;
			final int y = index1;
			final int z = type;
			final int zz = size;
			Runnable callFireIntervalAdded = new Runnable() {
				public void run() {
					fireIntervalAdded(x,y,z,zz);
				}
			};
			SwingUtilities.invokeLater(callFireIntervalAdded);
		}
	}

	/**
	 * Method fireIntervalRemoved is called when data is removed in the specified
	 * location.
	 * @param index0
	 * @param index1
	 */
	private void fireIntervalRemoved(int index0, int index1, int type, int size) {
		if (SwingUtilities.isEventDispatchThread()) {
			BufferListDataEvent event = new BufferListDataEvent(this,ListDataEvent.INTERVAL_REMOVED,index0,index1,type,size);
			if (eventListenerList.getListenerCount(ListDataListener.class) > 0) {
				ListDataListener[] listeners = (ListDataListener[]) eventListenerList.getListeners(ListDataListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].intervalRemoved(event);
				}
			}
			if (eventListenerList.getListenerCount(IListPositionListener.class) > 0) {
				IListPositionListener[] listeners =
					(IListPositionListener[]) eventListenerList.getListeners(
						IListPositionListener.class);
				for (int i = 0; i < listeners.length; i++) {
					listeners[i].intervalRemoved(event);
				}
			}
		}
		else {
			final int x = index0;
			final int y = index1;
			final int z = type;
			final int zz = size;
			Runnable callFireIntervalRemoved = new Runnable() {
				public void run() {
					fireIntervalRemoved(x,y,z,zz);
				}
			};
			SwingUtilities.invokeLater(callFireIntervalRemoved);
		}
	}

	private void fireDataNeeded(Object start, int direction, int length)
		throws BufferModelException {
		fireDataNeeded(start, direction, length, false);
	}

	private void fireDataNeeded(Object start,int direction,int length,boolean dataSearch)
		throws BufferModelException {
		if (eventListenerList.getListenerCount(IDataNeededListener.class)
			> 0) {
			DataNeededEvent event =
				new DataNeededEvent(this, start, direction, length, dataSearch);
			IDataNeededListener[] listeners =
				(IDataNeededListener[]) eventListenerList.getListeners(
					IDataNeededListener.class);
			for (int i = 0; i < listeners.length; i++) {
				try {
					listeners[i].getDataRequest(event);
				}
				catch (DataProviderException e) {
					throw new BufferModelException(e);
				}
			}
		}
	}

	private int getNext() throws BufferModelException {
		int result = NO_REQUEST;
		if (smallList.size()==0) return result;
		Object searchItem = smallList.getLast();
		int pos;
		if ((pos=buffer.indexOf(searchItem))<0) {
			throw new BufferModelException("Data request not found.");
		}
		pos++;
		if (pos+dataRequest.getSize()+1>buffer.size()) {
			if (!boundryKnown || (boundryKnown && !atBottom)) {
				fireDataNeeded(searchItem,FORWARD_DIRECTION,dataRequest.getSize()-buffer.size()+pos);
				result = NEXT_REQUEST;
			}
			else {
				appendSmallList(pos,dataRequest.getSize(),NEXT_REQUEST);
			}
		}
		else {
			appendSmallList(pos,dataRequest.getSize(),NEXT_REQUEST);
		}
		return result;
	}

	private void setNext() throws BufferModelException {
		int startPosition = 0;
		if ((buffer.size() > 0) && (dbDataReceived.length > 0)) {
			Object endItem = buffer.getLast();
			if (dbDataReceived[0].equals(endItem)) {
				startPosition = 1;
			}
		}
		if (dbDataReceived.length - startPosition < dbDataSizeExpected) {
			boundryKnown = true;
			atBottom = true;
		}
		for (int i = startPosition; i < dbDataReceived.length; i++) {
			buffer.addLast(dbDataReceived[i]);
		}
		if (buffer.size() > maxSize) {
			if (!atBottom) {
				boundryKnown = false;
			}
			atTop = false;
			while (buffer.size() > maxSize) {
				buffer.removeFirst();
			}
		}
		int pos;
		if ((pos=buffer.indexOf(smallList.getLast()))<0) {
			throw new BufferModelException("Invalid data in model:NEXT");
		}
		appendSmallList(pos+1,dataRequest.getSize(),dataRequest.getRequestID());
		if (pendingSearchPosition>-1) {
			fillSmallList(pendingSearchPosition, smallList.size(), SEARCH_REQUEST);
		}
	}

	private int getPrev() throws BufferModelException {
		int result = NO_REQUEST;
		if (smallList.size()==0) return result;
		Object searchItem = smallList.getFirst();
		int pos;
		if ((pos=buffer.indexOf(searchItem))<0) {
			throw new BufferModelException("Invalid data in model:PREV");
		}
		pos--;
		if (pos-dataRequest.getSize()+1<0) {
			if (!boundryKnown || (boundryKnown && !atTop)) {
				fireDataNeeded(searchItem,REVERSE_DIRECTION,dataRequest.getSize()-pos-1);
				result = PREV_REQUEST;
			}
			else {
				prependSmallList(pos,dataRequest.getSize(),PREV_REQUEST);
			}
		}
		else {
			prependSmallList(pos,dataRequest.getSize(),PREV_REQUEST);
		}
		return result;
	}

	private void setPrev() throws BufferModelException {
		int startPosition = 0;
		if ((buffer.size() > 0) && (dbDataReceived.length > 0)) {
			Object topItem = buffer.getFirst();
			if (dbDataReceived[0].equals(topItem)) {
				startPosition = 1;
			}
		}
		if (dbDataReceived.length - startPosition < dbDataSizeExpected) {
			boundryKnown = true;
			atTop = true;
		}
		for (int i = startPosition; i < dbDataReceived.length; i++) {
			buffer.addFirst(dbDataReceived[i]);
		}
		if (buffer.size() > maxSize) {
			if (!atTop) {
				boundryKnown = false;
			}
			atBottom = false;
			while (buffer.size() > maxSize) {
				buffer.removeLast();
			}
		}
		int pos;
		if ((pos=buffer.indexOf(smallList.getFirst()))<0) {
			throw new BufferModelException("Invalid data in model:PREV");
		}
		prependSmallList(pos-1,dataRequest.getSize(),dataRequest.getRequestID());
	}

	public Object[] getBufferFull() {
		return buffer.toArray();
	}

	private void setSearch() {
		resetBuffer();
		if (dbDataReceived.length < dbDataSizeExpected) {
			atBottom = true;
			boundryKnown = true;
			atTop = false;
		}
		else {
			atBottom = false;
			boundryKnown = false;
		}
		for (int i = 0; i < dbDataReceived.length; i++) {
			buffer.addLast(dbDataReceived[i]);
		}
		fillSmallList(0,defaultCallSize,dataRequest.getRequestID());

	}

	/**
	 * @see gov.va.med.hds.cd.spectrum.IBufferModel#getDefaultCallSize()
	 */
	public int getDefaultCallSize() {
		return this.defaultCallSize;
	}

	/**
	 * @see gov.va.med.hds.cd.spectrum.IBufferModel#setDefaultCallSize(int)
	 */
	public void setDefaultCallSize(int defaultCallSize) {
		this.defaultCallSize = defaultCallSize;
	}

	/**
	 * @see gov.va.med.hds.cd.spectrum.IBufferModel#reset()
	 */
	private void reset() {
		int size = buffer.size();
		resetBuffer();
		if (size > 0) {
			fireIntervalRemoved(0, size - 1,RESET_REQUEST,size);
		}
	}
	
	/**
	 * @see gov.va.med.hds.cd.spectrum.IBufferModel#setData(Object[], int)
	 */
	public void setData(Object[] items, int sizeExpected) {
		this.dbDataReceived = items;
		this.dbDataSizeExpected = sizeExpected;
		synchronized(modelThread) {
			modelThread.notify();
		}
	}

}
