package gov.va.med.hds.cd.spectrum.plaf.basic;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JButton;
import javax.swing.UIManager;
import javax.swing.border.Border;

import gov.va.med.hds.cd.spectrum.MultiSplitPane;

/**
 * Divider used by BasicMultiSplitPaneUI. Subclassers may wish to override
 * paint to do something more interesting.
 * The border effect is drawn in BasicMultiSplitPaneUI, so if you don't like
 * that border, reset it there.
 * To conditionally drag from certain areas subclass mousePressed and
 * call super when you wish the dragging to begin.
 */
public class BasicMultiSplitPaneDivider extends Container implements PropertyChangeListener {

	/**
	 * Handles mouse dragging message to do the actual dragging.
	 */
	protected DragController dragger;

	/**
	 * UI this instance was created from.
	 */
	protected BasicMultiSplitPaneUI splitPaneUI;

	/**
	 * Size of the divider.
	 */
	protected int dividerSize = 0; // default - SET TO 0???

	/**
	 * Divider that is used for noncontinuous layout mode.
	 */
	protected Component hiddenDivider;

	/**
	 * MultiSplitPane the receiver is contained in.
	 */
	protected MultiSplitPane pane;

	/**
	 * Handles mouse events from both this class, and the split pane.
	 * Mouse events are handled for the splitpane since you want to be able
	 * to drag when clicking on the border of the divider, which is not 
	 * drawn by the divider.
	 */
	protected MouseHandler mouseHandler;

	/**
	 * Orientation of the MultiSplitPane.
	 */
	protected int orientation;

	/**
	 * Button for quickly toggling the left component.
	 */
	protected JButton leftButton;

	/**
	 * Button for quickly toggling the right component.
	 */
	protected JButton rightButton;

	/**
	 * Cursor used for HORIZONTAL_SPLIT splitpanes.
	 */
	static final Cursor horizontalCursor = Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR);

	/**
	 * Cursor used for VERTICAL_SPLIT splitpanes.
	 */
	static final Cursor verticalCursor = Cursor.getPredefinedCursor(Cursor.S_RESIZE_CURSOR);

	/**
	 * Default cursor.
	 */
	static final Cursor defaultCursor = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

	/** Border. */
	private Border border;

	/**
	 * Creates an instance of BasicMultiSplitPaneDivider. Registers this
	 * instance for mouse events and mouse dragged events.
	 */
	public BasicMultiSplitPaneDivider(BasicMultiSplitPaneUI ui) {
		setBasicMultiSplitPaneUI(ui);
		setDividerSize(ui.getSplitPane().getDividerSize());
		orientation = pane.getOrientation();
		//		setBorder(BasicMultiSplitPaneDividerBorder.getSplitPaneDividerBorder());
		setCursor(
			(orientation == MultiSplitPane.HORIZONTAL_SPLIT) ? horizontalCursor : verticalCursor);
		setBackground(UIManager.getColor("SplitPane.background"));
	}

	/**
	 * Sets the MultiSplitPaneUI that is using the receiver.
	 */
	private void setBasicMultiSplitPaneUI(BasicMultiSplitPaneUI newUI) {
		if (pane != null) {
			pane.removePropertyChangeListener(this);
			if (mouseHandler != null) {
				pane.removeMouseListener(mouseHandler);
				pane.removeMouseMotionListener(mouseHandler);
				removeMouseListener(mouseHandler);
				removeMouseMotionListener(mouseHandler);
				mouseHandler = null;
			}
		}
		splitPaneUI = newUI;
		if (newUI != null) {
			pane = newUI.getSplitPane();
			if (pane != null) {
				if (mouseHandler == null)
					mouseHandler = new MouseHandler();
				pane.addMouseListener(mouseHandler);
				pane.addMouseMotionListener(mouseHandler);
				addMouseListener(mouseHandler);
				addMouseMotionListener(mouseHandler);
				pane.addPropertyChangeListener(this);
			}
		} else {
			pane = null;
		}
	}

	/**
	 * Returns the <code>MultiSplitPaneUI</code> the receiver is currently
	 * in.
	 */
	public BasicMultiSplitPaneUI getBasicSplitPaneUI() {
		return splitPaneUI;
	}

	/**
	 * Sets the size of the divider to <code>newSize</code>. That is
	 * the width if the splitpane is <code>HORIZONTAL_SPLIT</code>, or
	 * the height of <code>VERTICAL_SPLIT</code>.
	 */
	public void setDividerSize(int newSize) {
		dividerSize = newSize;
	}

	/**
	 * Returns the size of the divider, that is the width if the splitpane
	 * is HORIZONTAL_SPLIT, or the height of VERTICAL_SPLIT.
	 */
	public int getDividerSize() {
		return dividerSize;
	}

	/**
	 * Sets the border of this component.
	 * @since 1.3
	 */
	public void setBorder(Border border) {
		Border oldBorder = this.border;

		this.border = border;
	}

	/**
	 * Returns the border of this component or null if no border is
	 * currently set.
	 *
	 * @return the border object for this component
	 * @see #setBorder
	 * @since 1.3
	 */
	public Border getBorder() {
		return border;
	}

	/**
	 * If a border has been set on this component, returns the
	 * border's insets, else calls super.getInsets.
	 *
	 * @return the value of the insets property.
	 * @see #setBorder
	 */
	public Insets getInsets() {
		Border border = getBorder();

		if (border != null) {
			return border.getBorderInsets(this);
		}
		return super.getInsets();
	}

	/**
	 * Returns dividerSize x dividerSize
	 */
	public Dimension getPreferredSize() {
		// Ideally this would return the size from the layout manager,
		// but that could result in the layed out size being different from
		// the dividerSize, which may break developers as well as
		// BasicMultiSplitPaneUI.
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return new Dimension(getDividerSize(), 1);
		}
		return new Dimension(1, getDividerSize());
	}

	/**
	 * Returns dividerSize x dividerSize
	 */
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	/**
	 * Property change event, presumably from the MultiSplitPane, will message
	 * updateOrientation if necessary.
	 */
	public void propertyChange(PropertyChangeEvent e) {
		if (e.getSource() == pane) {
			if (e.getPropertyName().equals(MultiSplitPane.ORIENTATION_PROPERTY)) {
				orientation = pane.getOrientation();
				setCursor(
					(orientation == MultiSplitPane.HORIZONTAL_SPLIT)
						? horizontalCursor
						: verticalCursor);
				invalidate();
				validate();
			}
		}
	}

	/**
	 * Paints the divider.
	 */
	public void paint(Graphics g) {
		super.paint(g);
		paintBorder(g);
	}

	protected void paintBorder(Graphics g) {
		Border border = getBorder();
		if (border != null) {
			Dimension size = getSize();
			border.paintBorder(this, g, 0, 0, size.width, size.height);
		}
	}

	/**
	 * Message to prepare for dragging. This messages the BasicMultiSplitPaneUI
	 * with startDragging.
	 */
	protected void prepareForDragging() {
		splitPaneUI.startDragging(this);
	}

	/**
	 * Messages the BasicMultiSplitPaneUI with dragDividerTo that this instance
	 * is contained in.
	 */
	protected void dragDividerTo(int location) {
		splitPaneUI.dragDividerTo(this, location);
	}

	/**
	 * Messages the BasicMultiSplitPaneUI with finishDraggingTo that this instance
	 * is contained in.
	 */
	protected void finishDraggingTo(int location) {
		splitPaneUI.finishDraggingTo(this, location);
	}

	/**
	 * MouseHandler is responsible for converting mouse events
	 * (released, dragged...) into the appropriate DragController 
	 * methods.
	 * <p>
	 */
	protected class MouseHandler extends MouseAdapter implements MouseMotionListener {
		/**
		 * Starts the dragging session by creating the appropriate instance
		 * of DragController.
		 */
		public void mousePressed(MouseEvent e) {
			if ((e.getSource() == BasicMultiSplitPaneDivider.this || e.getSource() == pane)
				&& dragger == null
				&& pane.isEnabled()) {
				Component newHiddenDivider = splitPaneUI.getHiddenDivider();

				if (hiddenDivider != newHiddenDivider) {
					if (hiddenDivider != null) {
						hiddenDivider.removeMouseListener(this);
						hiddenDivider.removeMouseMotionListener(this);
					}
					hiddenDivider = newHiddenDivider;
					if (hiddenDivider != null) {
						hiddenDivider.addMouseMotionListener(this);
						hiddenDivider.addMouseListener(this);
					}
				}
				if (pane.getComponentCount() > 1) {
					if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
						dragger = new DragController(e);
					} else {
						dragger = new VerticalDragController(e);
					}
					if (!dragger.isValid()) {
						dragger = null;
					} else {
						prepareForDragging();
						dragger.continueDrag(e);
					}
				}
				e.consume();
			}
		}

		/**
		 * If dragger is not null it is messaged with completeDrag.
		 */
		public void mouseReleased(MouseEvent e) {
			if (dragger != null) {
				if (e.getSource() == pane) {
					dragger.completeDrag(e.getX(), e.getY());
				} else if (e.getSource() == BasicMultiSplitPaneDivider.this) {
					Point ourLoc = getLocation();

					dragger.completeDrag(e.getX() + ourLoc.x, e.getY() + ourLoc.y);
				} else if (e.getSource() == hiddenDivider) {
					Point hDividerLoc = hiddenDivider.getLocation();
					int ourX = e.getX() + hDividerLoc.x;
					int ourY = e.getY() + hDividerLoc.y;

					dragger.completeDrag(ourX, ourY);
				}
				dragger = null;
				e.consume();
			}
		}

		//
		// MouseMotionListener
		//

		/**
		 * If dragger is not null it is messaged with continueDrag.
		 */
		public void mouseDragged(MouseEvent e) {
			if (dragger != null) {
				if (e.getSource() == pane) {
					dragger.continueDrag(e.getX(), e.getY());
				} else if (e.getSource() == BasicMultiSplitPaneDivider.this) {
					Point ourLoc = getLocation();

					dragger.continueDrag(e.getX() + ourLoc.x, e.getY() + ourLoc.y);
				} else if (e.getSource() == hiddenDivider) {
					Point hDividerLoc = hiddenDivider.getLocation();
					int ourX = e.getX() + hDividerLoc.x;
					int ourY = e.getY() + hDividerLoc.y;

					dragger.continueDrag(ourX, ourY);
				}
				e.consume();
			}
		}

		/**
		 *  Resets the cursor based on the orientation.
		 */
		public void mouseMoved(MouseEvent e) {
		}
	}

	/**
	 * Handles the events during a dragging session for a
	 * HORIZONTAL_SPLIT oriented split pane. This continually
	 * messages <code>dragDividerTo</code> and then when done messages
	 * <code>finishDraggingTo</code>. When an instance is created it should be
	 * messaged with <code>isValid</code> to insure that dragging can happen
	 * (dragging won't be allowed if the two views can not be resized).
	 * <p>
	 * <strong>Warning:</strong>
	 * Serialized objects of this class will not be compatible with
	 * future Swing releases. The current serialization support is
	 * appropriate for short term storage or RMI between applications running
	 * the same version of Swing.  As of 1.4, support for long term storage
	 * of all JavaBeans<sup><font size="-2">TM</font></sup>
	 * has been added to the <code>java.beans</code> package.
	 * Please see {@link java.beans.XMLEncoder}.
	 */
	protected class DragController {
		/**
		 * Initial location of the divider.
		 */
		int initialX;

		/**
		 * Maximum and minimum positions to drag to.
		 */
		int maxX, minX;

		/**
		 * Initial location the mouse down happened at.
		 */
		int offset;

		protected DragController(MouseEvent e) {
			Component[] dividers = pane.getDividers();
			int divider = indexOfDivider(dividers);
			if (divider != -1) {
				Component leftC = pane.getComponentPrecedingDivider(divider);
				Component rightC = pane.getComponentSucceedingDivider(divider);

				initialX = getLocation().x;
				if (e.getSource() == BasicMultiSplitPaneDivider.this) {
					offset = e.getX();
				} else { // pane
					offset = e.getX() - initialX;
				}
				if (leftC == null || rightC == null || offset < -1 || offset >= getSize().width) {
					// Don't allow dragging.
					maxX = -1;
				} else {
					Insets insets = pane.getInsets();

					if (leftC.isVisible()) {
						minX = leftC.getMinimumSize().width;
						if (insets != null) {
							minX += insets.left;
						}
					} else {
						minX = 0;
					}
					if (rightC.isVisible()) {
						int right = (insets != null) ? insets.right : 0;
						maxX =
							Math.max(
								0,
								pane.getSize().width
									- (getSize().width + right)
									- rightC.getMinimumSize().width);
					} else {
						int right = (insets != null) ? insets.right : 0;
						maxX = Math.max(0, pane.getSize().width - (getSize().width + right));
					}
					if (maxX < minX)
						minX = maxX = 0;
				}
			}
		}

		protected int indexOfDivider(Component[] dividers) {
			int divider = -1;
			for (int i = 0; i < dividers.length; ++i) {
				if (dividers[i] == BasicMultiSplitPaneDivider.this) {
					divider = i;
					break;
				}
			}
			return divider;
		}

		/**
		 * Returns true if the dragging session is valid.
		 */
		protected boolean isValid() {
			return (maxX > 0);
		}

		/**
		 * Returns the new position to put the divider at based on
		 * the passed in MouseEvent.
		 */
		protected int positionForMouseEvent(MouseEvent e) {
			int newX =
				(e.getSource() == BasicMultiSplitPaneDivider.this)
					? (e.getX() + getLocation().x)
					: e.getX();

			newX = Math.min(maxX, Math.max(minX, newX - offset));
			return newX;
		}

		/**
		 * Returns the x argument, since this is used for horizontal
		 * splits.
		 */
		protected int getNeededLocation(int x, int y) {
			int newX;

			newX = Math.min(maxX, Math.max(minX, x - offset));
			return newX;
		}

		protected void continueDrag(int newX, int newY) {
			dragDividerTo(getNeededLocation(newX, newY));
		}

		/**
		 * Messages dragDividerTo with the new location for the mouse
		 * event.
		 */
		protected void continueDrag(MouseEvent e) {
			dragDividerTo(positionForMouseEvent(e));
		}

		protected void completeDrag(int x, int y) {
			finishDraggingTo(getNeededLocation(x, y));
		}

		/**
		 * Messages finishDraggingTo with the new location for the mouse
		 * event.
		 */
		protected void completeDrag(MouseEvent e) {
			finishDraggingTo(positionForMouseEvent(e));
		}
	} // End of BasicMultiSplitPaneDivider.DragController

	/**
	 * Handles the events during a dragging session for a
	 * VERTICAL_SPLIT oriented split pane. This continually
	 * messages <code>dragDividerTo</code> and then when done messages
	 * <code>finishDraggingTo</code>. When an instance is created it should be
	 * messaged with <code>isValid</code> to insure that dragging can happen
	 * (dragging won't be allowed if the two views can not be resized).
	 */
	protected class VerticalDragController extends DragController {
		/* DragControllers ivars are now in terms of y, not x. */
		protected VerticalDragController(MouseEvent e) {
			super(e);
			Component[] dividers = pane.getDividers();
			int divider = indexOfDivider(dividers);
			if (divider != -1) {
				Component leftC = pane.getComponentPrecedingDivider(divider);
				Component rightC = pane.getComponentSucceedingDivider(divider);
				initialX = getLocation().y;
				if (e.getSource() == BasicMultiSplitPaneDivider.this) {
					offset = e.getY();
				} else {
					offset = e.getY() - initialX;
				}
				if (leftC == null || rightC == null || offset < -1 || offset > getSize().height) {
					// Don't allow dragging.
					maxX = -1;
				} else {
					Insets insets = pane.getInsets();

					if (leftC.isVisible()) {
						minX = leftC.getMinimumSize().height;
						if (insets != null) {
							minX += insets.top;
						}
					} else {
						minX = 0;
					}
					if (rightC.isVisible()) {
						int bottom = (insets != null) ? insets.bottom : 0;

						maxX =
							Math.max(
								0,
								pane.getSize().height
									- (getSize().height + bottom)
									- rightC.getMinimumSize().height);
					} else {
						int bottom = (insets != null) ? insets.bottom : 0;

						maxX = Math.max(0, pane.getSize().height - (getSize().height + bottom));
					}
					if (maxX < minX)
						minX = maxX = 0;
				}
			}
		}

		/**
		 * Returns the y argument, since this is used for vertical
		 * splits.
		 */
		protected int getNeededLocation(int x, int y) {
			int newY;

			newY = Math.min(maxX, Math.max(minX, y - offset));
			return newY;
		}

		/**
		 * Returns the new position to put the divider at based on
		 * the passed in MouseEvent.
		 */
		protected int positionForMouseEvent(MouseEvent e) {
			int newY =
				(e.getSource() == BasicMultiSplitPaneDivider.this)
					? (e.getY() + getLocation().y)
					: e.getY();

			newY = Math.min(maxX, Math.max(minX, newY - offset));
			return newY;
		}
	}
}
