package gov.va.med.hds.cd.spectrum.plaf.basic;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager2;

import gov.va.med.hds.cd.spectrum.MultiSplitPane;

public class BasicMultiSplitPaneLayoutManager implements LayoutManager2 {

	/** Orientation. */
	private int orientation;

	BasicMultiSplitPaneLayoutManager(int orientation) {
		this.orientation = orientation;
	}

	public void layoutContainer(Container container) {
		Dimension containerSize = container.getSize();
		if (containerSize.height <= 0 || containerSize.width <= 0) {
			return;
		}
		if (container.getComponentCount() == 0)
			return;

		MultiSplitPane splitPane = (MultiSplitPane) container;
		int availableSize = getAvailableSize(splitPane);

		double[] ratios = splitPane.getRatios();
		if (!areRatiosSet(ratios)) {
			splitPane.setRatios(getPreferredRatios(splitPane, ratios, availableSize));
		}

		int[] sizes = calculateSizes(splitPane, ratios, availableSize);

		setComponentBounds(splitPane, sizes);
	}

	private int[] calculateSizes(
		MultiSplitPane splitPane,
		double[] ratios,
		int availableSize) {
		int[] sizes = new int[splitPane.getAllComponentCount()];
		for (int i = 0; i < sizes.length; ++i) {
			if (isDividerIndex(i)) {
				sizes[i] = splitPane.getDividerSize();
			} else {
				int componentIndex = i / 2;
				sizes[i] = (int) (ratios[componentIndex] * availableSize);
			}
		}
		return sizes;
	}

	private double[] getPreferredRatios(
		MultiSplitPane splitPane,
		double[] ratios,
		int availableSize) {
		int[] preferredSizes = getMinimumSizes(splitPane);
		for (int i = 0; i < preferredSizes.length; i++) {
			ratios[i] = preferredSizes[i] / availableSize;
		}
		return ratios;
	}

	private boolean areRatiosSet(double[] ratios) {
		if (ratios[0] < 0.0)
			return false;
		return true;
	}

	private boolean isDividerIndex(int i) {
		return (i % 2) != 0;
	}

	protected void setComponentBounds(MultiSplitPane splitPane, int[] sizes) {
		int nextLocation = getInitialLocation(splitPane.getInsets());
		for (int i = 0; i < splitPane.getAllComponentCount(); i++) {
			Component c = splitPane.getComponentFromAll(i);
			setComponentToSize(splitPane, c, sizes[i], nextLocation);
			nextLocation += sizes[i];
		}
	}

	/**
	 * Sets the width of the component c to be size, placing its
	 * x location at location, y to the insets.top and height
	 * to the containersize.height less the top and bottom insets.
	 */
	protected void setComponentToSize(
		MultiSplitPane splitPane,
		Component c,
		int size,
		int location) {
		Dimension containerSize = splitPane.getSize();
		Insets insets = splitPane.getInsets();
		if (insets != null) {
			if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
				c.setBounds(
					location,
					insets.top,
					size,
					containerSize.height - (insets.top + insets.bottom));
			} else {
				c.setBounds(
					insets.left,
					location,
					containerSize.width - (insets.left + insets.right),
					size);
			}
		} else {
			if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
				c.setBounds(location, 0, size, containerSize.height);
			} else {
				c.setBounds(0, location, containerSize.width, size);
			}
		}
	}

	/**
	 * Returns the available width based on the container size and
	 * Insets.
	 */
	int getTotalAvailableSize(MultiSplitPane splitPane) {
		Dimension containerSize = splitPane.getSize();
		Insets insets = splitPane.getInsets();
		if (insets == null)
			return getSizeForPrimaryAxis(containerSize);
		return (
			getSizeForPrimaryAxis(containerSize)
				- (getSizeForPrimaryAxis(insets, true)
					+ getSizeForPrimaryAxis(insets, false)));
	}

	int getAvailableSize(MultiSplitPane splitPane) {
		return getTotalAvailableSize(splitPane)
			- splitPane.getDividerSize() * splitPane.getDividerCount();
	}

	/**
	 * Returns the minimum size needed to contain the children.
	 * The width is the sum of all the childrens min widths and
	 * the height is the largest of the childrens minimum heights.
	 */
	public Dimension minimumLayoutSize(Container container) {
		int minPrimary = 0;
		int minSecondary = 0;
		Insets insets = container.getInsets();

		for (int i = 0; i < container.getComponentCount(); i++) {
			if (container.getComponent(i) != null) {
				Dimension minSize = container.getComponent(i).getMinimumSize();
				int secSize = getSizeForSecondaryAxis(minSize);

				minPrimary += getSizeForPrimaryAxis(minSize);
				if (secSize > minSecondary)
					minSecondary = secSize;
			}
		}
		if (insets != null) {
			minPrimary += getSizeForPrimaryAxis(insets, true)
				+ getSizeForPrimaryAxis(insets, false);
			minSecondary += getSizeForSecondaryAxis(insets, true)
				+ getSizeForSecondaryAxis(insets, false);
		}
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return new Dimension(minPrimary, minSecondary);
		}
		return new Dimension(minSecondary, minPrimary);
	}

	/**
	 * Returns the preferred size needed to contain the children.
	 * The width is the sum of all the childrens preferred widths and
	 * the height is the largest of the childrens preferred heights.
	 */
	public Dimension preferredLayoutSize(Container container) {
		int prePrimary = 0;
		int preSecondary = 0;
		Insets insets = container.getInsets();

		for (int i = 0; i < 3; i++) {
			if (container.getComponent(i) != null) {
				Dimension preSize =
					container.getComponent(i).getPreferredSize();
				int secSize = getSizeForSecondaryAxis(preSize);

				prePrimary += getSizeForPrimaryAxis(preSize);
				if (secSize > preSecondary)
					preSecondary = secSize;
			}
		}
		if (insets != null) {
			prePrimary += getSizeForPrimaryAxis(insets, true)
				+ getSizeForPrimaryAxis(insets, false);
			preSecondary += getSizeForSecondaryAxis(insets, true)
				+ getSizeForSecondaryAxis(insets, false);
		}
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return new Dimension(prePrimary, preSecondary);
		}
		return new Dimension(preSecondary, prePrimary);
	}

	public void addLayoutComponent(String name, Component component) {
		// NOOP
	}

	public void removeLayoutComponent(Component component) {
		// NOOP
	}

	/**
	* Adds the specified component to the layout, using the specified
	* constraint object.
	* @param comp the component to be added
	* @param constraints  where/how the component is added to the layout.
	*/
	public void addLayoutComponent(Component comp, Object constraints) {
		if ((constraints == null) || (constraints instanceof String)) {
			addLayoutComponent((String) constraints, comp);
		} else {
			throw new IllegalArgumentException(
				"cannot add to layout: "
					+ "constraint must be a "
					+ "string (or null)");
		}
	}

	/**
	 * Returns the left inset, unless the Insets are null in which case
	 * 0 is returned.
	 */
	protected int getInitialLocation(Insets insets) {
		if (insets != null)
			return getSizeForPrimaryAxis(insets, true);
		return 0;
	}

	/**
	 * Returns an array of the minimum sizes of the components.
	 */
	int[] getPreferredSizes(MultiSplitPane splitPane) {
		int[] retValue = new int[splitPane.getComponentCount()];
		for (int i = 0; i < splitPane.getComponentCount(); i++) {
			Component c = splitPane.getComponent(i);
			retValue[i] = getPreferredSizeOfComponent(c);
		}
		return retValue;
	}

	/**
	 * Returns an array of the minimum sizes of the components.
	 */
	int[] getMinimumSizes(MultiSplitPane splitPane) {
		int[] retValue = new int[splitPane.getComponentCount()];
		for (int i = 0; i < splitPane.getComponentCount(); i++) {
			Component c = splitPane.getComponent(i);
			retValue[i] = getMinimumSizeOfComponent(c);
		}
		return retValue;
	}

	/**
	 * Returns the alignment along the x orientation.  This specifies how
	 * the component would like to be aligned relative to other 
	 * components.  The value should be a number between 0 and 1
	 * where 0 represents alignment along the origin, 1 is aligned
	 * the furthest away from the origin, 0.5 is centered, etc.
	 */
	public float getLayoutAlignmentX(Container target) {
		return 0.0f;
	}

	/**
	 * Returns the alignment along the y orientation.  This specifies how
	 * the component would like to be aligned relative to other 
	 * components.  The value should be a number between 0 and 1
	 * where 0 represents alignment along the origin, 1 is aligned
	 * the furthest away from the origin, 0.5 is centered, etc.
	 */
	public float getLayoutAlignmentY(Container target) {
		return 0.0f;
	}

	/**
	 * Does nothing. If the developer really wants to change the
	 * size of one of the views JSplitPane.resetToPreferredSizes should
	 * be messaged.
	 */
	public void invalidateLayout(Container c) {
		// NOOP
	}

	/**
	 * Returns the maximum layout size, which is Integer.MAX_VALUE
	 * in both directions.
	 */
	public Dimension maximumLayoutSize(Container target) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	/**
	  * Returns the width of the passed in Components preferred size.
	  */
	int getPreferredSizeOfComponent(Component c) {
		return getSizeForPrimaryAxis(c.getPreferredSize());
	}

	/**
	 * Returns the width of the passed in Components minimum size.
	 */
	int getMinimumSizeOfComponent(Component c) {
		return getSizeForPrimaryAxis(c.getMinimumSize());
	}

	/**
	 * Returns the width of the passed in component.
	 */
	int getSizeOfComponent(Component c) {
		return getSizeForPrimaryAxis(c.getSize());
	}

	int getLocationOfComponent(Component c) {
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return c.getLocation().x;
		}
		return c.getLocation().y;
	}

	/**
	 * If the orientation == MultiSplitPane.HORIZONTAL_SPLIT, the width is returned, otherwise the height.
	 */
	int getSizeForPrimaryAxis(Dimension size) {
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return size.width;
		}
		return size.height;
	}

	/**
	 * If the orientation == MultiSplitPane.HORIZONTAL_SPLIT, the height is returned, otherwise the width.
	 */
	int getSizeForSecondaryAxis(Dimension size) {
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			return size.height;
		}
		return size.width;
	}

	/**
	 * Returns a particular value of the inset identified by the
	 * orientation and <code>isTop</code><p>
	 *   orientation isTop
	 *    0    true    - left
	 *    0    false   - right
	 *    1    true    - top
	 *    1    false   - bottom
	 */
	int getSizeForPrimaryAxis(Insets insets, boolean isTop) {
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			if (isTop) {
				return insets.left;
			}
			return insets.right;
		}
		if (isTop) {
			return insets.top;
		}
		return insets.bottom;
	}

	/**
	 * Returns a particular value of the inset identified by the
	 * orientation and <code>isTop</code><p>
	 *   orientation isTop
	 *    0    true    - left
	 *    0    false   - right
	 *    1    true    - top
	 *    1    false   - bottom
	 */
	int getSizeForSecondaryAxis(Insets insets, boolean isTop) {
		if (orientation == MultiSplitPane.HORIZONTAL_SPLIT) {
			if (isTop) {
				return insets.top;
			}
			return insets.bottom;
		}
		if (isTop) {
			return insets.left;
		}
		return insets.right;
	}
}