package gov.va.med.hds.cd.spectrum.plaf.basic;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.LookAndFeel;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.UIResource;

import gov.va.med.hds.cd.spectrum.MultiSplitPane;
import gov.va.med.hds.cd.spectrum.plaf.MultiSplitPaneUI;

public class BasicMultiSplitPaneUI extends MultiSplitPaneUI {

	/**
	 * JSplitPane instance this instance is providing
	 * the look and feel for.
	 */
	protected MultiSplitPane splitPane;

	/**
	 * List of the dividers for this MultiSplitPane.
	 */
	protected List dividers = new ArrayList();

	/**
	 * Instance of the PropertyChangeListener for this MultiSplitPane.
	 */
	protected PropertyChangeListener propertyChangeListener;

	/**
	 * Instance of the FocusListener for this MultiSplitPane.
	 */
	protected FocusListener focusListener;

	/**
	 * The size of the divider while the dragging session is valid.
	 */
	protected int dividerSize;

	/**
	 * Instance for the shadow of the divider during dragging
	 */
	protected Component hiddenDivider;

	/**
	 * Location of the divider when the dragging session began.
	 */
	protected int beginDragDividerLocation;

	private int lastDragLocation;

	/**
	 * LayoutManager that is created and placed into the split pane.
	 */
	protected BasicMultiSplitPaneLayoutManager layoutManager;

	/**
	 * Creates a new <code>BasicMultiSplitPaneUI</code> instance
	 */
	public static ComponentUI createUI(JComponent x) {
		return new BasicMultiSplitPaneUI();
	}

	/**
	 * Installs the UI.
	 */
	public void installUI(JComponent c) {
		this.splitPane = (MultiSplitPane) c;
		installDefaults();
		installListeners();
		setLastDragLocation(-1);
	}

	/**
	 * Installs the UI defaults.
	 */
	protected void installDefaults() {
		LookAndFeel.installColors(splitPane, "SplitPane.background", "SplitPane.foreground");

		installDividers();

		resetLayoutManager();

		/* Install the hiddenDivider here to avoid having to add/remove everything later. */
		if (hiddenDivider == null) {
			setHiddenDivider(createDefaultHiddenDivider());
		} else {
			setHiddenDivider(hiddenDivider);
		}
	}

	protected void installDividers() {
		this.dividerSize =  ((Integer) (UIManager.get("SplitPane.dividerSize"))).intValue();
        splitPane.setDividerSize(this.dividerSize);
		if (splitPane.getComponentCount() > 0 && splitPane.getDividerCount() == 0) {
			Component[] c = splitPane.getComponents();
			for (int i = 0; i < c.length - 1; ++i) {
				splitPane.add(createDefaultDivider(), MultiSplitPane.DIVIDER, (i * 2) + 1);
			}
		}
	}

	/**
	 * Installs the event listeners for the UI.
	 */
	protected void installListeners() {
		if ((propertyChangeListener = createPropertyChangeListener())
			!= null) {
			splitPane.addPropertyChangeListener(propertyChangeListener);
		}
		if ((focusListener = createFocusListener()) != null) {
			splitPane.addFocusListener(focusListener);
		}
	}

	/**
	* Uninstalls the UI.
	*/
	public void uninstallUI(JComponent c) {
		uninstallListeners();
		uninstallDefaults();
//		 dividerLocationIsSet = false;
		splitPane = null;
	}

	/**
	 * Uninstalls the UI defaults.
	 */
	protected void uninstallDefaults() {
		if (splitPane.getLayout() == layoutManager) {
			splitPane.setLayout(null);
		}

		if (hiddenDivider != null) {
			splitPane.remove(hiddenDivider);
		}

		splitPane.setBorder(null);
		
		LookAndFeel.uninstallBorder(splitPane);
		
		uninstallBorders();
		
		layoutManager = null;
		uninstallDividers();

		setHiddenDivider(null);
	}

	private void uninstallDividers() {
		splitPane.removeDividers();
		dividers.clear();
	}

	private void uninstallBorders() {
		for (Iterator i = dividers.iterator(); i.hasNext(); ) {
			BasicMultiSplitPaneDivider d = (BasicMultiSplitPaneDivider) i.next();
			Border b = d.getBorder();
			if (b instanceof UIResource) {
				d.setBorder(null);
			}
		}
	}

	/**
	 * Uninstalls the event listeners for the UI.
	 */
	protected void uninstallListeners() {
		if (propertyChangeListener != null) {
			splitPane.removePropertyChangeListener(propertyChangeListener);
			propertyChangeListener = null;
		}
		if (focusListener != null) {
			splitPane.removeFocusListener(focusListener);
			focusListener = null;
		}
	}

	/**
	 * Returns the last drag location of the JSplitPane.
	 */
	public int getLastDragLocation() {
		return lastDragLocation;
	}

	/**
	 * Set the last drag location of the JSplitPane.
	 */
	public void setLastDragLocation(int l) {
		lastDragLocation = l;
	}

	/**
	 * Creates a PropertyChangeListener for the MultiSplitPane UI.
	 */
	protected PropertyChangeListener createPropertyChangeListener() {
		return new PropertyHandler();
	}

	/**
	 * Creates a FocusListener for the MultiSplitPane UI.
	 */
	protected FocusListener createFocusListener() {
		return new FocusHandler();
	}

	/**
	* Returns the default non continuous layout divider, which is an
	* instanceof Canvas that fills the background in dark gray.
	*/
	protected Component createDefaultHiddenDivider() {
		return new Canvas() {
			public void paint(Graphics g) {
				if (getLastDragLocation() != -1) {
					Dimension size = splitPane.getSize();

					g.setColor(Color.darkGray);
					if (splitPane.getOrientation()
						== MultiSplitPane.HORIZONTAL_SPLIT) {
						g.fillRect(0, 0, dividerSize - 1, size.height - 1);
					} else {
						g.fillRect(0, 0, size.width - 1, dividerSize - 1);
					}
				}
			}
		};
	}

	/**
	 * Sets the divider to use when the splitPane is configured to
	 * not continuously layout. This divider will only be used during a
	 * dragging session. It is recommended that the passed in component
	 * be a heavy weight.
	 */
	protected void setHiddenDivider(Component newDivider) {
		//        rememberPaneSizes = rememberSizes;
		if (hiddenDivider != null && splitPane != null) {
			splitPane.remove(hiddenDivider);
		}
		this.hiddenDivider = newDivider;
	}

	/**
	* Resets the layout manager based on orientation and messages it
	* with invalidateLayout to pull in appropriate Components.
	*/
	private void resetLayoutManager() {
		layoutManager =
			new BasicMultiSplitPaneLayoutManager(splitPane.getOrientation());

		splitPane.setLayout(layoutManager);
		//		layoutManager.updateComponents();
		splitPane.revalidate();
		splitPane.repaint();
	}

	protected BasicMultiSplitPaneDivider createDefaultDivider() {
		return new BasicMultiSplitPaneDivider(this);
	}

	public void finishedPaintingChildren(MultiSplitPane jc, Graphics g) {
		int i = 0;
		if (jc == splitPane && getLastDragLocation() != -1) {
			Dimension size = splitPane.getSize();

			g.setColor(Color.darkGray);
			if (splitPane.getOrientation()
				== MultiSplitPane.HORIZONTAL_SPLIT) {
				g.fillRect(
					getLastDragLocation(),
					0,
					dividerSize - 1,
					size.height - 1);
			} else {
				g.fillRect(
					0,
					lastDragLocation,
					size.width - 1,
					dividerSize - 1);
			}
		}
	}

	public void addDivider(MultiSplitPane splitPane) {
		BasicMultiSplitPaneDivider divider = createDefaultDivider();
		divider.setDividerSize(this.dividerSize);
		dividers.add(divider);
		splitPane.add(divider, MultiSplitPane.DIVIDER);
		
	}

	/**
	 * Returns the divider to use when the splitPane is configured to
	 * not continuously layout. This divider will only be used during a
	 * dragging session.
	 */
	public Component getHiddenDivider() {
		return hiddenDivider;
	}

	public MultiSplitPane getSplitPane() {
		return splitPane;
	}

	/**
	 * Sets the location of the divider to location.
	 */
	public void setDividerLocation(
		MultiSplitPane jc,
		int divider,
		int location) {
		if (location >= getMinimumDividerLocation(splitPane, divider)
			&& location <= getMaximumDividerLocation(splitPane, divider)) {
			Component d = splitPane.getDivider(divider);
			int availableSpace = layoutManager.getAvailableSize(splitPane);
			double shift =
				(double) (location - layoutManager.getLocationOfComponent(d))
					/ (double) availableSpace;

			double[] ratios = splitPane.getRatios();
			ratios[divider] += shift;
			ratios[divider + 1] -= shift;

			splitPane.setRatios(ratios);
			splitPane.revalidate();
		}
		splitPane.repaint();
	}

	/**
	 * Should be messaged before the dragging session starts, resets
	 * lastDragLocation and dividerSize.
	 */
	protected void startDragging(Component divider) {
		beginDragDividerLocation =
			layoutManager.getLocationOfComponent(divider);
		setLastDragLocation(beginDragDividerLocation);
	}

	/**
	 * Messaged during a dragging session to move the divider to the
	 * passed in location. If continuousLayout is true the location is
	 * reset and the splitPane validated.
	 */
	protected void dragDividerTo(Component divider, int location) {
		if (getLastDragLocation() != location) {
			int lastLoc = getLastDragLocation();

			setLastDragLocation(location);
			if (splitPane.getOrientation()
				== MultiSplitPane.HORIZONTAL_SPLIT) {
				int splitHeight = splitPane.getHeight();
				splitPane.repaint(lastLoc, 0, dividerSize, splitHeight);
				splitPane.repaint(location, 0, dividerSize, splitHeight);
			} else {
				int splitWidth = splitPane.getWidth();
				splitPane.repaint(0, lastLoc, splitWidth, dividerSize);
				splitPane.repaint(0, location, splitWidth, dividerSize);
			}
		}
	}

	/**
	 * Messaged to finish the dragging session. If not continuous display
	 * the dividers location will be reset.
	 */
	protected void finishDraggingTo(Component divider, int location) {
		dragDividerTo(divider, location);
		setLastDragLocation(-1);
		splitPane.setDividerLocation(getDividerIndex(divider), location);
	}

	private int getDividerIndex(Component divider) {
		for (int i = 0; i < splitPane.getDividerCount(); ++i) {
			if (splitPane.getDivider(i).equals(divider))
				return i;
		}
		return -1;
	}

	/**
	* Returns the preferred size for the passed in component,
	* This is passed off to the current layoutmanager.
	*/
	public Dimension getPreferredSize(JComponent jc) {
		if (splitPane != null)
			return layoutManager.preferredLayoutSize(splitPane);
		return new Dimension(0, 0);
	}

	/**
	 * Returns the minimum size for the passed in component,
	 * This is passed off to the current layoutmanager.
	 */
	public Dimension getMinimumSize(JComponent jc) {
		if (splitPane != null)
			return layoutManager.minimumLayoutSize(splitPane);
		return new Dimension(0, 0);
	}

	/**
	 * Returns the maximum size for the passed in component,
	 * This is passed off to the current layoutmanager.
	 */
	public Dimension getMaximumSize(JComponent jc) {
		if (splitPane != null)
			return layoutManager.maximumLayoutSize(splitPane);
		return new Dimension(0, 0);
	}

	boolean isComponentValid(Component c) {
		return c != null && c.isVisible();
	}

	/**
	 * Gets the minimum location of the divider.
	 */
	protected int getMinimumDividerLocation(
		MultiSplitPane splitPane,
		int divider) {
		Component c = splitPane.getComponentPrecedingDivider(divider);
		return layoutManager.getLocationOfComponent(c)
			+ layoutManager.getMinimumSizeOfComponent(c);
	}

	/**
	 * Gets the maximum location of the divider.
	 */
	protected int getMaximumDividerLocation(MultiSplitPane jc, int divider) {
		Component c = splitPane.getComponentSucceedingDivider(divider);
		int extent =
			layoutManager.getLocationOfComponent(c)
				+ layoutManager.getSizeOfComponent(c);
		return extent - layoutManager.getMinimumSizeOfComponent(c);
	}

	/**
	 * Implementation of the PropertyChangeListener
	 * that the MultiSplitPane UI uses.
	 * <p>
	 * This inner class is marked &quot;public&quot; due to a compiler bug.
	 * This class should be treated as a &quot;protected&quot; inner class.
	 * Instantiate it only within subclasses of BasicSplitPaneUI.
	 */
	public class PropertyHandler implements PropertyChangeListener {
		/**
		 * Messaged from the <code>MultiSplitPane</code> the receiver is
		 * contained in.  May potentially reset the layout manager and cause a
		 * <code>validate</code> to be sent.
		 */
		public void propertyChange(PropertyChangeEvent e) {
			if (e.getSource() == splitPane) {
				String changeName = e.getPropertyName();

				if (changeName.equals(MultiSplitPane.ORIENTATION_PROPERTY)) {
					resetLayoutManager();
				} else if (
					changeName.equals(MultiSplitPane.DIVIDER_SIZE_PROPERTY)) {
//					setDividerSizes(splitPane.getDividerSize());
					splitPane.revalidate();
					splitPane.repaint();
				}
			}
		}
	}

	/**
	 * Implementation of the FocusListener that the MultiSplitPane UI uses.
	 * <p>
	 * This inner class is marked &quot;public&quot; due to a compiler bug.
	 * This class should be treated as a &quot;protected&quot; inner class.
	 * Instantiate it only within subclasses of BasicSplitPaneUI.
	 */
	public class FocusHandler extends FocusAdapter {
		public void focusGained(FocusEvent ev) {
			splitPane.repaint();
		}

		public void focusLost(FocusEvent ev) {
			splitPane.repaint();
		}
	}
}
