/*
 * Created on Apr 17, 2003
 */
package gov.va.med.hds.cd.spectrum.text;

import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.DocumentFilter;

/** General purpose document filter for restricting user input.
 */
public class GenericDocumentFilter extends DocumentFilter {

    private static final String DEFAULT_FORBIDDEN_CHARS = "^";
    private static final int NO_MAX_LENGTH = -1;

    private int maxLength;
    private boolean multiline;
    private String forbiddenChars;

    /** Create a generic document filter.
     * 
     * The filter created will enforce no maximum length.  It will forbid
     * the use of the "^" character.  It will strip carriage returns and
     * newlines from input.
     */
    public GenericDocumentFilter() {
        this(NO_MAX_LENGTH, DEFAULT_FORBIDDEN_CHARS, false);
    }

    /** Create a generic document filter. 
     * 
     * The filter created will forbid the use of the "^" character.  It will
     * strip carriage returns and newlines from input.
     * 
     * @param maxLength the maximum number of characters to allow.
     */
    public GenericDocumentFilter(int maxLength) {
        this(maxLength, DEFAULT_FORBIDDEN_CHARS, false);
    }

    /** Create a generic document filter. 
     * 
     * The filter created will strip carriage returns and newlines from input.
     * 
     * @param maxLength the maximum number of characters to allow.
     * @param forbiddenChars a string of characters the user should not be
     *    allowed to input.
     */
    public GenericDocumentFilter(int maxLength, String forbiddenChars) {
        this(maxLength, forbiddenChars, false);
    }

    /** Create a generic document filter. 
     * 
     * @param maxLength the maximum number of characters to allow.
     * @param forbiddenChars a string of characters the user should not be
     *    allowed to input.
     * @param multiline true if this is a multiline document.  If false, the
     *    filter will strip carriage returns and newlines from input.
     */
    public GenericDocumentFilter(
        int maxLength,
        String forbiddenChars,
        boolean multiline) {
        this.maxLength = maxLength;

        if (forbiddenChars != null && forbiddenChars.length() > 0)
            this.forbiddenChars = forbiddenChars;
        else
            this.forbiddenChars = null;

        this.multiline = multiline;
    }

    // implementation of the DocumentFilter.insertString() method.
    public void insertString(
        FilterBypass fb,
        int offset,
        String text,
        AttributeSet attrs)
        throws BadLocationException {

        if (test(text, fb, 0))
            super.insertString(fb, offset, clean(text), attrs);
        else
            beep();
    }

    // implementation of the DocumentFilter.replace() method.
    public void replace(
        FilterBypass fb,
        int offset,
        int length,
        String text,
        AttributeSet attr)
        throws BadLocationException {

        if (test(text, fb, length))
            super.replace(fb, offset, length, clean(text), attr);
        else
            beep();
    }

    /** Return true if the length of the document would be acceptable after
     * deleting some number of characters and replacing them with insText.
     */
    private boolean lengthOkay(
        String insText,
        FilterBypass fb,
        int delChars) {
        if (maxLength == NO_MAX_LENGTH)
            return true;

        int endingLength = fb.getDocument().getLength() - delChars;
        if (insText != null)
            endingLength += insText.length();
        return endingLength <= maxLength;
    }

    /** Return false if insText contains any forbidden characters.
     */
    private boolean charsOkay(String insText) {
        if (forbiddenChars == null
            || insText == null
            || insText.length() == 0)
            return true;

        for (int i = insText.length(); i-- > 0;)
            if (forbiddenChars.indexOf(insText.charAt(i)) != -1)
                return false;

        return true;
    }

    /** Test whether a modification should be allowed.
     */
    private boolean test(String insText, FilterBypass fb, int delChars) {
        return lengthOkay(insText, fb, delChars) && charsOkay(insText);
    }

    /** Strip carriage returns and newlines from text if necessary.
     */
    private String clean(String text) {
        if (multiline)
            return text;
        return text.replace('\r', ' ').replace('\n', ' ');
    }

    /** Alert the user that their input is unacceptable.
     */
    private void beep() {
        java.awt.Toolkit.getDefaultToolkit().beep();
    }

}
