package gov.va.med.hds.task.data.internal;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import gov.va.med.hds.cd.m.IMumps;
import gov.va.med.hds.cd.m.IMumpsFactory;
import gov.va.med.hds.cd.runtime.IConfigurationElement;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.task.ITaskPlugin;
import gov.va.med.hds.task.data.ILinkedItem;
import gov.va.med.hds.task.data.ITask;
import gov.va.med.hds.task.data.TaskDataAccessException;
import gov.va.med.hds.task.data.TaskID;
import gov.va.med.hds.task.internal.TaskEditorPlugin;
import gov.va.med.hds.util.mumps.StringUtil;

public class MTaskDAO extends MemoryTaskDAO {

	private static final String LINKED_ITEM_SEPARATOR = "~";
	private TaskEditorPlugin taskPlugin;
	private IMumps mumps;
    private ILogger logger = null;
    private static boolean CACHE_TASKS = false;

	public void setInitializationData(IRuntimeContext runtime, IConfigurationElement config, String propertyName, Object data) throws Exception {
		super.setInitializationData(runtime, config, propertyName, null);

		try {
			taskPlugin = (TaskEditorPlugin) runtime.getPluginInstance(TaskEditorPlugin.PLUGIN_ID);
			IPluginDescriptor desc = runtime.getPluginRegistry().getPluginDescriptor(ITaskPlugin.PLUGIN_ID);
			mumps = ((IMumpsFactory) runtime.getPluginInstance(IMumpsFactory.PLUGIN_ID)).create(desc);
			currDate = new MDateProvider(runtime);
			comparator = new TaskComparator(PRIORITIES, currDate);
		} catch (Exception e) {
			throw e;
		}
	}

	public ITask createNewTask(Object patientID, Object creatorID) {
		// parent behavior is acceptable.
		return super.createNewTask(patientID, creatorID);
	}

	protected AbstractTask newTaskObject() {
		return new MTask();
	}

	public String[] getDefinedTaskPriorities() throws TaskDataAccessException {
		// parent behavior is acceptable.
		return super.getDefinedTaskPriorities();
	}

	public ITask getTask(Object taskID) throws TaskDataAccessException {
		taskID = resolveTaskID(taskID);

		// ensure that the task is in the cache.
		if (CACHE_TASKS == false || !taskCache.containsKey(taskID))
			loadTasks(Collections.singletonList(taskID));

		// return the cached task.
		return super.getTask(taskID);
	}

	public List getTasks(List taskIDs) throws TaskDataAccessException {
		List tasksToLoad = new LinkedList();
		Iterator i = taskIDs.iterator();
		while (i.hasNext()) {
			Object taskID = resolveTaskID(i.next());
			if (CACHE_TASKS == false || !taskCache.containsKey(taskID))
				tasksToLoad.add(taskID);
		}

		if (!tasksToLoad.isEmpty())
			loadTasks(tasksToLoad);

		return super.getTasks(taskIDs);
	}

	public List getTasksForPatient(Object patientID) throws TaskDataAccessException {

		String result = null;
		try {
			result = mumps.callRPC(TASKS_BY_PATIENT_RPC_KEY, Collections.singletonList(patientID));
		} catch (Exception e) {
			throw new TaskDataAccessException(e);
		}
		debug(result);

		List tasks = parseTasks(result);
		Collections.sort(tasks, comparator);
		return tasks;
	}

	public void saveTask(ITask task) throws TaskDataAccessException {
		try {
			if (task.getID() == null)
				saveNewTask(task);
			else
				saveExistingTask(task);
		} finally {
			// remove the task from the local cache.  This will force the task
			// to be reloaded from the server the next time it is retrieved,
			// ensuring that the correct information is displayed.
			if (task != null)
				taskCache.remove(task.getID());
		}
	}

	private void saveExistingTask(ITask task) throws TaskDataAccessException {
		String parameter = "TSK:" + task.getID() + "^" + buildSaveString(task);
		debug("Editing task '" + parameter + "'");

		String result = null;
		try {
			result = mumps.callRPC(EDIT_TASK_RPC_KEY, Collections.singletonList(Collections.singletonList(parameter)));
		} catch (Exception e) {
			throw new TaskDataAccessException(e);
		}
	}

	private void saveNewTask(ITask task) throws TaskDataAccessException {
		String parameter = buildSaveString(task);
		debug("Saving task '" + parameter + "'");

		String result = null;
		try {
			result = mumps.callRPC(CREATE_TASK_RPC_KEY, Collections.singletonList(parameter));
		} catch (Exception e) {
			throw new TaskDataAccessException(e);
		}
		// save the new ID to the task
		TaskID newID = resolveTaskID(result);
		task.setID(newID);
		debug("Created new task with ID: " + task.getID());
	}

	private String buildSaveString(ITask task) {
		StringBuffer param = new StringBuffer();
		appendString(param, task.getSubject());
		appendDate(param, task.getCreationDate(), CREATE_DATE_FORMAT);
		appendString(param, resolveUser(task.getCreatingUserID()));
		appendString(param, task.getPatientID());
		appendString(param, formatPriority(task.getPriority()));
		appendDate(param, task.getDueDate(), DUE_DATE_FORMAT);
		appendDate(param, task.getCompletionDate(), COMPLETION_DATE_FORMAT);
		appendString(param, resolveUser(task.getCompletingUserID()));
		appendDate(param, task.getDeletionDate(), DELETION_DATE_FORMAT);
		appendString(param, resolveUser(task.getDeletingUserID()));

		List linkedItems = task.getLinkedItems();
		if (linkedItems != null && !linkedItems.isEmpty()) {
			Iterator i = linkedItems.iterator();
			while (i.hasNext()) {
				ILinkedItem linkedItem = (ILinkedItem) i.next();
				String linkedSaveStr = formatLinkedItem(linkedItem);
				if (linkedSaveStr != null) {
					param.append(linkedSaveStr);
					if (i.hasNext())
						param.append(LINKED_ITEM_SEPARATOR);
				}
			}
		}
		String parameter = param.toString();
		return parameter;
	}

	private static final String EDIT_TASK_RPC_KEY = "TASK EDIT";
	private static final String CREATE_TASK_RPC_KEY = "TASK ADD";

	private String formatLinkedItem(ILinkedItem linkedItem) {
		if (linkedItem == null)
			return null;
		return linkedItem.getItemID();
	}

	protected String formatPriority(String p) {
		if (p == null)
			return null;
		return p.substring(0, 1).toUpperCase();
	}

	private final void appendString(StringBuffer buf, Object s) {
		if (s != null)
			buf.append(s.toString());
		buf.append("^");
	}
	private final void appendDate(StringBuffer buf, Date d, DateFormat f) {
		if (d != null)
			buf.append(f.format(d));
		buf.append("^");
	}

	protected static final String TASKS_BY_ID_RPC_KEY = "TASKS BY ID";
	protected static final String TASKS_BY_PATIENT_RPC_KEY = "TASKS BY PATIENT";
	protected static final DateFormat CREATE_DATE_FORMAT = new SimpleDateFormat("yyyyMMddHHmmss");
	protected static final DateFormat DUE_DATE_FORMAT = new SimpleDateFormat("yyyyMMddHHmm");
	protected static final DateFormat COMPLETION_DATE_FORMAT = CREATE_DATE_FORMAT;
	protected static final DateFormat DELETION_DATE_FORMAT = CREATE_DATE_FORMAT;
	//        new SimpleDateFormat("yyyyMMddHHmm");
	protected static final DateFormat[] DATE_FORMATS =
		{ new SimpleDateFormat("yyyyMMddHHmmss"), new SimpleDateFormat("yyyyMMddHHmm"), new SimpleDateFormat("yyyyMMddHH"), new SimpleDateFormat("yyyyMMdd"), };

	private void loadTasks(List taskIDs) throws TaskDataAccessException {
		List parameters = new LinkedList();
		Iterator i = taskIDs.iterator();
		while (i.hasNext()) {
			Object taskID = i.next();
			if (taskID != null)
				parameters.add("TSK:" + taskID);
		}
		debug("load tasks parameter: '" + parameters + "'");

		String result = null;
		try {
			result = mumps.callRPC(TASKS_BY_ID_RPC_KEY, Collections.singletonList(parameters));
		} catch (Exception e) {
			throw new TaskDataAccessException(e);
		}
		debug(result);

		parseTasks(result);
	}

	private List parseTasks(String rpcResult) {
		BufferedReader in = new BufferedReader(new StringReader(rpcResult));
		List result = new LinkedList();
		String oneTaskLine;

		try {
			while ((oneTaskLine = in.readLine()) != null) {
				String[] taskData = StringUtil.split(oneTaskLine, '^');
				if (taskData.length != 12) {
                    getLogger().error("Invalid task data received: " + oneTaskLine);
					continue;
				}

				AbstractTask task = newTaskObject();
				task.taskID = resolveTaskID(taskData[0]);
				task.subject = parseString(taskData[1]);
				task.creationDate = parseCreationDate(taskData[2]);
				task.creatorID = parseString(taskData[3]);
				task.patientID = parseString(taskData[4]);
				task.priority = parsePriority(taskData[5]);
				task.dueDate = parseDueDate(taskData[6]);
				task.completionDate = parseCompletionDate(taskData[7]);
				task.completingUserID = parseString(taskData[8]);
				task.deletionDate = parseCompletionDate(taskData[9]);
				task.deletingUserID = parseString(taskData[10]);

				StringTokenizer items = new StringTokenizer(taskData[11], LINKED_ITEM_SEPARATOR);
				while (items.hasMoreTokens())
					addLinkedItemToTask(task, items.nextToken());

				taskCache.put(task.taskID, task);
				result.add(task);
			}
		} catch (IOException e) {
			// can't happen - we're reading from a StringReader!!
		}

		return result;
	}

	private void addLinkedItemToTask(AbstractTask task, String itemID) {
		if (itemID == null || itemID.length() == 0)
			return;
		AbstractLinkedItem result = new AbstractLinkedItem();

		String type = taskPlugin.getTypeForID(itemID);
		if (type == null) {
            getLogger().debug("Item type not recognized: " + itemID);
			task.addLinkedItem(null, itemID);
		} else
			task.addLinkedItem(type, itemID);
	}

	private final String parseString(String s) {
		if (s == null || s.length() == 0)
			return null;
		else
			return s;
	}


	private String parsePriority(String priority) {
		int priorityPos = "LMH".indexOf(priority);
		if (priorityPos == -1) {
            getLogger().warning("Unrecognized priority " + priority);
			priorityPos = 0;
        }

		return PRIORITIES[priorityPos];
	}

	private Date parseCreationDate(String date) {
		try {
			return parseDate(date);
		} catch (ParseException e) {
            getLogger().error("Couldn't parse task creation date: " + e);
		}
		return null;
	}

	private Date parseCompletionDate(String completion) {
		try {
			return parseDate(completion);
		} catch (ParseException e) {
			getLogger().error("Couldn't parse task completion date: " + e);
		}
		return null;
	}

	private Date parseDueDate(String due) {
		try {
			return parseDate(due);
		} catch (ParseException e) {
            getLogger().error("Couldn't parse task due date: " + e);
		}
		return null;
	}

	private Date parseDate(String date) throws ParseException {
		if (date == null || date.length() == 0)
			return null;

		int pos = date.indexOf('-');
		if (pos != -1)
			date = date.substring(0, pos); // discard time zone

		for (int i = 0; i < DATE_FORMATS.length; i++)
			try {
				DateFormat format = DATE_FORMATS[i];
				return format.parse(date);
			} catch (ParseException e) {
			}
		throw new ParseException("Couldn't parse date: " + date, 0);
	}

    protected ILogger getLogger() {
        if (logger == null)
            logger = runtime.getLogger(MTaskDAO.class);
        return logger;
    }

	private final void debug(String msg) {
        getLogger().debug(msg);
	}

}
