package gov.va.med.hds.task.internal;

import java.io.IOException;
import java.net.URL;
import java.net.URLEncoder;
import java.text.MessageFormat;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import gov.va.med.hds.cd.runtime.IConfigurationElement;

public class LinkableItemType {
    
    static final String TYPE_ATTRIBUTE_NAME = "typeName";
    static final String URL_FORMAT_NAME = "urlFormat";
    static final String ID_REGEXP_NAME = "idRegexp";
    static final String HEADER_CONTENT_NAME = "htmlHeaderContent";  
    static final String VALUE_NAME = "value";

    private static final Integer FALSE = new Integer(0);
    private static final Integer TRUE = new Integer(1);
    
    String type;
    
    MessageFormat urlFormat;
    
    Pattern[] idRegexps;
    
    String htmlHeaderContent;
    
    public LinkableItemType(IConfigurationElement ce) throws Exception {
        
        // get the linkable item type.
        this.type = ce.getAttribute(TYPE_ATTRIBUTE_NAME);
        if (type == null || type.length() == 0)
            throw new Exception(
                "Missing linkable item type in extension for plugin '"
                    + ce.getDeclaringExtension().getUniqueIdentifier() + "'.");        
       
        // get the format for item URLs.
        String format = ce.getAttribute(URL_FORMAT_NAME);
        if (format == null || format.length() == 0)
            throw new Exception(
                "The " + URL_FORMAT_NAME + 
                " for linkable item type '" + type + "' is missing.");
        try {
            this.urlFormat = new MessageFormat(format);
        } catch (IllegalArgumentException e) {
            throw new Exception(
                "The " + URL_FORMAT_NAME + " '" + format
                    + "' for linkable item type '" + type
                    + "' is not valid.");
        }
        
        // get the regular expresions for linked item IDs.
        IConfigurationElement[] children;
        children = ce.getChildren(ID_REGEXP_NAME);
        if (children == null || children.length == 0)
            throw new Exception(
                "Missing " + ID_REGEXP_NAME + " for linkable item type '"
                    + type + "'.");
        idRegexps = new Pattern[children.length];
        for (int i = 0; i < children.length; i++)
            idRegexps[i] = parseIDRegexp(children[i]);
        
        // get the html header content for this linked item type.
        htmlHeaderContent = "";
        children = ce.getChildren(HEADER_CONTENT_NAME);
        if (children != null)
            for (int i = 0; i < children.length; i++)
                htmlHeaderContent += children[i].getValue();
    }
    
    
    private Pattern parseIDRegexp(IConfigurationElement ce) throws Exception {
        // get the regular expression for item IDs.
        String regex = ce.getAttribute(VALUE_NAME);
        if (regex == null || regex.length() == 0)
            throw new Exception(
                "Missing " + ID_REGEXP_NAME + " for linkable item type '"
                    + type + "'.");
        try {
            return Pattern.compile(regex);
        } catch (PatternSyntaxException e) {
            throw new Exception(
                "The " + ID_REGEXP_NAME + " '" + regex
                    + "' for linkable item type '" + type
                    + "' is not valid.");
        } 
    }

    public String getType() {
        return type;
    }
    
    public String getHtmlHeaderContent() {
        return htmlHeaderContent;
    }
    
    public boolean matchesID(String id) {
        for (int i = 0; i < idRegexps.length; i++) {
            if (idRegexps[i].matcher(id).matches()) return true;
        }
        return false;
    }
    
    public URL getURL(String itemId, String patientID, boolean collapse) {
        if (!matchesID(itemId)) return null;
        try {
            Object[] args = new Object[3];
            args[0] = URLEncoder.encode(itemId, "UTF-8");
            args[1] = URLEncoder.encode(patientID, "UTF-8");            
            args[2] = (collapse ? TRUE : FALSE);
            String url = urlFormat.format(args, new StringBuffer(), null).toString();        
            return new URL(url);
        } catch (IOException ioe) {}
        return null;
    }
    
}
