package gov.va.med.hds.cd.url.internal;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Map;

import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.url.IServlet;

public class CDServletURLConnection extends URLConnection {

    private IServlet servlet;
    private CDServletRequest request;
    private CDServletResponse response;

    private InputStream clientInputStream = null;
    private OutputStream clientOutputStream = null;

    private Map responseHeaders = null;

    private AsynchronousServerThread responseThread = null;
    private IOException exceptionThrown = null;

    public CDServletURLConnection(
        IServlet servlet,
        IRuntimeContext runtime,
        URL u,
        String id,
        String extraPath,
        Map parameters) {

        super(u);
        this.servlet = servlet;
        this.request =
            new CDServletRequest(runtime, u, id, extraPath, parameters);
        this.response = null;
    }

    public synchronized void connect() throws IOException {
        if (exceptionThrown != null)
            throw exceptionThrown;

        if (connected)
            return;

        if (getDoOutput())
            doAsynchronousConnect();
        else
            doSynchronousConnect();

        connected = true;
    }

    /** Connect to the servlet, and run it in a separate thread.
     */
    private void doAsynchronousConnect() throws IOException {
        if (connected)
            throw new IllegalStateException("cannot connect to content provider twice");

        PipedInputStream pis;
        PipedOutputStream pos;

        // construct the forward connection
        pis = new PipedInputStream();
        pos = new PipedOutputStream(pis);
        clientOutputStream = pos;
        request.setInputStream(pis);

        // start the servlet in a separate thread
        responseThread = new AsynchronousServerThread(getURL());
        responseThread.start();
    }

    /** Connect to the servlet and run it in the current thread.
     */
    private void doSynchronousConnect() throws IOException {
        if (connected)
            throw new IllegalStateException("cannot connect to content provider twice");

        ByteArrayOutputStream out = new ByteArrayOutputStream();
        response = new CDServletResponse(out, getURL());
        service();

        clientInputStream = new ByteArrayInputStream(out.toByteArray());
        clientOutputStream = NULL_OUTPUT_STREAM;
    }

    private void service() throws IOException {
        try {
            servlet.service(request, response);
        } finally {
            response.close();
        }
    }

    private class AsynchronousServerThread extends Thread {
        private URL base;
        public AsynchronousServerThread(URL base) {
            this.base = base;
        }
        public void run() {
            try {
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                response = new CDServletResponse(out, base);
                service();
                InputStream result = new ByteArrayInputStream(out.toByteArray());
                synchronized (CDServletURLConnection.this) {
                    clientInputStream = result;
                    responseThread = null;
                    CDServletURLConnection.this.notifyAll();
                }
            } catch (IOException ioe) {
                exceptionThrown = ioe;
            }
        }
    }

    public synchronized InputStream getInputStream() throws IOException {
        if (exceptionThrown != null)
            throw exceptionThrown;

        if (!connected)
            connect();
            
        if (clientOutputStream != null) try {
            clientOutputStream.close();
        } catch (IOException ioe) {
            // ignore the exception - the output stream may have already
            // been closed.
        }

        if (responseThread != null) try {
            this.wait();
        } catch (InterruptedException e) {}

        return clientInputStream;
    }

    public synchronized OutputStream getOutputStream() throws IOException {
        if (exceptionThrown != null)
            throw exceptionThrown;

        if (!connected)
            connect();

        return clientOutputStream;
    }

    public String getHeaderField(String name) {
        if (!connected)
            try {
                connect();
            } catch (IOException ioe) {
                exceptionThrown = ioe;
            }

        if (responseHeaders == null)
            responseHeaders = response.getResponseHeaders();

        return (String) responseHeaders.get(name.toLowerCase());
    }

    private static class NullOutputStream extends OutputStream {
        public void write(int b) throws IOException {
        }
    }
    private static final OutputStream NULL_OUTPUT_STREAM =
        new NullOutputStream();

}
