package gov.va.med.hds.cd.url;

import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;

import gov.va.med.hds.cd.runtime.IRuntimeContext;

/**
 * Interface for a dynamic content provider. 
 * 
 * An IContentProvider acts as a URLConnection factory.  The URLPlugin will
 * typically create only one instance of each IContentProvider object, and
 * immediately call its init method.  It will maintain a handle to the
 * resulting object for the remainder of the session, and repeatedly call its
 * connect method to obtain URLConnection objects for various URLs.
 * 
 * Implementing the IContentProvider interface allows plugin developers to
 * respond to arbitrary URLs that cannot be foreseen at compile time.  This
 * flexibility is rarely required, so most plugin developers should
 * consider implementing the {@link IServlet} interface instead.  (Also note
 * that, if you extend the <code>contentProvider</code> extension-point using
 * the <code>package</code> attribute, a default implementation of
 * {@link IContentProvider} will automatically be supplied for you.)
 */
public interface IContentProvider {

    /**
     * Initializes the content provider.
     * 
     * The IRuntimeContext passed to connect will always be the same object
     * originally passed to init.  This allows each dynamic content provider to
     * choose its preferred initialization policy.  If the dynamic content
     * provider needs to perform any significant setup, it can do so in the
     * init method. Otherwise, it can have an empty init method and simply do
     * all of the work in the connect method.
     *
     * @param runtime the runtime context of the enclosing desktop.
     * @throws IOException if an error occurs while initializing this object.
     */
    public void init(IRuntimeContext runtime) throws IOException;

    /**
     * Opens a connection to the object referenced by the URL argument.
     * 
     * The IRuntimeContext passed to connect will always be the same object
     * originally passed to init.  This allows each dynamic content provider to
     * choose its preferred initialization policy.  If the dynamic content
     * provider needs to perform any significant setup, it can do so in the
     * init method. Otherwise, it can have an empty init method and simply do
     * all of the work in the connect method.
     * 
     * @param runtime the runtime context of the enclosing desktop. 
     * @param u the <code>URL</code> that this connects to.
     * @return URLConnection a <code>URLConnection</code> object for the
     * <code>URL</code>.
     * @throws IOException if an error occurs while opening the connection.
     */
    public URLConnection connect(IRuntimeContext runtime, URL u, String id)
        throws IOException;
}
