package gov.va.med.hds.cd.url;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Map;

import gov.va.med.hds.cd.runtime.IRuntimeContext;


/**
 * Defines an object to provide client request information to a servlet. The
 * embedded servlet container creates a ServletRequest object and passes it as
 * an argument to the servlet's service method.
 * 
 * A ServletRequest object provides data including the desktop runtime
 * context, parameter name and values, and an input stream.
 */
public interface IServletRequest {

    
    /** Returns the runtime context of the enclosing desktop.
     * 
     * @return IRuntimeContext the runtime context that the servlet can use to
     * locate and access other objects in the desktop.
     */
    public IRuntimeContext getRuntimeContext();

    
    /** Returns the URL originally requested by the client.
     * 
     * @return URL the full URL used by the client to initiate this request.
     */
    public URL getURL();

    
    /** Returns the ID of this servlet.
     * 
     * This will typically be the path portion of the URL, up to and including
     * the name of this servlet.
     */
    public String getID();
    
    
    /** Returns a <code>Map</code> containing all of the parameters passed
     * to this request.
     * 
     * Parameters can come from two sources:<ul>
     * 
     * <li>Query parameters are specified as part of the URL.  For example, the
     * url <code>foo?bar=1&baz=two</code> has two query parameters: the first
     * parameter, &quot;bar&quot; has the value &quot;1&quot;, and the second
     * parameter, &quot;baz&quot; has the value &quot;two&quot;.  Parameters
     * passed in this way will automatically be parsed and available to
     * servlets in the <code>Map</code> returned by this method.</li>
     * 
     * <li>Parameters can be posted in the body of a request by an HTML form. 
     * If a servlet would like to access parameters passed by a client in this
     * way, it must first call {@link #parsePostedFormData()}.</li>
     * </ul>
     */
    public Map getParameters();

    
    /** Returns the value of a request parameter as a String, or null if the
     * parameter does not exist.
     * 
     * You should only use this method when you are sure the parameter has only
     * one value. If the parameter might have more than one value, use
     * {@link #getParameterList(String)}.  If you use this method with a
     * multivalued parameter, the value returned is equal to the last value in
     * the array returned by {@link #getParameterList(String)}.
     * 
     * If the parameter data was sent in the request body, such as occurs with
     * POST-ed form data request, then the servlet must call {@link
     * #parsePostedFormData()} before calling this method.
     * 
     * @param name the name of the parameter
     * @return String the single value of the parameter
     */
    public String getParameter(String name);

    
    /** Returns an array of String objects containing all of the values the
     * given request parameter has, or null if the parameter does not exist.
     * 
     * If the parameter has a single value, the array has a length of 1.
     * 
     * @param name the name of the parameter whose value is requested
     * @return String[] an array of <code>String</code> objects containing the
     * parameter's values
     */
    public String[] getParameterList(String name);
    
    
    /** Retrieves the body of the request as binary data.
     * 
     * @return InputStream an <code>InputStream</code> object containing the
     * body of the request, or <code>null</code> if there was no request body.
     * @throws IOException if an input or output exception occurred.
     */
    public InputStream getInputStream() throws IOException;


    /** Reads posted form data if it is present, and adds any parameters found
     * to the parameters <code>Map</code>.
     * 
     * If no form data was posted to this servlet, this method will do nothing.
     * 
     * Subsequent calls to {@link #getParameters()}, {@link #getParameter
     * (String)}, or {@link #getParameterList(String)} will reflect the
     * parameters found in posted form data.
     * 
     * Parameters should not generally be supplied both via query parameters
     * and POST data;  however, this method attempts to take the union of the
     * two.  <b>However</b>, if a parameter with the same name is specified in
     * both places, the value or values in the form data will take precedence
     * and replace the values provided in the query string.
     * 
     * @throws IOException
     */    
    public void parsePostedFormData() throws IOException;

}
