package gov.va.med.hds.cd.url;

import java.io.IOException;
import java.io.OutputStream;

/** Provides an interface for a servlet to sending a response to the client.
 * 
 * Servlets will typically interact with a servlet reponse in the following
 * way:</ol>
 * 
 * <li>The servlet will call {@link #setContentType(String)} to declare the
 * mime type of the content it will generate.
 * <li>The servlet may set the values of other response headers using the
 * {@link #setHeader(String, String)} method.
 * <li>The servlet can call {@link #getWriter()} and write text content to the
 * client.  Alternatively, the servlet can call {@link #getOutputStream()} and
 * send binary data to the client.
 * 
 * </ol>
 * 
 */
public interface IServletResponse {


    /** Set the content type for the generated content.
     * 
     * If a servlet does not call this method, the default content type
     * &quot;text/html&quot; will be assumed.
     * 
     * @param type the MIME type of the content (for example,
     * <code>text/plain</code> or <code>image/jpeg</code>).
     * @throws IllegalStateException if the content type has already been
     * committed and sent to the client.
     */
    public void setContentType(String type) throws IllegalStateException;


    /** Sets a response header with the given name and value.
     * 
     * If the header had already been set, the new value overwrites the
     * previous one. 
     * 
     * @param name the name of the header
     * @param value the value of the header
     * @throws IllegalStateException if the response headers have already been
     * committed and sent to the client.
     */
    public void setHeader(String name, String value) throws IllegalStateException;


    /** Returns a {@link ServletWriter} suitable for writing text to the client.
     * 
     * The writer returned can be used to conveniently write text, HTML, or XML
     * to a client.  The writer will use the default platform encoding to
     * translate characters to bytes.  If you need to send binary data, or if
     * you need to control the character encoding, call {@link
     * #getOutputStream()} instead.
     * 
     * @return ServletWriter a {@link ServletWriter} object that can send
     * character data to the client.
     */
    public ServletWriter getWriter();


    /** Returns an <code>OutputStream</code> for sending binary data to the client.
     * 
     * If a servlet calls both {@link #getWriter()} and {@link
     * #getOutputStream()}, it must manually call <code>flush()</code> on the
     * two streams to ensure that data is not interleaved.
     * 
     * @return OutputStream an <code>OutputStream</code> object that can send
     * binary data to the client.
     */    
    public OutputStream getOutputStream();


    /** Instructs the embedded servlet container to execute a different
     * embedded servlet, and send its content back to the client.
     * 
     * (Note: the calling servlet should <b>not</b> write any data either to
     * the writer or to the output stream before calling this method.)
     * 
     * @param location a URL, which will be interpreted relative to the URL of
     * this servlet
     * @throws IllegalStateException if response headers have already been
     * committed and sent to the client
     * @throws IOException if an error is encountered when dispatching to or
     * executing the named servlet.
     */
    public void performInternalRedirect(String location)
        throws IllegalStateException, IOException;

}
