package gov.va.med.hds.util.date;

import gov.va.med.hds.util.mumps.StringUtil;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * This class encapsulates some relative date methods.
 */
public class RelativeDate {

	/**
	 * Returns a date range description string from a relative date range string.
	 * The referenceDate is used to call the relativeToDate function.  
	 * @param referenceDate
	 * @param relativeDateRange
	 * @return dateRangeDisplayString
	 */
	public static String getDateRangeDisplayText(Date referenceDate, String relativeDateRange) {
		StringBuffer result = new StringBuffer("");
		SimpleDateFormat dateFormat = new SimpleDateFormat("dd MMM yyyy");
		GregorianCalendar calendar = new GregorianCalendar();
		calendar.setTime(referenceDate);
		
		if (relativeDateRange.compareTo(":")==0) result.append("any date");
		if (relativeDateRange.compareTo("T:T")==0) result.append("during Today");
		if (relativeDateRange.compareTo("T-1:T")==0) result.append("during the Past 24 Hours");
		if (relativeDateRange.compareTo("T-7:T")==0) result.append("during the Past Week");
		if (relativeDateRange.compareTo("T-30:T")==0) result.append("during the Past Month");
		if (relativeDateRange.compareTo("YFQ")==0) result.append("during the Current Quarter");
		if (relativeDateRange.compareTo("YF")==0) result.append("during the Current Year");
		
		if (result.length()<1 && (relativeDateRange.startsWith("Y"))) {
			result.append("during ");
			String quarter = StringUtil.piece(relativeDateRange,'Q',2);
			if (quarter.compareTo("1")==0) result.append("1st Quarter ");
			else if (quarter.compareTo("2")==0) result.append("2nd Quarter ");
			else if (quarter.compareTo("3")==0) result.append("3rd Quarter ");
			else if (quarter.compareTo("4")==0) result.append("4th Quarter ");
			String yearType = null;
			if (relativeDateRange.startsWith("YF")) yearType = "FY "; else yearType = "";
			String x1 = StringUtil.piece(relativeDateRange,'Q',1);
			String x2 = null;
			if (x1.length()>2) {
				x2 = x1.substring(2);
			}
			else {
				x2 = "0";
			}
			int offset = Integer.parseInt(x2.replaceAll("\\+",""));
			int currentMonth = calendar.get(GregorianCalendar.MONTH);
			int currentYear = calendar.get(GregorianCalendar.YEAR);
			if (relativeDateRange.startsWith("YF") && (currentMonth>8)) {
				currentYear++;
			}
			result.append(yearType);
			result.append(Integer.toString(currentYear + offset));
		}

		if (result.length()<1 && (relativeDateRange.indexOf(':')>-1)) {
			String leftDate  = StringUtil.piece(relativeDateRange,':',1);
			String rightDate = StringUtil.piece(relativeDateRange,':',2);
			if (leftDate.length()>0) {
				result.append("from ");
				result.append(dateFormat.format(RelativeDate.getRelativeToDate(referenceDate,leftDate)));
			}
			if (rightDate.length()>0) {
				if (result.length()>0) {
					result.append(" ");
				}
				result.append("through ");
				result.append(dateFormat.format(RelativeDate.getRelativeToDate(referenceDate,rightDate)));
			}
		}

		return result.toString();
	}

	/**
	 * Returns a relative date string based on the referenceDate and the inputDate.
	 * @param date
	 * @return String
	 */
	public static String getRelativeDateString(Date referenceDate, Date inputDate) {
		if (inputDate==null) return "";
		String result = "T";
		GregorianCalendar cal = new GregorianCalendar();
		int diff = 0;
		cal.setTime(referenceDate);
		int curYear = cal.get(GregorianCalendar.YEAR);
		int curDayOfYear = cal.get(GregorianCalendar.DAY_OF_YEAR);
		cal.setTime(inputDate);
		int inputYear = cal.get(GregorianCalendar.YEAR);
		int inputDayOfYear = cal.get(GregorianCalendar.DAY_OF_YEAR);
		if (inputYear==curYear) {
			diff = inputDayOfYear - curDayOfYear;
		}
		else {
			if (inputYear > curYear) {
				cal.setTime(referenceDate);
				cal.set(GregorianCalendar.MONTH, GregorianCalendar.DECEMBER);
				cal.set(GregorianCalendar.DAY_OF_MONTH, 31);
				int dayOfYear = cal.get(GregorianCalendar.DAY_OF_YEAR);
				diff = dayOfYear - curDayOfYear + inputDayOfYear;
				if (inputYear-curYear > 1) {
					for (int i = curYear+1; i<inputYear; i++) {
						diff += getNumberOfDaysInYear(i);
					}
				}
			}
			else {
				cal.setTime(inputDate);
				cal.set(GregorianCalendar.MONTH, GregorianCalendar.DECEMBER);
				cal.set(GregorianCalendar.DAY_OF_MONTH, 31);
				int dayOfYear = cal.get(GregorianCalendar.DAY_OF_YEAR);
				diff = -(dayOfYear - inputDayOfYear + curDayOfYear);
				if (curYear-inputYear > 1) {
					for (int i = inputYear+1; i<curYear; i++) {
						diff -= getNumberOfDaysInYear(i);
					}
				}
			}
		}
		
		if (diff!=0) {
			result += (diff<0) ? "-":"+";
			result += Integer.toString(Math.abs(diff));
		}
		return result;
	}

	/**
	 * Returns the integer number of days in the year, passed in as an integer.
	 * @param i
	 * @return int
	 */
	public static int getNumberOfDaysInYear(int i) {
		GregorianCalendar cal = new GregorianCalendar(i, GregorianCalendar.DECEMBER, 31);
		int result = cal.get(GregorianCalendar.DAY_OF_YEAR);
		return result;
	}
	
	/**
	 * Returns a java.util.Date based on the relativeDate and referenceDate
	 * passed in. A relative date is a date denoted with a T and then a modifier
	 * such as +6, -7, -30, etc.  Example: T-7 is the day a week ago from today.
	 * @param relativeDate
	 * @return Date
	 */
	public static Date getRelativeToDate(Date referenceDate, String relativeDate) {
		Date result = referenceDate;
		if (relativeDate.equals("T")) return result;
		
		String value = StringUtil.piece(relativeDate,'T',2);
		int increment = 0;
		if (value.charAt(0)=='+') {
			increment = Integer.parseInt(value.substring(1));
		}
		else {
			increment = Integer.parseInt(value);
		}
		if (increment != 0) {
			GregorianCalendar calendar = new GregorianCalendar();
			calendar.setTime(result);
			calendar.add(GregorianCalendar.DATE, increment);
			result = calendar.getTime();
		}

		return result;
	}

}
