package gov.va.med.hds.cd.workspace.internal;

import java.beans.BeanInfo;
import java.beans.EventSetDescriptor;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.runtime.logging.ILoggingCapable;
import gov.va.med.hds.cd.workspace.IMulticaster;
import gov.va.med.hds.cd.workspace.IWorkspaceScope;
import gov.va.med.hds.cd.workspace.InvalidEventListenerException;
import gov.va.med.hds.cd.workspace.InvalidEventSourceException;

public class Multicaster implements IMulticaster {

	private static final String REMOVED_SOURCE_INFO_MESSAGE = "removed instance of ''{0}'' as an ''{1}' source from scope ''{2}''";
	private static final String ADDED_SOURCE_INFO_MESSAGE = "added instance of ''{0}'' as an ''{1}'' source in scope ''{2}''";
	private static final String ADDED_LISTENER_INFO_MESSAGE = "added instance of ''{0}'' as an ''{1}'' listener in scope ''{2}''";
	private static final String REMOVED_LISTENER_INFO_MESSAGE = "removed instance of ''{0}'' as an ''{1}'' listener from scope ''{2}''";
	private ILogger log;
	private IWorkspaceScope scope;
	private Map listenerMap = new HashMap();
	private Map sourceMap = new HashMap();
	private Map forwardingListenerMap = new HashMap();

	public Multicaster(ILoggingCapable logging, IWorkspaceScope scope) {
		super();
		this.log = logging.getLogger(getClass());
		this.scope = scope;
	}

	public IWorkspaceScope getScope() {
		return scope;
	}

	public void addSource(Class eventType, Object eventSource)
		throws InvalidEventSourceException {
		validateSource(eventType, eventSource);
		List sources = (List) sourceMap.get(eventType);
		if (sources == null) {
			sources = new LinkedList();
			sourceMap.put(eventType, sources);
		}
		sources.add(eventSource);
		try {
			addForwardingListener(eventType, eventSource);
		} catch (IllegalArgumentException e) {
			throw new InvalidEventSourceException(e);
		} catch (IllegalAccessException e) {
			throw new InvalidEventSourceException(e);
		} catch (InvocationTargetException e) {
			throw new InvalidEventSourceException(e);
		}
		
		Object[] args = new Object[] { eventSource.getClass().getName(), eventType.getName(), scope.getId() };
		log.info(MessageFormat.format(ADDED_SOURCE_INFO_MESSAGE, args));
	}

	private void validateSource(Class eventType, Object eventSource) throws InvalidEventSourceException {
		EventSetDescriptor event = findEventSetDescriptor(eventType, eventSource);
		if (event == null) throw new InvalidEventSourceException("The specified instance of " + eventSource.getClass().getName() + " is not a source of " + eventType.getName() + " events");
	}

	private void addForwardingListener(Class eventType, Object eventSource) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		EventSetDescriptor event = findEventSetDescriptor(eventType, eventSource);
		Object forwardingListener = createForwardingListener(eventType, event.getListenerType());
		putForwardingListener(eventSource, event.getListenerType(), forwardingListener);
		event.getAddListenerMethod().invoke(eventSource, new Object[] { forwardingListener });
	}

	private void putForwardingListener(
		Object eventSource,
		Class listenerType,
		Object forwardingListener) {
		Map listeners = (Map) forwardingListenerMap.get(eventSource);
		if (listeners == null) {
			listeners = new HashMap();
			forwardingListenerMap.put(eventSource, listeners);
		}
		listeners.put(listenerType, forwardingListener);
	}

	private void removeForwardingListener(Class eventType, Object eventSource) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
			EventSetDescriptor event =
				findEventSetDescriptor(
					eventType,
					eventSource);
			Map forwardingListeners =
				(Map) forwardingListenerMap.get(eventSource);
			if (forwardingListeners == null)
				return;
			Object forwardingListener =
				forwardingListeners.get(event.getListenerType());
			if (forwardingListener == null)
				return;
			event.getRemoveListenerMethod().invoke(
				eventSource,
				new Object[] { forwardingListener });
	}

	private EventSetDescriptor findEventSetDescriptor(
		Class eventType,
		Object eventSource) {
		BeanInfo eventSourceBeanInfo;
		try {
			eventSourceBeanInfo =
				Introspector.getBeanInfo(eventSource.getClass());
		} catch (IntrospectionException e) {
			return null;
		}
		if (eventSourceBeanInfo == null)
			return null;
		String event = getEventName(eventType);
		EventSetDescriptor[] descriptors =
			eventSourceBeanInfo.getEventSetDescriptors();
		for (int i = 0; i < descriptors.length; i++) {
			if (descriptors[i].getName().equals(event))
				return descriptors[i];
		}
		return null;
	}

	public void removeSource(Class eventClass, Object eventSource) {
		List sources = (List) sourceMap.get(eventClass);
		if (sources == null)
			return;
		sources.remove(eventSource);
		try {
			removeForwardingListener(eventClass, eventSource);
		} catch (IllegalArgumentException e) {
			log.error("Unable to remove forwarding listener", e);
		} catch (IllegalAccessException e) {
			log.error("Unable to remove forwarding listener", e);
		} catch (InvocationTargetException e) {
			log.error("Unable to remove forwarding listener", e);
		}
		Object[] args = new Object[] { eventSource.getClass().getName(), eventClass.getName(), scope.getId() };
		log.info(MessageFormat.format(REMOVED_SOURCE_INFO_MESSAGE, args));
	}

	public void addListener(Class eventType, Object listener)
		throws InvalidEventListenerException {
		validateListener(eventType, listener);
		List listeners = (List) listenerMap.get(eventType);
		if (listeners == null) {
			listeners = new LinkedList();
			listenerMap.put(eventType, listeners);
		}
		listeners.add(listener);
		
		Object[] args = new Object[] { listener.getClass().getName(), eventType.getName(), scope.getId() };
		log.info(MessageFormat.format(ADDED_LISTENER_INFO_MESSAGE, args));
	}

	private void validateListener(Class eventType, Object listener) throws InvalidEventListenerException {
		if (!isListenerOf(eventType, listener)) throw new InvalidEventListenerException("The specified instance of " + listener.getClass().getName() + " is not a listener of " + eventType.getName() + " events");
	}

	private boolean isListenerOf(Class eventType, Object listener) {
		Class listenerType = listener.getClass();
		Class[] listenerTypes = getListenerTypes(eventType);
		for (int i = 0; i < listenerTypes.length; i++) {
			if (listenerTypes[i].isAssignableFrom(listenerType)) return true;
		}
		return false;
	}

	private Class[] getListenerTypes(Class eventType) {
		List classes = new LinkedList();
		String packageName = getPackageName(eventType);
		String eventName = getEventName(eventType);
		ClassLoader classLoader = eventType.getClassLoader();
		try {
			Class c = classLoader.loadClass(getListenerInterfaceName(eventType));
			classes.add(c);
		} catch (ClassNotFoundException e) {
			// TODO_SB: Auto-generated catch block
		}
		try {
			Class c = classLoader.loadClass(getIListenerInterfaceName(eventType));
			classes.add(c);
		} catch (ClassNotFoundException e) {
			// TODO_SB: Auto-generated catch block
		}
		return (Class[]) classes.toArray(new Class[classes.size()]);
	}
	
	private String getListenerInterfaceName(Class eventType) {
		return getPackageName(eventType) + "." + capitalize(getEventName(eventType)) + "Listener";
	}
	
	private String getIListenerInterfaceName(Class eventType) {
		return getPackageName(eventType) + ".I" + capitalize(getEventName(eventType)) + "Listener";
	}
	
	private String getPackageName(Class eventClass) {
		return eventClass.getName().substring(0, eventClass.getName().lastIndexOf('.'));
	}
	
	private String getEventName(Class eventClass) {
		String eventClassName = eventClass.getName();
		eventClassName =
			eventClassName.substring(eventClassName.lastIndexOf('.') + 1);
		return decapitalize(
			eventClassName.substring(0, eventClassName.length() - 5));
	}

	public void removeListener(Class eventType, Object listener) {
		List listeners = (List) listenerMap.get(eventType);
		if (listeners == null) return;
		listeners.remove(listener);
		if (listeners.isEmpty()) listenerMap.remove(eventType);
		
		Object[] args = new Object[] { listener.getClass().getName(), eventType.getName(), scope.getId() };
		log.info(MessageFormat.format(REMOVED_LISTENER_INFO_MESSAGE, args));
	}

	private Object createForwardingListener(
		Class eventType,
		Class listenerClass) {
		return Proxy.newProxyInstance(
			listenerClass.getClassLoader(),
			new Class[] { listenerClass },
			new ForwardingInvoker(eventType));
	}

	private Object[] getListeners(Class eventType) {
		List listeners = (List) listenerMap.get(eventType);
		if (listeners == null)
			return new Object[0];
		return listeners.toArray();
	}

	/**
	 * Utility method to take a string and convert it to normal Java variable
	 * name capitalization.  This normally means converting the first
	 * character from upper case to lower case, but in the (unusual) special
	 * case when there is more than one character and both the first and
	 * second characters are upper case, we leave it alone.
	 * <p>
	 * Thus "FooBah" becomes "fooBah" and "X" becomes "x", but "URL" stays
	 * as "URL".
	 *
	 * @param  name The string to be decapitalized.
	 * @return  The decapitalized version of the string.
	 */
	private static String decapitalize(String name) {
		if (name == null || name.length() == 0) {
			return name;
		}
		if (name.length() > 1
			&& Character.isUpperCase(name.charAt(1))
			&& Character.isUpperCase(name.charAt(0))) {
			return name;
		}
		char chars[] = name.toCharArray();
		chars[0] = Character.toLowerCase(chars[0]);
		return new String(chars);
	}

	private static String capitalize(String name) {
		if (name == null || name.length() == 0) {
			return name;
		}
		if (name.length() > 1 && Character.isUpperCase(name.charAt(0))) {
			return name;
		}
		char chars[] = name.toCharArray();
		chars[0] = Character.toUpperCase(chars[0]);
		return new String(chars);
	}

	private class ForwardingInvoker extends DefaultInvoker {

		private Class eventType;

		public ForwardingInvoker(Class eventType) {
			this.eventType = eventType;
		}

		public Object invoke(Object proxy, Method method, Object[] args)
			throws Throwable {
			Object[] listeners = getListeners(eventType);
			for (int i = 0; i < listeners.length; i++) {
				method.invoke(listeners[i], args);
			}
			return super.invoke(proxy, method, args);
		}
	}
}
