package gov.va.med.hds.cd.workspace.internal;

import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;

import javax.swing.event.EventListenerList;

import gov.va.med.hds.cd.runtime.IPlugin;
import gov.va.med.hds.cd.runtime.IPluginDescriptor;
import gov.va.med.hds.cd.runtime.IRuntimeContext;
import gov.va.med.hds.cd.runtime.logging.ILogger;
import gov.va.med.hds.cd.workspace.EventRegistrar;
import gov.va.med.hds.cd.workspace.IWorkspace;
import gov.va.med.hds.cd.workspace.IWorkspaceListener;
import gov.va.med.hds.cd.workspace.IWorkspaceScope;
import gov.va.med.hds.cd.workspace.WorkspaceEvent;

public class Workspace implements IPlugin, IWorkspace {

	private IRuntimeContext runtime;
	private ILogger log;
	private EventListenerList listenerList = new EventListenerList();
	private Map scopes = new HashMap();
	private ModelRegistry modelRegistry;

	public Workspace() {
		addWorkspaceListener(new ModelRegistrar(SESSION_SCOPE));
		addWorkspaceListener(new ModelRegistrar(PATIENT_SCOPE));
		addWorkspaceListener(new ModelRegistrar(PERSPECTIVE_SCOPE));
		addWorkspaceListener(new ModelRegistrar(CUSTOM_SCOPE));
	}
	
	public void addWorkspaceListener(IWorkspaceListener l) {
		listenerList.add(IWorkspaceListener.class, l);
	}

	public void removeWorkspaceListener(IWorkspaceListener l) {
		listenerList.remove(IWorkspaceListener.class, l);
	}

	protected void fireWorkspaceScopeAdded(IWorkspaceScope scope) {
		final WorkspaceEvent event = new WorkspaceEvent(this, scope);
		IWorkspaceListener[] listeners =
			(IWorkspaceListener[]) listenerList.getListeners(
				IWorkspaceListener.class);
		for (int i = 0; i < listeners.length; i++) {
			listeners[i].workspaceScopeAdded(event);
		}
	}

	protected void fireWorkspaceScopeRemoved(IWorkspaceScope scope) {
		final WorkspaceEvent event = new WorkspaceEvent(this, scope);
		IWorkspaceListener[] listeners =
			(IWorkspaceListener[]) listenerList.getListeners(
				IWorkspaceListener.class);
		for (int i = 0; i < listeners.length; i++) {
			listeners[i].workspaceScopeRemoved(event);
		}
	}

	public void addScope(String scopeId) {
		if (scopes.containsKey(scopeId))
			return; // throw new IllegalArgumentException("dispatch scope ");
		if (isReservedScope(scopeId))
			return; // throw new IllegalArgumentException("dispatch scope ");
		scopes.put(scopeId, new WorkspaceScope(runtime, scopeId));
		if (log != null)
			log.info("added workspace scope: " + scopeId);
		fireWorkspaceScopeAdded(getScope(scopeId));
	}

	public void removeScope(String scopeId) {
		if (!scopes.containsKey(scopeId))
			return;
		IWorkspaceScope scope = getScope(scopeId);
		scopes.remove(scopeId);
		if (log != null)
			log.info("removed workspace scope: " + scopeId);
		fireWorkspaceScopeRemoved(scope);
	}

	public void startup(IRuntimeContext runtime, IPluginDescriptor plugin)
		throws Exception {
		this.runtime = runtime;
		log = runtime.getLogger(getClass());
		createModelRegistry();
	}

	public void shutdown() throws Exception {
		scopes = null;
	}

	private boolean isReservedScope(String scopeId) {
		return false;
	}

	public IWorkspaceScope getScope(String scopeId) {
		return (IWorkspaceScope) scopes.get(scopeId);
	}

	private ModelRegistry getModelRegistry() {
		return modelRegistry;
	}

	private void createModelRegistry() {
		ModelRegistryReader reader = new ModelRegistryReader(runtime);
		this.modelRegistry =
			reader.readModelRegistry(runtime.getPluginRegistry());
	}

	private void createModelsInScope(
		IWorkspaceScope scope,
		IModelDescriptor[] models) {
		for (int i = 0; i < models.length; i++) {
			Object model = null;
			try {
				model = models[i].createModel();
				scope.getModelSpace().putModel(models[i].getId(), model);
				Object[] args =
					new Object[] {
						models[i].getId(),
						models[i].getClassName(),
						scope.getId()};
				log.info(MessageFormat.format("created model ''{0}'' (instance of ''{1}'') in workspace scope ''{2}''", args));
			} catch (Exception e) {
				Object[] args =
					new Object[] {
						models[i].getId(),
						models[i].getClassName(),
						scope.getId()};
				log.error(
					MessageFormat.format(
						"error creating model ''{0}'' (instance of ''{1}'') in workspace scope ''{2}''",
						args),
					e);
			}
			if (model == null)
				return;
			try {
				EventRegistrar.connectSources(
					scope.getMulticaster(),
					model,
					models[i].getSourceOfEventDescriptors());
				EventRegistrar.connectListeners(
					scope.getMulticaster(),
					model,
					models[i].getListenerOfEventDescriptors());
			} catch (Exception e) {
				Object[] args =
					new Object[] {
						models[i].getId(),
						models[i].getClassName(),
						scope.getId()};
				log.error(
					MessageFormat.format(
						"error connecting events for model ''{0}'' (instance of ''{1}'') in worksace scope ''{2}''",
						args),
					e);
			}
		}
	}

	private class ModelRegistrar implements IWorkspaceListener {

		private String scopeCategory;
		
		public ModelRegistrar(String scope) {
			this.scopeCategory = scope;
		}

		public void workspaceScopeAdded(WorkspaceEvent e) {
			if (!isScopeCategory(e.getScope()))
				return;
			IModelDescriptor[] models =
				getModelRegistry().getModelDescriptorsInScope(scopeCategory);
			createModelsInScope(e.getScope(), models);
		}

		public void workspaceScopeRemoved(WorkspaceEvent e) {
			if (!isScopeCategory(e.getScope()))
				return;
		}

		private boolean isScopeCategory(IWorkspaceScope s) {
			return s.getId().startsWith(scopeCategory);
		}
	}
}
