package junit.extensions.jfcunit;

import java.awt.Toolkit;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.security.Permission;

import junit.framework.TestCase;

/**
 * Extend this class to create new tests for Swing based interfaces. This is a
 * subclass of TestCase, and therefore provides all of the facilities that you
 * would normally expect from TestCase.
 *
 * An important point to realize about coding with Swing classes is that most
 * methods are single threaded. This means that once a component has been
 * shown, its methods should normally only be accessed by the AWT thread.
 * While JFCTestCase runs its tests, the AWT thread is temporarily blocked to
 * prevent multi-threading issues. The effect of this is that any method calls
 * that you make on Swing components during a test will have no impact on the
 * GUI, until the AWT Thread is restarted. This can occur in one of two ways:
 * (1) The test runs to completion (2) The method "awtSleep" defined within
 * this class is called.
 *
 * @author Matt Caswell
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public class JFCTestCase extends TestCase {
    /**
     * Time to wait between lock attempts.
     */
    private static final long LOCKWAIT = 25L;

    /**
     * Default time to sleep
     */
    private static final long DEFAULTSLEEP = 100L;

    /**
     * Forcibly wait for full time
     */
    private boolean forcedWait = false;

    /**
     * Continuation flag.
     */
    private boolean cont = false;

    /**
     * Run the AWTThread free from interruption.
     */
    private volatile boolean pausingAWT = false;

    /**
     * Exception thrown by testcase.
     */
    private Throwable err;

    /**
     * Lock instance.
     */
    private Object lock = new Object();

    /**
     * Time to sleep.
     */
    private long sleepTime = DEFAULTSLEEP;

    /**
     * Waiting flag
     */
    private volatile boolean waiting = false;

    /**
     * The test helper to be used in testing. This has been made protected,
     * since most users already have a super class in their test hierarchy
     * where it is protected. Make sure to actually instantiate this in your
     * setUp() method or at the beginning of the test methods.
     */
    protected TestHelper helper;

    /*
     * Kick Start the WindowMonitor to hopefully
     * Get it started before any popups are displayed.
     */
    static {
        WindowMonitor.start();
    }

    /**
     * Constructs a new JFCTestCase (default Constructor if using JUnit ver 3.8)
     */
    public JFCTestCase() {
        super();
    }

    /**
     * Constructs a new JFCTestCase
     *
     * @param name The name of the test
     */
    public JFCTestCase(String name) {
        super(name);
    }

    /**
     * Default setUp which does nothing. Override this to set up the environment
     * for your test
     *
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void setUp() throws Exception {
        super.setUp();
    }

    /**
     * Default tearDown which does nothing. Override this to tear down the
     * environment for your test.
     *
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    /**
     * Sets the helper.
     *
     * @param helper TestHelper to be used.
     */
    public void setHelper(TestHelper helper) {
        this.helper = helper;
    }

    /**
     * Get the test helper.
     *
     * @return The test helper.
     */
    public TestHelper getHelper() {
        return helper;
    }

    /**
     * Sets the forcedWait
     *
     * @param forced New value for forcedWait
     */
    protected void setForcedWait(boolean forced) {
        forcedWait = forced;
    }

    /**
     * Resets the forcedWait to the default value (false)
     */
    protected void resetForcedWait() {
        forcedWait = false;
    }

    /**
     * Sets the sleepTime
     *
     * @param time New value for sleepTime
     */
    public void setSleepTime(long time) {
        sleepTime = time;
    }

    /**
     * Resets the sleepTime to the default value (DEFAULTSLEEP)
     */
    public void resetSleepTime() {
        sleepTime = DEFAULTSLEEP;
    }

    /**
     * Returns the value of sleepTime
     *
     * @return sleepTime
     */
    private long getSleepTime() {
        return sleepTime;
    }

    /**
     * Sets the cont
     *
     * @param c New value for cont
     */
    private void setContinue(boolean c) {
        cont = c;
    }

    /**
     * Returns the cont
     *
     * @return cont
     */
    private boolean getContinue() {
        return cont;
    }

    /**
     * Sets the err
     *
     * @param error New value for err
     */
    protected void setError(Throwable error) {
        err = error;
    }

    /**
     * Resets the err to the default value (null)
     */
    protected void resetError() {
        setError(null);
    }

    /**
     * Checks of the current test case has any errors
     *
     * @return true if the current test case has any errors
     */
    protected boolean hasError() {
        return (err != null && !(err instanceof ExitException));
    }

    /**
     * Returns the error
     *
     * @return error
     */
    protected Throwable getError() {
        return err;
    }

    /**
     * Suspends the test for up to a maximum period of time, and allows the
     * AWT Thread to resume temporarily. If the AWT Thread has not been paused,
     * this api has no effect. The time period is what was set by the user by
     * calling 'setSleepTime()' or 'awtSleep(sleepTime)'. If the notify occurs
     * before this time has elapsed, the thread will continue.
     */
    public final void awtSleep() {
        release();
        try {
            acquire();
        } catch (ExitException exe) {
            setError(exe);
        } catch (InterruptedException ie) {
            // shouldn't occur
        }
    }

    /**
     * Suspends the test for up to the specified (maximum) period of time, and allows the AWT
     * Thread to resume temporarily. If the AWT Thread has not been paused, this api has
     * no effect. Note that the sleep time for subsequent calls to awtSleep()
     * will use the time specified in the current call. If the notify occurs before
     * this time has elapsed, the thread will continue.
     *
     * @param sleepTime The number of ms to sleep
     */
    public final void awtSleep(long sleepTime) {
        setSleepTime(sleepTime);
        awtSleep();
    }

    /**
     * Suspends the test for a period of time, and allows the
     * AWT Thread to resume during this period if it has been
     * paused.
     *
     * @param delay The minimum amount of time the test case thread
     *              should be delayed.
     */
    public final void sleep(long delay) {
        long releaseTime = System.currentTimeMillis() + delay;
        releaseTime += delay;
        while (System.currentTimeMillis() < releaseTime) {
            release();
            try {
                Thread.currentThread().sleep(LOCKWAIT);
                acquire();
            } catch (InterruptedException ie) {
                // shouldn't occur
            }
        }
    }

    /**
     * Release the AWT Thread Control
     */
    private void release() {
        // when any events (usually window close events) try to call System.exit,
        // we need to trap the exception here
        try {
            lock.notify();
        } catch (ExitException exe) {
            // don't do anything here
        }
    }

    /**
     * Acquire AWT Thread Control
     *
     * @exception InterruptedException  Thrown when a thread is waiting, sleeping, or otherwise paused
     *                                  for a long time and another thread interrupts it using the
     *                                  <code>interrupt</code> method in class <code>Thread</code>.
     */
    private void acquire() throws InterruptedException {
        // when any events (usually window close events) try to call System.exit,
        // we need to trap the exception here
        try {
            if (pausingAWT) {
                lock.wait();
            } else {
                lock.wait(LOCKWAIT);
            }
        } catch (ExitException exe) {
            // don't do anything here
        }
    }

    /**
     * Pause the awt event queue until it is released by
     * releaseAWT() or the end of the test is reached.
     */
    public final void pauseAWT() {
        pausingAWT = true;
        awtSleep();
    }

    /**
     * Resume the awt event queue.
     */
    public final void resumeAWT() {
        awtSleep();
        pausingAWT = false;
        awtSleep();
    }

    /**
     * Flush all events currently in the AWTEventQueue.
     * The event queue is left running afterwards.
     */
    public final void flushAWT() {
        pauseAWT();
        resumeAWT();
    }

    /**
     * Get the AWT state.
     *
     * @return  boolean specifying whether the AWT Thread is still running or not.
     */
    public boolean isAWTRunning() {
        return !pausingAWT;
    }

    /**
     * Executes a test
     *
     * @exception Throwable exceptions thrown by code.
     */
    protected void runTest() throws Throwable {
        final Method runMethod;

        try {
            runMethod = getClass().getMethod(getName(), new Class[0]);
        } catch (NoSuchMethodException e) {
            fail("Method \"" + getName() + "\" not found");
            return;
        }
        if (runMethod != null && !Modifier.isPublic(runMethod.getModifiers())) {
            fail("Method \"" + getName() + "\" should be public");
            return;
        }

        runCode(new Runnable() {
            public void run() {
                try {
                    runMethod.invoke(JFCTestCase.this, new Object[0]);
                    flushAWT();
                } catch (ExitException exe) {
                    // might need it here due to the method invoked
                    setError(exe);
                } catch (InvocationTargetException e) {
                    setError(e.getTargetException());
                } catch (IllegalAccessException e) {
                    setError(e);
                }
            }
        });
    }

    /**
     * Sets up, executes and then tears down a test
     *
     * @exception Throwable exceptions thrown by code.
     */
    public void runBare() throws Throwable {
        SecurityManager oldsm = System.getSecurityManager();
        System.setSecurityManager(createNoExitSecurityManager());

        try {
            try {
                runCode(new Runnable() {
                    public void run() {
                        try {
                            JFCTestCase.this.setUp();
                            flushAWT();
                        } catch (ExitException exe) {
                            // might need it here due to the setUp method invocation
                            setError(exe);
                        } catch (Exception e) {
                            setError(e);
                        }
                    }
                });

                try {
                    runTest();
                } catch (ExitException exe) {
                    // might need it here due to the test being run
                    setError(exe);
                }
            } finally {
                runCode(new Runnable() {
                    public void run() {
                        try {
                            resumeAWT();
                            JFCTestCase.this.tearDown();
                            flushAWT();
                        } catch (ExitException exe) {
                            // might need it here due to the tearDown method invocation
                            setError(exe);
                        } catch (Exception e) {
                            setError(e);
                        }
                    }
                });
            }
        } finally {
            // put back the original security manager
            System.setSecurityManager(oldsm);
        }
    }

    /**
     * Note: This method will wait till the thread running code.run() completes
     * The above behavior is overridden if the setForcedWait() has been called with 'true'
     * Then the behavior is that the thread will wait for the time specified when awtSleep()
     * was called
     *
     * @param code Code which is to be executed.
     * @exception Throwable exceptions thrown by code.
     */
    protected synchronized void runCode(final Runnable code) throws Throwable {
        waiting = false;
        setContinue(true);
        resetError();
        new Thread(new Runnable() {
            public void run() {
                synchronized (lock) {
                    waiting = true;
                    try {
                        acquire();
                    } catch (InterruptedException ie) {
                        // shouldn't occur
                    }
                    try {
                        code.run();
                    } catch (ExitException exe) {
                        // might need it here due to the method invocation
                        setError(exe);
                    }
                    setContinue(false);
                    release();
                }
            }
        }, getName()).start();

        // here we have to wait for the code.run() to block
        while (!waiting) {
            try {
                Thread.currentThread().sleep(LOCKWAIT);
            } catch (ExitException exe) {
                // might need it here due to the method invocation
                setError(exe);
            } catch (InterruptedException ie) {
                // shouldn't occur
            }
        }

        boolean condition = false;
        boolean firstTime = true;
        long elapsedTime = 0L;
        do {
            if (firstTime && pausingAWT) {
                // flush out any other dialog boxes, etc which were opened by the call to code.run()
                Toolkit.getDefaultToolkit().getSystemEventQueue().invokeAndWait(new Runnable() {
                    public void run() {
                        synchronized (lock) {
                            release();
                            try {
                                acquire();
                            } catch (InterruptedException ie) {
                                // shouldn't occur
                            }
                        }
                    }
                });

                if (forcedWait) {
                    firstTime = false;
                }
            }

            try {
                elapsedTime += LOCKWAIT;
                Thread.currentThread().sleep(LOCKWAIT);
            } catch (InterruptedException ie) {
                // shouldn't occur
            }

            if (forcedWait) {
                condition = (elapsedTime <= getSleepTime());
            } else {
                condition = getContinue();
            }
        } while (condition);

        if (hasError()) {
            throw getError();
        }
    }

    /**
     * This method creates a replacement for the default system manager.
     * The goal is to intercept System.exit calls and make it throw an
     * exception instead so that a System.exit in a task does not
     * fully terminate the test run.
     *
     * @return  An instance of a SecurityManager that will "consume" a call to System.exit()
     */
    protected final SecurityManager createNoExitSecurityManager() {
        return new SecurityManager() {
            /**
             * @param status Status to be used.
             * @see java.lang.SecurityManager#checkExit(int)
             */
            public void checkExit(int status) {
                throw new ExitException(status);
            }

            /**
             * @param perm Permission to be checked.
             * @see java.lang.SecurityManager#checkPermission(java.security.Permission)
             */
            public void checkPermission(Permission perm) {
                // no permission to be validated
            }
        };
    }
}