package junit.extensions.jfcunit;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.lang.reflect.Constructor;
import junit.extensions.jfcunit.keyboard.JFCKeyStroke;


/**
 * Class that provides facilities for locating components within a GUI. To use
 * create a new instance of JFCTestHelper in your setUp. Windows can only be
 * located once they have been shown once to the user. Provide methods for
 * entering clicks and drag operations.
 *
 * Note: java.awt.dnd functions are tightly integrated with the OS and are
 * not dependent on the AWTEventQueue processing. Thus these interactions
 * can only be done with the RobotTestHelper which sends events in at the
 * OS level.
 *
 * @author Matt Caswell
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public class JFCTestHelper extends TestHelper {
    // though this is a helper class, these "instance" variables are used
    // to "display" the difference between the previous "state" and the
    // current one.
    /**
     * The x-coordinate where the last event was fired.
     */
    private int lastX = 0;

    /**
     * The y-coordinate where the last event was fired.
     */
    private int lastY = 0;

    /**
     * The mouse buttons that were pressed in the previous call.
     */
    private long lastPressed = 0;

    /**
     * The mouse buttons that were pressed in the previous call.
     */
    private int lastMouseModifiers = 0;

    /**
     * The popup that appeared in the previous call.
     */
    private boolean lastPopup = false;

    /** Last number of clicks */
    private int lastClicks;

    /**
     * The EventQueue on which all events are processed.
     */
    private EventQueue queue;

    /**
     * Constructor
     */
    public JFCTestHelper() {
        super();
        queue = Toolkit.getDefaultToolkit().getSystemEventQueue();
    }

    /**
     * Process a key press event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   stroke      The JFCKeyStroke to be performed.
     */
    protected void keyPressed(Component ultimate, JFCKeyStroke stroke) {
        postEvent(new KeyEvent(ultimate, KeyEvent.KEY_PRESSED, System.currentTimeMillis(), stroke.getModifiers(), stroke.getKeyCode(), stroke.getKeyChar()));
    }

    /**
     * Process a key release event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   stroke      The JFCKeyStroke to be performed.
     */
    protected void keyReleased(Component ultimate, JFCKeyStroke stroke) {
        if (stroke.isTyped()) {
            postEvent(new KeyEvent(ultimate, KeyEvent.KEY_TYPED, System.currentTimeMillis(), stroke.getModifiers(), KeyEvent.VK_UNDEFINED, stroke.getKeyChar()));
        }
        postEvent(new KeyEvent(ultimate, KeyEvent.KEY_RELEASED, System.currentTimeMillis(), stroke.getModifiers(), stroke.getKeyCode(), stroke.getKeyChar()));
    }

    /**
     * Process a mouse move event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   x           The x coordinate of the point where the mouse is being moved to.
     * @param   y           The y coordinate of the point where the mouse is being moved to.
     */
    protected void mouseMoved(Component ultimate, int x, int y) {
        int fx = x - ultimate.getLocationOnScreen().x;
        int fy = y - ultimate.getLocationOnScreen().y;

        int deltaX = (lastX < fx) ? 1 : -1;
        int deltaY = (lastY < fy) ? 1 : -1;

        int mouseEventType = (lastPressed == 0) ? MouseEvent.MOUSE_MOVED : MouseEvent.MOUSE_DRAGGED;
        while (lastX != fx || lastY != fy) {
            lastX = (lastX == fx) ? fx : lastX + deltaX;
            lastY = (lastY == fy) ? fy : lastY + deltaY;
            postEvent(new MouseEvent(ultimate, mouseEventType, System.currentTimeMillis(), lastMouseModifiers, lastX, lastY, 0, lastPopup));
        }
    }

    /**
     * Process a mouse press event on a component.
     *
     * @param   ultimate          The ultimate parent Component
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     */
    protected void mousePressed(Component ultimate, int modifiers, int click, boolean isPopupTrigger) {
        lastPressed = System.currentTimeMillis();
        lastPopup = isPopupTrigger;
        lastMouseModifiers = modifiers;
        if (modifiers != 0) {
            postEvent(new MouseEvent(ultimate, MouseEvent.MOUSE_PRESSED, lastPressed, modifiers, lastX, lastY, click, isPopupTrigger));
        }
    }

    /**
     * Process a mouse release event on a component.
     *
     * @param   ultimate          The ultimate parent Component
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     */
    protected void mouseReleased(Component ultimate, int modifiers, int click, boolean isPopupTrigger) {
        lastMouseModifiers = modifiers;
        lastPopup = isPopupTrigger;
        lastClicks = click;
        if (modifiers != 0) {
            postEvent(new MouseEvent(ultimate, MouseEvent.MOUSE_RELEASED, System.currentTimeMillis(), modifiers, lastX, lastY, click, isPopupTrigger));
            long delta = lastPressed - System.currentTimeMillis();
            if (click > 0 && delta < 100) {
                postEvent(new MouseEvent(ultimate, MouseEvent.MOUSE_CLICKED, System.currentTimeMillis(), modifiers, lastX, lastY, click, isPopupTrigger));
            }
        }
        lastPressed = 0;
    }

    /**
     * Simulate rotating the mouseWheel
     * Only supported in Java 1.4.
     *
     * @param ultimate Component to fire the events on.
     * @param amount Amount to rotate the wheel.
     * @param wheelRotation amount to rotate the wheel. positive for
     * rotation towards the user. Negative for rotation away from the user.
     * @exception Exception will be thrown if the JDK version
     * is less than 1.4.
     */
    protected void mouseWheel(Component ultimate, int amount, int wheelRotation) throws Exception {
        // This class is coded with reflection to allow the class
        // to compile on versions of the JDK less than 1.4.
        Class mouseWheelEventClass = Class.forName("java.awt.event.MouseWheelEvent");
        Constructor constructor =
            mouseWheelEventClass.getConstructor(
                new Class[] { Component.class, int.class, long.class, int.class, int.class, int.class, int.class, boolean.class, int.class, int.class, int.class });

        int scrollType = 0;

        int event_id = mouseWheelEventClass.getField("MOUSE_WHEEL").getInt(null);
        if ((lastMouseModifiers & MouseEvent.SHIFT_MASK) > 1) {
            scrollType = mouseWheelEventClass.getField("WHEEL_BLOCK_SCROLL").getInt(null);
        } else {
            scrollType = mouseWheelEventClass.getField("WHEEL_UNIT_SCROLL").getInt(null);
        }
        postEvent(
            (AWTEvent) constructor.newInstance(
                new Object[] {
                    ultimate,
                    new Integer(event_id),
                    new Long(System.currentTimeMillis()),
                    new Integer(lastMouseModifiers),
                    new Integer(lastX),
                    new Integer(lastY),
                    new Integer(lastClicks),
                    new Boolean(lastPopup),
                    new Integer(scrollType),
                    new Integer(amount),
                    new Integer(wheelRotation)}));

        //new MouseWheelEvent(Component source,
        //                    int id,
        //                    long when,
        //                    int modifiers,
        //                    int x,
        //                    int y,
        //                    int clickCount,
        //                    boolean popupTrigger,
        //                    int scrollType,
        //                    int scrollAmount,
        //                    int wheelRotation

        //throw new RuntimeException("Method is not supported by abstract class: mouseWheel");
    }

    /**
     * This method is just present so as to put debug statements in one central place,
     * without repeating everywhere.
     *
     * @param evt    The event to be posted.
     */
    private void postEvent(AWTEvent evt) {
        queue.postEvent(evt);
    }
}