package junit.extensions.jfcunit;

import java.awt.AWTException;
import java.awt.Component;
import java.awt.Robot;
import java.awt.event.InputEvent;
import java.lang.reflect.Method;
import junit.extensions.jfcunit.keyboard.JFCKeyStroke;

/**
 * Class that provides facilities for locating components within a GUI. To use
 * create a new instance of RobotTestHelper in your setUp. Windows can only be
 * located once they have been shown once to the user.
 *
 * After calling the helper.enterClickAndLeave() the JFCTestCase.sleep(long delay)
 * method should be called to prevent subsequent clicks from being treated as a
 * double click. A sleep dalay of 300ms can be used as a guideline.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public class RobotTestHelper extends TestHelper {
    /**
     * Robot instance.
     */
    private Robot robot = null;

    // Without this delay events tend to get lost somewhere between
    // the robot and the event queue.
    /**
     * Mandatory delay to get thing working.
     */
    private int delay = 5;

    /**
     * Last X location
     */
    private int lastX = -1;

    /**
     * Last Y location
     */
    private int lastY = -1;

    /**
     * Legal buttons.
     */
    private static int LEGAL_BUTTONS = (InputEvent.BUTTON1_MASK | InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK);

    /**
     * Construct a new RobotTestHelper
     *
     * @exception  AWTException   An instance of java.awt.AWTException can be thrown
     *                            when initializing the Robot.
     */
    public RobotTestHelper() throws AWTException {
        super();
        robot = new Robot();
    }

    /**
     * Process a key press event on a component.
     *
     * @param   ultimate    Not used in the robot.
     * @param   stroke      The JFCKeyStroke to be performed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void keyPressed(Component ultimate, JFCKeyStroke stroke) throws Exception {
        robot.delay(delay);
        robot.keyPress(stroke.getKeyCode());
    }

    /**
     * Process a key release event on a component.
     *
     * @param   ultimate    Not used in the robot.
     * @param   stroke      The JFCKeyStroke to be performed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void keyReleased(Component ultimate, JFCKeyStroke stroke) throws Exception {
        robot.delay(delay);
        robot.keyRelease(stroke.getKeyCode());
    }

    /**
     * Process a mouse move event on a component.
     *
     * @param   ultimate    Not used in the robot.
     * @param   x           The x coordinate of the point where the mouse is being moved to.
     * @param   y           The y coordinate of the point where the mouse is being moved to.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void mouseMoved(Component ultimate, int x, int y) throws Exception {
        int fx = x;
        int fy = y;

        int deltaX = (lastX < fx) ? 1 : -1;
        int deltaY = (lastY < fy) ? 1 : -1;

        while (lastX != fx || lastY != fy) {
            lastX = (lastX == fx) ? fx : lastX + deltaX;
            lastY = (lastY == fy) ? fy : lastY + deltaY;
            robot.delay(delay);
            robot.mouseMove(lastX, lastY);
        }
    }

    /**
     * Process a mouse press event on a component.
     *
     * @param   ultimate          Not used in the robot.
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void mousePressed(Component ultimate, int modifiers, int click, boolean isPopupTrigger) throws Exception {
        robot.delay(delay);
        if (modifiers != 0) {
            int buttons = (modifiers & LEGAL_BUTTONS);
            robot.mousePress(buttons);
        }
    }

    /**
     * Process a mouse release event on a component.
     *
     * @param   ultimate          Not used in the robot.
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected void mouseReleased(Component ultimate, int modifiers, int click, boolean isPopupTrigger) throws Exception {
        robot.delay(delay);
        if (modifiers != 0) {
            int buttons = (modifiers & LEGAL_BUTTONS);
            robot.mouseRelease(buttons);
        }
    }

    /**
     * Simulate rotating the mouseWheel
     * Only supported in Java 1.4.
     *
     * @param ultimate Component to fire the events on.
     * @param amount Amount to rotate the wheel. Ignored on the robot.
     *               The amount will be extracted from the OS Mouse configuration.
     * @param wheelRotation Rotation of the mouse wheel.
     * @exception Exception will be thrown if JRE is less than J2SE 1.4
     */
    protected void mouseWheel(Component ultimate, int amount, int wheelRotation) throws Exception {
        robot.delay(delay);
        Method method = Robot.class.getMethod("mouseWheel", new Class[] { int.class });
        method.invoke(robot, new Object[] { new Integer(wheelRotation)});
    }
}