package junit.extensions.jfcunit;

import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.HashMap;

import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;

import junit.extensions.jfcunit.finder.AbstractWindowFinder;
import junit.extensions.jfcunit.finder.Finder;
import junit.extensions.jfcunit.finder.ComponentFinder;
import junit.extensions.jfcunit.finder.FrameFinder;
import junit.extensions.jfcunit.finder.DialogFinder;
import junit.extensions.jfcunit.finder.NamedComponentFinder;
import junit.extensions.jfcunit.eventdata.KeyEventData;
import junit.extensions.jfcunit.eventdata.AbstractKeyEventData;
import junit.extensions.jfcunit.eventdata.AbstractMouseEventData;
import junit.extensions.jfcunit.eventdata.DragEventData;
import junit.extensions.jfcunit.eventdata.MouseWheelEventData;
import junit.extensions.jfcunit.eventdata.JComboBoxMouseEventData;
import junit.extensions.jfcunit.eventdata.JListMouseEventData;
import junit.extensions.jfcunit.eventdata.StringEventData;
import junit.extensions.jfcunit.keyboard.KeyMapping;
import junit.extensions.jfcunit.keyboard.JFCKeyStroke;
import junit.extensions.jfcunit.keyboard.DefaultKeyMapping;

/**
 * A helper class that provides facilities for locating and posting events to components
 * within a GUI. To use, create a new instance of TestHelper in your setUp.
 * Caveat: Windows can only be located once they have been shown once to the user.
 *
 * @author Matt Caswell
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public abstract class TestHelper {
    /**
     * The length of time that the test class will sleep for between each
     * event being posted.
     */
    public static final long DEFAULTSLEEP = 300L;

    /**
     * KeyMapping to be used with the helper.
     */
    private static KeyMapping keyMapping = null;

    /**
     * System Windows List. These windows should not be closed
     * when the cleanUp is called. The vector should contain
     * the Regular expression strings for matching the name to the
     * title.
     */
    private static HashMap systemWindows = new HashMap();

    /**
     * Lock for setting the key mapping.
     */
    private static final Object LOCK = new Object();

    /**
     * Constructor
     */
    protected TestHelper() {
    }

    /**
     * This method is used to clean up any existing open windows, etc
     * if a test fails midway. To be called in the tearDown() method.
     *
     * @param testCase The test case that is firing the event.
     * @see #addSystemWindow
     * @see #removeSystemWindow
     * @see #removeAllSystemWindows
     */
    public static void cleanUp(JFCTestCase testCase) {
        cleanUp(testCase, DEFAULTSLEEP);
    }

    /**
     * This method is used to clean up any existing open windows, etc
     * if a test fails midway. To be called in the tearDown() method.
     *
     * @param testCase The test case that is firing the event.
     * @param awtSleepTime
     *                 The wait time in ms between each event.
     */
    public static void cleanUp(JFCTestCase testCase, long awtSleepTime) {
        Iterator iter = getWindows().iterator();
        while (iter.hasNext()) {
            boolean ignore = false;
            Window w = (Window)iter.next();
            String title = null;
            if (w instanceof java.awt.Dialog) {
                title = ((java.awt.Dialog)w).getTitle();
            } else if (w instanceof Frame) {
                title = ((java.awt.Frame)w).getTitle();
            }
            if (title != null) {
                Iterator sw = systemWindows.values().iterator();
                while (sw.hasNext() && !ignore) {
                    RE re = (RE)sw.next();
                    if (re.match(title)) {
                        ignore = true;
                    }
                }
            }

            if (!ignore) {
                disposeWindow(w, testCase, awtSleepTime);
            }
        }
    }

    /**
     * This method does the actual work of destroying the window.
     *
     * @param window   The window that needs to be destroyed.
     * @param testCase The test case that is firing the event.
     */
    public static void disposeWindow(Window window, JFCTestCase testCase) {
        disposeWindow(window, testCase, DEFAULTSLEEP);
    }

    /**
     * This method does the actual work of destroying the window and any dialogs that it owns.
     *
     * @param window   The window that needs to be destroyed.
     * @param testCase The test case that is firing the event.
     * @param awtSleepTime
     *                 The wait time in ms between each event.
     */
    public static void disposeWindow(Window window, JFCTestCase testCase, long awtSleepTime) {
        if (window != null) {
            // need to leave the AWTThread in the same state that is in at the end
            boolean awt = testCase.isAWTRunning();
            if (awt) {
                testCase.pauseAWT();
            }

            try {
                if (!(window instanceof JDialog)) {
                    // dispose of any showing dialogs first
                    List dialogs = getShowingDialogs(window);
                    for (int i = 0; i < dialogs.size(); i++) {
                        disposeWindow((JDialog) dialogs.get(i), testCase, awtSleepTime);
                    }
                }
                window.setVisible(false);

                // try to clear the event queue
                testCase.flushAWT();
                if (!awt) {
                    testCase.pauseAWT();
                }
                try {
                    window.getToolkit().getSystemEventQueue().postEvent(new WindowEvent(window, WindowEvent.WINDOW_CLOSING));
                } catch (ExitException exe) {
                    // don't do anything here
                }
            } finally {
                // somehow, this sleeptime needs to be set after the first flush
                testCase.setSleepTime(awtSleepTime);

                // try to clear the event queue
                testCase.flushAWT();
                if (!awt) {
                    testCase.pauseAWT();
                }
                testCase.awtSleep();
            }
        }
    }

    /**
     * Add a regular expression to match the window titles
     * which should not be closed during cleanup.
     *
     * @param matchTitleRE Regular expression defining the windows
     *                     which should not be closed.
     * @throws RESyntaxException if the syntax is not a valid Regular expression.
     */
    public static final void addSystemWindow(String matchTitleRE) throws RESyntaxException{
        if (matchTitleRE == null) {
            return;
        }
        RE patternMatcher = new RE(matchTitleRE);
        systemWindows.put(matchTitleRE, patternMatcher);
    }

    /**
     * Remove a regular expression to match the window titles
     * which should not be closed during cleanup.
     *
     * @param matchTitleRE Regular expression defining the windows
     *                     which should not be closed.
     */
    public static final void removeSystemWindow(String matchTitleRE) {
        systemWindows.remove(matchTitleRE);
    }

    /**
     * Remove all System Window regular expressions.
     */
    public static final void removeAllSystemWindows() {
        systemWindows.clear();
    }

    /**
     * Returns a unmodifiable list of
     * the system window regular expressions.
     *
     * @return unmodifiable Set containing the regular
     *         expressions for matching windows.
     */
    public static final Set getSystemWindows() {
        return Collections.unmodifiableSet(systemWindows.keySet());
    }

    /**
     * A helper method to find the message being displayed in a dialog box.
     *
     * @param dialog The dialog box displaying the message
     * @return The messagebeing displayed
     */
    public String getMessageFromJDialog(JDialog dialog) {
        if (dialog == null || !dialog.isShowing()) {
            return null;
        }
        // since there are usually 2 JLabels, use a sufficiently high number ---> 10
        List list = findComponentList(new ComponentFinder(JLabel.class), dialog, new ArrayList(), 10);
        JLabel label;
        for (int i = 0; i < list.size(); i++) {
            label = (JLabel) list.get(i);
            if (label != null && label.getText() != null) {
                return label.getText();
            }
        }
        return null;
    }

    /**
     * Returns a single window that is showing with given string in the title.
     * If there is more than one window that match, then it is undefined which
     * one will be returned.
     *
     * @param title  The string that should be present in the title.
     * @return A showing window with the given title pattern.
     * @exception JFCTestException
     *                   Thrown if a window cannot be found.
     */
    public static Window getWindow(String title) throws JFCTestException {
        return getWindow(new FrameFinder(title));
    }

    /**
     * Returns a single window that is showing that passes the filtration by the Finder passed in.
     * If there is more than one window that match, then it is undefined which
     * one will be returned.
     *
     * @param finder  The AbstractWindowFinder which is used to filter
     * @return A showing window with the given title.
     * @exception JFCTestException
     *                   Thrown if a window cannot be found.
     */
    public static Window getWindow(AbstractWindowFinder finder) throws JFCTestException {
        Iterator iter = getWindows(finder).iterator();

        if (!iter.hasNext()) {
            throw new JFCTestException("Window with title '" + finder.getTitle() + "' not found");
        }
        return (Window) iter.next();
    }

    /**
     * Returns a set of all the Windows that are currently visible.
     *
     * @return Set of Window objects.
     */
    public static Set getWindows() {
        return getWindows((String) null);
    }

    /**
     * Returns a set of all the Windows that are currently visible and the title
     * contains the given titlematch string.
     *
     * @param titlematch The string pattern to search for in the window's title.
     * @return Set of Window objects.
     */
    public static Set getWindows(String titlematch) {
        return getWindows(new FrameFinder(titlematch));
    }

    /**
     * Returns a set of all the Windows that are currently visible based on the
     * Finder being passed in.
     *
     * @param finder  The AbstractWindowFinder which is used to filter
     * @return Set of Window objects.
     */
    public static Set getWindows(AbstractWindowFinder finder) {
        return getWindows(new HashSet(), getAllWindows(), finder);
    }

    /**
     * Returns a set of all the Windows that are currently visible and the title
     * contains the given titlematch string.
     *
     * @param ret        The list of already filtered and accepted windows.
     * @param windows    The array of windows to filter and add.
     * @param titlematch The string pattern to search for in the window's title.
     * @return Set of Window objects.
     */
    public static Set getWindows(Set ret, Window[] windows, String titlematch) {
        return getWindows(ret, windows, new FrameFinder(titlematch));
    }

    /**
     * Returns a set of all the Windows that are currently visible and the title
     * contains the given titlematch string.
     *
     * @param ret     The list of already filtered and accepted windows.
     * @param windows The array of windows to filter and add.
     * @param finder  The FrameFinder which is used to filter using a title match
     * @return Set of Window objects.
     */
    public static Set getWindows(Set ret, Window[] windows, Finder finder) {
        if (ret == null) {
            ret = new HashSet();
        }
        if (windows == null || finder == null) {
            return ret;
        }

        Frame frame;
        for (int i = 0; i < windows.length; i++) {
            if (!finder.testComponent(windows[i])) {
                continue;
            }

            frame = (Frame) windows[i];

            // if this test is run through any of the TestRunner classes which
            // brings up a GUI window, we need to skip that frame
            if ("JUnit".equalsIgnoreCase(frame.getTitle())) {
                continue;
            }

            if (!ret.contains(frame)) {
                ret.add(frame);
            }

            // add any windows owned by the current 'frame' object
            getWindows(ret, frame.getOwnedWindows(), finder);
        }
        return ret;
    }

    /**
     * Returns a set of all the Dialogs that are currently visible for all
     * opened, visible windows - including the special case of the shared, hidden
     * frame which is the owner of JDialogs whose owner is not specified.
     *
     * @return The full list of dialogs.
     */
    public static List getShowingDialogs() {
        return getShowingDialogs((String) null);
    }

    /**
     * Helper method to obtain a list of the dialogs owned by a given window.
     *
     * @param win    The window to find dialogs for.
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(Window win) {
        return getShowingDialogs(win, null);
    }

    /**
     * Returns a set of all the Dialogs that are currently visible and the title
     * contains the given titlematch string - including the special case of the
     * shared, hidden frame which is the owner of JDialogs whose owner is not specified.
     *
     * @param titlematch The string pattern to search for in the dialog's title.
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(String titlematch) {
        Set openWindows = getWindows();
        // We also have to add the shared, hidden frame used by JDialogs which did not
        // have an owner at the beginning
        openWindows.add((new JDialog()).getOwner());
        Window[] owners = new Window[openWindows.size() + 1];
        owners = (Window[]) openWindows.toArray(owners);
        return getShowingDialogs(owners, titlematch);
    }

    /**
     * Returns a set of all the Dialogs that are currently visible and the title
     * contains the given titlematch string.
     *
     * @param win        The window to find dialogs for.
     * @param titlematch The string pattern to search for in the dialog's title.
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(Window win, String titlematch) {
        return getShowingDialogs(new Window[] { win }, titlematch);
    }

    /**
     * Returns a set of all the Dialogs that are currently visible and the title
     * contains the given titlematch string - including the special case of the
     * shared, hidden frame which is the owner of JDialogs whose owner is not specified.
     *
     * @param owners     The array of windows to find dialogs for.
     * @param titlematch The string pattern to search for in the dialog's title.
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(Window[] owners, String titlematch) {
        return getShowingDialogs(new ArrayList(), owners, titlematch);
    }

    /**
     * Returns a set of all the Dialogs that are currently visible and the title
     * contains the given titlematch string - including the special case of the
     * shared, hidden frame which is the owner of JDialogs whose owner is not specified.
     *
     * @param ret        The list of already filtered and accepted dialogs.
     * @param owners     The array of windows to find dialogs for.
     * @param titlematch The string pattern to search for in the dialog's title.
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(List ret, Window[] owners, String titlematch) {
        return getShowingDialogs(ret, owners, new DialogFinder(titlematch));
    }

    /**
     * Returns a set of all the Dialogs that are currently visible and the title
     * contains the given titlematch string - including the special case of the
     * shared, hidden frame which is the owner of JDialogs whose owner is not specified.
     *
     * @param ret        The list of already filtered and accepted dialogs.
     * @param owners     The array of windows to find dialogs for.
     * @param finder     The DialogFinder which is used to filter using a title match
     * @return The list of dialogs owned by the given window (including the window if it is a Dialog itself).
     */
    public static List getShowingDialogs(List ret, Window[] owners, Finder finder) {
        if (ret == null) {
            ret = new ArrayList();
        }
        if (owners == null || finder == null) {
            return ret;
        }

        Window[] owned = null;
        for (int i = 0; i < owners.length; i++) {
            if (owners[i] == null) {
                continue;
            }

            if (finder.testComponent(owners[i]) && !ret.contains(owners[i])) {
                ret.add(owners[i]);
            }

            getShowingDialogs(ret, owners[i].getOwnedWindows(), finder);
        }

        return ret;
    }

    /**
     * Finds a JFileChooser being shown by a window.
     *
     * @param win    The window to owning the chooser.
     * @return The chooser that has been found, or null if there are none.
     */
    public JFileChooser getShowingJFileChooser(Window win) {
        List choosers = getShowingJFileChoosers(win);
        return (choosers.isEmpty() ? null : (JFileChooser) choosers.get(0));
    }

    /**
     * Finds a list of all the JFileChoosers being shown by a window.
     *
     * @param win    The window to owning the chooser.
     * @return The list of JFileChoosers.
     */
    public List getShowingJFileChoosers(Window win) {
        List ret = new ArrayList();
        List subComps;
        JFileChooser chooser;

        Window[] owned = win.getOwnedWindows();
        Finder finder = new ComponentFinder(JFileChooser.class);
        for (int i = 0; i < owned.length; i++) {
            if (!(owned[i] instanceof JDialog)) {
                continue;
            }

            subComps = findComponentList(finder, owned[i], new ArrayList(), 0);
            Iterator iter = subComps.iterator();
            while (iter.hasNext()) {
                chooser = (JFileChooser) iter.next();
                if (finder.testComponent(chooser)) {
                    ret.add(chooser);
                }
            }
        }

        return ret;
    }

    /**
     * Finds a component which is an instance of the given class by
     * searching from all top level windows.
     *
     * @param compCls Component class
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return Component Component found. Null if index components
     *         do not exist.
     */
    public static Component findComponent(Class compCls, int index) {
        return findComponent(compCls, null, index);
    }

    /**
     * Finds a component which is an instance of the given class within a given
     * container.
     *
     * @param compCls The class of the component.
     * @param cont    The container that the component will be in.
     * @param index   The index of the component. The first component matching the
     *                criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    public static Component findComponent(Class compCls, Container cont, int index) {
        return findComponent(new ComponentFinder(compCls), cont, index);
    }

    /**
     * Finds a component which is an instance of the given class by
     * searching from all top level windows.
     *
     * @param compCls Component class
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return Component Component found. Null if index components
     *         do not exist.
     */
    public static Component findNamedComponent(Class compCls, int index) {
        return findComponent(new NamedComponentFinder(compCls, null), index);
    }

    /**
     * Finds a indexed component which has the given name by
     * searching from all top level windows.
     *
     * @param name The pattern for the name of the component to find.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return Component Component found. Null if index components
     *         do not exist.
     */
    public static Component findNamedComponent(String name, int index) {
        return findComponent(new NamedComponentFinder(null, name), index);
    }

    /**
     * Finds a component which is an instance of the given class
     * and has the given name by searching from all top level windows.
     *
     * @param compCls Component class
     * @param name The pattern for the name of the component to find.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return Component Component found. Null if index components
     *         do not exist.
     */
    public static Component findNamedComponent(Class compCls, String name, int index) {
        return findComponent(new NamedComponentFinder(compCls, name), index);
    }

    /**
     * Finds a component which is an instance of the given class by
     * searching from all top level windows.
     *
     * @param finder Finder class to be used for locating components.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return Component Component found. Null if index components
     *         do not exist.
     */
    public static Component findComponent(Finder finder, int index) {
        Window[] wins = getAllWindows();
        boolean notfound = false;
        int idx = 0;
        for (int window = 0; window < wins.length; window++) {
            List comps = findComponentList(finder, wins[window], new ArrayList(), index - idx);
            if (comps.size() > index - idx) {
                return (Component) comps.get(index - idx);
            }
            idx += comps.size();
        }
        return null;
    }

    /**
     * Finds a component matching the given criteria within a given container
     *
     * @param name   The pattern for the name of the component to find.
     * @param cont   The container that the component will be in.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    public static Component findNamedComponent(String name, Container cont, int index) {
        return findNamedComponent(null, name, cont, index);
    }

    /**
     * Finds a component matching the given criteria within a given container.
     *
     * @param aClass The component's class.
     * @param cont   The container that the component will be in.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    public static Component findNamedComponent(Class aClass, Container cont, int index) {
        return findNamedComponent(aClass, null, cont, index);
    }

    /**
     * Finds a component matching the given criteria within a given container.
     *
     * @param aClass The component's class.
     * @param name   The pattern for the name of the component to find.
     * @param cont   The container that the component will be in.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    public static Component findNamedComponent(Class aClass, String name, Container cont, int index) {
        return findComponent(new NamedComponentFinder(aClass, name), cont, index);
    }

    /**
     * Finds a component matching the given criteria within a given container.
     *
     * @param finder The Finder that is used to test components for a
     *               match.
     * @param cont   The container that the component will be in.
     * @param index  The index of the component. The first component matching the
     *               criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    public static Component findComponent(Finder finder, Container cont, int index) {
        List ret = findComponentList(finder, cont, new ArrayList(), index);
        return ((ret.size() > index) ? (Component) ret.get(index) : null);
    }

    /**
     * Method that calls itself repetitively to build up a list of all components
     * in the container instance that is passed in.
     *
     * @param finder  An instance of the finder which implements the testComponent()
     *                method.
     * @param cont    The Container inside which the component is to be found.
     * @param outList The return list.
     * @param index   The index of the component. The first component matching the
     *                criteria will have index 0, the second 1, etc.
     * @return The component that has been found (or null if none found).
     */
    private static List findComponentList(Finder finder, Container cont, List outList, int index) {
        if (outList == null) {
            outList = new ArrayList();
        }
        if (cont == null || finder == null) {
            return outList;
        }

        Component children[] = cont.getComponents();

        for (int i = 0; i < children.length; i++) {
            if (finder.testComponent(children[i])) {
                if (!outList.contains(children[i])) {
                    outList.add(children[i]);
                }
                if (outList.size() > index) {
                    return outList;
                }
            } else if (children[i] instanceof Container) {
                findComponentList(finder, (Container) children[i], outList, index);
                if (outList.size() > index) {
                    return outList;
                }
            }
        }

        return outList;
    }

    /**
     * Convert the modifiers from mouse modifiers to
     * keyboard modifiers. ALT modifier overlaps with
     * BUTTON2 and META overlaps with BUTTON3.
     *
     * @param  modifiers   The mouse modifiers that need to be converted.
     * @return int         The corresponding keyboard modifiers.
     */
    protected static int mouseToKeyModifiers(int modifiers) {
        int keyModifiers = modifiers;

        int mouseButtons = InputEvent.BUTTON1_MASK + InputEvent.BUTTON2_MASK + InputEvent.BUTTON3_MASK;

        keyModifiers = keyModifiers & (-1 + mouseButtons);
        return keyModifiers;
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param comboBox The JComboBox that is being clicked.
     * @param element  The element that is to be selected.
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JComboBoxMouseEventData
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterDropDownClickAndLeave(JFCTestCase testCase, JComboBox comboBox, Object element, int numberOfClicks) throws Exception {
        enterDropDownClickAndLeave(testCase, comboBox, element, numberOfClicks, DEFAULTSLEEP);
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param comboBox The JComboBox that is being clicked.
     * @param element  The element that is to be selected.
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param awtSleepTime
     *                 The wait time in ms between each event.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JComboBoxMouseEventData
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterDropDownClickAndLeave(JFCTestCase testCase, JComboBox comboBox, Object element, int numberOfClicks, long awtSleepTime) throws Exception {
        int elementIndex = -1;
        for (int i = 0; i < comboBox.getItemCount(); i++) {
            if (comboBox.getItemAt(i) != null && comboBox.getItemAt(i).equals(element)) {
                elementIndex = i;
                break;
            }
        }
        enterClickAndLeave(new JComboBoxMouseEventData(testCase, comboBox, elementIndex, numberOfClicks, awtSleepTime));
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param list     The JList that is being clicked.
     * @param element  The element that is to be selected.
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JListMouseEventData
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterJListClickAndLeave(JFCTestCase testCase, JList list, Object element, int numberOfClicks) throws Exception {
        enterJListClickAndLeave(testCase, list, element, numberOfClicks, DEFAULTSLEEP);
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param list     The JList that is being clicked.
     * @param element  The element that is to be selected.
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param awtSleepTime
     *                 The wait time in ms between each event.
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JListMouseEventData
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterJListClickAndLeave(JFCTestCase testCase, JList list, Object element, int numberOfClicks, long awtSleepTime) throws Exception {
        int elementIndex = -1;
        for (int i = 0; i < list.getModel().getSize(); i++) {
            if (list.getModel().getElementAt(i) != null && list.getModel().getElementAt(i).equals(element)) {
                elementIndex = i;
                break;
            }
        }
        enterJListClickAndLeave(testCase, list, elementIndex, numberOfClicks, awtSleepTime);
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     *
     * @param testCase The test case that is firing these events.
     * @param list     The JList that is being clicked.
     * @param elementIndex
     *                 The element index (see ListDataModel).
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JListMouseEventData
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterJListClickAndLeave(JFCTestCase testCase, JList list, int elementIndex, int numberOfClicks) throws Exception {
        enterJListClickAndLeave(testCase, list, elementIndex, numberOfClicks, DEFAULTSLEEP);
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method.
     *
     * @param testCase The test case that is firing these events.
     * @param list     The JList that is being clicked.
     * @param elementIndex
     *                 The element index (see ListDataModel).
     * @param numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param awtSleepTime
     *                 The wait time in ms between each event.
     * @see #enterClickAndLeave(AbstractMouseEventData)
     * @see junit.extensions.jfcunit.eventdata.JListMouseEventData
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @deprecated as of 1.02, replaced by
     * <code>#enterClickAndLeave(AbstractMouseEventData)</code>
     */
    public void enterJListClickAndLeave(JFCTestCase testCase, JList list, int elementIndex, int numberOfClicks, long awtSleepTime) throws Exception {
        enterClickAndLeave(new JListMouseEventData(testCase, list, elementIndex, numberOfClicks, awtSleepTime));
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method can also differentiate between
     * upper and lower case characters in the specified string.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param comp     The component that is being clicked.
     * @param str      The string that has to be sent to the component.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendString(AbstractKeyEventData)
     * @see junit.extensions.jfcunit.eventdata.StringEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendString(AbstractKeyEventData)</code>
     */
    public void sendString(JFCTestCase testCase, Component comp, String str) throws Exception {
        sendString(testCase, comp, str, 0);
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method can also differentiate between
     * upper and lower case characters in the specified string.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase  The test case that is firing these events.
     * @param comp      The component that is being clicked.
     * @param str       The string that has to be sent to the component.
     * @param modifiers
     *                  The modifier keys down during the event.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendString(AbstractKeyEventData)
     * @see junit.extensions.jfcunit.eventdata.StringEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendString(AbstractKeyEventData)</code>
     */
    public void sendString(JFCTestCase testCase, Component comp, String str, int modifiers) throws Exception {
        sendString(testCase, comp, str, modifiers, DEFAULTSLEEP);
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method can also differentiate between
     * upper and lower case characters in the specified string.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param _testCase The test case that is firing these events
     * @param _comp     The component that is being clicked
     * @param _string   The string that has to be sent to the component
     * @param _modifiers
     *                  The modifier keys down during the event
     * @param _awtSleepTime
     *                  The wait time in ms between each event
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendString(AbstractKeyEventData)
     * @see junit.extensions.jfcunit.eventdata.StringEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendString(AbstractKeyEventData)</code>
     */
    public void sendString(JFCTestCase _testCase, Component _comp, String _string, int _modifiers, long _awtSleepTime) throws Exception {
        sendString(new StringEventData(_testCase, _comp, _string, _modifiers, _awtSleepTime));
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method should be used only to send Action key events.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase The test case that is firing these events.
     * @param comp     The component that is being clicked.
     * @param keyCode  The keyCode of the KeyEvent.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendKeyAction(KeyEventData)
     * @see junit.extensions.jfcunit.eventdata.KeyEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendKeyAction(KeyEventData)</code>
     */
    public void sendKeyAction(JFCTestCase testCase, Component comp, int keyCode) throws Exception {
        sendKeyAction(testCase, comp, keyCode, 0);
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method should be used only to send Action key events.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param testCase  The test case that is firing these events.
     * @param comp      The component that is being clicked.
     * @param keyCode   The keyCode of the KeyEvent.
     * @param modifiers
     *                  The modifier keys down during the event.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendKeyAction(KeyEventData)
     * @see junit.extensions.jfcunit.eventdata.KeyEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendKeyAction(KeyEventData)</code>
     */
    public void sendKeyAction(JFCTestCase testCase, Component comp, int keyCode, int modifiers) throws Exception {
        sendKeyAction(testCase, comp, keyCode, modifiers, DEFAULTSLEEP);
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method should be used only to send Action key events.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param _testCase  The test case that is firing these events
     * @param _comp      The component that is being clicked
     * @param _keyCode   The keyCode of the KeyEvent
     * @param _modifiers
     *                  The modifier keys down during the event
     * @param _awtSleep
     *                  The wait time in ms between each event
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see #sendKeyAction(KeyEventData)
     * @see junit.extensions.jfcunit.eventdata.KeyEventData
     * @deprecated as of 1.02, replaced by
     * <code>#sendKeyAction(KeyEventData)</code>
     */
    public void sendKeyAction(JFCTestCase _testCase, Component _comp, int _keyCode, int _modifiers, long _awtSleep) throws Exception {
        sendKeyAction(new KeyEventData(_testCase, _comp, _keyCode, _modifiers, _awtSleep));
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method should be used only to send Action key events.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param evtData  The event data container.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    public void sendKeyAction(KeyEventData evtData) throws Exception {
        sendString(evtData);
    }

    /**
     * Get all of the windows we know about including heavy weight
     * popups.
     *
     * @return Window[] All of the windows we know about.
     */
    public static Window[] getAllWindows() {
        return WindowMonitor.getWindows();
    }

    /**
     * Get the key mapping to be used.
     *
     * @return the key mapping.
     */
    public static KeyMapping getKeyMapping() {
        synchronized (LOCK) {
            if (keyMapping == null) {
                setKeyMapping(new DefaultKeyMapping());
            }
            LOCK.notifyAll();
        }
        return keyMapping;
    }

    /**
     * Set the key mapping to be used.
     *
     * @param km KeyMapping to be used.
     */
    public static void setKeyMapping(KeyMapping km) {
        synchronized (LOCK) {
            keyMapping = km;
            LOCK.notifyAll();
        }
    }

    /**
     * Send the modifier key press event for the modifiers.
     *
     * @param ultimate  Component to receive the events.
     * @param modifiers The modifiers to be pressed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected final void pressModifiers(Component ultimate, int modifiers) throws Exception {
        adjustModifiers(ultimate, 0, modifiers);
    }

    /**
     * Send the modifier key release event for the modifiers.
     *
     * @param ultimate  Component to receive the events.
     * @param modifiers The modifiers to be released.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected final void releaseModifiers(Component ultimate, int modifiers) throws Exception {
        adjustModifiers(ultimate, modifiers, 0);
    }

    /**
     * Send the modifier key release event for the modifiers.
     *
     * @param ultimate  Component to receive the events.
     * @param modifiers The modifiers to be released.
     * @param newModifiers The new modifiers which should be pressed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected final void adjustModifiers(Component ultimate, int modifiers, int newModifiers) throws Exception {
        if (modifiers == newModifiers) {
            return;
        }

        boolean s1 = (modifiers & InputEvent.SHIFT_MASK) > 0;
        boolean s2 = (newModifiers & InputEvent.SHIFT_MASK) > 0;
        boolean c1 = (modifiers & InputEvent.CTRL_MASK) > 0;
        boolean c2 = (newModifiers & InputEvent.CTRL_MASK) > 0;
        boolean m1 = (modifiers & InputEvent.META_MASK) > 0;
        boolean m2 = (newModifiers & InputEvent.META_MASK) > 0;
        boolean a1 = (modifiers & InputEvent.ALT_MASK) > 0;
        boolean a2 = (newModifiers & InputEvent.ALT_MASK) > 0;
        boolean g1 = (modifiers & InputEvent.ALT_GRAPH_MASK) > 0;
        boolean g2 = (newModifiers & InputEvent.ALT_GRAPH_MASK) > 0;

        int mods = modifiers;

        if (s1 && !s2) {
            // Release shift
            mods = mods & (-1 ^ InputEvent.SHIFT_MASK);
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_SHIFT);
            strokes[0].setModifiers(mods);
            keyReleased(ultimate, strokes[0]);
        } else if (!s1 && s2) {
            // Press shift
            mods = mods | InputEvent.SHIFT_MASK;
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_SHIFT);
            strokes[0].setModifiers(mods);
            keyPressed(ultimate, strokes[0]);
        }

        if (c1 && !c2) {
            // Release control
            mods = mods & (-1 ^ InputEvent.CTRL_MASK);
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_CONTROL);
            strokes[0].setModifiers(mods);
            keyReleased(ultimate, strokes[0]);
        } else if (!c1 && c2) {
            // Press control
            mods = mods | InputEvent.CTRL_MASK;
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_CONTROL);
            strokes[0].setModifiers(mods);
            keyPressed(ultimate, strokes[0]);
        }

        if (a1 && !a2) {
            // Release alt
            mods = mods & (-1 ^ InputEvent.ALT_MASK);
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_ALT);
            strokes[0].setModifiers(mods);
            keyReleased(ultimate, strokes[0]);
        } else if (!a1 && a2) {
            // Press alt
            mods = mods | InputEvent.ALT_MASK;
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_ALT);
            strokes[0].setModifiers(mods);
            keyPressed(ultimate, strokes[0]);
        }

        if (g1 && !g2) {
            // Release alt graph
            mods = mods & (-1 ^ InputEvent.ALT_GRAPH_MASK);
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_ALT_GRAPH);
            strokes[0].setModifiers(mods);
            keyReleased(ultimate, strokes[0]);
        } else if (!g1 && g2) {
            // Press alt graph
            mods = mods | InputEvent.ALT_GRAPH_MASK;
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_ALT_GRAPH);
            strokes[0].setModifiers(mods);
            keyPressed(ultimate, strokes[0]);
        }

        if (m1 && !m2) {
            // Release meta
            mods = mods & (-1 ^ InputEvent.META_MASK);
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_META);
            strokes[0].setModifiers(mods);
            keyReleased(ultimate, strokes[0]);
        } else if (!m1 && m2) {
            // Press meta
            mods = mods | InputEvent.META_MASK;
            JFCKeyStroke[] strokes = getKeyMapping().getKeyStrokes(KeyEvent.VK_META);
            strokes[0].setModifiers(mods);
            keyPressed(ultimate, strokes[0]);
        }
    }

    /**
     * This method is used to send KeyPressed, KeyTyped
     * and KeyReleased events (in that order) to the specified
     * component. This method can also differentiate between
     * upper and lower case characters in the specified string.
     * NOTE: This method will not call 'requestFocus()' on the component - the developer
     * should include it in the test code if needed.
     *
     * @param evtData  The event data container.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    public void sendString(AbstractKeyEventData evtData) throws Exception {
        if (evtData == null) {
            return;
        }
        evtData.getTestCase().pauseAWT();
        if (!evtData.prepareComponent()) {
            evtData.getTestCase().resumeAWT();
            return;
        }
        evtData.getTestCase().resumeAWT();

        Component ultimate = evtData.getRoot();
        int modifiers = evtData.getModifiers();

        // try to clear the event queue
        evtData.getTestCase().flushAWT();

        pressModifiers(ultimate, modifiers);
        JFCKeyStroke[] keyStroke = evtData.getKeyStrokes();
        for (int k = 0; k < keyStroke.length; k++) {
            adjustModifiers(ultimate, modifiers, keyStroke[k].getModifiers());
            modifiers = keyStroke[k].getModifiers();
            long t = System.currentTimeMillis();
            keyPressed(ultimate, keyStroke[k]);
            keyReleased(ultimate, keyStroke[k]);
        }
        releaseModifiers(ultimate, modifiers);

        // try to clear the event queue
        evtData.getTestCase().flushAWT();
    }

    /**
     * Helper method to fire appropriate events to click(s) a component with a
     * custom wait method
     *
     * @param evtData The event data container
     * @exception Exception may be thrown when processing Source or Destination
     *                   prepareComponent()
     */
    public void enterClickAndLeave(AbstractMouseEventData evtData) throws Exception {
        if (evtData == null) {
            return;
        }
        evtData.getTestCase().pauseAWT();
        if (!evtData.prepareComponent()) {
            evtData.getTestCase().resumeAWT();
            return;
        }
        evtData.getTestCase().resumeAWT();
        int numberOfClicks = evtData.getNumberOfClicks();
        int modifiers = evtData.getModifiers();
        boolean isPopupTrigger = evtData.getPopupTrigger();

        Component ultimate = evtData.getRoot();

        // try to clear the event queue
        evtData.getTestCase().flushAWT();

        // Translate the screen coordinates returned by the event to frame coordinates.
        Point screen = evtData.getLocationOnScreen();
        pressModifiers(ultimate, mouseToKeyModifiers(modifiers));

        // Insure a move event.
        mouseMoved(ultimate, screen.x - 5, screen.y - 5);
        mouseMoved(ultimate, screen.x, screen.y);

        for (int click = 1; click <= numberOfClicks; click++) {
            mousePressed(ultimate, modifiers, click, isPopupTrigger);
            mouseReleased(ultimate, modifiers, click, isPopupTrigger);
        }

        releaseModifiers(ultimate, mouseToKeyModifiers(modifiers));

        // try to clear the event queue
        evtData.getTestCase().flushAWT();
    }

    /**
     * Helper method to fire appropriate events to drag and drop a component with a
     * custom wait method.
     *
     * Note this will not work with java.awt.dnd API implementations.
     * The java.awt.dnd implementation is driven from a level below the
     * AWT EventQueue. Use the  RobotTestHelper to work with java.awt.dnd.
     *
     * @param srcEvtData The source  event data container
     * @param dstEvtData The destination event data container
     * @param incr       Amount to increment the coords while "moving" from
     *                   source to destination.
     * @exception Exception may be thrown when processing Source or Destination
     *                   prepareComponent()
     */
    public void enterDragAndLeave(AbstractMouseEventData srcEvtData, AbstractMouseEventData dstEvtData, int incr) throws Exception {
        enterDragAndLeave(new DragEventData(srcEvtData.getTestCase(), srcEvtData, dstEvtData));
    }

    /**
     * Helper method to fire appropriate events to drag and drop a component with a
     * custom wait method.
     *
     * Note this will not work with java.awt.dnd API implementations.
     * The java.awt.dnd implementation is driven from a level below the
     * AWT EventQueue. Use the  RobotTestHelper to work with java.awt.dnd.
     *
     * @param ded    The destination event data container
     * @exception Exception may be thrown when processing Source or Destination
     *                   prepareComponent()
     */
    public void enterDragAndLeave(DragEventData ded) throws Exception {
        // Is component 2 to be visible from the get go.
        // It may not be if autoScrolling is to take place.
        // If autoScrolling then when do we adjust the screen?
        // How do we calculate the target X/Y?

        // For now we will prepare the second component and
        // attempt to insure visibility

        ded.getTestCase().pauseAWT();

        if (!ded.prepareComponent()) {
            ded.getTestCase().resumeAWT();
            return;
        }
        ded.getTestCase().resumeAWT();

        int numberOfClicks = ded.getSource().getNumberOfClicks();
        int modifiers = ded.getSource().getModifiers();
        int keyModifiers = mouseToKeyModifiers(modifiers);
        int mouseButtonMask = modifiers & (InputEvent.BUTTON1_MASK | InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK);
        boolean isPopupTrigger = ded.getSource().getPopupTrigger();

        Component ultimate = ded.getSource().getComponent();
        Point p = ded.getSource().getLocationOnScreen();

        // try to clear the event queue
        ded.getTestCase().flushAWT();

        mouseMoved(ultimate, p.x - 5, p.y - 5);
        mouseMoved(ultimate, p.x, p.y);
        pressModifiers(ultimate, keyModifiers);
        mousePressed(ultimate, mouseButtonMask, 0, isPopupTrigger);

        Point[] pts = ded.getPoints();
        for (int i = 0; i < pts.length; i++) {
            mouseMoved(ultimate, pts[i].x, pts[i].y);
        }
        mouseReleased(ultimate, mouseButtonMask, numberOfClicks, isPopupTrigger);
        releaseModifiers(ultimate, keyModifiers);

        // try to clear the event queue
        ded.getTestCase().flushAWT();
    }

    /**
     * Enter a mouseWheel event.
     *
     * @param evtData MouseWheelEventData to be processed.
     * @exception Exception may be thrown if the JDK is less than 1.4
     */
    public void enterMouseWheel(MouseWheelEventData evtData) throws Exception {
        if (evtData == null) {
            return;
        }
        evtData.getTestCase().pauseAWT();
        if (!evtData.prepareComponent()) {
            evtData.getTestCase().resumeAWT();
            return;
        }
        evtData.getTestCase().resumeAWT();
        int numberOfClicks = evtData.getWheelRotation();
        int modifiers = evtData.getModifiers();
        boolean isPopupTrigger = evtData.getPopupTrigger();

        Component ultimate = evtData.getSource();

        // Translate the screen coordinates returned by the event to frame coordinates.
        Point screen = evtData.getLocationOnScreen();
        pressModifiers(ultimate, mouseToKeyModifiers(modifiers));

        // try to clear the event queue
        evtData.getTestCase().flushAWT();

        // Ensure a move event.
        mouseMoved(ultimate, screen.x - 5, screen.y - 5);
        mouseMoved(ultimate, screen.x, screen.y);

        int inc = (numberOfClicks > 0) ? 1 : -1;
        int mx = (numberOfClicks > 0) ? numberOfClicks : -1 * numberOfClicks;
        for (int i = 0; i < mx; i++) {
            mouseWheel(ultimate, evtData.getScrollAmount(), inc);
        }
        evtData.getTestCase().flushAWT();

        releaseModifiers(ultimate, mouseToKeyModifiers(modifiers));

        // try to clear the event queue
        evtData.getTestCase().flushAWT();
    }

    /**
     * Process a mouse move event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   x           The x coordinate of the point where the mouse is being moved to.
     * @param   y           The y coordinate of the point where the mouse is being moved to.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected abstract void mouseMoved(Component ultimate, int x, int y) throws Exception;

    /**
     * Process a mouse press event on a component.
     *
     * @param   ultimate          The ultimate parent Component
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected abstract void mousePressed(Component ultimate, int modifiers, int click, boolean isPopupTrigger) throws Exception;

    /**
     * Process a mouse release event on a component.
     *
     * @param   ultimate          The ultimate parent Component
     * @param   modifiers         The modifiers associated with this mouse event.
     * @param   click             The number of clicks associated with this mouse event.
     * @param   isPopupTrigger    Whether this mouse event will generate a popup.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected abstract void mouseReleased(Component ultimate, int modifiers, int click, boolean isPopupTrigger) throws Exception;

    /**
     * Simulate rotating the mouseWheel
     *
     * @param ultimate Component to fire the events on.
     * @param amount Amount to scroll for each rotation.
     * @param rotation clicks to rotate the mouse wheel.
     * @exception Exception will be thrown if JDK is less than J2SE 1.4.
     */
    protected abstract void mouseWheel(Component ultimate, int amount, int rotation) throws Exception;

    /**
     * Process a key press event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   stroke      The JFCKeyStroke to be performed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected abstract void keyPressed(Component ultimate, JFCKeyStroke stroke) throws Exception;

    /**
     * Process a key release event on a component.
     *
     * @param   ultimate    The ultimate parent Component
     * @param   stroke      The JFCKeyStroke to be performed.
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     */
    protected abstract void keyReleased(Component ultimate, JFCKeyStroke stroke) throws Exception;
}