package junit.extensions.jfcunit;

import java.awt.AWTEvent;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.AWTEventListener;
import java.awt.event.WindowEvent;
import java.util.Vector;

/**
 * The WindowMonitor class is used to monitor for windows
 * that might not show up in Frame.getFrames(). So we must watch for these
 * windows to be opened.
 *
 * @author Kevin Wilson
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class WindowMonitor implements AWTEventListener {
    /**
     * A handle to this class instance.
     */
    private static WindowMonitor singleton = null;

    /**
     * The dispatch thread used to process the event queue.
     */
    private static DispatchThread dt = null;

    /**
     * Event queue item or null if no pending items.
     * The event queue is a linked list.
     */
    private static WindowEventItem windowEventQueue = null;

    /**
     * Lock for the event queue and dispatch thread.
     */
    private static final Object QLOCK = new Object();

    /**
     * All Non-Popup Windows which have been found.
     */
    private static final Vector WINDOWS = new Vector();

    /**
     * All Popup Windows which have been found.
     */
    private static final Vector POPUPS = new Vector();

    /**
     * Get all of the windows/popups which are open from the passed in Vector
     *
     * @param    sourceVector    The source Vector containing the windows/popups
     * @return Window[] containing all of the windows/popups
     * which have been instantiated.
     */
    private static final Window[] filterWindows(Vector sourceVector) {
        flushAWT();
        Window[] results = new Window[sourceVector.size()];
        synchronized (QLOCK) {
            results = (Window[]) sourceVector.toArray(results);
            QLOCK.notifyAll();
        }
        return results;
    }

    /**
     * Get all of the windows which are open.
     *
     * @return Window[] containing all of the windows
     * which have been instantiated.
     */
    public static final Window[] getWindows() {
        return filterWindows(WINDOWS);
    }

    /**
     * return the singleton instance.
     * If the singleton has not been created,
     * then create one.
     *
     * @return WindowMonitor singleton instance.
     */
    private static final WindowMonitor getSingleton() {
        if (singleton == null) {
            singleton = new WindowMonitor();
        }
        return singleton;
    }

    /**
     * Constructor.
     *
     * Private use the static start() method.
     */
    private WindowMonitor() {
        super();
        if (dt == null) {
            dt = new DispatchThread("WindowMonitor-DispatchThread");
            dt.start();
            Toolkit.getDefaultToolkit().addAWTEventListener(this, AWTEvent.WINDOW_EVENT_MASK);
        }
    }

    /**
     * Start the monitor.
     */
    public static final void start() {
        getSingleton();
    }

    /**
     * Event dispatch implementation.
     *
     * @param theEvent AWTEvent which is to be processed.
     */
    public void eventDispatched(AWTEvent theEvent) {
        processEvent(theEvent);
    }

    /**
     * Processes the events by adding the event onto the
     * event queue to be processed by the dispatch thread.
     *
     * @param theEvent The event to be processed.
     */
    private static final void processEvent(AWTEvent theEvent) {
        switch (theEvent.getID()) {
            case WindowEvent.WINDOW_OPENED :
            case WindowEvent.WINDOW_ACTIVATED :
            case WindowEvent.WINDOW_DEACTIVATED :
            case WindowEvent.WINDOW_CLOSING :
                queueWindowEvent((WindowEvent) theEvent);
                break;
            default :
                break;
        }
    }

    /**
     * Queue the event.
     *
     * @param we WindowEvent to be queued.
     */
    private static void queueWindowEvent(WindowEvent we) {
        synchronized (QLOCK) {
            WindowEventItem qi = new WindowEventItem(we);
            if (windowEventQueue == null) {
                windowEventQueue = qi;
            } else {
                WindowEventItem q = windowEventQueue;
                while (true) {
                    if (q.getNext() != null) {
                        q = q.getNext();
                    } else {
                        break;
                    }
                }
                q.setNext(qi);
            }
            QLOCK.notifyAll();
        }
    }

    /**
     * Handle all Component events in a separate thread. The reason for
     * this is that WindowEvents tend to be used to do lots of processing
     * on the Window hierarchy. As a result, it can frequently result
     * in deadlock situations.
     */
    private class DispatchThread extends Thread {
        /**
         * Constructor
         *
         * @param name Name of the thread.
         */
        public DispatchThread(String name) {
            super(name);
            setDaemon(true);
        }

        /**
         * Run the Dispatch thread.
         * Process the events recieved and track the windows.
         */
        public void run() {
            WindowEvent we = null;
            while (true) {
                while (WindowMonitor.windowEventQueue == null) {
                    synchronized (WindowMonitor.QLOCK) {
                        try {
                            WindowMonitor.QLOCK.wait();
                        } catch (ExitException exe) {
                            // don't do anything, but still catch it
                        } catch (InterruptedException e) {
                            // don't do anything, but still catch it
                        }
                    }
                }
                synchronized (WindowMonitor.QLOCK) {
                    we = WindowMonitor.windowEventQueue.getEvent();
                    WindowMonitor.windowEventQueue = WindowMonitor.windowEventQueue.getNext();

                    switch (we.getID()) {
                        case WindowEvent.WINDOW_OPENED :
                        case WindowEvent.WINDOW_ACTIVATED :
                        case WindowEvent.WINDOW_DEACTIVATED :
                            if (!WINDOWS.contains(we.getWindow())) {
                                WINDOWS.add(we.getWindow());
                            }
                            break;
                        case WindowEvent.WINDOW_CLOSING :
                            WINDOWS.remove(we.getWindow());
                            break;
                    }
                    try {
                        WindowMonitor.QLOCK.notifyAll();
                    } catch (ExitException exe) {
                        // don't do anything, but still catch it
                    }
                }
            }
        }
    }

    /**
     * Flush the AWT Event Queue
     */
    private static void flushAWT() {
        // Invoke the below to insure that we have flushed the queue
        // of all current pending events.
        try {
            Toolkit.getDefaultToolkit().getSystemEventQueue().invokeAndWait(new Runnable() {
                public void run() {
                }
            });
        } catch (Exception e) {
            // don't do anything here
        }
    }

    /**
     * WindowEventItem is the basic type that handles the
     * queue for queueWindowEvent and the DispatchThread.
     */
    private static class WindowEventItem {
        /**
         * Event to be processed by the dispatch thread.
         */
        private WindowEvent event;

        /**
         * Next event item to be processed. Null if no more.
         */
        private WindowEventItem next;

        /**
         * Get the event.
         *
         * @return the event.
         */
        public WindowEvent getEvent() {
            return event;
        }

        /**
         * Get the next entry.
         *
         * @return WindowEventItem next item.
         */
        public WindowEventItem getNext() {
            return next;
        }

        /**
         * Set the next entry
         *
         * @param n next item
         */
        public void setNext(WindowEventItem n) {
            next = n;
        }

        /**
         * Constructor
         *
         * @param evt Event to construct the queue item for.
         */
        WindowEventItem(WindowEvent evt) {
            event = evt;
            next = null;
        }
    }
}