package junit.extensions.jfcunit.eventdata;

import junit.extensions.jfcunit.TestHelper;
import junit.extensions.jfcunit.keyboard.KeyMapping;
import junit.extensions.jfcunit.keyboard.JFCKeyStroke;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.event.KeyEvent;
import java.util.Arrays;
import java.util.Vector;

/**
 * Abstract data container class that holds most of the data necessary for jfcUnit to fire mouse events.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public abstract class AbstractKeyEventData extends AbstractEventData {
    /**
     * Key Strokes to be simulated.
     */
    private final Vector keyStrokes = new Vector();

    /**
     * LOCK for setting the keyMapping.
     */
    private static final Object LOCK = new Object();

    /**
     * KeyMapping used to translate char and keyCodes
     * to JFCKeyStrokes.
     */
    private static KeyMapping keyMapping = null;

    /**
     * Get the key strokes for this event.
     *
     * @return JFCKeyStroke[] to be simulated by this event.
     */
    public final JFCKeyStroke[] getKeyStrokes() {
        return (JFCKeyStroke[]) keyStrokes.toArray(new JFCKeyStroke[0]);
    }

    /**
     * Clear key Strokes
     */
    protected final void clearKeyStrokes() {
        keyStrokes.clear();
    }

    /**
     * Add key strokes to this event.
     *
     * @param strokes Key strokes to be added.
     */
    protected final void addKeyStrokes(JFCKeyStroke[] strokes) {
        for (int i = 0; i < strokes.length; i++) {
            keyStrokes.add(strokes[i]);
        }
    }

    /**
     * Apply the modifiers specified to all of the key strokes.
     *
     * @param modifiers Modifiers to be applied.
     */
    protected final void applyModifier(int modifiers) {
        for (int i = 0; i < keyStrokes.size(); i++) {
            JFCKeyStroke s = (JFCKeyStroke) keyStrokes.get(i);
            int mods = s.getModifiers() | modifiers;
            s.setModifiers(mods);
        }
    }

    /**
     * Set the attribute value.
     *
     * @param km    The new KeyMapping
     */
    public static void setKeyMapping(KeyMapping km) {
        TestHelper.setKeyMapping(km);
    }

    /**
     * Get the attribute value.
     *
     * @return    KeyMapping   The KeyMapping that is being currently used.
     */
    public static KeyMapping getKeyMapping() {
        return TestHelper.getKeyMapping();
    }

    /**
     * Get the attribute value.
     *
     * @param ae {@link java.awt.AWTEvent} to be processed.
     * @return boolean    The value of the attribute
     */
    public boolean consume(AWTEvent ae) {
        if (!(ae instanceof KeyEvent)) {
            return false;
        }

        KeyEvent ke = (KeyEvent) ae;
        int id = ke.getID();
        if ((id == KeyEvent.KEY_TYPED) || (id == KeyEvent.KEY_RELEASED) || (isMetaChar(ke.getKeyCode())) || ae.getSource().equals(getRoot())) {
            // Ignore the event.
            return true;
        }
        return false;
    }

    /**
     * Check if this event can consume the {@link java.awt.AWTEvent}.
     *
     * @param ae {@link java.awt.AWTEvent} to be consumed.
     * @return boolean true if the event can be consumed.
     */
    public boolean canConsume(AWTEvent ae) {
        if (!isValid()) {
            return true;
        }

        return (ae instanceof KeyEvent) && getRoot((Component) ae.getSource()).equals(getRoot());
    }

    /**
     * Compare to event datas and deterime if they are equal.
     *
     * @param o Object to be compared.
     * @return true if the events are the same.
     */
    public boolean equals(Object o) {
        if (!super.equals(o) || !(o instanceof AbstractKeyEventData)) {
            return false;
        }
        // isValid
        AbstractKeyEventData data = (AbstractKeyEventData) o;
        JFCKeyStroke[] dataStrokes = data.getKeyStrokes();
        JFCKeyStroke[] thisStrokes = getKeyStrokes();
        return Arrays.equals(dataStrokes, thisStrokes);
    }
}