package junit.extensions.jfcunit.eventdata;

import java.awt.Component;
import java.awt.event.KeyEvent;

import org.w3c.dom.Element;

import junit.extensions.jfcunit.JFCTestCase;

import junit.extensions.xml.XMLException;
import junit.extensions.xml.IXMLTestCase;
import junit.extensions.xml.elements.AbstractTagHandler;

import junit.extensions.jfcunit.xml.JFCXMLConstants;
/**
 * This class will handle the processing of &lt;key&gt; nodes.
 *
 * <h3>Description</h3>
 * <p>
 *   This element will be used to submit key codes and strings
 *   to a component.
 * </p>
 *
 * <h3>Parameters</h3>
 * <table border="1" cellpadding="2" cellspacing="0">
 *   <tr>
 *     <td valign="top"><b>Attribute</b></td>
 *     <td valign="top"><b>Description</b></td>
 *     <td align="center" valign="top"><b>Required</b></td>
 *     <td valign="top"><b>Default</b></td>
 *     <td valign="top"><b>Values</b></td>
 *   </tr>
 *   <tr>
 *     <td valign="top">refid</td>
 *     <td valign="top">Id of a object reference which has been previously found.</td>
 *     <td valign="top" align="center">Yes</td>
 *     <td valign="top">N/A</td>
 *     <td valign="top">N/A</td>
 *   </tr>
 *   <tr>
 *     <td valign="top">string</td>
 *     <td valign="top">The string to be entered.</td>
 *     <td valign="top" align="center">One string or code attribute is required</td>
 *     <td valign="top">empty string</td>
 *     <td valign="top">Any characters</td>
 *   </tr>
 *   <tr>
 *     <td valign="top">code</td>
 *     <td valign="top">The key code to be entered.</td>
 *     <td valign="top" align="center">One string or code attribute is required</td>
 *     <td valign="top">None</td>
 *     <td valign="top">The key codes to use should follow the VK_(codes) of KeyEventData. Ex: VK_F4</td>
 *   </tr>
 *   <tr>
 *     <td valign="top">modifiers</td>
 *     <td valign="top">The modifiers to be used for this event.</td>
 *     <td valign="top" align="center">No</td>
 *     <td valign="top">None</td>
 *     <td valign="top">alt altgr control meta shift</td>
 *   </tr>
 *   <tr>
 *     <td valign="top">sleeptime</td>
 *     <td valign="top">Maximum sleep time</td>
 *     <td valign="top" align="center">No</td>
 *     <td valign="top">500</td>
 *     <td valign="top">Positive Integer representing the milli-seconds</td>
 *   </tr>
 *
 * </table>
 * <h3>Example</h3>
 * <blockquote><pre>
 * &lt;key
 *    refid=&quot;FileMenuItem&quot;
 *    code=&quot;VK_F2&quot;
 *    modifiers=&quot;shift&quot;
 * /&gt;
 * </pre></blockquote>
 * <p>
 * The above simulates pressing shift F2.
 * </p>
 * @see junit.extensions.jfcunit.eventdata.KeyEventData
 * @author Kevin Wilson
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class KeyTagHandler extends AbstractTagHandler implements JFCXMLConstants {
    /**
     * Constructor for FindTagHandler.
     *
     * @param element     The element to be processed
     * @param testCase    The IXMLTestCase that uses this element
     */
    public KeyTagHandler(Element element, IXMLTestCase testCase) {
        super(element, testCase);
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#validateElement()
     */
    public void validateElement() throws XMLException {
        // do the default validations from the super class
        super.validateElement();

        // check the element tag name
        checkElementTagName(KEY);

        // reqd attribute: refid
        checkRequiredAttribute(REFID);

        // Only one of the following is required.
        checkOneRequiredAttribute(getElement(), new String[] { CODE, STRING });
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#processElement()
     */
    public void processElement() throws Exception {
        validateElement();

        String refid = getRefId();
        Component comp = (Component) getTestCase().getFoundObject(refid);
        if (comp == null) {
            throw new XMLException("Component not found for:" + refid);
        }

        int modifiers = getModifiers();
        long sleepTime = getSleepTime();

        String string = getString(STRING);
        AbstractKeyEventData data = null;
        if (string != null) {
            data = new StringEventData((JFCTestCase)getTestCase(), comp, string, modifiers, sleepTime);
        } else {
            int character = getCharacter();
            data = new KeyEventData((JFCTestCase)getTestCase(), comp, character, modifiers, sleepTime);
        }
        ((JFCTestCase)getTestCase()).getHelper().sendString(data);
    }

    /**
     * Returns the value of the REFID attribute for this element.
     * @return String  The value of the REFID attribute.
     */
    private String getRefId() {
        return getString(REFID);
    }

    /**
     * Returns the value of the MODIFIERS attribute for this element. Defaults to zero.
     * @return int  The value of the MODIFIERS attribute, zero if not specified.
     */
    private int getModifiers() {
        return getModifiers(0);
    }

    /**
     * Returns the value of the MODIFIERS attribute for this element. Defaults to
     * defaultValue.
     * @param defaultValue value to be returned if the element does not exist.
     * @return int  The value of the MODIFIERS attribute, defaultValue if not specified.
     */
    protected int getModifiers(int defaultValue) {
        int modifiers = 0;
        String s = getString(MODIFIERS);
        if (s == null) {
            return defaultValue;
        }

        try {
            // If we can parse a integer from the string then use the integer
            // value.
            return Integer.parseInt(s);
        } catch (NumberFormatException nfe) {
        }

        if (s.indexOf(SHIFT) != -1) {
            modifiers += KeyEvent.SHIFT_MASK;
        }
        if (s.indexOf(CONTROL) != -1) {
            modifiers += KeyEvent.CTRL_MASK;
        }
        if (s.indexOf(ALT) != -1) {
            modifiers += KeyEvent.ALT_MASK;
        }
        if (s.indexOf(ALTGR) != -1) {
            modifiers += KeyEvent.ALT_GRAPH_MASK;
        }
        if (s.indexOf(META) != -1) {
            modifiers += KeyEvent.META_MASK;
        }
        if (s.indexOf(BUTTON1) != -1) {
            modifiers += KeyEvent.BUTTON1_MASK;
        }
        if (s.indexOf(BUTTON2) != -1) {
            modifiers += KeyEvent.BUTTON2_MASK;
        }
        if (s.indexOf(BUTTON3) != -1) {
            modifiers += KeyEvent.BUTTON3_MASK;
        }

        return modifiers;
    }


    /**
     * Returns the value of the SLEEPTIME attribute for this element. Defaults to DEFAULT_SLEEPTIME.
     * @return String  The value of the SLEEPTIME attribute, DEFAULT_SLEEPTIME if not specified.
     */
    protected long getSleepTime() {
        return getLong(SLEEPTIME, DEFAULT_SLEEPTIME);
    }

    /**
     * Returns the value of the CODE attribute for this element.
     * @return String  The value of the CODE attribute.
     */
    private int getCharacter() {
        int keyCode = 0;
        String keyCodeStr = getString(CODE);
        try {
            Object code = KeyEvent.class.getField(keyCodeStr).get(null);
            return ((Integer)code).intValue();
        } catch (Exception e) {}
        return Integer.parseInt(keyCodeStr);
    }
}