package junit.extensions.jfcunit.eventdata;

import junit.extensions.jfcunit.JFCTestCase;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.lang.reflect.Method;

import javax.swing.JViewport;

/**
 * Data container class that holds all the data necessary for jfcUnit to fire mouse events.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class MouseWheelEventData extends MouseEventData {
    /**
     * Default value specifying whether the mouse event being fired
     * would trigger a popup or not.
     */
    public static final boolean DEFAULT_ISPOPUPTRIGGER = false;

    /**
     * The amount to scroll by default.
     */
    public static final int DEFAULT_MOUSE_CLICKS = 0;

    /**
     * MouseWheelEvent event id.
     */
    private static int mouseWheelEvent;

    /**
     * Class of the MouseWheelEvent.
     * Used so that the code is backward compatible.
     */
    private static Class mweClass;

    /**
     * Method for getting the rotation of the mouse from the
     * event. Used so that the code is backward compatible.
     */
    private static Method getWheelRotationMethod;

    /**
     * Method for getting the scroll amount from the
     * event. Used so that the code is backward compatible.
     */
    private static Method getScrollAmountMethod;

    /**
     * The Component on which to trigger the event.
     */
    private Component comp;

    /**
     * Number of wheel clicks to rotate.
     */
    private int wheelRotation = 0;

    /**
     * Amount to scroll for each wheel rotation.
     */
    private int scrollAmount = 0;

    static {
        try {
            mweClass = Class.forName("java.awt.event.MouseWheelEvent");
            mouseWheelEvent = mweClass.getField("MOUSE_WHEEL").getInt(null);
            getWheelRotationMethod = mweClass.getMethod("getWheelRotation", new Class[0]);
            getScrollAmountMethod = mweClass.getMethod("getScrollAmount", new Class[0]);
        } catch (Exception e) {
            mouseWheelEvent = 0;
        }
    }

    /**
     * Default Constructor
     */
    public MouseWheelEventData() {
        super();
    }

    /**
     * Constructor
     *
     * @param _testCase The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp     The component on which to trigger the event.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp) {
        this(_testCase, _comp, DEFAULT_SCROLL_AMOUNT, DEFAULT_WHEEL_ROTATION);
    }

    /**
     * Constructor
     *
     * @param _testCase The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp     The component on which to trigger the event.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _wheelRotation) {
        this(_testCase, _comp, DEFAULT_SCROLL_AMOUNT, _wheelRotation, DEFAULT_MOUSE_CLICKS);
    }

    /**
     * Constructor
     *
     * @param _testCase The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp     The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation) {
        this(_testCase, _comp, _amount, _wheelRotation, DEFAULT_MOUSE_CLICKS);
    }

    /**
     * Constructor
     *
     * @param _testCase The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp     The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, DEFAULT_MOUSE_MODIFIERS);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                  Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks, int _modifiers) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, _modifiers, DEFAULT_ISPOPUPTRIGGER);
    }

    /**
     * Constructor
     *
     * @param _testCase The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp     The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                 Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _isPopupTrigger
     *                 boolean specifying whether this event will show a popup.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks, boolean _isPopupTrigger) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, getDefaultModifiers(_isPopupTrigger), _isPopupTrigger);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                  Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _sleepTime
     *                  The wait time in ms between each event.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks, long _sleepTime) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, DEFAULT_MOUSE_MODIFIERS, DEFAULT_ISPOPUPTRIGGER, _sleepTime);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                  Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     * @param _isPopupTrigger
     *                  boolean specifying whether this event will show a popup.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks, int _modifiers, boolean _isPopupTrigger) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, _modifiers, _isPopupTrigger, DEFAULT_SLEEPTIME);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                  Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _isPopupTrigger
     *                  boolean specifying whether this event will show a popup.
     * @param _sleepTime
     *                  The wait time in ms between each event.
     */
    public MouseWheelEventData(JFCTestCase _testCase, Component _comp, int _amount, int _wheelRotation, int _numberOfClicks, boolean _isPopupTrigger, long _sleepTime) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, getDefaultModifiers(_isPopupTrigger), _isPopupTrigger, _sleepTime);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                   Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     * @param _isPopupTrigger
     *                   boolean specifying whether this event will show a popup.
     * @param _sleepTime
     *                   The wait time in ms between each event.
     */
    public MouseWheelEventData(
        JFCTestCase _testCase,
        Component _comp,
        int _amount,
        int _wheelRotation,
        int _numberOfClicks,
        int _modifiers,
        boolean _isPopupTrigger,
        long _sleepTime) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, _modifiers, _isPopupTrigger, _sleepTime, DEFAULT_POSITION, null);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                   Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     * @param _isPopupTrigger
     *                   boolean specifying whether this event will show a popup.
     * @param _sleepTime
     *                   The wait time in ms between each event.
     * @param _position  The relative mouse position within the cell.
     */
    public MouseWheelEventData(
        JFCTestCase _testCase,
        Component _comp,
        int _amount,
        int _wheelRotation,
        int _numberOfClicks,
        int _modifiers,
        boolean _isPopupTrigger,
        long _sleepTime,
        int _position) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, _modifiers, _isPopupTrigger, _sleepTime, _position, null);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                   Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     * @param _isPopupTrigger
     *                   boolean specifying whether this event will show a popup.
     * @param _sleepTime
     *                   The wait time in ms between each event.
     * @param _referencePoint     The CUSTOM mouse position within the cell.
     */
    public MouseWheelEventData(
        JFCTestCase _testCase,
        Component _comp,
        int _amount,
        int _wheelRotation,
        int _numberOfClicks,
        int _modifiers,
        boolean _isPopupTrigger,
        long _sleepTime,
        Point _referencePoint) {
        this(_testCase, _comp, _amount, _wheelRotation, _numberOfClicks, _modifiers, _isPopupTrigger, _sleepTime, CUSTOM, _referencePoint);
    }

    /**
     * Constructor
     *
     * @param _testCase  The JFCTestCase on whose thread <code>awtSleep()</code> has to be invoked.
     * @param _comp      The component on which to trigger the event.
     * @param _amount   The amount to scroll for each rotation.
     * @param _wheelRotation The amount to rotate the wheel positive
     * for values towards the user. Negative for away from the user.
     * @param _numberOfClicks
     *                   Number of clicks in the MouseEvent (1 for single-click, 2 for double clicks)
     * @param _modifiers The modifier key values that need to be passed onto the event.
     * @param _isPopupTrigger
     *                   boolean specifying whether this event will show a popup.
     * @param _sleepTime
     *                   The wait time in ms between each event.
     * @param _position  The relative mouse position within the cell.
     * @param _referencePoint
     *                   If _position is CUSTOM then the point is a offset from
     *                   the location of the component. If the _position is PERCENT
     *                   then the location is a percentage offset of the hight and width.
     *                   Otherwise, the _referencePoint is unused.
     */
    public MouseWheelEventData(
        JFCTestCase _testCase,
        Component _comp,
        int _amount,
        int _wheelRotation,
        int _numberOfClicks,
        int _modifiers,
        boolean _isPopupTrigger,
        long _sleepTime,
        int _position,
        Point _referencePoint) {
        setTestCase(_testCase);
        setSource(_comp);
        setScrollAmount(_amount);
        setWheelRotation(_wheelRotation);
        setNumberOfClicks(_numberOfClicks);
        setModifiers(_modifiers);
        setPopupTrigger(_isPopupTrigger);
        setSleepTime(_sleepTime);
        setPosition(_position);
        setReferencePoint(_referencePoint);
        setValid(true);
    }

    /**
     * Set the attribute value
     *
     * @param _comp   The new value of the attribute
     */
    public void setSource(Component _comp) {
        comp = _comp;
    }

    /**
     * Get the attribute value
     *
     * @return Component    The value of the attribute
     */
    public Component getSource() {
        return comp;
    }

    /**
     * The component on which the event has to be fired.
     *
     * @return The component
     */
    public Component getComponent() {
        // by default, the component is the same as the source
        return getSource();
    }

    /**
     * Set the scroll amount.
     *
     * @param scrollAmount Amount to scroll for each wheel click.
     */
    public void setScrollAmount(int scrollAmount) {
        this.scrollAmount = scrollAmount;
    }

    /**
     * Get the scroll amount.
     *
     * @return int scroll amount.
     */
    public int getScrollAmount() {
        return scrollAmount;
    }

    /**
     * Set the wheel rotation.
     *
     * @param wheelRotation amount to rotate the wheel.
     */
    public void setWheelRotation(int wheelRotation) {
        this.wheelRotation = wheelRotation;
    }

    /**
     * Get the wheel rotation.
     * @return int number of wheel rotations.
     */
    public int getWheelRotation() {
        return wheelRotation;
    }

    /**
     * Prepare the component for firing the mouse event.
     *
     * @return boolean true if the component is ready.
     */
    public boolean prepareComponent() {
        if (!isValidForProcessing(getSource())) {
            return false;
        }

        // Locate the ViewPort
        Container parent = null;
        if (getSource() instanceof Container) {
            parent = (Container) getSource();
        } else {
            parent = getSource().getParent();
        }
        while (parent.getParent() != null && !(parent instanceof JViewport)) {
            parent = parent.getParent();
        }

        if (!(parent instanceof JViewport)) {
            return false;
        }

        JFCTestCase testCase = getTestCase();
        if (testCase != null) {
            testCase.awtSleep(getSleepTime());
        }

        Rectangle bounds = parent.getBounds();
        setLocationOnScreen(calculatePoint(bounds));
        return true;
    }

    /**
     * Consume the event.
     *
     * @param ae AWTEvent to be consumed.
     * @return boolean true if the event was consumed.
     */
    public boolean consume(AWTEvent ae) {

        MouseEvent me = (MouseEvent) ae;
        try {
            int amount = ((Integer) getScrollAmountMethod.invoke(ae, new Object[0])).intValue();
            setScrollAmount(amount);

            int wheelRotation = ((Integer) getWheelRotationMethod.invoke(ae, new Object[0])).intValue();
            setWheelRotation(wheelRotation + getWheelRotation());
        } catch (Exception e) {
            e.printStackTrace();
        }

        // If the event is valid then we do not need to
        // reprocess the rest of the event.
        if (isValid()) {
            return true;
        }
        Component source = (Component) me.getSource();
        setSource(source);
        setModifiers(me.getModifiers());
        setNumberOfClicks(me.getClickCount());
        setPopupTrigger(me.isPopupTrigger());
        Point p = new Point(me.getX(), me.getY());

        Point screen = source.getLocationOnScreen();
        screen.translate(p.x, p.y);
        setLocationOnScreen(screen);

        setPosition(CENTER);
        setReferencePoint(null);
        setValid(true);
        return true;
    }

    /**
     * Returns true if the event can be consumed by
     * this instnace of event data.
     *
     * @param ae Event to be consumed.
     * @return true if the event was consumed.
     */
    public boolean canConsume(AWTEvent ae) {
        // Cannot consume if the event ID cannot be accessed
        // or the event is not a mouse wheel event.
        if (mouseWheelEvent == 0 || ae.getID() != mouseWheelEvent) {
            return false;
        }

        if (isValid()) {
            int modifiers = ((MouseEvent) ae).getModifiers();
            if (ae.getSource() != getSource()) {
                return false;
            }
            if (modifiers != getModifiers()) {
                return false;
            }
            try {
                int wheelRotation = ((Integer) getWheelRotationMethod.invoke(ae, new Object[0])).intValue();
                // Change of direction if result is negative.
                if (wheelRotation * getWheelRotation() < 0) {
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return true;
    }

    /**
     * Generate a string representation of the event.
     *
     * @return String form of the event data.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(1000);
        buf.append(super.toString());
        buf.append(" amount:" + getScrollAmount());
        buf.append(" rotation:" + getWheelRotation());
        return buf.toString();
    }

    /**
     * Compare to event data and deterime if they are equal.
     *
     * @param o Object to be compared.
     * @return true if the events are the same.
     */
    public boolean equals(Object o) {
        if (!(o instanceof MouseWheelEventData)) {
            return false;
        }
        MouseWheelEventData data = (MouseWheelEventData) o;
        return super.equals(o) && data.getWheelRotation() == getWheelRotation() && data.getScrollAmount() == getScrollAmount();
    }
}