package junit.extensions.jfcunit.finder;

import java.awt.Component;

import javax.swing.AbstractButton;
import javax.swing.Icon;

/**
 * Class for checking if the ({@link javax.swing.AbstractButton}) component being searched for has been found.
 * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
 * This class delegates the matching of the Icon to an instance of {@link junit.extensions.jfcunit.finder.IconMatcher}
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class AbstractButtonFinder extends Finder {
    /**
     * The text of the {@link javax.swing.AbstractButton} component.
     */
    private String label;

    /**
     * The matcher for the icon of the {@link javax.swing.AbstractButton} component.
     */
    private IconMatcher iconMatcher;

    /**
     * A boolean specifying whether the filtration is case insensitive.
     */
    private boolean caseIndependent = false;

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str    The desired pattern for the text of the component.
     */
    public AbstractButtonFinder(String str) {
        this(str, null, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str                The desired pattern for the text of the component.
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    public AbstractButtonFinder(String str, boolean caseIndependent) {
        this(str, null, caseIndependent);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param icon   The desired pattern for the icon of the component.
     */
    public AbstractButtonFinder(Icon icon) {
        this(null, icon, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str    The desired pattern for the text of the component.
     * @param icon   The desired pattern for the icon of the component.
     */
    public AbstractButtonFinder(String str, Icon icon) {
        this(str, icon, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str                The desired pattern for the text of the component.
     * @param icon               The desired pattern for the icon of the component.
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    public AbstractButtonFinder(String str, Icon icon, boolean caseIndependent) {
        this.label = str;
        try {
            this.iconMatcher = new IconMatcher(icon);
        } catch (InterruptedException ie) {
            this.iconMatcher = null;
        }
        this.caseIndependent = caseIndependent;
        createPatternMatcher(label, caseIndependent);
    }

    /**
     * Method that returns true if the given component matches the search
     * criteria
     *
     * @param comp   The component to test
     * @return true if this component is a match
     */
    public boolean testComponent(Component comp) {
        return (
            comp != null
                && isValidForProcessing(comp, AbstractButton.class)
                && (label == null || matchPattern(((AbstractButton) comp).getText(), label))
                && iconMatcher.matches(((AbstractButton) comp).getIcon()));
    }
}
