package junit.extensions.jfcunit.finder;

/**
 * Class for checking if the ({@link java.awt.Window}) component being searched for has been found.
 * This class can be used to check only windows that have titles (Frame/Dialog and their sub-classes)
 * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public abstract class AbstractWindowFinder extends Finder {
    /**
     * The title of the component.
     */
    private String title;

    /**
     * A boolean specifying whether the filtration is case insensitive.
     */
    private boolean caseIndependent = false;

    /**
     * Constructor accepting all arguments needed to filter component.
     *
     * @param str    The desired pattern for the title of the component.
     */
    public AbstractWindowFinder(String str) {
        this(str, false);
    }

    /**
     * Constructor accepting all arguments needed to filter component.
     *
     * @param str                The desired pattern for the title of the component.
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    public AbstractWindowFinder(String str, boolean caseIndependent) {
        title = str;
        this.caseIndependent = caseIndependent;
        createPatternMatcher(title, caseIndependent);
    }

    /**
     * This method checks whether the winTitle and the title strings are equal.
     * Note: If the title string is null, we need to return all showing windows,
     * so 'true' is returned.
     *
     * @param winTitle The title of the window being checked
     * @return True if either the filter string is null or if the
     *         filter string is a substring of the window title
     */
    protected final boolean checkIfTitleMatches(String winTitle) {
        return ((title == null) || (winTitle != null && matchPattern(winTitle, title)));
    }

    /**
     * Returns the title.
     *
     * @return String
     */
    public String getTitle() {
        return title;
    }

    /**
     * Returns the caseIndependent.
     * @return boolean
     */
    public boolean isCaseIndependent() {
        return caseIndependent;
    }
}