package junit.extensions.jfcunit.finder;

import java.awt.Component;
import java.awt.Container;
import java.awt.Window;

import javax.swing.Icon;
import javax.swing.ImageIcon;

import org.w3c.dom.Element;

import junit.extensions.jfcunit.TestHelper;

import junit.extensions.xml.XMLException;
import junit.extensions.xml.IXMLTestCase;
import junit.extensions.xml.elements.AbstractTagHandler;

import junit.extensions.jfcunit.xml.JFCXMLConstants;

/**
 * This class provides the basic handling for all of the FindTagHandler
 * implementations.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public abstract class BaseFindTagHandler extends AbstractTagHandler implements JFCXMLConstants {
    /**
     * Constructor for FindTagHandler.
     *
     * @param element     The element to be processed
     * @param testCase    The IXMLTestCase that uses this element
     */
    public BaseFindTagHandler(Element element, IXMLTestCase testCase) {
        super(element, testCase);
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#validateElement()
     */
    public void validateElement() throws XMLException {
        // do the default validations from the super class
        super.validateElement();

        // check the element tag name
        checkElementTagName(FIND);

        // reqd attribute: id
        checkRequiredAttribute(ID);
        // reqd attribute: index
        checkRequiredAttribute(INDEX);
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#validateElement()
     */
    public abstract void processElement() throws Exception;

    /**
     * Execute the common find on the TestHelper.
     * @param finder Finder to be used to locate the object.
     * @throws Exception may be thrown.
     */
    protected void find(Finder finder) throws Exception {
        int index = getIndex();
        String containerRefId = getContainerId();

        Component comp;
        if (containerRefId != null) {
            comp = TestHelper.findComponent(finder, (Container) getTestCase().getFoundObject(containerRefId), index);
        } else {
            comp = TestHelper.findComponent(finder, index);
        }
        String id = getId();
        getTestCase().addFoundObject(id, comp);
    }

    /**
     * Execute the common findWindow on the TestHelper.
     * @param wfinder Finder to be used to locate the object.
     * @throws Exception may be thrown.
     */
    protected void findWindow(AbstractWindowFinder wfinder) throws Exception {
        Window w = null;
        w = TestHelper.getWindow(wfinder);
        String id = getId();
        getTestCase().addFoundObject(id, w);
    }


    /**
     * Parse the element and create a JMenuItemFinder.
     *
     * @param element     The element to be processed
     * @return Finder     The JMenuItemFinder to be created.
     */
    private Finder getJMenuItemFinder(Element element) {
        String label = getLabel();
        boolean caseIndependent = getCaseIndependent();
        return new JMenuItemFinder(label, caseIndependent);
    }

    /**
     * Returns the value of the INDEX attribute for this element. Defaults to 0 (zero) if not found.
     * @return int  The value of the parsed INDEX attribute, zero if not found.
     */
    protected int getIndex() {
        return getInt(INDEX, 0);
    }

    /**
     * Returns the value of the NAME attribute for this element.
     * @return String  The value of the NAME attribute.
     */
    protected String getName() {
        return getString(NAME);
    }

    /**
     * Returns the value of the LABEL attribute for this element.
     * @return String  The value of the LABEL attribute.
     */
    protected String getLabel() {
        return getString(LABEL);
    }

    /**
     * Retrun the value of the TITLE attribute for this element.
     * @return String The value of the TITLE attribute.
     */
    protected String getTitle() {
        return getString(TITLE);
    }

    /**
     * Returns the value of the CLASS attribute for this element.
     * @return String  The value of the CLASS attribute.
     */
    protected String getClassName() {
        return getString(CLASS);
    }

    /**
     * Returns the value of the CONTAINER attribute for this element.
     * @return String  The value of the CONTAINER attribute.
     */
    protected String getContainerId() {
        return getString(CONTAINER);
    }

    /**
     * Returns the Icon described by the ICONFILE and ICONDESCRIPTION attributes for this element.
     * @return Icon  The ImageIcon described by the value of the ICONFILE and ICONDESCRIPTION
     * attributes.
     */
    protected Icon getIcon() {
        Icon icon = null;
        String file = getString(ICONFILE);
        String desc = getString(ICONDESCRIPTION);
        if (file != null && !file.equals("")) {
            icon = new ImageIcon(file, desc);
        }
        return icon;
    }

    /**
     * Returns the value of the ID attribute for this element.
     * @return String  The value of the ID attribute.
     */
    protected String getId() {
        return getString(ID);
    }

    /**
     * Returns the value of the CASEINDEPENDENT attribute for this element.
     * @return boolean  The value of the CASEINDEPENDENT attribute.
     */
    protected boolean getCaseIndependent() {
        return getBoolean(CASEINDEPENDENT);
    }
}