package junit.extensions.jfcunit.finder;

import java.awt.Component;

import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

/**
 * Abstract class for defining call back classes to test whether a {@link java.awt.Component}
 * that is being searched for has been found.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public abstract class Finder {
    /** This member is used to evaluate regular expressions */
    private RE patternMatcher;

    /**
     * This method should not be called inside the .equals() method of the
     * class (1st parameter) that is being checked - otherwise, you will have
     * a cyclical reference.
     *
     * @param objClass The class of the objects being compared
     * @param obj1     First object for comparison
     * @param obj2     Second object for comparison
     * @return boolean whether the two objects have the same values
     *         (ie, the .equals() method returns true)
     * @deprecated   Since we are now using pattern matching, use <code>matchPattern()</code>
     */
    protected boolean compareValues(Class objClass, Object obj1, Object obj2) {
        // if the type is different - return false
        if (objClass.isInstance(obj1) != objClass.isInstance(obj2)) {
            return false;
        }

        if ((obj1 != null) && (obj2 != null)) {
            // both are not null - call the .equals() method
            return obj1.equals(obj2);
        } else {
            // if one is null (or) both are null
            return (obj1 == obj2);
        }
    }

    /**
     * This method is used to filter components' attributes based on a pattern specified by the user.
     * For example, to search for all windows with title matching 'testWindow*'.
     * Note: If the patternString is null, we need to avoid the <code>PatternCompiler.compile()</code>
     * throwing a NullPointerException and in this case, return true if the componentAttribute
     * is also null.
     * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
     *
     * @param patternString      The pattern to match with
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    protected void createPatternMatcher(String patternString, boolean caseIndependent) {
        if (patternMatcher == null) {
            try {
                if (patternString != null) {
                    patternMatcher = new RE(patternString);
                } else {
                    patternMatcher = new RE("");
                }
                if (caseIndependent) {
                    patternMatcher.setMatchFlags(RE.MATCH_CASEINDEPENDENT);
                }
            } catch (RESyntaxException e) {
                patternMatcher = null;
            }
        }
    }

    /**
     * This method is used to filter components' attributes based on a pattern specified by the user.
     * For example, to search for all windows with title matching 'testWindow*'.
     * Note: If the patternString is null, we need to avoid the <code>PatternCompiler.compile()</code>
     * throwing a NullPointerException and in this case, return true if the componentAttribute
     * is also null.
     * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
     *
     * @param componentAttribute    The attribute text of the component to match against
     * @param patternString         The pattern to match with
     * @param caseIndependent       Whether the match should be case independent (true) or not (false)
     * @return boolean whether the pattern is contained within the components' attribute text
     * @deprecated use {@link #createPatternMatcher(String, boolean)} and {@link #matchPattern
     * (String, String)}
     */
    protected boolean matchPattern(String componentAttribute, String patternString, boolean caseIndependent) {
        createPatternMatcher(patternString, caseIndependent);
        return matchPattern(componentAttribute, patternString);
    }

    /**
     * This method is used to filter components' attributes based on a pattern specified by the user.
     * For example, to search for all windows with title matching 'testWindow*'.
     * Note: If the patternString is null, we need to avoid the <code>PatternCompiler.compile()</code>
     * throwing a NullPointerException and in this case, return true if the componentAttribute
     * is also null.
     * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
     *
     * @param componentAttribute    The attribute text of the component to match against
     * @param patternString         The pattern to match with
     * @return boolean whether the pattern is contained within the components' attribute text
     */
    protected boolean matchPattern(String componentAttribute, String patternString) {
        return matchPattern(componentAttribute, patternString, patternMatcher);
    }

    /**
     * This method is used to filter components' attributes based on a pattern specified by the user.
     * For example, to search for all windows with title matching 'testWindow*'.
     * Note: If the patternString is null, we need to avoid the <code>PatternCompiler.compile()</code>
     * throwing a NullPointerException and in this case, return true if the componentAttribute
     * is also null.
     * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
     *
     * @param componentAttribute    The attribute text of the component to match against
     * @param patternString         The pattern to match with
     * @param re                    The RE pattern matcher to be used
     * @return boolean whether the pattern is contained within the components' attribute text
     */
    protected boolean matchPattern(String componentAttribute, String patternString, RE re) {
        // if the patternString is null, we need to avoid a NullPointerException
        // but at the same time, return true if the componentAttribute is also null
        if (patternString == null) {
            return (componentAttribute == null);
        }

        return (componentAttribute != null && re != null && re.match(componentAttribute));
    }

    /**
     * This method is used to check that the window object is an instance
     * of the specified class and is also visible.
     *
     * @param comp   The {@link Component} to be checked
     * @param cls    The type of the component
     * @return true if the component is of the specified type and is visible.
     */
    protected final boolean isValidForProcessing(Component comp, Class cls) {
        return (comp != null && cls != null && cls.isInstance(comp) && comp.isShowing());
    }

    /**
     * Method that returns true if the given component matches the search
     * criteria
     *
     * @param comp   The {@link Component} to test
     * @return true if this {@link Component} is a match
     */
    public abstract boolean testComponent(Component comp);
}