package junit.extensions.jfcunit.finder;

import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.util.Arrays;

import javax.swing.Icon;

/**
 * Class for comparing that two {@link javax.swing.Icon}s match.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson</a>
 */
public class IconMatcher {
    /**
     * The icon of the first component.
     */
    private Icon icon = null;

    /**
     * iconData read from the icon given.
     */
    private int[] iconData = null;

    /**
     * width of the icon.
     */
    private int icw = 0;

    /**
     * Height of the icon.
     */
    private int ich = 0;

    /**
     * Increment for performing spot checks across the icons
     * to speed performance of not equals comparisons.
     */
    private int incr = 1;

    /**
     * Size of the icon.
     */
    private int size = 0;

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param icon   The desired pattern for the icon of the component.
     * @throws InterruptedException when the PixelGrabber.grabPixels()
     * is interrupted.
     */
    public IconMatcher(Icon icon) throws InterruptedException {
        this.icon = icon;

        if (icon != null) {
            iconData = iconToArray(icon);
            icw = icon.getIconWidth();
            ich = icon.getIconHeight();
            size = icw * ich;
            int min = Math.min(icw, ich);
            if (min > 32) {
                incr = min / 8;
            } else {
                incr = min / 4;
            }
        }
    }

    /**
     * Paint the icon into an array of int.
     * @param icon Icon to be painted.
     * @return int[] containing the painted icon.
     * @throws InterruptedException when
     * PixelGrabber.grabPixels() is interrupted.
     */
    private int[] iconToArray(Icon icon) throws InterruptedException {
        int width = icon.getIconWidth();
        int height = icon.getIconHeight();

        int[] array = new int[width * height];
        //BufferedImage bf = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        BufferedImage bf = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB);

        icon.paintIcon(null, bf.getGraphics(), 0, 0);

        PixelGrabber grab = new PixelGrabber(bf, 0, 0, width, height, array, 0, width);
        grab.grabPixels();
        return array;
    }

    /**
     * Method that returns true if the given icon matches the one in this matcher.
     *
     * @param other   The other Icon to be compared
     * @return true if this Icon is a match or the icon which this
     *              matcher is created for is null.
     */
    public boolean matches(Icon other) {
        // If the icon on this finder is null, then we treat the
        // finder as a don't care or always match.
        if (icon == null) {
            return true;
        }
        if (other == null) {
            return false;
        }

        if (icw != other.getIconWidth() || ich != other.getIconHeight()) {
            return false;
        }

        int[] otherData = null;
        try {
            otherData = iconToArray(other);
        } catch (InterruptedException ie) {
        }

        if (otherData == null || iconData.length != otherData.length) {
            return false;
        }

        // Spot check
        for (int i = 0; i < size; i += incr) {
            if (iconData[i] != otherData[i]) {
                return false;
            }
        }

        // If we make it this far then lets compare the whole array.
        return Arrays.equals(iconData, otherData);
    }
}