package junit.extensions.jfcunit.finder;

import java.awt.Component;

import javax.swing.Icon;
import javax.swing.JMenuItem;

/**
 * A generic component finder which uses just the type (class) for filtering.
 * The pattern syntax can be found at the Jakarta RegExp API Documentation in {@link org.apache.regexp.RE}.
 * This class delegates the matching of the Icon to an instance of {@link junit.extensions.jfcunit.finder.IconMatcher}
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class JMenuItemFinder extends Finder {
    /**
     * The text of the component.
     */
    private String text = null;

    /**
     * The matcher for the icon of the {@link javax.swing.AbstractButton} component.
     */
    private IconMatcher iconMatcher;

    /**
     * A boolean specifying whether the filtration is case insensitive.
     */
    private boolean caseIndependent = false;

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str    The desired pattern for the text of the component.
     */
    public JMenuItemFinder(String str) {
        this(str, null, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str                The desired pattern for the text of the component.
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    public JMenuItemFinder(String str, boolean caseIndependent) {
        this(str, null, caseIndependent);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param icon   The desired pattern for the icon of the component.
     */
    public JMenuItemFinder(Icon icon) {
        this(null, icon, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str    The desired pattern for the text of the component.
     * @param icon   The desired pattern for the icon of the component.
     */
    public JMenuItemFinder(String str, Icon icon) {
        this(str, icon, false);
    }

    /**
     * Constructor accepting all arguments needed to filter the component.
     *
     * @param str                The desired pattern for the text of the component.
     * @param icon               The desired pattern for the icon of the component.
     * @param caseIndependent    Whether the match should be case independent (true) or not (false)
     */
    public JMenuItemFinder(String str, Icon icon, boolean caseIndependent) {
        this.text = str;
        try {
            this.iconMatcher = new IconMatcher(icon);
        } catch (InterruptedException ie) {
            this.iconMatcher = null;
        }

        this.caseIndependent = caseIndependent;
        createPatternMatcher(text, caseIndependent);
    }

    /**
     * Method that returns true if the given component matches the search
     * criteria
     *
     * @param comp   The component to test
     * @return true if this component is a match
     */
    public boolean testComponent(Component comp) {
        // since menuItem might not be visible, we cannot use isValidForProcessing()
        return (
            comp != null
                && comp instanceof JMenuItem
                && (text == null || matchPattern(((JMenuItem) comp).getText(), text))
                && iconMatcher.matches(((JMenuItem) comp).getIcon()));
    }
}