package junit.extensions.jfcunit.keyboard;

/**
 * The JFCKeyStroke class holds all of the data pertaining to
 * a keystroke to be pressed/released by the TestHelper.
 *
 * @author Kevin Wilson
 */
public class JFCKeyStroke {
    /**
     * Key Character
     */
    private char keyChar;

    /**
     * Key Code.
     */
    private int keyCode;

    /**
     * Modifiers to be pressed for this event.
     */
    private int modifiers;

    /**
     * isTyped true if the key stroke is to emit a typed event.
     */
    private boolean isTyped;

    /**
     * Constructor
     *
     * @param c Character
     * @param code Key code.
     * @param modifiers Modifiers to be pressed when sending the keystroke.
     * @param typed true if a typed event should be emitted.
     */
    public JFCKeyStroke(char c, int code, int modifiers, boolean typed) {
        setKeyChar(c);
        setKeyCode(code);
        setModifiers(modifiers);
        setTyped(typed);
    }

    /**
     * Copy Constructor
     *
     * @param stroke Keystroke to be copied.
     */
    public JFCKeyStroke(JFCKeyStroke stroke) {
        setKeyChar(stroke.getKeyChar());
        setKeyCode(stroke.getKeyCode());
        setModifiers(stroke.getModifiers());
        setTyped(stroke.isTyped());
    }

    /**
     * Set the key character.
     *
     * @param c Character.
     */
    private final void setKeyChar(char c) {
        keyChar = c;
    }

    /**
     * Set the key code.
     *
     * @param code Key code
     */
    private final void setKeyCode(int code) {
        keyCode = code;
    }

    /**
     * Set the modifiers.
     *
     * @param modifiers Modifiers to be pressed for this keystroke.
     */
    public final void setModifiers(int modifiers) {
        this.modifiers = modifiers;
    }

    /**
     * Set the typed state.
     *
     * @param typed true if the keystroke is to emit a typed event.
     */
    private final void setTyped(boolean typed) {
        isTyped = typed;
    }

    /**
     * Get the key character.
     *
     * @return char key character.
     */
    public final char getKeyChar() {
        return keyChar;
    }

    /**
     * Get the key code.
     *
     * @return int key code
     */
    public final int getKeyCode() {
        return keyCode;
    }

    /**
     * Get the modifiers.
     *
     * @return int modifiers.
     */
    public final int getModifiers() {
        return modifiers;
    }

    /**
     * Is the keystroke to emit a typed event.
     *
     * @return true if the typed event is to be emitted.
     */
    public final boolean isTyped() {
        return isTyped;
    }

    /**
     * Create a string representation of the keystroke.
     *
     * @return String representation of the keystroke.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(100);
        buf.append(" Char:" + getKeyChar());
        buf.append(" Code:" + getKeyCode());
        buf.append(" Mods:" + getModifiers());
        buf.append(" IsTyped:" + isTyped());
        return buf.toString();
    }

    /**
     * Test for equality.
     *
     * @param o Object to be tested for equality.
     * @return true if the object is equal to this object.
     */
    public boolean equals(Object o) {
        if (o instanceof JFCKeyStroke) {
            JFCKeyStroke stroke = (JFCKeyStroke) o;
            return (
                (getKeyChar() == stroke.getKeyChar()) && (getKeyCode() == stroke.getKeyCode()) && (getModifiers() == stroke.getModifiers()) && (isTyped == stroke.isTyped()));
        }
        return false;
    }
}