package junit.extensions.xml;

import java.lang.reflect.Constructor;
import java.util.ResourceBundle;

import org.w3c.dom.Element;

import junit.extensions.xml.elements.AbstractTagHandler;

/**
 * This class is used to read in the properties and provide access to the tag handlers for various
 * elements based on their attribute values.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public class XMLTagResourceBundle {
    /**
     * Handle to the singleton instance.
     */
    private static XMLTagResourceBundle singleton;

    /**
     * The ResourceBundle used to read in the properties.
     */
    private ResourceBundle bundle = null;

    /**
     * A convenience method to get a handle to the AbstractTagHandler implementation.
     * @param element     The element for which we need the tag handler.
     * @param testCase    The testCase being run.
     * @param type        The type of element to be handled.
     * @return AbstractTagHandler    An implementation of the AbstractTagHandler which will be used
     * to process the element.
     * @throws XMLException if the tag handler cannot be instanciated.
     */
    public static AbstractTagHandler getTagHandler(Element element, IXMLTestCase testCase, String type) throws XMLException {
        return getSingleton().getTagHandlerImpl(element, testCase, type);
    }

    /**
     * A convenience method to get a handle to the AbstractTagHandler implementation.
     * @param element     The element for which we need the tag handler.
     * @param testSuite   The IXMLTestSuite being run.
     * @param type        The type of element to be handled.
     * @return AbstractTagHandler    An implementation of the AbstractTagHandler which will be used
     * to process the element.
     * @throws XMLException if the tag handle cannot be instanciated.
     */
    public static AbstractTagHandler getTagHandler(Element element, IXMLTestSuite testSuite, String type) throws XMLException {
        return getSingleton().getTagHandlerImpl(element, testSuite, type);
    }
    /**
     * The "default" method that is used to create and initialize a singleton instance.
     * @return XMLTagResourceBundle    The one and only instance of this class.
     */
    private static XMLTagResourceBundle getSingleton() {
        if (singleton == null) {
            singleton = new XMLTagResourceBundle();
        }
        return singleton;
    }

    /**
     * A default private constructor.
     * @see java.lang.Object#Object()
     */
    private XMLTagResourceBundle() {
        bundle = ResourceBundle.getBundle("TagMapping");
    }

    /**
     * Reads the short name (type) and creates an instance of the tag handler that corresponds to
     * the value from the resource bundle.
     * @param element     The element for which we need the tag handler.
     * @param testCase    The testCase being run.
     * @param type        The type of element to be handled.
     * @return AbstractTagHandler    An implementation of the AbstractTagHandler which will be used
     * to process the element.
     * @throws XMLException if the tag handler cannot be instanciated.
     */
    private AbstractTagHandler getTagHandlerImpl(Element element, IXMLTestCase testCase, String type) throws XMLException {
        String clazzName = bundle.getString(type);
        try {
            Class clazz = Class.forName(clazzName);
            Constructor cnstr = clazz.getConstructor(new Class[] { Element.class, IXMLTestCase.class });
            return (AbstractTagHandler) cnstr.newInstance(new Object[] { element, testCase });
        } catch (Exception e) {
            throw new XMLException("Could not create class:" + clazzName + "\n" + e);
        }
    }


    /**
     * Reads the short name (type) and creates an instance of the tag handler that corresponds to
     * the value from the resource bundle.
     * @param element     The element for which we need the tag handler.
     * @param testSuite   The testSuite being run.
     * @param type        The type of element to be handled.
     * @return AbstractTagHandler    An implementation of the AbstractTagHandler which will be used
     * to process the element.
     * @throws XMLException if the tag handler cannot be instanciated.
     */
    private AbstractTagHandler getTagHandlerImpl(Element element, IXMLTestSuite testSuite, String type) throws XMLException {
        String clazzName = bundle.getString(type);
        try {
            Class clazz = Class.forName(clazzName);
            Constructor cnstr = clazz.getConstructor(new Class[] { Element.class, IXMLTestSuite.class });
            return (AbstractTagHandler) cnstr.newInstance(new Object[] { element, testSuite });
        } catch (Exception e) {
            e.printStackTrace();
            throw new XMLException("Could not create class:" + clazzName);
        }
    }
}
