package junit.extensions.xml;

import java.util.HashMap;
import java.util.Map;

import org.w3c.dom.Element;

import junit.framework.TestCase;
import junit.framework.Test;

/**
 * Test Case for running XML Script based testing.
 *
 * @author Kevin Wilson
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class XMLTestCase extends TestCase implements IXMLTestCase, XMLConstants {
    /**
     * The name of the test xml script file - found from the classpath.
     */
    private String filename;

    /**
     * The element to be processed.
     */
    private Element element;

    /**
     * A Map of all objects that have been found - keyed by the string name.
     */
    private Map foundObjects;

    /**
     * The default constructor that is needed to createa a test case.
     *
     * @param filename    The name of the test xml script file (found from the classpath)
     * @param element     The Element to be processed
     * @throws Exception  An instance of java.lang.Exception can be thrown
     */
    public XMLTestCase(String filename, Element element) throws Exception {
        super(filename + ":" + XMLUtil.getPath(element));
        this.filename = filename;
        this.element = element;
    }

    /**
     * Sets up, executes and then tears down a test
     *
     * @exception Throwable exceptions thrown by code.
     */
    public void runBare() throws Throwable {
        setUp();
        try {
            runTest();
        } finally {
            tearDown();
        }
    }

    /**
     * Executes a test
     *
     * @exception Throwable exceptions thrown by code.
     */
    protected void runTest() throws Throwable {
        runXMLTest();
    }

    /**
     * Default setUp which does nothing. Override this to set up the environment
     * for your test
     *
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see junit.framework.TestCase#setUp()
     */
    public void setUp() throws Exception {
        super.setUp();
    }

    /**
     * Default tearDown which does nothing. Override this to tear down the
     * environment for your test.
     *
     * @exception  Exception   An instance of java.lang.Exception can be thrown
     * @see junit.framework.TestCase#setUp()
     */
    public void tearDown() throws Exception {
        super.tearDown();
    }

    /**
     * This method is the one that actually performs the test by processing the elements.
     *
     * @exception  Exception   An instance of java.lang.Exception can be thrown.
     */
    public final void runXMLTest() throws Exception {
        if (foundObjects == null) {
            foundObjects = new HashMap();
        } else {
            clearFoundObjects();
        }
        XMLTagResourceBundle.getTagHandler(element, this, TEST).processElement();
    }

    /**
     * Returns an empty suite.
     *
     * @return Test   An empty XMLTestSuite is returned by default
     * @exception  Exception   An instance of java.lang.Exception can be thrown.
     */
    public static Test suite() throws Exception {
        return new XMLTestSuite();
    }

    /**
     * For each test (method) element, the found objects has to be cleared
     */
    public void clearFoundObjects() {
        foundObjects.clear();
    }

    /**
     * Add each found object into the cache map.
     *
     * @param name    The name of the found object
     * @param obj     The actual object that was found
     */
    public void addFoundObject(String name, Object obj) {
        foundObjects.put(name, obj);
    }

    /**
     * Retrieve the object that was found previously.
     *
     * @param name    The name of the object that was found
     * @return    The object to be retrieved
     */
    public Object getFoundObject(String name) {
        return foundObjects.get(name);
    }
}
