package junit.extensions.xml;

import java.io.IOException;
import java.io.InputStream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * Interface layer to w3c dom.
 *
 * @author Kevin Wilson
 */
public class XMLUtil implements XMLConstants {
    /**
     * Private constructor: so that this class is not instantiated.
     */
    private XMLUtil() {
    }

    /**
     * This method is used to parse the xml file with the specified name.
     *
     * @param stream      The input stream of the xml file to be parsed.
     * @return Document   The root Document object.
     * @throws Exception  Exception may be throws when parsing the document.
     */
    public static Document parse(InputStream stream) throws Exception {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        //factory.setValidating(true);
        //factory.setNamespaceAware(true);
        try {
            return factory.newDocumentBuilder().parse(stream);
        } catch (SAXException sxe) {
            // Error generated during parsing
            Exception x = sxe;
            if (sxe.getException() != null) {
                x = sxe.getException();
            }
            x.printStackTrace();
            throw x;
        } catch (ParserConfigurationException pce) {
            // Parser with specified options can't be built
            pce.printStackTrace();
            throw pce;
        } catch (IOException ioe) {
            // I/O error
            ioe.printStackTrace();
            throw ioe;
        } finally {
            if (stream != null) {
                stream.close();
            }
        }
    }

    /**
     * A utility method to build the path of the Node (with '.' are the separators)
     *
     * @param node    The Node whose path is to be found.
     * @return String    The String representation of the path.
     */
    public static String getPath(Node node) {
        StringBuffer buf = new StringBuffer(500);
        if (node.getParentNode() != null && node.getParentNode() instanceof Element) {
            buf.append(getPath(node.getParentNode()));
            buf.append(".");
        }
        buf.append(getName(node));
        return buf.toString();
    }

    /**
     * The name of the Node. First, checks the attribute value, and if that is null, uses the
     * <code>getNodeName()</code> method on the Node interface.
     *
     * @param node    The Node whose name is to be found.
     * @return String    The name of the node (priority is given to the attribute called 'name')
     */
    public static String getName(Node node) {
        String name = null;
        if (node instanceof Element) {
            name = getAttribute(((Element) node), NAME);
        }
        return (name == null ? node.getNodeName() : name);
    }

    /**
     * A utility method that will get the attribute value, and if not found, will return null.
     *
     * @param element    The element to be searched
     * @param name       The name of the attribute to be searched for
     * @return String    The string value of the attribute, null if the value is not set or is an empty string
     */
    public static String getAttribute(Element element, String name) {
        String value = element.getAttribute(name);
        return ((value != null && value.trim().length() == 0) ? null : value);
    }

    /**
     * A utility method that will get the attribute value, and if not found, will return false.
     *
     * @param element    The element to be searched
     * @param name       The name of the attribute to be searched for
     * @return boolean   The boolean value of the attribute, false if the value is not set or is an
     * empty string
     */
    public static boolean getBooleanAttributeValue(Element element, String name) {
        String value = element.getAttribute(name);
        return new Boolean(value).booleanValue();
    }

    /**
     * A utility method to find a file from the classpath and read its contents.
     *
     * @param fileName     The name of the file to be read.
     * @return InputStream    The file contents as an InputStream.
     */
    public static InputStream readFileFromClasspath(String fileName) {
        return Thread.currentThread().getContextClassLoader().getResourceAsStream(fileName);
    }

    /**
     * A utility method to find a file from the context of the specified class and read its
     * contents. The difference between this method and the {@link #readFileFromClasspath(String)}
     * method is that this one uses the classpath, and then appends the package information of the
     * specified class and then looks for the specified file.
     *
     * @param clz             Class whos resource loader should be used.
     * @param fileName        The name of the file to be read.
     * @return InputStream    The file contents as an InputStream.
     */
    public static InputStream readFileFromClassContext(Class clz, String fileName) {
        return clz.getResourceAsStream(fileName);
    }
}