package junit.extensions.xml.elements;

import java.awt.Point;
import java.text.MessageFormat;
import java.text.ParseException;

import org.w3c.dom.Element;

import junit.extensions.xml.XMLException;
import junit.extensions.xml.IXMLTestCase;
import junit.extensions.xml.XMLUtil;

/**
 * Sub-classes of this class will handle the processing of elements.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 * @author Kevin Wilson
 */
public abstract class AbstractTagHandler {
    /**
     * The Element to be processed.
     */
    private Element element;

    /**
     * The IXMLTestCase that uses this element.
     */
    private IXMLTestCase testCase;

    /**
     * Constructor for AbstractTagHandler.
     *
     * @param element    The element to be processed
     * @param testCase   The testCase to be associated with the test.
     */
    public AbstractTagHandler(Element element, IXMLTestCase testCase) {
        this.element = element;
        this.testCase = testCase;
    }

    /**
     * Returns the element.
     *
     * @return Element
     */
    public Element getElement() {
        return element;
    }

    /**
     * Returns the testCase.
     *
     * @return IXMLTestCase
     */
    public IXMLTestCase getTestCase() {
        return testCase;
    }

    /**
     * Sub-classes should implement this method to provide validation
     * of the element attributes, etc.
     *
     * @throws XMLException    A validation exception is thrown.
     */
    protected void validateElement() throws XMLException {
        if (element == null) {
            throw new XMLException("Element could not be found.");
        }
    }

    /**
     * This method is used to process the xml Element (i.e. parsing and setting the attributes on
     * XYZ objects from those parsed values.
     *
     * @throws Exception Some kind of Exception might be thrown
     */
    public abstract void processElement() throws Exception;

    /**
     * Since all sub-classes will check for the element tag name,
     * this method can be used for consistent messages.
     *
     * @param expectedName    The non-null expected string name
     * @throws XMLException    A validation exception is thrown.
     */
    protected void checkElementTagName(String expectedName) throws XMLException {
        if (!expectedName.equals(element.getTagName())) {
            throw new XMLException("Cannot process " + getElement().getTagName() + " node, expected '" + expectedName + "'");
        }
    }

    /**
     * Since all sub-classes might check for a required attribute, this method can be used for consistent messages.
     *
     * @param attrName    The name of the attribute being checked
     * @throws XMLException    A validation exception is thrown.
     */
    protected void checkRequiredAttribute(String attrName) throws XMLException {
        String attrValue = XMLUtil.getAttribute(element, attrName);
        if (attrValue == null) {
            throw new XMLException("Missing the required attribute: " + attrName + " in element: " + element.getNodeName());
        }
    }

    /**
     * Since all sub-classes might check for a required attribute, this method can be used for consistent messages.
     *
     * @param root    The root Element whose attributes are being checked
     * @param attrName    The names of the attributes being checked
     * @throws XMLException    A validation exception is thrown.
     */
    protected void checkAtLeastOneRequiredAttribute(Element root, String[] attrName) throws XMLException {
        int found = 0;
        for (int i = 0; i < attrName.length; i++) {
            String attrValue = XMLUtil.getAttribute(root, attrName[i]);
            if (attrValue != null) {
                found++;
            }
        }
        if (found == 0) {
            String buf = "";
            for (int i = 0; i < attrName.length; i++) {
                buf = MessageFormat.format("{0} {1}", new Object[] { buf, attrName[i] });
            }
            throw new XMLException(root.getNodeName() + " is missing one of the required attributes: " + buf);
        }
    }

    /**
     * Since all sub-classes might check for a required attribute, this method can be used for consistent messages.
     *
     * @param root    The root Element whose attributes are being checked
     * @param attrName    The names of the attributes being checked
     * @throws XMLException    A validation exception is thrown.
     */
    protected void checkOneRequiredAttribute(Element root, String[] attrName) throws XMLException {
        int found = 0;
        if (attrName != null) {
            for (int i = 0; i < attrName.length; i++) {
                if (XMLUtil.getAttribute(root, attrName[i]) != null) {
                    found++;
                }
            }
        }
        if (found == 0) {
            String buf = "";
            for (int i = 0; i < attrName.length; i++) {
                buf = MessageFormat.format("{0} {1}", new Object[] { buf, attrName[i] });
            }
            throw new XMLException(root.getNodeName() + " is missing one of the required attributes: " + buf);
        }
        if (found > 1) {
            String buf = "";
            for (int i = 0; i < attrName.length; i++) {
                buf = MessageFormat.format("{0} {1}", new Object[] { buf, attrName[i] });
            }
            throw new XMLException("Too many attributes of:" + buf);
        }
    }

    /**
     * This method checks whether the current element contains an attribute
     * with the specified name.
     * @param attributeName    The name of the attribute being checked.
     * @return boolean    Whether this element contains this attribute.
     */
    private boolean isAvailable(String attributeName) {
        return getString(attributeName) != null;
    }

    /**
     * This method checks whether the specified attribute is present, and if so, returns the value
     * parsed as a primitive long, if not, then returns the default value specified.
     * @param attributeName    The name of the attribute whose value is needed.
     * @param defaultValue     The default value to use if such an attribute is not present.
     * @return long    The value of the attribute.
     * @throws NumberFormatException if the parse did not work.
     */
    protected long getLong(String attributeName, long defaultValue) throws NumberFormatException {
        return isAvailable(attributeName) ? Long.parseLong(getString(attributeName)) : defaultValue;
    }

    /**
     * This method checks whether the specified attribute is present, and if so, returns the value
     * parsed as a primitive int, if not, then returns the default value specified.
     * @param attributeName    The name of the attribute whose value is needed.
     * @param defaultValue     The default value to use if such an attribute is not present.
     * @return int    The value of the attribute.
     * @throws NumberFormatException if the parse did not work.
     */
    protected int getInt(String attributeName, int defaultValue) throws NumberFormatException {
        return isAvailable(attributeName) ? Integer.parseInt(getString(attributeName)) : defaultValue;
    }

    /**
     * This method checks whether the specified attribute is present, and if so, returns the value
     * parsed as a primitive boolean, if not, then returns false.
     * @param attributeName    The name of the attribute whose value is needed.
     * @return boolean    The value of the attribute (false if not found).
     */
    protected boolean getBoolean(String attributeName) {
        // this will default to false
        Boolean b = new Boolean(getString(attributeName));
        return b.booleanValue();
    }

    /**
     * This method checks whether the specified attribute is present, and if so, returns the value
     * parsed as a {@link Point}, if not, then returns the default value specified.
     * @param attributeName    The name of the attribute whose value is needed.
     * @param defaultValue     The default value to use if such an attribute is not present.
     * @return Point    The value of the attribute.
     * @throws ParseException if the parse did not work.
     */
    protected Point getPoint(String attributeName, Point defaultValue) throws ParseException {
        if (!isAvailable(attributeName)) {
            return defaultValue;
        }

        String s = getString(attributeName);
        // somehow, using this does not work!!!
        //MessageFormat mf = new MessageFormat("{0,number},{1,number}");
        MessageFormat mf = new MessageFormat("{0},{1}");
        Object[] obs = mf.parse(s);
        int x = Integer.parseInt((String) obs[0]);
        int y = Integer.parseInt((String) obs[1]);
        return new Point(x, y);
    }

    /**
     * This method returns the value of the specified attribute.
     * @param attributeName    The name of the attribute whose value is needed.
     * @return String    The value of the attribute.
     */
    protected String getString(String attributeName) {
        return XMLUtil.getAttribute(getElement(), attributeName);
    }
}
