package junit.extensions.xml.elements;

import org.w3c.dom.Element;

import junit.extensions.xml.XMLException;
import junit.extensions.xml.IXMLTestCase;
import junit.extensions.xml.XMLConstants;
import junit.framework.Assert;

/**
 * This class will handle the processing of &lt;assertnotnull&gt; nodes.
 *
 * @author <a href="mailto:vraravam@thoughtworks.com">Vijay Aravamudhan : ThoughtWorks Inc.</a>
 */
public class AssertEqualsTagHandler extends AbstractTagHandler implements XMLConstants {
    /**
     * Constructor for AssertEqualsTagHandler.
     *
     * @param element     The element to be processed
     * @param testCase    The IXMLTestCase that uses this element
     */
    public AssertEqualsTagHandler(Element element, IXMLTestCase testCase) {
        super(element, testCase);
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#validateElement()
     */
    public void validateElement() throws XMLException {
        // do the default validations from the super class
        super.validateElement();

        // check the element tag name
        checkElementTagName(ASSERTEQUALS);

        // reqd attribute: at least one of expected or expectedrefid
        checkAtLeastOneRequiredAttribute(getElement(), new String[] { EXPECTEDOBJ, EXPECTEDREFID });

        // reqd attribute: at least one of actual or actualrefid
        checkAtLeastOneRequiredAttribute(getElement(), new String[] { ACTUALOBJ, ACTUALREFID });
        // TODO: Check for one attribute being dependent on another attribute
    }

    /**
     * @see junit.extensions.xml.elements.AbstractTagHandler#processElement()
     */
    public void processElement() throws Exception {
        validateElement();

        Assert.assertEquals(getMessage(), getExpectedObject(), getActualObject());
    }

    /**
     * Returns the value of the MESSAGE attribute for this element.
     * @return String  The value of the MESSAGE attribute.
     */
    protected String getMessage() {
        return getString(MESSAGE);
    }

    /**
     * Returns the value of the EXPECTEDREFID attribute for this element.
     * @return String  The value of the EXPECTEDREFID attribute.
     */
    private String getExpectedRefId() {
        return getString(EXPECTEDREFID);
    }

    /**
     * Returns the value of the ACTUALREFID attribute for this element.
     * @return String  The value of the ACTUALREFID attribute.
     */
    private String getActualRefId() {
        return getString(ACTUALREFID);
    }

    /**
     * Returns the value of the EXPECTEDOBJ attribute for this element.
     * If the element contains an expected ref id attribute and that object
     * was found previously, then that object is returned.
     * @return String  The value of the EXPECTEDOBJ attribute. (If an expected ref id
     *                 object was found and was not null, then that value is returned.
     */
    private Object getExpectedObject() {
        if (getExpectedRefId() != null && getTestCase().getFoundObject(getExpectedRefId()) != null) {
            return getTestCase().getFoundObject(getExpectedRefId());
        } else {
            return getString(EXPECTEDOBJ);
        }
    }

    /**
     * Returns the value of the ACTUALOBJ attribute for this element.
     * If the element contains an actual ref id attribute and that object
     * was found previously, then that object is returned.
     * @return String  The value of the ACTUALOBJ attribute. (If an actual ref id
     *                 object was found and was not null, then that value is returned.
     */
    private Object getActualObject() {
        if (getActualRefId() != null && getTestCase().getFoundObject(getActualRefId()) != null) {
            return getTestCase().getFoundObject(getActualRefId());
        } else {
            return getString(ACTUALOBJ);
        }
    }
}