/**
 * Title:        MoneyBag
 * Description:  A MoneyBag defers exchange rate conversions
 * Copyright:    Copyright (c) 2000 Kent Beck and Erich Gamma
 *
 *   Permission to reproduce and create derivative works from the Software 
 *   ("Software Derivative Works") is hereby granted to you under the copyrights
 *   of Kent Beck and Erich Gamma. Kent Beck and Erich Gamma also grant you 
 *   the right to distribute the Software and Software Derivative Works.
 *
 *   Kent Beck and Erich Gamma licenses the Software to you on an "AS IS" basis,
 *   without warranty of any kind.  Kent Beck and Erich Gamma HEREBY
 *   EXPRESSLY DISCLAIMS ALL WARRANTIES OR CONDITIONS, EITHER EXPRESS OR
 *   IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OR 
 *   CONDITIONS OF MERCHANTABILITY, NON INFRINGEMENT AND FITNESS FOR A 
 *   PARTICULAR PURPOSE. You are solely responsible for determining the 
 *   appropriateness of using the Software and assume all risks associated with
 *   the use and distribution of this Software, including but not limited to 
 *   the risks of program errors, damage to or loss of data, programs or 
 *   equipment, and unavailability or interruption of operations. KENT BECK AND
 *   ERICH GAMMA WILL NOT BE LIABLE FOR ANY DIRECT DAMAGES OR FOR ANY SPECIAL, 
 *   INCIDENTAL, OR INDIRECT DAMAGES OR FOR ANY ECONOMIC CONSEQUENTIAL DAMAGES 
 *   (INCLUDING LOST PROFITS OR SAVINGS), EVEN IF KENT BECK AND ERICH GAMMA HAD
 *   BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. Kent Beck and Erich Gamma 
 *   will not be liable for the loss of, or damage to, your records or data, or
 *   any damages claimed by you based on a third party claim.
 *
 *   You agree to distribute the Software and any Software Derivatives under a
 *   license agreement that: 1) is sufficient to notify all licensees of the
 *   Software and Software Derivatives that Kent Beck and Erich Gamma assumes
 *   no liability for any claim that may arise regarding the Software or
 *   Software Derivatives, and 2) that disclaims all warranties, both express
 *   and implied, from Kent Beck and Erich Gamma regarding the Software and
 *   Software Derivatives. (If you include this Agreement with any distribution
 *   of the Software and Software Derivatives you will have met this
 *   requirement). You agree that you will not delete any copyright notices in
 *   the Software.
 *
 *   This Agreement is the exclusive statement of your rights in the Software
 *   as provided by Kent Beck and Erich Gamma. Except for the licenses
 *   granted to you in the second paragraph above, no other licenses are granted
 *   hereunder, by estoppel, implication or otherwise.
 *
 * @author Kent Beck
 * @author Erich Gamma
 *
 * $Id: MoneyBag.java,v 1.2 2001/12/14 11:27:27 aheilwag Exp $
 * 
 * $Log: MoneyBag.java,v $
 * Revision 1.2  2001/12/14 11:27:27  aheilwag
 * changes to the JUnit test cases moved to JUnitX and JUnitX now beeing a
 * module on top of JUnit
 *
 * Revision 1.1  2001/12/14 09:30:40  aheilwag
 * moved test cases from junit to junitx
 *
 * Revision 1.2  2000/11/10 19:57:33  aheilwag
 * added cvs tags
 *
 */


package tests.junitx.money.setup;


import java.util.*;


/**
 * A MoneyBag defers exchange rate conversions. For example adding 12 Swiss
 * Francs to 14 US Dollars is represented as a bag containing the two Monies
 * 12 CHF and 14 USD. Adding another 10 Swiss francs gives a bag with 22 CHF
 * and 14 USD. Due to the deferred exchange rate conversion we can later value
 * a MoneyBag with different exchange rates.
 *
 * A MoneyBag is represented as a list of Monies and provides different
 * constructors to create a MoneyBag. 
 */

public class MoneyBag implements IMoney
{
  private Vector _monies = new Vector (5);


  /**
   * hides the default constructor.
   */

  private MoneyBag ()
  {
  }


  /**
   * constructs a money bag.
   *
   * @param  aMoneyArray   an array of monies.
   */

  public MoneyBag (Money bag[])
  {
    for (int i = 0; i < bag.length; i++)
      appendMoney (bag[i]);
  }


  /**
   * constructs a money bag from two monies.
   *
   * @param   money1      first money
   * @param   money2      second money
   */

  public MoneyBag(Money money1, Money money2)
  {
    appendMoney (money1);
    appendMoney (money2);
  }


  /**
   * constructs a money bag from a money and a money bag
   *
   * @param   someMoney      some money
   * @param   aMoneyBag      a money bag
   */

  public MoneyBag (Money someMoney, MoneyBag aMoneyBag)
  {
    appendMoney (someMoney);
    appendBag   (aMoneyBag);
  }


  /**
   * constructs a money bag from two money bags.
   *
   * @param   moneyBag1      first money bag
   * @param   moneyBag2      second money bag
   */

  public MoneyBag (MoneyBag moneyBag1, MoneyBag moneyBag2)
  {
    appendBag (moneyBag1);
    appendBag (moneyBag2);
  }


  /**
   * adds a money to this money bag.
   *
   * @param   someMoney    some money
   *
   * @return  resulting money
   */

  public IMoney add (IMoney someMoney)
  {
    return someMoney.addMoneyBag (this);
  }


  /**
   * adds a money to this money bag.
   *
   * @param   someMoney    some money
   *
   * @return  resulting money bag
   */


  public IMoney addMoney (Money someMoney)
  {
    return (new MoneyBag (someMoney, this)).simplify ();
  }


  /**
   * adds a money bag to this money bag.
   *
   * @param   aMoneyBag    a money bag
   *
   * @return  resulting money bag
   */

  public IMoney addMoneyBag (MoneyBag aMoneyBag)
  {
    return (new MoneyBag (aMoneyBag, this)).simplify ();
  }


  /**
   * append a money bag to this money bag
   *
   * @param   aMoneyBag     a money bag
   */

  private void appendBag (MoneyBag aMoneyBag)
  {
    for (Enumeration e = aMoneyBag._monies.elements (); e.hasMoreElements (); )
      appendMoney ((Money)e.nextElement ());
  }


  /**
   * append some money to this money bag
   *
   * @param   someMoney     some money
   */

  private void appendMoney (Money someMoney)
  {
    IMoney old = findMoney (someMoney.currency ());

    if (old == null)
    {
      _monies.addElement (someMoney);
      return;
    }
    _monies.removeElement (old);

    IMoney sum = old.add (someMoney);
    if (sum.isNull ()) 
      return;

    _monies.addElement (sum);
  }


  /**
   * checks if the bag contains a certain money
   *
   * @param    someMoney    some money
   *
   * @return true if the bag contains a certain currency.
   */

  private boolean contains (Money someMoney)
  {
    return findMoney (someMoney.currency ()).amount () == someMoney.amount ();
  }


  /**
   * checks if two money bags are equal.
   *
   * @param   anObject    somebody implementing <code>IMoney</code>
   *
   * @return  true if equal
   */

  public boolean equals (Object anObject)
  {
    if (isNull ())
      if (anObject instanceof IMoney)
        return ((IMoney)anObject).isNull ();

    if (anObject instanceof MoneyBag)
    {
      MoneyBag aMoneyBag = (MoneyBag)anObject;
       
      if (aMoneyBag._monies.size() != _monies.size ())
        return false;

      for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
      {
        Money m = (Money)e.nextElement ();

        if (!aMoneyBag.contains (m))
          return false;
      }
      return true;
    }
    return false;
  }


  /**
   * checks if a certain currency is in the bag
   *
   * @param   aCurrency    a currency name
   *
   * @return  Money or null
   */

  private Money findMoney (String aCurrency)
  {
    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
    {
      Money m = (Money)e.nextElement ();
 
      if (m.currency ().equals (aCurrency))
        return m;
    }
    return null;
  }


  /**
   * @return hashcode for this money bey
   */

  public int hashCode ()
  {
    int hash = 0;

    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
      hash^= e.nextElement ().hashCode();

    return hash;
  }


  /**
   * @return true if the money bag has no value in it
   */

  public boolean isNull ()
  {
    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
    {
      Money m = (Money) e.nextElement();

      if (m.amount () != 0)
        return false;
    }
    return true;
  }


  /**
   * multiplies this money bag with a factor
   *
   * @param   aFactor    an integer factor
   *
   * @return  resulting money bag
   */

  public IMoney multiply (int aFactor)
  {
    MoneyBag result = new MoneyBag ();

    if (aFactor == 0)
      return result;

    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
    {
      Money m = (Money)e.nextElement ();

      result.appendMoney ((Money)m.multiply (aFactor));
    }

    return result;
  }


  /**
   * negate the current money bag
   *
   * @return the resulting money bag
   */

  public IMoney negate ()
  {
    MoneyBag result= new MoneyBag ();

    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
    {
      Money m = (Money) e.nextElement ();

      result.appendMoney ((Money)m.negate ());
    }
    return result;
  }


  /**
   * @return a money if the money bag contains only one currency, else
   *         returns the current money bag.
   */

  private IMoney simplify ()
  {
    if (_monies.size () == 1)
      return (IMoney)_monies.elements ().nextElement ();

    return this;
  }


  /**
   * subtract some money from this money bag
   *
   * @param    someMoney    some money
   *
   * @return   the resulting money bag
   */

  public IMoney subtract (IMoney someMoney)
  {
    return add (someMoney.negate ());
  }


  /**
   * @return a string representation of this money bag.
   */

  public String toString ()
  {
    StringBuffer buffer = new StringBuffer ("{");

    for (Enumeration e = _monies.elements (); e.hasMoreElements (); )
      buffer.append ((Money)e.nextElement ());

    buffer.append ("}");

    return buffer.toString ();
  }
}
